#ifndef BP2DICT_H
#define BP2DICT_H

class BppVirtualMachine;
class ValPtr;

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2Classes B++ Classes
@{
*/
#ifdef DOXY

/** Bidirectional iterator for Dictionary class.
This helper class allows sequentially walking through entries of a Dictionary object.
Instances of this class are created by Dictionary methods only. Each iterator object referres
to a specific Dictionary.
@see Dictionary, DictRIterator
*/
class DictIterator
{
   /** @publicsection */
   /** Increment operator function.
   Function walks to next entry of owning dictionary and returns iterator to it.
   If iterator points to end of dictionary before call, an error is raised.
   @return current iterator pointing to next entry (if any) or end
   @see next
   */
   DictIterator OP_INC();
   /// Prefix increment operator (same as OP_INC)
   DictIterator operator++();

   /** Post increment operator function.
   Function walks to next entry of owning dictionary and returns iterator to it.
   If iterator points to end of dictionary before call, an error is raised.
   @return new iterator pointing to next entry (if any) or end iterator
   */
   DictIterator OP_PINC();

   /// Postfix increment operator (same as OP_PINC)
   DictIterator operator++(var);

   /** Decrement operator function.
   Function walks to entry before current one in owning dictionary and returns iterator to it.
   If iterator points not to a valid dictionary entry before call, an error is raised.
   @return current iterator pointing to entry before (if any) or end
   @see before
   */
   DictIterator OP_DEC();
   /// Prefix decrement operator (same as OP_DEC)
   DictIterator operator--();

   /** Post decrement operator function.
   Function walks to to entry before current one in owning dictionary and returns iterator to it.
   If iterator points not to a valid dictionary entry before call, an error is raised.
   @return new iterator pointing to entry before (if any) or end iterator
   */
   DictIterator OP_PDEC();

   /// Postfix decrement operator (same as OP_PDEC)
   DictIterator operator--(var);


   /** Move iterator to next dictionary entry.
   Function walks to next entry of owning dictionary.<br>
   If iterator points to end of dictionary before call, an error is raised.
   @return current iterator pointing to next entry (if any) or end
   @note Function uses same implementation as increment operator. 
   @see OP_INC
   */
   DictIterator next();

   /** Move iterator to dictionary entry before.
   Function walks to entry before current one.<br>
   If iterator points not to a valid dictionary entry before before call, an error is raised.
   @return current iterator pointing to next entry (if any) or end
   @note Function uses same implementation as decrement operator.
   @see OP_DEC
   */
   DictIterator before();

   /** Implementation of equality and inequality operators.
   @return 1 on equality, otherwise 0
   @note Two iterators are equal if they are owned by same Dictionary object and
   point to same entry of that Dictionary object.
   */
   int equals(DictIterator comparison);

   /** Get key of referenced entry.
   Function returns key of entry where iterator points to. If iterator points to end of dictionary,
   an error is raised.
   @return key of referenced entry
   @note To prevent client code from modify key inside the dictionary, function returns a copy
   of the key rather than a reference to.
   */
   string getKey();

   /** Get value of referenced entry.
   Function returns value of entry where iterator points to. If iterator points to end of dictionary,
   an error is raised.
   @return value of referenced entry
   @note Result of function call is a shallow copy of entry value, i.e. if value is of reference type,
   modifications done on result may affect value in dictionary too.
   */
   var getValue();

   /** Set value of referenced entry.
   Function sets value of entry where iterator points to. If iterator points to end of dictionary,
   an error is raised.
   @param value new value
   */
   void setValue(var value);
   /** @protectedsection */
   /// dictionary entry where iterator currently points to
   var _current;
};

/** Bidirectional reverse iterator for Dictionary class.
This helper class allows reverse sequentially walking through entries of a Dictionary object.
Instances of this class are created by Dictionary methods only. Each iterator object referres
to a specific Dictionary.
@see Dictionary, DictIterator
*/
class DictRIterator
{
   /** @publicsection */

   /** Increment operator function.
   Function walks to entry before current one in owning dictionary and returns iterator to it.
   If iterator points not to a valid dictionary entry before call, an error is raised.
   @return current iterator pointing to entry before (if any) or end
   @see before
   */
   DictIterator OP_INC();
   /// Prefix increment operator (same as OP_INC)
   DictIterator operator++();

   /** Post increment operator function.
   Function walks to to entry before current one in owning dictionary and returns iterator to it.
   If iterator points not to a valid dictionary entry before call, an error is raised.
   @return new reverse iterator pointing to entry before (if any) or end iterator
   */
   DictIterator OP_PINC();

   /// Postfix increment operator (same as OP_NC)
   DictIterator operator++(var);

   /** Decrement operator function.
   Function walks to next entry of owning dictionary and returns iterator to it.
   If iterator points to end of dictionary before call, an error is raised.
   @return current iterator pointing to next entry (if any) or end
   @see next
   */
   DictIterator OP_DEC();
   /// Prefix decrement operator (same as OP_DEC)
   DictIterator operator--();

   /** Post decrement operator function.
   Function walks to next entry of owning dictionary and returns iterator to it.
   If iterator points to end of dictionary before call, an error is raised.
   @return new reverse iterator pointing to next entry (if any) or end iterator
   */
   DictIterator OP_PDEC();

   /// Postfix decrement operator (same as OP_PDEC)
   DictIterator operator++(var);



   /** Move iterator to next dictionary entry.
   Function walks to next entry of owning dictionary.<br>
   If iterator points to end of dictionary before call, an error is raised.
   @return current iterator pointing to next entry (if any) or end
   @note Function uses same implementation as decrement operator.
   @see OP_DEC
   */
   DictIterator next();

   /** Move iterator to dictionary entry before.
   Function walks to entry before current one.<br>
   If iterator points not to a valid dictionary entry before before call, an error is raised.
   @return current iterator pointing to next entry (if any) or end
   @note Function uses same implementation as decrement operator.
   @see OP_INC
   */
   DictIterator before();

   /** Implementation of equality and inequality operators.
   @return 1 on equality, otherwise 0
   @note Two iterators are equal if they are owned by same Dictionary object and
   point to same entry of that Dictionary object.
   */
   int equals(DictRIterator comparison);

   /** Get key of referenced entry.
   Function returns key of entry where iterator points to. If iterator points to end of dictionary,
   an error is raised.
   @return key of referenced entry
   @note To prevent client code from modify key inside the dictionary, function returns a copy
   of the key rather than a reference to.
   */
   string getKey();

   /** Get value of referenced entry.
   Function returns value of entry where iterator points to. If iterator points to end of dictionary,
   an error is raised.
   @return value of referenced entry
   @note Result of function call is a shallow copy of entry value, i.e. if value is of reference type,
   modifications done on result may affect value in dictionary too.
   */
   var getValue();

   /** Set value of referenced entry.
   Function sets value of entry where iterator points to. If iterator points to end of dictionary,
   an error is raised.
   @param value new value
    */
   void setValue(var value);
   /** @protectedsection */
   /// dictionary entry where iterator currently points to
   var _current;
};


/** Dictionary class.
Class encapsulates a <tt>dictionary</tt> and provides an interface similary to C++ std::map for using it.
*/
class Dictionary
{
   /** @publicsection */
   /// default ctor
   Dictionary();
   /** Create new dictionary from another one.
   Constructor builds new Dictionary object as a shallow copy of <i>assignment</i>.
   @param assignment Value to build dictionary from. The value must be of <tt>dictionary</tt> type or
   a <tt>Dictionary</tt> object.
   @note Constructor copies all key-value-pairs of <i>assignment</i> but makes no deep copy of
   values of reference type.
   */
   Dictionary(var assignment);


   /** Access operator (read).
   Operator returns value for specified <i>key</i>.
   @param key Key of element to get value for. If there is no entry for <i>key</i>
   in dictionary, a new (empty) one is created. The <i>key</i> must be either
   of <tt>string</tt> type or an object implementing <tt>string</tt> or <tt>toString</tt> method.
   @return Value for <i>key</i>. In case of empty value <tt>null</tt> is returned.
   */
   var OP_VREF(var key);
   /** Access operator (write).
   Operator sets value for dictionary entry specified by <i>key</i> to <i>val</i>.
   @param key Key of element to set value for. If there is no entry for <i>key</i>
   in dictionary, a new one is created. The <i>key</i> must be either
   of <tt>string</tt> type or an object implementing <tt>string</tt> or <tt>toString</tt> method.
   @param val New value.
   @return value of <i>val</i>
   */
   var OP_VSET(var key, var val);
   /** Get size of dictionary.
   @return Number of elements currently contained in dictionary.
   */
   uint size();

   /** Get iterator pointing to first element.
   @return iterator to first entry of dictionary (if any) or end-iterator otherwise
   @see end
   */
   DictIterator begin();

   /** Get iterator pointing behind last element.
   @return iterator to end of dictionary, i.e. behind last entry
   @note The end iterator is useful to check if end of dictionary was reached
   or find method returns without result.
   @par Example:
   @code
      Dictionary myDict;
      // fill dictionary
      // ...
      // iterate through dictionary
      for(auto it = myDict->begin(); it != myDict->end(); ++it)
      {
         // do anything
         // ...
      }
   @endcode
   */
   DictIterator end();

   /** Get reverse iterator pointing to last element.
   @return reverse iterator to last entry of dictionary (if any) or reverse end-iterator otherwise
   @see end
   */
   DictRIterator rbegin();

   /** Get reverse iterator pointing before first element.
   @return reverse iterator to end of dictionary, i.e. before first entry
   @note The reverse end iterator is useful to check if end of dictionary was reached
   for reverse iteration.
   @par Example:
   @code
   Dictionary myDict;
   // fill dictionary
   // ...
   // iterate through dictionary in reverse direction
   for(auto it = myDict->rbegin(); it != myDict->rend(); ++it)
   {
   // do anything
   // ...
   }
   @endcode
   */
   DictRIterator rend();

   /** Find dictionary entry by key.
   Function searches for dictionary entry specified by <i>key</i> and returns iterator to it.
   @param key Key of element to search for. The <i>key</i> must be either
   of <tt>string</tt> type or an object implementing <tt>string</tt> or <tt>toString</tt> method.
   @return Iterator to entry if found, otherwise an end-iterator.
   @par Example:
   @code
      myDict = new Dictionary();
      // fill dictionary
      // ...
      // search for entry named "mykey"
      it = myDict->find("mykey");
      if (it != myDict->end()) 
      {
         myval = it->getValue();
         // do anything
      }
      else // not found
      {
         // do anything else
      }
   @endcode
   */
   DictIterator find(var key);

   /** Removes all entries from dictionary.
   */
   void clear();

   /** Erase specified element from dictionary.
   Function removes entry specified by <i>elspec</i>.
   @param elspec Specification of element (dictionary entry) to erase. Valid types for
   <i>elspec</i> are:
   <ul>
     <li><tt>DictIterator</tt> object pointing to element to delete</li>
     <li><tt>string</tt> specifying key of element</li>
     <li>any object implementing <tt>string</tt> or <tt>toString</tt> method for specification of key</li>
   </ul>
   @return <tt>true</tt> if element was erased, otherwise <tt>false</tt>
   @note If <i>elspec</i> is an iterator, it is modified to point to element behind erased entry.
   */
   bool erase(var elspec);

   /** Convert current object into a plain dictionary value.
   @return internal dictionary value
   @note Returned value is the internal object. That means
   modifications done on result will affect the Dictionary object too.
   */
   dictionary dict();

   /** Conversion operator into plain dictionary value.
   Operator function is an alias of dict() methos.
   @return internal dictionary value
   @note Returned value is the internal object. That means
   modifications done on result will affect the Dictionary object too.
   */
   operator dictionary();

   /** @protectedsection */
   dictionary _dict;
}
#endif // DOXY

/// @}
#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

/// DictIterator::OP_INC implementation function
bool Bp2DictIterator_OP_INC(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::OP_PINC implementation function
bool Bp2DictIterator_OP_PINC(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::OP_DEC implementation function
bool Bp2DictIterator_OP_DEC(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::OP_PDEC implementation function
bool Bp2DictIterator_OP_PDEC(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::next implementation function
bool Bp2DictIterator_next(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::before implementation function
bool Bp2DictIterator_before(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::equals implementation function
bool Bp2DictIterator_equals(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::getKey implementation function
bool Bp2DictIterator_getKey(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::getValue implementation function
bool Bp2DictIterator_getValue(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictIterator::setValue implementation function
bool Bp2DictIterator_setValue(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Class creation function for DictIterator B++ class.
void Bp2DictIterator_CreateClass(BppVirtualMachine& vm);
/// DictRIterator::OP_PINC implementation function
bool Bp2DictRIterator_OP_PINC(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// DictRIterator::OP_PDEC implementation function
bool Bp2DictRIterator_OP_PDEC(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Class creation function for DictRIterator B++ class.
void Bp2DictRIterator_CreateClass(BppVirtualMachine& vm);

//@}
/// @}

/** @addtogroup Bp2Impl Implementation
@{
*/

/** @name Dictionary class implementation functions */
//@{

/// Dictionary constructor implementation function
bool Bp2Dictionary_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::OP_VREF implementation function
bool Bp2Dictionary_OP_VREF(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::OP_VSET implementation function
bool Bp2Dictionary_OP_VSET(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::begin implementation function
bool Bp2Dictionary_begin(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::end implementation function
bool Bp2Dictionary_end(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::rbegin implementation function
bool Bp2Dictionary_rbegin(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::rend implementation function
bool Bp2Dictionary_rend(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::find implementation function
bool Bp2Dictionary_find(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::size implementation function
bool Bp2Dictionary_size(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::clear implementation function
bool Bp2Dictionary_clear(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::erase implementation function
bool Bp2Dictionary_erase(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Dictionary::dict implementation function
bool Bp2Dictionary_dict(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Class creation function for Dictionary B++ class.
void Bp2Dictionary_CreateClass(BppVirtualMachine& vm);

//@}
/// @}
#endif // !DOXYUSER
/// @}

#endif // BP2DICT_H
