#ifndef BP2STRNG_H
#define BP2STRNG_H

class BppVirtualMachine;
class ValPtr;

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2Classes B++ Classes
@{
*/
#ifdef DOXY

/** String class.
This class encapsulates a <tt>string</tt>. It provides most important string operations vie object interface.
*/
class String
{
/** @publicsection */
   /// default ctor - creates empty string
   String();
   /** Create string as a copy of another one.
   Constructor makes a complete copy of <i>assignment</i> value.
   @param assignment Value to create String object from. Value must be either
   of <tt>string</tt> type implicitely convertible to <tt>string</tt>.
   */
   String(string assignment);
   /** Create string from format specification.
   Constructor builds string from format specification and any number
   of additional arguments.
   @param formatstr Format specification. Value must be either
   of <tt>string</tt> type or implicitely convertible to <tt>string</tt>.
   The format string has similary syntax as in C - see BppString description.
   @param ... More arguments to replace variables in format string. 
   @note Elements of variable argument list should be of simple types, i.e.
   integer, double, float or string (not String class!). For all other types
   return value of BppValue::addr() method is used for variable value.
   */
   String(string formatstr,...);
   /// Destructor releases internal string object.
   ~String();
   /** Left side concatenation operator implementation.
   Operator creates a new String object containing value of <i>suffix</i>
   appended to value of current string.
   @param suffix Value to append. The value may be a single character, a string
   or any object implementing a <tt>string</tt> or <tt>toString</tt> method.
   @return String object
   @par Usage example;
   @code
      str = new String("Hello ");
      str2 = str + "World!\n";
   @endcode
   */
   String OP_ADD(var suffix);
   /** Right side concatenation operator implementation.
   Operator creates a new String object containing value of <i>prefix</i>
   prepended to value of current string.
   @param prefix Value to prepend. The value may be a single character, a string
   or any object implementing a <tt>string</tt> or <tt>toString</tt> method.
   @return String object
   @par Usage example;
   @code
   str = new String("World!\n");
   str2 = "Hello " + str;
   @endcode
   */
   String OP_ADD_R(var prefix);
   /** Access operator (read).
   Operator returns single character at index <i>index</i>.
   @param index Index of character to return.
   @return character at <i>index</i> position
   @par Usage example;
   @code
   str = new String("Hello World!\n");
   c = str[1]; // returns 'e'
   @endcode
   */
   int OP_VREF(uint index);
   /** Access operator (write).
   Operator sets character at index position<i>index</i> to <i>charval</i>.
   @param index Index of character.
   @param charval New character value.
   @return <i>charval</i>
   @par Usage example;
   @code
   str = new String("Hello World!\n");
   str[1] = 'a'; // makes "Hello" to "Hallo"
   @endcode
   */
   int OP_VSET(uint index, int charval);
   /** Compare function (case sensitive).
   Function compares value of current string against <i>comparison</i>.
   Implicitely it immpelents compare operators for String class too.
    @param comparison Value to compare with. The value must be a string or
   implicitely convertible to <tt>string</tt>.
   @return <ul> 
   <li> &lt; 0 if current value is lower than <i>comparison</i></li>
   <li>  0 if current value equals to <i>comparison</i></li>
   <li> &gt; 0 if current value is greater than <i>comparison</i></li>
   </ul>
   */
   int compare(var comparison);
   /** Compare function (case insensitive).
   Function compares value of current string against <i>comparison</i>.
   @param comparison Value to compare with. The value must be a string or
   implicitely convertible to <tt>string</tt>.
   @return <ul> 
   <li> &lt; 0 if current value is lower than <i>comparison</i></li>
   <li>  0 if current value equals to <i>comparison</i></li>
   <li> &gt; 0 if current value is greater than <i>comparison</i></li>
   </ul>
   */
   int icompare(var comparison);
   /** Get length of current string.
   @return number of characters in string (without trailing '\\0' character)
   */
   uint length();
   /** Get C-string from current object.
   This function allows simple use of String objects for arguments
   in calls of system functions. It returns contents of string as
   a C-like character pointer.
   @return start address of internal character buffer
   @note <p>Client code must not change size of string nor delete the returned
   value.</p>
   <p>Character size of returned C-string depends on build type. i.e for UNICODE
   builds 16 bits wide characters, otherwise single byte characters are used.
   @par Usage example;
   @code
      msg = new String("Hello World!");
      title = new String("Message"); 
      hinst = LoadLibrary("user32.dll");
      fptr = GetProcAddress(hinst,"MessageBox?");
      CallLibFunc(fptr,0,msg->c_str(),title->c_str(),0);
      FreeLibrary(hinst);
   @endcode
   */
   intptr c_str();
   /** Get character buffer containing string contents.
   Function creates a character buffer (BppCharBuffer) containing a
   copy of contents of current String object.
   @return character buffer
   */
   charbuffer c_buf();
   /** Get byte buffer containing string contents (single byte).
   Function creates a buffer (BppBuffer) containing a
   copy of contents of current String object. Each character is represented
   by a single byte.
   @return buffer
   */
   buffer a_buf();
   /** Get byte buffer containing string contents (wide character).
   Function creates a buffer (BppBuffer) containing a
   copy of contents of current String object. Each character is represented
   by two bytes.
   @return buffer
   */
   buffer w_buf();
   /** Clears all contents from string.
    */
   void clear();
   /** Find position of substring or character.
   Function searches for substring or character <i>search</i> and returns
   position of its first occurrence.
   @param search String or character to search for. Instead of string any object
   implementing <tt>string</tt> or <tt>toString</tt> method may be used.
   @param start Character index to start search from. 
   @return Position of first occurrence if found, otherwise NPOS.
   */
   uint find(var search, uint start = 0);
   /** Replace part of string.
   Function replaces all occurrences of <i>search</i> by value of <i>replacement</i>.
   @param search character or substring to be replaced
   @param replacement character or string to be used for replacement
   @return current object
   @note Both parameters must be either single characters or strings (or objects
   convertible to string).
   */
   String replace(var search,var replacement);
   /** Get substring from current string.
   Function returns substring of current string, beginning at position <i>start</i>
   and of maximum <i>len</i> characters length.
   @param start Start position of substring.
   @param len Maximum size of substring. If len is omitted or greater as number of characters behind <i>start</i>,
   all characters of current string (beginning at <i>start</i> are included in result.
   @return new String object containing specified substring
   */
   String substr(uint start,uint len = NPOS);
   /** Erase part of string.
   Function deletes on or more characters from string.
   @param pos Index of first character to delete.
   @param cnt Number of characters to delete. If <i>cnt</i> id greater than number of characters behid <i>pos</i>,
	all characters from <i>pos</i> to end are removed.
   @return current object
   */
   String erase(int pos, uint cnt = NPOS);
   /** Insert specified string into current one.
   Function inserts value of <i>str</i> into current string at position <i>pos</i>.
   @param str String to insert. Value may be a string or any object implementing
   <tt>string</tt> or <tt>toString</tt> method.
   @param pos index position where <i>str</i> should be inserted
   @return current object
   */
   String insert(string str, uint pos);
   /** Insert character(s) into current string.
   Function inserts <i>cnt</i> characters <i>chr</i> into current string at position <i>pos</i>.
   @param chr character value to insert
   @param pos index position for insertion
   @param cnt number of characters to insert
   @return current object
   */
   String insert(int chr, uint pos, uint cnt = 1);
   /** Append string or single character to current string.
   Function appends either another string or a single character to current string value.
   @param suffix Value to append. The value may be a single character, a string
   or any object implementing <tt>string</tt> or <tt>toString</tt> method.
   @return current object
   @note Other than concatenation operator this function modifies current object
   instead of creating a new instance.
   @see OP_ADD
   */
   String append(var suffix);
   /** Convert current object into a string value.
   @return copy of internal string value
   @note Returned value is not the internal value of object but a copy of it. That means
   modifications done on result will not affect the string object.
   */
   operator string();
   /** Convert current object into a string value.
   Method is an alias of <tt>operator string()</tt>.
   @return copy of internal string value
   @note Returned value is not the internal value of object but a copy of it. That means
   modifications done on result will not affect the string object.
   */
   string toString();

/** @protectedsection */
   string _str;
}

#endif
/// @}
#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

/** @name String class implementation functions */
//@{
/** String constructor implementation.
@see String::String
*/
bool Bp2String_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::OP_ADD implementation function
bool Bp2String_OP_ADD(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::OP_ADD_R implementation function
bool Bp2String_OP_ADD_R(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::OP_VREF implementation function
bool Bp2String_OP_VREF(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::OP_VSET implementation function
bool Bp2String_OP_VSET(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::compare implementation function
bool Bp2String_compare(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::icompare implementation function
bool Bp2String_icompare(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::length implementation function
bool Bp2String_length(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::c_str implementation function
bool Bp2String_c_str(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::c_buf implementation function
bool Bp2String_c_buf(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::a_buf implementation function
bool Bp2String_a_buf(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::w_buf implementation function
bool Bp2String_w_buf(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::clear implementation function
bool Bp2String_clear(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::find implementation function
bool Bp2String_find(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::replace implementation function
bool Bp2String_replace(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::substr implementation function
bool Bp2String_substr(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::erase implementation function
bool Bp2String_erase(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::insert implementation function
bool Bp2String_insert(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::append implementation function
bool Bp2String_append(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// String::toString implementation function
bool Bp2String_toString(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Class creation function for String B++ class.
void Bp2String_CreateClass(BppVirtualMachine& vm);
//@}
/// @}
#endif //!DOXYUSER
/// @}
#endif // BP2STRNG_H