#ifndef BP2STRUC_H
#define BP2STRUC_H

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2Classes B++ Classes
@{
*/
#ifdef DOXY
/** Native structure class.
<p>Windows API functions often need pointers to structures for arguments. Because B++
doesn't support structure types directly they must be simulated using e.g. a buffer.<br>
This class helps for creating such a buffer.</p>
<p>All <tt>addXXX</tt> functions take a <i>name</i> (name of structure member), a <i>value</i>
(value to set for member) and optionally a <i>position</i>. The position is the byte offset where starts
member inside the native structure. If omitted, new member is appended to end of buffer.</p>
@note <ul>
  <li>Struct class doesn't fully support C-like unions. Union-like behavior can be 
  simulated for value types (but not for pointers) by placing multiple structure members
  at same position.</li>
  <li>Client code should take care not to overlap structure members of pointer types.</li>
  <li>Before reading values modified by a system function call, <tt>update</tt> method should
  be called to apply modified buffer contents to structure.</li>
</ul>
*/
class Struct
{
/** @publicsection */
   /// constructor
   Struct();
   /** Add 32-bits integer to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addInt(var name,int value, int pos=-1);
   /** Add pointer to 32-bits integer to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add address of
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addIntPtr(var name, int value, int pos=-1);
   /** Add 8-bits integer (char) to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addInt8(var name, int value, int pos=-1);
   /** Add 16-bits integer (short) to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addInt16(var name, int value, int pos=-1);
   /** Add double (8 bytes) to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addDouble(var name,double value, int pos=-1);
   /** Add pointer to double to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add address of
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addDoublePtr(var name, double value, int pos=-1);
   /** Add float (4 bytes) to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addFloat(var name, float value, int pos=-1);
   /** Add pointer to float to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add address of
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addFloatPtr(var name, float value, int pos=-1);
   /** Add string to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value Value to add. The value may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addString(var name, var value, int pos=-1);
   /** Add character buffer to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addCBuffer(var name, charbuffer value, int pos=-1);
   /** Add byte buffer to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addBuffer(var name, buffer value, int pos=-1);
   /** Add sub-structure to structure.
   @param name Member name, must be unique in entire structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @param value value to add
   @param pos Byte position in buffer. If omitted or lower than zero, value is appended to end of buffer.
   @return byte index in buffer where value starts
   */
   int addStruct(var name, Struct value, int pos=-1);

   /** Get value of a member by name.
   @param name Member name, must be contained in structure. The name may be a <tt>string</tt> or any object
   providing a <tt>toString</tt> method.
   @return value of structure member
   @note If native structure buffer could be modified, #update should be called before getting a value.
   */
   var getValue(var name);
   /** Get buffer of native structure data.
   @return native data buffer
   @note Use #addr function to get pointer to native structure buffer and #size function to 
   get its size.
   */
   buffer getData();
/** @protectedsection */
   /// buffer for storing native structure
   BppBuffer _data;
   /// dictionary for data access
   BppDictionary _dataDict;
}
#endif
/// @}

#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

class BppVirtualMachine;
class ValPtr;
/** @name Struct class implementation functions */
//@{

bool Bp2Struct_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addInt(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addIntPtr(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addInt8(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addInt16(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addDouble(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addDoublePtr(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addFloat(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addFloatPtr(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addString(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addCBuffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addBuffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_addStruct(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_getValue(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_getData(BppVirtualMachine* vm, int argc, ValPtr& sp);
bool Bp2Struct_update(BppVirtualMachine* vm, int argc, ValPtr& sp);
void Bp2Struct_CreateClass(BppVirtualMachine& vm);
//@}
/// @}
#endif //!DOXYUSER
/// @}

#endif // BP2STRUC
