#ifndef BP2VEC_H
#define BP2VEC_H

class BppVirtualMachine;
class ValPtr;

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2Classes B++ Classes
@{
*/
#ifdef DOXY
/** Vector class.
Class encapsualates a <tt>vector</tt> and provides an objekt riented interface to it.<br>
Vector class allows using of <tt>vector</tt> as an array, a list, a queue or a stack.
*/
class Vector
{
   /** @publicsection */
   /// default ctor
   Vector();
   /** Create new vector from another one.
   Constructor builds new Vector object as a shallow copy of <i>assignment</i>.
   @param assignment Value to build vector from. The value must be of <tt>vector</tt> type or
   a <tt>Vector</tt> object.
   @note Constructor copies all elements of <i>assignment</i> but makes no deep copy of
   element values of reference type.
   */
   Vector(var assignment);
   /** Create new vector of specified capacity.
   Constructor builds a new vector and optionally initializes its elements.
   @param capacity initial capacity of vector
   @param init Flag for vector initialization. If <tt>true</tt> (i.e. not 0),
   vector is initialized, i.e. vector size is set to be same as <i>capacity</i>
   and all elements are set to <tt>null</tt>.
   */
   Vector(int capacity, int init=0);

   /** Left side concatenation operator implementation.
   Operator creates a new Vector object containing value of <i>suffix</i>
   appended to current vector.
   @param suffix Value to append. If <i>suffix</i> is of type <tt>vector</tt> or a Vector object, elements of <i>suffix</i>
   are appended. Otherwise <i>suffix</i> is appended as a single element.
   @return Vector object
   */
   Vector OP_ADD(var suffix);
   /** Right side concatenation operator implementation.
   Operator creates a new Vector object containing value of <i>prefix</i>
   prepended to current vector.
   @param prefix Value to prepend. If <i>prefix</i> is of type <tt>vector</tt>, elements of <i>prefix</i>
   are prepended. Otherwise <i>prefix</i> is prepended as a single element.
   @return Vector object
   */
   Vector OP_ADD_R(var prefix);
   /** Access operator (read).
   Operator returns single vector element at index <i>index</i>.
   @param index Index of element to return.
   @return element at <i>index</i> position
   */
   var OP_VREF(uint index);
   /** Access operator (write).
   Operator element value at index position<i>index</i> to <i>val</i>.
   @param index Index of element.
   @param val New value.
   @return <tt>val</tt>
   */
   null OP_VSET(int index, var val);
   /** Get size of vector.
   @return Number of elements currently contained in vector.
   */
   uint size();
   /** Get capacity of vector.
   @return Maximum Number of elements currently space reserved for in vector.
   */
   uint capacity();
   /** Set capacity of vector.
   Function modifies capacity of Vector object.
   @param capacity Capacity to set.
   @return Maximum Number of elements currently space reserved for in vector, i.e. new capacity.
   @note Function doesn't delete any vector element. If <i>capacity</i> is lower than current size,
   vector capacity will be set to same as current size.
   */
   uint setCapacity(uint capacity);
   /** Shrink capacity to size.
   Function reduces capacity to number of currently used elements.
   @return new capacity
   */
   uint shrink();
   /** Push element to end of vector.
   Function appends a value (i.e. a new element) to end of vector.
   @param val value to append
   @return This Vector object itself.
   */
   Vector push_back(var val);
   /** Pop last vector element.
   Function removes last element from vector and returns its value.
   @return value of removed element
   */
   var pop_back();
   /** Push element to begin of vector.
   Function inserts a value (i.e. a new element) at begin of vector.
   @param val value to insert
   @return This Vector object itself.
   */
   Vector push_front(var val);
   /** Pop first vector element.
   Function removes first element from vector and returns its value.
   @return value of removed element
   */
   var pop_front();
   /** Remove all elements from vector.
   */
   void clear();

   /** Insert value(s) into vector.
   Function inserts one or more values into current Vector object.
   @param pos Position where value(s) should be inserted. If <i>pos</i> is greater than maximum index value(s) will be appended.
   @param val Value to insert or append.
   @param ... Comma separared list of more values to insert/append behind <i>val</i>.
   @return This Vector object itself.
   */
   Vector insert(uint pos, var val, ...);

   /** Get a sub-vector from current one.
   Function creates a new Vector object containing copy of specified part of current one.
   @param from Index of first element to copy.
   @param cnt Maximum number of elements to copy. If <i>cnt</i> is grater than number of elements behind <i>from</i>, all elements beginning at <i>from</i> are copied.
   @return New Vector object containing result.
   */
   Vector slice(uint from = 0, uint cnt = NPOS);

   /** Erase one or more element(s) from vector.
   Function removes specified number of elements, starting at <i>from</i>.
   @param from Index of first element to erase.
   @param cnt Maximum number of elements to erase. If <i>cnt</i> is grater than number of elements behind <i>from</i>, all elements beginning at <i>from</i> are removed.
   @return This Vector object itself.
   */
   Vector erase(int from, uint cnt = NPOS);

   /** Convert current object into a plain vector value.
   @return internal vector value
   @note Returned value is the internal object. That means
   modifications done on result will affect the Vector object too.
   */
   vector vec();

   /** Conversian to plain vector value.
   Operator function is an alias of vec() method.
   @return internal vector value
   @note Returned value is the internal object. That means
   modifications done on result will affect the Vector object too.
   */
   operator vector();

   /** @protectedsection */
   vector _vec;
}
#endif // DOXY
/// @}

#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

/** @name Vector class implementation functions */
//@{

/// Vector constructor implementation function
bool Bp2Vector_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::OP_ADD implementation function
bool Bp2Vector_OP_ADD(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::OP_ADD_R implementation function
bool Bp2Vector_OP_ADD_R(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::OP_VREF implementation function
bool Bp2Vector_OP_VREF(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::OP_VSET implementation function
bool Bp2Vector_OP_VSET(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::size implementation function
bool Bp2Vector_size(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::capacity implementation function
bool Bp2Vector_capacity(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::setCapacity implementation function
bool Bp2Vector_setCapacity(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::shrink implementation function
bool Bp2Vector_shrink(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::push_back implementation function
bool Bp2Vector_push_back(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::pop_back implementation function
bool Bp2Vector_pop_back(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::push_front implementation function
bool Bp2Vector_push_front(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::pop_front implementation function
bool Bp2Vector_pop_front(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::clear implementation function
bool Bp2Vector_clear(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::insert implementation function
bool Bp2Vector_insert(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::slice implementation function
bool Bp2Vector_slice(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::erase implementation function
bool Bp2Vector_erase(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Vector::vec implementation function
bool Bp2Vector_vec(BppVirtualMachine* vm, int argc, ValPtr& sp);
/// Class creation function for Vector B++ class.
void Bp2Vector_CreateClass(BppVirtualMachine& vm);

//@}
/// @}
#endif //!DOXYUSER
/// @}

#endif // BP2VEC_H
