#ifndef BP2WBASE_H
#define BP2WBASE_H

#ifdef WIN32

class BppVirtualMachine;
class ValPtr;

#include <bppstrng.h>
#include <windows.h>
#ifdef WINCE
  #include <winuser.h>
#endif

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2WinBase Windows Base Classes
@{
*/

#ifdef DOXY
/** Base class of installable message handlers.
Installable message handlers can be used alternative to or additional to overriding WinBase::handleMessage method.<br>
Each window can maintain a list of installable message handlers. WinBase::handleMessage calls handlers contained in list
before it does its own message handling.<br>
If one of installed handlers returns zero from its handleMessage method,
message is considered to be completely handeled, i. e. further handling is aborted.
*/
class WinMsgHandler
{ 
public:
   /** Ctor.
   @param owner Window that owns the handler.
   @note If Handler is installed by WinBase::addMessageHandler, owner is set using setOwner method, i.e.
   <i>owner</i> parameter not needs to be set.
   */
   WinMsgHandler(WinBase owner = null);
   /** Sets owner window reference.
   @param owner Window that owns the handler.
   @note  <ul>
   <li>In case of invalid <i>owner</i> value, exception of string type is raised.</li>
   <li>Reference counter ow <i>owner</i> is not incremented by funotion, i.e. handler doesn't block
   deletion of owner window.
   </ul>
   */
   void setOwner(WinBase owner);
   /// Gets owner window or <tt>null</tt> if none.
   WinBase getOwner() const;
   /** Handles message.
   Base implementation does nothing but return 1.<br>
   This method should be overridden in derived class.
   @parem sender Window which invokes message handler.
   @param hwnd Handle of destination window for message. Typical <i>hwnd</i> should same as 
   window handle of <i>sender</i>.
   @param msgId Windows message ID
   @param wParam wParam value of message
   @param lParam lParam of message
   @return Result of message handling. If result is zero, no further handling of message is done.
   */
   int handleMessage(WinBase sender, int hwnd, int msgId, int wParam, int lParam);
   /** Implementation of equality operators.
   @return 1 if <i>other</i> is the current object itself, otherwise 0.
   */
   int equals(const WinBase other) const;
protected:
   /// Internal (weak) refrence to owner window.
   var _owner;
};




/** Base class of all window objects.
This class is intended to be used for base of all windows, i.e.
all classes having its own window handle. It provides a set of functions common to
all windows as well as basic window and application management services.
*/
class WinBase
{
public:
   /// ctor
   WinBase();
   /// dtor
   ~WinBase();
   /** Creates wrapper for existing window.
	Method assigns existing window specified by <i>hwnd</i> to current object and registers it for use with B++.
	@param hwnd Handle of window to assign.
	@return true (1) on success, false (0) otherwise.
	@note Method fails (i.e. returns false) in case of invalid handle  or if <i>hwnd</i> already assigned to a
	B++ window.
	@note Function replaces window procedure for <i>hwnd</i> in order to allow B++ event handling for window.
	*/
   int createFromHandle(intptr hwnd);
   /** Creates window from window class.
   Method creates a new window of type specified by <i>wndClass</i>, initializes its message system and registers it for
   use with B++.<br>
   Typically method is used by a derived class rather then  directly from client code.
   @param wndClass Window class name
   @param hwndParent Parent window (subtype of WinBase) or handle of parent window (intptr). 
   @param x Initial x position of upper left corner relative to parent window, -1 means default position.
   @param y Initial y position of upper left corner ralative to parent window, -1 means default position.
   @param width Initial width in pixels. In case of negative value sign is ignored.
   @param height Initial height in pixels. In case of negative value sign is ignored.
   @param style Window style flags mask. See setStyle for possible values.
   @param extStyle Window extended style flags mask. See setExtStyle for possible values.
   @param hMenu Handle of menu used for window, 0 means none.
   @param text Initial window text, meaning depends on window class.
   @return true (1) on success, false (0) otherwise
   @note In case of invalid arguments exception of string type is thrown.
   */
   int create(string wndClass, var parent = 0, int x = -1, int y = -1, int width = 100, int height = 100, 
       int style = 0, int extStyle = 0, intptr hMenu_ID = 0, string text = "");
   /** Gets associated window handle.
   @return Handle (HWND) assiciated with window object, if any, otherwise zero.
   */
   intptr getHandle() const;
   /** Gets handle of parent window.
   Handle (HWND) of parent window, if any, otherwise zero.
   */
   intptr getParentHandle() const;
   /** Gest parent window.
   @return Parant windof, if any, otherwise <tt>null</tt>.
   */
   WinBase getParent() const;
   /** Adds child window to this window.
   @param child Window to make a child of current window.
   @return true (1) on success, otherwise false (0).
   @note In case of invalid <i>child</i> argument exeption of string type is raised.
   */
   int addChild(WinBase child);
   /** Get vector of all child window handles.
   @return vector containing handles of all direct child windows of current window.
   */
   vector getChildrenHandles() const;

   /** Sets window position relative to parent.
   @param x New x position.
   @param y new y position.
   @param repaint If <tt>true</tt>, WM_PAINT messages will be sent to window.
   @return true (1) on success, otherwise false (0).
   @note Function sends WM_WINDOWPOSCHANGING, WM_WINDOWPOSCHANGED, WM_MOVE, WM_SIZE and 
   WM_NCCALCSIZE messages to the window.
   */
   int setPosition(int x, int y, bool repaint = true);
   /** Sets window position relative to parent.
   @param pos New position.
   If <i>pos</i> not of type POINT, exception of string type is raised.
   @param repaint If <tt>true</tt>, WM_PAINT messages will be sent to window.
   @return true (1) on success, otherwise false (0).
   @note Function sends WM_WINDOWPOSCHANGING, WM_WINDOWPOSCHANGED, WM_MOVE, WM_SIZE and
   WM_NCCALCSIZE messages to the window.
   */
   bool setPosition(POINT pos, bool repaint = true);

   /** Sets absolute size of window in pixels.
   @param width New width in pixels.
   @param height New height in pixels.
   @param repaint If <tt>true</tt>, WM_PAINT messages will be sent to window.
   @return true (1) on success, otherwise false (0).
   @note Function sends WM_WINDOWPOSCHANGING, WM_WINDOWPOSCHANGED, WM_MOVE, WM_SIZE and
   WM_NCCALCSIZE messages to the window.
   */
   bool setSize(int width, int height, bool repaint = true);
   
   /** Sets absolute size of window in pixels.
   @param pos New size.
   If <i>size</i> not of type SIZE, exception of string type is raised.
   @param repaint If <tt>true</tt>, WM_PAINT messages will be sent to window.
   @return true (1) on success, otherwise false (0).
   @note Function sends WM_WINDOWPOSCHANGING, WM_WINDOWPOSCHANGED, WM_MOVE, WM_SIZE and
   WM_NCCALCSIZE messages to the window.
   */
   bool setSize(SIZE size, bool repaint = true);

   /** Gets position and size of window in pixels.
   @return RECT containing size and position of window.
   If there are a parent window, position is relative to parent.
   */
   RECT getWindowRect() const; 
   /** Gets position and size of client area in pixels.
   @return RECT containing size and position of client area.
   */
   RECT getClientRect() const;
   /** Gets window position.
   @return POINT containing position of window.
   If there are a parent window, position is relative to parent.
   */
   POINT getPosition() const;
   /** Gets position of client area.
   @return POINT containing position of client area.
   */
   POINT getClientPos() const;
   /** Gets absolute size in pixels.
   @return SIZE object containing width and height of window.
   */
   SIZE getSize() const;
   /** Gets absolute size of client area in pixels.
   @return SIZE object containing width and height of client area.
   */
   SIZE getClientSize() const;
   /// Gets x position (horizontal position).
   int getPosX() const;
   /// Gets y position (vertical position).
   int getPosY() const;
   /// Gets x position of client area.
   int getClientPosX() const;
   /// Gets y position of client area.
   int getClientPosY() const;
   /// Gets window width.
   int  getWidth() const;
   /// Gets window height.
   int getHeight() const;
   /// Gets width of client area.
   int getClientWidth() const;
   /// Gets height of client area.
   int getClientHeight() const;
 
   /** Gets window style flags.
   @return Combination of style flags actually set for window.
   For detailed description see setStyle documentation.
   @see setStyle, hasStyle
   */
   int getStyle() const;

   /** Checks if window has specified style(s).
   @param style Mask of style flag(s) to check for.
   @return true (1) if window has all style flags of mask <i>style</i>, otherwise false (0).
   @see getStyle
   */
   int hasStyle(int style) const;
   
   /** Sets window style.
   @param style Combination of following style flags to set for window:
   <table><tr><th>Flag constants</th><th>Description</th></tr>
   <tr><td>WS_BORDER</td><td>The window has a thin-line border.</td></tr>
   <tr><td>WS_CAPTION</td><td>The window has a title bar (includes the WS_BORDER style).</td></tr>
   <tr><td>WS_CHILD</td><td>The window is a child window. 
   A window with this style cannot have a menu bar. 
   This style cannot be used with the WS_POPUP style.</td></tr>
   <tr><td>WS_CHILDWINDOW</td><td>Same as the WS_CHILD style.</td></tr>
   <tr><td>WS_CLIPCHILDREN</td><td>Excludes the area occupied by child windows when drawing occurs 
   within the parent window. This style is used when creating the parent window.</td></tr>
   <tr><td>WS_CLIPSIBLINGS</td><td>Clips child windows relative to each other; that is, 
   when a particular child window receives a WM_PAINT message, the WS_CLIPSIBLINGS style clips 
   all other overlapping child windows out of the region of the child window to be updated. 
   If WS_CLIPSIBLINGS is not specified and child windows overlap, it is possible, 
   when drawing within the client area of a child window, to draw within the client 
   area of a neighboring child window.</td></tr>
   <tr><td>WS_DISABLED</td><td>The window is initially disabled. A disabled window cannot receive input from the user. 
   To change this after a window has been created, use the EnableWindow API function.</td></tr>
   <tr><td>WS_DLGFRAME</td><td>The window has a border of a style typically used with dialog boxes. 
   A window with this style cannot have a title bar.</td></tr>
   <tr><td>WS_GROUP</td><td>The window is the first control of a group of controls. 
   The group consists of this first control and all controls defined after it, up to the next 
   control with the WS_GROUP style. The first control in each group usually has the WS_TABSTOP 
   style so that the user can move from group to group. The user can subsequently change the 
   keyboard focus from one control in the group to the next control in the group by using the 
   direction keys.<br>
   You can turn this style on and off to change dialog box navigation. To change this style after 
   a window has been created, use the SetWindowLong function.</td></tr>
   <tr><td>WS_HSCROLL</td><td>The window has a horizontal scroll bar.</td></tr>
   <tr><td>WS_ICONIC</td><td>The window is initially minimized. Same as the WS_MINIMIZE style.</td></tr>
   <tr><td>WS_MAXIMIZE</td><td>The window is initially maximized.</td></tr>
   <tr><td>WS_MAXIMIZEBOX</td><td>The window has a maximize button. Cannot be combined with the WS_EX_CONTEXTHELP style. 
   The WS_SYSMENU style must also be specified.</td></tr>
   <tr><td>WS_MINIMIZE</td><td>The window is initially minimized. Same as the WS_ICONIC style.</td></tr>
   <tr><td>WS_MINIMIZEBOX</td><td>The window has a minimize button. Cannot be combined with the WS_EX_CONTEXTHELP style. 
   The WS_SYSMENU style must also be specified.</td></tr>
   <tr><td>WS_OVERLAPPED</td><td>The window is an overlapped window. An overlapped window has a title bar and a border. 
   Same as the WS_TILED style.</td></tr>
   <tr><td>WS_OVERLAPPEDWINDOW</td><td>(WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)<br>
   The window is an overlapped window. Same as the WS_TILEDWINDOW style</td></tr>
   <tr><td>WS_POPUP</td><td>The window is a pop-up window. This style cannot be used with the WS_CHILD style.</td></tr>
   <tr><td>WS_POPUPWINDOW</td><td>(WS_POPUP | WS_BORDER | WS_SYSMENU)<br>
   he window is a pop-up window. The WS_CAPTION and WS_POPUPWINDOW styles must be combined to 
   make the window menu visible.</td></tr>
   <tr><td>WS_SIZEBOX</td><td>The window has a sizing border. Same as the WS_THICKFRAME style.</td></tr>
   <tr><td>WS_SYSMENU</td><td>The window has a window menu on its title bar. The WS_CAPTION style must also be specified.</td></tr>
   <tr><td>WS_TABSTOP</td><td>The window is a control that can receive the keyboard focus when the user presses 
   the TAB key. Pressing the TAB key changes the keyboard focus to the next control with the WS_TABSTOP style.<br>
   You can turn this style on and off to change dialog box navigation.</td></tr>
   <tr><td>WS_THICKFRAME</td><td>The window has a sizing border. Same as the WS_SIZEBOX style.</td></tr>
   <tr><td>WS_TILED</td><td>The window is an overlapped window. An overlapped window has a title bar and a border. 
   Same as the WS_OVERLAPPED style.</td></tr>
   <tr><td>WS_TILEDWINDOW</td><td>(WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)<br>
   The window is an overlapped window. Same as the WS_OVERLAPPEDWINDOW style.</td></tr>
   <tr><td>WS_VISIBLE</td><td>The window is initially visible.<br>
   This style can be turned on and off by using the show method or SetWindowPos function.</td></tr>
   <tr><td>WS_VSCROLL</td><td>The window has a vertical scroll bar.</td></tr>
  </table>
   @return Previous style flags on success, otherwise 0.
   */
   int setStyle(int style);

   /** Adds specified style(s) to window.
   Does OR combination of <i>style</i> with style flag currently set.
   @param style Style(s) to add.
   @return Previous style flags on success, otherwise 0.
   @see getStyle, setStyle
   */
   int addStyle(int style);

   /** Removes specified style(s) from window if currently set.
   @param style Style(s) to remove.
   @return Previous style flags on success, otherwise 0.
   @see getStyle, setStyle
   */
   bool removeStyle(int style);

   /** Gets extended window style flags.
   @return Combination of extended style flags actually set for window.
   For detailed description see setExtStyle documentation.
   @see setExtStyle, hasExtStyle
   */
   int getExtStyle() const;

   /** Checks if window has specified extended style(s).
   @param style Mask of style flag(s) to check for.
   @return true (1) if window has all extended style flags of mask <i>style</i>, otherwise false (0).
   @see getExtStyle
   */
   bool hasExtStyle(int style) const;

   /** Sets extended window style.
   @param style Combination of following extended style flags to set for window:
   <table><tr><th>Flag constants</th><th>Description</th></tr>
      <tr><td>WS_EX_ACCEPTFILES</td><td>The window accepts drag-drop files.</td></tr>
	  <tr><td>WS_EX_APPWINDOW</td><td>Forces a top-level window onto the taskbar when the window is visible.</td></tr>
	  <tr><td>WS_EX_CLIENTEDGE</td><td>The window has a border with a sunken edge.</td></tr>
	  <tr><td>WS_EX_COMPOSITED</td><td>Paints all descendants of a window in bottom-to-top painting order 
	  using double-buffering. Bottom-to-top painting order allows a descendent window to have translucency 
	  (alpha) and transparency (color-key) effects, but only if the descendent window also has the 
	  WS_EX_TRANSPARENT bit set. Double-buffering allows the window and its descendents to be painted 
	  without flicker. This cannot be used if the window has a class style of either CS_OWNDC or CS_CLASSDC.<br>
      Note: This style is supported for Windows XP and newer.</td></tr>
	  <tr><td>WS_EX_CONTEXTHELP</td><td>The title bar of the window includes a question mark. 
	  When the user clicks the question mark, the cursor changes to a question mark with a pointer. 
	  If the user then clicks a child window, the child receives a WM_HELP message. The child window 
	  should pass the message to the parent window procedure, which should call the WinHelp function 
	  using the HELP_WM_HELP command. The Help application displays a pop-up window that typically 
	  contains help for the child window.<br>
      WS_EX_CONTEXTHELP cannot be used with the WS_MAXIMIZEBOX or WS_MINIMIZEBOX styles.</td></tr>
	  <tr><td>WS_EX_CONTROLPARENT</td><td>The window itself contains child windows that should take part 
	  in dialog box navigation. If this style is specified, the dialog manager recurses into children of 
	  this window when performing navigation operations such as handling the TAB key, an arrow key, or a 
	  keyboard mnemonic.</td></tr>
	  <tr><td>WS_EX_DLGMODALFRAME</td><td>The window has a double border; the window can, optionally, 
	  be created with a title bar by specifying the WS_CAPTION style in the dwStyle parameter.</td></tr>
	  <tr><td>WS_EX_LAYERED</td><td>The window is a layered window. This style cannot be used if the 
	  window has a class style of either CS_OWNDC or CS_CLASSDC.<br>
      Windows 8: The WS_EX_LAYERED style is supported for top-level windows and child windows. 
	  Previous Windows versions support WS_EX_LAYERED only for top-level windows.</td></tr>
	  <tr><td>WS_EX_LAYOUTRTL</td><td>If the shell language is Hebrew, Arabic, or another language that supports 
	  reading order alignment, the horizontal origin of the window is on the right edge. 
	  Increasing horizontal values advance to the left.</td></tr>
	  <tr><td>WS_EX_LEFT</td><td>The window has generic left-aligned properties. This is the default (value 0).</td></tr>
	  <tr><td>WS_EX_LEFTSCROLLBAR</td><td>If the shell language is Hebrew, Arabic, or another language that 
	  supports reading order alignment, the vertical scroll bar (if present) is to the left of the 
	  client area. For other languages, the style is ignored.</td></tr>
	  <tr><td>WS_EX_LTRREADING</td><td>The window text is displayed using left-to-right reading-order properties. 
	  This is the default (value 0).</td></tr>
	  <tr><td>WS_EX_MDICHILD</td><td>The window is a MDI child window.</td></tr>
	  <tr><td>WS_EX_NOACTIVATE</td><td>A top-level window created with this style does not become the foreground 
	  window when the user clicks it. The system does not bring this window to the foreground when the user 
	  minimizes or closes the foreground window.<br>
      The window should not be activated through programmatic access or via keyboard navigation by accessible 
	  technology, such as Narrator.<br>
      To activate the window, use the SetActiveWindow or SetForegroundWindow function.<br>
      The window does not appear on the taskbar by default. To force the window to appear on the taskbar, 
	  use the WS_EX_APPWINDOW style.</td></tr>
	  <tr><td>WS_EX_NOINHERITLAYOUT</td><td>The window does not pass its window layout to its child windows.</td></tr>
	  <tr><td>WS_EX_NOPARENTNOTIFY</td><td>The child window created with this style does not send the WM_PARENTNOTIFY 
	  message to its parent window when it is created or destroyed.</td></tr>
	  <tr><td>WS_EX_NOREDIRECTIONBITMAP</td><td>The window does not render to a redirection surface. This is for 
	  windows that do not have visible content or that use mechanisms other than surfaces to provide 
	  their visual.</td></tr>
	  <tr><td>WS_EX_OVERLAPPEDWINDOW</td><td>(WS_EX_WINDOWEDGE | WS_EX_CLIENTEDGE)<br>
	  The window is an overlapped window.</td></tr>
	  <tr><td>WS_EX_PALETTEWINDOW</td><td>(WS_EX_WINDOWEDGE | WS_EX_TOOLWINDOW | WS_EX_TOPMOST)<br>
	  The window is palette window, which is a modeless dialog box that presents an array of commands.</td></tr>
	  <tr><td>WS_EX_RIGHT</td><td>The window has generic "right-aligned" properties. This depends on the window class. 
	  This style has an effect only if the shell language is Hebrew, Arabic, or another language that supports 
	  reading-order alignment; otherwise, the style is ignored.<br>
      Using the WS_EX_RIGHT style for static or edit controls has the same effect as using the SS_RIGHT or 
	  ES_RIGHT style, respectively. Using this style with button controls has the same effect as using 
	  BS_RIGHT and BS_RIGHTBUTTON styles. </td></tr>
	  <tr><td>WS_EX_RIGHTSCROLLBAR</td><td>The vertical scroll bar (if present) is to the right of the client area. 
	  This is the default (value 0).</td></tr>
	  <tr><td>WS_EX_RTLREADING</td><td>If the shell language is Hebrew, Arabic, or another language that supports 
	  reading-order alignment, the window text is displayed using right-to-left reading-order properties. 
	  For other languages, the style is ignored.</td></tr>
	  <tr><td>WS_EX_STATICEDGE</td><td>The window has a three-dimensional border style intended to be used for 
	  items that do not accept user input.</td></tr>
	  <tr><td>WS_EX_TOOLWINDOW</td><td>The window is intended to be used as a floating toolbar. A tool window has 
	  a title bar that is shorter than a normal title bar, and the window title is drawn using a smaller font. 
	  A tool window does not appear in the taskbar or in the dialog that appears when the user presses ALT+TAB. 
	  If a tool window has a system menu, its icon is not displayed on the title bar. However, you can display 
	  the system menu by right-clicking or by typing ALT+SPACE. </td></tr>
	  <tr><td>WS_EX_TOPMOST</td><td>The window should be placed above all non-topmost windows and should stay 
	  above them, even when the window is deactivated. To add or remove this style, use the SetWindowPos function.</td></tr>
	  <tr><td>WS_EX_TRANSPARENT</td><td>he window should not be painted until siblings beneath the window 
	  (that were created by the same thread) have been painted. The window appears transparent because the 
	  bits of underlying sibling windows have already been painted.<br>
      To achieve transparency without these restrictions, use the SetWindowRgn function.</td></tr>
	  <tr><td>WS_EX_WINDOWEDGE</td><td>The window has a border with a raised edge.</td></tr>
   </table>
   @return Previous extended style flags on success, otherwise 0.
   @note Flags of value 0 can't be set or reset directly.
   */
   int setExtStyle(int style);

   /** Adds specified extended style(s) to window.
   Does OR combination of <i>style</i> with extended style flags currently set.
   @param style Extended style(s) to add.
   @return Previous extended style flags on success, otherwise 0.
   @see getExtStyle, setExtStyle
   */
   int addExtStyle(int style);

   /** Removes specified extended style(s) from window if currently set.
   @param style Extended style(s) to remove.
   @return Previous extended style flags on success, otherwise 0.
   @see getExtStyle, setExtStyle
   */
   int removeExtStyle(int style);

   /** Checks if window currently minimized.
   @return true (1) if minimized, otherwise false (0).
   */
   int isMinimized() const;

   /** Checks if window currently maximized.
   @return true (1) if maximized, otherwise false (0).
   */
   int isMaximized() const;

   /** Checks if window currently visible.
   @return true (1) if visible, otherwise false (0).
   */
   int isVisible() const;

   /** Minimizes window.
   @return true (1) if window was visible before, otherwise false (0).
   */
   int minimize();

   /** Maximizes window.
   @return true (1) if window was visible before, otherwise false (0).
   */
   int maximize();
 
   /** Sets show state of window.
   @param cmdShow Show state to set. It can be one of following values:
   <table><tr><th>Value</th><th>Meaning</th></tr>
   <tr><td>SW_HIDE</td><td>Hides the window and activates another window.</td></tr>
   <tr><td>SW_SHOWNORMAL<br>SW_NORMAL</td><td>Activates and displays a window. 
   If the window is minimized, maximized, or arranged, the system restores 
   it to its original size and position. An application should specify this 
   flag when displaying the window for the first time.</td></tr>
   <tr><td>SW_SHOWMINIMIZED</td><td>Activates the window and displays it as a minimized window.</td></tr>
   <tr><td>SW_SHOWMAXIMIZED<br>SW_MAXIMIZE</td><td>Activates the window and displays it as a maximized window.</td></tr>
   <tr><td>SW_SHOWNOACTIVATE</td><td>Displays a window in its most recent size and position. 
   This value is similar to SW_SHOWNORMAL, except that the window is not activated.</td></tr>
   <tr><td>SW_SHOW</td><td>Activates the window and displays it in its current size and position.</td></tr>
   <tr><td>SW_MINIMIZE</td><td>Minimizes the window and activates the next top-level window in the Z order.</td></tr>
   <tr><td>SW_SHOWMINNOACTIVE</td><td>isplays the window as a minimized window. This value is similar to 
   SW_SHOWMINIMIZED, except the window is not activated.</td></tr>
   <tr><td>SW_SHOWNA</td><td>Displays the window in its current size and position. 
   This value is similar to SW_SHOW, except that the window is not activated.</td></tr>
   <tr><td>SW_RESTORE</td><td>Activates and displays the window. If the window is minimized, 
   maximized, or arranged, the system restores it to its original size and position. 
   An application should specify this flag when restoring a minimized window.</td></tr>
   <tr><td>SW_SHOWDEFAULT</td><td>Sets the show state based on the SW_ value specified in the 
   STARTUPINFO structure passed to the CreateProcess function by the program that started the application.</td></tr>
   <tr><td>SW_FORCEMINIMIZE</td><td>Minimizes a window, even if the thread that owns the window is not responding. 
   This flag should only be used when minimizing windows from a different thread.</td></tr>
   </table>
   @return true (1) if window was visible before, otherwise false (0).
   */
   int show(int cmdShow = SW_SHOWNORMAL);

   /** Restores default show state of window.
   Does same as show(SW_RESTORE).
   @return true (1) if window was visible before, otherwise false (0).
   */
   int restore();

   /** Hides window.
   Does same as show(SW_HIDE).
   @return true (1) if window was visible before, otherwise false (0).
   */
   int hide();

   /** Message preprocessing function.
   Method is called from implementation of mainLoop for each message addressed to a known window. <br>
   Basic implementation processes accelerator keys for current and its parent windows.
   @param Message to process
   @return 1 (true) if accelerator found and processed, 0 otherwise.
   @note If return value is 1, a new nessage should have been appended to message queue ane mainLoop gets next message. Otherwise further processing of <i>msg</i> id done by mainLoop.<br>
   Derived classes may override this method in order to do any message manipulation before processing.
   */
   int preprocessMessage(MSG msg);
   /** Message handling function.
   Base implementation first calls installed message handlers contained in handler list. If a handler returns 0,
   function returns.  Otherwise it handles following messages:
   <table><tr><th>Message</th><th>Action</th></tr>
   <tr><td>WM_DESTROY</td><td>Deletes background brush (if defined).</td>
   <tr><td>WM_ERASEBKGND</td><td>If beckground brush defined, client area is filled using background brush and returned 0.</td></tr>
   <tr><td>WM_SIZE</td><td>If virtual size enabled, window is scrolled making virtual origin visible in client area.</td></tr>
   <tr><td>WM_HSCROLL<br>WM_VSCROLL</td><td>If virtual size enabled, window is scrolled depending on settings made by setScrollSteps.</td></tr>
   </table>
   Afterwarrds either original window procedure (stored by create or createFromHandle functions) or DefWindowProc is called for further handling.
	@param hwnd Handle of window
	@param msgId ID of message to handle 
	@param wParam Word parameter (meaning depends on <i>msgId</i>)
	@param lParam Long parameter (meaning depends on <i>msgId</i>)
	@return Result of message handling. Value depends on <i>msgId</i>, usually 0 means, message was handled completely.
    @note When method is overridden by a derived class, derived implementation should call base implementation,
    if message handling not completed.
	*/
   intptr handleMessage(intptr hwnd, uint msgId, intptr wParam, intptr lParam);

   /** Adds installable message handler to window.
   @param handler Message handler to install.
   @param atStart If <tt>true</tt> new handler will be inserted  at begin of handler list, otherwise handler
   is appended to end of list.
   @return <tt>true</tt> on success, otherwise (if handler already was installed) <tt>false</tt>.
   @note <ul>
   <li>If <i>handler</i> is <tt>null</tt> or of invalid type, exception of <tt>string</tt> type is raised.</li>
   <li>Same instance of message handler object can be installed on more than one window. Handler will be deleted when
   it is removed from all window instances.</li>
   <li>Message handler is removed from handler list when window is destroyed, i.e. it is not necessary to remove it by call
   to removeMessageHandler.</li>
   </ul>
   */
   bool addMsgHandler(WinMsgHandler handler, bool atStart = false);
   /** Removes specified message handler from window.
   @param handler Message handler to remove from handler list of window.
   @return true on success, i.e. <i>handler</i> was contained in handler list.
   @note If <i>handler</i> is <tt>null</tt> or of invalid type, exception of <tt>string</tt> type is raised.
   */
   int removeMsgHandler(WinMsgHandler handler);
   /// Removes all installable message handlers from window.
   void removeAllMsgHandlers();

   /** Gets array of accelerator definitions.
   @return Array (vector) containing accelerator definitions of type ACCEL.
   @note The returned vector is a reference to (not copy of) internal accelerator definitions of window, i.e.
   client code can modify it.<br>
   If client code modifies the definitions, it must call applyAccelerators to rebuils accelarator table of window.
   */
   vector getAcceleratorArray() const;
   /** Adds accelerator to window.
   Method creates a new ACCEL structure, adds it to internal accelerato definition table and calls applyAccelerators.
   @param key Specifies the accelerator key. This can be either a virtual-key code or a character code.
   @param cmd Specifies the accelerator identifier. This value is placed in the
   low-order word of the wParam parameter of the WM_COMMAND or WM_SYSCOMMAND
   message when the accelerator is pressed.
   @param flags Specifies the accelerator behavior. Value can be one or more of the following:
   <ul>
   <li>FALT: The ALT key must be held down when the accelerator key is pressed.</li>
   <li>FCONTROL: The CTRL key must be held down when the accelerator key is pressed.</li>
   <li>FNOINVERT: Specifies that no top-level menu item is highlighted when the accelerator is used.
   If this flag is not specified, a top-level menu item will be highlighted,
   if possible, when the accelerator is used.<br>
   This attribute is obsolete and retained only for backward compatibility with resource files designed for 16-bit
   Windows.</li>
   <li>FSHIFT: The SHIFT key must be held down when the accelerator key is pressed.</li>
   <li>FVIRTKEY: The key member specifies a virtual-key code. If this flag is not
   specified, key is assumed to specify a character code.</li>
   </ul>
   */
   void addAccelerator(int key, int cmd, int flags = 0);
   /// Applies all accelerator definitions to window.
   void applyAccelerators();

   /** Registers a window in internal window table.
   @return true on success (pwin has a valid handle).
   @note Normally registration is done by create or createFromHandle methods.
   Client code needs to call this function only, if none of create-methods is used. 
   */
   static bool registerWindow(WinBase pwin);

   /** Unregisters a window.
   @param Handle of window to unregister
   @return true on success (<i>hwnd</i> contained in internal window mapping).
   @note Normally unregister is done by destructor, so client code needs not call this method directly.
   */
   static bool unregisterWindow(intptr hwnd);
   /** Find window from window handle.
   @param hwnd Handle to find window object for.
   @return Window object for <i>hwnd</i> on success, otherwise <tt>null</tt>.
   */
   static WinBase findWindow(intptr hwnd);
   /// Forces termination of message loop.
   static void forceQuit();
   /// set automatic quit mode
   static void setAutoQuit(int autoQuit);
   /** Sets applications return code.
   @param Return code. The value is stored in _returnCode class variable and defines result of
   mainLoop function.
   @see mainLoop
   */
   static void setReturnCode(int returnCode);

   /** Sets text of window. 
   * Using of window text depends on windoe type, e.g. used for caption on top window woth frame.
   @param text Text to set for window. Value must be convertible into <tt>string</tt>.
   @return true on success.
   */
   bool setWindowText(var text);
   /// Gets text of window.
   string getWindowText();
   /** Sets color of window background.
   Function creates a solid brush of specified color and calls setBackgroundBrush.
   @param backColor RGB color specification (Windows COLORREF) of new background color.
   @return true on success.
   @see setBackgroundBrush
   */
   bool setBackgroundColor(uint backColor);
   /** Sets brush for drawing window background.
   @param hbrush Handle of windows brush to set for drawing background.
   @return true on success (hbrush was not zero).
   @note <ul><li>Brush set by this function is used by handleMessage method for drawing background when WM_ERASEBACKGROUND
   message occurs. If brush was not set before, handling WM_ERASEBACKGROUND is done by default, depending on window type.</li>
   <li>Brush specified by <i>hbrush</i> will be deleted be destructor or when another brush is set. Client code must
   not call DeleteObject API function on <i>hbrush</i>.</li>
   <li>Setting new background brush invalidates entire client area and updates the window.</li>
   </ul>
   */
   bool setBackgroundBrush(intptr hbrush);
   /** Sets foreground color. 
   Function does nothing but set member variable _foreColor to specified color value.
   Value e.g. may be used by derived classes for drawing text or other elements.
   @param foreColor Color to set.
   @return true on success (<i>foreColor</i> is a valid CPLORREF value).
   */
   bool setForegroundColor(uint foreColor);

   /** Implementation of application main loop.
   Function implements main loop of a Windows application. It normally is called by <tt>main</tt> function
   of B++ application. Function returns when main message loop of application ends.
   @return Return code stored in Winbase::_returnCode
   @par Usage example;
   @code
		int main()
		{
		   // Create main window of appliocation
		   auto win = new MainWin();
		   win->show(SW_NORMAL);
		   // Enter main event loop and return return code from WinBase
		   return WinBase::mainLoop();
		}  
   @endcode
   */
   static int mainLoop();
   /// Sets module instance handle of bp2class library (internal used).
   static void setHInstance(int hInstance);
   /// Gets module instance handle of bp2class library (needed e.g. for resource access).
   static intptr getHInstance();
   /// Helper for ID creation (for internal use)
   static int createID();
   /** Helper for initialization of common controls.
   Function encapsualates call to <a href="https://learn.microsoft.com/en-us/windows/win32/api/commctrl/nf-commctrl-initcommoncontrolsex"
   target="_blank">InitCommonControlsEx</a> API function.
   @param dwICC set of bit flags that indicate which common control classes will be loaded from the Comctl32 DLL. 
   This can be a combination of the following values.
   <table><tr><th>Value</th><th>Meaning</th></tr>
   <tr><td>ICC_ANIMATE_CLASS</td><td>Load animate control class.</td></tr>
   <tr><td>ICC_BAR_CLASSES</td><td>Load toolbar, status bar, trackbar, and tooltip control classes.</td></tr>
   <tr><td>ICC_COOL_CLASSES</td><td>Load rebar control class.</td></tr>
   <tr><td>ICC_DATE_CLASSES</td><td>Load date and time picker control class.</td></tr>
   <tr><td>ICC_HOTKEY_CLASS</td><td>Load hot key control class.</td></tr>
   <tr><td>ICC_INTERNET_CLASSES</td><td>Load IP address class.</td></tr> 
   <tr><td>ICC_LINK_CLASS</td><td>Load a hyperlink control class.</td></tr>
   <tr><td>ICC_LISTVIEW_CLASSES</td><td>Load list-view and header control classes.</td></tr>
   <tr><td>ICC_NATIVEFNTCTL_CLASS</td><td>Load a native font control class.</td></tr>
   <tr><td>ICC_PAGESCROLLER_CLASS</td><td>Load pager control class.</td></tr>
   <tr><td>ICC_PROGRESS_CLASS</td><td>Load progress bar control class.</td></tr>
   <tr><td>ICC_STANDARD_CLASSES</td><td>Load one of the intrinsic User32 control classes. 
   The user controls include button, edit, static, listbox, combobox and scroll bar.</td></tr>
   <tr><td>ICC_TAB_CLASSES</td><td>Load tab and tooltip control classes.</td></tr>
   <tr><td>ICC_TREEVIEW_CLASSES</td><td>Load tree-view and tooltip control classes.</td></tr>
   <tr><td>ICC_UPDOWN_CLASS</td><td>Load up-down control class.</td></tr>
   <tr><td>ICC_USEREX_CLASSES</td><td>Load ComboBoxEx class.</td></tr>
   <tr><td>ICC_WIN95_CLASSES</td><td>Load animate control, header, hot key, list-view, progress bar, 
   status bar, tab, tooltip, toolbar, trackbar, tree-view, and up-down control classes. </td></tr>
   </table>
   @return true on success.
   */
   static bool initCommonControls(uint dwICC);
   /** Allows or disallows virtual size, i.e. use of scrollbars.
   If virtual size is allowed, usable area of window can be larger than client size, i.e. content client area 
   can be scrolled using scrollbars.
   @param enable Determines, if virtual size is allowed (<tt>true</tt> or disallowed (<tt>false</tt>.
   @note If using virtual size is switched to enabled, virtual size is set to same as client size.
   @return true on success
   */
   bool enableVirtualSize(bool enable = true);
   /** Allows or disallows window scrolling while move of scrollbar.
   @param lifeScroll Determines, if life scroll is allowed (<tt>true</tt> or disallowed (<tt>false</tt>.
   @note Function simply sets internal flag, used by handleMesssage for processing scrollbar messages.
   */
   void setLifeScroll(bool lifeScroll = true);
   /** Sets virtual window size using values.
   @param width New virtual width of window.
   @param height New virtual height of window.
   @return true on success.
   @note Virtual window size must be enabled before call this function.
   @see enableVirtualSize
   */
   bool setVirtualSize(int width, int height);
   /** Set virtual window size using SIZE struct.
   @param size New virtual size of window.
   @return true on success.
   @note Virtual window size must be enabled before call this function.
   @see enableVirtualSize
   */
   bool setVirtualSize(const SIZE size);
   /** Gets virtual window size.
   @return Virtual size of window.
   @note If using virtual size disabled, value of result is undefined.
   */
   SIZE getVirtualSize();
   /** Fits virtual window size to window contents.
   Function computes bounding box around all childs windows and adds <i>boederWidth</i> to width and height of computed box,
   sets virtual size to result and rearranges scrollbars.
   @param borderWidth Additional number of pixxels to add to virtual width and height.
   @return true on success,
   @note Use of virtual size must be enabled when call this function.
   @see enableVirtualSize, applyScrollBars
   */
   bool fitVirtualSizeToContents(int borderWidth = 0);
   /** Gets origin position of wirtual area.
   @return Origin position of virtual area relative to origin of client area.
   @note If using virtual size disabled, value of result is undefined.
   @note Use of virtual size must be enabled when call this function.
   @see enableVirtualSize, applyScrollBars
   */
   POINT getVirtualOrigin();

   /** Scrolls virtual area for specified number of pixels in horizontal direction.
   @param offset Number of pixels to scroll. Positive value scrolls to right, negative to left.
   @return Number uf pixels scrolled.
   @note Use of virtual size must be enabled when call this function.
   @see enableVirtualSize, scrollVertical
   */
   int scrollHorizontal(int offset);
   /** Scrolls virtual area for specified number of pixels in vertical direction.
   @param offset Number of pixels to scroll. Positive value scrolls down, negative up.
   @return Number uf pixels scrolled.
   @note Use of virtual size must be enabled when call this function.
   @see enableVirtualSize, scrollHorizontal
   */
   int scrollVertical(int offset);

   /** Defines Scrollbars to be shown.
   @param scrollBarMode Selects scrollbars to be shown. Following values are possible:
   <table><tr><th>Value</th><th>Meaning</th></tr>
   <tr><td>SBM_NONE</td><td>Scrollbars not shown.</td></tr>
   <tr><td>SBM_VERTICAL</td><td>Shows vertical scrollbar.</td></tr>
   <tr><td>SBM_HORIZONTAL</td><td>Shows horizontal scrollbar.</td></tr>
   <tr><td>SBM_BOTH</td><td>Shows both, horizontal and vertical scrollbars.</td></tr>
   <tr><td>SBM_AUTO</td><td>If using virtual size is disabled, it hides scrollbars, otherwise does nothing.</td></tr>
   </table>
   Value is stored in _scrollBarMode member variable and used by applyScrollBars method.
   @return true on success.
   */
   bool setScrollBarMode(uint scrollBarMode);
   /** Defines step width for scrollbars.
   Function sets step widths in horizontal and vertical direction.
   @param sx Horizontal step width.
   @param sy Vertical step width.
   @note Values are used by handleMessage for handling WM_HSCROLL and WM_VSCROLL messages.
   */
   void setScrollSteps(int sx=10, int sy=10);
   /** Gets current scrollbar mode setting.
   @return Current scrollbar mode.
   @see setScrollBarMode
   */
   uint getScrollBarMode() const;

   /** Applies current settings to scrollbars.
   */
   void applyScrollBars();

   /** Starts Paint Operation
   @return Handle of device context on success (as intptr), otherwise 0
   @note Method may be used inside handling of WM_PAINT messages in order to get
   paint DC for window. When painting operation done, client code must call endPaint().
   @see endPaint
   */
   intptr beginPaint();

   /** Ends Paint Operation
   @return Nonzero value on success.
   @note Method is used handling of WM_PAINT messages in order to end operation.
   @see beginPaint
   */
   int endPaint();

protected:
   intptr _hwnd; ///< window handle
   vector _handlerList; ///< list of installed message handlers
   intptr _savedWndProc; ///< original window procedure of associated window

   RECT _virtualRect;
   int _useVirtualSize;
   int _scrollBarMode;
   int _sx;
   int _sy;
   int _lifeScroll;

   uint _foreColor; ///< current foreground color (COLORREF)
   uint _backColor; ///< current background color (COLORREF)
   intptr _backBrush; ///< current background brush handle


   /// array of accelerator definitions
   vector _accelArray;
   /// handle of accelerator table currently in use
   intptr _haccel;
   /// handle of internal PAINTSTRUCT currently in use
   intptr _paintStruct; 
public:
   static int _maxID; ///< maximum ID used (helper for createID)
   /// return code when application exits
   static int _returnCode;
   /// auto quit flag
   static int _autoQuit;
};

#endif // DOXY
/// @}
#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

/** @name WinBaseImpl Windows base class implementation */
//@{

class BppObject;

class Bp2WinMsgHandler
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static int _hdlOwnerIdx;
   static const BppString _className;
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setOwner(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getOwner(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool handleMessage(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool equals(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2WinMsgHandler();
   ~Bp2WinMsgHandler();
};

class Bp2WinBase
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static const BppString _className;
   static int _hwndIdx; ///< window handle
   static int _handlerListIdx; ///< list of installed message handlers
   static int _savedWndProcIdx; ///< original window procedure of associated window

   static int _virtualRectIdx;
   static int _useVirtualSizeIdx;
   static int _scrollBarModeIdx;
   static int _sxIdx;
   static int _syIdx;
   static int _lifeScrollIdx;

   static int _foreColorIdx; ///< current foreground color (COLORREF)
   static int _backColorIdx; ///< current background color (COLORREF)
   static int _backBrushIdx; ///< current background brush handle


   /// array of accelerator definitions
   static int _accelArrayIdx;
   /// handle of accelerator table currently in use
   static int _haccelIdx;
   static int _paintStructIdx; ///< PAINTSTRUCT handle

   static const BppString _maxIDNm; ///< maximum ID used (helper for createID)
  /// return code when application exits
   static const BppString _returnCodeNm;
   /// auto quit flag
   static const BppString _autoQuitNm;

   static void setHInstance(HINSTANCE hInstance);
   static HINSTANCE getHInstance();
   /// global message distribution callback
   static LRESULT CALLBACK globalMsgCallback(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool dtor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool createFromHandle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool create(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHandle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getParentHandle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getParent(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addChild(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getChildrenHandles(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setPosition(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWindowRect(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientRect(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPosition(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientPos(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPosX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPosY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientPosX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientPosY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWidth(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHeight(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientWidth(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getClientHeight(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool hasStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool hasExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeExtStyle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool isMinimized(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool isMaximized(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool isVisible(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool minimize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool maximize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool show(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool restore(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool hide(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool preprocessMessage(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool handleMessage(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addMsgHandler(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeMsgHandler(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool removeAllMsgHandlers(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getAcceleratorArray(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool addAccelerator(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool applyAccelerators(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool registerWindow(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool unregisterWindow(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool findWindow(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool forceQuit(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setAutoQuit(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setReturnCode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setWindowText(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWindowText(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setBackgroundColor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setBackgroundBrush(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setForegroundColor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool mainLoop(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setHInstance(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHInstance(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool createID(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool enableVirtualSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setLifeScroll(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setVirtualSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getVirtualSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool fitVirtualSizeToContents(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getVirtualOrigin(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool scrollHorizontal(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool scrollVertical(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setScrollBarMode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setScrollSteps(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getScrollBarMode(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool applyScrollBars(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool beginPaint(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool endPaint(BppVirtualMachine* vm, int argc, ValPtr& sp);

   static bool initCommonControls(BppVirtualMachine* vm, int argc, ValPtr& sp);

   /// module instance handle of this library
   static HINSTANCE _hInstance;

   static BppObject* _pWinCreate;
};
//@}
/// @}
#endif //!DOXYUSER
/// @}

#endif // WIN32
#endif // !BP2WBASE_H
