#ifndef BP2WDLG_H
#define BP2WDLG_H

#ifdef WIN32

class BppVirtualMachine;
class ValPtr;

#include <bppstrng.h>
#include <windows.h>
#ifdef WINCE
  #include <winuser.h>
#endif
/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2WinDialog Windows Common Dialog Classes
@{
*/

#ifdef DOXY

/** File dialog class.
Class builds wrapper around GetOpenFileName and GetSaveFileName
Common Dialog function calls. Furthermore it encapsulates using of OPENFILENAME structure
and provides some handy functions for work with file paths and patterns.
@note For documentation types of return values are used in signature. 
*/
class FileDialog
{
public:
   /// Constructor
   FileDialog();
   //~FileDialog();
   /** Gets Filter vector.
   @return Filter vector of dialog object.
   @note Return value is a reference to dialogs filter vector (not a copy),
   i.e. modifications of its contents affects dialog behavior.
   @see setFilter
   */
   vector getFilter();
   /** Sets file Filter for dialog.
   @param filter Vector of file filters (type <tt>vector</tt> or object of Vector class).
   Elements of <i>filter</i> are strings, consisting of readable description, followes
   by a pipe symbol (|) and file pattern(s).
   @par Example:
   @code
   var myFilters = ["Text Files (*.txt; *.ini)|*.txt;*.ini","All Files (*.*)|*.*"]; // Vector constructed by literal
   var dlg = new FileDialog();
   dlg->setFilter(myFilters);
   // // ...  more code here
   @endcode
   @note Empty filter vector means "no filter", i.e. same as pattern *.* .<br>
   No copy of <i>filter</i> is made while assignment, i.e. modifications
   of its contents after setFilter call may affect behavior of file dialog too.
   */
   null setFilter(var filter);
   /** Gets directory part of file path.
   @return Directory path. Path may be absolute or relative. If path
   is not empty nor ends with colon (: , i.e. contains drive specification
   only), last character is a backslash (\).
   @note Return value is a copy of _dir member value.
   @see setDirectory
   */
   string getDirectory();
   /** Sets directory part of file path.
   @param dir Directory path value to set (must be of string type or an object convertible into string).
   @note Function makes a copy of <i>dir</i> value and normalizes it, i.e. slashes are
   replaced by backslashes. If <i>dir</i> is not empty nor ends with colon,
   last character is a backslash (\).
   @see getDirectory
   */
   null setDirectory(var dir);
   /** Gets name part of file path (without extension).
   @return File name (copy of _fname member value).
   @see setFileName
   */
   string getFileName();
   /** Sets name part of file path (without extension).
   @param name File name (must be of string type or an object convertible into string).
   @note Function makes a copy of <i>name</i> value and assigns it to _fname member without
   any modifications.
   @see getFileName
   */
   null setFileName(var name);
   /** Gets name part of file path including extension.
   @return File name including extension.
   @see setFullName
   */
   string getFullName();
    /** Sets name part of file path including extension.
   @param name File name including extension (must be of string type or an object convertible into string).
   @note Function sets _fname and _ext members. Extension is the part behind last dot (if any)
   in <i>name</i>.
   @see getFullName
   */
   null setFullName(var name);
   /** Gets extension part of file path.
   @return Copy of _ext member value.
   */
   string getExtension();
   /** Sets extension part of file path.
   @param ext Extension (must be of string type or an object convertible into string).
   @see getExtension
   */
   null setExtension(var ext);
   /** Gets fully qualified file path.
   @return File path. Path may be absolute or relative..
   @note Return value is a concatenation  of _dir, _fname and _ext member values.
   @see setFullPath
   */
   string getFullPath();
   /** Sets fully qualified of file path.
   @param path File Path (must be of string type or an object convertible into string).
   @note Function parses <i>path</i> value and sets _dir, _fname and _ext members.
   @see getFullPath
   */
   null setFullPath(var path);
   /** Shows dialog in Open mode.
   @param parent Handle of parent window (long) or parent window (derived from WinBase). If
   <i>parent</i> is null (or omitted) dialog is shown without owner window.
   @return TRUE if dialog was closed with OK, FALSE otherwise.
   @note If TRUE was returned, selected file name is stored in object, i.e.
   _dir, _fname and _ext member are set. Client code may use getXXX functions
   to retrieve values needed.
   @see showSave
   */
   long showOpen(var parent = null);
   /** Shows dialog in Save mode.
   @param parent Handle of parent window (long) or parent window (derived from WinBase). If
   <i>parent</i> is null (or omitted) dialog is shown without owner window.
   @return TRUE if dialog was closed with OK, FALSE otherwise.
   @note If TRUE was returned, selected file name is stored in object, i.e.
   _dir, _fname and _ext member are set. Client code may use getXXX functions
   to retrieve values needed.
   @see showSave
   */
   long showSave(var parent = null);
   /** Gets dialog title.
   @return Copy of _title member value.
   @see setTitle
   */
   string getTitle();
   /** Sets dialog title.
   @param title Value to use for dialog title (caption), must be of string type or an object convertible into string.
   @note Function makes a copy of <i>title</i> value and assigs it to _title member.
   @see getTitle
   */
  null setTitle(var title);

private:
   var _filter; ///< filter array
   var _dir; ///< directory path
   var _fname; ///< file name without extension
   var _ext; ///< extension
   var _title ///< file dialog title
};
#endif // DOXY
/// @}
#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

class Bp2FileDialog
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static const BppString _className;
   static int _filterIdx; ///< filter array
   static int _dirIdx; ///< directory
   static int _fnameIdx; ///< file name without extension
   static int _extIdx; ///< extension
   static int _titleIdx; ///< dialog title
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getFilter(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setFilter(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getDirectory(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setDirectory(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getFileName(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setFileName(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getFullName(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setFullName(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getExtension(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setExtension(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getFullPath(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setFullPath(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool showOpen(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool showSave(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getTitle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setTitle(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2FileDialog();
   ~Bp2FileDialog();
};
/// @}
#endif //!DOXYUSER
/// @}

#endif // WIN32
#endif // BP2WDLG_H
