#ifndef BP2WSTRUC_H
#define BP2WSTRUC_H

#ifdef WIN32

class BppVirtualMachine;
class ValPtr;
#include <bppstrng.h>

/** @addtogroup Bp2ClassLib
@{
*/

/** @addtogroup Bp2WinHelper Common Windows Structures
@{
*/

#ifdef DOXY
/** POINT class.
This class encapsulates a 2d-POINT structure of Windows API.
*/
class POINT
{
public:
   /** Constructor, optionally with initialization of coordinates.
   @param x x-coordinate
   @param y y-coordinate
   */
   POINT(int x=-1, int y = -1);
   /** Construction as a copy of another POINT.
   @param cpy POINT to get coordinates from.
   */
   POINT(POINT cpy);
   /** Gets x-coordinate.
   @return x-coordinate
   */
   int getX();
   /** Sets x-coordinate.
   @param x New value of x-coordinate.
   */
   null setX(int x);
   /** Gets y-coordinate.
   @return y-coordinate
   */
   int getY();
   /** Sets y-coordinate.
   @param y New value of y-coordinate.
   */
   null setY(int y);
   /** Addition operator implementation.
   @param p Right operand for addition.
   @return Result of addition as new object.
	@note Because argument always is of type POINT, we need not implement right version of operator.
	*/
   POINT operator+(POINT p);
   /** Subtraction operator implementation.
   @param p Right operand for subtraction.
   @return Result of subtraction as new object.
	@note Because argument always is of type POINT, we need not implement right version of operator.
   */
   POINT operator-(POINT p);
   /** Checks equality.
   @param p POINT to check equality with.
   @return 1 on equality, otherwise 0
	*/
   int equals(POINT p);
   /** Gets pointer to Win-API POINT-structure.
   @return Address of POINT structure. The result may be passed directly to Windows-API functions.
   */
   int ptr();
protected:
   /// internal data buffer
   buffer _pt;
};

/** SIZE class.
This class encapsulates a SIZE structure of Windows API. SIZE ist mostly same as
POINT. However it differs from POINT in interpretation of coordinates. In SIZE
structure x-coordinate means width while y-coordinate means height. Furthermore
a SIZE structure is invalid if at least one of with or height is negative or zero.
*/
class SIZE : POINT
{
public:
   /** Constructor, optionally with initialization.
   @param w width
   @param h height
   */
   SIZE(int w=-1, int h = -1);
   /** Construction as a copy of another SIZE- or POINT-object.
   @param cpy object to get coordinates from.
   */
   SIZE(POINT cpy);
   /** Gets width.
   @return width
   */
   int getW();
   /** Sets width.
   @param w New width.
   */
   null setW(int w);
   /** Gets height.
   @return height
   */
   int getH();
   /** Sets height.
   @param h New height.
   */
   null setH(int h);
   /** Checks for valid size.
   @return 1 if valid, otherwise 0.
   */
   int ok();
};

/** RECT class.
This class encapsulates a RECT structure of Windows API. 
*/
class RECT
{
public:
   /** Constructor, optionally with initialization.
   @param l coordinate of left border
   @param t coordinate of top border
   @param r coordinate of right border
   @param b coordinate of bottom border
   */
   RECT(int l=-1, int t = -1, int r = -1, int b=-1);
   /** Construction as a copy of another RECT object.
   @param cpy object to get coordinates from.
   */
   RECT(RECT cpy);
   /** Gets left border coordinate.
   @return coordinate of left border.
   */
   int getLeft();
   /** Sets left border coordinate.
   @param l new coordinate of left border.
   */
   null setLeft(int l);
   /** Gets top border coordinate.
   @return coordinate of top border.
   */
   int getTop();
   /** Sets top border coordinate.
   @param t new coordinate of top border.
   */
   null setTop(int t);
   /** Gets right border coordinate.
   @return coordinate of right border.
   */
   int getRight();
   /** Sets right border coordinate.
   @param r new coordinate of right border.
   */
   null setRight(int r);
   /** Gets bottom border coordinate.
   @return coordinate of bottom border.
   */
   int getBottom();
   /** Sets bottom border coordinate.
   @param b new coordinate of bottom border.
   */
   null setBottom(int b);
   /** Gets position of rectangles upper left corner.
   @return POINT containing upper left corner coordinates.
   */
   POINT getPos();
   /** Moves rectangle to specified position.
   @param pos New position of upper left corner.
   */
   null setPos(POINT pos);
   /** Gets current size of rectangle.
   @return SIZE object containing width and height of rectangle.
   */
   SIZE getSize();
   /** Sets size of rectangle.
   @param size New size of rectangle.
   @note Resizing is done by modifying bottom right corner of rectangle.
   */
   null setSize(SIZE size);
   /** Checks equality.
   @param r RECT to check equality with.
   @return 1 on equality, otherwise 0
   */
   int equals(RECT r);
   /** Checks for valid rectangle.
   @return 1 if valid, otherwise 0.
   @note A rectangle is valid if its size in both, horizontal and vertical direction, is
   greater than zero.
   */
   int ok();
   /** Gets pointer to Win-API RECT-structure.
   @return Address of RECT structure. The result may be passed directly to Windows-API functions.
   */
   int ptr();
private:
   /// internal data buffer
   buffer _bounds;
};

/** MSG class.
This class encapsulates a Windows Message structure. 
*/
class MSG
{
public:
   /** Default ctor.
   Constructor initializes all members to zero.
   */
   MSG();
   /** Construction as a copy of another MSG object.
   @param cpy object to copy from.
   */
   MSG(const MSG cpy);
   /** Sets window handle field.
   @param hwnd window handle (0 for none)
   */
   null setHwnd(intptr hwnd);
   /** Gets window handle field value.
   @return window handle (0 for none)
   */
   intptr getHwnd() const;
   /** Sets message ID field.
   @param msgId message ID
   */
   null setMsgId(uint msgId);
   /** Gets message ID field value.
   @return message ID
   */
   uint getMsgId() const;
   /** Sets WPARAM field of message.
   @param wParam value to set
   */
   null setWParam(uintptr wParam);
   /** Gets WPARAM field value.
   @return value of WPARAM field
   */
   uintptr getWParam();
   /** Sets LPARAM field of message.
   @param lParam value to set
   */
   null setLParam(intptr lParam);
   /** Gets LPARAM field value.
   @return value of LPARAM field
   */
   intptr getLParam() const;
   /** Sets time field of message.
   @param time value to set
   @note If Windows creates messages, it uses milliseconds elapsed since last reboot
   for time value. To do the same in your own massages, use GetTickCount API function
   to set proper value.
   */
   null setTime(int time);
   /** Gets time field value.
   @return value of time field
   */
   int getTime() const;
   /** Sets pt field of message.
   @param pt value to set
   @note When creating your own message, value should be set to cursor position (mouse position)
   in screen coordinates.
   */
   null setPoint(const POINT pt);
   /** Sets pt field of message using separate coordinates.
   @param x x-coordinate of pt
   @param y y-coordinate of pt
   */
   null setPoint(int x, int y);
   /** Gets pt field value as new POINT object.
   @return value of pt field
   */
   POINT getPoint() const;
   /** Gets x-position.
   @return x-coordinate of pt field
   */
   int getX() const;
   /** Sets x-position.
   @param x new x-coordinate of pt field
   */
   null setX(int x);
   /** Gets y-position.
   @return y-coordinate of pt field
   */
   int getY() const;
   /** Sets y-position.
   @param y new y-coordinate of pt field
   */
   null setY(int y);
   /** Gets pointer to Win-API MSG-structure.
   @return Address of MSG structure. The result may be passed directly to Windows-API functions.
   */
   intptr ptr() const;
private:
   /// internal data buffer
   buffer _msg;
};

/** MSG class.
Class encapsulates a Windows ACCEL structure. 
*/
class ACCEL
{
public:
   /** Default constructor.
   Ctor initializes all attributes to zero. Client code may use
   setXXX methods to assign values.
   */
   ACCEL();
   /** Initialized constructor.
   Ctor initializes attributes to specified values.
   @param key Specifies the accelerator key. This member can be either a virtual-key code or a character code. 
   @param cmd Specifies the accelerator identifier. This value is placed in the 
   low-order word of the wParam parameter of the WM_COMMAND or WM_SYSCOMMAND 
   message when the accelerator is pressed.
   @param flags Specifies the accelerator behavior. Value can be one or more of the following:
   <ul>
   <li>FALT: The ALT key must be held down when the accelerator key is pressed.</li>
   <li>FCONTROL: The CTRL key must be held down when the accelerator key is pressed.</li>
   <li>FNOINVERT: Specifies that no top-level menu item is highlighted when the accelerator is used. 
   If this flag is not specified, a top-level menu item will be highlighted, 
   if possible, when the accelerator is used.<br>
   This attribute is obsolete and retained only for backward compatibility with resource files designed for 16-bit 
   Windows.</li> 
   <li>FSHIFT: The SHIFT key must be held down when the accelerator key is pressed.</li>
   <li>FVIRTKEY: The key member specifies a virtual-key code. If this flag is not 
   specified, key is assumed to specify a character code.</li>
   </ul>
   */
   ACCEL(int key, int cmd, int flags = 0);
   /** Gets accelerator behavior flags.
   @return behavior flags
   @see setFVirt
   */
   int getFVirt();
   /** Sets accelerator behavior flags.
   @param fVirt Specifies the accelerator behavior. Value can be one or more of the following:
   <ul>
   <li>FALT: The ALT key must be held down when the accelerator key is pressed.</li>
   <li>FCONTROL: The CTRL key must be held down when the accelerator key is pressed.</li>
   <li>FNOINVERT: Specifies that no top-level menu item is highlighted when the accelerator is used. 
   If this flag is not specified, a top-level menu item will be highlighted, 
   if possible, when the accelerator is used.<br>
   This attribute is obsolete and retained only for backward compatibility with resource files designed for 16-bit 
   Windows.</li> 
   <li>FSHIFT: The SHIFT key must be held down when the accelerator key is pressed.</li>
   <li>FVIRTKEY: The key member specifies a virtual-key code. If this flag is not 
   specified, key is assumed to specify a character code.</li>
   </ul>
   @see getFVirt
   */
   null setFVirt(int fVirt);
   /** Gets accelerator key code.
   @return key code
   @see setKey
   */
   int getKey();
   /** Sets accelerator key code.
   @param key Specifies the accelerator key. This member can be either a virtual-key code or a character code. 
   @see getKey
   */
   null setKey(int key);
   /** Gets accelerator message identifier.
   @return message identifier
   @see setCmd
   */
   int getCmd();
   /** Sets accelerator message identifier.
   @return message identifier
   @param cmd Specifies the accelerator identifier. This value is placed in the 
   low-order word of the wParam parameter of the WM_COMMAND or WM_SYSCOMMAND 
   message when the accelerator is pressed.
   @see getCmd
   */
   null setCmd();
   /** Gets pointer to Win-API ACCEL-structure.
   @return Address of ACCEL structure. The result may be passed directly to Windows-API functions.
   */
   int ptr();
private:
   /// internal data buffer
   buffer _accel;
}
/// @}


#endif // DOXY
#ifndef DOXYUSER
/** @addtogroup Bp2Impl Implementation
@{
*/

class Bp2Size;
/** @name POINT implementation class */
//@{

class Bp2Point
{
friend class Bp2Size;
public:
   static void CreateClass(BppVirtualMachine& vm);
   static int _ptIdx;
   static const BppString _className;
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool OP_ADD(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool OP_SUB(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool equals(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool ptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2Point();
   ~Bp2Point();
};
//@}

/** @name SIZE implementation class */
//@{

class Bp2Size
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static const BppString _className;
private:
   //bool Bp2Size_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   //bool Bp2Size_getW(BppVirtualMachine* vm, int argc, ValPtr& sp);
   //bool Bp2Size_setW(BppVirtualMachine* vm, int argc, ValPtr& sp);
   //bool Bp2Size_getH(BppVirtualMachine* vm, int argc, ValPtr& sp);
   //bool Bp2Size_setH(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool ok(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2Size();
   ~Bp2Size();
};
//@}

/** @name RECT implementation class */
//@{

class Bp2Rect
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static int _rectIdx;
   static const BppString _className;
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getLeft(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setLeft(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getTop(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setTop(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getRight(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setRight(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getBottom(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setBottom(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPos(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setPos(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool ok(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool equals(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool ptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2Rect();
   ~Bp2Rect();
};
//@}

/** @name MSG implementation class */
//@{

class Bp2Msg
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static int _msgIdx;
   static const BppString _className;
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setHwnd(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getHwnd(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setMsgId(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getMsgId(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setWParam(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getWParam(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setLParam(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getLParam(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setTime(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getTime(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setPoint(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getPoint(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getX(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getY(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool ptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2Msg();
   ~Bp2Msg();
};
//@}

/** @name ACCEL implementation class */
//@{

class Bp2Accel
{
public:
   static void CreateClass(BppVirtualMachine& vm);
   static int _accelIdx;
   static const BppString _className;
private:
   static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getFVirt(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setFVirt(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getKey(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setKey(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool getCmd(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool setCmd(BppVirtualMachine* vm, int argc, ValPtr& sp);
   static bool ptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
   Bp2Accel();
   ~Bp2Accel();
};
//@}
/// @}
#endif //!DOXYUSER

/// @}

#endif // WIN32
#endif // !BP2WSTRUC_H
