#include "bp2dict.h"

#include <bppvm.h>
#include <bppclass.h>
#include <bppvalue.h>
#include <bppdict.h>
#include <bppvec.h>
#include <bppstrng.h>
#include <bppobjct.h>
#include <bppint.h>
#include <bppprint.h>
#include <bppbuf.h>

#include "bp2cldef.h"

static unsigned _currentIndex = 0;
static unsigned _dictIndex = 0;

/** Internal helper for dictionary iterator creation.
*/
static BppObject* _createDictIterator(BppVirtualMachine* vm, BppDictionaryEntry* entry, bool reverse)
{
	static const BppString key = _T("DictIterator");
	static const BppString rkey = _T("DictRIterator");
	BppDictionaryEntry* clentry = vm->classes().find(reverse ? rkey : key);
	if (clentry == 0) return 0;
	BppClass* cl = clentry->getValue();
	if (cl == 0) return 0;
	BppObject* result = new BppObject(cl);
	if (result != 0)
	{
		BppValue& v = (*result->getMemberData())[_currentIndex];
		if (entry != 0)
			v = entry;
		else
			v.dispose();
	}
	return result;
}

bool Bp2DictIterator_OP_INC(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var = var->getNext();
	if (var)
		varVal = var;
	else
		varVal.dispose();
	*bp = pThis;
	sp = bp; 
	return true;
}

bool Bp2DictIterator_OP_PINC(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var = var->getNext();
	BppObject* result = _createDictIterator(vm,var,false);
	if (result == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = result;
	sp = bp; 
	return true;
}

bool Bp2DictIterator_OP_DEC(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var = var->getBefore();
	if (var)
		varVal = var;
	else
		varVal.dispose();
	*bp = pThis;
	sp = bp;
	return true;
}

bool Bp2DictIterator_OP_PDEC(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var = var->getBefore();
	BppObject* result = _createDictIterator(vm, var,false);
	if (result == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = result;
	sp = bp;
	return true;
}

bool Bp2DictIterator_next(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return Bp2DictIterator_OP_INC(vm,argc,sp);
}

bool Bp2DictIterator_before(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	return Bp2DictIterator_OP_DEC(vm, argc, sp);
}

bool Bp2DictIterator_equals(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2,2);
	BppObject* cmp = bp[2];
	if (cmp == 0)
		return interpreter->badType(bp[2].getType(),BVT_OBJECT);
	if (!cmp->getClass()->isSubtypeOf(pThis->getClass()))
		THROW(_T("Comparison is not a dictionary iterator."),_T("DictIterator::equals"));
	BppValue& varVal = MEMBER(_currentIndex);
	BppValue& cmpVal = (*(cmp->getMemberData()))[_currentIndex];
	long equ = 0;  
	BppValueType tval = varVal.getType();
	if (tval == cmpVal.getType())
	{
		if (tval == VT_NULL) // both are end iterators
			equ = 1;
		else // both iterators point to dictionary entries
		{
			if (((BppDictionaryEntry*)varVal) == ((BppDictionaryEntry*)cmpVal))
				equ = 1;
		}
	}
	*bp = equ;
	sp = bp;
	return true;
}

bool Bp2DictIterator_getKey(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	*bp = new BppString(var->getKey()->c_str());
	sp = bp;
	return true;
}

bool Bp2DictIterator_getValue(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	*bp = var->getValue();
	sp = bp;
	return true;
}

bool Bp2DictIterator_setValue(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2,2);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var->setValue(bp[2]);
	bp->dispose();
	sp = bp;
	return true;
}


void Bp2DictIterator_CreateClass(BppVirtualMachine& vm)
{
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(_T("DictIterator"),pClass)) return;
	pClass->addDataMember(_T("_current"));
	pClass->addMethod(_T("OP_INC"),&Bp2DictIterator_OP_INC);
	pClass->addMethod(_T("OP_PINC"),&Bp2DictIterator_OP_PINC);
	pClass->addMethod(_T("OP_DEC"), &Bp2DictIterator_OP_DEC);
	pClass->addMethod(_T("OP_PDEC"), &Bp2DictIterator_OP_PDEC);
	pClass->addMethod(_T("next"),&Bp2DictIterator_next);
	pClass->addMethod(_T("before"), &Bp2DictIterator_before);
	pClass->addMethod(_T("equals"),&Bp2DictIterator_equals);
	pClass->addMethod(_T("getKey"),&Bp2DictIterator_getKey);
	pClass->addMethod(_T("getValue"),&Bp2DictIterator_getValue);
	pClass->addMethod(_T("setValue"),&Bp2DictIterator_setValue);
	pClass->fixupDataDefs();
	_currentIndex = pClass->getMemberIndex(_T("_current"));
}

bool Bp2DictRIterator_OP_PINC(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var = var->getBefore();
	BppObject* result = _createDictIterator(vm, var, true);
	if (result == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = result;
	sp = bp;
	return true;
}

bool Bp2DictRIterator_OP_PDEC(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppValue& varVal = MEMBER(_currentIndex);
	BppDictionaryEntry* var = varVal;
	if (var == 0)
		THROW(_T("Invalid iterator."));
	var = var->getNext();
	BppObject* result = _createDictIterator(vm, var, true);
	if (result == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = result;
	sp = bp;
	return true;
}

void Bp2DictRIterator_CreateClass(BppVirtualMachine & vm)
{
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(_T("DictRIterator"), pClass)) return;
	pClass->addDataMember(_T("_current"));
	pClass->addMethod(_T("OP_INC"), &Bp2DictIterator_OP_DEC);
	pClass->addMethod(_T("OP_PINC"), &Bp2DictRIterator_OP_PINC);
	pClass->addMethod(_T("OP_DEC"), &Bp2DictIterator_OP_INC);
	pClass->addMethod(_T("OP_PDEC"), &Bp2DictRIterator_OP_PDEC);
	pClass->addMethod(_T("next"), &Bp2DictIterator_next);
	pClass->addMethod(_T("before"), &Bp2DictIterator_before);
	pClass->addMethod(_T("equals"), &Bp2DictIterator_equals);
	pClass->addMethod(_T("getKey"), &Bp2DictIterator_getKey);
	pClass->addMethod(_T("getValue"), &Bp2DictIterator_getValue);
	pClass->addMethod(_T("setValue"), &Bp2DictIterator_setValue);
	pClass->fixupDataDefs();
	_currentIndex = pClass->getMemberIndex(_T("_current"));
}


bool Bp2Dictionary_ctor(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1,2);
	BppValue& dictVal = MEMBER(_dictIndex);
	if (argc == 1)
	{
		dictVal = new BppDictionary();
	}
	else
	{
		BppValue& assignment = bp[2];
		BppDictionary* pdict = assignment;
		if (pdict != 0)
		{
			BppDictionary* newdict = new BppDictionary();
			dictVal = newdict;
			for (BppDictionaryEntry* p=pdict->getFirst();p!=0;p=p->getNext())
				(*newdict)[*(p->getKey())] = p->getValue();
		}
		else
		{
			BppObject* po = assignment;
			if (po != 0)
			{
				BppClass* pc = po->getClass();
				if (!pc->isSubtypeOf(pThis->getClass()))
					THROW(_T("Argument is not of class type Dictionary."),pc->getClassName()->c_str());
				pdict = (*(po->getMemberData()))[_dictIndex];
				if (pdict == 0)
					THROW(_T("Structure of argument damaged."));
				BppDictionary* newdict = new BppDictionary();
				dictVal = newdict;
				for (BppDictionaryEntry* p=pdict->getFirst();p!=0;p=p->getNext())
					(*newdict)[*(p->getKey())] = p->getValue();
			}
			else
				return interpreter->badType(assignment.getType(),BVT_DICTIONARY);
		}
	}
	*bp = pThis;
	sp = bp;
	return true;
}

bool Bp2Dictionary_OP_VREF(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2,2);
	BppDictionary* pdict = MEMBER(_dictIndex);
	BppString* key = interpreter->getStringFromValueExtern(bp[2]);
	if (key == 0)
		return interpreter->badType(bp[2].getType(),BVT_STRING);
	BppDictionaryEntry& p = (*pdict)[*key];
	*bp = p.getValue();
	sp = bp;
	return true;
}

bool Bp2Dictionary_OP_VSET(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3,3);
	BppDictionary* pdict = MEMBER(_dictIndex);
	BppString* key = interpreter->getStringFromValueExtern(bp[2]);
	if (key == 0)
		return interpreter->badType(bp[2].getType(),BVT_STRING);
	BppDictionaryEntry& p = (*pdict)[*key];
	p.setValue(bp[3]);
	*bp = bp[3];
	sp = bp;
	return true;
}

bool Bp2Dictionary_begin(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppDictionary* pdict = MEMBER(_dictIndex);
	BppObject* it = _createDictIterator(vm,pdict->getFirst(),false);
	if (it == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = it;
	sp = bp;
	return true;
}

bool Bp2Dictionary_rbegin(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppDictionary* pdict = MEMBER(_dictIndex);
	BppObject* it = _createDictIterator(vm, pdict->getLast(), true);
	if (it == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = it;
	sp = bp;
	return true;
}


bool Bp2Dictionary_end(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppObject* it = _createDictIterator(vm,0,false);
	if (it == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = it;
	sp = bp;
	return true;
}

bool Bp2Dictionary_rend(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppObject* it = _createDictIterator(vm, 0, true);
	if (it == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = it;
	sp = bp;
	return true;
}

bool Bp2Dictionary_find(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2,2);
	BppDictionary* pdict = MEMBER(_dictIndex);
	BppString* key = interpreter->getStringFromValueExtern(bp[2]);
	if (key == 0)
		return interpreter->badType(bp[2].getType(),BVT_STRING);
	BppObject* it = _createDictIterator(vm,pdict->find(*key),false);
	if (it == 0)
		THROW(_T("Creation of dictionary iterator failed"));
	*bp = it;
	sp = bp;
	return true;
}

bool Bp2Dictionary_size(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppDictionary* pdict = MEMBER(_dictIndex);
	*bp = (BppUInt) pdict->count();
	sp = bp;
	return true;
}

bool Bp2Dictionary_clear(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppDictionary* pdict = MEMBER(_dictIndex);
	pdict->clear();
	bp->dispose();
	sp = bp;
	return true;
}

bool Bp2Dictionary_erase(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2,2);
	BppDictionary* pdict = MEMBER(_dictIndex);
	BppValue& val = bp[2];
	BppObject* po = val;
	long result = -1;
	if (po != 0)
	{
		if (po->getClass()->isSubtypeOf(_T("DictIterator")))
		{

			BppValue& peval = (*(po->getMemberData()))[_currentIndex];
			BppDictionaryEntry* pe = peval;
			if (pe != 0)
			{
				BppDictionaryEntry* pnext = pe->getNext();
				if (pdict->erase(pe))
				{
					if (pnext) peval = pnext;
					else peval.dispose();
					result = 1;
				}
				else
					result = 0;
			}
		}
	}
	if (result < 0) // arg was not an iterator
	{
		BppString* key = interpreter->getStringFromValueExtern(val);
		if (key == 0)
			return interpreter->badType(bp[2].getType(),BVT_STRING);
		result = (pdict->erase(*key)) ? 1 : 0;
	}
	*bp = BppValue::makeBool(result != 0);
	sp = bp;
	return true;
}

bool Bp2Dictionary_dict(BppVirtualMachine* vm, int argc, ValPtr& sp)
{
	CHKVM;
	CHKOBJECT;
	ARGCMAX(1);
	BppDictionary* pdict = MEMBER(_dictIndex);
	*bp= pdict;
	sp = bp;
	return true;
}

void Bp2Dictionary_CreateClass(BppVirtualMachine& vm)
{
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(_T("Dictionary"),pClass)) return;
	pClass->addDataMember(_T("_dict"));
	pClass->addMethod(_T("Dictionary"),&Bp2Dictionary_ctor);
	pClass->addMethod(_T("OP_VREF"),&Bp2Dictionary_OP_VREF);
	pClass->addMethod(_T("OP_VSET"),&Bp2Dictionary_OP_VSET);
	pClass->addMethod(_T("begin"),&Bp2Dictionary_begin);
	pClass->addMethod(_T("rbegin"), &Bp2Dictionary_rbegin);
	pClass->addMethod(_T("end"),&Bp2Dictionary_end);
	pClass->addMethod(_T("rend"), &Bp2Dictionary_end);
	pClass->addMethod(_T("find"),&Bp2Dictionary_find);
	pClass->addMethod(_T("size"),&Bp2Dictionary_size);
	pClass->addMethod(_T("clear"),&Bp2Dictionary_clear);
	pClass->addMethod(_T("erase"),&Bp2Dictionary_erase);
	pClass->addMethod(_T("dict"),&Bp2Dictionary_dict);
	pClass->addMethod(_T("dictionary"),&Bp2Dictionary_dict);
pClass->fixupDataDefs();
	_dictIndex = pClass->getMemberIndex(_T("_dict"));
}

