#ifndef BPPCLASS_H
#define BPPCLASS_H

#include <bppvec.h>

class BppDictionary;
class BppDictionaryEntry;
class BppDictionaryIO;
class BppString;
class BppVirtualMachine;
class BppObject;


/** Class for internal representation of a B++ class.
Please note that in B++ a class always lives in context of a virtual machine.
In order to ensure VM binding, this class has no default constructor.
*/
class BPPLIB_DLLEXP BppClass : public BppRefObject
{
public:
	/// uninitialized ctor for multi-step construction
	BppClass(BppVirtualMachine& vm);
	/// single-step construction from name pointer
	BppClass(BppVirtualMachine& vm,BppString* name, BppClass* baseClass = 0, BppClass* ownerClass = 0);
	/// single-step construction from name reference
	BppClass(BppVirtualMachine& vm,const BppString& name, BppClass* baseClass = 0, BppClass* ownerClass = 0);
	/// dtor
	virtual ~BppClass();
	/// access to data member dictionary
private:
	BppDictionary& dataMembers() BPPNOEXCEPT;
public:
	/// access to function member dictionary
	BppDictionary& funcMembers() BPPNOEXCEPT;
	/// get pointer to base class descriptor
	BppClass* getBaseClass() const BPPNOEXCEPT;
	/// set base class
	bool setBaseClass(BppClass* baseClass) BPPNOEXCEPT;
	/// get pointer to owner class descriptor
	BppClass* getOwnerClass() const BPPNOEXCEPT;
	/// set owner class
	bool setOwnerClass(BppClass* ownerClass) BPPNOEXCEPT;
	/// check for inheritance from specified class
	bool isSubtypeOf(BppClass* cls) const BPPNOEXCEPT;
	/// check for inheritance from specified class name
	bool isSubtypeOf(const BppString& className) const BPPNOEXCEPT;

	/** Gets local name of class.
	@return name of class (or NULL if not set)
	*/
	BppString* getClassName() const BPPNOEXCEPT
	{
		return _className;
	}
	/// Gets fully qualified classname path.
	BppString getClassPath() const;
	/// set name of class from string pointer
	bool setClassName(BppString* className) BPPNOEXCEPT;
	/// set name of class from constant string
	bool setClassName(const BppString& className);
	/// Compute number of non static data members of a single object instance of this class.
	int computeDataSize() BPPNOEXCEPT;
	/// Gets number of non static data members of a single object instance of this class without new computation.
	int getDataSize() const BPPNOEXCEPT;
	/// search for a class member by name
	BppDictionaryEntry* findMember(const BppClass* contextClass, const BppString& name,  bool includeBase = true, bool requirePublic = false) const BPPNOEXCEPT;
	/// search for a class member by name assuming own context
	BppDictionaryEntry* findOwnMember(const BppString& name, bool includeBase = true) const BPPNOEXCEPT
	{
		return findMember(this, name, includeBase);
	}
	/// search for a class member by name
	BppDictionaryEntry* findMemberPL(unsigned short rprotLevel, const BppString& name, bool includeBase = true) const BPPNOEXCEPT;
	/// search for a data member name by index
	const BppChar* findDataMemberFromIndex(unsigned idx, bool includeBase = true) const;
	/// send delete notification to specified object
	void notifyObjDelete(BppObject* obj);
	/// Notification about creatiom of specified object
	void notifyObjCreation(BppObject* obj, bool copied);
	/// Notification about serialization od deserialization of specified object
	void notifyObjSerialization(BppObject* obj, bool serialize);
	/// returns clone of vector containing default values of data members
	const BppVector& getDefaultDataMembers();
	/// gets vector of data member names
	BppVector getDataMemberNames() const;
// native extension helper

/** @name Native extension helpers.
Functions in this section are intended do provide an easy to use interface
for creating natively implemented classes.
 */
//@{
	/// add an initialized named data member to class
	bool addDataMember(const BppChar* name);
	/// add a named data member to class
	bool addDataMember(const BppChar* name, const BppValue& initValue);
	/// add or redefine a named data member of class
	bool defineDataMember(const BppChar* name, const BppValue& initValue);
	/// add an uninitialized static data member to class
	bool addStaticDataMember(const BppChar* name);
	/// add an initialized static data member to class
	bool addStaticDataMember(const BppChar* name, const BppValue& initValue);
	/// add or redefine an initialized static data member of class
	bool defineStaticDataMember(const BppChar* name, const BppValue& initValue);
	/// do indexing of data members and compute size needed for an instance of class
	void fixupDataDefs();
	/// add a native member function to class
	bool addMethod(const BppChar* name, BPPUSERFUNC pfunc, unsigned char flags = 0);
	/// add a native class function to class
	bool addStaticMethod(const BppChar* name, BPPUSERFUNC pfunc, unsigned char flags = 0);
	/// get index of a data member by its name
	int getMemberIndex(const BppString& name) const;
	/// mark class to be native
	void setIsNative(bool isNative);
	/// check if class is marked native
	bool getIsNative() const;

	/// Gets installed notification function for construction and destruction of objects (NULL if none).
	BppObjNotifyProc getNotifyProc() const { return _notifyProc;  }
	/// Installs notification function for construction and destruction of objects
	void setNotifyProc(BppObjNotifyProc notifyProc) { _notifyProc = notifyProc; };
	/// Gets access to VM owning this class.
	BppVirtualMachine& getVM() const;
//@}
private:
	friend class BppDictionaryIO;
	/// copy forbidden
	BppClass(const BppClass&);
	/// assignment forbidden
	BppClass& operator=(const BppClass&);
	/// builds vector of default non-static data values
	void buildDefaultValues(BppVector* defValues = 0);
	/// builds vector of member data names
	void fillMemberDataNames(BppVector& names) const;
	/// add a native function to class
	bool _addMethod(bool isStatic, const BppChar* name, BPPUSERFUNC pfunc, unsigned char flags = 0);
	/// class name
	BppString* _className;
	/// pointer to base class
	BppClass* _baseClass;
	/// pointer to owner class
	BppClass* _ownerClass;
	/// dictionary of data members
	BppDictionary* _dataMembers;
	/// dictionary of function members
	BppDictionary* _funcMembers;
	/// size (i.e. number of value objects) needed for a class
	int _size;
	BppVector _defaultValues;
	/// virtual machine owning this class
	BppVirtualMachine& _vm;
	/// native flag
	bool _isNative;
	/// notification proc, called when object is created or destroyed
	BppObjNotifyProc _notifyProc;
};

#endif // BPPCLASS_H
