#ifndef BPPCOMP_H
#define BPPCOMP_H

#include <bppdef.h>
#include <bppscan.h>
#include <bppstrng.h>
#include <bppdict.h>
#include <bppconst.h>
#include <bppvec.h>
#include <bppbuf.h>
class BppVirtualMachine;


/** Opcode byte structure.
An opcode byte in B++ consists of 6 bits opcode and 2 bits defining
size of additional data - may be byte, word (16 bits) or dword (16 bits).
This structure provides some acess helpers.
@see BppOpCode, BppOpDataType
*/
struct BppOpCodeByte
{
	union V {
		struct O {
			/// Opcode (6 bits)
			unsigned opCode : 6;
			/// Type od additional data (2 bits)
			unsigned opDataType : 2;
		}_op;
		 BppByte _ocb;
	}_v;
	/** Construction from opcode and additional data type.
	@param oc Opcode
	@aram odt Data type specification
	*/
	BppOpCodeByte(BppOpCode oc, BppOpDataType odt = ODT_NONE)
		 { _v._op.opCode= oc; _v._op.opDataType = odt; }
	/** Construction from single byte.
	@param ocb Byte containing opcode (6 bits) and additional data type (2 bits)
	*/
	BppOpCodeByte(BppByte ocb)
		/*opCode(ocb & 0x3F), opDataType(ocb >> 6)*/ { _v._ocb = ocb; }
	/// Conversion to byte
	inline operator BppByte() const {return _v._ocb/*return (BppByte)((opDataType << 6) | opCode)*/; }
	/// Conversion to BppOpCode.
	inline operator BppOpCode() const { return (BppOpCode) _v._op.opCode; }
	/// Conversion to BppOpDataType
	inline operator BppOpDataType() const { return (BppOpDataType) _v._op.opDataType; }
	inline BppOpCode opCode() const { return (BppOpCode) _v._op.opCode; }
	inline BppOpDataType opDataType() const {  return (BppOpDataType) _v._op.opDataType; }
};

/** Variable access function codes.
This internal helper type is used by some of code generation procedures to find out
correct opcode for access to variable.
*/
enum BppVarAccessType
{
	VA_NONE = 0,
	VA_LOAD	= 1,
	VA_STORE	= 2,
	VA_PUSH = 3,
	VA_DUP = 4
};

struct BppVarInfo;
struct BppFuncContext;
class IBppPrinter;
class BppDebugger;
class BppClass;
class BppValue;
/** Internal helper type for code generation.
This type is used by some of code generation procedures to find out
correct opcode for access to variable.
*/
enum BppVarType
{
	/// nothing, variable doesn't exist
	VAR_NONE = 0,
	/// argument
	VAR_ARG,
	/// temporary
	VAR_LOCAL,
	/// static local variable in function
	VAR_SLOCAL,
	/// global variable or static member variable
	VAR_VAR,
	/// member variable
	VAR_MEMBER,
	/// index variable
	VAR_INDEX,
	/// property member
	VAR_PROP,
	/// ignore - variable already processed
	VAR_DONE
};

/** B++ Bytecode compiler.
This class provides single pass compiler for B++ language. It uses a recursive
descending LL(1) parser.
@note Bytecode compiler is used by virtual machine to compile any sources into internal
bytecode representation. Client code (i.e. an application) should never need to
create instances of this class directly.
@see BppVirtualMachine::compileFile, BppVirtualMachine::compileText
*/
class BPPLIB_DLLEXP BppCompiler
{
public:
	/// ctor
	BppCompiler(BppVirtualMachine& vm, IBppPrinter* errPrinter = 0);
	/// dtor
	~BppCompiler();

	/** @name Conpiler main functions */
	//@{
	/// compile file source
	bool compile(FILE* fp, const BppChar* srcName = _T(""));
	/// compile string source
	bool compile(const BppChar* src, const BppChar* srcName = _T(""));
	/// compile a single function without creating a symbol table entry
	BppVector* compileFunction(const BppChar* src, const BppChar* args = _T(""), const BppChar* rtype=_T(""), const BppChar* srcName = _T(""),BppClass* pclass=0);
	/// parses a literal expression
	bool litExpr(const BppChar* src, BppValue& result, const BppChar* srcName = _T(""));
	//@}

	/** @name Debugging helper functions */
	//@{
	/// set _showInfo flag
	void setShowInfo(bool showInfo);
	/// get value of _showInfo flag
	bool getShowInfo() const;
	/// set _genLineInfo flag
	void setGenerateLineInfo(bool genInfo);
	/// get value of _genLineInfo flag
	bool getGenerateLineInfo() const;
	/// set _genLocalVarInfo flag
	void setGenerateLocalVarInfo(bool genInfo);
	/// Gets value of _usecDefine flag.
	bool getUseCDefine() const;
	///  Sets value od _useCDefine flag
	void setUseCDefine(bool useCDefine);
	/// Pushs current _useCDefine flag to stack.
	void pushUseCDefine();
	/// Restores _useCDefine flog from stack.
	bool popUseCDefine();
	/// get value of _genLocalVarInfo flag
	bool getGenerateLocalVarInfo() const;
	/// set a debugger for disassembling generated code
	void setDebugger(BppDebugger* debugger);
	//@}
	void onScannerDelete(BppScanner* scanner);
	bool notifyPreproc(int& errCnt);
	bool notifyToken(BppToken token, int& errCnt);
private:
	bool compile();
	void initCompiler();
	const BppChar* _getOperatorFunc(int& errCnt, BppClass* ownerClass = 0);
	/// compile a function
	bool _doFunctionOrVariable(bool isConst, int& errCnt, BppValueType rtype=BVT_NULL, BppClass* prclass = 0);
	/// compile a class
	bool _doClass(int& errCnt, BppClass* ownerClass = 0, unsigned short protectionLevel = 0, bool isnamespace = false);
	/// handle a preprocessing instruction
	bool _doPreproc(int& errCnt);
   /// handle ignored source part while preprocessing
   bool _doPreprocIgnore(int& errCnt);
   /// handle include preprocessing instruction
	bool _doInclude(int& errCnt);
	
	/// compile global variable definition
	bool _doGlobalVariable(const BppString& name, bool isConst, int& errCnt, BppValueType vtype=BVT_NULL, BppClass* pvClass = 0);
	/// compile a regualar function
	bool _doRegularFunction(const BppString& name, int& errCnt, bool constResult, BppValueType rtype=BVT_NULL, BppClass* prClass = 0);
	/// compile a member function
	bool _doMemberFunction(const BppString& name, int& errCnt, bool constResult, BppValueType rtype=BVT_NULL, BppClass* prClass = 0);
	/// get literal value from current token
	bool _doLiteralValue(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// handle include preprocessing instruction
	bool _doReadBinFile(const BppString& fname, BppBuffer& buffer, int& errCnt);
	/// evaluate literal expression
	bool _doLitExpr(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 1
	bool _doLitExpr1(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 2
	bool _doLitExpr2(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 3
	bool _doLitExpr3(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 4
	bool _doLitExpr4(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 5
	bool _doLitExpr5(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 6
	bool _doLitExpr6(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 6
	bool _doLitExpr7(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 8
	bool _doLitExpr8(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 9
	bool _doLitExpr9(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 10
	bool _doLitExpr10(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 11
	bool _doLitExpr11(BppToken token, BppDictionaryEntry& entry, int& errCnt); 
	/// evaluate literal expression of priority 12
	bool _doLitExpr12(BppToken token, BppDictionaryEntry& entry, int& errCnt);
	/// Parses and stores single macro definition
	bool _doMacroDefinition(const BppString& name, int& errCnt);
	/** Vector of token descriptors.
	Struct encapsulates a vector of tokens (token descriptors), containing contant of a macro.
	*/
	struct TokenListVec {
		~TokenListVec();
		/// Adds pointer to token descriptor at end of vector.
		void push_back(BppSavedToken* lst);
		/// Gets token descriptor at apecified position.
		BppSavedToken* at(size_t index) const;
		/// Merges tokens beginning at specified position.
		void mergeFrom(size_t index);
		/// Creates descriptor of comma token.
		BppSavedToken* createCommaToken() const;
		/// Gets number of entries in vector.
		size_t size() const;
		/// Removes all entries from vector
		void clear();
		BppVector _listVec; ///< Vector of entries.
	};

	bool _parseMacroArgumentList(TokenListVec& tokenListVec, int& errCnt);
	/// Expands specified macro.
	bool _doMacroExpansion(const BppString& name, int& errCnt);
	/// Puts information about current surce file and line number into compiled code.
	void _codeSourceInfo(BppFuncContext& context, int& errCnt);
	/// compile body of a function
	BppVector* _codeFunctionBody(const BppString& name, int& errCnt, bool constResult, BppClass* ownerClass = 0, BppValueType rtype=BVT_NULL, BppClass* prClass = 0);
	/// compile a block statement
	void _codeBlock(BppFuncContext& context, int& errCnt);
	/// compile a single statement
	void _codeStatement(BppFuncContext& context, int& errCnt);
	/// compile var statement
	void _codeVarDefinition(BppFuncContext& context, bool isConst, int& errCnt, BppValueType defType = BVT_NULL, BppClass* reqClass = 0);
	/// compile static [var] statement
	void _codeStaticVarDefinition(BppFuncContext& context, bool isConst, int& errCnt, BppValueType defType = BVT_NULL, BppClass* defClass = 0);
	/// compile try-catch statement
	void _codeTry(BppFuncContext& context, int& errCnt);
	/// compile try-catch statement
	void _codeThrow(BppFuncContext& context, int& errCnt);
	/// compile if statement
	void _codeIf(BppFuncContext& context, int& errCnt);
	/// compile while statement
	void _codeWhile(BppFuncContext& context, int& errCnt);
	/// compile do-while statement
	void _codeDoWhile(BppFuncContext& context, int& errCnt);
	/// compile for statement
	void _codeFor(BppFuncContext& context, int& errCnt);
	/// compile switch-case statement
	void _codeSwitch(BppFuncContext& context, int& errCnt);
	/// compile break
	void _codeBreak(BppFuncContext& context, int& errCnt);
	/// compile continue
	void _codeContinue(BppFuncContext& context, int& errCnt);
	/// compile return statement
	void _codeReturn(BppFuncContext& context, bool isYield, int& errCnt);
	/// compile a condition, i.e. '(' expression ')'
	void _codeCondition(BppFuncContext& context, int& errCnt);
	/// compile an expression
	void _codeExpr(BppFuncContext& context, int& errCnt);
	/// compile an expression of priority 1

protected:
	void _codeExpr1_comma(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 2
private:
	BppScanner* getActiveScanner() const;
   bool _active() const;
	void _codeExpr2_assign(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// code assignment (helper for combined assignment in _codeExpr2)
	void _codeAssign(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 3
	void _codeExpr3_ternop(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 4
	void _codeExpr4_logor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 5
	void _codeExpr5_logand(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 6
	void _codeExpr6_binor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 7
	void _codeExpr7_xor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 8
	void _codeExpr8_binand(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 9
	void _codeExpr9_eqops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 10
	void _codeExpr10_cmpops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 11
	void _codeExpr11_shift(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 12
	void _codeExpr12_addops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 13
	void _codeExpr13_mulops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 14
	void _codeExpr14_preops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile an expression of priority 15
	void _codeExpr15_post(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile ++ and -- prefix operators
	void _codePreIncDec(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt);
	/// compile ++ and -- postfix operators
	void _codePostIncDec(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt);
	/// compile new operator
	void _codeNew(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile delete operator
	void _codeDelete(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile primary expression
	void _codePrimary(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// fixup reference chain
	void _fixupRefChain(BppFuncContext& context, BppNWord startIdx, BppNWord destIdx);
	/// compile a object-method call
	void _codeMethodCall(const BppChar* name, BppVarInfo& thisInfo,  BppFuncContext& context, int& errCnt);
	/// compile a function call
	void _codeFunctionCall(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// Gets info about function to be called from current context
	BppDictionaryEntry* _getCalledFunctionInfo(const BppFuncContext& context, size_t varIndex, BppVector*& defargs, size_t& ndefargs, size_t& nrequargs, bool isMemberfunc= false) const;
	/// compile a indexed operation
	void _codeIndexedOp(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// compile a property operation
	void _codePropertyOp(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt);
	/// read comma separated list of identifiers
	int _getIdList(BppVector& dest, BppClass* contextClass, const BppChar* termStr, int& errCnt,BppVector* defaults = 0,bool useTypeNames = true, const BppChar* ellipseSymbol = 0);
	/// search for a class specified by name
	BppClass* _findClass(const BppString& name, int& errCnt,BppClass* ownerClass = 0, bool allowNotFound = false);
	/// search for variable in specified context
	BppVarInfo _findVariable(const BppChar* name, BppFuncContext& context, bool global, int& errCnt);
	/// search for class variable in specified context
	BppVarInfo _findClassVariable(const BppChar* name, BppFuncContext& context, int& errCnt, BppClass* refClass);
	/// code rvalue of a partial expression 
	void _codeRValue(BppVector& codeVec, BppVarInfo& varInfo);
	/// code from varInfo of a partial expression 
	void _codeVarInfo(BppVector& codeVec, BppVarAccessType act, BppVarInfo& varInfo);
	/// compile an argument
	void _codeArgument(BppVector& codeVec, BppVarAccessType act, BppNWord index);
	/// compile a temporary (local variable)
	void _codeTemporary(BppVector& codeVec, BppVarAccessType act, BppNWord index);
	/// compile a atatic local variable reference
	void _codeStaticLocal(BppVector& codeVec, BppVarAccessType act, BppNWord index);
	/// compile a variable reference
	void _codeVariable(BppVector& codeVec, BppVarAccessType act, BppNWord index);
	/// compile a member variable reference
	void _codeMember(BppVector& codeVec, BppVarAccessType act, BppNWord index);
	/// compile a literal reference
	void _codeLiteral(BppVector& codeVec, BppNWord index);
	/// compile an indexed reference
	void _codeIndexRef(BppVector& codeVec, BppVarAccessType act);
	/// compile a property reference
	void _codePropertyRef(BppVector& codeVec, BppVarAccessType act);
	/// code operation with index reference, size of operand depends on index value
	void _codeIndexedOpcode(BppVector& codeVec, BppOpCode opcode, BppNWord index, bool forceNativeSize = false);
	/// code operation without index reference
	size_t _codeOpcode(BppVector& codeVec, BppOpCode opcode);
	/// get new token and check for required type
	bool _fetchRequire(const BppChar* func, int& errCnt, BppToken rtoken, BppChar ch = _T('\0'));
	/// check token for required type
	bool _require(const BppChar* func, BppToken rtoken, BppToken token, BppChar ch = _T('\0'));
	/// check for lvalue
	bool _checkLValue(const BppVarInfo& varInfo);
	// Appends OP_POPC to buffer if last operation is assignment.
	//void _adjustAssignment(BppBuffer& codeBuffer,size_t startIdx);
	/// check if class has derived classes
 //  bool _isBaseClass(BppClass* pclass);
	void _error(const BppChar* msg, int* errCnt = 0);

	BppValueType findVTypeFromName(const BppChar* func,const BppString& name, int& errCnt, bool includeClasses = false, BppClass** classType = 0, BppClass* ownerClass = 0);

	/// Parses qualified name (delimited by ::)
	BppString parseNamePath(BppToken tok, int& errCnt);
	BppString parseNamePathParts(BppToken tok, int& errCnt,BppVector& parts, bool storeLastPart, bool returnFullPath);
#ifndef DOXY
	/// default ctor forbidden
	BppCompiler();
	/// copy ctor forbidden
	BppCompiler(const BppCompiler&);
	/// assignment forbidden
	BppCompiler& operator=(const BppCompiler&);
#endif
	/// scanner used by compiler
	BppScanner* _scanner;
	/// dictionary of global symbols (provided by _vm)
	BppDictionary& _symbols;
	/// dictionary of classes (provided by _vm)
	BppDictionary& _classes;
	/// dictionary of named literals
	BppDictionary _namedLiterals;


	struct BPPLIB_DLLEXP DefinedMacros {
		/** Indexes of items in value vector of macro definition entry.
	    @note Value of macro definition entry is a vector of two elements,
		where first element is vector of arguments of macro and second
		contains pointer macro content (list of token descriptors).
		*/
		enum {
			IARGS = 0, ///< Index of argument vector
			ILIST = 1 ///< Index of token descriptor list
		};
		DefinedMacros();
		~DefinedMacros();
		/// Deletes all macro definitions.
		void clear();
		/// Adds single macro definition.
		void defineMacro(const BppString& name, BppVector* definedArgs, BppSavedToken* tokenlist);
		/// Searches for macro definition by name.
		BppDictionaryEntry* find(const BppString& name);
		/// Gets Vector of defined arguments for specified macro.
		BppVector* getDefinedArgs(BppDictionaryEntry* entry) const;
		/// Gets content of defined macro.
		BppSavedToken* getTokenList(BppDictionaryEntry* entry) const;
		/// Erases nacro of specified name.
		bool erase(const BppString& name);
		/// Erases macro.
		bool erase(BppDictionaryEntry* entry);
		//unsigned count() const BPPNOEXCEPT;

	   /// dictionary of #define macros
	   BppDictionary _defined;
	} _definedMacros;
	/// stack for conditional compilation
   BppBuffer _condCompStack;
	/// reference to virtual machine
	BppVirtualMachine& _vm;
	/// printer for error messages
	IBppPrinter& _print;
	/// debugger for disassembling
	BppDebugger* _debugger;
	/// flag for showing compile informatins
	bool _showInfo;
	/// flag for generate line info in bytecode
	bool _genLineInfo;
	/// flag for generate local variables info in bytecode
	bool _genLocalVarInfo;
	/// Flag for using C-like defines
	bool _useCDefine;
	BppBuffer _useCDefStack;
	/// current line number
	long _currentLine;
	/// current source name
	BppString _currentSource;
	/// preprocessing level
	int _preprocLevel;
	/// Stack of source directories from start of compiling
	BppVector _srcDirStack;
	/// Flag set while macro definition
	bool _suppressMacroExpansion;
};

#endif // BPPCOMP_H
