#ifndef BPPCONST_H
#define BPPCONST_H

/** @file bppconst.h
Common constants for B++ library.
This header defines some constants used by B++ compiler, VM and debugger.
*/
#ifndef DOXY
#include <bppdef.h>
#endif


/// Indexes for special items in a bytecode vector.
enum BppByteCodeIndex
{
	/// reserved index of exception variable in temporaries vector
	IDX_EXCEPTION = 0,
	/// index of code buffer in a bytecode vector
	IDX_CODE = 0,
	/// index of class pointer in a bytecode vector
	IDX_CLASS = 1,
	/// index of stored function state in bytecode vector
	IDX_FUNCSTATE = 2,
	/// index of function name in a bytecode vector
	IDX_NAME = 3,
	/// index of first literal in a bytecode vector (same as IDX_NAME)
	IDX_FIRSTLITERAL = 3,
	/// index of vector containing default argument values
	IDX_DEFAULTARGS = 4,
	/// index of reference to bytecode vector itself, used for self-calls of a function
	IDX_SELFREFERENCE = 5
};


/** Opcodes defined in B++.
@see BppOpCodeByte
*/
enum BppOpCode
{
	/* opcodes */
	OP_NOP = 0x00, ///< no operation
	OP_BRT = 0x01, ///< branch on true 
	OP_BRF = 0x02, ///< branch on false 
	OP_BR = 0x03, ///< branch unconditionally 
	OP_NIL = 0x04, ///< load top of stack with nil 
	OP_PUSH = 0x05, ///< push nil onto stack 
	OP_NOT = 0x06, ///< logical negate top of stack 
	OP_NEG = 0x07, ///< negate top of stack 
	OP_ADD = 0x08, ///< add top two stack entries 
	OP_SUB = 0x09, ///< subtract top two stack entries 
	OP_MUL = 0x0A, ///< multiply top two stack entries 
	OP_DIV = 0x0B, ///< divide top two stack entries 
	OP_REM = 0x0C, ///< remainder of top two stack entries 
	OP_BAND = 0x0D, ///< bitwise and of top two stack entries 
	OP_BOR = 0x0E, ///< bitwise or of top two stack entries 
	OP_XOR = 0x0F, ///< bitwise xor of top two stack entries 
	OP_BNOT = 0x10, ///< bitwise not of top two stack entries 
	OP_SHL = 0x11, ///< shift left top two stack entries 
	OP_SHR = 0x12, ///< shift right top two stack entries 
	OP_LT = 0x13, ///< less than 
	OP_LE = 0x14, ///< less than or equal to 
	OP_EQ = 0x15, ///< equal to 
	OP_NE = 0x16, ///< not equal to 
	OP_GE = 0x17, ///< greater than or equal to 
	OP_GT = 0x18, ///< greater than 
	OP_INC = 0x19, ///< increment 
	OP_DEC = 0x1A, ///< decrement 
	OP_LIT = 0x1B, ///< load literal 
	OP_RETURN = 0x1C, ///< return from interpreter 
	OP_CALL = 0x1D, ///< call a function 
	OP_REF = 0x1E, ///< load a variable value 
	OP_SET = 0x1F, ///< set the value of a variable 
	OP_VREF = 0x20, ///< load a vector element 
	OP_VSET = 0x21, ///< set a vector element 
	OP_MREF = 0x22, ///< load a member variable value 
	OP_MSET = 0x23, ///< set a member variable 
	OP_AREF = 0x24, ///< load an argument value 
	OP_ASET = 0x25, ///< set an argument value 
	OP_TREF = 0x26, ///< load a temporary variable value 
	OP_TSET = 0x27, ///< set a temporary variable 
	OP_TSPACE = 0x28, ///< allocate temporary variable space 
	OP_SEND = 0x29, ///< send a message to an object 
	OP_DUP2 = 0x2A, ///< duplicate top two elements on the stack 
	OP_NEW = 0x2B, ///< create a new class object 
	OP_DELETE = 0x2C,    ///< delete object instance 
	OP_DUP = 0x2D,  ///< duplicate TOS
	OP_POP = 0x2E, ///< remove topmost entry from stack
	OP_MKREF = 0x2F, ///< make reference from reference counted value
	OP_TRY = 0x30, ///< enter try block
	OP_ENDTRY = 0x31, ///< leave try block
	OP_THROW = 0x32, ///< raise an exception
	OP_ENDCATCH = 0x33, ///< reset exception on end of catch block
	OP_SWITCH = 0x34,   ///< enter switch block
	OP_SREF = 0x35,    ///< load static variable
	OP_SSET = 0x36,    ///< set static variable
	OP_PINC = 0x37,    ///< post increment
	OP_PDEC = 0x38,    ///< post decrement
	OP_PREF = 0x39,    ///< load property
	OP_PSET = 0x3A,    ///< set property
	OP_POPC = 0x3B,    ///< copy TOS and POP
	OP_PLUS = 0x3C,    ///< unary plus
	OP_UNREF = 0x3D,   ///< make value from reference
	OP_STOS = 0x3E,    ///< store TOS value to tmp value in procInfo
	OP_LDTOS = 0x3F,   ///< load TOS from tmp value in procInfo
	OP_MKCONST = 0x40, ///< set temporary variable to constant
	OP_SPCONST = 0x41, ///< set value at TOS to constant
	OP_YIELD = 0x42, ///< Yield from coroutine
	// debug opcodes
	OP_TRON = 0xFC, ///< switch trace on
	OP_TROFF = 0xFD, ///< switch trace off
	OP_TRSTEP = 0xFE, ///< switch to single step mode
	OP_NEWLINE = 0xFF, ///< notify source line changed
	// prefix codes
	//OP_RBYTE = 0xF0,  ///< next operand is byte 
	//OP_RWORD = 0xF1,  ///< next operand is word instead of byte
	//OP_RDWORD = 0xF2  ///< next operand is dword instead of byte
};

/** Definition of additional data size attribute specified in an opcode byte.
@see BppOpCodeByte
*/
enum BppOpDataType
{
	ODT_NONE = 0,  ///< opcode without index data
	ODT_BYTE = 1,  ///< opcode with following data byte (8 bits)
	ODT_WORD = 2,  ///< opcode with following data word (16 bits)
	ODT_DWORD = 3, ///< opcode with following data dword (32 bits)
	/// opcode with following data of native word size (16 bits or 32 bits)
	ODT_NWORD = 
#if (defined(GCC) || (INT_MAX != SHRT_MAX))
	3
#else
	2
#endif
}; 





#endif // BPPCONST_H

