#ifndef BPPFIMPL_H
#define BPPFIMPL_H
#include <bppdef.h>

BPPLIB_DLLEXP extern const BppChar* FIMPL_MISSING_INTERPRETER;
BPPLIB_DLLEXP extern const BppChar* FIMPL_TOO_FEW_ARGS;
BPPLIB_DLLEXP extern const BppChar* FIMPL_TOO_MANY_ARGS;
BPPLIB_DLLEXP extern const BppChar* FIMPL_INVALID_THIS_PTR;
BPPLIB_DLLEXP extern const BppChar* FIMPL_CONST_ARG_NOT_ALLOWED;
/** @name Helper macros for native class implementation.
Macros defined in this section are intended to simplify implementation
of native B++ functions and methods.
*/
//@{

/// References default printer interface of virtual machine
#define PR (vm->printer())
#ifdef _DEBUG
/** Check for valid virtual machine.
In DEBUG builds this macro checks validity of virtual machine pointer.
In all build types it defines a local variable<br>
<tt>BppInterpreter* interpreter</tt> for use in implementation of native functions.
*/
#define CHKVM if (vm == 0) return false;\
   BppInterpreter* interpreter = vm->interpreter();\
   if (interpreter == 0) { PR(FIMPL_MISSING_INTERPRETER); return false;}
#else
/** Check for valid virtual machine.
In DEBUG builds this macro checks validity of virtual machine pointer.
In all build types it defines a local variable<br>
<tt>BppInterpreter* interpreter</tt> for use in implementation of native functions.
*/
#define CHKVM BppInterpreter* interpreter = vm->interpreter();
#endif
/// Leave current function and emit an error message.
#define THROW return interpreter->returnError
/** Check type of a value.
Macro checks if <i>VAL</i> is of data type <i>TYPE</i>. If not so,
function is aborted.
@param VAL any value (BppValue)
@param TYPE required type of value (BppValueType)
*/
#define CHKTYPE(VAL,TYPE) if (!interpreter->checkType(VAL,TYPE)) return false
/** Check for minimum number of function arguments.
@param MINCNT number of required arguments
*/
#define ARGCMIN(MINCNT) if (argc < MINCNT) THROW(FIMPL_TOO_FEW_ARGS)
/** Check for maximum number of function arguments.
@param MAXCNT maximum number of arguments
*/
#define ARGCMAX(MAXCNT) if (argc > MAXCNT) THROW(FIMPL_TOO_MANY_ARGS)
/** Check for minimum and maximum number of function arguments.
@param CMIN number of required arguments
@param CMAX maximum number of arguments
*/
#define CHKARGC(CMIN,CMAX) ARGCMIN(CMIN); ARGCMAX(CMAX)

/** Check if specified value is not constant.
@param VAL value to check
*/
#define CHKVAR(VAL) if ((VAL).getFlags() & BVF_CONST) THROW(FIMPL_CONST_ARG_NOT_ALLOWED)


/** Get reference to <i>I</i>th argument.
@param I argument index (index of first argument is 0)
@note CHKOBJECT macro must be used before first use of ARG.
@see CHKOBJECT
*/
#define ARG(I) (bp[(I)+1])

/** Check for objects <tt>this</tt>-pointer.
This macro is for use in native non-static member function implementations.
It defines a local variable <tt>bp</tt> of ValPtr type. The bp variable
points to value before first argument on stack.<br>
Furthermore it defines a variable <tt>pThis</tt> containing pointer to
current B++ object (type BppObject).
@note The <tt>bp</tt> variable is intended to be used for access to function arguments.
First argument is accessed by <tt>bp[1]</tt>. In non-static member functions
the first argument always is the object itself and bp[0] referres to method name.
@see ARG(I)
*/
#define CHKOBJECT \
   ValPtr bp = sp; bp-=argc; \
   BppObject* pThis = bp[1]; \
   if (pThis == 0) THROW(FIMPL_INVALID_THIS_PTR);

/** Get a member variable of own object by index.
@param IDX index of variable
@return reference to variable (BppValue&)
*/
#define MEMBER(IDX) (*(pThis->getMemberData()))[IDX]

/** Get a member entry of own objects class (including base classes) by name.
@param NAME name of member
@return pointer to member descriptor (BppDictionaryEntry*) if found, otherwise NULL
*/
#define CLMEMBER(NAME) (pThis->getClass()->findMember(NAME))

/** Get a member variable of any object by index.
@param pOBJ pointer to BppObject
@param IDX index of variable
@return reference to variable (BppValue&)
@see BppClass::findMember
*/
#define OBJMEMBER(pOBJ, IDX) (*((pOBJ)->getMemberData()))[IDX]
//@}

#endif // BPPFIMPL_H