// Documentation only - don't include it

/** @addtogroup Bp2Functions B++ predefined functions
This section describes predefined function available in each B++ program.
The Functions are grouped according to their typical areas of use.
@note
<ul>
 <li>Predefined functions normally are not statically typed. Thus, when assigning values to the arguments, there are generally no implicit type conversions.
 With the exception of integral types, which are always compatible with each other, application code is responsible for
 correct conversion.<br>
 If individual functions or function groups carry out such conversions, this is mentioned in their descriptions.</li>
 <li>The implementations of the functions check the validity of arguments passed at runtime and can trigger exceptions of type <tt>string</tt> in case of an error</li>
 <li>The types of parameters and return values are specified in the signatures.</li>
</ul>
@{
*/

/** @addtogroup ContainerFuncs Container related functions
The functions summarized here are used to handle the predefined container types <tt>buffer</tt>, <tt>charbuffer</tt>, <tt>dictionary</tt> and <tt>vector</tt>.
@{
*/

/** Clones a value of any type.
Function creates a clone of <i>val</i>. It may be used to create full (deep) copy of
any container type, string or object instance. For all other types (including
weak references) it simply returns argument itself.
@param val Value to clone.
@return cloned value
@note Function also creates a copy of bytecode functions (type BVT_BYTECODE) having static variables. In this case the result is
another function with same implementation as the original but own instances of static variables.
*/
var clone(const var val);

/** Creates a new vector.
Function creates a new vector and optionally initializes its elements.
@param capacity initial capacity of vector
@param init Flag for vector initialization. If <tt>true</tt>,
vector is initialized, i.e. vector size is set to be same as <i>capacity</i>
and all elements are set to <tt>null</tt>.
@return vector object
@note Setting capacity and initialization on creation
may increase speed because there is no need to resize vector at runtime.
@par Usage example:
@code
vec = newvector(10);
for (i=1;i<=10;++i) vec[i-1] = i*i;
@endcode
@see insert,verase, reserve
*/
vector newvector(uint capacity=0, bool init=true);

/** Creates buffer as a copy of another buffer or string.
Function makes a complete copy of <i>assignment</i> value.
@param assignment Value to create new buffer from. Value must be of type
<tt>string</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@return new buffer
@note In UNICODE builds wide characters will be converted to single bytes
discarding high byte of each character.
*/
buffer newbuffer(const var assignment);
/** Creates buffer filled with a specified byte.
@param size Number of bytes in buffer to be created.
@param fillbyte Byte value to initialize buffer with.
@return new buffer
*/
buffer newbuffer(uint size, byte fillbyte=0);


/** Creates character buffer as a copy of another buffer or string.
Function makes a complete copy of <i>assignment</i> value.
@param assignment Value to create new character buffer from. Value must be of type
<tt>string</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@return new character buffer
@note In UNICODE builds if <i>assignment</i> is of type <tt>buffer</tt>
each byte of <i>assignment</i> is converted into wide character
setting zero for high byte.
*/
charbuffer newcbuffer(const var assignment);


/** Creates character buffer filled with a specified character.
@param size Number of bytes in buffer to be created.
@param fillchar Character value to initialize buffer with.
@return new character buffer
*/
charbuffer newcbuffer(uint size, int fillchar=0);

/** Creates new empty dictionary.
@param sorted Flag for sorted dictionary. If <tt>true</tt> entries in dictionary will
be ordered by keys, otherwise new elements are appended to end.
@return new dictionary
@note <ul><li>Value of <i>sorted</i> flag affects results of dictgetkeys and dictgetvalues
functions. If set these functions return entries key ordered, otherwise in order of
insertion.</li>
<li>In DOS and WindowsCE builds element access will be faster in sorted than non-sorted
dictionaries.</li>
</ul>
*/
dictionary newdict(bool sorted = false);

/** Checks for dictionary entry.
Function checks if an entry for <i>key</i> is contained in <i>dict</i>.
@param dict dictionary to search
@param key key to search for
@return <tt>true</tt> if found, otherwise <tt>false</tt>
*/
bool dictcontains(const dictionary dict, const string key);

/** Removes entry from dictionary.
Function removes entry of key <i>key</i> from <i>dict</i> if contained.
@param dict dictionary to remove entry from
@param key key of entry to remove
@return <tt>true</tt> on success, otherwise <tt>false</tt>
*/
bool dicterase(dictionary dict, const string key);

/** Removes all entries from specified dictionary.
@param dict dictionary to remove entries from
@note Simply an alias to clear(var container).
*/
void dictclear(dictionary dict);

/** Gets vector of all keys contained in specified dictionary.
@param dict dictionary to get keys from
@return vector containing all keys of <i>dict</i>
@note <p>Returned vector contains copies of keys in dictionary. Therefore client code can modify
strings in returned vector without sideeffects for <i>dict</i>.</p>
<p>Returned vector may be used instead of an iterator to walk through the dictionary.</p>
<p>Order of keys in returned vector depends on sorted state of dictionary. In a sorted
dictionary keys are in alphabetical order, otherwise in order of insertion.
@par Usage example;
@code
   dictionary myDict = newdict();  // create unsorted dictionary
   // add any entries here
   // ...
   vector keyvec = dictgetkeys(myDict);
   uint n = size(keyvec);
   for (uint i=0; i<n;++i)
      print(keyvec[i],":",myDict[keyvec[i]],"\n");
@endcode
@see newdict, dictgetvalues
*/
vector dictgetkeys(const dictionary dict);

/** Gets const vector of all values contained in specified dictionary.
@param dict dictionary to get values from
@return constant vector containing all values of <i>dict</i>
@note
<p>Returned vector may be used instead of an iterator to walk through the dictionary.</p>
<p>Order of values in returned vector conforms to order of keys.</p>
@par Usage example;
@code
   dictionary myDict = newdict();  // create unsorted dictionary
   // add any entries here
   // ...
   // make constant alias of myDict
   // cdict is declared as const to prevent from making a copy.
   const auto cdict = mkconst(myDict);
   vector keyvec = dictgetkeys(cdict);
   const vector valvec = dictgetvalues(cdict);
   uint n = size(keyvec); // same as size(valvec)
   for (uint i=0; i<n;++i)
      print(keyvec[i],":",valvec[i],"\n");
@endcode
@see newdict, dictgetkeys
*/
const vector dictgetvalues(const dictionary dict);

/** Gets vector of all values contained in specified dictionary.
@param dict dictionary to get values from
@return vector containing all values of <i>dict</i>
@note
<p>Returned vector may be used instead of an iterator to walk through the dictionary.</p>
<p>Order of values in returned vector conforms to order of keys.</p>
<p>Modification of elements of result vector having reference types by client code will modify <i>dict</i> values too.
@par Usage example;
@code
   dictionary myDict = newdict();  // create unsorted dictionary
   // add any entries here
   // ...
   vector keyvec = dictgetkeys(myDict);
   vector valvec = dictgetvalues(myDict);
   uint n = size(keyvec); // same as size(valvec)
   for (uint i=0; i<n;++i)
      print(keyvec[i],":",valvec[i],"\n");
@endcode
@see newdict, dictgetkeys
*/
vector dictgetvalues(dictionary dict);


/** Gets number of elements in a container.
@param container Any container, i.e. value of type <tt>buffer</tt>,
<tt>charbuffer</tt>, <tt>vector</tt> or <tt>dictionary</tt>, or <tt>string</tt>.
@return number of elements in <i>container</i>
*/
uint size(const var container);

/** Shrinks capacity of container to current size.
@param container Value of type <tt>buffer</tt>,
<tt>charbuffer</tt> or <tt>vector</tt>.
@return Capacity (number of elements) of <i>container</i>
*/
uint shrink(var container);

/** Gets number of elements in a vector.
@param vec vector to get size from
@return number of elements in <i>vec</i>
@see size
*/
uint vecsize(const vector vec);

/** Sorts content of a container using QuickSort algorithm.
Function behaves like qsort function of C library.
@param container Container to sort. <i>container</i> may be of
type <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>vector</tt>.
@param callback Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked from qsort algorithm with two arguments (elements
of <tt>container</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>callback</i> is not specified, comparison is done using default rules.
@par Usage example;
@code
	class comparer
	{
	   comparer() {}
	   operator()(a,b) { return b-a; } // makes descending order
	};

	void main()
	{
		var vec = [3,1,5,7,2,0,9];
		var sz = size(vec);
		// sort ascending using anonymous function for callback
		qsort(vec,function(a,b) { return a-b; });
		// prints 0123579
		for (var i=0;i<sz;++i) print(vec[i]);
		print("\n");
		// sort descending using comparer object
		qsort(vec,new comparer());
		// prints 9753210
		for (i=0;i<sz;++i) print(vec[i]);
	}
@endcode
*/
void qsort(var container, var callback = null);

/** Enumerates a container and calls callback for each of its items.
@param container Container to enumarate. <i>container</i> may be of
type <tt>buffer</tt>, <tt>charbuffer</tt>,<tt>vector</tt>, <tt>dictionary</tt>
<tt>string</tt> or any class type defining its own <tt>foreach</tt> method of signature<br>
<tt>uint foreach(var callback, ...)</tt>.
@param callback Callback for enumeration. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked from <tt>foreach</tt> with at least two arguments (key and value
of current container element). It should return an integer. If result of <i>callback</i>
is zero (means <tt>false</tt>), enumeration is cancelled.
@param ... If <tt>foreach</tt> called with more than two arguments, additional argumants are
passed to <tt>callback</tt>.
@return Number of items enumerated.
@note <ul>
<li>If <i>container</i> is of <tt>dictionary</tt> type, first argument of <i>callback</i> is
the key of item. For all other types, first argument is the index in <i>container</i>.</li>
<li>Second argument of <i>callback</i> is the value of container item. If item value is of
a reference type, <i>callback</i> can modify it if needed.</li>
<li><i>callback</i> shouldn't modify size of <i>container</i> of enumerated items. If it does
so, it must return <tt>false</tt> to cancel enumeration.</li>
<li>If <i>container</i> is an object, function calls <tt>foreach</tt> method of that
object, passing <i>callback</i> and all additional arguments.</li>
<li>Function allows (i.e. ignores) use of constant object for <i>container</i>. In this case <i>callback</i> should not modify 
content of container.</li>
</ul>
@par Usage example;
@code
void main()
{
   // fill sorted dictionary from an unsorted one
   auto dict = { t:'t', e:'e',m:'m',p:'p',o:'o'}; // unsorted
   dictionary dict2 = newdict(true); // sorted
   // use dict2 for additional element
   foreach(dict,
           function(i,v,d) { d[i] = v; return true; },
           dict2);
   // callback for print container items
   auto iter = function(i, v) { print("I: ",i," V: ",v,"\n"); return true;};
   foreach(dict,iter); // list dict
   foreach(dict2,iter); // list dict2
   // enumerate a string
   foreach("Hello World",iter);
}
@endcode
*/
uint foreach(var container, var callback, ...);

/** Finds position of value in a container.
Function searches in <i>cont</i> for value equal to <i>searchVal</i> and returns
position of its first occurrence.
@param cont Container to search in, must be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param searchVal Value to search for. If <i>cont</i> is of <tt>string</tt> type, <i>searchVal</i> may be of
type <tt>strinh</tt> too or of a integral type (interpreted as character code).
@param pos position to start search from.
@return position of first occurrence if found, otherwise NPOS
@note Function does low level compare, i.e. it doesn't call equals- or compare-methods for objects.
*/
uint find(const var cont, const var searchVal, uint pos = 0);

/** Finds last position of value in a container.
Function searches backwards in <i>cont</i> for value equal to <i>searchVal</i> and returns
position of its last occurrence.
@param cont Container to search in, must be of type <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param searchVal Value to search for.
@param pos Position to start backwards search from.
@return Position of last occurrence if found, otherwise NPOS
@note Function does low level compare, i.e. it doesn't call equals- or compare-methods for objects.
*/
uint rfind(const var cont, const var searchVal, uint pos = NPOS);

/** Finds lower bound for specified value in specified range of a sorted container.
@param container Container to search in. Possible types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param startpos Start index of search interval in <i>cont</i>.
@param endpos End index of search interval in <i>cont</i>. If <i>endpos</i> is greater or equal to
size of <i>cont</i> search is done to end.
@param val Value to search for. In case of <tt>buffer</tt> or <tt>charbuffer</tt> as type of
<i>cont</i>, <i>val</i> should be of integral type.
@param comparer Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked with two arguments (elements
of <tt>cont</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>comparer</i> is not specified, comparison is done using default rules.
@return Returns index of first element in range [startpos,endpos] not less than <i>val</i>.<br>
If there is no such element, result is <i>endpos</i>+1.
@note Function assumes range [startpos,endpos] to be sorted in ascending order
with respect to result of behaviour of <i>comparer</i>, otherwise result is undefined.
*/
uint lbound(const var cont, uint startpos, uint endpos, const const var val, const var comparer = null);

/** Finds lower bound for specified value in a sorted container.
@param container Container to search in. Possible types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param val Value to search for. In case of <tt>buffer</tt> or <tt>charbuffer</tt> as type of
<i>cont</i>, <i>val</i> should be of integral type.
@param comparer Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked with two arguments (elements
of <tt>cont</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>comparer</i> is not specified, comparison is done using default rules.
@return Returns index of first element in <i>cont</i> not less than <i>val</i>.<br>
If there is no such element, result is same as #size(cont).
@note Function assumes <i>cont</i> to be sorted in ascending order
with respect to result of behaviour of <i>comparer</i>, otherwise result is undefined.
*/
uint lbound(const var cont, const var val, const var comparer = null);

/** Finds upper bound for specified value in specified range of a sorted container.
@param container Container to search in. Possible types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param startpos Start index of search interval in <i>cont</i>.
@param endpos End index of search interval in <i>cont</i>. If <i>endpos</i> is greater or equal to
size of <i>cont</i> search is done to end.
@param val Value to search for. In case of <tt>buffer</tt> or <tt>charbuffer</tt> as type of
<i>cont</i>, <i>val</i> should be of integral type.
@param comparer Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked with two arguments (elements
of <tt>cont</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>comparer</i> is not specified, comparison is done using default rules.
@return Returns index of first element in range [startpos,endpos] greater than <i>val</i>.<br>
If there is no such element, result is <i>endpos</i>+1.
@note Function assumes range [startpos,endpos] to be sorted in ascending order
with respect to result of behaviour of <i>comparer</i>, otherwise result is undefined.
*/
uint ubound(const var cont, uint startpos, uint endpos, const var val, const var comparer = null);

/** Finds upper bound for specified value in a sorted container.
@param container Container to search in. Possible types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param val Value to search for. In case of <tt>buffer</tt> or <tt>charbuffer</tt> as type of
<i>cont</i>, <i>val</i> should be of integral type.
@param comparer Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked with two arguments (elements
of <tt>cont</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>comparer</i> is not specified, comparison is done using default rules.
@return Returns index of first element in <i>cont</i> greater than <i>val</i>.<br>
If there is no such element, result is same as #size(cont).
@note Function assumes <i>cont</i> to be sorted in ascending order
with respect to result of behaviour of <i>comparer</i>, otherwise result is undefined.
*/
uint ubound(const var cont, const var val, const var comparer = null);

/** Does a binary search for specified value in specified range of a sorted container.
@param container Container to search in. Possible types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param startpos Start index of search interval in <i>cont</i>.
@param endpos End index of search interval in <i>cont</i>. If <i>endpos</i> is greater or equal to
size of <i>cont</i> search is done to end.
@param val Value to search for. In case of <tt>buffer</tt> or <tt>charbuffer</tt> as type of
<i>cont</i>, <i>val</i> should be of integral type.
@param comparer Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked with two arguments (elements
of <tt>cont</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>comparer</i> is not specified, comparison is done using default rules.
@return Returns index of first element in range [startpos,endpos] equal to <i>val</i>.<br>
If there is no such element, result is NPOS.
@note Function assumes range [startpos,endpos] to be sorted in ascending order
with respect to result of behaviour of <i>comparer</i>, otherwise result is undefined.
*/
uint bsearch(const var cont, uint startpos, uint endpos, const var val, const var comparer = null);

/** Does a binary search for specified value in a sorted container.
@param container Container to search in. Possible types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param val Value to search for. In case of <tt>buffer</tt> or <tt>charbuffer</tt> as type of
<i>cont</i>, <i>val</i> should be of integral type.
@param comparer Compare callback. The callback may be either
a function or a functor instance (an object defining operator()).
The callback is invoked with two arguments (elements
of <tt>cont</tt> to compare. It should return a negative number
if left argument is lower than right argument, a positive number if left
argument is greater than right argument or zero in case of equality.<br>
If <i>comparer</i> is not specified, comparison is done using default rules.
@return Returns index of first element in <i>cont</i> equal to <i>val</i>.<br>
If there is no such element, result is NPOS.
@note Function assumes <i>cont</i> to be sorted in ascending order
with respect to result of behaviour of <i>comparer</i>, otherwise result is undefined.
*/
uint bsearch(const var cont, const var val, const var comparer = null);


/** Create a vector from n-tupel of any values.
Function creates a vector from argument list.
@param ... list of any values
@return vector created from values
@note Other than in vector constructor ([...]) arguments of this function
are not restricted to literals.
@par Usage example;
@code
   string str = "Hello World!\n";
   vector vec = T(str,"This is B++\n",strlen(str);
   uint n = size(vec);
   for (uint i=0;i<n;++i) print(vec[i]);
@endcode
*/
vector T(...);

/** Create a vector from n-tupel of any values.
Function is a synonym to T(...) function.
@param ... list of any values
@return vector created from values
@see T(...)
*/
vector Vec(...);

/** Create a character buffer from n-tupel of values.
Function builds a character buffer from argument list.
@param ... List of values. Elements of argument list
may be of types <tt>int</tt> (interpreted as character), <tt>buffer</tt>,
<tt>charbuffer</tt> or <tt>string</tt>. Arguments of other types are ignored.
@return character buffer created from values
@par Usage example;
@code
   string str = "Hello World!";
   charbuffer myCBuf = CBuf(str,0x20,"Let's ",'b','u','i','l',"d a buffer!",'\n');
   print(string(myCBuf));
@endcode
*/
charbuffer CBuf(...);

/** Inserts copy of value one or more times into copy of container at specified position.
@param dst container to copy and insert value into, may  be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param pos Position in <i>dst</i> where insertion is done. If <i>pos</i> is greater than maximum valid index in <i>dst</i>, <i>insertion</i> is appended.
@param insertion Value to insert. 
If <i>dst</i> is of string type and <i>insertion</i> too and <i>cnt</i> is 1, string value of <i>insertion</i>
is inserted at <i>pos</i>.<br>
In all other cases either <i>dst</i> must be of type <tt>vector</tt> or <i>insertion</i> of integral type.<br>
If <i>dst</i> is of type <tt>string</tt> or <tt>charbuffer</tt> and value of <i>insertion</i> is of any integral type, it is converted into internal character type.
@param cnt Number of elements to insert.
@return modified copy of container
*/
var insert(const var dst, uint pos, const var insertion, uint cnt = 1);

/** Inserts value one or more times into container at specified position.
@param dst container to insert value into, may  be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param pos Position in <i>dst</i> where insertion is done. If <i>pos</i> is greater than maximum valid index in <i>dst</i>, <i>insertion</i> is appended.
@param insertion Value to insert. If <i>dst</i> is of vector type and is itself is used for
<i>insertion</i> argument, value is cloned before insertion.<br>
If <i>dst</i> is of string type and <i>insertion</i> too and <i>cnt</i> is 1, string value of <i>insertion</i>
is inserted at <i>pos</i>.<br>
In all other cases <i>insertion</i> must be of integral type.<br>
If <i>dst</i> is of type string or charbuffer and value of <i>insertion</i> is of any character type, it is converted into internal character type.
@param cnt Number of elements to insert.
@return modified container
*/
var insert(var dst, uint pos, var insertion, uint cnt = 1);

/** Inserts range of elements from one container into copy of another (or same) container of same type
at specified position.
@param dst container to insert values into copy of, may  be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param pos Position in <i>dst</i> where insertion is done. If <i>pos</i> is greater than maximum valid index in <i>dst</i>, values are appended.
@param src Container to copy range from. Type of <i>src</i> must be same as of <i>dst</i>.
@param startpos First index of range to copy from <i>src</i>. If <i>startpos</i> is invalid, i.e. greater than last possible index in <i>src</i>, function does nothing.
@param endpos Last index of range to copy from <i>src</i>. If <i>endpos</i> is greater than last possible index in <i>src</i>, last possible index is used instead.<br>
In <i>endpos</i> is less than <i>startpos</i>, insertion is done in reverse order.
@return new container as result of insertion
@note This function should be used rather than <tt>insert</tt> to insert multiple elements from one container into onother one, because it needs
only a single move operation in <i>dest</i>.
*/
var insrange(const var dst, uint pos, const var src, uint startpos = 0, uint endpos = NPOS);

/** Inserts range of elements from one container into another (or same) container of same type
at specified position.
@param dst container to insert values into, may  be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param pos Position in <i>dst</i> where insertion is done. If <i>pos</i> is greater than maximum valid index in <i>dst</i>, values are appended.
@param src Container to copy range from. Type of <i>src</i> must be same as of <i>dst</i>.<br>
If <i>src</i> and <i>dest</i> refer to same container and a part of range to copy is behind <i>pos</i>,
temporary shallow copy of <i>src</i> is made before insertion operation.
@param startpos First index of range to copy from <i>src</i>. If <i>startpos</i> is invalid, i.e. greater than last possible index in <i>src</i>, function does nothing.
@param endpos Last index of range to copy from <i>src</i>. If <i>endpos</i> is greater than last possible index in <i>src</i>, last possible index is used instead.<br>
In <i>endpos</i> is less than <i>startpos</i>, insertion is done in reverse order.
@return modified container
@note This function should be used rather than <tt>insert</tt> to insert multiple elements from one container into onother one, because it needs
only a single move operation in <i>dest</i>.
*/
var insrange(var dst, uint pos, var src, uint startpos = 0, uint endpos = NPOS);


/** Modifies capacity of container.
Function tries to reserve memory for <i>capacity</i> elements in <i>cont</i>. It doesn't
remove any data, i.e. if current size (number of elements) is greater than <i>capacity</i> it does nothing.
@param cont Container to modify. Possible container types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param capacity New capacity of container (number of elements).
@return Capacity of container.
*/
uint reserve(var cont, uint capacity);

/** Modifies capacity of container.
Function resizes <i>cont</i>. to <i>size</i> elements. If new size is less than current size, elements are removed from end of container.
If new size is greater than current, new empty elements will be appended to <i>cont</i>
@param cont Container to resize. Possible container types are <tt>vector</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>.
@param size New size of container (number of elements).
@parem fill Value to use for initialization of new (appended) elements. If <i>cont</i> is of type
<tt>buffer</tt> or <tt>charbuffer</tt>, <i>fill</i> must be of an integral type.<br>
If <i>fill</i> is omitted, <tt>null</tt> is used for default value of <tt>vector</tt> elements and <tt>0</tt> for
elements of <tt>buffer</tt> or <tt>charbuffer</tt>.
@return true on success.
*/
bool resize(var cont, uint size, var fill = null);

/** Inserts value one or more times into vector at specified position.
@param dst vector to insert value into.
@param pos Position in <i>cont</i> where insertion is done. If <i>pos</i> is negative
or greater than maximum valid index in <i>cont</i>, <i>insertion</i> is appended.
@param insertion Value (of any type) to insert. If <i>vec</i> itself is used for
<i>insertion</i> argument, value is cloned before insertion.
@param cnt Number of elements to insert.
@return modified vector
@note Simply an alias of insert(var dst,pos,insertion,cnt).
*/
vector vinsert(vector dst, int pos, var insertion, uint cnt = 1);

/** Erases one or more element(s) from copy of container.
@param cont container to erase element(s) from, may  be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param pos Position index of first element to erase.
@param cnt Maximum number of elements to erase.
@return modified copy of container
*/
var erase(const var cont, uint pos = 0, uint cnt = NPOS);

/** Erases one or more element(s) from container.
@param cont container to erase element(s) from, may  be of type <tt>vector</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>string</tt>.
@param pos Position index of first element to erase.
@param cnt Maximum number of elements to erase.
@return modified container
*/
var erase(var cont, uint pos = 0, uint cnt = NPOS);

/** Erases specified element or all elements from copy of dictionary.
@param dict dictionary to erase element(s) from.
@param key Key of element to erase, if not specified all elements are removed from dictionary.
@return modified copy of dictionary
@note If key not found, result is simply a copy of dict.
*/
dictionary erase(const dictionary dict, const string key = null);

/** Erases specified element or all elements from dictionary.
@param dict dictionary to erase element(s) from.
@param key Key of element to erase, if not specified all elements are removed from dictionary.
@return modified dictionary
@note If key not found, <i>dict</i> is not modified.
*/
dictionary erase(dictionary dict, const string key = null);

/** Erases one or more element(s) from copy of vector.
@param vec vector to erase element(s) from.
@param pos Position index of first element to erase.
@param cnt Maximum number of elements to erase.
@return modified copy of vector
@note Simply an alias of erase(const var container,pos,cnt).
*/
vector verase(const vector vec, uint pos = 0, uint cnt = NPOS);

/** Erases one or more element(s) from vector.
@param vec vector to erase element(s) from.
@param pos Position index of first element to erase.
@param cnt Maximum number of elements to erase.
@return modified vector
@note Simply an alias of erase(var container,pos,cnt).
*/
vector verase(vector vec, uint pos = 0, uint cnt = NPOS);

/** Exchanges specified elements in copy of a container.
@param cont Container containing elements to exchange. Possible types are <tt>vector</tt>,
<tt>buffer</tt> or <tt>charbuffer</tt>.
@param pos1 Position index of first element.
@param pos2 Position index of second element.
@return modified copy of container.
@note Function throws exception of string type in case of invalid index or invalid type of <i>container</i>.
@par Example;
@code
const vector arr = [1,2,3,4,5,6];
vector arr2 = swap(arr,4,1);
// arr2 is now [1,5,3,4,2,6]
@endcode
*/
var swap(const var cont, uint pos1, uint pos2);

/** Exchanges specified elements of a container.
@param cont Container containing elements to exchange. Possible types are <tt>vector</tt>,
<tt>buffer</tt> or <tt>charbuffer</tt>.
@param pos1 Position index of first element.
@param pos2 Position index of second element.
@return modified container.
@note Function throws exception of string type in case of invalid index or invalid type of <i>container</i>.
@par Example;
@code
vector arr = [1,2,3,4,5,6];
swap(arr,4,1);
// arr is now [1,5,3,4,2,6]
@endcode
*/
var swap(var cont, uint pos1, uint pos2);


/** Exchanges specified elements in copy of vector.
@param vec vector containing elements to exchange.
@param pos1 Position index of first element.
@param pos2 Position index of second element.
@return modified copy of vector vector.
@note Simply an alias of swap(const var cont ,pos1, pos2).
*/
vector vswap(const vector vec, uint pos1, uint pos2);

/** Exchanges specified elements of vector.
@param vec vector containing elements to exchange.
@param pos1 Position index of first element.
@param pos2 Position index of second element.
@return modified vector.
@note Simply an alias of swap(var cont ,pos1, pos2).
*/
vector vswap(vector vec, uint pos1, uint pos2);

/** Moves an element inside copy of container.
@param cont Container containing element to move.
Possible types are <tt>vector</tt>,
<tt>buffer</tt> or <tt>charbuffer</tt>.
@param srcpos Position index of element to move.
@param dstpos Destination position index for move operation.
@return modified copy of container.
@note Function throws exception of string type in case of invalid index or invalid type of <i>cont</i>.
@par Example
@code
const vector arr = [1,2,3,4,5,6];
auto arr2 = move(arr,4,1);
// arr2 is now [1,5,2,3,4,6]
@endcode
*/
var move(const var cont, uint srcpos, uint dstpos);

/** Moves an element inside a container.
@param cont Container containing element to move.
Possible types are <tt>vector</tt>,
<tt>buffer</tt> or <tt>charbuffer</tt>.
@param srcpos Position index of element to move.
@param dstpos Destination position index for move operation.
@return modified container.
@note Function throws exception of string type in case of invalid index or invalid type of <i>cont</i>.
@par Example
@code
vector arr = [1,2,3,4,5,6];
move(arr,4,1);
// arr is now [1,5,2,3,4,6]
@endcode
*/
var move(var cont, uint srcpos, uint dstpos);

/** Moves an element inside copy of vector.
@param vec vector containing element to move.
@param srcpos Position index of element to move.
@param dstpos Destination position index for move operation.
@return modified copy of vector.
@note Simply an alias of move(const var cont, srcpos, dstpos).
*/
vector vmove(const vector vec, uint srcpos, uint dstpos);

/** Moves an element inside a vector.
@param vec vector containing element to move.
@param srcpos Position index of element to move.
@param dstpos Destination position index for move operation.
@return modified vector.
@note Simply an alias of move(var cont, srcpos, dstpos).
*/
vector vmove(vector vec, uint srcpos, uint dstpos);

/** Extracts a subsequence of elements from container.
@param cont container to build subsequence from. Possible types are <tt>vector</tt>,
<tt>buffer</tt> or <tt>charbuffer</tt>.
@param from Index of first element to extract in <i>vec</i>.
@param cnt Number of elements to extract. If <i>cnt</i> greater than 
number of elements behind <i>from</i>, all elements to end of <i>cont</i> are included into result.
@return New container of same type as <i>cont</i> containing selected Elements of <i>vec</i>.
*/
var slice(const var cont, uint from = 0, uint cnt = NPOS);

/** Extracts a subsequenze of elements from vector.
@param vec vector to build subsequence from.
@param from Index of first element to extract in <i>vec</i>.
@param cnt Number of elements to extract. If <i>cnt</i> greater than
number of elements behind <i>from</i>, all elements to end of <i>vec</i> are icluded into result.
@return New vector containing selected Elements of <i>vec</i>.
@note Simply an alias of slice(var cont, from, cnt).
*/
vector vslice(const vector vec, uint from = 0, uint cnt = NPOS);


/** Removes all elements from specified container.
@param container container of type vector, charbuffer, buffer, string or dictionary to clear,
*/
void clear(var container);


/** Removes all elements from specified vector.
@param vec vector to clear,
@note Simply an alias to clear(var container).
*/
void vclear(vector vec);

/// @}

/** @addtogroup StringFuncs String functions
@{
*/
/** Create string as a copy of another string or buffer.
Function makes a complete copy of <i>assignment</i> value.
@param assignment Value to create new string from. Value must be of type
<tt>string</tt>, <tt>buffer</tt> or <tt>charbuffer</tt>. In case of
<tt>buffer</tt> each byte contained in <i>assignment</i> is assumed to
be a single byte character.
@return new string
*/
string newstring(const var assignment);

/** Create string from format specification.
Function builds new string from format specification and any number
of additional arguments.
@param formatstr Format specification. Value must be of <tt>string</tt> type.
The format string has similary syntax as in C - see BppString description.
@param ... More arguments to replace variables in <i>formatstring</i>.
@return new string
@note Elements of variable argument list normally should be of simple types, i.e.
characters, numbers or strings. For all other types
return value of BppValue::addr() method is used for variable value.
*/
string newstring(const string formatstr,...);

/** Creates string filled with a specified character.
@param size Number of characters in string to be created (without trailing '\\0'
character.
@param fillchar Character to initialize string with.
@return new string
*/
string newstring(uint size, int fillchar=' ');

/** Get length of a string.
@param str <tt>string</tt> value to get length of
@return number of characters in <i>str</i> (without trailing zero)
*/
uint strlen(const string str);

/** Compare function (case sensitive).
@param str1 left <tt>string</tt> for compare
@param str2 right <tt>string</tt> for compare
@return <ul>
<li> &lt; 0 if <i>str1</i> is lower than <i>str2</i></li>
<li>  0 if <i>str1</i> equals to <i>str2</i></li>
<li> &gt; 0 if <i>str1</i> is greater than <i>str2</i></li>
</ul>
*/
int strcmp(const string str1, const string str2);

/** Compare function (case insensitive).
@param str1 left <tt>string</tt> for compare
@param str2 right <tt>string</tt> for compare
@return <ul>
<li> &lt; 0 if <i>str1</i> is lower than <i>str2</i></li>
<li>  0 if <i>str1</i> equals to <i>str2</i></li>
<li> &gt; 0 if <i>str1</i> is greater than <i>str2</i></li>
</ul>
*/
int stricmp(const string str1, const string str2);

/** String tokenize function.
Function splits <i>str</i> into tokens using delimiters defined in <i>delimiter</i>.
@param str <tt>string</tt> to tokenize
@param delimiter Set of characters (type <tt>string</tt>) to be used for delimiters
@param storeEmptyTokens If <tt>true</tt>, empty tokens are stored in result too.
@param storeDelimitChar If <tt>true</tt>, delimiter character between tokens are stored in result too.
@return vector containing tokens
@note <ul>
  <li>If <i>str</i> is empty, result vector is empty too.</li>
  <li>Delimiters (if <i>storeDelimitChar</i> set) are atored in result as character codes (i.e. of type <tt>int</tt>).</li>
  <li>If <i>storeEmptyTokens</i> is set and last character of <i>str</i> is a delimiter, empty
  string is appended to result.
 </ul>
@par Usage example;
@code
void listvec(vector v)
{
   int n = size(v);
   print("size: ",n,"\n");
   for (int i=0;i<n;++i) print(""+v[i],"\n");
   print("-----\n");
}

main() {
   string str = "This is a  sample string.";
   listvec(strsplit(str," ."));
   listvec(strsplit(str," .",false,true));
   listvec(strsplit(str," .",true));
   listvec(strsplit(str," .",false,true));
   listvec(strsplit(str," .",true,true));
}
@endcode
*/
vector strsplit(const string str, string delimiter, bool storeEmptyTokens = false, bool storeDelimitChar= false);

/** Find position of substring or character.
Function searches <i>str</i> for substring or character <i>search</i> and returns
position of its first occurrence.
@param str string to search
@param search string or character to search for
@param start character position to start search from.
@return position of first occurrence if found, otherwise NPOS
*/
uint strfind(const string str, const var search, uint start = 0);

/** Erase part from copy of string.
Function deletes one or more characters from a copy of string.
@param str source string
@param pos index of first character to delete
@param cnt Number of characters to delete. If value is greater than number of
characters behind <i>pos</i>, all characters to end of string will be deleted.
@return modified copy of <i>str</i>
@note Function modifies argument <i>str</i>. The return value is the
modified string itself, not a copy of it.
*/
string strerase(const string str, uint pos, uint cnt = NPOS);

/** Erase part of string.
Function deletes one or more characters from string.
@param str string to erase character(s) from
@param pos index of first character to delete
@param cnt Number of characters to delete. If value is greater than number of
characters behind <i>pos</i>, all characters to end of string will be deleted.
@return modified copy of <i>str</i>.
*/
string strerase(string str, uint pos, uint cnt = NPOS);

/** Insert string into copy of another one.
Function inserts value of <i>insstr</i> into <i>str</i> at position <i>pos</i>.
@param str string to copy and modify
@param pos Index position where <i>insstr</i> should be inserted. Use length
of <i>str</i> for <i>pos</i> to append <i>insstr</i> value.
@param insstr string to insert
@return modified copy of <i>str</i>
*/
string strinsert(const string str, uint pos, const string insstr);

/** Insert string into another one.
Function inserts value of <i>insstr</i> into <i>str</i> at position <i>pos</i>.
@param str string to modify
@param pos Index position where <i>insstr</i> should be inserted. Use length
of <i>str</i> for <i>pos</i> to append <i>insstr</i> value.
@param insstr string to insert
@return modified string
@note Function modifies argument <i>str</i>. The return value is the
modified string itself, not a copy of it.
*/
string strinsert(string str, uint pos, const string insstr);

/** Insert character(s) into string.
Function inserts <i>cnt</i> characters <i>insc</i> into <i>str</i> at position <i>pos</i>.
@param str string to modify
@param pos Index position for insertion. Use length
of <i>str</i> for <i>pos</i> to append characters.
@param insc character value to insert
@param cnt number of characters to insert
@return modified string
@note Function modifies argument <i>str</i>. The return value is the
modified string itself, not a copy of it.
*/
string strinsert(string str, uint pos, int insc, uint cnt = 1);

/** Replace part of copy of string.
Function replaces all occurrences of <i>search</i> by value of <i>replacement</i>.
@param str string to copy and modify
@param search character or substring to be replaced
@param replacement character or string to be used for replacement
@return result of replacement
@note Both, <i>search</i> and <i>replacement</i>, must be either single characters or strings.
*/
string strreplace(const string str, const var search, const var replacement);

/** Replace part of string.
Function replaces all occurrences of <i>search</i> by value of <i>replacement</i>.
@param str string to modify
@param search character or substring to be replaced
@param replacement character or string to be used for replacement
@return current object
@note Both, <i>search</i> and <i>replacement</i>, must be either single characters or strings.
*/
string strreplace(string str, const var search, const var replacement);

/** Get substring from string.
Function returns substring of <i>str</i>, beginning at position <i>start</i>
and of maximum <i>len</i> characters length.
@param str string to get substrung from
@param start Start position of substring.
@param len Maximum size of substring. If len is omitted,
all characters of current string (beginning at <i>start</i>) are included in result.
@return new string containing specified substring
*/
string substr(const string str, uint start, uint len = NPOS);


/** Make string upper case.
Function converts letters contained in copy of <i>str</i> to upper case.
@param str string to convert
@return converted string
@note Function behaves like C-function of same name, i.e. result depends on locale settings.
@see setlocale, strlwr
*/
string strupr(const string str);

/** Make string upper case.
Function converts letters contained in <i>str</i> to upper case.
@param str string to convert
@return converted string
@note Function modifies content of <i>str</i> and returns reference to modified string.<br>
It behaves like C-function of same name, i.e. result depends on locale settings.
@see setlocale, strlwr
*/
string strupr(string str);

/** Make string lower case.
Function converts letters contained in copy of <i>str</i> to lower case.
@param str string to convert
@return converted string
@note Function behaves like C-function of same name, i.e. result depends on locale settings.
@see setlocale, strupr
*/
string strlwr(const string str);

/** Make string lower case.
Function converts letters contained in <i>str</i> to lower case.
@param str string to convert
@return converted string
@note Function modifies content of <i>str</i> and returns reference to modified string.<br>
It behaves like C-function of same name, i.e. result depends on locale settings.
@see setlocale, strupr
*/
string strlwr(string str);

/// @}

/** @addtogroup MathFuncs Math functions

All functions of this section accept argument(s) of any numeric type. <br>
Return value is (with except of #abs function)
of <tt>double</tt> type.
@note If an argument is of object type, functions (with except of #abs) try to
convert it using double conversion operator of argument.
@{
*/

/** Calculates sin(<i>x</i>).
@param x argument (radiant)
@return sinus of <i>x</i>
*/
double sin(var x);

/** Calculates cos(<i>x</i>).
@param x argument (radiant)
@return cosinus of <i>x</i>
*/
double cos(var x);

/** Calculates tan(<i>x</i>).
@param x argument (radiant)
@return tangens of <i>x</i>
*/
double tan(var x);

/** Calculates arctan(<i>x</i>).
@param x argument
@return arcus tangens of <i>x</i> (radiant)
*/
double atan(var x);

/** Calculates square root.
@param x argument, must not be negative
@return square root of <i>x</i>
*/
double sqrt(var x);

/** Calculates natural logarithm ln(<i>x</i>).
@param x argument, must be positive
@return natural logarithm of <i>x</i>
*/
double log(var x);

/** Calculates exponential function <b>e</b><sup><i>x</i></sup>.
@param x argument
@return <b>e</b> to the power <i>x</i>
*/
double exp(var x);

/** Calculates exponential function <i>x</i><sup><i>y</i></sup>.
@param x base
@param y exponent
@return <i>x</i> to the power of <i>y</i>
*/
double pow(var x, var y);

/** Calculates absolute value of <i>x</i>.
@param x argument
@return absolute value of <i>x</i>
@note Return type is same as type of <i>x</i>.
*/
var abs(var x);

/// @}

/** @addtogroup IOFuncs I/O functions
@{
*/

/** @name Common I/O functions
@note Functions of this section are available in all B++ builds.
*/
//@{

/** Print list of values to default output device.
@param ... Comma separated list of values. Output depends on value type as follows:
<ul>
   <li><tt>string</tt> : Value is written to output without any changes.</li>
   <li>all integer types (<tt>sbyte ... ulong</tt>): Value is written as unformatted integer or unsigned integer (depending on type).</li>
   <li><tt>double</tt> or <tt>float</tt> : Value is written as unformatted 
   double of automatic notation (like %g-format pf printf).</li>
   <li><tt>intptr</tt> : Value is written as address in hexadecimal representation.</li>
   <li>all other types : Type name and address of value are written to output.</li>
</ul>
@note <ul>
<li>Output is done using IBppPrinter implementation set for default printer of
virtual machine.</li>
<li>All arguments are assumed to be <tt>const</tt>.</li>
*/
void print(...);


/** Open a file.
Function tries to open file <i>filename</i> in specified open mode. It behaves
exactly as same-named C-function.
@param filename Name of file to open. The name may contain relative or absolute
peth specification.
@param mode Specification of open mode, may contain following elements:
<ul>
   <li>r : Open an existing file for read only.</li>
   <li>w : Create file and open it for write. If specified file already exists it is overwritten.</li>
   <li>a : Open a file for append data. If file doesn't exist it will be created.</li>
	<li>a+ : Open a file for read and append data. If file doesn't exist it will be created.</li>
	<li>r+ : Open existinng file for read and write.</li>
   <li>w+ : Create file and open it for read and write. If specified file already exists it is overwritten.</li>
   <li>t : Open file in text mode (may be combined with a, r, r+, w and w+).</li>
   <li>b : Open file in binary mode (may be combined with a, r, r+, w and w+).</li>
</ul>
@return File descriptor on success, otherwise <tt>null</tt>.
@note Files opened by this function should be closed using fclose if no longer needed.
@see fclose
*/
FILE fopen(const string filename, const string mode);

/** Reopen a file stream with different file and/or mode.
Function closes <i>fp</i> tries to reopen it  for file <i>filename</i> in specified open mode. It behaves
exactly as same-named C-function.
@param filename Name of file to open. The name maycontain relative or absolute
peth specification.
@param mode Specification of open mode, may contain following elements:
<ul>
<li>r : Open an existing file for read only.</li>
<li>w : Create file and open it for write. If specified file already exists it is overwritten.</li>
<li>a : Open a file for append data. If file doesn't exist it will be created.</li>
<li>a+ : Open a file for read and append data. If file doesn't exist it will be created.</li>
<li>r+ : Open existinng file for read and write.</li>
<li>w+ : Create file and open it for read and write. If specified file already exists it is overwritten.</li>
<li>t : Open file in text mode (may be combined with a, r, r+, w and w+).</li>
<li>b : Open file in binary mode (may be combined with a, r, r+, w and w+).</li>
</ul>
@param fp Descriptor of file to reopen.
@return File descriptor on success, otherwise <tt>null</tt>.
@note Files opened by this function should be closed using fclose if no longer needed.
@see fclose
*/
FILE freopen(const string filename, const string mode, FILE fp);


/** Close a file.
Function closes a file previously opened by fopen.
@param fp descriptor of file to be closed
@return 0 on success, otherwise non-zero value
@note Function must not be used with default streams stdin, stdout or stderr.
*/
int fclose(FILE fp);

/** Flushes buffer of file stream.
Function behaves like same named C-function.
@param fp descriptor of file to be flushed
@return 0 on success, otherwise non-zero value
@note If <i>fp</i> is null, all files opened in write mode are flushed.
*/
int fflush(const FILE fp);

/** Check for end of file.
Function checks if end of specified file is reached.
@param fp descriptor of file 
@return non-zero value if end of file is reached, otherwise 0
*/
int feof(const FILE fp);

/**Moves the file pointer to a specified position.
Function behaves like C-function of same name.
@param fp descriptor of file
@param offset Number of bytes from <i>origin</i>
@param origin Initial position, must be one of SEEK_SET (0), SEEK_CUR (1) or SEEK_END (2).
@return 0 on success, non zero value otherwise
*/
int fseek(FILE fp, int offset, int origin);

/** Gets current position in specified file.
Function behaves like C-function of same name.
@param fp descriptor of file
@return Current position of file pointer in <i>fp</i>.
*/
int ftell(const FILE fp);


/** Read single character from opened file.
Function reads a single-byte character from specified file.
@param fp file descriptor
@return character code (-1 if end of file reached)
*/
int getc(FILE fp);

/** Read single character input object installed for virtual machine.
Function reads a single character from input object installed for virtual machine.
@return character code (-1 if nothing read)
@note By default virtual machine uses <tt>stdin</tt> device to get the character.
However, another input object (IBppInput implementation) may be installed by client code.
*/
int getc();

/** Checks if any key was pressed.
@return true if a key with a character code was pressed.
@note Behavior of function differs depending on OS where program is running on:
<ul>
<li>For MS-DOS and Windows console applications, it behaves like <a href="https://learn.microsoft.com/en-us/cpp/c-runtime-library/reference/kbhit" target="_blank"><tt>kbhit</tt> C-function</a>.</li>
<li>For Windows applications without console, function is emulated using <a href="https://learn.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getasynckeystate" target="_blank"><tt>GetAsyncKeyState</tt> function</a>.<br>
Please note that there is no character available on <tt>stdin</tt> when function returns <i>true</i> in that case.</li>
<li>On UNIX platforms it's emulated too. It behaves as for Windows console application if current process is connected to a terminal window. Otherwise behavior is undefined.</li>
</ul>
*/
bool kbhit();

/** Write single character to opened file.
Function writes a single-byte character to specified file.
@param c character to write
@param fp file descriptor
@return character code written (-1 on error)
*/
int putc(int c, FILE fp);

/** Read string from standard input.
Function by default reads string from stdin device (usually the console). It returns
on new line (i.i.erase. Enter key). The trailing line feed is not contained
in return value.
@return string read
@note Function normally works on systems having a console only. However, an input object (IBppInput implemantation)
may be installed for virtual machine to get input from another device than <tt>stdin</tt>.
@see fgets
*/
string gets();

/** Read string from file.
Function reads a string from an opened file, beginning at current position.
If returns on new line or if end of file is reached.
@param fp descriptor of file to read from
@return string read
@note <p>Trailing line feed character is not part of result.</p>
<p>Function should be used only for reading from files opened in text mode.</p>
@see gets, fread
*/
string fgets(FILE fp);

/** Write string to file.
Function writes <i>text</i> into opened file <i>fp</i>, beginning at current position.
@param text string to write
@param fp descriptor of file to write into
@return on success code of last character written, otherwise EOF (-1).
@note Function should be used for writing to files opened in text mode only.
@see fwrite
*/
int fputs(const string text, FILE fp);

/** Read value from file.
Function reads any B++ value from opened binary file, beginning at current position.
@param fp Descriptor of file to read from. The file must be in format as created by
fwriteval function.
@return on success value read, otherwise <tt>null</tt>
@note Client code should assign returned value to a variable for further use.
Please refer to #fwriteval documentation for a usage example.
@see fwriteval
*/
var freadval(FILE fp);

/** Write value to file.
Function writes any B++ value into opened binary file, beginning at current position.
@param value value to write
@param fp descriptor of file to write into
@return <tt>true</tt> on success, otherwise <tt>false</tt>
@note Using #fwriteval and #freadval functions allows to serialize and deserialize
any B++ data structure in a simple way. Espacially that includes complete object
hierarchies too. Any cross references between objects inside a hierarchy are
saved on serialization and resolved on deserialization.
@par Example:
@code
// Serialization example using fwriteval and freadval.
// first we define a list entry class

class ListEntry
{
   static _maxID = 0;
   _a;
   _b;
   _ID;
   _next;
}

// ctor
ListEntry::ListEntry(a,b)
{
   _ID = ++_maxID;
   _a=a;
   _b=b;
   _next = null;
}

ListEntry::getA() { return _a; }

// print contents of entry
ListEntry::pr()
{
   print("ID: ",_ID," ",getA(), " ",_b,"\n");
}

ListEntry::getNext() { return _next; }

ListEntry::setNext(next) { _next = next; }

ListEntry::append(v)
{
   if (_next)
      _next->append();
   else
   {
      _next = v;
   }
}

// dtor
ListEntry::~ListEntry()
{
   print("destroyed ",_ID,"\n");
}

main()
{
   var lst = new ListEntry(-1,0);
   for(var i=0;i<9;i++)               // append more objects to lst
   {
      lst->append(new ListEntry(i,i+1));
   }
   for(var p = l;p;p=p->getNext())    // print contents
      p->pr();
   var fp=fopen("obj.dat","wb");      // open file for write
   fwriteval(lst,fp);             // write entire list to file
   fclose(fp);
   lst = null;                    // free list
   fp=fopen("obj.dat","rb");      // open file for read
   lst=freadval(fp);              // read contents and assign to lst
   fclose(fp);
   for(p = lst;p;p=p->getNext())  // print contents
      p->pr();
}
@endcode
@see freadval
*/
bool fwriteval(const var value, FILE fp);

/** Read file.
Function appends contents of file specified by <i>fp</i> to <i>readbuf</i>.
@param readbuf Destination buffer.
@param fp Descriptor of opened file to read from.
@return number of bytes read
@note <ul>
  <li>Contents of <i>fp</i> is read from current file position to end of file.</li>
  <li>Capacity of <i>readbuf</i> is enlarged if neccessary.</li>
</ul>
*/
uint fread(buffer readbuf, FILE fp);

/** Read specified number of bytes from file.
Function appends up to <i>cnt</i> bytes from contents of file specified by <i>fp</i>
to <i>readbuf</i>.
@param readbuf Destination buffer.
@param cnt Maximum number of bytes to read.
@param fp Descriptor of opened file to read from.
@return number of bytes read
@note <ul>
   <li>Capacity of <i>readbuf</i> is enlarged if neccessary.</li>
   <li>Reading starts at current file position</li>
</ul>
*/
uint fread(buffer readbuf,  uint cnt, FILE fp);

/** Read specified number of elements of fixed size from file.
Function appends up to <i>cnt</i> elements of <i>size</i> bytes length from contents
of file specified by <i>fp</i> to <i>readbuf</i>.
@param readbuf Destination buffer.
@param size size of single element to read
@param cnt Maximum number of elements to read.
@param fp Descriptor of opened file to read from.
@return number of elements read
@note <ul>
   <li>Capacity of <i>readbuf</i> is enlarged if neccessary.</li>
   <li>Reading starts at current file position</li>
</ul>
*/
uint fread(buffer readbuf, uint size, uint cnt, FILE fp);

/** Write buffer to file.
Function writes complete contents of <i>buf</i> into file, starting at current
file position.
@param buf Buffer containing contents to write.
@param fp Descriptor of opened file to write to.
@return number of bytes written
*/
uint fwrite(const buffer buf, FILE fp);

/** Write specified number of bytes to file.
Function writes up to <i>cnt</i> bytes from <i>buf</i> to file specified by <i>fp</i>.
@param buf Buffer containing contents to write.
@param cnt Maximum number of bytes to write.
@param fp Descriptor of opened file to write to.
@return number of bytes written
@note Function writes into file starting at current file position.
*/
uint fwrite(const buffer buf, uint cnt, FILE fp);

/** Write specified number of fixed sized elements to file.
Function writes up to <i>cnt</i> elements of <i>size</i> bytes length from <i>buf</i>
to file specified by <i>fp</i>.
@param buf Buffer containing contents to write.
@param size size of e single element
@param cnt Maximum number of elements to write.
@param fp Descriptor of opened file to write to.
@return number of elements written
@note Function writes into file starting at current file position.
*/
uint fwrite(const buffer buf, uint size, uint cnt, FILE fp);

///@}

/** @name Wide character I/O functions
@note Functions of this section are available in Windows builds only.
*/
//@{

/** Read single wide character from opened file.
Function reads a wide character from specified file.
@param fp file descriptor
@return character code (-1 if end of file reached)
*/
int getwc(FILE fp);

/** Write single wide character to opened file.
Function writes a wide character to specified file.
@param c character to write
@param fp file descriptor
@return character code written (-1 on error)
*/
int putwc(int c, FILE fp);

/** Read wide character string from file.
Function reads a wide character string from an opened file, beginning at current position.
If returns on new line or if end of file is reached.
@param fp descriptor of file to read from
@return string read
@note <p>Trailing line feed character is not part of result.</p>
<p>Function should be used only for reading from files opened in text mode.</p>
@see fread
*/
string fgetws(FILE fp);

/** Write wide character string to file.
Function writes <i>text</i> into opened file <i>fp</i>, beginning at current position.
@param text string to write
@param fp descriptor of file to write into
@return on success code of last character written, otherwise EOF (-1).
@note Function should be used only for writing to files opened in text mode.
@see fwrite
*/
int fputws(const string text, FILE fp);
///@}

/// @}

/** @addtogroup RTTIFuncs RTTI and type conversion
@{
*/

/** Convert value to boolean value.
Function converts argument of any type to <tt>bool</tt> value.
@param value Value to convert.
@return boolean representation for <i>value</i>.<br>
Result is <tt>false</tt> for numeric types of value 0, the <tt>null</tt> value and
all reference type having data pointer NULL (i.e. without any data). In all other cases
result value is <tt>true</tt>, if argument is not an object. For argument of object type
its conversion operators <tt>bool</tt> or <tt>int</tt> are called if defined.
@note 
  For conversion of objects, function searches for <tt>bool</tt> operator first, then
  for <tt>int</tt> operator.
@see size, gettype, int
*/
bool bool(const var value);

/** Converts value to signed byte.
Function tries to convert argument into value of <tt>sbyte</tt> type (8 bits signed integer).
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>sbyte</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>sbyte</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
sbyte sbyte(const var value);

/** Converts value to signed char.
Function tries to convert argument into value of <tt>char</tt> type (single byte character).
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>char</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>char</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
char char(const var value);

/** Converts value to unsigned byte (unsigned char).
Function tries to convert argument into value of <tt>byte</tt> type (8 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>byte</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
byte byte(const var value);

/** Converts value to unsigned char.
Function tries to convert argument into value of <tt>uchar</tt> type (single byte character).
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>char</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>uchar</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
uchar uchar(const var value);

/** Converts value to short integer.
Function tries to convert argument into value of <tt>short</tt> type (16 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>short</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
short short(const var value);

/** Converts value to unsigned short integer.
Function tries to convert argument into value of <tt>ushort</tt> type (16 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>ushort</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
ushort ushort(const var value);

/** Converts value to UTF-16le character.
Function tries to convert argument into value of <tt>char16_t</tt> type (16 bits character).
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>char16_t</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
char16_t char16_t(const var value);

/** Converts value to integer.
Function tries to convert argument into value of <tt>int</tt> type (32 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>int</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
int int(const var value);

/** Converts value to unsigned integer.
Function tries to convert argument into value of <tt>uint</tt> type (32 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>uint</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
uint uint(const var value);

/** Converts value to UTF-32le character.
Function tries to convert argument into value of <tt>char32_t</tt> type (32 bits character).
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>char32_t</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
char32_t char32_t(const var value);

/** Converts value to wide character.
Function tries to convert argument into value of <tt>wchar_t</tt> type.
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>The size of <tt>wchar_t</tt> depends on runtime environment used.</li>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>wchar_t</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
wchar_t wchar_t(const var value);

/** Converts value to UTF-8 unicode character.
Function tries to convert argument into value of <tt>char8_t</tt> type.
@param value Value to convert. The value must be of any number or character type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>The internal representation of <tt>char8_t</tt> is (other than in C++20) the representation of UTF-8 byte sequence
  and is stored as 32 bits value.</li>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, and size of buffer is lower or equal to 4, its content is assumed to be a byte sequence representing 
  the bytes of UTF-8 character.</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>char8_t</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
char8_t char8_t(const var value);

/** Converts value to long integer.
Function tries to convert argument into value of <tt>long</tt> type (64 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>long</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
long long(const var value);

/** Converts value to unsigned long integer.
Function tries to convert argument into value of <tt>ulong</tt> type (64 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>ulong</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
ulong ulong(const var value);

/** Converts value to integer of size same as address pointers of system.
Function tries to convert argument into value of <tt>intptr</tt> type (32 or 64 bits signed integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>intptr</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
intptr intptr(const var value);

/** Converts value to unsigned integer of size same as address pointers of system.
Function tries to convert argument into value of <tt>uintptr</tt> type (32 or 64 bits unsigned integer).
@param value Value to convert. The value must be of any number type, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>int</tt>
@note <ul>
  <li>A <tt>string</tt> value should contain string representation of a decimal, octal,
  hexadecimal or dual integer number. Hexadecimal numbers are prefixed by '0x'or '0X', dual numbers by '0b' or '0B'. An octal number starts
  with 0, followed by a digit between 1 and 7. In all other cases a decimal number is assumed.</li>
  <li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 8, 16, 32 or
  64 bits integer value (i.e. size of buffer must be 1, 2, 4 or 8).</li>
  <li>In case of <tt>object</tt> for <i>value</i>, the specified object must define an <tt>intptr</tt>
  conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
uintptr uintptr(const var value);

/** Convert value to <tt>float</tt>.
Function tries to convert argument into value of <tt>float</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>string</tt>, <tt>buffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>float</tt>
@note <ul>
<li>A <tt>string</tt> value must be string representation of a <tt>float</tt> number in format<br>
<tt>[<i>whitespace</i>] [<i>sign</i>] [digits] [.digits] [ {d | D | e | E }[sign]digits]</tt><br>
where a whitespace is a space or tab character and sign is either plus (+) or minus (-).</li>
<li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 32 bits (float) or
64 bits (double) value (i.e. size of buffer must be 4 or 8).</li>
<li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>float</tt>
or <tt>double</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
float float(const var value);

/** Convert value to <tt>double</tt>.
Function tries to convert argument into value of <tt>double</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>string</tt> or <tt>buffer</tt>.
@return <i>value</i> converted to <tt>double</tt>
@note <ul>
<li>A <tt>string</tt> value must be string representation of a <tt>double</tt> number in format<br>
<tt>[<i>whitespace</i>] [<i>sign</i>] [digits] [.digits] [ {d | D | e | E }[sign]digits]</tt><br>
where a whitespace is a space or tab character and sign is either plus (+) or minus (-).</li>
<li>If <i>value</i> is of <tt>buffer</tt> type, the buffer must contain a 32 bits (float) or
64 bits (double) value (i.e. size of buffer must be 4 or 8).</li>
<li>In case of <tt>object</tt> for <i>value</i>, the specified object must define a <tt>double</tt>
or <tt>float</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
double double(const var value);

/** Convert value to <tt>string</tt>.
Function tries to convert argument into value of <tt>string</tt> type.
@param value Value to convert. The value must be of any numeric type or character type, <tt>charbuffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>string</tt>
@note <ul>
<li>Numeric values of integral types (except <tt>intptr</tt> and <tt>uintptr</tt>) are converted to decimal string representation.</li>
<li>Values of type <tt>intptr</tt> or <tt>uintptr</tt> are represented as haxadecimal addresses (like <tt>printf("%p",...)</tt> in C).
<li>Values of types <tt>float</tt> and <tt>double</tt> will be converted to string
without any special format (and without expotantial notation).</li>
<li>Values of character types are converted into internal character type and result in a string containing single character.</li>
<li>In case of a <tt>charbuffer</tt> value a new string containing all characters of buffer is returned.</li>
<li>Values of <tt>object</tt> type must have a <tt>string</tt> conversion operator or a
<tt>toString</tt> method for conversion.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
string string(const var value);

/** Convert value to <tt>buffer</tt>.
Function tries to convert argument into value of <tt>buffer</tt> type.
@param value Value to convert. The value must be of type <tt>int</tt>, <tt>float</tt>,
<tt>double</tt>, <tt>buffer</tt>, <tt>charbuffer</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>buffer</tt>
@note <ul>
<li>For numeric types (integral numbers, characters, <tt>float</tt> and <tt>double</tt>) returned buffer
contains a bytewise copy of argument and has same langth as native size of argument type.</li>
<li>In case of a <tt>charbuffer</tt> or <tt>string</tt> all characters of argument, including 
trailing zero, are copied to result buffer. Size of result is number of characters copied
multipled by number of bytes used for a single character.</li>
<li>In case of a <tt>buffer</tt> function simply returns a reference (not a copy!) to argument. If argument is <tt>const</tt>,
result is <tt>const</tt> too.</li>
<li>Values of <tt>object</tt> type must have a <tt>buffer</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
buffer buffer(const var value);

/** Convert value to <tt>charbuffer</tt>.
Function tries to convert argument into value of <tt>charbuffer</tt> type.
@param value Value to convert. The value must be of a numeric integral type, a character type, <tt>float</tt>,
<tt>double</tt>, <tt>buffer</tt>, <tt>charbuffer</tt>, <tt>string</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>charbuffer</tt>
@note <ul>
<li>For numeric types (except <tt>intptr</tt> and <tt>uintptr</tt>) returned charbuffer
contains a string representation of argument. Floating point values are converted
using %g format specification.</li>
<li>Values of types <tt>intptr</tt> and <tt>uintptr</tt> are interpreted as addresses and result in a hexadecimal
represantation like <tt>printf</tt> C-function using format specification %f.</li>
<li>Values of character types are converted into internal character type and result in a character buffer containing single character.</li>
<li>In case of a <tt>buffer</tt> argument all bytes of argument are interpreted as
single character values, independent of current character size. </li>
<li>In case of a <tt>charbuffer</tt> function simply returns a reference (not a copy!) to argument. If argument is <tt>const</tt>,
result is <tt>const</tt> too. </li>
<li>Values of <tt>object</tt> type must have a <tt>charbuffer</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
charbuffer charbuffer(const var value);

/** Convert value to <tt>vector</tt>.
Function tries to convert argument into value of <tt>vector</tt> type.
@param value Value to convert. The value must be of type <tt>vector</tt>, 
<tt>buffer</tt>, <tt>charbuffer</tt>, <tt>dictionary</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>vector</tt>
@note <ul>
<li>In case of a <tt>vector</tt> function simply returns a reference (not a copy!) of argument. If argument is <tt>const</tt>,
result is <tt>const</tt> too.</li>
<li>In case of a <tt>buffer</tt> argument each byte of argument is converted into vector
element of type <tt>byte</tt>. </li>
<li>In case of a <tt>charbuffer</tt> argument each character of argument is converted into vector
element of internal character type. </li>
<li>In case of a <tt>dictionary</tt> argument result vector consists of values from dictionary. </li>
<li>Values of <tt>object</tt> type must have a <tt>vector</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
vector vector(const var value);

/** Convert value to <tt>dictionary</tt>.
Function tries to convert argument into value of <tt>dictionary</tt> type.
@param value Value to convert. The value must be of type <tt>vector</tt>, 
 <tt>dictionary</tt> or <tt>object</tt>.
@return <i>value</i> converted to <tt>vector</tt>
@note <ul>
<li>In case of a <tt>vector</tt> result contains vector elements as values where keys
are string representations of element indexes in vector.</li>
<li>In case of a <tt>dictionary</tt> function simply returns a reference (not a copy!) of argument. If argument is <tt>const</tt>,
result is <tt>const</tt> too.</li>
<li>Values of <tt>object</tt> type must have a <tt>dictionary</tt> conversion operator.</li>
<li>In case of invalid argument type exception of string type is thrown. Use #gettype to
check type of value.</li>
</ul>
@see size, gettype
*/
dictionary dictionary(const var value);

/** Get type id of argument value.
@param value value to get type id from
@return type id of <i>value</i>
@note Use #gettypename function to convert returned id into readable string representation.
@see gettypename, BppValueType
*/
int gettype(const var value);


/** Get human readable name of a data type.
@param typeId id of data type to get name of
@param prefixed Specifies if result is prefixed by 'BVT_' in order to get
internal symbol name of type. Prefix is used if \e prefixed can be evaluated to \c true.
@return Name of data type. Following table shows possible names:
<table border="1">
<tr><td><b>Type id (HEX)</b></td><td><b>Type name</b></td><td><b>Type</b></td></tr>
<tr><td>0 (00)</td><td>NULL</td><td><tt>null</tt></td></tr>
<tr><td>1 (01)</td><td>VOID</td><td>void</td></tr>
<tr><td>2 (02)</td><td>AUTO</td><td>auto</td></tr>
<tr><td>64 (40)</td><td>SBYTE</td><td><tt>sbyte</tt></td></tr>
<tr><td>65 (41)</td><td>BOOL</td><td><tt>bool</tt></td></tr>
<tr><td>66 (42)</td><td>CHAR</td><td><tt>char</tt></td></tr>
<tr><td>68 (44)</td><td>BYTE</td><td><tt>bool</tt></td></tr>
<tr><td>70 (46)</td><td>UCHAR</td><td><tt>uchar</tt></td></tr>
<tr><td>71 (47)</td><td>CHAR8</td><td><tt>char8_t</tt></td></tr>
<tr><td>72 (48)</td><td>SHORT</td><td><tt>short</tt></td></tr>
<tr><td>76 (4C)</td><td>USHORT</td><td><tt>ushort</tt></td></tr>
<tr><td>78 (4E)</td><td>CHAR16</td><td><tt>char16_t</tt></td></tr>
<tr><td>79 (4F)</td><td>WCHAR</td><td><tt>wchar_t</tt></td></tr>
<tr><td>80 (50)</td><td>INT</td><td><tt>int</tt></td></tr>
<tr><td>81 (51)</td><td>INTPTR</td><td><tt>intptr</tt></td></tr>
<tr><td>84 (54)</td><td>UINT</td><td><tt>uint</tt></td></tr>
<tr><td>85 (55)</td><td>UINTPTR</td><td><tt>uintptr</tt></td></tr>
<tr><td>86 (56)</td><td>CHAR32</td><td><tt>char32_t</tt></td></tr>
<tr><td>88 (58)</td><td>LONG</td><td><tt>long</tt></td></tr>
<tr><td>92 (5C)</td><td>ULONG</td><td><tt>ulong</tt></td></tr>
<tr><td>112 (70)</td><td>FLOAT</td><td><tt>float</tt></td></tr>
<tr><td>120 (78)</td><td>DOUBLE</td><td><tt>double</tt></td></tr>
<tr><td>128 (80)</td><td>FILE</td><td><tt>FILE</tt></td></tr>
<tr><td>129 (81)</td><td>CODE</td><td>native function</td></tr>
<tr><td>130 (82)</td><td>VAR</td><td>variable, dictionary entry</td></tr>
<tr><td>160 (A0)</td><td>OBJECT</td><td><tt>object</tt></td></tr>
<tr><td>161 (A1)</td><td>STRONGOBJECT</td><td><tt>strong typed object</tt></td></tr>
<tr><td>168 (A8)</td><td>CLASS</td><td><tt>Class</tt></td></tr>
<tr><td>176 (B0)</td><td>STRING</td><td><tt>string</tt></td></tr>
<tr><td>177 (B1)</td><td>BUFFER</td><td><tt>buffer</tt></td></tr>
<tr><td>178 (B2)</td><td>CHARBUFFER</td><td><tt>charbuffer</tt></td></tr>
<tr><td>180 (B4)</td><td>VECTOR</td><td><tt>vector</tt></td></tr>
<tr><td>181 (B5)</td><td>BYTECODE</td><td>B++ function</td></tr>
<tr><td>182 (B6)</td><td>LOCALVERINFO</td><td>local variable info (internal used)</td></tr>
<tr><td>184 (B8)</td><td>DICTIONARY</td><td><tt>dictionary</tt></td></tr>
<tr><td>224 (E0)</td><td>OBJECTREF</td><td>free reference to <tt>object</tt></td></tr>
<tr><td>232 (E8)</td><td>CLASSREF</td><td>free reference to <tt>Class</tt></td></tr>
<tr><td>240 (F0)</td><td>STRINGREF</td><td>free reference to <tt>string</tt></td></tr>
<tr><td>241 (F1)</td><td>BUFFERREF</td><td>free reference to <tt>buffer</tt></td></tr>
<tr><td>242 (F2)</td><td>CHARBUFFERREF</td><td>free reference to <tt>charbuffer</tt></td></tr>
<tr><td>244 (F4)</td><td>VECTORREF</td><td>free reference to <tt>vector</tt></td></tr>
<tr><td>245 (F5)</td><td>VARINFOREF</td><td>free reference to B++ function</td></tr>
<tr><td>246 (F6)</td><td>CODEREF</td><td>free reference to native function</td></tr>
<tr><td>248 (F8)</td><td>DICTIONARYREF</td><td>free reference to <tt>dictionary</tt></td></tr>
</table>
@note Predefined literals of symbolc type names should be used for <i>typeid</i> value rather than
numeric constants, e.g. BVT_INT instead of 72.
@see gettype
*/
string gettypename(int typeId, bool prefixed = false);

/** Get class name of object or class.
@param value <tt>object</tt> or <tt>Class</tt> to get class name from
@return class name
@note 
<ul>
<li>In case of invalid argument type exception of string type is thrown.</li>
<li>Because return value is only a reference (not a copy) of class name stored in BppClass instance,
client code should not modify const state and change contents of return value.</li>
</ul>
@see dynamic_cast, gettype, getclass
*/
const string getclassname(const var value);

/** Gets class of object or finds class for specified name.
@param value <tt>object</tt>, <tt>Class</tt> or <tt>string</tt> to get class for.
@return class for <i>value</i>
@note 
<ul>
<li>In case of invalid argument type, exception of string type is thrown.</li>
<li>If <i>value</i> is of <tt>string</tt> type and no class found for (qualified) name specified by <i>value</i>, result is <tt>null</tt>.
<li>If <i>value</i> is already a class, function does nothing but return the argument.</li>
</ul>
@see dynamic_cast, gettype, getclassname, newobject
*/
Class getclass(const var value);


/** Creates new object instance of specified class.
Function behaves like <tt>new</tt> operator but allows dynamic specification of class to create instance of.<br>
It is intended for use as a helper for generic object creation at runtime.
@param classtype Class to use for object creation.
@param ... List of arguments for constructor call.
@return object created
@note In case of invalid argument type or value <tt>null</tt> for <i>classtype</i>, exception of string type is thrown.
@par Example
@code
object factoryFunc(string classname, ...) {
	vector args = argv();
	// replace class name by class reference in arguments
	args[0] = getclass(classname);
	return call(newobject,args);
}
@endcode
@see dynamic_cast, gettype, getclassname, getclass
*/
object newobject(Class classtype, ...);

/** Gets a global symbol (variable, function or class) by name at runtime.
@param name Name of symbol to search for.
@param isclass If set to <tt>true</tt> function searches for a class, otherwise for
global symbol of any other type.
@return If a symbol for <i>name</i> was found, function returns its value, e.g. function
pointer for a function, otherwise <tt>null</tt>.
@note This function is mainly intended for using different system functions, depending
on platform used, in same program without compiler error.
@par Example
@code
// Gets current working directory
string getCwd()
{
   string buf = null;
   if (MSDOS)
   {
      buf = newstring(256);
      function segread = getsymbol("segread");
      function setreg = getsymbol("setreg");
      function int86x = getsymbol("int86x");
      segread();
      setreg("ah",0x47);
      setreg("dl",0);
      intptr pbuf = addr(buf);
      setreg("ds",pbuf >> 16);
      setreg("si",pbuf & 0xFFFF);
      int86x(0x21);
      return buf;
   }
   else if (WIN32)
   {
      function GetFunc = getsymbol("GetKernel32Func");
      intptr pfunc = GetFunc("GetCurrentDirectory?");
      uint siz = CallLibFunc(pfunc,0,0);
      buf = newstring(siz);
      CallLibFunc(pfunc,siz,addr(buf));
      return buf;
   }
   throw "Invalid OS";
}
@endcode
@see dynamic_cast, gettype
*/
var getsymbol(const string name, bool isclass = false);

/** Do dynamic cast to specified destination class.
Function behaves very similarly to <tt>dynamic_cast</tt> operator in C++.<br>
First it evaluates expression <i>expr</i>. If evaluation results in an object of class
type <i>desttype</i> or derived class, it returns the object, otherwise <tt>null</tt>.
@param desttype class to cast <i>expr</i> to
@param expr expression of any type
@return object of class <i>desttype</i> (or derived) on success, otherwise <tt>null</tt>
@note <ul>
  <li>Other than <tt>dynamic_cast</tt> operator in C++ there is no exeption thrown
if <i>expr</i> is not an object. In this case function simply returns <tt>null</tt>.
Use #gettype to check a value for its basic type.</li>
<li>Function preserves constant state of <i>expr</i>, i.e., if result of <i>expr</i> is a constant object, 
result of function is <tt>const</tt> too. In that case object is cloned if result is assigned to a non-const variable.</li>
</ul>
@par Example:
@code
   class Base
   {
      Base();
   }
   Base::Base() { }

   class Derived : Base
   {
      Derived();
   }
   Derived::Derived() { }

   main(;obj1, obj1, o, i)
   {
      obj1 = new Base();
      obj2 = new Derived();
      o = dynamic_cast(Base, obj1);      // returns obj1
      o = dynamic_cast(Base, obj2);      // returns obj2
      o = dynamic_cast(Derived, obj1);   // returns null
      o = dynamic_cast(Derived, obj2);   // returns obj2
      i = dynamic_cast(Base, 17);        // returns null (not 0)
   }
@endcode
@see gettype, getclass, dynamic_cast(string,var)
*/
object dynamic_cast(const Class desttype, var expr);

/** Do dynamic cast to destination class specified by name.
Function behaves very similarly to <tt>dynamic_cast</tt> operator in C++.<br>
First it evaluates expression <i>expr</i>. If evaluation results in an object of class
specified by <i>classname</i> or derived class, it returns the object, otherwise <tt>null</tt>.
@param classname name of class to cast <i>expr</i> to
@param expr expression of any type
@return object of class <i>desttype</i> (or derived) on success, otherwise <tt>null</tt>
@note <ul>
  <li>Other than <tt>dynamic_cast</tt> operator in C++ there is no exeption thrown
if <i>expr</i> is not an object. In this case function simply returns <tt>null</tt>.
Use #gettype to check a value for its basic type.</li>
<li>Function preserves constant state of <i>expr</i>, i.e., if result of <i>expr</i> is a constant object,
result of function is <tt>const</tt> too. In that case object is cloned if result is assigned to a non-const variable.</li>
</ul>
@par Example:
@code
   class Base
   {
      Base();
   }
   Base::Base() { }

   class Derived : Base
   {
      Derived();
   }
   Derived::Derived() { }

   main(;obj1, obj1, o, i)
   {
      obj1 = new Base();
      obj2 = new Derived();
      o = dynamic_cast("Base", obj1);               // returns obj1
      o = dynamic_cast("Base", obj2);               // returns obj2
      o = dynamic_cast("Derived", obj1);            // returns null
      o = dynamic_cast("Derived", obj2);            // returns obj2
      o = dynamic_cast(getclassname(obj2), obj1);   // returns null
      o = dynamic_cast(getclassname(obj1), obj2);   // returns obj2
      i = dynamic_cast("Base", 17);                 // returns null (not 0)
   }
@endcode
@see gettype, getclassname, dynamic_cast(Class, var)
*/
object dynamic_cast(string classname, var expr);

/** Checks if specified value is const.
@param expr any expression, typical variable or result of function call, to check for constant
@return true if <i>expr</i> is constant, false otherwise
@see mkconst
*/
bool isconst(var expr);

/** Modifies constant state of exprerssion result.
@param expr any expression, typical variable or result of function call, change constant state for
@param setconst new constant state
@return value of <i>expr</i> with modified constant state
@note Function never modifies state of an l-value (variable) specified by <i>expr</i>. Result is
either a copy of value in <i>expr</i> (in case of simple value) or a reference (in case of reference type value) with constant state 
as specified by <i>setconst</i>.
@see isconst
*/
var mkconst(var expr, bool setconst = true);

/// @}

/** @addtogroup TimeFuncs Date and time functions
@{
*/
/** Get C-timestamp.
@return number of seconds elapsed since 1970/01/01 0:00 GMT
@note Function behaves like <tt>time</tt> function of C standard library, on 16-Bit-platforms return type is <i>int</i> (32 bits)
instead of <i>long</i>.
*/
long time();

/** Gets elapsed time.
This function is intended to be used for time measurements.<br>
On Windows-CE platform it returns elapsed time since start of system, on all other platforms 
elapsed time since start of current program (more precise: time since start of B++ runtime).
@return elapsed time in milliseconds
@note Depending on platform, resolution of timer may be lower than 1 millisecond.
@par Example:
@code
   main()
   {
      uint t1 = timer();
      // do anything here
      ...
      uint t2 = timer();
      print(t2-t1, " ms needed for operation.\n");
   }
@endcode
*/
uint timer();


/** Get string representation of a C-timestamp.
Function makes a readable string from timestamp. It behaves like same named
C function.
@param t C-timestamp, e.g. result of a #time() call. On 16-Bit-platforms argument type is <i>int</i> (32 bits).
@return string representation of argument, e.g. <i>Sat May 10 21:52:54 2008</i>
*/
string ctime(long t);

/** Make a localized time point from C-timestamp and split it into
different parts.
Behavior of function is similarly to same named C function. However,
because B++ doesn't support structure types it uses a vector for result.
@param t Timestamp to localize, on 16-Bit-platforms argument type is <i>int</i> (32 bits).
@return vector of following components:
<table border="1">
   <tr><td align="center"><b>Index</b></td><td align="center"><b>Value</b></td></tr>
   <tr><td>0</td><td>Seconds after minute (0 .. 59).</td></tr>
   <tr><td>1</td><td>Minutes after hour (0 .. 59).</td></tr>
   <tr><td>2</td><td>Hours after midnight (0 .. 23).</td></tr>
   <tr><td>3</td><td>Day of month (1 .. 31).</td></tr>
   <tr><td>4</td><td>Month (0 .. 11; January = 0).</td></tr>
   <tr><td>5</td><td>Year (current year minus 1900).</td></tr>
   <tr><td>6</td><td>Day of week (0 .. 6; Sunday = 0).</td></tr>
   <tr><td>7</td><td>Day of year (0 .. 365; January 1 = 0), -1 if not supported.</td></tr>
   <tr><td>8</td><td>Positive value if daylight saving time is in effect;<br>
   0 if daylight saving time is not in effect;<br>negative value if status is unknown.</td></tr>
</table>
@note On Windows CE informations about day of year (index 7) and daylight saving time
are currently not supported, i.e. both values will be -1.
see time, ctime
*/
vector localtime(long t);

/** Function converts date and time information from a given calendar time tm to character string according to format string fmt.
Behavior of function is similarly to same named C function.
@param fmt Format string.<br>
The format string consists of zero or more conversion specifiers and ordinary characters 
(except %). All ordinary characters, including the terminating null character, are copied 
to the output string without modification. Each conversion specification begins with 
% character, followed by the character that determines the behavior of the specifier. 
At least the following format specifiers are available:
<table>
<tr><th>Specifier</th><th>Meaning</th></tr>
<tr><td align="center"> %% </td><td> Character % </td></tr>
<tr><td align="center"> \%a </td><td> Abbreviated weekday name </td></tr>
<tr><td align="center"> \%A </td><td> Full weekday name </td></tr>
<tr><td align="center"> \%b </td><td> Abbreviated month name </td></tr>
<tr><td align="center"> \%B </td><td> Full month name </td></tr>
<tr><td align="center"> \%c </td><td> Date and time representation </td></tr>
<tr><td align="center"> \%d </td><td> Day of the month (01-31) </td></tr>
<tr><td align="center"> \%H </td><td> Hour in 24h format (00-23) </td></tr>
<tr><td align="center"> \%I </td><td> Hour in 12h format (01-12) </td></tr>
<tr><td align="center"> \%j </td><td> Day of the year (001-366) </td></tr>
<tr><td align="center"> \%m </td><td> Month as a decimal number (01-12) </td></tr>
<tr><td align="center"> \%M </td><td> Minute (00-59) </td></tr>
<tr><td align="center"> \%p </td><td> Either AM or PM according to the given time value </td></tr>
<tr><td align="center"> \%S </td><td> Second (00-59) </td></tr>
<tr><td align="center"> \%U </td><td> Week number (00-53), Sunday as 1st day of week </td></tr>
<tr><td align="center"> \%w </td><td> Day of week (0-6), 0=Sunday </td></tr>
<tr><td align="center"> \%W </td><td> Week number (00-53), Monday as 1st day of week </td></tr>
<tr><td align="center"> \%x </td><td> Preferred date representation </td></tr>
<tr><td align="center"> \%X </td><td> Preferred time representation </td></tr>
<tr><td align="center"> \%y </td><td> Year without century (00-99) </td></tr>
<tr><td align="center"> \%Y </td><td> Year with century </td></tr>
<tr><td align="center"> \%Z </td><td> Time zone name, empty if no time zone </td></tr>
</table>
Depending on platform of runtime there may be more specifiers supported. Please refer to
documentation of strftime function in C-runtime of compiler used to build B++ library.
@param tm Localized time point from C-timestamp (vector of 9 elements like result of localtime) or C-timestamp (like result of time() function).
@return Converted date and time information on success, otherwise empty string.
@note In case of invalid argument type exception of string type is thrown.
@see localtime, time, ctime
*/
string strftime(const string fmt, var tm);
/// @}

/** @addtogroup RandFuncs Random number functions
@{
*/
/** Get random number.
Function computes a pseudo-random number. It behaves like same named C function.
@return random number between 0 and 0x7FFF
@note Multiple calls to this function will produce an approximately uniformly distributed
series of random numbers. However, random number generator should be initialized
using #srand function to prevent from generating same series for each run of your program.
@see srand
*/
int rand();

/** Initialize random number generator.
Function initializes random number generator by value of <i>seed</i>.
It behaves like same named C function.
@param seed value to initialize random number generator with
@note On start of program random number generator is initialized by value 1.
In order to get "really random" series of random numbers, usually a time depended
value for <i>seed</i> is used in calls to this function.
@par Random number generator example:
@code
   main(;i,j,l,seed)
   {
      // create 5 different series of random numbers
      for (l = 0; l< 5; ++l)
      {
         seed = timer(); // use elapsed time for initialization
         srand(seed);    // initialize generator
         print("initial value; ",seed,"\n");
         // create series of 100 numbers and print
         // 10 numbers per line
         for (i=0; i<10; ++i)
         {
            for (j=0; j<10; ++j) print(rand()," ");
            print("\n");
         }
      }
   }
@endcode
@see rand, timer, print
*/
void srand(int seed);

/// @}

/** @addtogroup ProgFuncs Program related functions
@{
*/

/** Get command line arguments passed to B++.
Function builds a vector of all command line arguments passed to B++ on run.
When B++ command line version is running, first argument is always the program
name itself (e.g. bp2.exe).
@return Vector of all command line arguments. All elements of returned
vector are of <tt>string</tt> type.
@note When B++ is used as a library, i.e. inside any program, owner program is resposnible
for passing arguments to B++ if needed.
@see getusrargs, BppVirtualMachine::execute
*/
vector getargs();

/** Get command line user arguments passed to B++.
Function builds a vector of all command line user arguments passed to B++ on run.
First element of result vector is first command line argument behind hash sign(#)
in command line.
@return Vector of all command line user arguments. All elements of returned
vector are of <tt>string</tt> type.
@note When B++ is used as a library, i.e. inside any program, owner program is resposnible
for passing arguments to B++ if needed.
@par Example for using getargs and getusrargs functions:
@code
   main(;vec1,vec2,n,i)
   {
      vec1 = getargs();    // get all arguments
      vec2 = getusrargs(); // get user arguments
      print("args:\n");
      n = vecsize(vec1);
      for (i=0;i<n;++i) print(vec1[i],"\n");
      print("usrargs:\n");
      n = vecsize(vec2);
      for (i=0;i<n;++i) print(vec2[i],"\n");
   }
@endcode
@see getargs, BppVirtualMachine::execute
*/
vector getusrargs();

/** Get number of function arguments.
Function retrieves number of arguments passed to calling function.
@return number of arguments of caller
@note If called from <tt>main</tt> function  in a command line program, all command line arguments (including programm name)
are passed as string parameters to function, i.e. arguments are same as result of #getargs function.
@see arg, getargs getusrargs
*/
int argcnt();

/** Get function argument of caller by index.
Function returns argument of specified index from calling function. It may
be used in conjunction with #argcnt to construct functions of variable sized
argument lists.
@param idx index of argument to get
@return callers function argument for index <i>idx</i>. If argument is <tt>const</tt>, result is <tt>const</tt> too.
@note In case of invalid <i>idx</i> value an error is reported.
@par Example for using arg and argcnt functions
@code
   // define a function
   var f(var x)   // only first argument has got a name
   {
      uint n = argcnt();            // get number of arguments passed
      print("\n",n," arguments:\n");
      for (uint i=0;i<n;++i)
         print("\t",arg(i));   // output argument values
      print("\n---\n");
      return x; // return value of named (first) argument
   }

   main()
   {
      print(f(1,3,"test"));
      // should print
      // 3 arguments:
      //    1  3  test
      // ---
      // 1
   }
@endcode
@see argcnt
*/
var arg(int idx);

/** Gets vector of functions actual parameters.
Function returns vector of all arguments passed to calling function. It may
be used in conjunction with #call function in order to pass arguments
to another function.
@return Vector containing all actual parameters  of caller. If #argv is called
from an element function, first element of result is the calling object. Constant state of arguments is preserved in result vector.
@see argcnt, arg
*/
vector argv();

/** Calls function <i>f</i> with actual parameters from <i>args</i>.
Function does a generic call to another function passing elements of <i>args</i>
as actual parameters to called function. It is mainly intended to dispatch
variable number of arguments to any function.
@param f Function or functor object to be invoked
@param args Vector of arguments to pass to <i>f</i>.
@return Returns return value of <i>f</i> if any, otherwise result is undefined.
@note Function may be used for call to element functions of objects too. In that case
first element of <i>args</i> must be the object itself.<br>
If <i>f</i> is an object, it must define <tt>operator()</tt> (or element function <tt>OP_CALL</tt>).
@par Examples for using call function
<b>Building a C-like printf function:</b>
@code
int printf(string format)
{
   vector v = argv();
   print(call(newstring,v));
   return int(size(v))-1;
}
@endcode
<b>Simple implementation of a .NET-like delegate</b>
@code
class Delegate
{
private:
   object _obj;
   function _func;
public:
   Delegate(object obj, function func)
   {
      _obj = obj;
      _func = func;
   }
   operator ()() {
     vector v = argv();
     if (_obj) v[0] = _obj; // Object is first parameter for call.
     else verase(v,0,1);
     return call(_func,v);
  }
}
@endcode
*/
var call(var f, vector args = null);

/** Compile source code at runtime.
Function compiles source code at runtime and integrates it to running program.
@param src string containing B++ source code to compile
@param srcname Name of source. If specified <i>srcname</i> is included in
debug output in case of an error during compiling.
@param debug If <i>true</i>, generated bytecode contains additional information about local
variable names and line nmbers in source code for debugging purposes.
@return <tt>true</tt> if <i>src</i> was compiled successfully, otherwise <tt>false</tt>
@note <ul>
<li>Other than <tt>\#include</tt> processing instructon code is compiled and integrated
at runtime.</li>
<li>This function may be called anywhere in your program. However, result is mapped into
globally context replacing already existing symbols of same names (if any) by new
versions. That allows modifying behavior of running programm (e.g. depending on
any conditions).</li>
</ul>
@par Example:
@code
   // define a test function
   testfunc(n) {
      print("testfunc(\"",n,"\") called.\n");
   }

   main(;fstr) {
      // call testfunc
      testfunc("Hello World");
      // create source of alternate function
      fstr = 
		  R"(testfunc(n) {
			   print("alternate version of testfunc called:",n,"\n");
			 })";
      if (compile(fstr))
         testfunc("Hello World");
   }
@endcode
@see compile(FILE,string), loadmodule, compilefn
*/
bool compile(const string src, const string srcname = "", bool debug = false);

/** Compile source file at runtime.
Function compiles source file at runtime and integrates it to running program.
@param fp file opened for binary read
@param srcname Name of source. If specified <i>srcname</i> is included in
debug output in case of an error during compiling.
@param debug If <i>true</i>, generated bytecode contains additional information about local 
variable names and line nmbers in source code for debugging purposes.
@return <tt>true</tt> if contents of <i>fp</i> was compiled successfully, otherwise <tt>false</tt> 
@note <ul>
<li>Other than for <tt>\#include</tt> processing instructon code is compiled and integrated
at runtime.</li>
<li>This function may be called anywhere in your program. However, result is mapped into
globally context replacing already existing symbols of same names (if any) by new
versions. That allows modifying behavior of running programm (e.g. depending on
any conditions).</li>
</ul>
@see compile(string,string), loadmodule, compilefn
*/
bool compile(FILE fp, const string srcname = "", bool debug = false);

/** Compiles source of a single function or element function at runtime.
Function compiles source code of a function body at runtime without integration into running program.
@param src Source code of function body to compile.
@param args Contents of function argument list in same notation as for a 'normal' function decvlaration.
@param rtype Name of return type of function. Empty value means no specific type.
@param ownerClass Type or class name of owner class, i.e. class to create element function for. If
not specified, a simple function is created. In case of class name, argument may be <ii>const</ii>.
@param srcname Name of source. If specified <i>srcname</i> is included in
debug output in case of an error during compiling.
@param debug If <i>true</i>, generated bytecode contains additional information about local
variable names and line nmbers in source code for debugging purposes.
@return Compiled function (or element function) if contents of <i>src</i> was compiled successfully, otherwise <tt>null</tt>.
@note <ul>
<li>Function allows dynamic craetion of any function at runtime and assign it e.g. to a variable or call it directly without 
sideeffects for code of running program.</li>
<li>Inside body (contents of <i>src</i>) all globally defined symbols (but not local symbols of caller) and
members of ownerclass (if specified by <i>ownerClass</i>) can be accessed.</li>
<li>If a member function is created, result must be called in context of an object of 
<i>ownerClass</i> type.</li>
</ul>
@par Example
@code
   // Evaluate expressions at runtime
   var evalExpr(string expr)
   {
      var f = compilefn("return ("+expr+");");
      if (f) return f();
      return null;
   }
   void main()
   {
      while (true)
      {
	     print("Expression: ");
	     string s = gets();
		 if (size(s) == 0) break;
		 print(" Result: ",evalExpr(s),"\n")
	  }
   }
@endcode
 compile(string,string), loadmodule
*/
function compilefn(const string src, const string args = "", const string rtype = "", var ownerClass = null, const string srcname = "", bool debug = false);

/** Parses a literally expression or single literally value from string.
@param src Source string to parse
@param srcname Source name for use in error messages
@return Result of expression in <i>src</i>
@note If parsing fails, i.e. in case of invalid expression, exception of type <tt>string</tt> containing error description is thrown.
*/
var litexpr(const string src, const string srcname= "");

/** Load module at runtime.
Function loads a pre-compiled byte code module at runtime and integrates it to running program.
@param filename Name of file containing module
@return <tt>true</tt> on success, otherwise <tt>false</tt>
@note Other than for <tt>\#use</tt> processing instructon code is integrated
at runtime.<br>
This function may be called anywhere in your program. However, result is mapped into
globally context replacing already existing symbols of same names (if any) by new
versions. That allows modifying behavior of running programm (e.g. depending on
any conditions).
@see compile(string,string), compile(FILE, string)
*/
bool loadmodule(const string filename);

/// @}

/** @addtogroup SysFuncs System functions
@{
*/

/** Get address of value data.
Function returns address of data for value <i>val</i>. It is intended for
use with system specific functions which require variable pointers for arguments.
@param any value or variable of any type except <tt>null</tt>
@return pointer representing value address. The returned value depends on
data type of <i>any</i> as follows:
<ul>
  <li>For types <tt>string</tt>, <tt>buffer</tt> and <tt>charbuffer</tt>
  function returns start address of data area. The result can be interpreted
  as an array or string in C-style.</li>
  <li>For all numeric types current function does same as explicit conversion into <tt>intptr</tt> type.</li>
  <li>In case of any other type, value itself is returned as a pointer.
  This allows passing of reference data (e.g. file pointer or object) to
  native functions.</li>
</ul>
*/
intptr addr(const var any);

/** Does raw copy of specified block in memory.
Function behaves like same named C function, i.e. it copies <i>n</i> bytes
from <i>src</i> to <i>dst</i>, where overlapping of memory areas is allowed.
@param dst Destination address of copy operation
@param src Source address of copy operation
@param n Number of bytes to copy
@return Destination address (same as <i>dst</i>)
@note
<ul>
<li>Function does implicit cast of argument types.</li>
<li>On DOS platforms function allows copying blocks larger than 64kB too.</li>
<li>In case of invalid arguments or an error while copy operation, it 
throws exception of <tt>string</tt> type.</li>
<li><b>Caution:</b> Function allows manipulation of any memory area and
may crash system!</li>
</ul>
*/
intptr memmove(intptr dst, intptr src, uint n);

/** Runs a command line as a synchronous child process.
Function invokes command line interpreter with specified command line and returns execution result of call.
@param cmd Command line passed to system. Line may contain name (and optionally path) of programm to run
or name of internal command followed by any parameters.
@param ignoreErr If true (default) call will accept return value -1, otherwise exception of string type is thrown in that case.
@return Execution result value of command.
@note Function is not available in Windows-CE builds. Furthermore call fails, if there is no command shell specified
by  environment variable (COMSPEC on DOS/Windows, SHELL on UNIX systems) of system (e.g. on DIP-DOS of Atari Portfolio).<br>
If call fails, function throws exception.
*/
int system(const string cmd, bool ignoreErr = true);

/** Sets or gets current runtime localization scheme.
Function behaves like same named C function (and internally uses that function).<br>
@param category Category to set/get localization parameters for. Valid values are
<ul>
<li><tt>LC_ALL</tt> (0): all categories</li>
<li><tt>LC_COLLATE</tt> (1): string comparison</li>
<li><tt>LC_CTYPE</tt> (2): character fuctions</li>
<li><tt>LC_MONETARY</tt> (3): currency format</li>
<li><tt>LC_NUMERIC</tt> (4): numeric formats</li>
<li><tt>LC_TIME</tt> (5): time format</li>
</ul>
@param locale Locale string, e.g. "en-US", "German_Germany.437", ".OCP" or "C".<br>
Empty string means 'System default settings', <tt>null</tt> returns current settings
without any changes.
@return String of settings after the call or emty string in case of an error.
@note Function (especially valid values for <i>locale</i> string) depends on Operating System.<br>
On DOS-Platforms it is useless, because only "C"-locale is available.
*/
string setlocale(int category, const string locale);


/** @name Environment access functions 
@note <p>For Windows CE builds environment is emulated in registry
(key HKLM\\Environment). There is only one globally defined environment, i.e. 
modification of an environment variable may affect other processes too.</p>
<p>In DOS and Windows Desktop builds process local environment area is used.</p>
*/
///@{

/** Gets dictionary containing copy of environment variables.
@return Ordered dictionary containing copies of all environment variable values.
Keys are the names of the environment variables.
@par Example
@code
   // list names and values of all environment variables:
   void main()
   {
      auto env = getenv();
      vector names = dictgetkeys(env);
      uint n = size(names);
      for (uint i=0;i<n;++i) print(names[i],"=",env[names[i]],"\n");
   }
@endcode
@note Function won't work for Windows CE (always returns empty dictionary).
@see getenv(string), setenv
*/
dictionary getenv();

/** Get value of specified environment variable.
Function returns value of environment variable <i>varname</i>.
@param varname Name of environment variable to get value from.
@return Copy of environment variable value if it exists or
empty string otherwise.
@par Example
@code
   // list all parts of "Path" environment variable:
   void main()
   {
      string path = getenv("Path");
      vector parts = strsplit(path,";");
      uint n = size(parts);
      for (uint i=0;i<n;++i) print(parts[i],"\n");
   }
@endcode
@see getenv(), setenv
*/
string getenv(const string varname);

/** Set value of an environment variable.
Function defines a new environment variable or changes value of an existing one.
@param varname Name of environment variable so define or change.
@param value Value to set for environment variable. If specified value is an empty string,
variable is removed from environment.
@return <tt>true</tt> on success, otherwise <tt>false</tt>
@see getenv
*/
bool setenv(const string varname, const string value);
///@}

/** @name Error reporting functions
*/
///@{

/** Gets code of last system error.
Function reads errno value of runtime.
@return Code of last error occurred while a runtime function call.
@note Function may be used to get error information in case of failure of a call to function in
runtime library, e.g. an I/O-function.
@see strerror(), errno(int)
*/
int errno();

/** Sets code of last system error.
Function sets errno value of runtime and returns changed value.
@param err New value of system error code
@return Value of system error code after setting (should be same as <i>err</i>).
@note Function may be used to reset pending system error (e.g. after a numerical overflow)
@see errno()
*/
int errno(int err);

/** Gets number of system error codes.
@return Number of error codes defined by system.
@note Function may be used to validate an error code or enumerate system error messages.
@par Example for enumerate system error codes:
@code
void main()
{
    int n = nerr();
    for (int i=0;i<n;++i)
       print(i,"\t",strerror(i),"\n");
}
@endcode
@see strerror(int errcode)
*/
int nerr();

/** Gets last system error message.
Function calls errno() and returns error message accociated to its result.
@return Description of last error occurred while a runtime function call.
@note Function may be used to get error information in case of failure of a call to function in
runtime library, e.g. an I/O-function.
@see errno(), strerror(int errcode)
*/
string strerror();

/** Gets error message for specified code.
Function returns error message accociated to <i>errcode</i>.
@param errcode Code of error to get description for
@return Description of specified error code. On invalid coed it returns "Unknown error".
@see errno(), strerror()
*/
string strerror(int errcode);

///@}

/** @name Dynamic library (DLL) functions.
@note For DOS builds handling of Dynamic Link Libraries (DLLs) is
supported using Kiez<i>Soft</i> DOSDLL library. Please refer to
documentation of this library for restrictions and detailed using
instructions.
*/
///@{

/** Load a dynamic library.
Function loads dynamic library specified by <i>dllname</i> into current process
and returns a handle to it.
@param dllname Name of library. The name specified is the file name of library module (including extension if any).
It may contain a relative or absolute path specification.
@return handle to loaded library or 0 (if not found)
@note <ul><li>The returned handle is needed as argument for following calls to
<tt>GetProcAddress</tt> and <tt>FreeLibrary</tt> functions.</li>
<li>A library loaded by <tt>LoadLibrary</tt> function should be released using
<tt>FreeLibrary</tt> if no longer needed.</li>
<li>If no absolute full path is specified in <i>dllname</i>, a system dependend serach for 
library is done. Please refer to platform documentation (or DOSDLL manual for DOS builds)
for search strategy used.</li>
<li>If <i>dllname</i> contains an asterisk (*), the asterisk character is replaced by build specific name 
extension, consisting of following (optional) parts:
<ul>
   <li>u : marks UTF-16 build</li>
   <li>d : marks Debug build</li>
   <li>_64 : marks build for 64 Bits platform</li>
</ul>
</li>
</ul>
@par Example for using dynamic library functions:
@code
   // This sample program works for Win32 Desktop platforms.
   // For Windows CE load coredll.dll instead of user32.dll .
   #define MB_ICONINFORMATION 0x40
   // function shows a simple message box
   ShowMessage(string title, string msg)
   {
      intptr hinst = LoadLibrary("user32.dll");
      // ? will be replaced by 'W' or 'A', depending if unicode build or not
      intptr fptr = GetProcAddress(hinst,"MessageBox?");
      // use addr function to get C-character pointers from strings
      CallLibFunc(fptr,0,addr(msg),addr(title),MB_ICONINFORMATION);
      FreeLibrary(hinst);
   }

   main()
   {
      ShowMessage("Hello World","This is a message\nfrom B++.");
   }
@endcode
@see FreeLibrary, GetProcAddress, CallLibFunc, addr
*/
intptr LoadLibrary(const string dllname);

/** Release dynamic library.
Function releases a dynamic library previously loaded by LoadLibrary function.
@param hdll Handle of library to release.
@return 1 on success, otherwise (i.e. on invalid handle) 0
@see LoadLibrary
*/
int FreeLibrary(intptr hdll);

/** Get address of function or data area exported from dynamic library.
Function returns address of symbol <i>name</i> (function or data area) exported by
a dynamic library.
@param hdll library handle
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>In Windows builds <i>name</i> may contain a question sign (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
</ul>
@see LoadLibrary, CallLibFunc
*/
intptr GetProcAddress(intptr hdll, const string name);

/** Call a function inside dynamic library.
Function invokes library function at address specified by <i>procaddr</i>.
@param procaddr Adress of function to invoke. Use #GetProcAddress to get address from
function name.
@param ... Comma separated list of parameters needed by called function.
@return result of function call
@note<ul>
   <li>Because B++ knows nothing about function called inside the library, there is
   no way to do any checks. So programmer is responsible to take care about
   passing correct parameters.</li>
   <li>All parameters must be passed as values of same width as <tt>intptr</tt> type,  values
   convertible to <tt>intptr</tt> may be passed directly where for all other types addresses should be used.</li>
   <li>On Windows x86 or DOS platforms functions inside dynamic library must be implemented using C- or STDCALL-calling conventions. 
   Furthermore they should return a result value of <tt>intptr</tt> size.</li>
   <li>Meaning of returned value depends on implemantation of library function. If library
   function returns nothing, result is undefined.</li>
   <li>Number of function arguments is restricted to a maximum of 16 for platforms other than Windows x86 or DOS.</li>
</ul>
@par Example:
In the following we assume a dynamic library "database.dll" implementing some database access functions.
This library exports a login function of following C-signature:
@code
   int login(const char* userName, const char* passwd);
@endcode
This function needs two parameters (user name and password). On success it returns a session id for database access,
otherwise 0.<br>
Sample program below shows all steps for call the function.
@code
   // Example for DLL function call
   void main()
   {
      // first load dynamic library
      intptr hdll = LoadLibrary("database.dll");
      // get address of exported function
      intptr proc = GetProcAddress(hdll,"login");
      // now ask user for her/his name and password
      print("Enter your name: ");
      string name = gets(); 
      print("Enter your password: ");
      string pwd = gets();
      // try to login
      // dont forget to pass addresses of strings
      intptr session = CallLibFunc(proc,addr(name),addr(pwd));
      if (session)
      {
         // do anything
      }
      else
         print("Database access denied!\n");
      // release library
      FreeLibrary(hdll); 
   }
@endcode
@see CallCLibFunc, LoadLibrary, GetProcAddress, FreeLibrary, addr
*/
intptr CallLibFunc(intptr procaddr, ...);
///@}

/** @name Callback support functions.
Beginning at version 1.5 B++ provides two sets of native callback functions intended to use as callback arguments 
of system API functions.<br>
Callbacks are handled by user defined functor objects, i.e, by object instance providing <tt>operator()</tt>.<br>
First set assumes last callback argument to be user data containing pointer to the functor object. This is the
typical callback type used mostly by Win32 API functions. This set allows callbacks up to 8 arguments 
(including functor pointer) and any number of callbacks.<br>
Second set supports callbacks without user data. Therefore functor used for callback handler must be registered from
client code. This set is restricted to a maximum of 4 arguments per callback and 5 callbacks of same type (i.e. number of arguments)
per process.<br>
In both cases prototypes of native callback functions assume all parameters and return value to be of <ttintptr</tt> type.
This restricts size of arguments and return value to pointer size, i.e., 4 bytes or 8 bytes (in 64-Bit builds).
@note In DOS builds callback support is disabled by default, because system calls usually
are done by interrupt calls and B++ is not intended to write interrupt service routines. If needed, B++ must be compiled
with defined symbol USECALLBACK in compiler options.
*/
///@{

/** Gets pointer to native callback function for use as parameter of system functions.
Each of these functions assumes last argument to be a B++ functor, i.e, an object instance providing call operator.
When callback is invoked from system function, functor is called with callback arguments (without last argument).<br>
Therefore system function must support user data pointer and call the callback with user data as last argument.<br>
Return type and type of all parameters for callback functions are assumed to be if <tt>intptr</tt> type, i.e. they have
size of native pointer type. From view of client code all simple types up to size of <tt>intptr</tt> are usable for arguments and return type.
Values of reference types must be passed as adresses, i.e. using <tt>addr()</tt> function.
@param numArgs Number of arguments for callback, including functor object. Currently a value between 1 and 8 is allowed.
@return Address of callback function.
@par Example: Implementation of a windows enumerator for Win32 API
@code
// Functor for callback
class WinEnumerator {
   private:
     charbuffer txtbuf = null;
     intptr func = 0;
   public:
   WinEnumerator() {
      txtbuf = newcbuffer(256); // initialize buffer
      func = GetUser32Func("GetWindowText?"); // get address of GetWindowText function
   }
   // call operator
   uint operator()(intptr hwnd) {
      // get window text for handle
      CallLibFunc(func,hwnd,addr(txtbuf),size(txtbuf)-1);
      auto str = string(txtbuf); // convert buffer to string
      if (size(str) > 0)
         print(ulong(hwnd),":",str,"\n");
      return 1;
   }
};

void main() {
    intptr pfunc = GetUser32Func("EnumWindows");
    // get callback function of 2 arguments (first for window handle, second for enumerator object)
    intptr pcallback = GetCallbackFunc(2);
    WinEnumerator we; // build instance of functor object
    // call EnumWindows with callback function and address of functor object
    CallLibFunc(pfunc,pcallback,addr(we));
}
@endcode
*/
intptr GetCallbackFunc(uint numArgs);

/** Gets pointer to static native callback function for use as parameter of system functions.
@param numArgs Number of arguments for callback. Currently a value between 0 and 4 is allowed.
@param idxInstance Index of instance of callback function; value between 0 and 4 allowed.
@return Address of callback function.
@note Before use of callback, client code must install a functor object by call to SetStaticCallbackhandler.
@see SetStaticCallbackHandler
*/
intptr GetStaticCallbackFunc(uint numArgs, uint idxInstance);


/** Installs functor object for one of static callback functions.
@param uint numArgs Number of arguments for callback. Currently a value between 0 and 4 is allowed.
@param idxInstance Index of instance of callback function; value between 0 and 4 allowed.
@param handler Functor object used to handle callback. Argument can be an <tt>object</tt>, a weak reference to object or address
(type intptr) of object.
@note  <ul>
<li>If <i>handler</i> is <tt>object</tt>, its reference counter is incremented, i.e. it can't be released
until it is unregistered, either by set another handler or call SetStaticCallbackHandler with <tt>null</tt> as <tt>handler</tt> argument.</li>
<li>If the <i>handler</i> is a weak reference or an address, the client code must make sure that the handler exists for as long as it is needed.</li>
<li>In case of invalid values for <i>numArgs</i> or <i>idxInstance</i> an exception of <tt>string</tt> type is raised.</li>
</ul>
@see GetStaticCallbackFunc
*/
void SetStaticCallbackHandler(uint numArgs, uint idxInstance, var handler);
///@}



/** @addtogroup WINFuncs Windows specific system functions.
All functions of this section are available in Windows builds only.
@{
*/
/** Get address of function or data area exported from user32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetUser32Func(const string name);

/** Get address of function or data area exported from shell32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetShell32Func(const string name);

/** Get address of function or data area exported from gdi32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetGdi32Func(const string name);

/** Get address of function or data area exported from kernel32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetCoreFunc
*/
intptr GetKernel32Func(const string name);

/** Get address of function or data area exported from ole32.dll library.
Function returns address of symbol <i>name</i> (function or data area) exported by
the library. It is intended to simplify using of Windows-API functions.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress
*/
intptr GetOle32Func(const string name);

/** Alias of GetUser32Func.
Because in most functions of WindowsCE-API are contained in coredll.dll library,
there are no user32, shell32, kernel32 and gdi32 libraries in WindowsCE.<br>
This function alias is intended to be used in programms running on WinCE only.
@param name name of exported symbol
@return address of symbol on success, otherwise 0
@note <ul>
   <li>Returned function address can be used for argument of #CallLibFunc call.</li>
   <li>In case of a data area returned address points to start of area.</li>
   <li>If Function fails, i.e. if requested symbol doesn't exist, 0 is returned.</li>
   <li>The <i>name</i> argument may contain a question mark (?). This sign
   is replaced by 'W' in unicode versions and by 'A' in non-unicode versions.</li>
   <li>In WindowsCE builds symbol is retrieved from coredll.dll library.</li>
</ul>
@see LoadLibrary, CallLibFunc, GetProcAddress, GetUser32Func
*/
intptr GetCoreFunc(const string name);

/** Get last error code.
Function simply wraps Windows API function of same name.
@return Last error code of calling thread.
@see GetSysErrorMsg
*/
uint GetLastError();

/** Get readable message for system error code.
Function retrieves error message for last system error code.
@return Error message for last system error code, empty string if no message available.
@note GetLastError is called internally to get error code.
@see GetLastError
*/
string GetSysErrorMsg();

/** Get readable message for system error code.
Function retrieves error message for specified code.
@param errcode Error code to get message for (optional)
@return Error message for specified code, empty string if no message available.
@see GetLastError
*/
string GetSysErrorMsg(uint errcode);

/// @}
/** @addtogroup DOSFuncs DOS specific system functions.
All functions of this section are available in DOS builds only.
@{
*/



/** Get size of available memory. 
Function returns amount of free memory on DOS heap.
@return number of free bytes in memory
*/
int memsize();

/** Set register value.
Function sets value of register specified by <i>regname</i>.
@param regname Name of register to set; possible names are:<br>
<tt>ax, bx, cx, dx, di, si, al, ah, bl, bh, cl, ch, dl, dh,
   cflag, flags, ds, es, ss, cs</tt>.
@param value New value of specified register. For 8-bit registers
lowest byte and for 16-bit registers low word of <i>value</i> is used.
@return null
@note <ul>
   <li>Register names in <i>regname</i> argument are not case sensitive, i.e. 
   "AX" means same as "ax".</li>
   <li>Function stores values in internal register structure rather immediately
   modifying processor registers. Contents of this structure is applied to
   processor registers on calls to #int86 and #int86x functions.</li>
</ul>
@see getreg, segread, int86, int86x
*/
null setreg(const string regname, int value);

/** Get register value.
Function gets value of register specified by <i>regname</i>.
@param regname Name of register to get value from; possible names are:<br>
<tt>ax, bx, cx, dx, di, si, al, ah, bl, bh, cl, ch, dl, dh,
cflag, flags, ds, es, ss, cs</tt>.
@return value of specified register
@note <ul>
<li>Register names in <i>regname</i> argument are not case sensitive, i.e. 
"AX" means same as "ax".</li>
<li>Function reads values from internal register structure rather directly reading 
processor registers. Contents of this structure is modified explicitely using
functions #setreg or #segread and implicitely by #int86 and #int86x
functions.</li>
</ul>
@see setreg, segread, int86, int86x
*/
int getreg(const string regname);

/** Read segment registers.
Function reads current values of processors segment registers (DS, ES, SS, CS)
and stores them into internal register data structure.
@return null
@note Use #getreg to access values from your program.
@see getreg, setreg
*/
null segread();

/** Do x86 interrupt call.
Function does a call to x86 interrupt routine specified by <i>intnr</i>. Before call
is done it loads processor registers AX, BX, CX, DX, DI, SI and FLAGS with values from
internal register data structure. After call it saves all registers into iternal structure
and restores original values of processor registers.
@param intnr interrupt number
@return value of AX register after interrupt call
@see setreg, getreg, int86x
*/
int int86(int intnr);

/** Do extended x86 interrupt call.
Function does a call to x86 interrupt routine specified by <i>intnr</i>. Before call
is done it loads processor registers AX, BX, CX, DX, DI, SI, DS, ES and FLAGS with values from
internal register data structure. After call it saves all registers into iternal structure
and restores original values of processor registers.
@param intnr interrupt number
@return value of AX register after interrupt call
@note Before call to this function values of processors segment registers should be
transferred to internal register structure using #segread function.
@par Example:
@code
   // interrupt call example - prints current directory path 
   main(;buf,bufptr)
   {
      buf = newstring(200);       // buffer for result 
      segread();                  // init variables for seg-registers
      setreg("ah",0x47);          // dosfunc nr.
      setreg("dl",0);             // get path for current drive 
      bufptr = addr(buf);         // get address of buffer contents
      setreg("ds",bufptr >> 16);  // high word of buffer address -> segment
      setreg("si",bufptr & 0xFF); // low word of buffer address -> offset
      int86x(0x21);               // call DOS-Interrupt
      print(buf,"\n");
   }
@endcode
@see setreg, getreg, int86, segread
*/
int int86x(int intnr);

/** @name Port I/O functions */
//@{

/** Read byte from I/O port.
Function reads single byte from I/O port specified by <i>portid</i>.
@param portid id (number) of I/O port to read byte from
@return byte read
*/
int inportb(int portid);

/** Write byte to I/O port.
Function writes single byte (lowest byte of <i>value</i>) to port <i>portid</i>.
@param portid id (number) of I/O port to write byte to
@param value value to write
@return null
*/
null outportb(int portid, int value);

/** Read word from I/O port.
Function reads 16 bits word from I/O port specified by <i>portid</i>.
@param portid id (number) of I/O port to read word from
@return word read
*/
int inport(int portid);

/** Write word to I/O port.
Function writes 16 bits word (low word of <i>value</i>) to port <i>portid</i>.
@param portid id (number) of I/O port to write word to
@param value value to write
@return null
*/
null outport(int portid, int value);
//@}

/** @name Graphic BIOS functions.
Functions of this section allow rudimentary graphic programming using BIOS interrupt calls.
If more functions needed, they may be implemented via #int86 or #int86x calls.
*/
//@{
/** Set screen mode.
Function clears the screen and switches display into specified <i>mode</i>. After call
cursor position is (0,0).
@param mode Screen mode to switch display into. Meaning of value depends on display adaptor.
Usually mode 7 specifies monochrome text mode (MDA).
@return display mode used by adaptor after call.
@see setpixel, line
*/
int setscrmode(int mode);

/** Set pixel in graphics mode.
Function sets a single graphics pixel to specified <i>color</i>.
@param x x-coordinate of pixel to set
@param y y-coordinate of pixel to set
@param color new color value to set for specified pixel
@return null
@note <ul>
  <li>Display must be switched into a graphics mode for use this function.</li>
  <li>Coordinates of pixel both should be non-negative values. Maximum possible
  coordinate values are determined by resolution of screen (depending on selected
  graphics mode.<br>The coordinate pair (0,0) specifies upper left corner of screen.</li>
  <li>Because maximum color depth of BIOS graphic modes is restricted to 125, only lowest byte 
  of <i>color</i> argument is significant.</li>
</ul>
@see setscrmode, line
*/
null setpixel(int x, int y, int color);

/** Draw a line.
Function draws a graphics line from point (<i>x1</i>,<i>y1</i>) to point (<i>x2</i>,<i>y2</i>).
@param x1 x-coordinate of start point
@param y1 y-coordinate of start point
@param x2 x-coordinate of end point
@param y2 y-coordinate of end point
@param color color to use for drawing the line
@return null
@note <ul>
<li>Display must be switched into a graphics mode for use this function.</li>
<li>Coordinates of pixel both should be non-negative values. Maximum possible
coordinate values are determined by resolution of screen (depending on selected
graphics mode.<br>The coordinate pair (0,0) specifies upper left corner of screen.</li>
<li>Because maximum color depth of BIOS graphic modes is restricted to 125, only lowest byte 
of <i>color</i> argument is significant.</li>
</ul>
@par Example for using graphic functions:
@code
   // fill screen with lines
   main()
   {
      // Resolution is restricted to 240x64 pixels here. This allows running the
      // sample on ATARI-Portfolio too.
      var px=240;  // pixels horizontal
      var py=64;   // pixels vertical
      // switch to graphic mode  
      setscrmode(4);  // should work on most display adapters
      // get current time 
      var ms = timer();
      var n=0;
      // draw some lines 
      for (var x=0;x<px;x+=2)
      {
         n++;
         var x2= px -x;
         line(x,0,x2,py-1,1);
      }
      for (var y=0;y<py;y+=2)
      {
         n++;
         var y2= py-1 -y;
         line(0,y,px-1,y2,2);
      }
      // get elapsed time 
      ms = timer() - ms;
      var r = n*1000 / ms;
      gets();
      // switch back to text mode 
      setscrmode(7);
      print(r, " lines per second\n");
   }
@endcode
@see setscrmode, setpixel
*/
null line(int x1, int y1, int x2, int y2, int color);
///@}

/// @}

/// @}


/// @}
