#ifndef BPPINPUT_H
#define BPPINPUT_H
#ifndef DOXY
#include <bppdef.h>
#endif
#include <bppstrng.h>
#include <stdio.h>


/** Interface for reading data character data from any input device.
Implementations of IBppInput are used by BppVirtualMachine to get
character inputs, e.g. console or any input control.
*/
class BPPLIB_DLLEXP IBppInput
{
public:
	virtual ~IBppInput() {}

   /** Gets single character from input device.
   @return Code of character read or _TEOF if no character available. 
   */
   virtual int getChar() = 0;

   /** Reads string from input device.
   @return String (i.e. line) read or empty string if no data available.
   @note Returned string should not contain trailing EOL character.
   */
   virtual BppString getString() = 0;
};

/** Character reader for file source.
This class implements IBppInput for read character data from file or opened
character device, e.g. stdin.
*/
class BPPLIB_DLLEXP BppFileStreamInput : public IBppInput
{
public:
   /// Construction from opened file or device.
   BppFileStreamInput(FILE* fp, bool autoClose = false);
   /// Construction from file name.
   BppFileStreamInput(const BppChar* fileName, const BppChar* mode = _T("rt"));
   /// dtor
   virtual ~BppFileStreamInput();
   /// Gets single character.
   virtual int getChar();
   /// Reads string.
   virtual BppString getString();
private:
#ifndef DOXY
   BppFileStreamInput(const BppFileStreamInput&);
   BppFileStreamInput& operator=(const BppFileStreamInput&);
#endif
   FILE* _fp; ///< opened file pointer
   bool _autoClose; ///< flag for close file in destructor
};

/** Character reader using callback functions.
This class implements IBppInput for read character data using callback functions.
*/
class BPPLIB_DLLEXP BppCallbackInput : public IBppInput
{
public:
   /// Construction
   BppCallbackInput(BppCharInputCallback charCallback = 0,BppStringInputCallback stringCallback = 0);
   /// Gets single character.
   virtual int getChar();
   /// Reads string.
   virtual BppString getString();
   BppCharInputCallback _charCallback; ///< character read callback function
   BppStringInputCallback _stringCallback; ///< string read callback function
};

#ifdef WIN32
#include <windows.h>
/** Character reader using Windows messages.
Class implements IBppInput for read character data MS-Windows message system.
*/
class BPPLIB_DLLEXP BppWinMsgInput : public IBppInput
{
public:
   BppWinMsgInput(HWND hwnd);
   virtual int getChar();
   virtual BppString getString();
   void setChar(char chr);
   void setChar(wchar_t chr);
   void setString(const char* str);
   void setString(const wchar_t* str);
   /// Gets ID of string input message.
   static UINT getStringInputMsg();
   /// Gets ID of character input message.
   static UINT getCharInputMsg();
private:
   HWND _hwnd;
   BppString _strBuffer;
   int _chrBuffer;
   static UINT _stringInputMsg;
   static UINT _charInputMsg;

};
#endif

#endif // BPPINPUT_H
