#ifndef BPPINT_H
#define BPPINT_H
#ifndef DOXY
#include <bppdef.h>
#endif
#include <bppvec.h>
#include <bppstrng.h>
#include <bppvalue.h>
#include <bppconst.h>

class BppVirtualMachine;
class BppDictionary;
class IBppPrinter;
class BppDebugger;


class BppInterpreter;

struct BppErrorHandler;

/** Helper class containing information about a function currently running in VM.
Instances of this class are created by BppInterpreter for support of debugging and exception handling.
*/
class BPPLIB_DLLEXP BppProcInfo
{
public:
	/// Ctor.
	BppProcInfo(BppInterpreter& interpreter, BppVector* bytecode,int nargs,ValPtr bp, ValPtr fp,bool nocaller = false) BPPNOEXCEPT;
	/// DTor.
	~BppProcInfo();
	/** Gets read access to part of Bytecode vector by index.
	@param idx Index of part to get, where IDX_CODE (0) is the code buffer, IDX_CLASS (1) the class,
			 IDX_NAME (2) the function name and IDX_DEFAULTARGS (4) vector of default argument values.
	@return Constant reference to part.
	@see BppByteCodeIndex
	*/
	inline const BppValue& operator[](unsigned idx) const
	{
		return (*_bytecode)[idx];
	}

	/** Gets access to part of Bytecode vector by index.
	@param idx Index of part to get, where IDX_CODE (0) is the code buffer, IDX_CLASS (1) the class,
	IDX_NAME (2) the function name and IDX_DEFAULTARGS (4) vector of default argument values.
	@return Reference to part.
	@see BppByteCodeIndex
	*/
	inline BppValue& operator[](unsigned idx)
	{
		return (*_bytecode)[idx];
	}

	/// Local variable types.
	enum LocalVarType {
		LVT_ARG = 0, ///< Argument
		LVT_VAR, ///< Function level (temporary) variable
		LVT_STATIC ///< Function level static variable
	};

	/// Gets access to local variables info vector (if any).
	const BppVector* getLocalVarInfo() const;

	/// Gets name of a local variable, specified by index and variable type.
	const BppChar* getLocalVarName(unsigned idx, LocalVarType lvt) const;

	/// Gets declared name of <i>idx</i>th argument.
	inline const BppChar* getArgumentName(unsigned idx) const { return getLocalVarName(idx, LVT_ARG); }
	/// Gets declared name of <i>idx</i>th local (temporary) variable.
	inline const BppChar* getLocalVarName(unsigned idx) const { return getLocalVarName(idx, LVT_VAR); }
	/// Gets declared name of <i>idx</i>th local static variable.
	inline const BppChar* getLocalStaticVarName(unsigned idx) const { return getLocalVarName(idx, LVT_STATIC); }

	/// Gets pointer to code buffer contents.
	operator BppByte*() const BPPNOEXCEPT;
	/// Gets number of arguments.
  inline int getNargs() const BPPNOEXCEPT { 	return _nargs; }
	/// Gets function name.
	const BppChar* getName() const BPPNOEXCEPT;

	inline int& frameTop() BPPNOEXCEPT { return _frameTop;  }

	inline BppValue& tosTmp() BPPNOEXCEPT { return _tosTmp; }

	/// Gets class defining function.
	BppClass* getClass() const BPPNOEXCEPT;

	/// bp points before 1st argument (i.e. to object in method calls).
	inline ValPtr& bp() BPPNOEXCEPT { return _bp; }

	/// fp points before 1st local (temporary) variable).
	inline ValPtr& fp() BPPNOEXCEPT { return _fp; }

	/** Gets Descriptor of function calling the current one.
	@return Pointer to caller function descriptor or 0, if current function has no caller.
	@note Function behaves similary to getParent function. The only difference is checking a flag signaling
	that there is no caller.
	*/
	inline BppProcInfo* getCaller() const BPPNOEXCEPT
	{
		return _nocaller ? 0 : _parent;
	}

	/** Gets parent item of function in call stack.
	@return Pointer to parent item or 0, if there is no parent.
	*/
	inline BppProcInfo* getParent() const BPPNOEXCEPT
	{
		return _parent;
	}

	/** Explicitely sets parent of current function.
	@note Function is intended for internal use and called by BppInterpreter.
	Client applications never should call it, however, debuggers may use it to implement
	manipulations of call stack.
	*/
	inline void setParent(BppProcInfo* parent) BPPNOEXCEPT
	{
		_parent = parent;
	}

	/** Pushes error handler to error errorhandler stack
	@param pc Program counter (index in bytecode buffer) of start of handler code (catch block).
	@param pStack Stack pointer value at begin of code section protected by handler (start of try block)
	@see BppErrorHandler
	*/
	void pushErrorHandler(BppNWord pc, const ValPtr& pStack);
	/// Removes top of errorhandler stack.
	bool popErrorHandler();
	/// Gets top of errorhandler stack.
	inline BppErrorHandler* getEHStack() const BPPNOEXCEPT
	{
		return _errHandlerStack;
	}

private:
#ifndef DOXY
	BppProcInfo(const BppProcInfo&);
	BppProcInfo& operator=(const BppProcInfo&);
#endif
	BppInterpreter& _interpreter; ///< Reference to interpreter.
	BppErrorHandler* _errHandlerStack; ///< Top of exception handler stack.
	BppVector* _bytecode; ///< ByteCode vector of function
	ValPtr _bp; ///< Base pointer, points before 1st argument (i.e. to object in method calls)
	ValPtr _fp; ///< Frame pointer, points before 1st local (temporary) variable)
	int _nargs; ///< Number of arguments
	int _frameTop; ///< Index of last temporary variable on stack
	BppValue _tosTmp; ///< Temporary store for TOS value
	BppProcInfo* _parent; ///< Pointer to parent entry on call stack.
	bool _nocaller; ///< Flag for extern called function (e.g. from inside a native function)
};

struct __BppProcHelper;

class BppInterpreter;
/// Type of Opcode handler function.
typedef bool (BppInterpreter::*BppOpcodeFunc)(unsigned& pc, BppUInt data);

/** B++ Bytecode Interpreter.
@note Bytecode interpreter is used by VM of B++.
Client code normally should prefer calling methods of VM instead of using interpreter directly if possible.
*/
class BPPLIB_DLLEXP BppInterpreter
{
	friend class BppVirtualMachine;
	friend struct __BppProcHelper;
public:
	BppInterpreter(BppVirtualMachine& vm) BPPNOEXCEPT;
	~BppInterpreter();

	/// Gets reference to VM owning the interpreter
	inline BppVirtualMachine& getVirtualMachine() BPPNOEXCEPT
	{
		return _vm;
	}


	/// execute entry point function
	bool execute(const BppChar* name, int argc, const BppChar* argv[]);
	/// call a function
	bool call(int nargs,BppVector* byteCode);
	/// call object method
	bool send(int nargs, const BppChar* sel = 0, const BppValue* funcVal = 0);
	// reset interpreter
	void reset();
	/// Access to error message buffer
	inline BppString& errors() BPPNOEXCEPT
	{
		return _errBuffer;
	}   
	/// Sets debugger for this interpreter.
	void setDebugger(BppDebugger* debugger);
	/// Gets information about current procedure.
	inline BppProcInfo* getProcInfo() const BPPNOEXCEPT
	{
		return _procStack;
	}
	/// Helper function for getting string from object value.
	BppString* getStringFromValue(BppClass* ownClass, BppValue& val);

	/// Helper function for getting string from object value without class context
	BppString* getStringFromValueExtern(BppValue& val)
	{
		return getStringFromValue(0, val);
	}
	/// Helper function for conversion to numeric value
	bool doNumCast(BppValue& dest, const BppValue& src, BppValueType dsttype, const BppChar* fname=_T(""));
	/// Helper function for conversion to numeric value
	bool doStrCast(BppValue& dest, const BppValue& src);
	/// Invokes any function specified by funcval.
	bool invokeFunc(BppValue& result, const BppValue& funcval, BppVector* args = 0);
	/// Invokes a global function.
	bool invoke(BppValue& result, const BppChar* name, BppVector* args = 0);
	/// Invokes member function of an object.
	bool invokeMember(BppValue& result, BppObject* pobj, const BppChar* name, BppVector* args = 0);
	/// Invokes static method or function assigned to static member variable of a class.
	bool invokeStatic(BppValue& result, BppClass* pclass, const BppChar* name, BppVector* args = 0);
	/// Creates object instance of class specified by name.
	bool createObject(BppValue& result, const BppString& className, BppVector* args = 0);
	/// Creates object instance of class.
	bool createObject(BppValue& result, BppClass* pclass, BppVector* args = 0);
	/// Builds error value and return from current procedure
	bool returnError(const BppChar* msg, const BppChar* context = 0);
	/// return error value from current procedure
	bool returnError(const BppValue& errValue);
	/// return method not found error
	bool noMethod(const BppChar* name);
	/// check value having specified data type
	bool checkType(const BppValue& val, BppValueType vtype);
	/// return invalid type error
	bool badType(BppValueType type, BppValueType expected);
	/// return invalid type error
	bool badType(const BppValue& val, const BppValue& expected);
	/// Checks if break requests are allowed.
	inline bool getAllowBreakRequest() const BPPNOEXCEPT
	{
		return _allowBreakRequest;
	}

	/// Enables/disables break requests
	inline void setAllowBreakRequest(bool val) BPPNOEXCEPT
	{
		_allowBreakRequest = val;
	}

	/// Sets cancel request flag.
	inline void cancel() BPPNOEXCEPT
	{
		_requestCancel = true;
	}

#ifndef DOXY
	static bool _num_addOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_subOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_mulOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_divOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_remOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_bandOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_borOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_xorOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_shlOp(BppValueType rt, BppValue& left, BppValue& right);
	static bool _num_shrOp(BppValueType rt, BppValue& left, BppValue& right);
#endif
private:
	/// Helper function for decoding single instruction (opcode) and its additional data.
	BppOpCode _decodeOpcode(BppByte* pcode, unsigned int& pc, BppDWord& opdata) const;
	/// Procedure entry helper function.
	void enterProc(BppVector* bytecode,int nargs,ValPtr bp, ValPtr fp);
	/// Procedure leave helper function.
	void leaveProc();
	/// Helper for typesafe assignment.
	bool safeAssign(BppValue& dest, const BppValue& src);
	/// Processes arguments in function call.
	bool handleCallArguments(int& nargs, int definedArgs, ValPtr &BP, BppVector * defargs, BppVector* funcState);
	/// Unwinds call stack on exception.
	bool processException(unsigned int &pc, bool &pending);
	/** @name Opcode handler functions */
	//@{
		/// No Operation.
		bool _OP_NOP(unsigned& pc, BppUInt data); // 0x00
		/// Branch if TOS value is <tt>true</tt>.
		bool _OP_BRT(unsigned& pc, BppUInt data); // 0x01
		/// Branch if TOS value is <tt>false</tt>.
		bool _OP_BRF(unsigned& pc, BppUInt data); // 0x02
		/// Branch without condition.
		bool _OP_BR(unsigned& pc, BppUInt data); // 0x03
		/// Set TOS value to <tt>null</tt>.
		bool _OP_NIL(unsigned& pc, BppUInt data); // 0x04
		/// Push empty value to stack.
		bool _OP_PUSH(unsigned& pc, BppUInt data); // 0x05
		/// Logical negation of value on TOS.
		bool _OP_NOT(unsigned& pc, BppUInt data); // 0x06
		/// Arithmetic negation of value on TOS.
		bool _OP_NEG(unsigned& pc, BppUInt data); // 0x07
		/// Add topmost two values on stack.
		bool _OP_ADD(unsigned& pc, BppUInt data); // 0x08
		/// Subtract value on TOS from value on TOS-1.
		bool _OP_SUB(unsigned& pc, BppUInt data); // 0x09
		/// Multiply topmost two values on stack.
		bool _OP_MUL(unsigned& pc, BppUInt data); // 0x0A
		/// Divide value on TOS-1 by value on TO.
		bool _OP_DIV(unsigned& pc, BppUInt data); // 0x0B
		/// Compute reminder of division of value on TOS-1 by value on TO.
		bool _OP_REM(unsigned& pc, BppUInt data); // 0x0C
		/// Bitwise AND of topmost two values on stack.
		bool _OP_BAND(unsigned& pc, BppUInt data); // 0x0D
		/// Bitwise OR of topmost two values on stack.
		bool _OP_BOR(unsigned& pc, BppUInt data); // 0x0E
		/// Bitwise XOR of topmost two values on stack.
		bool _OP_XOR(unsigned& pc, BppUInt data); // 0x0F
		/// Bitwise NOT of values on TOS.
		bool _OP_BNOT(unsigned& pc, BppUInt data); // 0x10
		/// Left shift of value on TOS-1 by value on TOS.
		bool _OP_SHL(unsigned& pc, BppUInt data); // 0x11
		/// Right shift of value on TOS-1 by value on TOS.
		bool _OP_SHR(unsigned& pc, BppUInt data); // 0x12
		/// Check if value on TOS-1 is lower than value on TOS.
		bool _OP_LT(unsigned& pc, BppUInt data); // 0x13
		/// Check if value on TOS-1 is lower than or equally to value on TOS.
		bool _OP_LE(unsigned& pc, BppUInt data); // 0x14
		/// Check if value on TOS-1 equals to value on TOS.
		bool _OP_EQ(unsigned& pc, BppUInt data); // 0x15
		/// Check if value on TOS-1 is not equally to value on TOS.
		bool _OP_NE(unsigned& pc, BppUInt data); // 0x16
		/// Check if value on TOS-1 is greater than or equally to value on TOS.
		bool _OP_GE(unsigned& pc, BppUInt data); // 0x17
		/// Check if value on TOS-1 is greater than value on TOS.
		bool _OP_GT(unsigned& pc, BppUInt data); // 0x18
		/// Increment value on TOS.
		bool _OP_INC(unsigned& pc, BppUInt data); // 0x19
		/// Increment value on TOS.
		bool _OP_DEC(unsigned& pc, BppUInt data); // 0x1A
		/// Load literal value to TOS.
		bool _OP_LIT(unsigned& pc, BppUInt data); // 0x1B
		/// Return from function call.
		bool _OP_RETURN(unsigned& pc, BppUInt data); // 0x1C
		/// Do function call.
		bool _OP_CALL(unsigned& pc, BppUInt data); // 0x1D
		/// Load value of a global variable to TOS.
		bool _OP_REF(unsigned& pc, BppUInt data); // 0x1E
		/// Store value on TOS into a global variable.
		bool _OP_SET(unsigned& pc, BppUInt data); // 0x1F
		/// Load indexed element of a container to TOS.
		bool _OP_VREF(unsigned& pc, BppUInt data); // 0x20
		/// Set indexed element of a container to value on TOS.
		bool _OP_VSET(unsigned& pc, BppUInt data); // 0x21
		/// Load value of a member variable to TOS.
		bool _OP_MREF(unsigned& pc, BppUInt data); // 0x22
		/// Set member variable to value on TOS.
		bool _OP_MSET(unsigned& pc, BppUInt data); // 0x23
		/// Load value of a function argument to TOS.
		bool _OP_AREF(unsigned& pc, BppUInt data); // 0x24
		/// Set function argument value to value on TOS.
		bool _OP_ASET(unsigned& pc, BppUInt data); // 0x25
		/// Load value of a temporary (local) variable to TOS.
		bool _OP_TREF(unsigned& pc, BppUInt data); // 0x26
		/// Store value on TOS into a temporary (local) variable.
		bool _OP_TSET(unsigned& pc, BppUInt data); // 0x27
		/// Reserve space for temporary variables on stack.
		bool _OP_TSPACE(unsigned& pc, BppUInt data); // 0x28
		/// Call a member function (method).
		bool _OP_SEND(unsigned& pc, BppUInt data); // 0x29
		/// Duplicate the two topmost values on stack.
		bool _OP_DUP2(unsigned& pc, BppUInt data); // 0x2A
		/// Create new object instance.
		bool _OP_NEW(unsigned& pc, BppUInt data); // 0x2B
		/// Delete object instance.
		bool _OP_DELETE(unsigned& pc, BppUInt data); // 0x2C
		/// Duplicate topmost value on stack.
		bool _OP_DUP(unsigned& pc, BppUInt data); // 0x2D
		/// Remove topmost value from stack.
		bool _OP_POP(unsigned& pc, BppUInt data); // 0x2E
		/// Build weak reference for value on TOS.
		bool _OP_MKREF(unsigned& pc, BppUInt data); // 0x2F
		/// Enter protected (try) block.
		bool _OP_TRY(unsigned& pc, BppUInt data); // 0x30
		/// End of protected (try) block.
		bool _OP_ENDTRY(unsigned& pc, BppUInt data); // 0x31
		/// Raise an exception.
		bool _OP_THROW(unsigned& pc, BppUInt data); // 0x32
		/// End of exception handler (catch) block.
		bool _OP_ENDCATCH(unsigned& pc, BppUInt data); // 0x33
		/// Begin of a <tt>switch</tt> statement.
		bool _OP_SWITCH(unsigned& pc, BppUInt data); // 0x34
		/// Load value of a static local variable to TOS.
		bool _OP_SREF(unsigned& pc, BppUInt data); // 0x35
		/// Store value on TOS into a static local variable.
		bool _OP_SSET(unsigned& pc, BppUInt data); // 0x36
		/// Postfix increment.
		bool _OP_PINC(unsigned& pc, BppUInt data); // 0x37
		/// Postfix decrement.
		bool _OP_PDEC(unsigned& pc, BppUInt data); // 0x38
		/// Load property value to TOS.
		bool _OP_PREF(unsigned& pc, BppUInt data); // 0x39
		/// Set property value to value on TOS.
		bool _OP_PSET(unsigned& pc, BppUInt data); // 0x3A
		/// Conditional POP - removes values behind top of current call frame from stack.
		bool _OP_POPC(unsigned& pc, BppUInt data); // 0x3B
		/// Positive sign operator.
		bool _OP_PLUS(unsigned& pc, BppUInt data); // 0x3C
		/// Make value from weak reference on TOS.
		bool _OP_UNREF(unsigned& pc, BppUInt data); // 0x3D
		/// Store TOS value to tmp value in procInfo
		bool _OP_STOS(unsigned& pc, BppUInt data); // 0x3E
		/// Load TOS from tmp value in procInfo
		bool _OP_LDTOS(unsigned& pc, BppUInt data); // 0x3F
		/// Make tamporary const
		bool _OP_MKCONST(unsigned& pc, BppUInt data); // 0x3F
		/// Make TOS const
		bool _OP_SPCONST(unsigned& pc, BppUInt data); // 0x3F

	/// Handles binary arithmetic ans bit operators.
	bool doBinOp(int binOpIdx);
	/// Handles compare operators.
	bool doCompare(BppClass* ownClass);
	/// Handles equality operators.
	bool doEquals(BppClass* ownClass);
	//@}
	/// copy ctor (forbidden)
	BppInterpreter(const BppInterpreter&);
	/// assignment operator (forbidden)
	BppInterpreter& operator=(const BppInterpreter&);
	/// interpreter stack
	BppVector _stack;
	/// interpreter stack pointer
	ValPtr _sp;
	/// stack of pending exceptions
	BppVector _errStack;
	/// dictionary of defined classes
	BppDictionary& _classes;
	/// dictionary of globally defined symbols
	BppDictionary& _symbols;
	/// reference to default printer
	IBppPrinter& _printer;
	/// reference to virtual machine owning this interpreter
	BppVirtualMachine& _vm;
	/// error message buffer
	BppString _errBuffer;
	/// procedure call stack
	BppProcInfo* _procStack;
	/// debugger currently used
	BppDebugger* _debugger;
	/// stores last opcode processed
	BppOpCode _lastOpCode;
	/// flag for trace execution
	bool _trace;
	/// flag for single step mode
	bool _step;
	/// flag for allowing break requests
	bool _allowBreakRequest;
	/// flag for cancel request
	bool _requestCancel;
	/// Table of opcode handler functions.
	static BppOpcodeFunc opFuncTable[];
};

#endif // BPPINT_H
