#ifndef BPPPRINT_H
#define BPPPRINT_H
#ifndef DOXY
#include <bppdef.h>
#endif
#include <bppstrng.h>
#include <stdio.h>

/** Text output interface for B++.
Implementations of IBppPrinter are used by VM for text outputs. This includes normal text outputs 
as soon as debug and error messages.
@see BppVirtualMachine, BppDummyPrinter, BppFileStreamPrinter, BppCallbackPrinter,
BppStringPrinter, BppWinMsgPrinter
*/
class BPPLIB_DLLEXP IBppPrinter
{
public:
   /// Dtor does nothing.
   virtual ~IBppPrinter() {}
   /** Formatted output operator.
   Implementation of operator should behave like printf C-function.
   @param format Format string.
   @param ... Arguments depending on <i>format</i>.
   @return  Number of characters written, or a negative value if an output error occurs.
   */
   virtual int operator()(const BppChar* format, ...) = 0;
};

/// Dummy printer (discards all output).
class BPPLIB_DLLEXP BppDummyPrinter : public IBppPrinter
{
public:
   /** Formatted output operator.
   Operator does nothing but return 0.
   */
   virtual int operator()(const BppChar* format, ...);
};

/** Output to file stream.
Class implements IBppPrinter for output into a text file.
*/
class BPPLIB_DLLEXP BppFileStreamPrinter : public IBppPrinter
{
public:
   /// Ctor using already opened file.
   BppFileStreamPrinter(FILE* fp, bool autoClose = false);
   /// Ctor using file name.
   BppFileStreamPrinter(const BppChar* fileName, const BppChar* mode = _T("wt"));
   /// Dtor.
   virtual ~BppFileStreamPrinter();
   /// Formatted output operator.
   virtual int operator()(const BppChar* format, ...);
private:
   BppFileStreamPrinter(const BppFileStreamPrinter&);
   BppFileStreamPrinter& operator=(const BppFileStreamPrinter&);
   FILE* _fp;
   bool _autoClose;
};

/** Output using a callback function.
Class implements IBppPrinter for output redirection to a user defined callback function.
*/
class BPPLIB_DLLEXP BppCallbackPrinter : public IBppPrinter
{
public:
   /// Constructor
   BppCallbackPrinter(BppPrinterCallback = 0);
   /// Formatted output operator.
   virtual int operator()(const BppChar* format, ...);
   /// Callback installed by constructor
   BppPrinterCallback _callback;
};

/** Output using a string buffer.
Class implements IBppPrinter for output to a string buffer.
*/
class BPPLIB_DLLEXP BppStringPrinter : public IBppPrinter
{
public:
   /// Constructor
   BppStringPrinter();
   /// Formatted output operator.
   virtual int operator()(const BppChar* format, ...);
   /// Gets contents of buffer.
   const BppChar* c_str() const;
   /// Gets contents of buffer as single byte string.
   const char* a_str() const;
   /// Gets contents of buffer as wide character string.
   const wchar_t* w_str() const;
   /// Clears output buffer.
   void clear();
private:
   BppString _buffer;
};

#ifdef WIN32
#include <windows.h>
class BPPLIB_DLLEXP BppWinMsgPrinter : public IBppPrinter
{
public:
   BppWinMsgPrinter(HWND hwnd);
   virtual int operator() (const BppChar* format, ...);
   static UINT getPrintStringMsg();
private:
   HWND _hwnd;
   static UINT _printStringMsg;
};
#endif

#endif // BPPPRINT_H
