#ifndef BPPREGEX_H
#define BPPREGEX_H

/** @addtogroup bpp_regex B++ Regular Expressions
@{

<p>B++ provides a simple implementation of Regular Expressions, based on T-REX Library from Alberto Demichelis (see
<a href="http://tiny-rex.sourceforge.net" title="T - Rex a tiny regular expression library" target="_blank">http://tiny-rex.sourceforge.net</a>).<br>
It supports essentially a subset of ECMAScript syntax (see <a href="http://www.cplusplus.com/reference/regex/ECMAScript/" target="_blank">Refernce at www.cplusplus.com</a>).
</p>
@par Description below contains syntax elements supported by B++:

<h3>Syntax</h3>

<h4>Special Characters</h4>

<p>
<tt>\\ </tt>Quotes next (meta)character<br>
<tt>^ </tt>Matches begin of string<br>
<tt>. </tt>Matches any character<br>
<tt>\\t </tt>Matches tab character<br>
<tt>\\v </tt>Matches vertical tab character<br>
<tt>\\n </tt>Matches new line character<br>
<tt>\\r </tt>Matches carriage return character<br>
<tt>\\f </tt>Matches form feed character<br>
<tt>| </tt>Separates two alternative patterns or subpatterns
</p>

<p>Note:<br>
Escape sequences for character codes, e.g. \\x<i>hh</i>, \\u<i>hhhh</i>, are not supported, but 'normal' B++ escape sequences may be used in an regualar expression.<br>
<br>
<i>Example:</i> <tt>&quot;^\\x20&quot;</tt> matches a space character at begin of text, where <tt>&quot;^\\\\s&quot;</tt> matches any whitespace character on start of text.
</p>

<h4>Assertions</h4>
<p>
<tt>^ </tt>Matches begin of target text<br>
<tt>$ </tt>Matches end of target text<br>
<tt>\\b </tt>Matches a word boundary<br>
<tt>\\B </tt>Matches position inside a word, i. e. both, character before and behind must be word characters
</p>

<h4>Brackets</h4>
<p>
<tt>() </tt>Grouping (creates a stored capture)<br>
<tt>(?:) </tt>Builds group (subexpression) too, but creates no stored capture<br>
<tt>[] </tt>Defines character class
</p>

<p>Note:<br>
Due to a bug in T-REX library stored capture groups are not correctly filled, if alternative patterns (separated by '|') are used. In that case groups of submatches are filled if first alternative matches, otherwise only main match is filled.
</p>

<h4>Predefined Character Classes</h4>
<p>
<tt>\\l		</tt>lowercase characters<br>
<tt>\\u		</tt>uppercase characters<br>
<tt>\\a		</tt>letters<br>
<tt>\\A		</tt>non letters<br>
<tt>\\w		</tt>word characters, includes decimal digits, letters and underline character (_)<br>
<tt>\\W		</tt>non word characters<br>
<tt>\\s		</tt>whitespaces<br>
<tt>\\S		</tt>non whitespaces<br>
<tt>\\d		</tt>digits<br>
<tt>\\D		</tt>non digits<br>
<tt>\\x		</tt>hexadecimal digits<br>
<tt>\\X		</tt>non hexadecimal digits<br>
<tt>\\c		</tt>control characters<br>
<tt>\\C		</tt>non control characters<br>
<tt>\\p		</tt>punctation<br>
<tt>\\P		</tt>non punctation<br>
</p>

<p>Note:<br>
T-REX library uses character classification functions from C-library (isalpha, isalnum ...) for character classes. Therefore members of character classes may be locale dependent.
</p>

<h4>Quantifiers</h4>
<p>
<tt>* </tt>Matches 0 or more times<br>
<tt>+ </tt>Matches 1 or more times<br>
<tt>? </tt>Matches 1 or 0 times<br>
<tt>{n} </tt>Matches exactly n times<br>
<tt>{n,} </tt>Matches at least n times<br>
<tt>{n,m} </tt>Matches at least n but not more than m times
</p>

*/




#ifdef DOXY

/** B++ Regular Expression class.
Class provides support of Regular Expressions for B++. It is a wrapper around native class BppRegEx.
@note Other than most predefined functions, arguments of public members of this class behave like
strong typed arguments in B++ sources, i. e. they use implicit type conversion if needed (and possible).
*/
class RegEx {
public:
	/** Ctor.
	@param pattern Pattern to prepare. If specified and not empty, constructor calls
	#compile method.
	@param flags Flags specification for expression, it may contain letters 'm' (multi line), 'i' (ignore case) and 'g' (global search); all other characters are ignored.
	@see compile, setFlags
	*/
	RegEx(const string pattern = null, const string flags = null);
	/// Destructor releases native RegEx object.
	~RegEx();
	/** Sets flags for evaluation behaviour of expression.
	@param flags Flags specification for expression, it may contain letters 'm' (multi line), 'i' (ignore case) and 'g' (global search); all other characters are ignored.
	@note Flags may be modified after compiling expression too.
	*/
	void setFlags(const string flags);
	/** Gets currently active flags for evaluation behaviour of expression.
	@return Flags specification for expression, result may contain letters 'm' (multi line), 'i' (ignore case) and 'g' (global search).
	@see setFlags, compile
	*/
	string getFlags() const;
	/** Compiles regular expression (pattern).
	@param pattern Pattern to prepare. If specified and not empty, constructor calls
	#compile method.
	@param flags Flags specification for expression, it may contain letters 'm' (multi line), 'i' (ignore case) and 'g' (global search); all other characters are ignored. If not specified, flags remain unchanged.
	@return <tt><i>true</i></tt> on success. If return value is <tt><i>false</i></tt>, caller get an error description by getError() method.
	@note <i>flags</i> are stored only. They don't affect compilation of pattern and may be changed
	later by setFlags method if needed.
	@see getError, setFlags
	*/
	bool compile(const string pattern, const string flags = null);
	/** Gets error message.
	@return Error description of last error occurred while call to compile method.
	@see compile
	*/
	string getError() const;
	
	/** Matches entire text.
	Function checks if expression matches entire text.
	@return true on match. On match a call to getSubExpressions(vector) can be done to get
	any submatches.
	@par Example
	@code
	// checks if adr is a valid mail address
	bool checkMailAdress(string adr)
	{
		RegEx rex("\w+[\\w\\.]*\w+@\\w+\\.\\a+");
		return rex.match(adr);
	}
	@endcode
	*/
	bool match(const string txt);
	
	/** Searches for first match.
	Function searches for first match in specified text.
	@param txt Text to search in.
	@param start Start position for search in <i>txt</i>.
	@param len Length of substring in <i>text</i> used for search. If <i>len</i> is
	larger than length of <i>txt</i> searching is done until end of <i>txt</i>.
	@return true if a match was found. If function returns <tt>true</tt>, #getSubExpressions(vector)
	can be called do get positions and values of match and its submatches (if any). Function #next()
	may be called to get next match.
	@note This function searches for first match only, regardless of 'g' flag.
	@par Example
	@code
	// Count number of digits in string
	RegEx rex;
	if (rex.compile("\\d")) {
		int cnt = 0;
		for(bool ok = rex.search("abc123x4y5z"); ok; ok=rex.next()) ++cnt;
		print("Number of matches: ",cnt, "\n");
	}
	else print(getError(),"\n");
	@endcode
	@see next, getSubExpressions
	*/
	bool search(const string txt, uint start = 0, uint len = NPOS);
	
	/** Finds next match.
	Function is intended for use together with #search. It finds next match.
	@return true if another match was found.
	@see search
	*/
	bool next();
	
	/** Fills vector with texts of all matches.
	Function gets matching substrings in <i>txt</i>.
	@param txt Text to search for matches.
	@param dst Vector to store results in. Funktion simply appends new string for each match
	to <i>dest</i>.
	@param start Start position of search in <i>txt</i>.
	@param len Maximum length of search, if <i>len</i>+<i>start</i> exceeds lemgth of <i>txt</i>
	search is done to end of <i>txt</i>.
	@return Number of matching substrings appended to <i>dst</i>.
	@note Behavior of method depends on 'g' flag. If it is set, all matches are
	appended to <i>dst</i>, otherwise search is aborted after first match.
	@see getMatchPositions, search, next
	*/
	uint getMatches(const string txt, vector dst, uint start = 0, uint len = NPOS);
	
	/** Fills vector with positions and sizes of all matches.
	Function gets positions and length of matching substrings in <i>txt</i>.
	@param txt Text to search for matches.
	@param dst Vector to store results in. For each match function appends a pair
	(vector of two elements) to <i>dest</i>, where first element represents
	position of match in <i>txt</i> and second its length.
	@param start Start position of search in <i>txt</i>.
	@param len Maximum length of search, if <i>len</i>+<i>start</i> exceeds lemgth of <i>txt</i>
	search is done to end of <i>txt</i>.
	@return Number of matching substrings appended to <i>dst</i>.
	@note Behavior of method depends on 'g' flag. If it is set, all matches are
	appended to <i>dst</i>, otherwise search is aborted after first match.
	@see getMatches, search, next
	*/
	uint getMatchPositions(const string txt, vector dst, uint start = 0, uint len = NPOS);
	
	/** Gets descriptions for current match and all its submatches.
	Function gets submatch descriptors of last match.
	@param dst Vector to store results in. For each submatch a dictionary of two elements is appended to <i>dst</i>, where
	item 'pos' (BppInt) contains position and item 'text' (BppString) the matching text. First element appended is the main match itself.
	@return Number of elements appended to <i>dst</i>.
	@note Method is intended for use with #search, #next and #match functions.
	@par Example
	@code
	// Find numbers behind a letter
	RegEx rex;
	if (rex.compile("\\a(\\d+)")) {
		for(bool ok = rex.search("abc123x4y5z");ok;ok=rex.next()) 
		{
			vector v = [];
			uint cnt = rex.getSubExpressions(v);
			for (uint i=0;i<cnt;++i) {
			 dictionary d = v[i];
			  print(i, " pos: ", d.pos, " text: ", d.text, "\n");
			}
		}
	}
	else print(getError(),"\n");
	@endcode
	@see search, next, match
	*/
	uint getSubExpressions(vector dst);
	
	/** Gets descriptions for all matches and their submatches.
	Function combines #search, #next and #getSubExpressions functions. It iterates through specified text collecting all matches and their submatches.
	@param txt Text to analyze.
	@param dst Vector to store results into. For each match a new vector is appended to <i>dst</i>, each containing a vector of submatches.
	@param start Start position of search in <i>txt</i>.
	@param len Maximum length of search, if <i>len</i>+<i>start</i> exceeds lemgth of <i>txt</i>
	search is done to end of <i>txt</i>.
	@return Number of items (matches found) appended to <i>dst</i>.
	@see getSubExpressions(vector)
	*/
	uint getSubExpressions(string txt, vector dst, uint start = 0, uint len = NPOS);
	
	/** Replaces matches in a string producing a new string.
	Function builds new string from <i>txt</i> replacing first match or all matches in <i>txt</i> by <i>repl</i>.
	@param txt Text to use for replace matches.
	@param repl Replacement expression. The expression is a string, optionally containing following placeholders/variables:
	<ul>
	<li><b>$&</b> or <b>$0</b> - Inserts entire matching text.</li>
	<li><b>$`</b> (back tick) - Inserts all text before match.</li>
	<li><b>$'</b> (apostrophe) - Inserts all text behind match.</li>
	<li><b>$<i>n</i></b> - Inserts text of <i>n</i>-th submatch, where <i>n</i> is a positive decimal number. If <i>n</i> exceeds number of submatches
	available, string representation of <i>n</i> is inserted.</li>
	<li><b>$$</b> - Inserts Dollar character ($).
	</ul>
	@return Result of replace operation
	@note Behavior depends on 'g' flag. If set, all matches are replaced, otherwise only first match.
	@see strReplace
	*/
	string replace(string txt, string repl);
	
	/** Replaces matches in a string  modifying the source string.
	Function replaces first match or all matches in <i>txt</i> by <i>repl</i>.
	@param txt Text to replace matches in.
	@param repl Replacement expression. The expression is a string, optionally containing following placeholders/variables:
	<ul>
	  <li><b>$&</b> or <b>$0</b> - Inserts entire matching text.</li>
	  <li><b>$`</b> (back tick) - Inserts all text before match.</li>
	  <li><b>$'</b> (apostrophe) - Inserts all text behind match.</li>
	  <li><b>$<i>n</i></b> - Inserts text of <i>n</i>-th submatch, where <i>n</i> is a positive decimal number. If <i>n</i> exceeds number of submatches
	  available, string representation of <i>n</i> is inserted.</li>
	  <li><b>$$</b> - Inserts Dollar character ($).
	</ul>
	@return Number of matches replaced
	@note <p>Function modifies <i>txt</i> argument. Use #replace to create new string instead.</p>
	<p>Behavior depends on 'g' flag. If set, all matches are replaced, otherwise only first match.</p>
	@see replace
	*/
	uint strReplace(string txt, string repl);
	
	/** Searches for first match.
	Operator is simply an alias of #search method, intended for simplified iteration through matches.
	@param txt Text to search in.
	@param start Start position for search in <i>txt</i>.
	@param len Length of substring in <i>text</i> used for search. If <i>len</i> is
	larger than length of <i>txt</i> searching is done until end of <i>txt</i>.
	@return true if a match was found. If operator returns <tt>true</tt>, #getSubExpressions(vector)
	can be called do get positions and values of match and its submatches (if any). Operator may be called
	without arguments to get next match.
	@note Operator searches for first match only, regardless of 'g' flag.
	@par Example
	@code
	// Count number of digits in string using operator()
	RegEx rex;
	if (rex.compile("\\d")) {
		int cnt = 0;
		for(bool ok = rex("abc123x4y5z"); ok; ok=rex()) ++cnt;
		print("Number of matches: ",cnt, "\n");
	}
	else print(getError(),"\n");
	@endcode
	@see next, getSubExpressions
	*/
		bool operator() (const string txt, uint start = 0, uint len = NPOS);
	/** Searches for next match.
	Operator is simply an alias of #next method, intended for simplified iteration through matches.
	@return true if another match was found.
	@see operator()(string,uint,int)
	*/
	bool operator() ();
	
	/** Allows simplified access to some properties of RegEx object.
	Operator provides read access to following properties:
	<table><tr><th>Key</th><th>Result type</th><th>Description</th></tr>
	<tr><td>flags</td><td>string</td><td>Returns flags currently set as string, may contain 'g', 'm' and 'i' character.</td></tr>
	<tr><td>global</td><td>bool</td><td>Returns <tt>true</tt> if 'g' flag is set.</td></tr>
	<tr><td>ignoreCase</td><td>bool</td><td>Returns <tt>true</tt> if 'i' flag is set.</td></tr>
	<tr><td>multiline</td><td>bool</td><td>Returns <tt>true</tt> if 'm' flag is set.</td></tr>
	<tr><td>source</td><td>string</td><td>Returns copy of pattern string used for expression.</td></tr>
	<tr><td>text</td><td>string</td><td>Returns copy of text string used for last search operation.</td></tr>
	<tr><td>pos</td><td>uint</td><td>Returns position for start of next search.</td></tr>
	<tr><td>len</td><td>uint</td><td>Returns search length.</td></tr>
	</table>
	@param key Key of property to get.
	@return Result value dependent on <i>key</i>.
	@exception
		Operator throws exception of <tt>string</tt> type on invalid <i>key</i>. 
	*/
	var operator . (const string key) const;
	
	/** Allows simplified access to some properties of RegEx object.
	Operator provides write access to following properties:
	<table><tr><th>Key</th><th>Value type</th><th>Description</th></tr>
	<tr><td>flags</td><td>string</td><td>Sets all flags of RegEx object (does same as #setFlags).</td></tr>
	<tr><td>global</td><td>bool</td><td>Sets/Resets 'g' flag.</td></tr>
	<tr><td>ignoreCase</td><td>bool</td><td>Sets/Resets 'i' flag.</td></tr>
	<tr><td>multiline</td><td>bool</td><td>Sets/Resets 'm' flag.</td></tr>
	<tr><td>text</td><td>string</td><td>Sets text string used for next search operation.</td></tr>
	<tr><td>pos</td><td>uint</td><td>Sets position for start of next search.</td></tr>
	<tr><td>len</td><td>uint</td><td>Sets search length.</td></tr>
	</table>
	@param key Key of property to set.
	@param value New value of property specified by <i>key</i>.
	@return value
	@exception
	  Operator throws exception of <tt>string</tt> type on invalid <i>key</i> or invalid type of <i>value</i>. 
	*/
	var operator .= (const string key, var value);
private:
	/// stored pattern source
	string _pattern;
	/// stored flag mask
	uint _flags;

};

#endif // DOXY 



/// @}


#include <bppstrng.h>
#include <TRexpp.h>
class BppVector;
class Bpp_RegEx;


/// Wrapper around T-REX library
class BPPLIB_DLLEXP BppRegEx
{
	friend class Bpp_RegEx;
public:

	// Flags for modifying match behavior - may be ORed
	enum Flags {
		RXF_NONE = 0,        ///< nothing, default behavior
		 RXF_IGNORECASE = 1,  ///< ignore character case
		RXF_MULTILINE = 2,   ///< multiline match (^ and $ match begin and end of line too)
		RXF_FINDALL = 4      ///< find all matches (not the first only)
	};
	/// Default ctor
	BppRegEx();
	/// Copy ctor
	BppRegEx(const BppRegEx& cpy);
	/// Assignment
	BppRegEx& operator=(const BppRegEx& cpy);
	/// dtor
	~BppRegEx();
	/// Compiles expression.
	bool compile(const BppString& pat, unsigned flags = 0);
	/// Sets flags bitmask.
	bool setFlags(unsigned flags);
	/// Gets flags bitmask.
	unsigned getFlags() const;
	/// Gets description of last error occurred.
	const BppString& getError() const;
	/// Matches entire text.
	bool match(const BppString& txt);
	/// Searches for first match.
	bool search(const BppString& txt, size_t start = 0, size_t len = BPP_NNPOS);
	/// Finds next match.
	bool next();
	/// Fills vector with texts of all matches.
	unsigned getMatches(const BppString& txt, BppVector& dst, size_t start = 0, size_t len = BPP_NNPOS);
	/// Fills vector with positions and sizes of all matches.
	unsigned getMatchPositions(const BppString& txt, BppVector& dst, size_t start = 0, size_t len = BPP_NNPOS);
	/// Gets descriptions for current match and all its submatches.
	unsigned getSubExpressions(BppVector& dst);
	/// Gets descriptions for all matches and their submatches.
	unsigned getSubExpressions(const BppString& txt, BppVector& dst, size_t start = 0, size_t len = BPP_NNPOS);
	/// Replaces matches in a string producing a new string.
	BppString replace(const BppString& txt, const BppString& repl);
	/// Replaces matches in a string  modifying the source string.
	unsigned strReplace(BppString& txt, const BppString& repl);

private:
	TRexpp& _trex;        ///< Wrapped T-REX object
	BppString _error;    ///< Error message buffer
	BppString _pattern;  ///< Pattern buffer
	unsigned _flags;     ///< Flags bitmask
	BppString _ptxt;     ///< Buffers current search text
	size_t _pos;         ///< Position of corrent match
	size_t _len;           ///< Length of current match
};

#ifndef DOXY

// B++ language binding for BppRegEx
class BPPLIB_DLLEXP Bpp_RegEx {
public:
	static bool ctor(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool dtor(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool setFlags(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getFlags(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool compile(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getError(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool match(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool search(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool next(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getMatches(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getMatchPositions(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getSubExpressions(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool replace(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool strReplace(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_call(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_pref(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_pset(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static void createClass(BppVirtualMachine& vm);
	static void BPPAPI notifyProc(BppVirtualMachine* vm, BppObject* obj, void** nativeObject, BppObjectNotifyType reason);
private:
	static unsigned _idxPattern;
	static unsigned _idxFlags;
};

#endif

#endif // BPPREGEX_H

