#ifndef BPPSTRNG_H
#define BPPSTRNG_H

#include <bpprobj.h>
class BppBuffer;
class BppCharBuffer;
class BppVector;

/** B++ string class.
This class is the internal representation of a character string.
It manages an array of characters where character size depends on build
(Unicode or single byte code). For both types of builds class allows
conversions between Unicode and single byte strings. Furthermore
it provides a set of useful string operations.
*/
class BPPLIB_DLLEXP BppString : public BppRefObject
{
public:
   /// default ctor
	BppString() BPPNOEXCEPT;
   /// copy ctor
	BppString(const BppString& cpy);
#ifdef CPP11
   /// move ctor
   BppString(BppString&& cpy) BPPNOEXCEPT;
#endif
   /// construction from single byte C-string
	BppString(const char* src);
   /// construction from byte buffer
	BppString(const BppBuffer& src);
   /// construction from character buffer
	BppString(const BppCharBuffer& src);
   /// formatted construction
   BppString(const BppChar* fmt,unsigned nargs, ValPtr& args);
#ifdef WCHAR_MAX
   /// construction from Unicode C-string
	BppString(const wchar_t* src);
#endif
   /// dtor
	virtual ~BppString();
   /// create clone of current object
	BppString* clone() const;
   /// assignment
	BppString& operator=(const BppString& assignment);
#ifdef CPP11
   /// move assignment
   BppString& operator=(BppString&& assignment);
#endif
   /// assignment fron C string
	BppString& operator=(const BppChar* assignment);
   /// append string
	BppString& operator+=(const BppString& appendix);
   ///append C-string
	BppString& operator+=(const BppChar* appendix);
   /// append single character
	BppString& operator+=(BppChar appendix);
   /// concatenation
	BppString  operator+(const BppString& appendix) const;
 
   /** Read/write access to character.
   @param idx index of character to get
   @return read/write reference to specified character
   @note This operator does no bounds checking, i.e. client code is responsible
   t do it.
   */
   inline BppChar& operator[](size_t idx)
   {
      return _data[idx];
   }

   /** Read only access to character.
   @param idx index of character to get
   @return read only reference to specified character
   @note This operator does no bounds checking, i.e. client code is responsible
   t do it.
   */
   inline const BppChar& operator[](size_t idx) const
   {
      return _data[idx];
   }

   /// lower-than-operator (for use in STL maps)
   bool operator<(const BppString& cmp) const BPPNOEXCEPT;
   /// equality operator
   bool operator== (const BppString& cmp) const BPPNOEXCEPT;
   /// unequality operator
   bool operator!= (const BppString& cmp) const BPPNOEXCEPT;
   /// compare function
	int compare(const BppString& comparison) const BPPNOEXCEPT;
   /// case insensitive compare function
   int icompare(const BppString& comparison) const BPPNOEXCEPT;
   /// get length of string (number of contained characters)
   size_t length() const BPPNOEXCEPT;
//   operator const BppChar*() const;

/** Gets C-string of build dependent character size.
   @return C-string representation of current value
   */
   inline const BppChar* c_str() const BPPNOEXCEPT
   {
      return _data ? _data : _T("");
   }
   /// get single byte C-string
   const char* a_str() const;
   /// get Unicode C-string
   const wchar_t* w_str() const;
   /// remove all characters
   void clear();
   /// find position of substring
   size_t find(const BppString& s, size_t off = 0) const BPPNOEXCEPT;
   /// find position of a character
   size_t find(BppChar c, size_t off = 0) const BPPNOEXCEPT;
   /// replace substring by another one
   BppString& replace(const BppString& str, const BppString& replacement);
   /// replace a character by another one
   BppString& replace(BppChar c, BppChar replacement) BPPNOEXCEPT;
   /// get substring
   BppString substr(size_t start, size_t len = BPP_NNPOS) const;
   /// erase specified part from string
   BppString& erase(size_t pos, size_t cnt = BPP_NNPOS);
   /// insert another string into current one
   BppString& insert(size_t pos, const BppString& insertion);
   /// insert character into string
   BppString& insert(size_t pos, BppChar insertion, size_t cnt = 1);
   /// split into tokens
   size_t split(BppVector& dest, const BppString& delim, bool storeEmptyTokens = false, bool storeDelimiter = false) const;
   /// convert string to lower case
	BppString& lower();
   /// convert string to upper case
	BppString& upper();
   /// build formatted string
   static BppString format(const BppChar* fmt,...);
   /// build string from lower case representation of str
	static BppString strlower(const BppChar* str);
   /// build string from upper case representation of str
   static BppString strupper(const BppChar* str);
private:
   /// pointer to character buffer
	BppChar* _data;
};

#ifndef DOXY

// On Win32 desktop systems (but not WinCE) GetProcAddress function always 
// uses ANSI strings. So we need to convert strings in Unicode builds.
#if (defined(BUILD_BPPLIB) || defined(__MSDOS__) || defined(BPP_STATIC))
  #if defined (WIN32) && (_MSC_VER >=1400 || !defined(WINCE))
    #define p_str a_str
    #define _PT(s) s
  #else
    #define p_str c_str
    #define _PT(s) _T(s) 
  #endif 
#endif // BUILD_BPPLIB

#endif // !DOXY (undocumented)

#endif // BPPSTRNG_H
