#ifndef BPPSTRUCT_H
#define BPPSTRUCT_H

/** @addtogroup bpp_cstruct B++ POD C-Struct Emulation
@{ 
B++ emulates C structures using a predefined base class.

Many operating system functions use references to C data structures as arguments.<br>
Because B++ does not have structures compatible with C, they have to be emulated, typically using a buffer.<br>
The correct mapping of a structure to a linear buffer can be difficult because the data types of the structure elements have to be mapped correctly and rules for the alignment of the structure have to be considered.<br>

To simplify this, there is a predefined abstract class CStruct in B++, which enables the emulation of C-compatible structures using a simple descriptor.<br>
Class CStruct realizes the mapping of C-compatible structures, including nested structures, to B++ buffers and easy access to individual structure elements. By default the standard rules are used for the alignment of the elements, 
i.e. simple data types are aligned to multiples of their type size, structures to multiples of their largest simple element. By overriding #getStructAlignment derived
class may define special alignment of members.<br>
The limitation is that only numeric types (including the intptr type for mapping pointers), embedded structures and fix sized arrays of these types can be used as structure elements.<br>
Embedded structures are allowed to be of union type. They are declared like embedded arrays of size 0 and a structure for element type.<br>
Dynamic sized arrays may be emulated in derived classes by modifying structure descriptor at runtime.<br>
An array element can't have another array as element type, i.e. only arrays of single dimension are supported.<br>

To create a C compatible structure, a class derived from CStruct has to be defined, which implments
at least a constructor and the #getStructDesc method.<br>
The constructor must call the derived create method and #getStructDesc must return a dictionary containing
structure descriptor as follows:
<ul>
   <li>Each element of dictionary describes a member of struct, where order of entries in 
	dictionary defines order of structure members.</li>
	<li>The name of an entry must be a valid identifier.</li>
	<li>The value of an entry is either the internal type identifier of any number type (type literal)
	or a nested dictionary, defining a nested structure, or a nested vector of two elements defining a nested array, 
   where first element of vector is the number of elements in array and second element is type of elements 
   (internal identifier of any number type or dictionary defining element structure).</li>
</ul>
In special case of a C structure having an array of dynamic size as last element, #getStructureDesc may return modified descriptor depending on array size.
In that case #create should be called with <tt>true</tt> for second argument (<i>isdynamic</i>) in order to caching internal data descriptor.

The following example defines a simple rectangle structure, containing coordinates x and y and 
a nested structure for size.
@code

class Rect : CStruct
{
public:
   Rect(dictionary data = null) {
	  create(data);
	}
protected:
   dictionary getStructDesc() {
	   return {
		  x : BVT_INT,
		  y : BVT_INT,
		  ext : {
		     w : BVT_UINT,
			 h : BVT_UINT
		  }
		};
	}
};
@endcode

Initialization of a structure can be done either by argument of constructor or by writing value to each
structure member. Without initialization all members are initialized to zero.

Member access is done by point notation, i.e. using point operator.

Next example fills a Rect structure using constructor, modifies values and
prints content of result.
@code
void main() {
	Rect r({ x : 10, y : 10, ext : { w : 20, h : 30}});
	r.x *= -1;
	r.ext.w += 5;
	print("x:", r.x, " y:", r.y, " w:", r.ext.w, " h:", r.ext.h,"\n");
}
@endcode

Following example demonstrates use of array inside a structure and definition of a structure by named literal.
It emulates RECT and PAINTSTRUCT structures of MS-Windows API.
@code
#define DescRECT { left : BVT_INT, top: BVT_INT, right : BVT_INT, bottom : BVT_INT }

class RECT: CStruct
{
public:
   RECT(dictionary data = null) { 
      create(data);
   }
protected:
   dictionary getStructDesc() { 
      static dictionary desc = DescRECT;
      return desc;
   }
};

class PAINTSTRUCT: CStruct
{
public:
   PAINTSTRUCT(dictionary data = null) { 
      create(data);
   }
protected:
   dictionary getStructDesc() {
      static dictionary desc = {
         hdc : BVT_INTPTR,
         fErase : BVT_INT,
         rcPaint : DescRECT,
         fRestore : BVT_INT,
         fIncUpdate : BVT_INT,
         rgbReserved : [32,BVT_BYTE]
      };
      return desc;
  }
};
@endcode

Following example demonstrates user defined structure alignment. Therefore getStructuerAlignment is overridden.
It emulates BITMAPFILEHEADER structure of Windows GDI:
@code
  class BITMAPFILEHEADER : CStruct
   {
	public:
	   BITMAPFILEHEADER(dictionary data = null) {
		  create(data);
	   }
	protected:
	     // define WORD alignment
		 uint getStructPacking() { return 2; }

		 dictionary getStructDesc() {
			static dictionary desc = {
			   bfType : BVT_USHORT,
			   bfSize : BVT_UINT,
			   bfReserved1 : BVT_USHORT,
			   bfReserved2 : BVT_USHORT,
			   bfOffBits : BVT_UINT
			};
			return desc;
		  }
   };

Following example demonstrates a dynamic sized structure.
It emulates EXTLOGPEN structure of Windows GDI:
@code
   class EXTLOGPEN : CStruct
   {
	private:
	  int _numentries = 1;
	public:
	   EXTLOGPEN(dictionary data = null, int numentries = 1) {
	      // Get size from data if available
		  if (data && dictcontains(data,"elpStyleEntry"))
			 _numentries = size(data.elpStyleEntry);
		  else
		  _numentries = numentries; // Otherwise use numentries argument
		  // call create with isdynamic=true
		  create(data,true);
	   }
	protected:
		 dictionary getStructDesc() {
		    // default descriptor
			static dictionary desc = {
			  elpPenStyle : BVT_UINT,
			  elpWidth : BVT_UINT,
			  elpBrushStyle : BVT_UINT,
			  elpColor : BVT_UINT,
			  elpHatch : BVT_UINTPTR,
			  elpNumEntries : BVT_UINT,
			  elpStyleEntry : [1,BVT_UINT] //  DWORD elpStyleEntry[1]
			};
			if (_numentries > 1) {
			   // make copy of descriptor
			   auto res = clone(desc);
			   // modify size of array definition for elpStyleEntry
			   res.elpStyleEntry[0] = _numentries;
			   return res;
			}
			return desc;
		  }
   };
@endcode
*/

#ifdef DOXY
/** Helper for nested POD structure access.
This class is an implementation helper of CStruct class. Its instances are temporary proxies around
a substructure of a structure defined by a subclass of CStruct. It allows access to members of substuct.<br>
Instances of __SubstructProxy may be returned as result of operator. of CStruct derived class or of __SubstructProxy instances containing another substructure.<br>
Because there is no public constructor, client code can't  create instances of this class exlpicitely.
*/
class __SubstructProxy
{
public:
	/** Read access to structure member.
	@param key member name
	@return Value of member, in case of simple data member, or __SubstructProxy instance for a substructure.
	@exception Exception of string type thrown in case of invalid <i>key</i>.
	*/
   var operator.(const string key) const;
	/** Write access to structure member.
	@param key member name
	@param val value to assign
	@return copy of <i>val</i>
	@exception Exception of string type thrown in case of invalid <i>key</i> or if <i>val</i> can't be converted into member type.
	*/
	var operator.=(const string key, var val);
	/** Gets dictionary containing copy of all members of substructure.
	@return Returns a dictionary of member data, where keys are member names and values copies of member data. Result
	may contain values of dictionary type if there are substructures.
	*/
	dictionary getData() const;
	/** Gets size of a structure member.
	@param name member name
	@return size of member specified by <i>name</i> in bytes.
	@exception Exception of string type thrown in case of invalid <i>name</i>.
	*/ 
	uint sizeOf(const string name) const;
	/** Gets offset of a structure member.
	@param name member name
	@return offset of member specified by <i>name</i> in bytes. Result is related to start of
	outermost structure, i.e. of a structure derived from CStruct.
	@exception Exception of string type thrown in case of invalid <i>name</i>.
	*/
	uint offsetOf(const string name) const;
	/** Gets signed pointer to start of native substruct representation.
	@return Address inside of data area of internal buffer where substruct starts.
	@note Operator should be called to get pointer to dispatch as argument of a native function.
	*/
	operator intptr() const;
	/** Gets unsigned pointer to start of native substruct representation.
	@return Address inside of data area of internal buffer where substruct starts.
	@note Operator should be called to get pointer to dispatch as argument of a native function.
	*/
    operator uintptr() const; 
private:
   buffer _sub_buf; ///< Buffer used
	uint _sub_offs; ///< Offset of substructure in _sub_buf
	dictionary _sub_desc; ///< Descriptor of substructure
};

/** Helper for nested POD array access.
This class is an implementation helper of CStruct class. Its instances are temporary proxies around
a C-array member of structure defined by a subclass of CStruct. It allows access to elements of arrayt.<br>
Instances of __SubArrayProxy may be returned as result of operator. of CStruct derived class or of __SubstructProxy instances containing array mamber.<br>
Because there is no public constructor, client code can't  create instances of this class exlpicitely.
*/
class __SubArrayProxy
{
public:
	/** Read access to array element.
	@param idx element index member name
	@return Value of element, in case of simple data element, or __SubstructProxy instance for structured element type.
	@exception Exception of string type thrown in case of invalid <i>idx</i> value.
	*/
	var operator[](uint idx) const;
	/** Write access to array element.
	@param idx element index
	@param val value to assign
	@return copy of <i>val</i>
	@exception Exception of string type thrown in case of invalid <i>idx</i> value or if <i>val</i> can't be converted into element type type.
	*/
	var operator[]= (uint idx, var val);
	/** Gets vector containing copy of all elements of array.
	@return Returns a vector of element data. Result
	may contain values of dictionary elements as substructures.
	*/
	vector getData() const;
	/** Gets number of elements in array.
	@return Array size (number of elements).
	*/
	uint size() const;
	/** Gets size of single array element.
	@return Size of array element in bytes.
	*/
	uint elSize() const;
	/** Gets offset specified array element.
	@param idx element index
	@return Offset of element specified by <i>idx</i> in bytes. Result is related to start of
	outermost structure, i.e. of a structure derived from CStruct.
	@exception Exception of string type thrown in case of invalid <i>idx</i> value.
	*/
	uint offsetOf(uint idx) const;
	/** Gets signed pointer to start of native array representation.
	@return Address inside of data area of internal buffer where array starts.
	@note Operator should be called to get pointer to dispatch as argument of a native function.
	*/
	operator intptr() const;
	/** Gets unsigned pointer to start of native array representation.
	@return Address inside of data area of internal buffer where array starts.
	@note Operator should be called to get pointer to dispatch as argument of a native function.
	*/
	operator uintptr() const;
private:
	buffer _sub_buf; ///< Buffer used
	uint _sub_offs; ///< Offset of array in _sub_buf
	dictionary _sub_desc; ///< Descriptor of array
};

/** Abstract base class for emulation of C-compatible data structures.
This class provides emulation of any C-compatible POD structure using a buffer.<br>
Emulation always uses default alignment and padding rules.<br>
Members of structure may be of numeric types or substructures.<br>
In order to access members of substructures it uses __SubstructProxy instances.
@note Structure emulation is intended to simplyfy use of structured arguments for
arguments of system function calls. It is not recommended to use it in order to save some bytes
of memory for data representation, because access to members is slower than normal class member access.
*/
class CStruct
{
public:
	/** Gets signed pointer to native structure representation.
	@return Address of data area of internal buffer, containing structure data.
	@note Operator should be called to get structure pointer to dispatch as argument of a native function.
	*/
	operator intptr() const;
	/** Gets unsigned pointer to native structure representation.
	@return Address of data area of internal buffer, containing structure data.
	@note Operator should be called to get structure pointer to dispatch as argument of a native function.
	*/
	operator uintptr() const;
	/** Gets internal buffer containing structure representation..
	@return Internal buffer containing structure data.
	@note Result allows access to internal data representation. This may be useful, e.g., to get buffer size or
	to allow copy ecternal data to buffer. Client code never should modify size of buffer.
	*/
	buffer getDataBuffer() const;

	/** Gets size of internal buffer.
	@return Site of internal buffer in bytes.
	@see getDataBuffer
	*/
	uint getDataSize() const;
	/** Read access to structure member.
	@param key member name
	@return Value of member, in case of simple data member, __SubstructProxy instance for a substructure or
	__SubArrayProxy for array member.
	@exception Exception of string type thrown in case of invalid <i>key</i>.
	*/
	var operator.(const string key) const;
	/** Write access to structure member.
	@param key member name
	@param val Value to assign. If <i>val</i> is of dictionary type, <i>key</i> must refer to a substructure and
	content of <i>val</i> must match that structure. If <i>val</i> is of vector type, <i>key</i> must refer to
	an array mamber and content must match that member.
	@return copy of <i>val</i>
	@exception Exception of string type thrown in case of invalid <i>key</i> or if <i>val</i> can't be converted into member type.
	*/
	var operator.= (const string key, var val);
	/** Gets dictionary containing copy of all members of structure.
	@return Returns a dictionary of member data, where keys are member names and values copies of member data. Result
	may contain values of dictionary type if there are substructures.
	*/
	dictionary getData() const;
	/** Sets data of multiple structure members from dictionary.
	Function iterates through <i>data</i> and sets members of structure
	from its contents using operator.=(key,value).
	@param data Dictionary to set member data from. Elements of <i>data</i> must match
	members of structure.
	@exeption Exception of string type thrown if <i>data</i> contains elements nut matching
	defined members of structure.
	*/
	void applyData(dictionary data);
	/** Gets size of a structure member.
	@param name member name
	@return size of member specified by <i>name</i> in bytes.
	@exception Exception of string type thrown in case of invalid <i>name</i>.
	*/
	uint sizeOf(const string name);
	/** Gets offset of a structure member.
	@param name member name
	@return offset of member specified by <i>name</i> in bytes.
	@exception Exception of string type thrown in case of invalid <i>name</i>.
	*/
	uint offsetOf(const string name) const;

protected:
	/** Creates data buffer and optionally fills it.
	@param data Dictionary containing initially data
	@param isdynamic Flag for sttucture of dynamic size, e.g. containing array of variable length.
	@note Function creates internal data descriptor and data buffer. Normally data descriptos are chached,
	i.e. creation is needed only for first instance of a specifig class. If <i>isdynamic</i> is set, no caching is done. 
	This allows construction of structures of different size, depending on implementation of getStructureDesc method.
	*/
	void create(dictionary data = null, bool isdynamic = false);
	/** @pure Gets Structure definition descriptor.
	*/
	dictionary getStructDesc() const;
	/** Gets packing of structure.
	@return Packing rule in bytes, default value is 0 and means default structure member alignment.<br>
	If packing rule is different from 0, members will be aligned to multiple of specified value in bytes.<br>
	The return value shoultd be 0, 1, 2, 4, 8 or 16.
	@note Derived classes may override this method in order to determine special packing of structure members.
	*/
	uint getStructPacking() const;
private:
	buffer _dataBuffer; ///< Buffer for plain structure data.
	dictionary _bufDesc; ///< Structure descriptor.
	static dictionary _bufDescriptors; ///< Global mapping of all structure descriptors.
};

#endif
/// @}

#ifndef DOXY
#include <bppdef.h>
class BppClass;
class BppBuffer;
class BppDictionary;
class BppDictionaryEntry;
class BppVirtualMachine;

class  Bpp_CStruct;
class Bpp_ArrayHelper;
/// Implementation of __SubstructProxy class
class BPPLIB_DLLEXP Bpp_StructHelper {
	friend class Bpp_CStruct;
	friend class Bpp_ArrayHelper;
public:
	static bool op_pref(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_pset(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getData(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool offsetOf(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool sizeOf(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool intptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static void createClass(BppVirtualMachine& vm);
private:
	static BppObject* _createInstance(BppClass& cls, BppBuffer* buf, BppUInt offs, BppDictionary* desc);
	static unsigned _idxDataBuffer;
	static unsigned _idxOffs;
	static unsigned _idxDesc;

};

/// Implementation of __SubArrayProxy class
class BPPLIB_DLLEXP Bpp_ArrayHelper {
	friend class Bpp_CStruct;
	friend class Bpp_StructHelper;
public:
	static bool op_vref(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_vset(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getData(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool offsetOf(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool size(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool elSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool intptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static void createClass(BppVirtualMachine& vm);
private:
	static BppObject* _createInstance(BppClass& cls, BppBuffer* buf, BppUInt offs, BppDictionary* desc);
	static unsigned _idxDataBuffer;
	static unsigned _idxOffs;
	static unsigned _idxDesc;

};


/// Implementation of CStruct class
class BPPLIB_DLLEXP Bpp_CStruct {
public:
	static bool create(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getStructDesc(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getStructPacking(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool intptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool uintptr(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getDataBuffer(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getDataSize(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_pref(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool op_pset(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool getData(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool applyData(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool offsetOf(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static bool sizeOf(BppVirtualMachine* vm, int argc, ValPtr& sp);
	static void createClass(BppVirtualMachine& vm);
private:
	static bool createBuffer(BppVirtualMachine* vm, BppObject* pThis, bool isdynamic);
	static bool buildDescItem(BppVirtualMachine* vm,BppDictionaryEntry& item, BppDictionary& cont, BppUInt packing, bool isUnion);
	static bool buildBufDesc(BppVirtualMachine* vm, BppValue& result, BppDictionary& structDesc, BppUInt packing, bool isUnion);
	static unsigned _idxBufDesc;
	static unsigned _idxDataBuffer;


};
#endif // !DOXY
#endif // !BPPSTRUCT_H

