#ifndef BPPVALUE_H
#define BPPVALUE_H

#include <stdio.h>
#ifndef DOXY
#include "bppdef.h"
#endif
/// Type catagory classification.
enum BppVTypeCategory {
	BTC_PSEUDO = 0, ///< Pseuto types (null, void, auto and internal pseudo types)
	BTC_NUMBER = 0x40, ///< Number types 
	BTC_WEAKREFFLAG = 0x40, ///< Marks weak references for reference types.
	BTC_REFTYPE = 0x80, ///< Reference types
	BTC_WEAKREF = BTC_REFTYPE | BTC_WEAKREFFLAG ///< Weak reference types
};


/// Sub-category of number type.
enum BppNumberCategory {
	BNC_INTEGRAL = 0, ///< Integral type
	BNC_REAL = 0x20 ///< Real number type
};

/// Signed property of integral types.
enum BppSignedType {
	BST_SIGNED = 0, ///< Signed number
	BST_UNSIGNED = 4 ///< Unsigned number
};

enum BppIntegralSubtype {
	BIST_NONE = 0,
	BIST_SPECIAL = 1,
	BIST_CHAR = 2
};
/// Size type flags of integral numbers.
enum NumSizeShift {
	BNS_BYTE = 0, ///< Single byte
	BNS_WORD = 8, ///< Word (16 bits)
	BNS_DWORD = 0x10, ///< DWord (32 bits)
	BNS_INTPTR = BNS_DWORD | BIST_SPECIAL, ///< Address (32 or 64 bits)
	BNS_QWORD = 0x18 ///< QWord (64 bits)
};

/// Subtype of reference types.
enum BppRefTypeCategory {
	BRTC_PTRTYPE = 0, ///< Simple pointer type
	BRTC_REFCOUNTED = 0x20 ///< Reference counted type
};

/// Subtype of reference counted type.
enum RefCountedCategory {
	BRCC_SINGLE = 0, ///< Single element
	BRCC_SEQUENCE = 0x10 ///< Sequence of elements
};

/// Subtype of single reference counted type.
enum BppRefSingleCategory {
	BRSC_OBJECTTYPE = 0, ///< Object
	BRSC_CLASS = 0x08 ///< Class
};

/// Subtype of reference counted sequence type.
enum BppRefSequenceCategory {
	BRSQC_SIMPLE = 0, ///< Simple sequence (strings, buffers)
	BRSQC_VECTOR = 0x04, ///< Vector
	BRSQC_DICT = 0x08 ///< Dictionary
};

enum BppTypeMasks {
	WEAKMASK = ~BTC_WEAKREF | BTC_REFTYPE,
	SPECIALMASK = ~BIST_SPECIAL
};


/** Enumeration of B++ internal data types.
*/
enum BppValueType
{
	_BVTMIN = 0,
	BVT_NULL = BTC_PSEUDO, ///< null type (not initialized value)
	BVT_VOID, ///< defines no value (for return type of functions only)
	BVT_AUTO, ///< defines strong typed value, where type is determined on first assignment
	BVT_NIL, ///< defines strong null type (for return value and to prevent null variables)
	/** @name Pseudo types (for internal typechecks only.
	@see BppInterpreter::checkType
	*/
	//@{
	BVT_NUMBER,
	BVT_INTNUMBER,
	BVT_UINTNUMBER,
	BVT_INTEGRALTYPE,
	BVT_PTRTYPE,
	BVT_REALNUMBER,
	BVT_REFCNTTYPE,
	BVT_REFERENCE,
	//@}
	_BVTNUMBERBASE = BTC_NUMBER,
	_BVTINTEGRALBASE = _BVTNUMBERBASE | BNC_INTEGRAL,
	BVT_SBYTE = _BVTINTEGRALBASE | BNS_BYTE | BST_SIGNED, ///< 8 bits integer
	BVT_INTEGRAL_MIN = BVT_SBYTE,
	BVT_BOOL = BVT_SBYTE | BIST_SPECIAL,
	BVT_CHAR = BVT_SBYTE | BIST_CHAR,
	BVT_BYTE = _BVTINTEGRALBASE | BNS_BYTE | BST_UNSIGNED, ///< 8 bits unsigned integer
	BVT_UCHAR = BVT_BYTE | BIST_CHAR,
	BVT_CHAR8 = BVT_UCHAR | BIST_SPECIAL,
	BVT_SHORT = _BVTINTEGRALBASE | BNS_WORD | BST_SIGNED, ///< 16 bits signed integer
	BVT_USHORT = _BVTINTEGRALBASE | BNS_WORD | BST_UNSIGNED, ///<16 bits unsigned integer
	BVT_CHAR16 = BVT_USHORT | BIST_CHAR,
	BVT_WCHAR = BVT_CHAR16 | BIST_SPECIAL,
	BVT_TCHAR = sizeof(BppChar) > 1 ? BVT_WCHAR : BVT_CHAR,
	BVT_INT = _BVTINTEGRALBASE | BNS_DWORD | BST_SIGNED, ///< 32 bits signed integer
	BVT_INTPTR = _BVTINTEGRALBASE | BNS_INTPTR | BST_SIGNED, ///< signed integer of same size as pointer
	BVT_UINT = _BVTINTEGRALBASE | BNS_DWORD | BST_UNSIGNED, ///< 32 bits unsigned integer
	BVT_UINTPTR = _BVTINTEGRALBASE | BNS_INTPTR | BST_UNSIGNED, ///< unsigned integer of same size as pointer
	BVT_CHAR32 = BVT_UINT | BIST_CHAR,
	BVT_WCHAR_INTERNAL =
#if defined(WCHAR_MAX) && (WCHAR_MAX > USHRT_MAX)
	BVT_CHAR32
#elif defined(__OLDCPP__)
	BVT_CHAR8
#else
	BVT_CHAR16
#endif
	,
	BVT_LONG = _BVTINTEGRALBASE | BNS_QWORD | BST_SIGNED, ///< 64 bits signed integer
	BVT_ULONG = _BVTINTEGRALBASE | BNS_QWORD | BST_UNSIGNED, ///< 64 bits unsigned integer
	BVT_INTEGRAL_MAX = BVT_ULONG,
	_BVTREALNUMBASE = _BVTNUMBERBASE | BNC_REAL,
	BVT_FLOAT = _BVTREALNUMBASE | BNS_DWORD, ///< single precision real type (32 bits)
	BVT_REAL_MIN = BVT_FLOAT,
	BVT_DOUBLE = _BVTREALNUMBASE | BNS_QWORD, ///< double precision real type (64 bits)
	BVT_REAL_MAX = BVT_DOUBLE,
	_BVTVALMAX = BVT_DOUBLE,
	_BVTREFTYPEBASE = BTC_REFTYPE,
	_BVTREFPTRBASE = _BVTREFTYPEBASE | BRTC_PTRTYPE,
	BVT_FILE = _BVTREFPTRBASE, ///< file descriptor pointer
	BVT_CODE, ///< native code (function) pointer
	BVT_VAR, ///< variable (contained in dictionaries)
	_BVTREFCNTBASE = _BVTREFTYPEBASE | BRTC_REFCOUNTED,
	_BVTREFCNTSINGLEBASE = _BVTREFCNTBASE | BRCC_SINGLE,
	_BVTOBJECTBASE = _BVTREFCNTSINGLEBASE | BRSC_OBJECTTYPE,
	BVT_OBJECT = _BVTOBJECTBASE, ///< B++ object (instance of class)
	BVT_RCNT_MIN = BVT_OBJECT,
	BVT_STRONGOBJECT, ///< strong typed holder of B++ object (instance of class)
	_BVTCLASSBASE = _BVTREFCNTSINGLEBASE | BRSC_CLASS,
	BVT_CLASS = _BVTCLASSBASE,  ///< B++ class descriptor object
	_BVTSEQUENCEBASE = _BVTREFCNTBASE | BRCC_SEQUENCE,
	_BVTSEQSIMPLEBASE = _BVTSEQUENCEBASE | BRSQC_SIMPLE,
	BVT_STRING = _BVTSEQSIMPLEBASE, ///< string value
	BVT_BUFFER, ///< byte buffer
	BVT_CHARBUFFER, ///< character buffer
	_BVTVECTORBASE = _BVTSEQUENCEBASE | BRSQC_VECTOR,
	BVT_VECTOR = _BVTVECTORBASE, ///< vector (dynamic array) of values
	BVT_BYTECODE, ///< bytecode vector
	BVT_LOCALVARINFO, ///< vector of local variable descriptions
	_BVTDICTBASE = _BVTSEQUENCEBASE | BRSQC_DICT,
	BVT_DICTIONARY = _BVTDICTBASE, ///< dictionary (maps key strings to values)
	BVT_RCNT_MAX = BVT_DICTIONARY,
	_BVTMAX = BVT_RCNT_MAX,
	_BVTWEAKREFBASE = _BVTREFCNTBASE | BTC_WEAKREFFLAG,
	BVT_OBJECTREF = BVT_OBJECT | BTC_WEAKREFFLAG, ///< reference to an object
	_BVTREFMIN = BVT_OBJECTREF,
	BVT_STRONGOBJECTREF = BVT_STRONGOBJECT | BTC_WEAKREFFLAG,
	BVT_CLASSREF = BVT_CLASS | BTC_WEAKREFFLAG, ///< reference to class descriptor
	BVT_STRINGREF = BVT_STRING | BTC_WEAKREFFLAG, ///< reference to string
	BVT_BUFFERREF = BVT_BUFFER | BTC_WEAKREFFLAG, ///< reference to byte buffer
	BVT_CHARBUFFERREF = BVT_CHARBUFFER | BTC_WEAKREFFLAG, ///< reference to character buffer
	BVT_VECTORREF = BVT_VECTOR | BTC_WEAKREFFLAG, ///< reference to vector
	BVT_BYTECODEREF = BVT_BYTECODE | BTC_WEAKREFFLAG, ///< reference to bytecode vector
	BVT_VARINFOREF = BVT_LOCALVARINFO | BTC_WEAKREFFLAG,
	BVT_DICTIONARYREF = BVT_DICTIONARY | BTC_WEAKREFFLAG, ///< dictionary reference
	_BVTREFMAX = BVT_DICTIONARYREF
};

inline BppValueType maskWeakRef(BppValueType t) { return (t & BTC_REFTYPE) ? BppValueType(t & ~BTC_WEAKREFFLAG) : t; }

inline BppValueType mkWcharInternal(BppValueType t) { return  t == BVT_WCHAR ? BVT_WCHAR_INTERNAL : t; }
/** Additionally attribute flags for B++ values.
@note Not all flags are used in current implementation - some of them are reserved
for further versions.
*/
enum BppValueFlags
{
	BVF_NONE = 0,
	BVF_CONST = 1, ///< constant
	BVF_STRONGTYPED = 2,///< strong typed value
	BVF_BYREF = 4, ///< byref argument (reserved)
	BVF_MUTABLE = 8, ///< mutable member variable
	BVF_SEALED = 0x10, ///< sealed class or method (reserved)
	BVF_INVALID = 0x18, ///< marks value invalid
	BVF_PRIVATE = 0x20, ///< private member
	BVF_PROTECTED = 0x40, ///< protected member
	BVF_PUBLIC = BVF_PRIVATE | BVF_PROTECTED, ///< public member
	BVF_STATIC = 0x80 /// < marks value static

};

/// Helper function for getting readable type name from type id.
const BPPLIB_DLLEXP BppChar* getValueTypeName(BppValueType type, bool getPrefix = false, bool noInvalidType = false);
#ifndef DOXY
class BppClass;
class BppObject;
class BppVector;
class BppBytecode;
class BppBuffer;
class BppCharBuffer;
class BppRefObject;
class BppString;
class BppDictionaryEntry;
class BppDictionary;
class BppInterpreter;
class BppDefaultFunctions;

	#define v_type v_exttype._type
	#define v_flags v_exttype._flags
#endif

/** B++ value object class.
All values in B++ are represented by instances of this class. Furthermore
such value objects are used by compiler and virtual machine itself.<br>
Class provides a set of constructors to create object from different data objects
and a set of conversion operators to get these data objects bach from value object.
*/
class BPPLIB_DLLEXP BppValue
{
public:
  /** @name Constructors. */
  //@{
  /// Default ctor.
  BppValue() BPPNOEXCEPT;
  /// Constructs special typed empty value.
  BppValue(BppValueType vtype, BppValueFlags flags, bool createDefaultObjects);
  /// Creation from any reference counted value by explicit type specification.
  BppValue(BppRefObject* val,  BppValueType vtype) BPPNOEXCEPT;
  /// Creation from class.
  BppValue(BppClass*, bool isRef = false) BPPNOEXCEPT;
  /// Creation from object.
  BppValue(BppObject*, bool isRef = false) BPPNOEXCEPT;
  /// Creation from Vector
  BppValue(BppVector*,bool isBytecode = false, bool isRef = false) BPPNOEXCEPT;
  /// Creation from string pointer.
  BppValue(BppString*, bool isRef = false) BPPNOEXCEPT;
  /// Creation from string.
  BppValue(const BppString&);
  /// Creation from byte buffer.
  BppValue(BppBuffer*, bool isRef = false) BPPNOEXCEPT;
  /// Creation from character buffer.
  BppValue(BppCharBuffer*, bool isRef = false) BPPNOEXCEPT;
  /// Creation from Ditionary.
  BppValue(BppDictionary*, bool isRef = false) BPPNOEXCEPT;
  /// Creation from dictionary entry (variable descriptor).
  BppValue(BppDictionaryEntry*) BPPNOEXCEPT;
  /// Creation from native (user defined) function.
  BppValue(BPPUSERFUNC) BPPNOEXCEPT;
  /// Creation from BppSByte (8 bits integer) value.
  BppValue(BppSByte) BPPNOEXCEPT;
  /// Creation from BppByte (8 bits unsigned integer) value.
  BppValue(BppByte) BPPNOEXCEPT;
  /// Creation from BppShort (16 bits integer) value.
  BppValue(BppShort) BPPNOEXCEPT;
  /// Creation from BppUShort (16 bits unsigned integer) value.
  BppValue(BppUShort) BPPNOEXCEPT;
  /// Creation from BppInt (32 bits integer) value.
  BppValue(BppInt, bool isptr = false) BPPNOEXCEPT;
  /// Creation from BppUInt (32 bits unsigned integer) value.
  BppValue(BppUInt, bool isptr = false) BPPNOEXCEPT;
  /// Creation from BppLong (64 bits integer) value.
  BppValue(BppLong, bool isptr = false) BPPNOEXCEPT;
  /// Creation from BppULong (64 bits unsigned integer) value.
  BppValue(BppULong, bool isptr = false) BPPNOEXCEPT;
  /// Creation from float (single precision real) value.
  BppValue(float) BPPNOEXCEPT;
  /// Creation from double (double precision real) value.
  BppValue(double) BPPNOEXCEPT;
  /// Creation from file descriptor.
  BppValue(FILE*) BPPNOEXCEPT;
  /// Copy constructor.
  BppValue(const BppValue& cpy, bool clone=false);
#ifdef CPP11
  BppValue(BppValue&& cpy) BPPNOEXCEPT;
#endif
  //@}
  /// Destructor.
  ~BppValue();
  /// Assignment operator.
  BppValue& operator=(const BppValue&);
#ifdef CPP11
  BppValue& operator=(BppValue&&);
#endif
  /// Assignment operator.
  bool assign(const BppValue&);

  /** Gets value type.
  @return value type attribute
  */
  inline BppValueType getType() const BPPNOEXCEPT
  {
	  return (BppValueType) v_type;
  }

  /** Gets value flags.
  @return value flags attribute.
  @see BppValueFlags
  */
  inline unsigned char getFlags() const BPPNOEXCEPT
  {
	  return v_flags;
  }

  /** Sets value flags.
  @param flags new value flags attribute.
  @see BppValueFlags
  */
  inline void setFlags(unsigned char flags) BPPNOEXCEPT
  {
	  v_flags = flags;
  }
  
  /** @name Type conversion operators. */
  //@{
  /// Conversion to class.
  operator BppClass*() const BPPNOEXCEPT;
  /// Conversion to object.
  operator BppObject*() const BPPNOEXCEPT;
  /// Conversion to vector.
  operator BppVector*() const BPPNOEXCEPT;
  /// Conversion to string.
  operator BppString*() const BPPNOEXCEPT;
  /// Conversion to byte buffer.
  operator BppBuffer*() const BPPNOEXCEPT;
  /// Conversion to character buffer.
  operator BppCharBuffer*() const BPPNOEXCEPT;
  /// Conversion to dictionary.
  operator BppDictionary*() const BPPNOEXCEPT;
  /// Conversion to variable descriptor.
  operator BppDictionaryEntry*() const BPPNOEXCEPT;
  /// Conversion to native function pointer.
  operator BPPUSERFUNC() const BPPNOEXCEPT;
  /// Conversion to file descriptor.
  operator FILE*() const BPPNOEXCEPT;
  /// Conversion to 8 bits integer.
  operator BppSByte() const BPPNOEXCEPT;
  /// Conversion to 8 bits unsigned integer.
  operator BppByte() const BPPNOEXCEPT;
  /// Conversion to 16 bits integer.
  operator BppShort() const BPPNOEXCEPT;
  /// Conversion to 16 bits unsigned integer.
  operator BppUShort() const BPPNOEXCEPT;
  /// Conversion to 32 bits integer.
  operator BppInt() const BPPNOEXCEPT;
  /// Conversion to 32 bits unsigned integer.
  operator BppUInt() const BPPNOEXCEPT;
  /// Conversion to 64 bits integer.
  operator BppLong() const BPPNOEXCEPT;
  /// Conversion to 64 bits unsigned integer.
  operator BppULong() const BPPNOEXCEPT;
  /// Conversion to single precision real.
  operator float() const BPPNOEXCEPT;
  /// Conversion to double precision real.
  operator double() const BPPNOEXCEPT;
  //@}
  /// Generic compare function
  int compare(const BppValue& cmp) const;
  /// Equality operator
  bool operator==(const BppValue& cmp) const;
  /// Non-equality operator
  bool operator!=(const BppValue& cmp) const;

  /** Get (re-interpret) value as untyped pointer.
  @return Stored value as raw pointer.
  @note Function simply re-interprets value as a native untyped pointer, i.e. in case
  of any reference type it returns pointer to data object. In case of a value type
  it behaves like a reinterpret-cast in C++.
  @see addr
  */
  void* ptr() const BPPNOEXCEPT
  {
	  return _v.v_addr;
  } 

  void* v_ptr() const BPPNOEXCEPT
  {
	  return (void*)  &_v;
  }

  bool asBool() const BPPNOEXCEPT;
  BppChar asBppChar() const BPPNOEXCEPT;

  /// Get address of data (depending on value type).
  void* addr() const BPPNOEXCEPT;
  /// Dispose value.
  void dispose(); 
  /// Assign reference type data vvalue.
  void assignRef(BppRefObject*);

  /** Gets reference type data value.
  @return Pointer to value interpreted as reference(counted) type.
  @note Function doesn't do any type check, i.e. it behaves like C++ reinterpret cast.
  */
  inline BppRefObject* getRef() const BPPNOEXCEPT
  {
	  return _v.v_refObj;
  }
  /// Get descriptor of strong typed object.
  BppVector* getStrongObjDesc() const BPPNOEXCEPT;
  /// Set protection attribute.
  void setProtection(short protflags) BPPNOEXCEPT;
  
  /**
  @result Protection flag mask of value, i.e. BVF_PRIVATE or BVF_PROTECTED or BVF_PUBLIC or BVF_NONE.
  @see setProtection
  */
  inline unsigned short getProtection() const BPPNOEXCEPT
  {
	  return v_flags & BVF_PUBLIC;
  }

  /// Checks if value type is in specified range.
  inline bool inTypeRange(BppValueType from, BppValueType to) const BPPNOEXCEPT
  {
	  return v_type >= from && v_type <= to;
  }
  /// Checks if value type is integral number.
  inline bool isIntegralType() const BPPNOEXCEPT
  {
	  return v_type >= BVT_INTEGRAL_MIN && v_type <= BVT_INTEGRAL_MAX;
  }

  inline bool isCharacterType() const BPPNOEXCEPT
  {
	  return isIntegralType() && ((v_type & BIST_CHAR) != 0);
  }

  /// Checks if value type is unsigned integral number.
  inline bool isUIntType() const BPPNOEXCEPT
  {
	  return (v_type &  (_BVTINTEGRALBASE | BST_UNSIGNED)) == (_BVTINTEGRALBASE | BST_UNSIGNED);
  }
  /// Checks if value type is signed integral number.
  inline bool isSIntType() const BPPNOEXCEPT
  {
	  return (v_type &  (_BVTINTEGRALBASE | BST_UNSIGNED)) == _BVTINTEGRALBASE;
  }

  /// Checks if value type is real number.
 inline bool isRealType() const BPPNOEXCEPT
  {
	  return v_type >= BVT_REAL_MIN && v_type <= BVT_REAL_MAX;
  }

 /// Checks if value is reference counted.
 inline bool isRefCounted() const BPPNOEXCEPT
 {
	 return v_type >= BVT_RCNT_MIN && v_type <= BVT_RCNT_MAX;
 }

 /** Checks if value equals to <tt>null</tt>.
 @return true if value type is BVT_NULL or addr() returns NULL.
 */
 inline bool isNull() const {
	 return v_type == BVT_NULL || (v_type > _BVTVALMAX && (!addr()));
 }

  /** Does raw copy of value members.
  @param dst Destination value
  @param src Source value
  @note Function copies data without any typechecking and without calling dispose()
  on destination before. It is intended for internal use or implementation of native
  extensions. 
  */
  static inline void rawcopy(BppValue& dst, const BppValue& src) BPPNOEXCEPT
  {
	  dst._v = src._v;
	  dst.v_exttype = src.v_exttype;
  }

  static BppValue makeBool(bool val);
  static BppValue makeChar(BppUInt val, BppValueType vt = BVT_TCHAR);
  static BppValue makeChar(const BppValue& val, BppValueType vt = BVT_TCHAR);
private:
#ifndef DOXY
	operator const BppString& () const;
#endif
	friend class BppInterpreter;
	friend class BppDefaultFunctions;
	/// Type descriptor structure.
	struct BPPLIB_DLLEXP ExtType
	{
		/// Value type.
		unsigned short _type : 8;
		/// Additional attribute flags for value.
		unsigned short _flags : 8;
		/// Default ctor.
		ExtType()  BPPNOEXCEPT: _type(BVT_NULL),_flags(0) {}
		/// Initialized ctor.
		ExtType(BppValueType type,unsigned short flags) BPPNOEXCEPT : _type(type),_flags(flags) {}
	} v_exttype; ///< Type specification.
  /// Data area.
  union
  {			/* value */
	 BppRefObject* v_refObj;
	 BppClass* v_class;
	 BppObject* v_object;
	 BppVector* v_vector;
	 BppString* v_string;
	 BppBuffer* v_buffer;
	 BppCharBuffer* v_charbuffer;
	 BppDictionary* v_dict;
	 BppDictionaryEntry* v_var;
	 BPPUSERFUNC v_code;
	BppSByte v_sbyte;
	BppByte v_byte;
	BppShort v_short;
	BppUShort v_ushort;
	BppWChar v_wchar;
	 BppInt v_integer;
	BppUInt v_uint;
	_Int64 v_long;
	_UInt64 v_ulong;
	BppIntPtr v_intptr;
	BppUIntPtr v_uintptr;
	float v_float;
	 double v_double;
	 void* v_addr;
	 FILE* v_fp;
	 BppUInt v_dwords[2];
	 BppUShort v_words[4];
	 BppByte v_bytes[8];
  } _v;
};

#endif // BPPVALUE_H
