#ifndef BPPVEC_H
#define BPPVEC_H

#include <bpprobj.h>
#include <bppvalue.h>

class ValPtr;
class BppDictionaryEntry;

/** B++ vector class.
This class ist the internal representation of vectors in B++ language. Furthermore
vector type is used in implementation of B++ itselfs, e.g. stack of bytecode interpreter
is implemented by a BppVector object.<br>
The BppVector class provides a dynamic array of BppValue objects.
*/
class BPPLIB_DLLEXP BppVector : public BppRefObject
{
   friend class ValPtr;
   friend class BppValue;
public:
   /// default ctor
   BppVector();
   /// ctor for vectors of initial size
   BppVector(size_t capacity,bool init=false);
   /// copy ctor
   BppVector(const BppVector& cpy);
#ifdef CPP11
   /// move ctor
   BppVector(BppVector&& cpy) BPPNOEXCEPT;
#endif
   /// dtor
   virtual ~BppVector();
   /// create clone of current object
   BppVector* clone() const;
   /// assignment
   BppVector& operator=(const BppVector& assignment);
#ifdef CPP11
   BppVector& operator=(BppVector&& assignment);
#endif

   /** Index bases element access (read/write).
   @param index index of vector element to get reference to
   @return reference (read/write) to element at index position <i>index</i>
   @note Operator does no bounds check for <i>index</i>.
   */
   inline BppValue& operator[](size_t index)
   {
      return _data[index];
   }

   /** Index based element access (read only).
   @param index index of vector element to get reference to
   @return read only reference to element at index position <i>index</i>
   @note Operator does no bounds check for <i>index</i>.
   */
   inline const BppValue& operator[](size_t index) const
   {
      return _data[index];
   }

   /** Gets current number of elements.
   @return number of elements currently contained in vector
   */
   inline size_t size() const BPPNOEXCEPT
   {
      return _size;
   }

   /** Get current capacity.
   @return current capacity ov vector
   */
   inline size_t capacity() const BPPNOEXCEPT
   {
      return _capacity;
   }
   /// set capacity
   size_t setCapacity(size_t capacity);
   /// shrink capacity to current size
   size_t shrink();
	/// modifies vector size
	bool resize(size_t size, const BppValue& fill = BppValue());

   /// create concatenation of two vectors
   BppVector operator+(const BppVector& app) const;
   /// append another vector to current one
   BppVector& operator+=(const BppVector& app);
   /// erase specified part from vector
   BppVector& erase(size_t pos, size_t cnt = BPP_NNPOS);
   /// insert value into vector
   BppVector& insert(size_t pos, const BppValue& insertion, size_t cnt = 1);
   /// swaps two values in vector
   BppVector& swap(size_t pos1, size_t pos2) BPPNOEXCEPT;
   /// moves value inside vector
   BppVector& move(size_t srcpos, size_t dstpos) BPPNOEXCEPT;
   /// Gets a sub-vector from current one.
   BppVector slice(size_t from = 0, size_t cnt = BPP_NNPOS) const;
   /// remove last element
   void pop_back();
   /// insert element at begin of vector
   void push_front(const BppValue& val);
   /// remove first element
   void pop_front();
   /// append element to end of vector
   void push_back(const BppValue& val);
#ifdef CPP11
   void push_back(BppValue&& val);
#endif
   /// get first element
   ValPtr first();
   /// get last element
   ValPtr last();
   /// remove all elements
   void clear();

   int compare(const BppVector& cmp) const;
   /// equality operator
   bool operator==(const BppVector& cmp) const;
   /// unequality operator
   bool operator!=(const BppVector& cmp) const;
   /// finds first index of <i>val</i> in vector beginning at <i>pos</i>;
   size_t find(const BppValue& val, size_t pos=0) const;
   /// finds last index of <i>val</i> in vector beginning at <i>pos</i>;
   size_t rfind(const BppValue& val, size_t pos= BPP_NNPOS) const;

   /// get index of string element in vector
   size_t getStringIndex(const BppChar* str) const BPPNOEXCEPT;
   /// get index of dictionary entry element in vector
   size_t getVarIndex(const BppDictionaryEntry* var) const;

   inline BppValue* data() const BPPNOEXCEPT
   {
      return _data;
   }
private:
   /// current size, i.e. number of elements
   size_t _size;
   /// current capacity
   size_t _capacity;
   /// pointer to data field
   BppValue* _data;
};

/** Class for vector element pointer.
This class provides encapsulation of pointer to an element inside a vector. 
It is very similar to a bidirectional iterator. Additional to iterator behavior
ValPtr class allows increase of size of related vector object.
*/
class BPPLIB_DLLEXP ValPtr
{
public:
   /// default ctor
   inline ValPtr() BPPNOEXCEPT:_owner(0), _current(-1)  { }
   /// initialized ctor
   inline ValPtr(BppVector* owner,int current)  BPPNOEXCEPT :_owner(owner), _current(current) { }
   /// copy ctor
   inline ValPtr(const ValPtr& cpy) BPPNOEXCEPT :_owner(cpy._owner), _current(cpy._current)   {  }
   /// dtor
   ~ValPtr() BPPNOEXCEPT;
   /// assignment
   ValPtr& operator=(const ValPtr& assignment) BPPNOEXCEPT;
   /// reference operator
   inline BppValue* operator->() BPPNOEXCEPT
   {
      return (BppValue*) *this;
   }

   /// dereference operator
   inline BppValue& operator*() BPPNOEXCEPT
   {
      return *(BppValue*) *this;
   }
   /// conversion to value pointer
   operator BppValue*() BPPNOEXCEPT;
   /// multi step increment
   ValPtr& operator+=(int offs);
   /// multi step decrement
   ValPtr& operator-=(int offs) BPPNOEXCEPT;
   /// pre increment
   ValPtr& operator++();
   /// pre decrement
   ValPtr& operator--() BPPNOEXCEPT;
   /// relative access by index
   inline BppValue& operator[](int offs)
   {
      return (*_owner)[_current + offs];
   }
#ifndef __OLDCPP__
   /// post increment
   const ValPtr operator++(int);
   /// post decrement
   const ValPtr operator--(int) BPPNOEXCEPT;
#endif
   /// Gets index of element currently pointing to in related vector.
   inline int current() const BPPNOEXCEPT
   {
      return _current;
   }
   /// equality operator
   bool operator==(const ValPtr& cmp) const BPPNOEXCEPT;
   /// inequality operator
   bool operator!=(const ValPtr& cmp) const BPPNOEXCEPT;
private:
   /// pointer to related vector
   BppVector* _owner;
   /// index of element pointing to in related vector
   int _current;
};

#endif // BPPVEC_H
