#ifndef BPPVM_H
#define BPPVM_H

#ifdef __cplusplus 

#include <bppdict.h>
#include <bppvalue.h>
#include <bppinput.h>
#include <bppstrng.h>

class IBppPrinter;
class BppObject;
class BppInterpreter;
class BppClass;

/** Virtual machine class.

This class provides a virtual machine. It provides all methods needed for compile and execute
a B++ program as soon as support for storing and reading compiled bytecode modules.
*/
class BPPLIB_DLLEXP BppVirtualMachine
{
public:
   /// Constructor.
   BppVirtualMachine(IBppPrinter* printer = 0);
   /// Clone (i.e. copy) all symbols from another virtual machine instance.
   void cloneSymbols(const BppVirtualMachine& other);
   /// Destructor.
   ~BppVirtualMachine();
   /// set _genLineInfo flag
   void setGenerateLineInfo(bool genInfo);
   /// get value of _genLineInfo flag
   bool getGenerateLineInfo() const;
   /// set _genLocalVarInfo flag
   void setGenerateLocalVarInfo(bool genInfo);
   /// get value of _genLocalVarInfo flag
   bool getGenerateLocalVarInfo() const;
   /// get current source name while bytecode interpreter is running
   const BppString& getCurrentSource() const;
   /// get current source line while bytecode interpreter is running
   long getCurrentLine() const;
   /// Compile source from file into bytecode.
   bool compileFile(const BppChar* filename, bool showinfo = false, bool decode = false, IBppPrinter* errPrinter = 0, IBppPrinter* dbgPrinter = 0); 
   /// Compile source from string into bytecode.
   bool compileText(const BppChar* src, bool showinfo = false, bool decode = false, IBppPrinter* errPrinter = 0, IBppPrinter* dbgPrinter = 0); 
   /// Execute a B++ function.
   bool execute(const BppChar* funcName, int argc, const BppChar** argv, bool trace = false, IBppPrinter* errPrinter = 0, IBppPrinter* dbgPrinter = 0); 
   /// Save a precompiled bytecode module to file.
   bool writeModule(const BppChar* filename);
   /// Save a precompiled bytecode module to file.
   bool writeModule(FILE* fp);

   /// Read a precompiled bytecode module from file.
   bool readModule(const BppChar* filename);
   /// Read a precompiled bytecode module from file.
   bool readModule(FILE* fp);
   /// Get result of execution.
   const BppValue& getResult() const;
   /// Handle deletion of object.
   void notifyObjDelete(BppObject* obj);
   /// Do complete reinitialization of virtual machine.
   void reinit();
   /// Access to class table.
   BppDictionary& classes();
   /// Access to symbol table.
   BppDictionary& symbols();
   /// Access to native module table.
   BppDictionary& modules();
   /// Access to currently used printer interface.
   IBppPrinter& printer();
   /// Get bytecode interpreter instance.
   BppInterpreter* interpreter() const;
   /// Access to number of command line arguments.
   int& cmdargc();
   /// Access to array of command line arguments.
   BppChar**& cmdargv();
   /// Register a native function for use with B++.
   bool registerNativeFunction(const BppChar* name, BPPUSERFUNC pfunc);
   /// Register a native class for use with B++.
   bool registerNativeClass(const BppChar* name, BppClass* pClass);
#ifndef NO_DYNLIB
   /// load native module dll
   bool loadNativeModule(const BppChar* dllName);
#endif
   /// Invoke a global function.
   bool invoke(BppValue& result, const BppChar* name, BppVector* args = 0);
   /// Invoke member function or function assigned to member variable of an object.
   bool invokeMember(BppValue& result, BppObject* pobj, const BppChar* name, BppVector* args = 0);
   /// Invoke static method or function assigned to static member variable of a class.
   bool invokeStatic(BppValue& result, BppClass* pclass, const BppChar* name, BppVector* args = 0);
   /// call installed debug wait procedure (if any)
   bool dbgWait() const;
   /// call installed debug break request procedure (if any)
   bool breakRequest() const;
   /// call installed line changed notification procedure (if any)
   bool notifySourceLineChanged(const BppChar* sourceName, long line);
   /// install debug wait procedure 
   BppDbgWaitProc setDbgWaitProc(BppDbgWaitProc dbgWaitProc);
   /// install debug break request procedure
   BppBreakRequestProc setBreakRequestProc(BppBreakRequestProc breakRequestProc);
   /// install debug line changed notification procedure
   BppLineChangedProc setLineChangedProc(BppLineChangedProc lineChangedProc);
   /// cancel running byte code interpreter
   void cancel();
   bool isInDbg() const;
   void setInDbg(bool inDbg);
	/// Gets access to table of predefined symbols.
   BppDictionary& predefines() { return _predefined; }
	/// Adds or replaces entry to table of predefined symbols. 
   void addPredefinedSymbol(const BppChar* name, const BppChar* val = 0);
	/// Removes entry from table of predefined symbols.
   bool removePredefinedSymbol(const BppChar* name);
#ifdef WIN32
   /// set handle of window for processing debug wait and break request messages
   void setDbgWindow(HWND hwndDbg);
   /// get id of break request message
   static UINT getBreakRequestMsg();
   /// get id of debug wait message
   static UINT getDbgWaitMsg();
   /// get id of line changed notification message
   static UINT getLineChangedMsg();
#endif
   /// get current input object
   IBppInput* getCurrentInput() const;
   /// set current input object
   IBppInput* setCurrentInput(IBppInput* currentInput);
   /// Searches for value type by type name in source code.
   BppValueType findVTypeFromName(const BppString& name, bool includeClasses = false, BppClass** classType = 0, BppClass* ownerClass = 0);
   /// Gets type name (notation in sourcecode) for specified value type.
   const BppChar* getTypeNameFromVType(BppValueType vtype);
private:
   /// copy ctor forbidden
   BppVirtualMachine(const BppVirtualMachine&);
   /// assignment forbidden
   BppVirtualMachine& operator=(const BppVirtualMachine&);
   /// flag for compiling line info into bytecode
   bool _genLineInfo;
   /// flag for generate local variables info in bytecode
   bool _genLocalVarInfo;
   /// class table
   BppDictionary _classes;
   /// symbol table
   BppDictionary _symbols;
   /// module table
   BppDictionary _modules;
   /// table of additional predefined literals for compiler
   BppDictionary _predefined;
   /// pointer to printer interface
   IBppPrinter* _printer;
   /// pointer to bytecode interpreter
   BppInterpreter* _interpreter;
   /// flag for ownership of printer
   bool _ownsPrinter;
   /// execution result
   BppValue _result;
   /// number of command line arguments
   int _cmdargc;
   /// array of commend line arguments
   BppChar** _cmdargv;
   /// installable wait procedure for debugging purposes
   BppDbgWaitProc _dbgWaitProc;
   /// installable wait procedure for debugging purposes
   BppBreakRequestProc _breakRequestProc;
   /// installable line changed notification procedure for debugging purposes
   BppLineChangedProc _lineChangedProc;
   /// Input object for reading from stdin
#ifdef WIN32
   /** Handle of debug window. 
   Debug window handle may be used alternatively to callbacks for
   break requests and debug wait handling.
   */
   HWND _hwndDbg;
   /// message id of debug wait message
   static UINT _dbgWaitMsg;
   /// message id of break request message
   static UINT _breakRequestMsg;
   /// message id of line changed notification message
   static UINT _lineChangedMsg;
#endif
   BppFileStreamInput _defaultInput;
   /// Pointer to currently set input object
   IBppInput* _currentInput;
   /// Buffer for current source name
   BppString _currentSource;
   /// Buffer for current line number
   long _currentLine;
   /// Flag set when in debugger callback.
   bool _inDbg;
};

#else
   #include <bppdef.h>
#endif // __cplusplus

#ifndef NO_CAPI
#ifdef __cplusplus
extern "C" {
#endif

/** @defgroup C_API B++ C-API
In order to make B++ useable for language environments other than C++, a set of
C-functions is provided.<br>
For all functions using string parameters there are two versions: One for
single byte characters (suffix A) and another for wide characters (2 bytes, suffix W).
@note On WIN32 platforms all of these functions must be called by 
STDCALL-convention (like Windows API functions) where C-convention is used for other 
platforms, i.e.
<ul>
   <li>Arguments are expected on stack in reverse order, i.e. last argument
       must be pushed t stack first.</li>
   <li>Return value is contained in EAX register for 32-bit processors.
   For 16-bit processors results of 16 bits with are returned in AX register where
   32-bit results are returned in DX:AX register pair.</li>
   <li>On WIN32 functions remove their arguments from stack itself.
   For other platforms caller is responsible to do that.</li>
</ul>
@{
*/
  /// Create new virtual machine.
   void* BPPAPI Bpp_CreateVM(void* pPrinter);
  /// Release a virtual machine.
   int BPPAPI Bpp_ReleaseVM(void* pVM);
  /// Get value of line info generation flag.
   int BPPAPI Bpp_GetGenerateLineInfo(void* pVM);
  /// Set value of line info generation flag.
   void BPPAPI Bpp_SetGenerateLineInfo(void* pVM, int bGenInfo);
  /// Get value of line info generation flag.
   int BPPAPI Bpp_GetGenerateLocalVarInfo(void* pVM);
  /// Set value of line info generation flag.
   void BPPAPI Bpp_SetGenerateLocalVarInfo(void* pVM, int bGenInfo);
   /// Adds predefined symbol for compilation (single byte version).
   void BPPAPI Bpp_AddPredefinedSymbolA(void* pVM, char* pcszName, char* pcszVal);
   /// Removes definition of predefined symbol (single byte version).
   int BPPAPI Bpp_RemovePredefinedSymbolA(void* pVM, char* pcszName);
	/// Removes all predefined symbols.
	void BPPAPI Bpp_ClearPredefinedSymbols(void* pVM);
   /// Compile source from file (single byte character version).
   int BPPAPI Bpp_CompileFileA(void* pVM,
                                            char* pcszFileName, 
                                            int bShowInfo, 
                                            int bDecode, 
                                            void* pErrPrinter, 
                                            void* pDbgPrinter);  
  /// Compile source from string (single byte character version).
   int BPPAPI Bpp_CompileTextA(void* pVM,
                                            char* pcszSrc, 
                                            int bShowInfo, 
                                            int bDecode, 
                                            void* pErrPrinter, 
                                            void* pDbgPrinter);  
  /// Execute a function (single byte character version).
   int BPPAPI Bpp_ExecuteA(void* pVM,
                                       char* pcszFuncName, 
                                       int iArgc, 
                                       char** ppcszArgv,
                                       int bTrace,
                                       void* pErrPrinter, 
                                       void* pDbgPrinter); 
  /// Save compiled module to file (single byte character version).
   int BPPAPI Bpp_WriteModuleA(void* pVM,
                                           char* pcszFileName);  
  /// Read compiled module from file (single byte character version).
   int BPPAPI Bpp_ReadModuleA(void* pVM,
                                           char* pcszFileName); 
  /// Get current source name while execution (single byte character version).
   char* BPPAPI Bpp_GetCurrentSourceA(void* pVM);
  /// Get current source line while execution.
   long BPPAPI Bpp_GetCurrentLine(void* pVM);
#ifndef __OLDCPP__
	/// Adds predefined symbol for compilation (wide character version).
   void BPPAPI Bpp_AddPredefinedSymbolW(void* pVM, wchar_t* pcszName, wchar_t* pcszVal);
   /// Removes definition of predefined symbol (wide character).
   int BPPAPI Bpp_RemovePredefinedSymbolW(void* pVM, wchar_t* pcszName);
   /// Compile source from file (wide character version).
   int BPPAPI Bpp_CompileFileW(void* pVM,
                                            wchar_t* pcszFileName, 
                                            int bShowInfo, 
                                            int bDecode, 
                                            void* pErrPrinter, 
                                            void* pDbgPrinter);  
  /// Compile source from string (wide character version).
   int BPPAPI Bpp_CompileTextW(void* pVM,
                                            wchar_t* pcszSrc, 
                                            int bShowInfo, 
                                            int bDecode, 
                                            void* pErrPrinter, 
                                            void* pDbgPrinter);  
  /// Execute a function (wide character version).
   int BPPAPI Bpp_ExecuteW(void* pVM,
                                       wchar_t* pcszFuncName, 
                                       int iArgc, 
                                       wchar_t** ppcszArgv,
                                       int bTrace,
                                       void* pErrPrinter, 
                                       void* pDbgPrinter);
  /// Save compiled module to file (wide character version).
   int BPPAPI Bpp_WriteModuleW(void* pVM,
                                           wchar_t* pcszFileName);  
  /// Read compiled module from file (wide character version).
   int BPPAPI Bpp_ReadModuleW(void* pVM,
                                           wchar_t* pcszFileName); 
  /// Get current source name while execution (wide character version).
   wchar_t* BPPAPI Bpp_GetCurrentSourceW(void* pVM);
#endif // !NO_TCHAR_H
  /// Get execution result as integer
   int BPPAPI Bpp_GetIntResult(void* pVM);
  /// Do reinitialization of virtual machine.
   void BPPAPI Bpp_Reinit(void* pVM);

  /// Create a console printer.
   void* BPPAPI Bpp_CreateConsolePrinter();
  /// Create a callback printer.
   void* BPPAPI Bpp_CreateCallbackPrinter(BppPrinterCallback pCallback);
#ifdef WIN32
  /// Create a windows message based printer.
   void* BPPAPI Bpp_CreateWinMsgPrinter(HWND hwndDest);
  /// Get ID of print string message.
   UINT BPPAPI Bpp_GetPrintStringMsg();
#endif
  /// Release a printer instance.
   int BPPAPI Bpp_ReleasePrinter(void* pPrinter);
  /// Create a callback based input object
   void* BPPAPI Bpp_CreateCallbackInput(BppCharInputCallback pcCallback, BppStringInputCallback psCallback);
#ifdef WIN32
  /// Create a windows message based input object
   void* BPPAPI Bpp_CreateWinMsgInput(HWND hwndDest);
  /// Get ID of string input message.
   UINT BPPAPI Bpp_GetStringInputMsg();
  /// Get ID of string input message.
   UINT BPPAPI Bpp_GetCharInputMsg();
  /// Set string input value for specified WinMsgInput instance (single byte character version)
   void BPPAPI Bpp_SetWinInputStringA(void* pMsgWinInput, char* pcszInputStr);
  /// Set string input value for specified WinMsgInput instance (wide character version)
   void BPPAPI Bpp_SetWinInputStringW(void* pMsgWinInput, wchar_t* pcszInputStr);
  /// Set character input value for specified WinMsgInput instance (single byte character version)
   void BPPAPI Bpp_SetWinInputCharA(void* pMsgWinInput, char cInputChr);
  /// Set character input value for specified WinMsgInput instance (wide character version)
   void BPPAPI Bpp_SetWinInputCharW(void* pMsgWinInput, wchar_t cInputChr);
  /// Set handle of window processing break request and debug wait messages.
   void BPPAPI Bpp_SetDbgWindow(void* pVM, HWND hwndDbg);
  /// Get ID of break request message.
   UINT BPPAPI Bpp_GetBreakRequestMsg();
  /// Get ID of debug wait message.
   UINT BPPAPI Bpp_GetDbgWaitMsg();
  /// Get ID of line changed notification message.
   UINT BPPAPI Bpp_GetLineChangedMsg();
#endif
  /// Release an input object instance.
   int BPPAPI Bpp_ReleaseInput(void* pInput);
  /// Set input object for VM.
   void* BPPAPI Bpp_SetCurrentInput(void* pVM, void* pInput);
  /// Get currently used input object for specified VM
   void* BPPAPI Bpp_GetCurrentInput(void* pVM);
  /// Install debug wait procedure.
   BppDbgWaitProc BPPAPI Bpp_SetDbgWaitProc(void* pVM, BppDbgWaitProc dbgWaitProc);
  /// Install debug break request procedure.
   BppBreakRequestProc BPPAPI Bpp_SetBreakRequestProc(void* pVM, BppBreakRequestProc breakRequestProc);
  /// Install debug line changed notificationt procedure.
   BppLineChangedProc BPPAPI Bpp_SetLineChangedProc(void* pVM, BppLineChangedProc lineChangedProc);
  /// Cancel running byte code interpreter.
   void BPPAPI Bpp_Cancel(void* pVM);
   /// Gets Value of environment variable (single byte version)
   char* BPPAPI Bpp_GetEnvA(char* pcszName);
   /// Sets Value of environment variable (single byte version)
   void BPPAPI Bpp_SetEnvA(char* pcszName,char* pcszValue);
#ifndef __OLDCPP__
   /// Gets Value of environment variable (wide character version)
   wchar_t* BPPAPI Bpp_GetEnvW(wchar_t* pcszName);
   /// Sets Value of environment variable (wide character version)
   void BPPAPI Bpp_SetEnvW(wchar_t* pcszName,wchar_t* pcszValue);
#endif
#ifdef __cplusplus
}
#endif


#ifdef _UNICODE
  #define Bpp_AddPredefinedSymbol Bpp_AddPredefinedSymbolW
  #define Bpp_RemovePredefinedSymbol Bpp_RemovePredefinedSymbolW
  #define Bpp_CompileFile Bpp_CompileFileW
  #define Bpp_CompileText Bpp_CompileTextW
  #define Bpp_Execute Bpp_ExecuteW
  #define Bpp_WriteModule Bpp_WriteModuleW
  #define Bpp_ReadModule Bpp_ReadModuleW
  #define Bpp_GetCurrentSource Bpp_GetCurrentSourceW
#else
  #define Bpp_AddPredefinedSymbol Bpp_AddPredefinedSymbolA
  #define Bpp_RemovePredefinedSymbol Bpp_RemovePredefinedSymbolA
  #define Bpp_CompileFile Bpp_CompileFileA
  #define Bpp_CompileText Bpp_CompileTextA
  #define Bpp_Execute Bpp_ExecuteA
  #define Bpp_WriteModule Bpp_WriteModuleA
  #define Bpp_ReadModule Bpp_ReadModuleA
  #define Bpp_GetCurrentSource Bpp_GetCurrentSourceA
#endif

/// @}
#endif // !NO_CAPI
#endif // BPPVM_H
