
#include "bppdef.h"
#include "bppbuf.h"
#include "bppufunc.h"
#include <stdlib.h>

/// a buffer at least can contain one element
static const size_t MINCAPACITY = 1;

/**
Default constructor allocates memory for a buffer of minimum capacity. Elements are initialized to
null.
@see MINCAPACITY
*/
BppBuffer::BppBuffer()
:_size(0),_capacity(MINCAPACITY)
{
	_data = (BppByte*) allocmem(_capacity);
	memset(_data,0,_capacity);
}

/**
This constructor allocates memory for a buffer of specified capacity. Elements are initialized to
null.
@param capacity initially capacity of buffer in bytes
@param size initially size of buffer in bytes
@note If <i>size</i> is greater than <i>capacity</i> it is set to value of <i>capacity</i>.
@see MINCAPACITY
*/
BppBuffer::BppBuffer(size_t capacity, size_t size)
:_size(size),_capacity(capacity>=MINCAPACITY ? capacity : MINCAPACITY)
{
   if (_size > _capacity) _size = _capacity;
	_data = (BppByte*) allocmem(_capacity);
	if (_data != 0)
		memset(_data,0,_capacity);
}

/**
@param cpy buffer instance to create copy from
*/
BppBuffer::BppBuffer(const BppBuffer& cpy)
:_size(cpy._size),_capacity(cpy._capacity)
{
	_data = (BppByte*) allocmem(_capacity);
	if (_data)
	  memcpy(_data,cpy._data,_capacity);
}

#ifdef CPP11

BppBuffer::BppBuffer(BppBuffer&& cpy) BPPNOEXCEPT
	:_size(cpy._size), _capacity(cpy._capacity),_data(cpy._data)
{
	_size = 0;
	_capacity = 0;
	_data = 0;
}
#endif
BppBuffer::~BppBuffer()
{
	if (_data != 0)
		freemem(_data);
	_data = 0;
}

/**
@return new copy of current buffer
@note Caller is responsible for delete result if no longer needed.
*/
BppBuffer* BppBuffer::clone() const
{
	return new BppBuffer(*this);
}

/**
@param assignment buffer instance to assign data from
@return Self-reference of buffer object.
@note <ul>
<li>Operator makes a copy of all data stored in <i>assignmant</i>.</li>
<li>Address of internal data buffer may be changed by this operator.</li>
</ul>
*/
BppBuffer& BppBuffer::operator=(const BppBuffer& assignment)
{
	if (&assignment == this)
		return *this;
	BppByte* newdata = (BppByte*) allocmem(assignment._capacity);
	if (newdata) memcpy(newdata,assignment._data,assignment._capacity);
	if (_data) freemem(_data);
	_data = newdata;
	_capacity = assignment._capacity;
	_size = assignment._size;
	return *this;
}
#ifdef CPP11
BppBuffer & BppBuffer::operator=(BppBuffer && assignment)
{
	if (&assignment == this)
		return *this;
	if (_data && _data != assignment._data) {
		freemem(_data);
	}
	_data = assignment._data;
	_capacity = assignment._capacity;
	_size = assignment._size;
	assignment._data = 0;
	assignment._capacity = 0;
	assignment._size = 0;
	return *this;

}
#endif


/**
@param pos Index of word in buffer.
@return Word (16 bits) at position <i>pos</i>.
*/
BppWord BppBuffer::getW(size_t pos) const
{
#ifndef WINCE
	return *(BppWord*)(&(_data[pos]));
#else
	BppWord result;
	memcpy(&result,&(_data[pos]),sizeof(BppWord));
	return result;
#endif
}

/**
@param pos Index of dword in buffer.
@return DWord (32 bits) at position <i>pos</i>.
*/
BppDWord BppBuffer::getDW(size_t pos) const
{
#ifndef WINCE
	return *(BppDWord*)(&(_data[pos]));
#else
	BppDWord result;
	memcpy(&result,&(_data[pos]),sizeof(BppDWord));
	return result;
#endif
}

/**
@param pos Index of native word in buffer.
@return Native word (16 or 32 bits, depending on platform) at position <i>pos</i>.
*/
BppNWord BppBuffer::getNW(size_t pos) const
{
#ifndef WINCE
	return *(BppNWord*)(&(_data[pos]));
#else
	BppNWord result;
	memcpy(&result,&(_data[pos]),sizeof(BppNWord));
	return result;
#endif
}

/**
@param pos Index of word in buffer.
@param val Value to write at position <i>pos</i> into buffer.
*/
void BppBuffer::setW(size_t pos, BppWord val)
{
	memcpy(&(_data[pos]),&val,sizeof(val));
}

/**
@param pos Index of dword in buffer.
@param val Value to write at position <i>pos</i> into buffer.
*/
void BppBuffer::setDW(size_t pos, BppDWord val)
{
	memcpy(&(_data[pos]),&val,sizeof(val));
}

/**
@param pos Index of native word in buffer.
@param val Value to write at position <i>pos</i> into buffer.
*/
void BppBuffer::setNW(size_t pos, BppNWord val)
{
	memcpy(&(_data[pos]),&val,sizeof(val));
}


/**
Function changes capacity of buffer to specified number of elements. If new capacity is larger than
old one, additional allocated mamory is initialized with NULL bytes.
@param requested new capacity
@return really available new capacity after resizing
@note Function never discards elements managed by buffer, i.e. if <i>capacity</i> is lower than
current size (result of size() method) new capacity will be equal to size.
*/
size_t BppBuffer::setCapacity(size_t capacity)
{
	if (capacity == 0) capacity = MINCAPACITY;
	if (capacity < _size)
	{
		capacity = _size;
	}
	void* p = reallocmem(_data, capacity);
	_data = (BppByte*)p;
	if (capacity > _capacity)
	{
		memset(&_data[_capacity], 0, capacity - _capacity);
	}
	_capacity = capacity;
	return _capacity;
}


/**
Function adjusts capacity to current size, i.e. it releases unused memory.
@return new capacity
*/
size_t BppBuffer::shrink()
{
	return setCapacity(_size);
}

/**
Function erases maximum <i>cnt</i> bytes from current buffer, starting
at position <i>pos</i>.
@param pos start position for erase
@param cnt maximum number of bytes to delete
@return reference to current buffer
@note If <i>cnt</i> greater than number of bytes behind <i>pos</i>,
all bytes from <i>pos</i> to end of buffer will be erased.
*/
BppBuffer & BppBuffer::erase(size_t pos, size_t cnt)
{
	if ((_size == 0) || (pos >= _size)) return *this;
	if (cnt >(_size - pos))
		cnt = _size - pos;
	if (cnt > 0)
	{
		size_t moveCnt = _size - (pos + cnt);
		for (size_t i = 0; i<moveCnt; ++i)
		{
			_data[pos + i] =  _data[pos + cnt + i];
		}
		while (cnt > 0)
		{
			_data[--_size] = 0;
			--cnt;
		}
	}
	return *this;
}

/**
Function inserts <i>cnt</i> bytes of value <i>insertion</i> into current buffer.
@param pos insertion position
@param insertion byte to insert
@param cnt number of bytes to insert
@return reference to current buffer
*/
BppBuffer & BppBuffer::insert(size_t pos, BppByte insertion, size_t cnt)
{
	if (cnt > 0)
	{
		if (pos > _size) pos = _size;
		size_t newsize = _size + cnt;
		if (_capacity < newsize) setCapacity(newsize);
		if (_capacity < newsize) errorExit(BP_ERR_NOMEM);
		BppByte* p2 = &_data[newsize];
		BppByte* p1 = &_data[_size];
		size_t ccnt = _size - pos;
		for (size_t i = 0; i<ccnt; ++i, --p1, --p2)
		{
			*p2 = *p1;
		}
		p1 = _data + pos;
		for (size_t j = 0; j<cnt; ++j, ++p1) *p1 = insertion;
		_size = newsize;
	}
	return *this;
}

/**
@param pos1 position of first element (Byte)
@param pos2 position of second element (Byte)
@return reference to buffer
@note In case of invalid value of <i>pos1</i> or <i>pos2</i> method does nothing.
*/
BppBuffer& BppBuffer::swap(size_t pos1, size_t pos2) BPPNOEXCEPT
{
	if (pos1 != pos2 && pos1 < _size && pos2 < _size)
	{
		BppByte m = _data[pos1];
		_data[pos1] = _data[pos2];
		_data[pos2] = m;
	}
	return *this;
}


/**
@param srcpos position of element (byte) to move
@param dstpos destination position
@return reference to buffer
@note In case of invalid value of <i>srcpos</i> or <i>srcpos</i> method does nothing.
*/
BppBuffer& BppBuffer::move(size_t srcpos, size_t dstpos) BPPNOEXCEPT
{
	if (srcpos == dstpos || srcpos >= _size || dstpos >= _size) return *this;
	BppByte p = _data[srcpos];
	if (srcpos < dstpos)  // shift left
	{
		for (size_t i = srcpos; i<dstpos; ++i)
			_data[i] = _data[i + 1];
	}
	else // shift right
	{
		for (size_t i = srcpos; i>dstpos; --i)
			_data[i] = _data[i - 1];
	}
	_data[dstpos] = p;
	return *this;
}



/**
Operator creates a new buffer object containing concatenation of own value and value
of <i>app</i>. Capacity (and size) of result is the sum of sizes of both operands.
@param app other buffer to concatenate with this one
@return new buffer containing concatenation of operands
*/
BppBuffer BppBuffer::operator+(const BppBuffer& app)
{
	BppBuffer tmp(*this);
	return tmp += app;
}

/**
Operator appends contents of <i>app</i> to current one.
@param app buffer to append contents from
@return reference to this object
@note If current capacity is lower than sum of both buffer sizes, it is enlarged to this sum.
*/
BppBuffer& BppBuffer::operator+=(const BppBuffer& app)
{
	size_t newsize = _size + app._size;
	if (_capacity < newsize)
		setCapacity(newsize);
	memcpy(_data+_size,app._data,app._size);
	_size += app._size;
	return *this;
}

/**
Function removes last byte from buffer.
@return byte removed.
@note This function does not affect capacity. Memory space of removed byte is set to null.
*/
BppByte BppBuffer::pop_back()
{
  if (_size == 0)
	  return 0;
  BppByte tmp = _data[--_size];
  _data[_size] = 0;
  return tmp;
}

/**
Function removes last word (i.e. two bytes) from buffer.
@return word removed.
@note This function does not affect capacity. Memory space of removed word is set to null.
*/
BppWord BppBuffer::pop_backW()
{
	BppWord w = pop_back() << 8;
	w |= pop_back();
	return w;
}

/**
Function removes last double word (i.e. four bytes) from buffer.
@return double word removed.
@note This function does not affect capacity. Memory space of removed double word is set to null.
*/
BppDWord BppBuffer::pop_backDW()
{
	BppDWord dw = pop_backW() << 16;
	dw |= pop_backW();
	return dw;
}

/**
Function removes last native word from buffer. Number of bytes removed depends on word size
of machine (or operating system). On a 16-Bit machine this function is same as pop_backW where
on a 32-Bit machine same as pop_backDW.
@return native word removed.
@note This function does not affect capacity. Memory space of removed native word is set to null.
*/
BppNWord BppBuffer::pop_backNW()
{
#if (!defined(GCC) && (INT_MAX == SHRT_MAX))
	return pop_backW();
#else
	return pop_backDW();
#endif
}

/**
Function appends any number of bytes from <i>psrc</i> to end of buffer.
@param psrc Address of soruce to copy bytes from
@param bytes Number of bytes to append
@return size of buffer before appending.
*/
size_t  BppBuffer::append(const void * psrc, size_t bytes)
{
   size_t s = _size;
   if (bytes) {
      _size += bytes;
      if (_size > _capacity) {
         setCapacity(bytes <= 8 ? _size + 10 * bytes : _size);
      }
      BppByte* pdst = _data + s;
      memcpy(pdst, psrc, bytes);
   }
   return s;
}

/**
Function pushes (i.e. appends) a single byte to end of buffer.
@param b byte to push
@return size of buffer before appending <i>b</i>
*/
size_t BppBuffer::push_back(BppByte b)
{
	if (_size == _capacity)
		setCapacity(10+_size);
	_data[_size] = b;
	return (size_t) _size++;
}

/**
Function pushes (i.e. appends) a word (two bytes) to end of buffer.
@param w word to push
@return size of buffer (in bytes) before appending <i>w</i>
*/
size_t BppBuffer::push_backW(BppWord w)
{
  return append(&w, sizeof(BppWord));
}

/**
Function pushes (i.e. appends) a double word (four bytes) to end of buffer.
@param dw double word to push
@return size of buffer (in bytes) before appending <i>dw</i>
*/
size_t BppBuffer::push_backDW(BppDWord dw)
{
   return append(&dw, sizeof(BppDWord));
}

/**
Function pushes (i.e. appends) a native word to end of buffer. Number of bytes appended depends on
word size of machine (or operating system). On a 16-Bit machine this function is same as push_backW where
on a 32-Bit machine same as push_backDW.
@param nw native word to push
@return size of buffer (in bytes) before appending <i>nw</i>
*/
size_t BppBuffer::push_backNW(BppNWord nw)
{
#if (!defined(GCC) && (INT_MAX == SHRT_MAX))
	return push_backW(nw);
#else
	return push_backDW(nw);
#endif
}

/**
Function removes all managed elements of buffer, i.e. it sets current size to 0 and overwrites
all elements with null value.
*/
void BppBuffer::clear()
{
	if ((_data != 0) && (_size != 0))
		memset(_data,0,_size);
	_size = 0;
}



/**
Function simply sets internal _size variable without changing any data nor capacity.
@param size new size of buffer
@return true on sucess, false if <i>size</i> is greater than current capacity
*/
bool BppBuffer::setSize(size_t size)
{
	if (size > _capacity) return false;
	_size = size;
	return true;
}

/**
Function  resizes buffer to <i>size</i> bytes.
If <i>size</i> is less than current size, bytes are erased from end of buffer.
If <i>size</i> is greater than current size, new elements of value <i>fill</i> are appended.
@param size New size (number of bytes) of buffer.
@param fill Value to use for new elements.
@return true on success, false in case of memory error.
*/
bool BppBuffer::resize(size_t size, BppByte fill)
{
	if (size > _capacity) {
		if (setCapacity(size) < size) return false;
	}
	while (_size < size) {
		_data[_size] = fill;
		++_size;
	}
	return setSize(size);
}

/**
@param from Index position of start of subsequence.
@param cnt Maximum number of elements to copy into result.
@return Returns new <tt>buffer</tt> containing copy of specified part from current one.
*/
BppBuffer BppBuffer::slice(size_t from, size_t cnt) const
{
	size_t maxcnt = (from <= _size) ? _size - from : 0;
	if (cnt > maxcnt) cnt = maxcnt;
	if (cnt == 0) return BppBuffer();
	BppBuffer result(cnt,cnt);
	for (size_t i = 0; i<cnt; ++i)
		result[i] = _data[from + i];
	return result;
}

int BppBuffer::compare(const BppBuffer& comparison) const BPPNOEXCEPT
{
	if (&comparison == this) return true;
	if (_size == 0) return comparison._size ? -1 : 0;
	if (comparison._size == 0) return 1;
	size_t s = _size > comparison._size ? comparison._size : _size;
	int r = memcmp(_data,comparison._data,s);
	if (r != 0) return r;
	return _size == comparison._size ? 0 : s < _size ? 1 : -1; 
}

bool BppBuffer::operator==(const BppBuffer& cmp) const BPPNOEXCEPT
{
	if (&cmp == this) return true;
	if (_size != cmp._size) return false;
	if (_size == 0) return true;
	 return memcmp(_data,cmp._data,_size) == 0;
}

bool BppBuffer::operator!=(const BppBuffer& cmp) const BPPNOEXCEPT
{
	return (!(this->operator==(cmp)));
}

size_t BppBuffer::find(BppByte val, size_t pos) const
{
	if (pos >= _size) return BPP_NNPOS;
	BppByte* p = (BppByte*) memchr(_data + pos, val, _size - pos);
	return p ? size_t(p - _data) : BPP_NNPOS;
}

size_t BppBuffer::rfind(BppByte val, size_t pos) const
{
	if (_size == 0) return BPP_NNPOS;
	if (pos >= _size) pos = _size - 1;
	for (BppByte* p = _data + pos; p >= _data; --p, --pos) {
		if (*p == val) return pos;
		if (p == _data) break;
	}
	return BPP_NNPOS;
}

/**
Default constructor allocates memory for a buffer of minimum capacity. Elements are initialized to
null character.
@see MINCAPACITY
*/
BppCharBuffer::BppCharBuffer()
:_size(0),_capacity(MINCAPACITY)
{
	_data = (BppChar*) allocmem(_capacity * sizeof(BppChar));
	memset(_data,0,_capacity * sizeof(BppChar));
}

/**
This constructor allocates memory for a buffer of specified capacity. Elements are initialized to
null character.
@param capacity initially capacity of buffer in characters
@param size initially size of buffer in bytes
@note If <i>size</i> is greater than <i>capacity</i> it is set to value of <i>capacity</i>.
@see MINCAPACITY
*/
BppCharBuffer::BppCharBuffer(size_t capacity,size_t size)
:_size(size),_capacity(capacity>=MINCAPACITY ? capacity : MINCAPACITY)
{
   if (_size > _capacity) _size = _capacity;
	_data = (BppChar*) allocmem((1+_capacity) * sizeof(BppChar));
	memset(_data,0,(1+_capacity) * sizeof(BppChar));
}

/**
@param cpy buffer instance to create copy from
*/
BppCharBuffer::BppCharBuffer(const BppCharBuffer& cpy)
:_size(cpy._size),_capacity(cpy._capacity)
{
	_data = (BppChar*) allocmem((_capacity+1) * sizeof(BppChar));
	if (_data) {
		memcpy(_data, cpy._data, _capacity * sizeof(BppChar));
		_data[_capacity] = 0;
	}
	//memset(_data,0,_capacity * sizeof(BppChar));
	//for (size_t i=0;i<_size;i++)
	//   _data[i] = cpy[i];
}
#ifdef CPP11

BppCharBuffer::BppCharBuffer(BppCharBuffer&& cpy) BPPNOEXCEPT
	:_size(cpy._size), _capacity(cpy._capacity), _data(cpy._data)
{
	cpy._size = 0;
	cpy._capacity = 0;
	cpy._data = 0;
}
#endif

BppCharBuffer::BppCharBuffer(const BppChar* cstr)
:_size(0),_capacity(MINCAPACITY)
{
	size_t len = cstr ? _tcslen(cstr) : 0;
	if (len > 0) 
	{
		_capacity = len+1;
		_size = len;
		_data = (BppChar*) allocmem((1+_capacity) * sizeof(BppChar));
		_tcscpy(_data,cstr);
	}
	else
	{
		_data = (BppChar*) allocmem((1+_capacity) * sizeof(BppChar));
		memset(_data,0,_capacity * sizeof(BppChar));
	}
	_data[_capacity] = 0;
}


BppCharBuffer::~BppCharBuffer()
{
	if (_data != 0)
		freemem(_data);
	_data = 0;
}

/**
@return new copy of current buffer
@note Caller is responsible for delete result if no longer needed.
*/
BppCharBuffer* BppCharBuffer::clone() const
{
	return new BppCharBuffer(*this);
}
#ifdef CPP11
BppCharBuffer & BppCharBuffer::operator=(BppCharBuffer && assignment)
{
	if (&assignment == this)
		return *this;
	if (_data && _data != assignment._data) freemem(_data);
	_data = assignment._data;
	_capacity = assignment._capacity;
	_size = assignment._size;
	assignment._data = 0;
	assignment._capacity = 0;
	assignment._size = 0;
	return *this;
}
#endif
/**
@param assignment character buffer instance to assign data from
@return Self-reference of buffer object.
@note <ul>
<li>Operator makes a copy of all data stored in <i>assignmant</i>.</li>
<li>Address of internal data buffer may be changed by this operator.</li>
</ul>
*/
BppCharBuffer& BppCharBuffer::operator=(const BppCharBuffer& assignment)
{
	if (&assignment == this || assignment._data == _data)
		return *this;
	BppChar* newdata = (BppChar*) allocmem((assignment._capacity+1) * sizeof(BppChar));
	if (newdata) {
		memmove(newdata, assignment._data, assignment._capacity * sizeof(BppChar));
		if (_data) freemem(_data);
		_data = newdata;
		_capacity = assignment._capacity;
		_size = assignment._size;
		_data[_capacity] = 0;
	}
	return *this;
}

/**
@param cstr C character string to assign data from
@return Self-reference of buffer object.
@note <ul>
<li>Operator makes a copy of all characters stored in <i>cstr</i>.</li>
<li>Address of internal data buffer may be changed by this operator.</li>
</ul>
*/
BppCharBuffer& BppCharBuffer::operator=(const BppChar* cstr)
{
	if (cstr == _data)
		return *this;
	size_t len = cstr ? _tcslen(cstr) : 0;
	if (!len) 
	{
		 clear();
		return *this;
	}
	setCapacity(len+1);
	_tcscpy(_data, cstr);
	_size = len;
	return *this;
}

/**
@param index Index of character to access to. The index must be lower than current capacity.
@return writable reference to specified character
@note Operator directly accesses the size_ternal data buffer without any range checking.
*/
BppChar& BppCharBuffer::operator[](size_t index)
{
	return _data[index];
}

/**
@param index Index of character to access to. The index must be lower than current capacity.
@return value of specified character
@note Operator directly accesses the size_ternal data buffer without any range checking.
*/
BppChar BppCharBuffer::operator[](size_t index) const
{
	return _data[index];
}

/**
@return number of characters currently contained in buffer
*/
size_t BppCharBuffer::size() const BPPNOEXCEPT
{
	return _size;
}

/**
@return maximum number of characters buffer can hold without resizing
*/
size_t BppCharBuffer::capacity() const BPPNOEXCEPT
{
	return _capacity;
}

/**
Function changes capacity of buffer to specified number of characters. If new capacity is larger than
old one, additional allocated memory is initialized with NULL characters.
@param requested new capacity
@return really available new capacity after resizing
@note Function never discards any characters contained in buffer, i.e. if <i>capacity</i> is lower than
current size (result of size() method) new capacity will be equal to size.
*/
size_t BppCharBuffer::setCapacity(size_t capacity)
{
	if (capacity == 0) capacity = MINCAPACITY;
	if (capacity == _capacity) return _capacity;
	if (capacity < _size)
	{
		capacity = _size;
	}
	void* p = reallocmem(_data, (capacity+1) * sizeof(BppChar));
	_data = (BppChar*)p;
	if (capacity > _capacity)
	{
		memset(&_data[_capacity], 0, (capacity - _capacity) * sizeof(BppChar));
	}
	_capacity = capacity;
	_data[_capacity] = 0;
	return _capacity;
}


/**
Function adjusts capacity to current size, i.e. it releases unused memory.
@return new capacity
*/
size_t BppCharBuffer::shrink()
{
	return setCapacity(_size);
}

/**
Function erases maximum <i>cnt</i> characters from current buffer, starting
at position <i>pos</i>.
@param pos start position for erase
@param cnt maximum number of characters to delete
@return reference to current buffer
@note If <i>cnt</i> greater than number of characters behind <i>pos</i>,
all characters from <i>pos</i> to end of buffer will be erased.
*/
BppCharBuffer & BppCharBuffer::erase(size_t pos, size_t cnt)
{
	if ((_size == 0) || (pos >= _size)) return *this;
	if (cnt >(_size - pos))
		cnt = _size - pos;
	if (cnt > 0)
	{
		size_t moveCnt = _size - (pos + cnt);
		for (long i = 0; i<moveCnt; ++i)
		{
			_data[pos + i] = _data[pos + cnt + i];
		}
		while (cnt > 0)
		{
			_data[--_size] = 0;
			--cnt;
		}
	}
	return *this;
}

/**
Function inserts <i>cnt</i> characters of value <i>insertion</i> into current buffer.
@param pos insertion position
@param insertion character value to insert
@param cnt number of characters to insert
@return reference to current buffer
*/
BppCharBuffer & BppCharBuffer::insert(size_t pos, BppChar insertion, size_t cnt)
{
	if (cnt > 0)
	{
		if (pos > _size) pos = _size;
		size_t newsize = _size + cnt;
		if (_capacity < newsize) setCapacity(newsize);
		if (_capacity < newsize) errorExit(BP_ERR_NOMEM);
		BppChar* p2 = &_data[newsize];
		BppChar* p1 = &_data[_size];
		size_t ccnt = _size - pos;
		for (size_t i = 0; i<ccnt; ++i, --p1, --p2)
		{
			*p2 = *p1;
		}
		p1 = _data + pos;
		for (size_t j = 0; j<cnt; ++j, ++p1) *p1 = insertion;
		_size = newsize;
	}
	return *this;
}

/**
Operator creates a new character buffer containing concatenation of own value and value
of <i>app</i>. Capacity (and size) of result is the sum of sizes of both operands.
@param app other character buffer to concatenate with this one
@return new character buffer containing concatenation of operands
*/
BppCharBuffer BppCharBuffer::operator+(const BppCharBuffer& app)
{
	BppCharBuffer tmp(*this);
	return tmp += app;;
}

/**
@param pos1 position of first element (BppChar)
@param pos2 position of second element (BppChar)
@return reference to charbuffer
@note In case of invalid value of <i>pos1</i> or <i>pos2</i> method does nothing.
*/
BppCharBuffer& BppCharBuffer::swap(size_t pos1, size_t pos2) BPPNOEXCEPT
{
	if (pos1 != pos2 && pos1 < _size && pos2 < _size)
	{
		BppChar m = _data[pos1];
		_data[pos1] = _data[pos2];
		_data[pos2] = m;
	}
	return *this;
}


/**
@param srcpos position of element (BppChar) to move
@param dstpos destination position
@return reference to buffer
@note In case of invalid value of <i>srcpos</i> or <i>srcpos</i> method does nothing.
*/
BppCharBuffer& BppCharBuffer::move(size_t srcpos, size_t dstpos) BPPNOEXCEPT
{
	if (srcpos == dstpos || srcpos >= _size || dstpos >= _size) return *this;
	BppChar p = _data[srcpos];
	if (srcpos < dstpos)  // shift left
	{
		for (size_t i = srcpos; i<dstpos; ++i)
			_data[i] = _data[i + 1];
	}
	else // shift right
	{
		for (size_t i = srcpos; i>dstpos; --i)
			_data[i] = _data[i - 1];
	}
	_data[dstpos] = p;
	return *this;
}

/**
Operator appends contents of <i>app</i> to current character buffer.
@param app character buffer to append contents from
@return reference to this object
@note If current capacity is lower than sum of both buffer sizes, it is enlarged to this sum.
*/
BppCharBuffer& BppCharBuffer::operator+=(const BppCharBuffer& app)
{
	size_t newsize = _size + app._size;
	if (_capacity < newsize)
		setCapacity(newsize);
	memmove(&_data[_size], app._data, app._size * sizeof(BppChar));
	//memcpy(&_data[_size],app._data,app._size * sizeof(BppChar));
	//for (size_t i=0;i<app._size;i++)
	//   _data[_size+i] = app._data[i];
	_size += app._size;
	return *this;
}

/**
Function removes last character from buffer.
@return character removed.
@note This function does not affect capacity. Memory space of removed character is set to null.
*/
BppChar BppCharBuffer::pop_back()
{
  if (_size == 0)
	  return 0;
  _size--;
  BppChar tmp = _data[_size];
  _data[_size] = 0;
  return tmp;
}

/**
Function pushes (i.e. appends) a single character to end of buffer.
@param b character to push back
*/
void BppCharBuffer::push_back(BppChar b)
{
	if (_size >= _capacity)
		setCapacity(2*_size);
	_data[_size] = b;
	++_size;
}


/**
Function removes all characters from buffer, i.e. it sets current size to 0 and overwrites
all characters with null value.
*/
void BppCharBuffer::clear()
{
	if ((_data != 0) && (_size != 0))
		memset(_data,0,_size*sizeof(BppChar));
	_size = 0;
}

/**
@return pointer to data area
@note This function is mainly for internal use. Client code must not delete nor resize returned data space.<br>
Implementation of class ensures returned pointer to be usable like a C-style character string.
*/
BppChar* BppCharBuffer::data() const BPPNOEXCEPT
{
	return _data;
}

/**
Function sets internal size attribute to number of characters of C-string at start of internal buffer.
It is intended for internal use only.
*/
void BppCharBuffer::adjustSize()
{
	if (_data != 0)
	{
		//_data[_capacity-1] = 0;
	  _size = (size_t) _tcslen(_data);
	}
	else
		_size = 0;
}



/**
Function simply sets internal _size variable without changing any data nor capacity.
@param size new size of charbuffer
@return true on sucess, false if <i>size</i> is greater than current capacity
*/
bool BppCharBuffer::setSize(size_t size)
{
	if (size > _capacity) return false;
	_size = size;
	return true;
}

/**
Function resizes character buffer to <i>size</i> characters.
If <i>size</i> is less than current size, characters are erased from end of buffer.
If <i>size</i> is greater than current size, new characters of value <i>fill</i> are appended.
@param size New size (number of characters) of buffer.
@param fill Value to use for new elements.
@return true on success, false in case of memory error.
*/
bool BppCharBuffer::resize(size_t size, BppChar fillchar)
{
	if (size > _capacity) {
		if (setCapacity(size) < size) return false;
	}
	while (_size < size) {
		_data[_size] = fillchar;
		++_size;
	}
	return setSize(size);
}

/**
@param from Index position of start of subsequence.
@param cnt Maximum number of elements to copy into result.
@return Returns new <tt>charbuffer</tt> containing copy of specified part from current one.
*/
BppCharBuffer BppCharBuffer::slice(size_t from, size_t cnt) const
{
	size_t maxcnt = (from <= _size) ? _size - from : 0;
	if (cnt > maxcnt) cnt = maxcnt;
	if (cnt == 0) return BppCharBuffer();
	BppCharBuffer result(cnt, cnt);
	for (size_t i = 0; i<cnt; ++i)
		result[i] = _data[from + i];
	return result;
}


int BppCharBuffer::compare(const BppCharBuffer& comparison) const BPPNOEXCEPT
{
	if (&comparison == this) return 0;
	if (_size == 0) return comparison._size ? -1 : 0;
	if (comparison._size == 0) return 1;
	size_t s = _size > comparison._size ? comparison._size : _size;
	//s *= sizeof(BppChar);
	int r = 
#ifdef _UNICODE
		wmemcmp(_data,comparison._data,s);
#else
		memcmp(_data, comparison._data, s);
#endif
	if (r != 0) return r;
	return _size == comparison._size ? 0 : s < _size ? 1 : -1; 
}

bool BppCharBuffer::operator==(const BppCharBuffer& cmp) const BPPNOEXCEPT
{
	if (&cmp == this) return true;
	if (_size != cmp._size) return false;
	if (_size == 0) return true;
#ifdef _UNICODE
	return wmemcmp(_data, cmp._data, _size) == 0;
#else
	return memcmp(_data, cmp._data, _size) == 0;
#endif
//	return memcmp(_data,cmp._data,_size*sizeof(BppChar)) == 0;
}

bool BppCharBuffer::operator!=(const BppCharBuffer& cmp) const BPPNOEXCEPT
{
	return !(this->operator==(cmp));
}

size_t BppCharBuffer::find(BppChar val, size_t pos) const
{
	if (pos >= _size) return BPP_NNPOS;
	for (BppChar* p = _data + pos; pos < _size; ++p, ++pos) {
		if (*p == val) return pos;
	}
	return BPP_NNPOS;
}

size_t BppCharBuffer::rfind(BppChar val, size_t pos) const
{
	if (_size == 0) return BPP_NNPOS;
	if (pos >= _size) pos = _size - 1;
	for (BppChar* p = _data + pos; p >= _data; --p, --pos) {
		if (*p == val) return pos;
		if (p == _data) break;
	}
	return BPP_NNPOS;
}

#ifndef DOXY
BppDWord BytePtr2DW(BppByte* p)
{
#ifndef WINCE
	return *(BppDWord*) p;
#else
	BppDWord tmp;
	memcpy(&tmp,p,sizeof(tmp));
	return tmp;
#endif
}

BppWord BytePtr2W(BppByte* p)
{
#ifndef WINCE
	return *(BppWord*) p;
#else
	BppWord tmp;
	memcpy(&tmp,p,sizeof(tmp));
	return tmp;
#endif
}

BppNWord BytePtr2NW(BppByte* p)
{
#ifndef WINCE
	return *(BppNWord*) p;
#else
	BppNWord tmp;
	memcpy(&tmp,p,sizeof(tmp));
	return tmp;
#endif
}

#endif // !DOXY

