#include "bppclass.h"
#include "bppstrng.h"
#include "bppdict.h"
#include "bppvm.h"
#include "bppvec.h"
#include "bppobjct.h"
/**
Constructor creates an empty unnamed class. At least the class name
must be specified by a call to one of the #setClassName methods.
@param vm reference to virtual machine to become owner of class
*/
BppClass::BppClass(BppVirtualMachine& vm)
:_vm(vm),_className(0),_baseClass(0),_ownerClass(0),_size(-1),_isNative(false),_notifyProc(0)
{
	_dataMembers = new BppDictionary();
	_funcMembers = new BppDictionary(true);
	_dataMembers->_ownerClass = this;
	_funcMembers->_ownerClass = this;
	_dataMembers->addRef();
	_funcMembers->addRef();

}

/**
Constructor creates a named class. Optionally a base class an owner class may be specified.
@param vm reference to virtual machine to become owner of class
@param name pointer to an existing string object containing the (local) class name
@param baseClass pointer to an existing class object to become base class of this class
@param ownerClass pointer to an existing class object to become owner (outer) class of this class
@note There is no checking made for validity of specified class name nor base class, i.e.
client code is responsible to take care about it.<br>
The constructor increments reference counter of <i>name</i> and <i>baseClass</i> arguments
if not NULL.
*/
BppClass::BppClass(BppVirtualMachine& vm,BppString* name, BppClass* baseClass, BppClass* ownerClass)
:_vm(vm),_className(name),_baseClass(baseClass),_ownerClass(ownerClass),_size(-1),_isNative(false), _notifyProc(0)
{
	_dataMembers = new BppDictionary();
	_funcMembers = new BppDictionary(true);
	_dataMembers->_ownerClass = this;
	_funcMembers->_ownerClass = this;
	_dataMembers->addRef();
	_funcMembers->addRef();
	if (_className != 0) _className->addRef();
	if (_baseClass != 0) _baseClass->addRef();
	if (_ownerClass != 0) _ownerClass->addRef();
}

/**
Constructor creates a named class. Optionally a base class an owner class may be specified.
@param vm reference to virtual machine to become owner of class
@param name string object containing the (local) class name
@param baseClass pointer to an existing class object to become base class of this class
@param ownerClass pointer to an existing class object to become owner (outer) class of this class
@note There is no checking made for validity of specified class name nor base class, i.e.
client code is responsible to take care about it.<br>
The constructor increments reference counter of <i>baseClass</i> (if not NULL) and
copies <i>name</i>.
*/
BppClass::BppClass(BppVirtualMachine& vm,const BppString& name, BppClass* baseClass, BppClass* ownerClass)
:_vm(vm),_className(0),_baseClass(baseClass),_ownerClass(ownerClass),_size(-1),_isNative(false), _notifyProc(0)
{
	if (name.length() > 0)
		_className = new BppString(name);
	_dataMembers = new BppDictionary();
	_funcMembers = new BppDictionary(true);
	_dataMembers->_ownerClass = this;
	_funcMembers->_ownerClass = this;
	_dataMembers->addRef();
	_funcMembers->addRef();
	if (_className != 0) _className->addRef();
	if (_baseClass != 0) _baseClass->addRef();
	if (_ownerClass != 0) _ownerClass->addRef();
}

BppClass::~BppClass()
{
	if (_className != 0) _className->release();
	if (_baseClass != 0) _baseClass->release();
	if (_ownerClass != 0) _ownerClass->release();
	_dataMembers->_ownerClass = 0;
	_funcMembers->_ownerClass = 0;
	_dataMembers->release();
	_funcMembers->release();
}

/**
@return reference to data member dictionary
*/
BppDictionary& BppClass::dataMembers() BPPNOEXCEPT
{
	return *_dataMembers;
}

/**
@return reference to function member dictionary
*/
BppDictionary& BppClass::funcMembers() BPPNOEXCEPT
{
	return *_funcMembers;
}

/**
@return pointer to base class object (or NULL if none)
*/
BppClass* BppClass::getBaseClass() const BPPNOEXCEPT
{
	return _baseClass;
}

/**
@param baseClass pointer to base class
@return true on success, i.e. if base class not already set
@note Function increments reference count of base class
*/
bool BppClass::setBaseClass(BppClass* baseClass) BPPNOEXCEPT
{
	if (_baseClass != 0) return false;
	if (baseClass != 0)
	{
		baseClass->addRef();
		_baseClass = baseClass;
	}
	return true;
}

/**
@return pointer to owner class object (or NULL if none)
*/
BppClass* BppClass::getOwnerClass() const BPPNOEXCEPT
{
	return _ownerClass;
}

/**
@param ownerClass pointer to base class
@return true on success, i.e. if owner class not already set
@note Function increments reference count of owner class
*/
bool BppClass::setOwnerClass(BppClass* ownerClass) BPPNOEXCEPT
{
	if (_ownerClass != 0) return false;
	if (ownerClass != 0)
	{
		ownerClass->addRef();
		_ownerClass = ownerClass;
	}
	return true;
}

/**
@func Function checks if current class is same as <i>cls</i> or
derived from it.
@param cls class to check inheritance for
@return true if current class is subtype of <i>cls</i>
@see isSubTypeOf(const BppString&)
*/
bool BppClass::isSubtypeOf(BppClass* cls) const BPPNOEXCEPT
{
	if (cls == this) return true;
	if (_baseClass == 0) return false;
	return _baseClass->isSubtypeOf(cls);
}

/**
@func Function does a name based inheritance check.
@param className name of class to check inheritance for
@return true if current class is subtype of class named <i>className</i>
@see isSubTypeOf(BppClass*)
*/
bool BppClass::isSubtypeOf(const BppString& className) const BPPNOEXCEPT
{
	if (_className == 0) return false;
	if (_className->compare(className) == 0) return true;
	if (_baseClass == 0) return false;
	return _baseClass->isSubtypeOf(className);
}

/**
* return Fully qualified name of class.
*/
BppString BppClass::getClassPath() const
{
	if (_ownerClass) {
		return BppString::format(_T("%s::%s"), _ownerClass->getClassPath().c_str(), (_className) ?_className->c_str() : _T(""));
	}
	if (_className) return *_className;
	return _T("");
}

/**
@func Function sets (local) class name from pointer to existing string.
@param className pointer to string object containing class name
@return true on success, otherwise, i.e. if class name already set or
<i>className</i> argument was NULL, false
@note Reference counter of <i>className</i> is incremented.
@see setClassName(const BppString&)
*/
bool BppClass::setClassName(BppString* className) BPPNOEXCEPT
{
	if (_className != 0) return false;
	if (className == 0) return false;
	_className = className;
	_className->addRef();
	return true;
}

/**
@func Function sets (local) class name from string.
@param className class name
@return true on success, otherwise, i.e. if class name already set or
<i>className</i> argument was empty, false
@note Function makes a copy of <i>className</i> argument value.
@see setClassName(BppString*)
*/
bool BppClass::setClassName(const BppString& className)
{
	if (_className != 0) return false;
	if (className.length() == 0) return false;
	_className = new BppString(className);
	_className->addRef();
	return true;
}

/**
@func Function counts number of non static data members fo class (including all base classes),
sets internal _size variable and returns its value.
@return Number of non static data members (including base class members)
@see getDataSize
*/
int BppClass::computeDataSize() BPPNOEXCEPT
{
	_size = 0;
	for (BppDictionaryEntry* p=_dataMembers->getFirst();p!=0;p=p->getNext())
	{
		if (p->getSymbolType() == ST_DATA) 
			_size++;
	}
	if (_baseClass != 0)
		_size += _baseClass->computeDataSize();
	return _size;
}

/**
@return Vector containing default values for all non-static data members (including flags, if any). Result
also contains items of base classes. First item (index 0) in result is the class itself. Indexes
of data items correspond to definition order.<br>
To get name of a data member use <tt>findDataMemberFromIndex</tt> with indec in result vector as argument.
@note Vector returned by this function is built only when needed, i.e. on first call to getDefaultDataMembers.
@see findDataMemberFromIndex, buildDefaultValues
*/
const BppVector& BppClass::getDefaultDataMembers()
{
	if (_defaultValues.size() == 0)
		  buildDefaultValues();
	return _defaultValues;
}

/**
@return Vector of non-static data member names. Order of names corresponds to order of
values in vector returned by BppObject::getMemberData() method. First entry (at index 0)
always is "this".
@note Function is intended for debugging purposes. It allows retrieving of member names
and name based access to member variables of an object.<br>
Entries in result are references to keys in _dataMembers dictionary of class. Therefore
client code must not delete nor modify them.
*/
BppVector BppClass::getDataMemberNames() const
{
	static BppValue thisval = new BppString(_T("this"));
	BppVector result(_size + 1, true);
	result[0] = thisval;
	fillMemberDataNames(result);
	return result;
}

void BppClass::buildDefaultValues(BppVector* defValues)
{
	if (!defValues)
	{
		if (_size < 0) _size = 0;
	  _defaultValues = BppVector(_size+1,true);
	  _defaultValues[0] = this;
	  defValues = &_defaultValues;
	}
	for (BppDictionaryEntry* p=_dataMembers->getFirst();p!=0;p=p->getNext())
	{
		if (p->getSymbolType() == ST_DATA)
	  {
		  BppVector* v = p->getValue();
		  if (v)
		  {
			  unsigned i = (unsigned) (BppInt) (*v)[0];
			  if (i>0 && i<=_size) 
			  {
				  (*defValues)[i] = (*v)[1];
				  (*defValues)[i].setFlags((*v)[1].getFlags());
			  }
		  }
	  }
	}
	if (_baseClass != 0)
	  _baseClass->buildDefaultValues(defValues);

}

void BppClass::fillMemberDataNames(BppVector & names) const
{
	for (BppDictionaryEntry* p = _dataMembers->getFirst(); p != 0; p = p->getNext())
	{
		if (p->getSymbolType() == ST_DATA)
		{
			BppVector* v = p->getValue();
			if (v)
			{
				unsigned i = (unsigned)(BppInt)(*v)[0];
				if (i>0 && i <= _size)
				{
					names[i] = p->getKey();
				}
			}
		}
	}
	if (_baseClass != 0)
		_baseClass->fillMemberDataNames(names);
}

/**
@return number of non static data members
@note Other than computeDataSize this function simply returns current value of
internal _size variable.
@see computeDataSize
*/
int BppClass::getDataSize() const BPPNOEXCEPT
{
	return _size;
}

/**
@func Function searches for class member <i>name</i>. Optionally base classes are included
in search.
@param contextClass Specifies context class for member search. If NULL or a class not derived
fron current class, only public members are acessible. For a derived class, protected members are found too,
in case of current class itself, all members can be accessed.
@param name Name of member to search for
@param includeBase If true, base classes are included in search.
@param requirePublic If true, public methods are found. This flag is mainly intended for internal use
and set by findMember method itself while recursive search.
@return dictionary entry containing member description if found, otherwise NULL
@note Function first searches for data member, than for function member. If found
nothing but <i>includeBase</i> is <tt>true</tt> search is done in base class (if any)
recursively.
*/
BppDictionaryEntry* BppClass::findMember(const BppClass* contextClass, const BppString& name, bool includeBase, bool requirePublic) const BPPNOEXCEPT
{
	BppDictionaryEntry* result = _dataMembers->find(name);
	if (!result) result = _funcMembers->find(name);
	unsigned short rprotLevel = BVF_PUBLIC;
	if (result != 0)
	{
		unsigned short protlevel = result->getValue().getProtection();
		if (!protlevel)
		{
			protlevel = result->getSymbolType() == ST_DATA ? BVF_PROTECTED : BVF_PUBLIC;
		}
		if ((!requirePublic) && contextClass && protlevel < rprotLevel)
		{
			if (contextClass == this) rprotLevel = BVF_PRIVATE;
			else if (contextClass->isSubtypeOf((BppClass*)this)
				|| this->isSubtypeOf((BppClass*)contextClass)) rprotLevel = BVF_PROTECTED;
		}
		if (protlevel < rprotLevel) result = 0;
	}
	if (!result && includeBase && (_baseClass != 0)) {
		if (contextClass && contextClass != this) {
		if (!contextClass->isSubtypeOf((BppClass*)this)
			&& !this->isSubtypeOf((BppClass*)contextClass)) requirePublic = true;
		}
		result = _baseClass->findMember(contextClass,name,true,requirePublic);
	}
	if (!result && (_ownerClass != 0)) {
		//BppDictionaryEntry* myEntry = _vm.classes().find(getClassPath());
		//if (!myEntry) return 0;
		//unsigned short myProtLevel = myEntry->getValue().getProtection();
		result = _ownerClass->findMember((contextClass && contextClass == this) ? _ownerClass : contextClass, name, includeBase, requirePublic);
		if (result) {
			BppSymbolType st = result->getSymbolType();
			if (st != ST_SDATA && st != ST_SFUNCTION) return 0;
		}
	}
	return result;
}


/**
@func Function searches for class member <i>name</i> using explicitely specified protection level. Optionally base classes are included
in search.
@param rprotLevel Specifies protection level class for member search. If BVV_PUBLIC, only public members are acessible. 
For BFV_PROTECTED protected members are found too,
in case BFV_PRIVATE, all members can be accessed.
@param name Name of member to search for
@param includeBase If true, base classes are included in search.
@return dictionary entry containing member description if found, otherwise NULL
@note Function first searches for data member, than for function member. If found
nothing but <i>includeBase</i> is <tt>true</tt> search is done in base class (if any)
recursively.
*/
BppDictionaryEntry* BppClass::findMemberPL(unsigned short rprotLevel, const BppString& name, bool includeBase) const BPPNOEXCEPT
{
	BppDictionaryEntry* result = _dataMembers->find(name);
	//if (result != 0)
	//	return result;
	if (!result) result = _funcMembers->find(name);
	if (result != 0)
	{
		unsigned short protlevel = result->getValue().getProtection();
		if (!protlevel)
		{
			protlevel = result->getSymbolType() == ST_DATA ? BVF_PROTECTED : BVF_PUBLIC;
		}
		if (protlevel < rprotLevel) return 0;
		return result;
	}
	if (includeBase && (_baseClass != 0))
		result = _baseClass->findMemberPL(rprotLevel,name, true);
	if (!result && _ownerClass != 0) {
		result = _ownerClass->findMemberPL(rprotLevel, name, includeBase);
		if (result) {
			BppSymbolType st = result->getSymbolType();
			if (st != ST_SDATA && st != ST_SFUNCTION) return 0;
		}
	}
	return result;
}

/**
@func Function searches for data member at index <i>idx</i>. Optionally base classes are included
in search.<br>
It iterates through data member dictionary. If entry having value same as <i>idx</i>
was found, key of that entry is returned.<br>
If found nothing but <i>includeBase</i> is <tt>true</tt> search is done in base class (if any)
recursively.
@return name (identifier) of data member if found, otherwise NULL
@note This function is used by BppDebugger class to retrieve member names while tracing.
*/
const BppChar* BppClass::findDataMemberFromIndex(unsigned idx, bool includeBase) const
{
	BppDictionaryEntry* p = 0;
	for (p = _dataMembers->getFirst(); p != 0; p=p->getNext())
	{
		BppVector* v = p->getValue();
		if (v && idx == ((unsigned) (BppInt) (*v)[0]))
			return p->getKey()->c_str();
	}
	if (includeBase && (_baseClass != 0))
		return _baseClass->findDataMemberFromIndex(idx,true);
	return 0;
}

/**
@func Function is called from destructor of BppObject instances, i.e. immediately
before an object is destroyed. It invokes the notifyObjDelete method of virtual
machine in order to execute destructor code of B++ object.
Furthermore, if notification callback functiom installed, callback is called with
reason BPP_OBJ_DELETE.
@see BppVirtualMachine::notifyObjDelete, notifyObjSerialization, setNotifyProc, BppObjNotifyProc, BppObjectNotifyType
*/
void BppClass::notifyObjDelete(BppObject* obj)
{
	_vm.notifyObjDelete(obj);
	if (_notifyProc) _notifyProc(&_vm, obj, &(obj->_nativeObject),BPP_OBJ_DELETE);
}

/**
@func Function is called from public constructors of BppObject instances, immediately
after initialization of member data vector. If notification callback functiom installed, 
callback is called with reason BPP_OBJ_CREATE or BPP_OBJ_COPY, depending on <i>copied</i> flag.
@see notifyObjDelete, notifyObjSerialization, setNotifyProc, BppObjNotifyProc, BppObjectNotifyType
*/
void BppClass::notifyObjCreation(BppObject * obj, bool copied)
{
	if (_notifyProc) _notifyProc(&_vm, obj, &(obj->_nativeObject), copied ? BPP_OBJ_COPY : BPP_OBJ_CREATE);
}

/**
@func Function is called immediately before serialization and after deserialization. 
If notification callback functiom installed,
callback is called with reason BPP_OBJ_BEFORESERIALIZE or BPP_OBJ_AFTERDESERIALIZE, depending on <i>serialize</i> flag.<br>
This allows copying data of a wrapped native object from/to data members to support
serialization of native objects too.
@see notifyObjDelete, notifyObjCreate, setNotifyProc, BppObjNotifyProc, BppObjectNotifyType
*/
void BppClass::notifyObjSerialization(BppObject * obj, bool serialize)
{
	if (_notifyProc) _notifyProc(&_vm, obj, &(obj->_nativeObject), serialize ? BPP_OBJ_BEFORESERIALIZE : BPP_OBJ_AFTERDESERIALIZE);
}

/**
@func Function adds a data member to class.
@param name name of member
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty or
there is already a member of same name, false
@see addStaticDataMember(const BppChar*)
*/
bool BppClass::addDataMember(const BppChar* name)
{
	return addDataMember(name,BppValue());
}

/**
@func Function adds an initialized non-static data member to class.
@param name name of member
@param initValue initial value of member
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty or
there is already a member of same name, false
@note Flags of <i>initValue</i> are assigned to member.
@see addStaticDataMember(const BppChar*)
*/
bool BppClass::addDataMember(const BppChar* name, const BppValue& initValue)
{
	BppString* pname = new BppString(name);
	pname->addRef();
	bool ok =  ((pname->length() > 0) && (!findMemberPL(BVF_PRIVATE,*pname,false)));
	if (ok)
	{
		BppDictionaryEntry* pe = (*_dataMembers)[pname];
	  BppVector* v = new BppVector(2,true);
	  (*v)[1] = initValue;
	  BppValue vv = v;
	  vv.setFlags(initValue.getFlags());
	  pe->setValue(vv,false,true);
	  ((*v)[1]).setFlags(initValue.getFlags());
		pe->setSymbolType(ST_DATA);
	}
	pname->release();
	return ok;
}

/**
@func Function adds an initialized non-static data member to class or redefines an existing one.
@param name name of member
@param initValue initial value of member
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty or
there is already a member of same name but other than non-static data member, false
@note Flags of <i>initValue</i> are assigned to member.
@see defineStaticDataMember
*/
bool BppClass::defineDataMember(const BppChar* name, const BppValue& initValue)
{
	BppString* pname = new BppString(name);
	BppValue nv = pname;
	if (pname->length() == 0) return false;
	BppDictionaryEntry* pe = findMemberPL(BVF_PRIVATE,*pname,false);
	if (!pe)
	{
		pe = (*_dataMembers)[pname];
		 pe->setSymbolType(ST_DATA);
	}
	else if (pe->getSymbolType() != ST_DATA) return false;
	
	BppVector* v = new BppVector(2,true);
	(*v)[1] = initValue;
	BppValue vv = v;
	vv.setFlags(initValue.getFlags());
	pe->setValue(vv, false, true);
	 ((*v)[1]).setFlags(initValue.getFlags());

	return true;
}

/**
@func Function adds a static data member to class.
@param name name of member
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty or
there is already a member of same name, false
@see addDataMember(const BppChar*)
*/
bool BppClass::addStaticDataMember(const BppChar* name)
{
	return addStaticDataMember(name,BppValue());
}

/**
@func Function adds an initialized static data member to class.
@param name name of member
@param initValue initial value of member
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty or
there is already a member of same name, false
@note Flags of <i>initValue</i> are assigned to member.
@see addStaticDataMember(const BppChar*), addDataMember(const BppChar*,const BppValue&)
*/
bool BppClass::addStaticDataMember(const BppChar* name, const BppValue& initValue)
{
	BppString* pname = new BppString(name);
	pname->addRef();
	bool ok =  ((pname->length() > 0) && (!findMemberPL(BVF_PRIVATE,*pname,false)));
	if (ok)
	{
		BppDictionaryEntry* pe = (*_dataMembers)[pname];
		pe->setSymbolType(ST_SDATA);
		pe->setValue(initValue,false,true);
	}
	pname->release();
	return ok;
}

/**
@func Function adds an initialized static data member to class or redefines an existing one.
@param name name of member
@param initValue initial value of member
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty or
there is already a member of same name but other than static data member, false
@note Flags of <i>initValue</i> are assigned to member.
@see addStaticDataMember(const BppChar*)
*/
bool BppClass::defineStaticDataMember(const BppChar* name, const BppValue& initValue)
{
	BppString* pname = new BppString(name);
	BppValue nv = pname;
	if (pname->length() == 0) return false;
	BppDictionaryEntry* pe = findMemberPL(BVF_PRIVATE,*pname,false);
	if (!pe)
	{
		pe = (*_dataMembers)[pname];
		 pe->setSymbolType(ST_SDATA);
	}
	else if (pe->getSymbolType() != ST_SDATA) return false;
	
	pe->setValue(initValue,false,true);
 
	return true;
}

/**
@func This function is called by compiler at end of compile a class definition.
It enumerates non-static data members, builds data member index and computes
data size for class instances.
@see computeDataSize
*/
void BppClass::fixupDataDefs()
{
	int n = (_baseClass != 0) ? _baseClass->computeDataSize() : 0;
	for (BppDictionaryEntry* pe = _dataMembers->getFirst(); pe != 0; pe = pe->getNext())
	{
		if (pe->getSymbolType() == ST_DATA)
		{
			n++;
		 BppVector* v = pe->getValue();
		 if (v) (*v)[0] = (BppInt)n;
			//pe->setValue((long)n);
		}
	}
	computeDataSize();
}

/**
@func Function adds a native non-static member function to class.
@param name name of member function
@param pfunc address of function
@param flags Flags for function value (use to specify protection level)
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty,
there is already a member of same name or <i>pfunc</i> is NULL, false
@see addStaticMethod, BppValueFlags
*/
bool BppClass::addMethod(const BppChar* name, BPPUSERFUNC pfunc, unsigned char flags)
{
	return _addMethod(false, name, pfunc, flags);
}

/**
@func Function adds a native static member function to class.
@param name name of member function
@param pfunc address of function
@param flags Flags for function value (use to specify protection level)
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty,
there is already a member of same name or <i>pfunc</i> is NULL, false
@see addMethod, BppValueFlags
*/
bool BppClass::addStaticMethod(const BppChar* name, BPPUSERFUNC pfunc, unsigned char flags)
{
	return _addMethod(true,name, pfunc, flags);
}

/**
@param isStatic Flag specifies if <i>pfunc</i> is class or element function.
@param name name of member function
@param pfunc address of function
@param flags Flags for function value (use to specify protection level)
@return true on success, otherwise, i.e. if <i>name</i> was NULL or empty,
there is already a member of same name or <i>pfunc</i> is NULL, false
@see addMethod, BppValueFlags
*/
bool BppClass::_addMethod(bool isStatic, const BppChar * name, BPPUSERFUNC pfunc, unsigned char flags)
{
	if (pfunc == 0) return false;
	BppString* pname = new BppString(name);
	pname->addRef();
	bool ok = ((pname->length() > 0) && (!findMemberPL(BVF_PRIVATE, *pname, false)));
	if (ok)
	{
		BppDictionaryEntry* pe = (*_funcMembers)[pname];
		pe->setSymbolType(isStatic ? ST_SFUNCTION : ST_FUNCTION);
		BppValue v = pfunc;
		v.setFlags(flags);
		pe->setValue(v, false, true);
	}
	pname->release();
	return ok;
}

/**
@func Function gets index of specified non-static data member
in member table of class.
@param name name of member
@return member index if found, otherwise -1
*/
int BppClass::getMemberIndex(const BppString& name) const
{
	BppDictionaryEntry* pe = findMember((BppClass*)this,name);
	if ((pe != 0) && (pe->getSymbolType() == ST_DATA))
	{
		BppVector* v = pe->getValue();
		if (v)
			  return (int) (BppInt) (*v)[0];
	}
	return -1;
}

/**
@func Function allows a class to be marked as native.
@param isNative new value of _isNative flag
@note Classes marked as native are assumed always to be loaded.
They are not included in modules of compiled bytecode. 
@see BppDictionaryIO::writeDictEntry, getIsNative
*/
void BppClass::setIsNative(bool isNative)
{
	_isNative = isNative;
}

/**
@return value of internal _isNative flag
@see setIsNative
*/
bool BppClass::getIsNative() const
{
	return _isNative;
}

BppVirtualMachine& BppClass::getVM() const
{
	return _vm;
}

