#include "bppcomp.h"
#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include "bppvec.h"
#include "bppclass.h"
#include "bppscan.h"
#include "bppvalue.h"
#include "bppbuf.h"
#include "bppufunc.h"
#include "bppprint.h"
#include "dictio.h"
#include "bppvm.h"
#include "bppdbg.h"
#include "bppint.h"
#include "bppconst.h"
#include "bppmsg.h"

/// maximum number of errors before abort compiling
#define MAXERRS 10

// literal ::= "null" | "nil" | <string> | <intnumber> | <floatnumber> | <arrayliteral> | <dictionaryliteral> | 
//              <funcliteral> | <literalexpr> | <litid> .
// arrayliteral ::= "[" [ <literal> {"," <literal> } ] "]" .
// dictionaryliteral ::= "{" [ <keyident> ":" <literal> {"," <keyident> ":" <literal> } ] "}" .
// keyident ::= <identifier> | <string> .
// funcliteral ::= "function" <functionbody> .
// literalexpr ::= "#" "(" <litexpr> ")" .
// litid ::= <identifier> . // previously defined named literal value
bool BppCompiler::_doLiteralValue(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	int savedErrCnt = errCnt;
	switch (token)
	{
	case T_NIL:
		entry.setValue(BppValue());
		break;
	case T_NUMBER:
	case T_FNUMBER:
		entry.setValue(_scanner->getNumberValue());
		break;
	case T_STRING:
		entry.setValue(new BppString(_scanner->getTokenString()));
		break;
	case T_FUNCTION:
	{
		BppClass* pclass = 0;
		token = _scanner->getToken(errCnt);
		if ((token == T_ANY) && (_scanner->getChar() == _T('[')))
		{
			token = _scanner->getToken(errCnt);
			if ((token != T_STRING) && (!_require(T__DFUNCTION, T_IDENTIFIER, token)))
			{
				errCnt++;
				break;
			}
			pclass = _findClass(_scanner->getTokenString(), errCnt);
			if (!_fetchRequire(T__DFUNCTION,errCnt, T_ANY, _T(']')))
			{
				errCnt++;
				break;
			}

		}
		else _scanner->back();
		BppValueType rtype = BVT_NULL;
		BppClass* prClass = 0;
		token = _scanner->getToken(errCnt);
		bool constResult = token == T_CONST;
		if (constResult) token = _scanner->getToken(errCnt);
		switch (token)
		{
		case T_VAR:
		case T_NIL:
			break;
		case T_VOID:
		case T_FUNCTION:
		case T_IDENTIFIER:
			rtype = findVTypeFromName(T__DFUNCTION,_scanner->getTokenString(), errCnt, true, &prClass);
			if (rtype == BVT_NULL)
			{
				_error(BppString::format(CMP_INVALID_TYPENAME_FMT, _scanner->getTokenString()).c_str(), &errCnt);
			}
			break;
		default:
			_scanner->back();
		}

		if (!_fetchRequire(T__DFUNCTION,errCnt, T_ANY, _T('(')))
		{
			errCnt++;
		}
		else
		{
			BppVector* codevec = _codeFunctionBody(_T("<anonymous>"), errCnt, constResult, pclass, rtype, prClass);
			if (codevec != 0)
			{
				entry.setValue(BppValue(codevec, true));
			}
		}
	}
	break;
	case T_LITEXPR:
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('(')))
		{
			errCnt++;
			break;
		}
		token = _scanner->getToken(errCnt);
		if (!_doLitExpr(token, entry, errCnt)) return false;
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
		break;
	case T_SHL: // buffer literal
	{
		BppDictionaryEntry dummyEntry(0, BppValue());
		token = _scanner->getToken(errCnt);
		if (!_doLitExpr(token, dummyEntry, errCnt)) return false;
		BppString* pfname = dummyEntry.getValue();
		if (!pfname) {
			_error(CMP_INCLUDE_FILENAME_EXPECTED, &errCnt);
			break;
		}
		BppBuffer* pbuf = new BppBuffer();
		BppValue v(pbuf);
		if (_doReadBinFile(*pfname, *pbuf, errCnt)) entry.setValue(v);
	}
	break;
	case T_ANY:
	{
		if (_scanner->getChar() == _T('[')) // array literal
		{
			BppVector* pvec = new BppVector();
			entry.setValue(pvec);
			BppDictionaryEntry dummyEntry(0, BppValue());
			for (token = _scanner->getToken(errCnt);
				!((token == T_ANY) && (_scanner->getChar() == _T(']')));
				token = _scanner->getToken(errCnt))
			{
				// read single literal value
				_doLitExpr(token, dummyEntry, errCnt);
				pvec->push_back(dummyEntry.getValue());
				if (errCnt > savedErrCnt)
					break;
				// value must be followed by , or ]
				token = _scanner->getToken(errCnt);
				if (token != T_ANY) // not a single character
				{
					_error(CMP_COMMA_SQBRACKET_EXPECTED, &errCnt);
					break;
				}
				if (_scanner->getChar() == _T(']')) // return if end is reached
					break;
				if (!_require(T__DFUNCTION, T_ANY, token, _T(','))) // return if comma missing
				{
					errCnt++;
					break;
				}
			} // end for
			break;
		} // end array literal
		else if (_scanner->getChar() == _T('{')) // dictionary literal
		{
			token = _scanner->getToken(errCnt);
			bool sorted = false;
			if (token == T_ANY && _scanner->getChar() == _T('!')) sorted = true;
			else _scanner->back();
			BppDictionary* pdict = new BppDictionary(sorted);
			entry.setValue(pdict);
			for (token = _scanner->getToken(errCnt);
				!((token == T_ANY) && (_scanner->getChar() == _T('}')));
				token = _scanner->getToken(errCnt))
			{
				if (!((token == T_IDENTIFIER) || _require(T__DFUNCTION, T_STRING, token)))
				{
					errCnt++;
					break;
				}
				BppDictionaryEntry& entryref = (*pdict)[_scanner->getTokenString()];
				if (!_fetchRequire(T__DFUNCTION,errCnt, T_ANY, _T(':')))
				{
					errCnt++;
					break;
				}
				_doLitExpr(_scanner->getToken(errCnt), entryref, errCnt);
				if (errCnt > savedErrCnt)
					break;
				// value must be followed by , or }
				token = _scanner->getToken(errCnt);
				if (token != T_ANY) // not a single character
				{
					_error(CMP_COMMA_CURBRACKET_EXPECTED, &errCnt);
					break;
				}
				if (_scanner->getChar() == _T('}')) // return if end is reached
					break;
				if (!_require(T__DFUNCTION, T_ANY, token, _T(','))) // return if comma missing
				{
					errCnt++;
					break;
				}
			} // end for
			break;
		} // end dictionary literal
		_require(T__DFUNCTION, T_IDENTIFIER, token);
		errCnt++;
		break;
	} // end T_ANY
	case T_ERROR:
		_print(_scanner->getTokenString());
		errCnt++;
		return false;
	case T_IDENTIFIER:
	{
		const BppChar* str = _scanner->getTokenString();
		if (_tcscmp(str, SYMN_defined) == 0)
		{
			FlagHlp hlp(_suppressMacroExpansion, true);
			token = _scanner->getToken(errCnt);
			bool bracket = false;
			if ((token == T_ANY) && (_scanner->getChar() == _T('('))) {
				bracket = true;
				token = _scanner->getToken(errCnt);
			}
			if (!_require(T__DFUNCTION, T_IDENTIFIER, token)) errCnt++;
			BppString nm = _scanner->getTokenString();
			BppDictionaryEntry* lit = _namedLiterals.find(nm);
			if (!lit) lit = _definedMacros.find(nm);
			entry.setValue(BppValue::makeBool(lit ? true : false));
			if (bracket && !_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
			break;
		}
		BppDictionaryEntry* lit = _namedLiterals.find(str);
		if (lit != 0)
		{
			entry.setValue(lit->getValue());
			break;
		}
	}
	default:
		_error(CMP_LITERAL_EXPECTED, &errCnt);
	} // end switch
	return true;
}

bool BppCompiler::_doReadBinFile(const BppString& fname, BppBuffer& buffer, int& errCnt)
{
	BppString path = fname;
	if (path.length() == 0)
	{
		_error(CMP_INCLUDE_FILENAME_EXPECTED, &errCnt);
		return false;
	}
	BppString fullpath = path;

	FILE* fp = _tfopen(path.c_str(), _T("rb"));
	if (fp == 0)
	{
		size_t n = _srcDirStack.size();
		for (size_t i = 0; (!fp) && i < n; ++i)
		{
			BppString* s = _srcDirStack[i];
			const BppChar* fpath = searchDirList(s->c_str(), path.c_str());
			if (fpath)
			{
				fullpath = fpath;
				fp = _tfopen(fullpath.c_str(), _T("rb"));
			}
		}
	}
	if (fp == 0)
	{
		const BppChar* inc = searchEnv(SYMN_BPPINC, path.c_str());
		if (inc != 0)
		{
			fullpath = (const BppChar*)inc;
			fp = _tfopen(fullpath.c_str(), _T("rb"));
		}
	}
	if (fp == 0)
	{
		_scanner->parseError(CMP_FILE_NOT_FOUND);
		_print(_T("%s %s"), fullpath.c_str(), _scanner->getTokenString());
		errCnt++;
		return false;
	}
	//fread()
	static BppByte buf[200];
	static const size_t bufsiz = sizeof(buf);
	buffer.clear();
	size_t cnt = 0;
	do {
		cnt = fread(buf, 1, bufsiz, fp);
		if (cnt > 0) buffer.append(buf, cnt);
	} while (cnt == bufsiz);
	fclose(fp);
	return true;
}

static BppValueType getCommonLitExprType(BppValueType tl, BppValueType tr)
{
	if (tl == BVT_NULL || tr == BVT_NULL) return BVT_NULL;
	if ((tl == BVT_STRING || tl == BVT_STRINGREF)
		&& (tr == BVT_STRING || tr == BVT_STRINGREF)) return BVT_STRING;
	BppValueType rt = tl < tr&& tr != BVT_BOOL ? tr : tl;
	if (rt == BVT_STRING || rt == BVT_STRINGREF) return BVT_STRING;
#ifdef BPP_I64
	if (rt == BVT_INTPTR)
	{
		if (tl == BVT_ULONG || tr == BVT_ULONG) rt = BVT_ULONG;
	}
	else
#endif
		if (rt < BVT_INT && rt != BVT_BOOL) rt = BVT_INT;
	if (rt > _BVTVALMAX) rt = BVT_INTPTR;
	return rt;
}


/// litexpr ::= litexpr1 { "?" litexpr ":" litexpr } .
bool BppCompiler::_doLitExpr(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (errCnt > MAXERRS)
		return false;
	if (!_doLitExpr1(token, entry, errCnt)) return false;
	while (((token = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('?')))
	{
		BppDictionaryEntry de1(0, BppValue()), de2(0, BppValue());
		if (!_doLitExpr(_scanner->getToken(errCnt), de1, errCnt)) return false;
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(':')))
		{
			++errCnt;
			break;
		}
		if (!_doLitExpr(_scanner->getToken(errCnt), de2, errCnt)) return false;
		if (entry.getValue().asBool()) entry.setValue(de1.getValue());
		else entry.setValue(de2.getValue());
	}
	_scanner->back();
	return true;
}

/// litexpr1 ::= litexpr2 { "||" litexpr2 } .
bool BppCompiler::_doLitExpr1(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr2(token, entry, errCnt)) return false;
	while (_scanner->getToken(errCnt) == T_OR)
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr2(_scanner->getToken(errCnt), de, errCnt)) return false;
		if (!entry.getValue().asBool()) entry.setValue(de.getValue());
	}
	_scanner->back();
	return true;
}

/// litexpr2 ::= litexpr3 { "&&" litexpr3 } .
bool BppCompiler::_doLitExpr2(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr3(token, entry, errCnt)) return false;
	while (_scanner->getToken(errCnt) == T_AND)
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr3(_scanner->getToken(errCnt), de, errCnt)) return false;
		if (entry.getValue().asBool()) entry.setValue(de.getValue());
	}
	_scanner->back();
	return true;
}

/// litexpr3 ::= litexpr4 { "|" litexpr4 } .
bool BppCompiler::_doLitExpr3(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr4(token, entry, errCnt)) return false;
	while (((token = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('|')))
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr4(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppValue& left = (BppValue&)entry.getValue();
		BppValue& right = (BppValue&)de.getValue();
		BppValueType rt = getCommonLitExprType(left.getType(), right.getType());
		if (!BppInterpreter::_num_borOp(rt, left, right)) {
			_error(CMP_LITERAL_EXPECTED, &errCnt);
		}
	}
	_scanner->back();
	return true;
}

/// litexpr4 ::= litexpr5 { "^" litexpr5 } .
bool BppCompiler::_doLitExpr4(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr5(token, entry, errCnt)) return false;
	while (((token = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('^')))
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr5(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppValue& left = (BppValue&)entry.getValue();
		BppValue& right = (BppValue&)de.getValue();
		BppValueType rt = getCommonLitExprType(left.getType(), right.getType());
		if (!BppInterpreter::_num_xorOp(rt, left, right))
			_error(CMP_LITERAL_EXPECTED, &errCnt);
	}
	_scanner->back();
	return true;
}

/// litexpr5 ::= litexpr6 { "&" litexpr6 } 
bool BppCompiler::_doLitExpr5(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr6(token, entry, errCnt)) return false;
	while (((token = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('&')))
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr6(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppValue& left = (BppValue&)entry.getValue();
		BppValue& right = (BppValue&)de.getValue();
		BppValueType rt = getCommonLitExprType(left.getType(), right.getType());
		if (!BppInterpreter::_num_bandOp(rt, left, right))
			_error(CMP_LITERAL_EXPECTED, &errCnt);
	}
	_scanner->back();
	return true;
}


#define cmpres(T) T(vl) < T(vr) ? -1 : T(vr) < T(vl) ? 1 : 0
static BppInt cmpLiteralValues(const BppValue& vl, const BppValue& vr)
{
	BppValueType tl = vl.getType();
	BppValueType tr = vr.getType();
	if (tl == BVT_NULL) return (tr == BVT_NULL) ? 0 : -1;
	if (tr == BVT_NULL) return (tl == BVT_NULL) ? 0 : 1;
	BppValueType rt = getCommonLitExprType(tl, tr);
	BppInt result;
	switch (rt)
	{
	case BVT_BOOL:
		result = cmpres(BppByte);
		break;
	case BVT_INT:
		result = cmpres(BppInt);
		break;
	case BVT_UINT:
		result = cmpres(BppUInt);
		break;
	case BVT_LONG:
		result = cmpres(BppLong);
		break;
	case BVT_ULONG:
		result = cmpres(BppULong);
		break;
	case BVT_INTPTR:
#ifdef GCC
		result = cmpres(BppLong);
#else
		result = cmpres(BppIntPtr);
#endif   
		break;
	case BVT_FLOAT:
	case BVT_DOUBLE:
		result = cmpres(double);
		break;
	case BVT_STRING:
	{
		BppString* sl = vl;
		BppString* sr = vr;
		if (sl && sr)
		{
			result = (BppInt)(sl->compare(*sr));
			break;
		}
	}
	default:
#ifdef GCC
		result = cmpres(BppLong);
#else
		result = cmpres(BppIntPtr);
#endif   
		break;
	}
	return result;
}
#undef cmpres

/// litexpr6 ::= litexpr7 { ("==" | "!=") litexpr7 } .
bool BppCompiler::_doLitExpr6(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr7(token, entry, errCnt)) return false;
	while ((token = _scanner->getToken(errCnt)) == T_EQ || token == T_NE)
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr7(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppInt cmp = cmpLiteralValues(entry.getValue(), de.getValue());
		if (token == T_EQ)
			entry.setValue(BppValue::makeBool(cmp == 0));
		else
			entry.setValue(BppValue::makeBool(cmp != 0));
	}
	_scanner->back();
	return true;
}

/// litexpr7 ::= litexpr8 { ("<" | ">" | "<=" | ">="} litexpr8 } .
bool BppCompiler::_doLitExpr7(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr8(token, entry, errCnt)) return false;
	BppChar c = 0;
	while ((token = _scanner->getToken(errCnt)) == T_LE
		|| token == T_GE
		|| ((token == T_ANY) && (((c = _scanner->getChar()) == _T('>')) || (c == _T('<'))))
		)
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr8(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppInt cmp = cmpLiteralValues(entry.getValue(), de.getValue());
		switch (token)
		{
		case T_LE:
			entry.setValue(BppValue::makeBool(cmp <= 0));
			break;
		case T_GE:
			entry.setValue(BppValue::makeBool(cmp >= 0));
			break;
		default:
			if (c == _T('<'))
				entry.setValue(BppValue::makeBool(cmp < 0));
			else
				entry.setValue(BppValue::makeBool(cmp > 0));
		}
	}
	_scanner->back();
	return true;
}

/// litexpr8 ::= litexpr9 { ("<<" | ">>") litexpr9 } .
bool BppCompiler::_doLitExpr8(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr9(token, entry, errCnt)) return false;
	while ((token = _scanner->getToken(errCnt)) == T_SHL || token == T_SHR)
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr9(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppValue& left = (BppValue&)entry.getValue();
		BppValue& right = (BppValue&)de.getValue();
		BppValueType rt = getCommonLitExprType(left.getType(), right.getType());
		bool ok = false;
		if (token == T_SHL)
			ok = BppInterpreter::_num_shlOp(rt, left, right);
		else
			ok = BppInterpreter::_num_shrOp(rt, left, right);
		if (!ok) _error(CMP_LITERAL_EXPECTED, &errCnt);
	}
	_scanner->back();
	return true;
}




/// litexpr9 ::= litexpr10 { ("+" | "-") litexpr10 } .
bool BppCompiler::_doLitExpr9(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr10(token, entry, errCnt)) return false;
	BppChar c = 0;
	while (((token = _scanner->getToken(errCnt)) == T_ANY) && (((c = _scanner->getChar()) == _T('+')) || (c == _T('-'))))
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr10(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppValue& left = (BppValue&)entry.getValue();
		BppValue& right = (BppValue&)de.getValue();
		BppValueType rt = getCommonLitExprType(left.getType(), right.getType());
		bool ok = false;
		if (c == _T('+')) {
			if (rt == BVT_STRING) {
				BppString* sl = left;
				BppString* sr = right;
				if (sl) {
					if (sr) {
						BppString* s = sl->clone();
						(*s) += (*sr);
						left = s;
						ok = true;
					}
					else if (right.isIntegralType()) {
						BppString* s = sl->clone();
						(*s) += BppValue::makeChar(right).asBppChar();
						left = s;
						ok = true;
					}
				}
				else if (sr && left.isIntegralType()) {
					BppString* s = sr->clone();
					s->insert(0, BppValue::makeChar(left).asBppChar());
					left = s;
					ok = true;
				}
			}
			else
			   ok = BppInterpreter::_num_addOp(rt, left, right);
		}
		else
			ok = BppInterpreter::_num_subOp(rt, left, right);
		if (!ok) _error(CMP_LITERAL_EXPECTED, &errCnt);
	}
	_scanner->back();
	return true;
}

/// litexpr10 ::= litexpr11 { ("*" | "/" | "%") litexpr11 } .
bool BppCompiler::_doLitExpr10(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (!_doLitExpr11(token, entry, errCnt)) return false;
	BppChar c = 0;;
	while (((token = _scanner->getToken(errCnt)) == T_ANY) && (((c = _scanner->getChar()) == _T('*')) || (c == _T('/')) || (c == _T('%'))))
	{
		BppDictionaryEntry de(0, BppValue());
		if (!_doLitExpr11(_scanner->getToken(errCnt), de, errCnt)) return false;
		BppValue& left = (BppValue&)entry.getValue();
		BppValue& right = (BppValue&)de.getValue();
		BppValueType rt = getCommonLitExprType(left.getType(), right.getType());
		bool ok = false;
		if (c == _T('*'))
			ok = BppInterpreter::_num_mulOp(rt, left, right);
		else if (c == _T('/'))
			ok = BppInterpreter::_num_divOp(rt, left, right);
		else
			ok = BppInterpreter::_num_remOp(rt, left, right);
		if (!ok) _error(CMP_LITERAL_EXPECTED, &errCnt);
	}
	_scanner->back();
	return true;
}

/// litexpr11 ::= [ "+" | "-" | "!" | "~" ] litexpr12 .
bool BppCompiler::_doLitExpr11(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	if (token == T_ANY)
	{
		BppChar c = _scanner->getChar();
		if ((c == _T('+')) || (c == _T('-')) || (c == _T('!')) || (c == _T('~')))
		{
			BppDictionaryEntry de(0, BppValue());
			if (!_doLitExpr11(_scanner->getToken(errCnt), de, errCnt)) return false;
			const BppValue& v = de.getValue();
			BppValueType vt = v.getType();
			switch (c)
			{
			case _T('+'):
				if (v.inTypeRange(_BVTINTEGRALBASE, _BVTVALMAX))
					entry.setValue(v);
				else (entry.setValue(BppInt(v)));
				break;
			case _T('-'):
				switch (vt)
				{
				case BVT_SBYTE:
					entry.setValue(BppSByte(-((BppSByte)v)));
					break;
				case BVT_BYTE:
					entry.setValue(BppByte(-((BppByte)v)));
					break;
				case BVT_SHORT:
					entry.setValue(BppShort(-((BppShort)v)));
					break;
				case BVT_USHORT:
					entry.setValue(BppUShort(-((BppUShort)v)));
					break;
				case BVT_INT:
				case BVT_BOOL:
					entry.setValue(-((BppInt)v));
					break;
				case BVT_UINT:
					entry.setValue(BppUInt(-((BppInt)v)));
					break;
				case BVT_LONG:
					entry.setValue(-((BppLong)v));
					break;
				case BVT_ULONG:
					entry.setValue(BppULong(-((BppLong)v)));
					break;
				case BVT_INTPTR:
#ifdef GCC
					entry.setValue(-((BppLong)v));
#else
					entry.setValue(-((BppIntPtr)v));
#endif 
					break;
				case BVT_FLOAT:
					entry.setValue(-((float)v));
					break;
				case BVT_DOUBLE:
					entry.setValue(-((double)v));
					break;
				default:
					entry.setValue(-((BppInt)v));
				}
				break;
			case _T('!'):
				entry.setValue(BppValue::makeBool(!v.asBool()));
				break;
			case _T('~'):
				switch (vt)
				{
				case BVT_SBYTE:
					entry.setValue(BppSByte(~((BppSByte)v)));
					break;
				case BVT_BYTE:
					entry.setValue(BppByte(~((BppByte)v)));
					break;
				case BVT_SHORT:
					entry.setValue(BppShort(~((BppShort)v)));
					break;
				case BVT_USHORT:
					entry.setValue(BppUShort(~((BppUShort)v)));
					break;
				case BVT_INT:
				case BVT_BOOL:
					entry.setValue(~((BppInt)v));
					break;
				case BVT_UINT:
					entry.setValue(~((BppUInt)v));
					break;
				case BVT_LONG:
					entry.setValue(~((BppLong)v));
					break;
				case BVT_ULONG:
					entry.setValue(~((BppULong)v));
					break;
				case BVT_INTPTR:
#ifdef GCC
					entry.setValue(-((BppLong)v));
#else
					entry.setValue(-((BppIntPtr)v));
#endif 
				default:
					entry.setValue(~((BppInt)v));
				}
				break;
			}
			return true;
		}
	}
	return _doLitExpr12(token, entry, errCnt);
}

/// litexpr12 ::= ( "(" litexpr ")" ) | literal 
bool BppCompiler::_doLitExpr12(BppToken token, BppDictionaryEntry& entry, int& errCnt)
{
	BppDictionaryEntry de(0, BppValue());
	if ((token == T_ANY) && (_scanner->getChar() == _T('(')))
	{
		if (!_doLitExpr(_scanner->getToken(errCnt), de, errCnt)) return false;
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) ++errCnt;
	}
	else if (!_doLiteralValue(token, de, errCnt)) return false;
	entry.setValue(de.getValue());
	return true;
}



bool BppCompiler::_doMacroDefinition(const BppString& name, int& errCnt)
{
	if (errCnt > MAXERRS)
		return false;
	FlagHlp inMacroDef(_suppressMacroExpansion, true);

	if (_definedMacros.find(name)) {
		_scanner->parseError(BppString::format(_T("Warning: Redefinition of macro '%s'"), name.c_str()).c_str());
		_print(_T("%s"), _scanner->getTokenString());
	}
	BppVector* argvec = 0; //new BppVector();
	BppValue tmp; // deletes argvec if not used
	bool hasArgs = _scanner->peekChar() == _T('(');
	BppToken tkn = _scanner->getToken(errCnt, false, false);
	if (hasArgs && tkn == T_ANY && _scanner->getChar() == _T('(')) {
		argvec = new BppVector();
		tmp = argvec;
		_getIdList(*argvec, 0, _T(")"), errCnt, 0, false, _T("__VA_ARGS__"));
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) ++errCnt;
	}
	else {
		//hasArgs = false;
		_scanner->back();
	}
	if (errCnt > 0) return false;

	BppSavedToken* tokenList = 0;
	// gibt es keine Argumentliste, so wird das erste Token aus dem Stack geholt.
	// Zeilenende wird in Makros ausgewertet
	for (tkn = _scanner->getToken(errCnt, false, false);
		tkn != T_EOF && !(tkn == T_ANY && _scanner->getChar() == _T('\n'));
		tkn = _scanner->getToken(errCnt, false, false)) {
		_scanner->back(true,&tokenList);
	}
	if (errCnt > 0) return false;
	_definedMacros.defineMacro(name, argvec, tokenList);
	return true;
}

bool BppCompiler::_parseMacroArgumentList(TokenListVec& tokenListVec, int& errCnt)
{
	FlagHlp inMacroDef(_suppressMacroExpansion, true);
	BppScanner* pscan = getActiveScanner();
	if (!pscan->_tokenStack && pscan->peekChar() != _T('(')) return true;
	BppToken tkn = pscan->getToken(errCnt);
	if (!(tkn == T_ANY && _scanner->getChar() == _T('('))) {
		// keine Argumentliste
		pscan->back();
		return true;
	}
	int brLevel = 1;
	BppSavedToken* tokenStack = 0;
	bool listStarted = true;
	while (brLevel > 0) {
		tkn = pscan->getToken(errCnt/*, true*/);
		switch (tkn) {
		case T_EOF:
			_error(CMP_EOF_IN_ARGUMENTLIST, &errCnt);
			brLevel = 0;
			break;
		case T_ANY:
			switch (pscan->getChar()) {
			case _T('('):
				++brLevel;
				break;
			case _T(')'):
				--brLevel;
				break;
			case _T(','):
				if (brLevel == 1) {
					tokenListVec.push_back(tokenStack);
					tokenStack = 0;
					listStarted = true;
					continue;
				}
				break;
			}
			if (brLevel > 0) {
				pscan->back(true, &tokenStack);
				listStarted = false;
			}
			break;
		default:
			pscan->back(true, &tokenStack);
			listStarted = false;
		}
	}
	if (listStarted || (tokenStack != 0))
		tokenListVec.push_back(tokenStack);
	size_t nvecs = tokenListVec.size();
	for (size_t i = 0; i < nvecs; ++i) {
		BppSavedToken* tklist = tokenListVec.at(i);
		BppSavedToken* ptbefore = 0;
		for (BppSavedToken* pt = tklist; pt; ptbefore = pt, pt = pt ? pt->next : tklist) {
			if (pt->ttype == T_IDENTIFIER) {
				BppScanner* pscan = getActiveScanner();
				BppSavedToken* savedTokStack = pscan->_tokenStack;
				pscan->_tokenStack = pt->next ? pt->next : 0;
				if (_doMacroExpansion(pt->stringBuffer.data(), errCnt)) {
					pt->next = 0;
					BppSavedToken* replacelist = pscan->_tokenStack;
					if (replacelist) {
						BppSavedToken* p = replacelist;
						while (p->next) p = p->next;
						p->next = pt->next;
					}
					if (ptbefore) {
						if (replacelist) ptbefore->next = replacelist;
						else ptbefore->next = pt->next;
						pt->next = 0;
						delete pt;
						pt = ptbefore;
					}
					else // erstes Element
					{
						pt->next = 0;
						delete pt;
						pt = 0;
						tklist = replacelist;
					}
				}
				pscan->_tokenStack = savedTokStack;
				tokenListVec._listVec[i] = BppValue(INTPTRCAST(tklist), true);
			}
		}
	}
	return true;
}

bool BppCompiler::_doMacroExpansion(const BppString& name, int& errCnt)
{
	if (errCnt > MAXERRS) return false;
	static const BppString VA_ARGS = _T("__VA_ARGS__");
	BppDictionaryEntry* pentry = _definedMacros.find(name);
	if (!pentry) return false;

	BppVector* args = _definedMacros.getDefinedArgs(pentry);
	TokenListVec tokenListVec; // Vektor der Tokenlisten der aktuellen Argumente
	if (args) _parseMacroArgumentList(tokenListVec, errCnt);
	// hier ist die Argumentliste mit Klammern vom Parser entfernt
	size_t ndefargs = args ? args->size() : 0; // Anzahl definierte formaler Parameter
	size_t nrequargs = ndefargs; // geforderte Mindestzahl
	if (ndefargs > 0) {
		BppString* lastArg = (*args)[ndefargs - 1];
		if (lastArg && lastArg->compare(VA_ARGS) == 0) 
			nrequargs--; // letztes Argument ist Ellipse
	}
	size_t nTokLists = tokenListVec.size();
	if (nTokLists < nrequargs) {
		// zu wenige Argumente
		_scanner->parseError(BppString::format(_T("Warning: Too few arguments for macro '%s'"), name.c_str()).c_str());
		_print(_T("%s"), _scanner->getTokenString());
	}
	else if (nTokLists > ndefargs && nrequargs == ndefargs) {
		// zu viele Argumente
		_scanner->parseError(BppString::format(_T("Warning: Too many arguments for macro '%s'"), name.c_str()).c_str());
		_print(_T("%s"), _scanner->getTokenString());
	}

	// Tokenliste des geparsten Makrotextes (Body ohne Kopf)
	BppSavedToken* replaceList = _definedMacros.getTokenList(pentry);
	if (!replaceList) return true; // leeres Makro durch nichts ersetzen

	if (nTokLists > ndefargs && nrequargs < ndefargs) {
		// Tokenlisten der variablen Argumentliste zusammenfassen
		tokenListVec.mergeFrom(nrequargs);
		nTokLists = ndefargs;
	}

	// formale Parameter durch Tokenlisten der Argumente ersetzen

	// Kopie der Tokenliste der Makrodefinition
	BppSavedToken* tklist = replaceList->clone();
	BppSavedToken* itemBefore = 0;
	for (BppSavedToken* item = tklist; item != 0; itemBefore = item, item = item ? item->next : tklist) {
		if (item->ttype == T_IDENTIFIER) {
			BppString nm = item->stringBuffer;
			for (size_t i = 0; i < ndefargs; ++i) {
				BppString* s = (*args)[i];
				if (s && ((*s) == nm)) {
					BppSavedToken* ptoklist = (i < nTokLists) ? tokenListVec.at(i) : 0;
					BppSavedToken* clonedlist = ptoklist ? ptoklist->clone() : 0;
					if (clonedlist) {
						if (itemBefore) itemBefore->next = clonedlist;
						else tklist = clonedlist;
						BppSavedToken* p = clonedlist;
						while (p->next) {
							p = p->next;
						}
						p->next = item->next;
						item->next = 0;
						delete item;
						if (!itemBefore) {
							item = 0;
							tklist = clonedlist;
							break;
						}
						item = itemBefore;
						break;
					}
					else {
						item->ttype = T_NOTOKEN;
						break;
					}
				}
			}
		}
	} // end for item
	// leere Token entfernen
	BppSavedToken* pbefore = 0;
	for (BppSavedToken* pnt = tklist; pnt;  pbefore = pnt, pnt ? pnt = pnt->next : tklist) {
		if (pnt->ttype == T_ANY && pnt->charBuffer == _T(',')) {
			// leeres Token hinter Verkettung nach Komma
			if (pnt->next && pnt->next->ttype == T_CONCAT && pnt->next->next && pnt->next->next->ttype == T_NOTOKEN) {
				BppSavedToken* behind = pnt->next->next;
				pnt->next->next = 0;
				delete pnt;
				pnt = 0;
				if (pbefore) {
					pbefore->next = behind;
					pnt = pbefore;
				}
				else tklist = behind;
			}
			continue;
		}
		if (pnt->ttype == T_NOTOKEN) {
			BppSavedToken* behind = pnt->next;
			if (behind && behind->ttype == T_CONCAT && behind->next && behind->next->ttype == T_ANY && behind->next->charBuffer == _T(',')) {
				// leeres Token vor Verkettung mit Komma
				BppSavedToken* p = behind->next->next; // p zeigt auf Token hinter dem Komma
				behind->next->next = 0;
				delete pnt;
				behind = p;
			}
			else {
				pnt->next = 0;
				delete pnt;
			}
			pnt = 0;
			if (pbefore) {
				pbefore->next = behind;
				pnt = pbefore;
			}
			else tklist = behind;
		}
	}
	// Verkettungen
	pbefore = 0;
	for (BppSavedToken* ptc = tklist; ptc && ptc->next; pbefore = ptc, ptc ? ptc = ptc->next : tklist) {
		if (ptc->ttype != T_CONCAT) continue;
		if (!pbefore) {
			tklist = ptc->next;
			ptc->next = 0;
			delete ptc;
			ptc = 0;
			continue;
		}
		BppToken tb = pbefore->ttype;
		BppToken t = ptc->next->ttype;
		if (tb == T_IDENTIFIER || (tb >= T_CLASS && tb <= T_AUTO)) {
			if (t == T_IDENTIFIER || t == T_NUMBER || (tb >= T_CLASS && tb <= T_AUTO)) {
				pbefore->stringBuffer += ptc->next->stringBuffer;
				if (!ptc->next->next || ptc->next->next->ttype != T_CONCAT) {
					pbefore->stringBuffer.push_back(_T('\0'));
					//buf.adjustSize();
					BppToken kt = BppScanner::getKeywordTypeFromString(pbefore->stringBuffer.data());
					pbefore->ttype = (kt == T_NOTOKEN) ? T_IDENTIFIER : kt;
				}
				//pbefore->stringBuffer = buf.data();
				pbefore->next = ptc->next->next;
				ptc->next->next = 0;
				delete ptc;
				ptc = pbefore;
			    continue;
			}
		}
		if (tb == T_ANY && t == T_ANY) {
			BppChar buf[3] = { 0,0,0 };
			buf[0] = pbefore->charBuffer;
			buf[1] = ptc->next->charBuffer;
			BppToken kt = BppScanner::getKeywordTypeFromString(buf);
			if (kt != T_NOTOKEN) {
				pbefore->stringBuffer = buf;
				pbefore->ttype = kt;
				pbefore->next = ptc->next->next;
				ptc->next->next = 0;
				delete ptc;
				ptc = pbefore;
			}
		}
		else {
			pbefore->next = ptc->next;
			ptc->next = 0;
			delete ptc;
			ptc = pbefore;
		}
	}
	// Stringify
	pbefore = 0;
	for (BppSavedToken* pts = tklist; pts; pbefore = pts, pts ? pts = pts->next : tklist) {
		if (!(pts->ttype == T_ANY && pts->charBuffer == _T('#'))) continue;
		BppSavedToken* behind = pts->next;
		if (behind && behind->ttype >= T_STRING) {
			pts->ttype = T_STRING;
			switch (behind->ttype) {
			case T_IDENTIFIER:
			case T_STRING:
			case T_NUMBER:
			case T_FNUMBER:
				pts->stringBuffer = behind->stringBuffer;
				pts->stringBuffer.push_back(_T('\0'));
				break;
			case T_ANY:
				pts->stringBuffer.clear();
				pts->stringBuffer.push_back(behind->charBuffer);
				pts->stringBuffer.push_back(_T('\0'));
				break;
			default:
				pts->stringBuffer = _scanner->getTokenName(behind->ttype);
			}
			pts->next = behind->next;
			behind->next = 0;
			delete behind;
			continue;
		}
		pts -> next = 0;
		if (pbefore) {
			pbefore->next = behind;
			delete pts;
			pts = pbefore;
		}
		else {
			tklist = behind;
			delete pts;
			pts = 0;
		}
	}
	// concat strings
	pbefore = 0;
	for (BppSavedToken* ps = tklist; ps; pbefore = ps, ps ? ps = ps->next : tklist) {
		if (!(ps->ttype == T_STRING && ps->next && ps->next->ttype==T_STRING)) continue;
		BppSavedToken* behind = ps->next;
		ps->stringBuffer += (behind->stringBuffer.data());
		ps->stringBuffer.push_back(_T('\0'));
		ps->next = behind->next;
		behind->next = 0;
		delete behind;
		ps = (pbefore) ? pbefore : 0;
	}
	if (!tklist) return true;
	BppSavedToken* lend = tklist;
	while (lend->next) lend = lend->next;
	BppScanner* pscan = getActiveScanner();
	lend->next = pscan->_tokenStack;
	pscan->_tokenStack = tklist;
	return true;
}

BppCompiler::DefinedMacros::DefinedMacros()
	:_defined(true)
{
}

BppCompiler::DefinedMacros::~DefinedMacros()
{
	clear();
}

void BppCompiler::DefinedMacros::clear()
{
	for (BppDictionaryEntry* p = _defined.getFirst(); p; p = p->getNext())
		delete getTokenList(p);
	_defined.clear();
}

/**
* @param name Macro nane.
* @param definedArgs Vector of macro arguments
* @param tokenList List of token descriptors representing content of macro.
*/
void BppCompiler::DefinedMacros::defineMacro(const BppString& name, BppVector* definedArgs, BppSavedToken* tokenlist)
{
	BppDictionaryEntry* pe = find(name);
	if (pe) delete getTokenList(pe);
	else pe = &_defined[name];
	BppVector* ev = new BppVector(2, true);
	(*ev)[IARGS] = definedArgs;
	(*ev)[ILIST] = BppValue(INTPTRCAST(tokenlist), true);
	pe->setValue(ev);
}

BppDictionaryEntry* BppCompiler::DefinedMacros::find(const BppString& name)
{
	return _defined.find(name);
}

BppVector* BppCompiler::DefinedMacros::getDefinedArgs(BppDictionaryEntry* entry) const
{
	if (!entry || entry->getDictionary() != &_defined)
		return 0;
	BppVector* pv = entry->getValue();
	if (pv) return ((*pv)[IARGS]);
	return 0;
}

/** 
@param entry Entry of dictionary of defined macros
@return Pointer to descriptor of first token of macro content on success, otherwise 0.
@note Client code must not delete result.
@see _defined
*/
BppSavedToken* BppCompiler::DefinedMacros::getTokenList(BppDictionaryEntry* entry) const
{
	if (!entry || entry->getDictionary() != &_defined)
		return 0;
	BppVector* pv = entry->getValue();
	if (pv && pv->size() > ILIST) return (BppSavedToken*)((*pv)[ILIST]).ptr();
	return 0;
}

bool BppCompiler::DefinedMacros::erase(const BppString& name)
{
	return erase(find(name));
}

bool BppCompiler::DefinedMacros::erase(BppDictionaryEntry* entry)
{
	if (!entry || entry->getDictionary() != &_defined)
		return false;
	delete getTokenList(entry);
	return _defined.erase(entry);
}

enum CompCondFlags {
	CCF_NONE = 0,
	CCF_ACTIVE = 1,
	CCF_ELSE = 2,
	CCF_ELIF = 4
};
bool BppCompiler::_active() const
{
	size_t s = _condCompStack.size();
	return (s == 0 || (_condCompStack[s - 1] & CCF_ACTIVE));
}

#ifndef DOXY
struct ppLock {
	int& _v;
	ppLock(int& v) :_v(v) { ++_v; }
	~ppLock() { --_v; }
};
#endif

/**
<pre>preprocinstruction ::= "#" ( <use> | <import> | <endsrc> | <defvar> | <deflit> |
								<literaldef> | <undeflit> | <include> | <condcompdef> | <define> | <undef> |
								<error> | <warning> | <pragma> )
   use ::= "use" """ <modulenameorpath> """
   import ::= "import" """ <dllnameorpath> """
   endsrc ::= "endsrc"
   defvar ::= "defvar" <identifier> <literal>
   deflit ::= "deflit" <identifier> <literal>
   literaldef ::= "literal" <identifier> <literal>
   undeflit ::= "undeflit" <identifier>
   include ::= "include" ( "<" <filenameorpath> ">") | (""" <filenameorpath> """ )
	condcompiledef ::= <if> | <elif> | <ifdef> | <ifndef> | "else" | "endif"
	if ::= "if" <litexpr>
	elif ::= "elif" <litexpr>
	ifdef ::= "ifdef" <identifier>
	ifndef ::= "ifndef" <identifier>
	define ::= "define" <identifier> ["(" <argumentlist> ")"] <replacementtext>
	undef ::= "undef" <identifier>
	error ::= "error" <txtexpr> 
	warning ::= "warning" <txtexpr>
	txtexpr ::= <litexpr of result type string>
	pragma ::= "pragma" <pragmafunc>
	pragmafunc ::= <cdefine> | <message> | <codepage>
	cdefine ::= "cdefine" "on"|"off"|"push"|"pop"
	message ::= "message" <txtexpr>
	codepage ::= "codepage" <litexpr>
</pre>
@param errCnt Error counter
@return true if scanner should get next token before further processing, false if not.
*/
bool BppCompiler::_doPreproc(int& errCnt)
{
	if (errCnt > MAXERRS)
		return false;
	ppLock pplock(_preprocLevel);
	/* get the preprocessor symbol */
	BppToken token = _scanner->getToken(errCnt);
	switch (token) {
	case T_IF:
	{
		BppString* s = new BppString(_T("if"));
		BppDictionaryEntry de(s, BppValue());
		if (!_doLitExpr(_scanner->getToken(errCnt), de, errCnt)) return true;
		bool active = de.getValue().asBool();
		_condCompStack.push_back(active ? CCF_ACTIVE : CCF_NONE);
		if (!active) return _doPreprocIgnore(errCnt);
		return true;
	}
	break;
	case T_ELSE:
	{
		size_t s = _condCompStack.size();
		if (!s || (_condCompStack[s - 1] & CCF_ELSE))
		{
			_error(CMP_UNEXPECTED_ELSE_DIRECTIVE, &errCnt);
			return true;
		}
		_condCompStack[s - 1] = CCF_ELSE;
		return  _doPreprocIgnore(errCnt);
	}
	break;
	default:
		if (!_require(T__DFUNCTION, T_IDENTIFIER, token))
		{
			errCnt++;
			return true;
		}
	}
	const BppChar* tkstr = _scanner->getTokenString();
	if (_tcscmp(tkstr, _T("elif")) == 0) {
		size_t s = _condCompStack.size();
		if (!s || (_condCompStack[s - 1] & CCF_ELSE))
		{
			_error(CMP_UNEXPECTED_ELIF_DIRECTIVE, &errCnt);
			return true;
		}
		_condCompStack[s - 1] = CCF_ELIF;
		return  _doPreprocIgnore(errCnt);
	}
	else if (_tcscmp(tkstr, _T("ifdef")) == 0) {
		FlagHlp hlp(_suppressMacroExpansion, true);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppString* s = new BppString(tkstr);
		BppDictionaryEntry de(s, BppValue());
		bool active = (_namedLiterals.find(tkstr) != 0 || _definedMacros.find(tkstr) != 0);
		_condCompStack.push_back(active ? CCF_ACTIVE : CCF_NONE);
		if (!active) return _doPreprocIgnore(errCnt);
	}
	else if (_tcscmp(tkstr, _T("ifndef")) == 0) {
		FlagHlp hlp(_suppressMacroExpansion, true);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppString* s = new BppString(tkstr);
		BppDictionaryEntry de(s, BppValue());
		bool active = (_namedLiterals.find(tkstr) == 0 && _definedMacros.find(tkstr) == 0);
		_condCompStack.push_back(active ? CCF_ACTIVE : CCF_NONE);
		if (!active) return _doPreprocIgnore(errCnt);
	}
	else if (_tcscmp(tkstr, _T("endif")) == 0) {
		if (_condCompStack.size() == 0)
			_error(CMP_UNEXPECTED_ENDIF_DIRECTIVE, &errCnt);
		else _condCompStack.pop_back();
		return true;
	}
	else if (_tcscmp(tkstr, _T("use")) == 0)
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_STRING))
		{
			errCnt++;
			return true;
		}
		const BppChar* modulename = _scanner->getTokenString();
		const BppChar* fname = searchEnv(SYMN_BPPLIB, modulename);
		if (fname == 0)
			fname = modulename;
		BppDictionaryIO reader(_vm);
		if (!reader.readModule(fname))
		{
			_error(CMP_LIBRARY_IMPORT_FAILED, &errCnt);
		}
	}
	else if (_tcscmp(tkstr, _T("import")) == 0)
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_STRING))
		{
			errCnt++;
			return true;
		}
		const BppChar* modulename = _scanner->getTokenString();
#ifndef NO_DYNLIB
		if (!_vm.loadNativeModule(modulename))
#endif
		{
			_error(CMP_NMODULE_IMPORT_FAILED, &errCnt);
		}
	}
	else if (_tcscmp(tkstr, _T("endsrc")) == 0)
	{
		getActiveScanner()->_token = T_EOF;
		return false;
	}
	else if (_tcscmp(tkstr, _T("defvar")) == 0)
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppString* s = new BppString(tkstr);
		BppDictionaryEntry* entry = _symbols[s];
		BppDictionaryEntry dummy(*s, BppValue());
		if (entry->getSymbolType() == ST_SDATA) entry = &dummy;
		else entry->setSymbolType(ST_SDATA);

		bool ok = _doLitExpr(_scanner->getToken(errCnt), dummy, errCnt);
		if (ok) {
			entry->setValue(BppValue(dummy.getValue(), true));
		}
		return ok;
	}
	else if ((_tcscmp(tkstr, _T("deflit")) == 0) || (!_useCDefine && _tcscmp(tkstr, _T("define")) == 0))
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppString s(tkstr);
		BppDictionaryEntry& entry = _namedLiterals[s];
		return _doLitExpr(_scanner->getToken(errCnt), entry, errCnt);
	}
	else if (_useCDefine && _tcscmp(tkstr, _T("define")) == 0)
	{
		FlagHlp hlp(_suppressMacroExpansion, true);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppString s(tkstr);
		if (_definedMacros.find(s)) {
			_scanner->parseError(BppString::format(_T("Warning: Redefinition of macro '%s'"), s.c_str()).c_str());
			_print(_T("%s"), _scanner->getTokenString());
		}
		return _doMacroDefinition(_scanner->getTokenString(), errCnt);
	}
	else if (_tcscmp(tkstr, _T("literal")) == 0)
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppString* s = new BppString(tkstr);
		BppDictionaryEntry de(s, BppValue());
		if (!_doLitExpr(_scanner->getToken(errCnt), de, errCnt)) return false;
		if (_namedLiterals.find(*s) == 0)
		{
			BppDictionaryEntry* entry = _namedLiterals[s];
			entry->setValue(de.getValue());
		}
		return true;
	}

	else if (_useCDefine && _tcscmp(tkstr, _T("undef")) == 0)
	{
		FlagHlp hlp(_suppressMacroExpansion, true);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		_definedMacros.erase(tkstr);
		return true;
	}
	else if ((_tcscmp(tkstr, _T("undeflit")) == 0) || (!_useCDefine && _tcscmp(tkstr, _T("undef")) == 0))
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		tkstr = _scanner->getTokenString();
		BppDictionaryEntry* entry = _namedLiterals.find(tkstr);
		if (entry != 0)
			_namedLiterals.erase(entry);
		return true;
	}
	else if (_tcscmp(tkstr, _T("include")) == 0)
		return _doInclude(errCnt);
	else if (_tcscmp(tkstr, _T("pragma")) == 0) {
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
		{
			errCnt++;
			return true;
		}
		const BppChar* ptype = _scanner->getTokenString();
		if (_tcscmp(ptype, _T("cdefine")) == 0) {
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
			{
				errCnt++;
				return true;
			}
			const BppChar* verb = _scanner->getTokenString();
			if ((_tcscmp(verb, _T("on")) == 0))
				setUseCDefine(true);
			else if ((_tcscmp(verb, _T("off")) == 0))
				setUseCDefine(false);
			else if ((_tcscmp(verb, _T("push")) == 0))
				pushUseCDefine();
			else if ((_tcscmp(verb, _T("pop")) == 0)) {
				if (!popUseCDefine())
				{
					_error(CMP_INVALID_POP_CDEFINE, &errCnt);
					return true;
				}
			}
		}
		else if (_tcscmp(ptype, _T("message")) == 0) {
			BppDictionaryEntry de(_T(""), BppValue());
			if (!_doLitExpr(_scanner->getToken(errCnt), de, errCnt)) return true;
			BppString* s = de.getValue();
			if (s) _error(s->c_str());
		}
		else if (_tcscmp(ptype, _T("codepage")) == 0) {
			BppDictionaryEntry de(_T(""), BppValue());
			if (!_doLitExpr(_scanner->getToken(errCnt), de, errCnt)) return true;
			const BppValue& v = de.getValue();
			BppUShort cp = 0;
			if (v.isIntegralType()) cp = v;
			else {
				BppString* s = v;
				if (s) cp = getCpIdFromName(s->c_str(),getActiveCp());
			}
			if (cp != 0) getActiveScanner()->_srcCodePage = cp;
		}
	}
	else if (_tcscmp(tkstr, _T("error")) == 0) {
		errCnt = MAXERRS;
		BppDictionaryEntry de(_T(""), BppValue());
		_doLitExpr(_scanner->getToken(errCnt), de, errCnt);
		BppString* s = de.getValue();
		_error(s ? s->c_str() : _T("error"));
		getActiveScanner()->_token = T_EOF;
		return false;
	}
	else if (_tcscmp(tkstr, _T("warning")) == 0) {
		BppDictionaryEntry de(_T(""), BppValue());
		_doLitExpr(_scanner->getToken(errCnt), de, errCnt);
		BppString* s = de.getValue();
		_error(s ? s->c_str() : _T("warning"));
		//_scanner->_token = T_EOF;
		return true;
	}
	else
	{
		_error(CMP_UNDEFINED_PREPROC_SYMBOL, &errCnt);
	}
	return true;
}


bool BppCompiler::_doPreprocIgnore(int& errCnt)
{
	for (BppToken t = _scanner->getToken(errCnt); ; t = _scanner->getToken(errCnt)) {
		if (t == T_EOF) {
			_error(CMP_UNEXPECTED_EOF, &errCnt);
			return false;
		}
		if (t != T_ANY || _scanner->getChar() != _T('#')) continue;
		size_t ssize = _condCompStack.size();
		bool parentActive = ssize < 2 || (_condCompStack[ssize - 2] & CCF_ACTIVE);
		t = _scanner->getToken(errCnt);
		switch (t) {
		case T_IF:
			_condCompStack.push_back(CCF_NONE);
			break;
		case T_ELSE:
			if (_condCompStack[ssize - 1] & CCF_ELSE) {
				_error(CMP_UNEXPECTED_ELSE_DIRECTIVE, &errCnt);
				return true;
			}
			_condCompStack[ssize - 1] = parentActive ? CCF_ACTIVE | CCF_ELSE : CCF_ELSE;
			if (ssize < 2 || parentActive) return true;
			break;
		case T_IDENTIFIER:
		{
			const BppChar* tkstr = _scanner->getTokenString();
			if (!_tcscmp(tkstr, _T("elif"))) {
				if (_condCompStack[ssize - 1] & CCF_ELSE) {
					_error(CMP_UNEXPECTED_ELIF_DIRECTIVE, &errCnt);
					return true;
				}
				if (_condCompStack[ssize - 1] & CCF_ELIF) break;
				if (ssize < 2) {
					BppDictionaryEntry de(tkstr, BppValue());
					if (!_doLitExpr(_scanner->getToken(errCnt), de, errCnt)) return true;
					bool active = de.getValue().asBool();
					_condCompStack[ssize - 1] = (active ? CCF_ACTIVE : CCF_NONE);
					if (active) return true;
				}
			}
			else if (!_tcscmp(tkstr, _T("ifdef"))) {
				_condCompStack.push_back(CCF_NONE);
			}
			else if (!_tcscmp(tkstr, _T("ifndef"))) {
				_condCompStack.push_back(CCF_NONE);
			}
			else if (!_tcscmp(tkstr, _T("endif"))) {
				if (ssize < 1) {
					_error(CMP_UNEXPECTED_ENDIF_DIRECTIVE, &errCnt);
					return true;
				}
				_condCompStack.pop_back();
				if (_active()) return true;
			}
		}
		break;
		}
	}
	return true;
}


// include ::= "include" ( "<" <filenameorpath> ">") | (""" <filenameorpath> """ )
// First notation searches in BPPINC environment variable path for specified file.
bool BppCompiler::_doInclude(int& errCnt)
{
	FlagHlp cdefine(_useCDefine, true);
	BppString path;
	BppToken tkn = _scanner->getToken(errCnt);
	bool isLibPath = false;
	if (tkn == T_STRING)
		path = _scanner->getTokenString();
	else if ((tkn == T_ANY) && (_scanner->getChar() == _T('<')))
	{
		isLibPath = true;
		if ((tkn = _scanner->getToken(errCnt)) == T_STRING)
			path = _scanner->getTokenString();
		else
		{
			_scanner->back();
			while (((tkn = _scanner->getToken(errCnt)) >= T_IDENTIFIER) && (tkn <= T_NIL))
			{
				path += _scanner->getTokenString();
				tkn = _scanner->getToken(errCnt);
				if (tkn == T_ANY)
				{
					BppChar c[2];
					c[1] = 0;
					c[0] = _scanner->getChar();
					if ((c[0] == _T('/')) || (c[0] == _T('.')))
					{
						path += (const BppChar*)c;
						continue;
					}
				}
				_scanner->back();
				break;
			} // end while
		} // end else
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('>')))
		{
			errCnt++;
			return false;
		}
	} // end !T_STRING
	if (path.length() == 0)
	{
		_error(CMP_INCLUDE_FILENAME_EXPECTED, &errCnt);
		return false;
	}
	BppString fullpath = path;

	FILE* fp = _tfopen(path.c_str(), _T("rt"));
	if (fp == 0 && (!isLibPath))
	{
		size_t n = _srcDirStack.size();
		for (size_t i = 0; (!fp) && i < n; ++i)
		{
			BppString* s = _srcDirStack[i];
			const BppChar* fpath = searchDirList(s->c_str(), path.c_str());
			if (fpath)
			{
				fullpath = fpath;
				fp = _tfopen(fullpath.c_str(), _T("rt"));
			}
		}
	}
	if (fp == 0)
	{
		const BppChar* inc = searchEnv(SYMN_BPPINC, path.c_str());
		if (inc != 0)
		{
			fullpath = (const BppChar*)inc;
			fp = _tfopen(fullpath.c_str(), _T("rt"));
		}
	}
	if (fp == 0)
	{
		_scanner->parseError(CMP_FILE_NOT_FOUND);
		_print(_T("%s %s"), fullpath.c_str(), _scanner->getTokenString());
		errCnt++;
		return false;
	}
	_scanner->push(fp, fullpath.c_str());
	_srcDirStack.push_front(new BppString(getFileDir(fullpath.c_str(), true)));
	return true;
}
