#include "bppcomp.h"
#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include <locale.h>
#include "bppclass.h"
#include "bppobjct.h"
#include "bppvalue.h"
#include "bppufunc.h"
#include "bppprint.h"
#include "dictio.h"
#include "bppvm.h"
#include "bppdbg.h"
#include "bppmsg.h"
#include "bpptchar.h"
#ifdef GCC
#include <errno.h>
#endif
/// maximum number of errors before abort compiling
#define MAXERRS 10


/** Internal helper type for code generation.
This type is used for temporary store of information about a variable
to be coded.
*/
struct BppVarInfo
{
	/// variable type
	BppVarType type;
	/// index inside a container
	size_t index;
	/// value type
	BppValueType varvtype;
	/// const flag
	bool isConst;
	/// mutable flag
	bool isMutable;
	/// requested class (for strong typed objects)
	BppClass* reqClass;
	/// ctor
	BppVarInfo() :type(VAR_NONE), index(BPP_NNPOS), varvtype(BVT_NULL), isConst(false), isMutable(false), reqClass(0) {}
};




/** Helper for compiling a function.
This helper struct contains data needed during compile a function.
It's used for argument type by some of the production functions.
*/
struct BppFuncContext
{
	/// ctor - does initialization
	BppFuncContext() : temporaries(1, true), codeVec(0), codeBuffer(0), catchLevel(0),
		returnType(BVT_NULL),returnConst(false), prClass(0), contextClass(0), objClass(0),constThis(false) {}
	/// vector of function arguments
	BppVector arguments;
	/// vector of functions temporary (i.e. local) variables
	BppVector temporaries;
	/// break stack
	BppBuffer breakStack;
	/// continue stack
	BppBuffer contStack;
	/// destination vector for result of compilation (contains codeBuffer and variables)
	BppVector* codeVec;
	/// buffer for bytecode produced on compiling (part of codeVec)
	BppBuffer* codeBuffer;
	/// Dictionary of static variables - value is index of variable in codeVec
	BppDictionary statics;
	/// level of catch blocks - must be > 0 for throw statement without following expression
	unsigned catchLevel;

	/// return type of function (BVT_NULL if not defined) 
	BppValueType returnType;
	/// flag for constant return value
	bool returnConst;
	/// optionally class if return type is of object type
	BppClass* prClass;
	/// optionally context class for check member protection
	BppClass* contextClass;
	/// optionally class of object calling operator, e.g. '->' or '.'
	BppClass* objClass;
	/// flag vor constant this reference
	bool constThis;


	/** Push entry to break stack.
	@param cbIdx index in codeBuffer where to store jump destination index
	@return old size of stack (i.e. index of new entry)
	*/
	BppNWord pushBreak(BppNWord cbIdx)
	{
		return breakStack.push_backNW(cbIdx);
	}
	/** Push entry to continue stack.
	@param cbIdx index in codeBuffer where to store jump destination index
	@return old size of stack (i.e. index of new entry)
	*/
	BppNWord pushContinue(BppNWord cbIdx)
	{
		return contStack.push_backNW(cbIdx);
	}

	/** Removes break from stack.
	@param oldIdx old index from stack (result of previous pop-call)
	@param savedVal value to be returned if no entry is removed (index in codeBuffer)
	@return If <i>oldIdx</i> is lower than index of TOS, value on TOS is
	returned (and removed), otherwise <i>savedVal</i>
	*/
	BppNWord popBreak(BppNWord oldIdx, BppNWord savedVal)
	{
		return (oldIdx < breakStack.size()) ? breakStack.pop_backNW() : savedVal;
	}

	/** Removes continue from stack until <i>oldIdx</i> reached.
	@param oldIdx old index from stack (result of previous pop-call)
	@return value of last removed entry (index in codeBuffer) or 0
	*/
	BppNWord popContinue(BppNWord oldIdx)
	{
		BppNWord result = 0;
		BppNWord curr = contStack.size();
		while (curr > oldIdx)
		{
			result = contStack.pop_backNW();
			curr = contStack.size();
		}
		return result;
	}

	/** Peek top of break stack.
	@return value of item on TOS if any, otherwise 0
	*/
	BppNWord peekBreak()
	{
		BppNWord s = breakStack.size();
		if (s > 0) return breakStack.getNW(s - sizeof(BppNWord));
		return 0;
	}

	/** Peek top of continue stack.
	@return value of item on TOS if any, otherwise 0
	*/
	BppNWord peekContinue()
	{
		BppNWord s = contStack.size();
		if (s > 0) return contStack.getNW(s - sizeof(BppNWord));
		return 0;
	}
};



/**
Constructor assigns reference to virtual machine, dictionaies of classes
and globally symbols as soon as printer interface provided by virtual machine.
@param vm Virtual machine owning the compiler.
@param errPrinter Pointer to printer interface for error message reporting.
@note If <i>errPrinter</i> not specified (or NULL), default output of virtual
machine is used for reporting errors.
@see BppVirtualMachine::printer
*/
BppCompiler::BppCompiler(BppVirtualMachine& vm, IBppPrinter* errPrinter)
	:_scanner(0),  _symbols(vm.symbols()), _classes(vm.classes()), _namedLiterals(true),_vm(vm),
	_print((errPrinter) ? *errPrinter : vm.printer()),
	_showInfo(false), _debugger(0), _genLineInfo(false), _genLocalVarInfo(false), _useCDefine(true), _currentLine(-1), _preprocLevel(0),
	_suppressMacroExpansion(false)
{
	_currentSource.addRef();
#ifndef WINCE
	errno = 0;
#endif
}

/**
does nothing
*/
BppCompiler::~BppCompiler()
{
}

/**
@return value of _showInfo flag
@see setShowInfo
*/
bool BppCompiler::getShowInfo() const
{
	return _showInfo;
}

/**
If flag set, names of classes and functions are written to output during compilation.
@param showInfo new value of flag
@see _doClass,_doRegularFunction,_doMemberFunction, hetShowInfo
*/
void BppCompiler::setShowInfo(bool showInfo)
{
	_showInfo = showInfo;
}

/**
@return current value of flag
@see setGenerateLineInfo
*/
bool BppCompiler::getGenerateLineInfo() const
{
	return _genLineInfo;
}

/**
If flag is set to <i>true</i> (should be done before call to compile-method), additional information
about source files and line numers are inserted into bytecode for debugging purposes.
@param genInfo new value of flag
@note Insertion of line information means additional opcodes and data entries in compile result,
i.e. it makes code bigger and slower. Therefore line imformation should be generated
for debug builds only.
@see getGenerateLineInfo
*/
void BppCompiler::setGenerateLineInfo(bool genInfo)
{
	_genLineInfo = genInfo;
}

/**
@return current value of flag
@see setGenerateLocalVarInfo
*/
bool BppCompiler::getGenerateLocalVarInfo() const
{
	return _genLocalVarInfo;
}

/**
If <i>genInfo</i> is <tt>true</tt> name information is appended to end of
bytecode vector of a compiled function.
Name information consists of a two-elements vector, where both elements are vectors too.
First element contains names of declared arguments, second names of local (temporary)
variables. Indexes of names in element vectors correspond to relative position of
variables inside stack frame of function while execution.<br>
If Function contains static variables, an additional dictionary is appended to second
element. Keys in that dictionary are names of local static variables, values are their
indexes in bytecode vector of compiled function.
@param genInfo Value of flag for include information about local variable names in generated code.
@note Creation of information about local variables is useful for debugging purposes.
It increases code size but doesn't affect execution speed.
@see getGenerateLocalVarInfo
*/
void BppCompiler::setGenerateLocalVarInfo(bool genInfo)
{
	_genLocalVarInfo = genInfo;
}

/**
@see setUseCDefine
*/
bool BppCompiler::getUseCDefine() const
{
	return _useCDefine;
}

/**
If flag is set (default), #define directive is handled like done by C preprocessor.
Otherwise #define and #undef are aliases of #deflit and #undeflit directives.
@param New value of flag.
@note In source code flag may be set using #pragma cdefine on|off directive.
*/
void BppCompiler::setUseCDefine(bool useCDefine)
{
	_useCDefine = useCDefine;
}

/**
Current state of _useCDefine is pushed to a stack to allow restore it by popUseCDefine later.
@see setUseCDefine, popUseCDefine
*/
void BppCompiler::pushUseCDefine()
{
	_useCDefStack.push_back(_useCDefine ? 1 : 0);
}

/**
Restores state of _useCDefine flag from stack and removes top item of stack.
@return <i>true</i> on success, otherwise (stack was empty) <i>false</i>.
*/
bool BppCompiler::popUseCDefine()
{
	if (_useCDefStack.size() == 0) return false;
	_useCDefine = _useCDefStack.pop_back();
	return true;
}

void BppCompiler::setDebugger(BppDebugger* debugger)
{
	_debugger = debugger;
}

void BppCompiler::onScannerDelete(BppScanner* scanner)
{
	if (!scanner) return;
	size_t i = _srcDirStack.getStringIndex(getFileDir(scanner->_srcName.c_str(), true));
	if (i != BPP_NNPOS) _srcDirStack.erase(i, 1);
}

bool BppCompiler::notifyPreproc(int& errCnt)
{
	if (errCnt < MAXERRS && !_preprocLevel) return _doPreproc(errCnt);
	return false;
}

bool BppCompiler::notifyToken(BppToken token, int& errCnt)
{
	if (_suppressMacroExpansion || token != T_IDENTIFIER || !_active()) return false;
	BppString id = _scanner->getTokenString();
	return _doMacroExpansion(id, errCnt);
}

bool BppCompiler::compile(FILE* fp, const BppChar* srcName)
{
	if (fp == 0)
	{
		_print(CMP_INVALID_FILEHANDLE);
		return false;
	}
#ifndef WINCE
	errno = 0;
#endif
	BppScanner scanner(fp, srcName);
	_srcDirStack.clear();
	_srcDirStack.push_back(new BppString(getFileDir(srcName, true)));
	_scanner = &scanner;
	BppScanner::_notifyCompiler = this;
	bool ok = compile();
	BppScanner::_notifyCompiler = 0;
	return ok;
}

bool BppCompiler::compile(const BppChar* src, const BppChar* srcName)
{
	if (src == 0)
	{
		_print(CMP_INVALID_FILESOURCE);
		return false;
	}
#ifndef WINCE
	errno = 0;
#endif
	BppScanner scanner(src, srcName);
	_scanner = &scanner;
	BppScanner::_notifyCompiler = this;
	bool ok = compile();
	BppScanner::_notifyCompiler = 0;
	return ok;
}

BppVector* BppCompiler::compileFunction(const BppChar* src, const BppChar* args, const BppChar* rtype, const BppChar* srcName, BppClass* pclass)
{
	if (!src)
	{
		compile(src, srcName);
		return 0;
	}
	C_LOCALE;
	initCompiler();
#ifndef WINCE
	errno = 0;
#endif
	BppString strSrc = BppString::format(_T("%s){\n%s\n}"), args, src);
	BppValueType rvaltype = BVT_NULL;
	BppClass* resultClass = 0;
	int errcnt = 0;
	BppScanner* psc = _scanner;
	BppCompiler* pscomp = BppScanner::_notifyCompiler;
	BppScanner scanner(strSrc.c_str(), srcName);
	_scanner = &scanner;
	BppScanner::_notifyCompiler = this;
	if (rtype && *rtype)
	{
		rvaltype = findVTypeFromName(T__DFUNCTION,rtype, errcnt, true, &resultClass);
		if (rvaltype == BVT_NULL && _tcscmp(rtype, TN_var) && _tcscmp(rtype, TN_null))
			_error(BppString::format(CMP_INVALID_TYPENAME_FMT, rtype).c_str(), &errcnt);
	}
	BppVector* result = _codeFunctionBody(_T("<anonymous>"), errcnt, false, pclass, rvaltype, resultClass);
	if (result && (errcnt > 0))
	{
		delete result;
		result = 0;
	}
	BppScanner::_notifyCompiler = pscomp;
	_scanner = psc;
	return result;
}

bool BppCompiler::litExpr(const BppChar* src, BppValue& result, const BppChar* srcName)
{
	if (!src)
	{
		compile(src);
		return false;
	}
	C_LOCALE;
	initCompiler();
#ifndef WINCE
	errno = 0;
#endif
	BppString strSrc = src;
	int errcnt = 0;
	BppScanner* psc = _scanner;
	BppCompiler* pscomp = BppScanner::_notifyCompiler;
	BppScanner scanner(strSrc.c_str(), srcName);
	_scanner = &scanner;
	BppScanner::_notifyCompiler = this;
	BppDictionaryEntry de(strSrc, BppValue());
	_doLitExpr(_scanner->getToken(errcnt), de, errcnt);
	result = de.getValue();
	if (!_fetchRequire(T__DFUNCTION,errcnt, T_EOF)) errcnt++;
	BppScanner::_notifyCompiler = pscomp;
	_scanner = psc;
	return errcnt == 0;
}


void BppCompiler::initCompiler()
{
	_useCDefine = true;
	_namedLiterals.clear();
	_definedMacros.clear();
	_condCompStack.clear();
	_useCDefStack.clear();
	BppDictionary& predefs = _vm.predefines();
	for (BppDictionaryEntry* pe = predefs.getFirst(); pe; pe = pe->getNext()) {
		_namedLiterals[*pe->getKey()] = pe->getValue();
	}
	for (unsigned i = _BVTMIN; i <= _BVTREFMAX; ++i)
	{
		const BppChar* vtn = getValueTypeName((BppValueType)i, true, true);
		if (!vtn) continue;
		_namedLiterals[vtn] = (BppInt)i;
	}
	_namedLiterals[LN_EDOM] = (BppInt)EDOM;
	_namedLiterals[LN_ERANGE] = (BppInt)ERANGE;
	_namedLiterals[LN_LC_ALL] = (BppInt)LC_ALL;
	_namedLiterals[LN_LC_COLLATE] = (BppInt)LC_COLLATE;
	_namedLiterals[LN_LC_CTYPE] = (BppInt)LC_CTYPE;
	_namedLiterals[LN_LC_MONETARY] = (BppInt)LC_MONETARY;
	_namedLiterals[LN_LC_NUMERIC] = (BppInt)LC_NUMERIC;
	_namedLiterals[LN_LC_TIME] = (BppInt)LC_TIME;
	_namedLiterals[LN_SEEK_SET] = (BppInt)SEEK_SET;
	_namedLiterals[LN_SEEK_CUR] = (BppInt)SEEK_CUR;
	_namedLiterals[LN_SEEK_END] = (BppInt)SEEK_END;
	_namedLiterals[LN_true] = BppValue::makeBool(true);
	_namedLiterals[LN_false] = BppValue::makeBool(false);
	_namedLiterals[LN_UNICODE] =
#ifdef _UNICODE
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_WIN32] =
#ifdef WIN32
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_WINCE] =
#ifdef WINCE
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_MSDOS] =
#ifdef __MSDOS__
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_UNIX] =
#ifdef __unix__
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_LINUX] =
#ifdef __linux__
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_BPPI64] =
#ifdef BPP_I64
	(BppInt) 1;
#else
		(BppInt)0;
#endif
	_namedLiterals[LN_NPOS] = BPP_NPOS;
	_namedLiterals[LN_PTRSIZE] = (BppUInt)sizeof(BppIntPtr);
	_namedLiterals[LN_CHARSIZE] = (BppUInt)sizeof(BppChar);
	_namedLiterals[LN_WCHARSIZE] = (BppUInt)sizeof(wchar_t);
	_namedLiterals[LN_bplusplus] = (BppUInt)
#ifdef _BPP_VER
		_BPP_VER;
#else   
		120;
#endif
	_currentLine = -1;

}

// start symbol
// bppprog ::= { <function> | <classdeclaration> | <preprocinstruction> } .
bool BppCompiler::compile()
{
	C_LOCALE;
	initCompiler();   /* process statements until end of file */
	int errCnt = 0;
	BppToken tkn;
	bool isConst = false;
	bool keepConst = false;
	BppValueType savedrtype = BVT_NULL;
	BppClass* savedprclass = 0;
	while ((errCnt < MAXERRS) && ((tkn = _scanner->getToken(errCnt)) != T_EOF))
	{
		if (isConst && !keepConst)
			_error(CMP_UNEXPECTED_CONST, &errCnt);
		keepConst = false;
		BppValueType rtype = BVT_NULL;
		switch (tkn)
		{
		case T_ERROR:
			errCnt++;
			_print(_scanner->getTokenString());
			break;
		case T_CLASS:
		case T_NAMESPACE:
			if (!_doClass(errCnt,0,0,tkn == T_NAMESPACE)) return false;
			break;
		case T_CONST:
		{
			if (isConst)
				_error(CMP_UNEXPECTED_CONST, &errCnt);
			isConst = true;
			BppToken t = _scanner->getToken(errCnt);
			if (t != T_VAR) _scanner->back();
		}
		case T_VAR:
			if (!_fetchRequire(T__DFUNCTION,errCnt, T_IDENTIFIER))
			{
				errCnt++;
				_scanner->back();
				break;
			}
		case T_NIL:
			rtype = BVT_NIL;
		case T_VOID:
		case T_AUTO:
		case T_FUNCTION:
		case T_IDENTIFIER:
		{
			BppClass* prClass = 0;
			if (tkn != T_NIL) rtype = findVTypeFromName(T__DFUNCTION,_scanner->getTokenString(),errCnt, tkn == T_IDENTIFIER, &prClass);
			if (rtype != BVT_NULL && prClass) {
				// erster Identifikator bezeichnet eine Klasse
				// Bei der Bedeutung gibt es unterschiedliche Fälle:
				// - einfacher Klassenbezeichner
				//  - ... als Ergebnistyp eines Funktionsaufrufs --> Folgetoken ist Identifier
				//  - ... als globale Konvertierungsfunktion --> Folgetiken ist '('
				// - Anfang eines zusammengesetzten Bezeichners --> ganzen Bezeichner lesen
				//    - Klassenbezeichner als Ergebnistyp --> Bezeichner ist Klasse und Folgetoken ist '('
				//    - einfacher oder zusammengesetzter Klassenbezeichner mit folgendem Elementfunktionsnamen
				//         - speziell kann der Funktionsname 
				//           - ein Konstruktor 
				//           - oder Destruktor sein (dann folgt auf '::' ein '~')
				BppString name = _scanner->getTokenString();
				BppToken tnext = _scanner->getToken(errCnt);
				_scanner->back();
				getActiveScanner()->_stringBuffer = name.c_str();
				if (tnext == T_CC) { // zusammengesetzter Bezeichner
					BppVector parts;
					name = parseNamePathParts(tkn, errCnt, parts, true, true);
					BppDictionaryEntry* pe = _classes.find(name);
					if (pe) {
						// Pfad ergibt Klasse
						prClass = pe->getValue();
						tnext = _scanner->getToken(errCnt);
						BppChar ch = _scanner->getChar();
						_scanner->back();
						if (tnext == T_OPERATOR || (tnext == T_ANY && ch == _T('~'))) {
							// dtor
							// Klassenpfad in _scanner->stringBuffer
							getActiveScanner()->_stringBuffer = name.c_str();
							// Doppelpunkt zum Erkennen einer Member-Funktion
							getActiveScanner()->_token = T_CC;
							_scanner->back();
							rtype = BVT_NULL;
						}
						else {
							// Ergebnistyp
							// tnext muss funktionsname sein
							if (!_require(T__DFUNCTION,T_IDENTIFIER,tnext)) ++errCnt;
							if (savedrtype != BVT_NULL)
								_error(CMP_RETURNVAL_NOT_ALLOWED, &errCnt);
							savedrtype = rtype;
							savedprclass = prClass;
							keepConst = true;
							continue;
						}
					}
					else // vollständiger Pfad ergibt keine Klasse
					{
						const BppChar* pch = _tcsrchr(name.c_str(), _T(':'));
						if (pch) {
							// letzten Namensteil abspalten
							size_t idx = (--pch) - name.c_str();
							name.erase(idx);
						}
						pe = _classes.find(name);
						if (pe) {
							prClass = 0; // pe->getValue();
							rtype = BVT_NULL;
						}
						else 
							_error(BppString::format(_T("%s %s"), CMP_UNDEFINED_CLASS, name.c_str()).c_str(), &errCnt);
						const BppChar* lastname = ((BppString*)*parts.last())->c_str();
						getActiveScanner()->_token = _tcscmp(lastname,_T("operator")) == 0 ? T_OPERATOR : T_IDENTIFIER;
						getActiveScanner()->_stringBuffer = lastname;
						_scanner->back();
						getActiveScanner()->_stringBuffer = name.c_str();
						getActiveScanner()->_token = T_CC;
						_scanner->back();
					}
				}
				else {
					// einfacher Bezeichner vom Klassentyp
					if (tnext == T_ANY && _scanner->getChar() == _T('(')) {
						getActiveScanner()->_stringBuffer = name.c_str();
						rtype = BVT_NULL;
					}
					else {
						savedrtype = rtype;
						savedprclass = prClass;
						keepConst = true;
						continue;
					}
				}
			}
			else if (rtype != BVT_NULL)
			{
				BppString typeName = _scanner->getTokenString();
				BppToken t = _scanner->getToken(errCnt);
				if (t == T_ANY && _scanner->getChar() == _T('(')) {
  				    _scanner->back();
					// Typkonvertierungsfunktion
					getActiveScanner()->_stringBuffer = typeName.c_str();
				    rtype = BVT_NULL;
				}
			}

// tokenString in _scanner ist beim Aufruf der Name der Funktion oder Variable oder der Klassenpfad (bei Elementfunktionen)
// prClass ist ggf. die Klasse des Ergebnistyps
// Eine Elementfunktion wird daran erkannt, dass das nächstfolgende Token ein Doppelpunkt ist.
// Ist das nächstfolgende Token kein Doppelpunkt, muss es eines der Zeichen <tt>(,;=</tt> sein, dann wird eine Variable definiert.
// Im Falle einer Variablen enthält rtype den Variablentyp, sonst den Rückgabetyp der Funktion

			if (savedprclass) prClass = savedprclass;
			if (savedrtype != BVT_NULL) rtype = savedrtype;
			bool ok = _doFunctionOrVariable(isConst, errCnt, rtype, prClass);
			isConst = false;
			if (!ok) return false;
			savedprclass = 0;
			savedrtype = BVT_NULL;
		}
		break;
		default:
			_error(CMP_DECLARATION_EXPECTED, &errCnt);
		}
	}
	return (errCnt == 0);
}


// func ::= <regularfunc> | <globalvardef> | <memberfunction>
// tokenString in _scanner ist beim Aufruf der Name der Funktion oder der letzte Teil des Klassenpfades (bei Elementfunktionen)
// Eine Elementfunktion wird daran erkannt, dass das nächstfolgende Token ein Doppelpunkt ist.
// Bei einem Destruktor steht hinter dem Doppelpunkt noch ein '~'
// Ist das nächstfolgende Token kein Doppelpunkt, muss es eines der Zeichen <tt>(,;=</tt> sein, dann wird eine Variable definiert.
// Im Falle einer Variablen enthält rtype den Variablentyp, sonst den Rückgabetyp der Funktion
bool BppCompiler::_doFunctionOrVariable(bool isConst, int& errCnt, BppValueType rtype, BppClass* prClass)
{
	BppString name(_scanner->getTokenString());
	BppToken tkn = _scanner->getToken(errCnt);
	switch (tkn)
	{ 
	case T_ANY:
	{
		BppChar c = _scanner->getChar();
		if (rtype == BVT_NIL && c != _T('(')) _error(CMP_VAR_NULL_NOT_ALLOWED, &errCnt);
		switch (c)
		{
		case _T('('):
			if (rtype == BVT_AUTO)
			{
				_error(CMP_RETURN_AUTO_NOT_ALLOWED, &errCnt);
			}
			return _doRegularFunction(name, errCnt, isConst, rtype, prClass);
		case _T(','):
		case _T(';'):
			if (rtype == BVT_AUTO)
			{
				_error(CMP_MUST_INITIALIZE_AUTO, &errCnt);
			}
		case _T('='):
			_scanner->back();
			return _doGlobalVariable(name, isConst, errCnt, rtype, prClass);
		}
	}
	break;
	case T_CC:
		if (rtype == BVT_AUTO)
		{
			_error(CMP_RETURN_AUTO_NOT_ALLOWED, &errCnt);
		}
		//if (isConst)
		//	_error(CMP_UNEXPECTED_CONST, &errCnt);
		return _doMemberFunction(name, errCnt, isConst, rtype, prClass);
	case T_ERROR:
		_print(_scanner->getTokenString());
		break;
	default:



		break;
	}
	_error(CMP_DECLARATION_FUNC_OR_VAR_EXPEXTED, &errCnt);
	return true;
}

const BppChar* BppCompiler::_getOperatorFunc(int& errCnt, BppClass* ownerClass)
{
	BppToken tkn = _scanner->getToken(errCnt); // 'L', 'R' or operator symbol
	bool isRop = false;
	bool hasPrefix = false;
	bool incdec = false;
	bool subop = false;
	bool addop = false;
	bool mulop = false;
	const BppChar* result = 0;
	if (tkn == T_IDENTIFIER)
	{
		BppString namepath = parseNamePath(tkn, errCnt);
		//const BppChar* ps = _scanner->getTokenString();
		BppClass* pclass = 0;
		BppValueType casttype = findVTypeFromName(T__DFUNCTION,namepath, errCnt, true, &pclass, ownerClass);
		if (casttype == BVT_STRONGOBJECT)
			result = pclass->getClassName()->c_str();
		else if ((casttype != BVT_NULL) && (casttype != BVT_VOID))
			result = _vm.getTypeNameFromVType(casttype);
		switch (casttype)
		{
		case BVT_INT:
		case BVT_UINT:
		case BVT_SHORT:
		case BVT_USHORT:
		case BVT_LONG:
		case BVT_ULONG:
		case BVT_SBYTE:
		case BVT_BYTE:
		case BVT_BOOL:
		case BVT_CHAR:
		case BVT_UCHAR:
		case BVT_CHAR8:
		case BVT_CHAR16:
		case BVT_CHAR32:
		case BVT_WCHAR:
		case BVT_INTPTR:
		case BVT_UINTPTR:
		case BVT_FLOAT:
		case BVT_DOUBLE:
		case BVT_BUFFER:
		case BVT_CHARBUFFER:
		case BVT_VECTOR:
		case BVT_STRING:
		case BVT_DICTIONARY:
		case BVT_FILE:
		case BVT_STRONGOBJECT:
			//    case BVT_OBJECT: 
			break;
		default:
			result = 0;
			{
				hasPrefix = true;
				if (_tcscmp(namepath.c_str(), _T("R")) == 0)
					isRop = true;
				else if (_tcscmp(namepath.c_str(), _T("L")) != 0)
				{
					_error(CMP_INVALID_OPERATOR_PREFIX, &errCnt);
				}
				tkn = _scanner->getToken(errCnt);
			}
		}
	}
	if (hasPrefix)
	{
		switch (tkn)
		{
		case T_SHL:
		case T_SHR:
			break;
		case T_ANY:
			if (_tcschr(_T("+-*/%&|^"), _scanner->getChar()) != 0)
				break;
		default:
			_error(CMP_INVALID_OPERATOR_BEHIND_PREFIX, &errCnt);
		}
	}
	switch (tkn)
	{
	case T_IDENTIFIER:
		if (result != 0)
			break;
	case T_SHL:
		result = isRop ? _T("OP_SHL_R") : _T("OP_SHL");
		break;
	case T_SHR:
		result = isRop ? _T("OP_SHR_R") : _T("OP_SHR");
		break;
	case T_INC:
		result = _T("OP_INC");
		incdec = true;
		break;
	case T_DEC:
		result = _T("OP_DEC");
		incdec = true;
		break;
	case T_ANY:
	{
		BppChar c = _scanner->getChar();
		switch (c)
		{
		case _T('+'):
			result = isRop ? _T("OP_ADD_R") : _T("OP_ADD");
			if (!hasPrefix) addop = true;
			break;
		case _T('-'):
			result = isRop ? _T("OP_SUB_R") : _T("OP_SUB");
			if (!hasPrefix) subop = true;
			break;
		case _T('*'):
			result = isRop ? _T("OP_MUL_R") : _T("OP_MUL");
			if (!hasPrefix) mulop = true;
			break;
		case _T('/'):
			result = isRop ? _T("OP_DIV_R") : _T("OP_DIV");
			break;
		case _T('%'):
			result = isRop ? _T("OP_REM_R") : _T("OP_REM");
			break;
		case _T('&'):
			result = isRop ? _T("OP_BAND_R") : _T("OP_BAND");
			break;
		case _T('|'):
			result = isRop ? _T("OP_BOR_R") : _T("OP_BOR");
			break;
		case _T('^'):
			result = isRop ? _T("OP_XOR_R") : _T("OP_XOR");
			break;
		case _T('~'):
			result = _T("OP_BNOT");
			break;
		case _T('!'):
			result = _T("OP_NOT");
			break;
		case _T('('):
			result = _T("OP_CALL");
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')')))
				++errCnt;
			break;
		case _T('['):
		{
			BppToken t = _scanner->getToken(errCnt);
			if ((t == T_VAR) || (t == T_IDENTIFIER && _tcscmp(_scanner->getTokenString(), _T("int")) == 0))
			{
				result = _T("OP_VSET");
				t = _scanner->getToken(errCnt);
			}
			else
				result = _T("OP_VREF");
			if (!_require(T__DFUNCTION, T_ANY, t, _T(']')))
				++errCnt;
			t = _scanner->getToken(errCnt);
			if ((t == T_ANY) && (_scanner->getChar() == _T('=')))
			{
				result = _T("OP_VSET");
			}
			else
			{
				_scanner->back();
			}
		}
		break;
		case _T('.'):
		{
			BppToken t = _scanner->getToken(errCnt);
			if ((t == T_ANY) && (_scanner->getChar() == _T('=')))
			{
				result = _T("OP_PSET");
			}
			else
			{
				result = _T("OP_PREF");
				_scanner->back();
			}
		}
		break;
		default:
			_error(CMP_INVALID_OPERATOR_DEFINITION, &errCnt);
		}
	}
	break;
	default:
		_error(CMP_INVALID_OPERATOR_DEFINITION, &errCnt);
	}
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('(')))
		++errCnt;
	if (addop || subop || mulop)
	{
		tkn = _scanner->getToken(errCnt);
		if ((tkn == T_ANY) && (_scanner->getChar() == _T(')')))
			result = subop ? _T("OP_NEG") : addop ? _T("OP_PLUS") : _T("OP_UNREF");
		_scanner->back();
	}
	else if (incdec)
	{
		BppToken t = _scanner->getToken(errCnt);
		if ((t == T_VAR) || (t == T_IDENTIFIER && _tcscmp(_scanner->getTokenString(), TN_int) == 0))
		{
			if (tkn == T_INC)
				result = _T("OP_PINC");
			else
				result = _T("OP_PDEC");
		}
		else
			_scanner->back();
	}
	return result;
};

// classdeclaration ::= "class" <identifier> (";" | ( [ <baseclassdecl> ] "{" { <preprocinstruction> | <memberdeclaration> } "}" [";"]) )
// baseclassdecl ::= ":" <identifier>
// memberdeclaration ::= <staticvarmemberdecl> | <varmemberdecl> | <memberfuncdecl>
// staticvarmemberdecl ::= "static" ["var" | <typeident>] <identifier> ["=" <literal>] { ","<identifier> ["=" <literal>] } ";"
// staticconstmemberdecl ::= ["static"] "const" ["var"| <typeident>] <identifier> ["=" <literal>] { ","<identifier> ["=" <literal>] } ";"
// varmemberdecl ::= ["var" | <typeident>] <identifier> { "," <identifier> } ";"
// memberfuncdecl ::= [ "static" ] ["var" | "null" | <typeident>] <identifier> "(" [ <argumentlist ] [ ";" <localvarlist> ] ")" ";"
// Rules:
//   - Constructor is a member function of same name as class.
//   - Destructor is a member function without any parameters and defined as ~<classname>() or dtor().
//   - Constructor and destructor can't be static.
//   - Redefinition of base class is forbidden.
//   - In redeclaration of an already existing class adding new non-static data members is forbidden.
//   - Member variables already defined as non-static can't be redefined to be static.
//   - Declaration of non-static member functions is optional and for better readability only. Any arguments
//     specified in their declaration are ignored.
//   - Static data members may be initialized in declaration by any literal value,
//   - If static modifier is specified on start of a variable list, all elements of list are static
bool BppCompiler::_doClass(int& errCnt, BppClass* ownerClass, unsigned short protectionLevel, bool isnamespace)
{
	/* get the class name */
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
	{
		errCnt++;
		return false;
	}
	BppString className = _scanner->getTokenString();
	BppString classPath = className;
	if (ownerClass) classPath =
		BppString::format(_T("%s::%s"), ownerClass->getClassPath().c_str(), className.c_str());
	BppDictionaryEntry* clsEntry = _classes.find(classPath);
	BppClass* pclass = (clsEntry != 0) ? (BppClass*)clsEntry->getValue() : 0;
	const BppChar* cltype = isnamespace ? _T("namespace") : _T("class");

	BppToken tkn = _scanner->getToken(errCnt);
	if (tkn == T_ERROR)
		_print(_scanner->getTokenString());
	// check for forward declaration
	bool isForward = (tkn == T_ANY) && (_scanner->getChar() == _T(';'));

	BppClass* pBaseClass = 0;
	if (!isForward)
	{
		if (!isnamespace && (tkn == T_ANY) && (_scanner->getChar() == _T(':'))) // base class
		{
			BppString basePath;
			tkn = _scanner->getToken(errCnt);
			if (tkn == T_CC) {
				basePath = _scanner->getTokenString();
				tkn = _scanner->getToken(errCnt);
			}
			if (!_require(T__DFUNCTION, T_IDENTIFIER, tkn)) {
				errCnt++;
				return false;
			}
			basePath += parseNamePath(tkn, errCnt);
			pBaseClass = _findClass(basePath, errCnt, ownerClass);
			if (pBaseClass != 0)
			{
				if (pBaseClass == pclass)
				{
					_error(CMP_CLASS_CANT_BE_OWN_BASE, &errCnt);
					return false;
				}
				if (_showInfo)
					_print(_T("%s %s : %s\n"), cltype, classPath.c_str(), pBaseClass->getClassPath().c_str());
			}


		}// end if baseclass
		else
		{
			_scanner->back();
			if (_showInfo)
				_print(_T("%s %s\n"), cltype, classPath.c_str());
		}
	} // end !isForward
	else
	{
		if (_showInfo)
			_print(_T("%s %s forward declaration\n"), cltype, className.c_str());
	}

	// Flag if class is already used as base of another one. If so, additional functions but not
	// any new member variables may be defined.
	bool isBaseClass = false;
	// If class is already defined we skip its registration in _classes dictionary.
	if (!clsEntry) // class was not already defined
	{
		BppString* pname = new BppString(className);
		clsEntry = &_classes[classPath];
		clsEntry->setSymbolType(ST_CLASS);
		pclass = new BppClass(_vm, pname);
		clsEntry->setValue(pclass);
		pclass->setOwnerClass(ownerClass);
		((BppValue&)(clsEntry->getValue())).setProtection(protectionLevel && !isnamespace ? protectionLevel : BVF_PUBLIC);
	}
	else // check if class is already base of another one
	{
		for (BppDictionaryEntry* pe = _classes.getFirst(); pe != 0; pe = pe->getNext())
		{
			BppClass* pc = pe->getValue();
			if ((pc != 0) && (pc->getBaseClass() == pclass))
			{
				isBaseClass = true;
				break;
			}
		}
	}

	if (pBaseClass) // set base class if specified
	{
		BppClass* pcurrbase = pclass->getBaseClass();
		if (!pcurrbase)
			pclass->setBaseClass(pBaseClass);
		else if (pcurrbase != pBaseClass)
		{
			// Note: It's not an error to define same (but only same) base class once more!
			_error(CMP_ILLEGAL_REDEFINITION_OF_BASE, &errCnt);
		}
	}
	if (isForward)
	{
		pclass->computeDataSize();
		return true;
	}

	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('{')))
	{
		errCnt++;
		return false;
	}
	unsigned short protlevel = 0;
	/* handle each variable declaration */
	BppChar tknChar = 0;
	bool isConst = false;
	bool isMutable = false;
	bool preserveConst = false;
	bool preserveMutable = false;
	while (!(((tkn = _scanner->getToken(errCnt)) == T_ANY) && ((tknChar = _scanner->getChar()) == _T('}'))))
	{
		if (tkn == T_CONST) {
			if (isConst)
				_error(CMP_UNEXPECTED_CONST, &errCnt);
			isConst = true;
			preserveConst = true;
			continue;
		}
		if (tkn == T_MUTABLE) {
			if (isMutable || isnamespace)
				_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
			isMutable = true;
			preserveMutable = true;
			continue;
		}
		if (!preserveConst)
			isConst = false;
		if (!preserveMutable)
			isMutable = false;
		if (isConst && isMutable) {
			_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
		}
		preserveConst = false;
		preserveMutable = false;
		bool varPrefix = false;
		bool nullPrefix = false;
		bool isDtor = false;
		BppToken type = tkn;
		switch (type) {
		case T_ERROR:
			_print(_scanner->getTokenString());
			break;
		case T_EOF:
			_error(CMP_UNEXPECTED_EOF, &errCnt);
			return false;
			break;
		case T_NAMESPACE:
			if (!isnamespace)
				_error(CMP_ILLEGAL_NAMESPACE_IN_CLASS, &errCnt);
		case T_CLASS:
			if (isConst)
				_error(CMP_UNEXPECTED_CONST, &errCnt);
			if (isMutable)
				_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
			_doClass(errCnt, pclass, protlevel ? protlevel : BVF_PUBLIC, type == T_NAMESPACE);
			tkn = _scanner->getToken(errCnt);
			if (!(tkn == T_ANY && _scanner->getChar() == _T(';')))
				_scanner->back();
			continue;
			/* check for static members */
		case T_STATIC:
		  if (isConst)
			  _error(CMP_UNEXPECTED_CONST,&errCnt);
		  if (isMutable)
			  _error(CMP_UNEXPECTED_MUTABLE, &errCnt);
		  tkn = _scanner->getToken(errCnt);
			if (tkn == T_CONST) {
				tkn = _scanner->getToken(errCnt);
				isConst = true;
			}
			varPrefix = tkn == T_VAR;
			break;
		case T_VAR:
			varPrefix = true;
			break;
		case T_PUBLIC:
		case T_PROTECTED:
		case T_PRIVATE:
			if (isnamespace)
				_error(CMP_UNEXPECTED_PROTECTION_SPECIFIER, &errCnt);
			if (isConst)
				_error(CMP_UNEXPECTED_CONST, &errCnt);
			if (isMutable)
				_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
			protlevel = type == T_PUBLIC ? BVF_PUBLIC : type == T_PROTECTED ? BVF_PROTECTED : BVF_PRIVATE;
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(':'))) ++errCnt;
			continue;
		case T_NIL:
			nullPrefix = true;
			break;
		case T_ANY:
			isDtor = tknChar == _T('~');
			break;
		}
		BppValueType vtype = BVT_NULL;
		BppClass* pvClass = 0;
		// skip next token for var, null and dtor, otherwise check for type modifier
		if (isDtor || varPrefix || nullPrefix)
			tkn = _scanner->getToken(errCnt);
		else
		{
			bool globaltype = tkn == T_CC;
			if (globaltype) {
				tkn = _scanner->getToken(errCnt);
				if (!_require(T__DFUNCTION, T_IDENTIFIER, tkn)) ++errCnt;
			}
			if (tkn == T_IDENTIFIER || tkn == T_VOID || tkn == T_FUNCTION || tkn == T_AUTO)
			{
				BppString name = parseNamePath(tkn, errCnt);
				vtype = findVTypeFromName(T__DFUNCTION,name, errCnt, true, &pvClass, globaltype ? 0 : pclass);
				if (vtype != BVT_NULL)
				{
					BppString typeName = name;
					BppToken t = tkn;
					tkn = _scanner->getToken(errCnt);
					if (tkn == T_IDENTIFIER || tkn == T_OPERATOR)
						varPrefix = true;
					else
					{
						_scanner->back();
						getActiveScanner()->_token = t;
						tkn = t;
						getActiveScanner()->_stringBuffer = typeName.c_str();
						vtype = BVT_NULL;
						varPrefix = false;
					}
				}
			}
		}

		/* check for destructor or optionally var or null keyword */
		/* get the first identifier */
		bool opdef = false;
		bool funcdef = false;
		if (tkn == T_OPERATOR)
		{
			if (type == T_STATIC || isnamespace)
			{
				_error(CMP_STATIC_OPERATORS_NOT_ALLOWED, &errCnt);
			}
			if (isMutable)
				_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
			opdef = true;
		}
		else if (tkn != T_IDENTIFIER)
		{
			if (tkn == T_ERROR)
				_print(_scanner->getTokenString());
			_error(CMP_MEMBER_DECLARATION_EXPECTED, &errCnt);
			continue;
		}
		BppString id = opdef ? _getOperatorFunc(errCnt) : _scanner->getTokenString();
		if (opdef && errCnt == 0)
		{
			BppClass* prClass = 0;
			BppValueType rtype = findVTypeFromName(T__DFUNCTION, id, errCnt, true, &prClass, pclass);
			if (vtype != BVT_NULL && rtype != BVT_NULL)
			{
				_error(CMP_ILLEGAL_TYPE_BEFORE_CONVOP, &errCnt);
			}
			else if (rtype != BVT_NULL) { vtype = rtype; pvClass = prClass; }
		}
		bool isCtor = false;
		if (isDtor)
		{
			if (id.compare(className) != 0)
			{
				_error(CMP_DTOR_DECLARATION_EXPECTED, &errCnt);
			}
			else
				id = SYMN_dtor;
		}
		else
		{
			if (id.compare(className) == 0)
				isCtor = true;
			else if (id.compare(SYMN_dtor) == 0)
				isDtor = true;
		}
		if (type == T_STATIC || isnamespace)
		{
			if (isMutable)
				_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
			if (isDtor)
			{
				_error(CMP_DTOR_CANT_BE_STATIC, &errCnt);
			}
			else if (isCtor)
			{
				_error(CMP_CTOR_CANT_BE_STATIC, &errCnt);
			}
		}
		else if (protlevel && isDtor && protlevel != BVF_PUBLIC)
		{
			_error(CMP_DTOR_MUST_BE_PUBLIC, &errCnt);
		}
		if (varPrefix || nullPrefix)
		{
			if (isDtor)
			{
				_error(CMP_DTOR_CANT_RETURN_VALUE, &errCnt);
			}
			else if (isCtor)
			{
				_error(CMP_CTOR_CANT_DEFINE_RETURNTYPE, &errCnt);
			}
		}
		/* check for a member function declaration */
		if (!opdef) // open bracket already read by _getOperatorFunc
		{
			tkn = _scanner->getToken(errCnt);
			if (tkn == T_ERROR)
				_print(_scanner->getTokenString());
		}
		if (opdef || ((tkn == T_ANY) && (_scanner->getChar() == _T('('))))
		{
			if (isMutable)
				_error(CMP_UNEXPECTED_MUTABLE, &errCnt);
			BppDictionaryEntry& entry = pclass->funcMembers()[id];
			entry.setSymbolType(isnamespace || (type == T_STATIC) ? ST_SFUNCTION : ST_FUNCTION);
			((BppValue&)entry.getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
			getActiveScanner()->_funcName = BppString::format(_T("%s::%s"), pclass->getClassPath().c_str(), id.c_str());
			if (_showInfo)
				_print(_T("member function %s\n"), _scanner->_funcName.c_str());
			/* compile the body of the function */
			funcdef = true;
			//if (isConst)
			//	_error(CMP_UNEXPECTED_CONST, &errCnt);
			BppVector* code = _codeFunctionBody(id, errCnt, isConst, pclass, vtype, pvClass);
			getActiveScanner()->_funcName = _T("");
			if (code)
			{
				/* store the new method */
				entry.setValue(BppValue(code, true));
				((BppValue&)entry.getValue()).setProtection(protlevel && !isnamespace ? protlevel : BVF_PUBLIC);
				/* build OP_PINC / OP_PDEC methods if needed */
				const BppString& key = *(entry.getKey());
				BppString* selector = 0;
				if (key.compare(_T("OP_INC")) == 0)
					selector = new BppString(_T("OP_PINC"));
				else if (key.compare(_T("OP_DEC")) == 0)
					selector = new BppString(_T("OP_PDEC"));
				if (selector != 0)
				{
					selector->addRef();
					BppDictionaryEntry* pentry = pclass->findMemberPL(BVF_PROTECTED, *selector, true);
					if (pentry == 0) // add post increment/decrement method if it not exists
					{
						pentry = pclass->funcMembers()[selector];
						pentry->setSymbolType(ST_FUNCTION);
						pentry->setValue(BppValue(code, true));
						((BppValue&)pentry->getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
					}
					selector->release();
				}
			}
		}
		/* handle data members */
		else if (!isnamespace && type != T_STATIC && !isConst)
		{
			if (nullPrefix)
			{
				_error(CMP_UNEXPECTED_NULL_BEFORE_MEMBER, &errCnt);
			}
			if (isBaseClass)
			{
				_error(CMP_CLASS_MEMBERDEF_NOT_ALLOWED, &errCnt);
			}
			while (1)
			{
				BppDictionaryEntry tmpEntry(id, BppValue());
				bool initialized = false;
				if ((tkn == T_ANY) && (_scanner->getChar() == _T('=')))
				{
					tkn = _scanner->getToken(errCnt);
					if (tkn == T_ERROR)
						_print(_scanner->getTokenString());
					initialized = _doLitExpr(tkn, tmpEntry, errCnt);
					tkn = _scanner->getToken(errCnt);
					if (tkn == T_ERROR)
						_print(_scanner->getTokenString());
				}
				BppValue v(vtype, vtype != BVT_NULL ? BVF_STRONGTYPED : BVF_NONE, initialized ? false : true);
				if (vtype == BVT_STRONGOBJECT && pvClass)
				{
					BppVector* pv = new BppVector(2, true);
					(*pv)[IDX_CLASS] = pvClass;
					BppValue vdesc((BppRefObject*)pv, vtype);
					v = vdesc;
				}
				if (initialized && (!v.assign(BppValue(tmpEntry.getValue(), true))))
				{
					_error(BppString::format(CMP_TYPE_MISMATCH_FMT3,
						id.c_str(),
						getValueTypeName(v.getType()),
						getValueTypeName(tmpEntry.getValue().getType())).c_str(), &errCnt);
				}
				if (vtype == BVT_AUTO && (!initialized))
				{
					_error(CMP_MUST_INITIALIZE_AUTO, &errCnt);
				}
				v.setProtection(protlevel ? protlevel : BVF_PROTECTED);

				if (isMutable)
					v.setFlags(v.getFlags() | BVF_MUTABLE);
				if (!pclass->addDataMember(id.c_str(), v))
				{

					if (!pclass->defineDataMember(id.c_str(), v))
					{
						_error(CMP_ILLEGAL_REDEF_TO_DATA_MEMBER, &errCnt);
						errCnt++;
					}
					else _error(CMP_WARN_REDEF_DATA_MEMBER);
				}
				if ((tkn == T_ANY) && (_scanner->getChar() == _T(',')))
				{
					if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER)) errCnt++;
					else
						id = _scanner->getTokenString();
					tkn = _scanner->getToken(errCnt);
				}
				else
				{
					_scanner->back();
					break;
				}
			} // end while
			pclass->fixupDataDefs();
		}
		else // static datamember(s)
		{
			if (nullPrefix)
			{
				_error(CMP_UNEXPECTED_NULL_BEFORE_CLASSVAR, &errCnt);
			}
			while (1)
			{
				BppDictionaryEntry tmpEntry(id, BppValue());
				bool initialized = false;
				if ((tkn == T_ANY) && (_scanner->getChar() == _T('=')))
				{
					tkn = _scanner->getToken(errCnt);
					if (tkn == T_ERROR)
						_print(_scanner->getTokenString());
					initialized = _doLitExpr(tkn, tmpEntry, errCnt);
					tkn = _scanner->getToken(errCnt);
					if (tkn == T_ERROR)
						_print(_scanner->getTokenString());
				}
				else if (isConst)
					_error(CMP_MUST_INITIALIZE_CONST, &errCnt);
				BppValue v(vtype, vtype != BVT_NULL ? BVF_STRONGTYPED : BVF_NONE, initialized ? false : true);
				if (vtype == BVT_STRONGOBJECT && pvClass)
				{
					BppVector* pv = new BppVector(2, true);
					(*pv)[IDX_CLASS] = pvClass;
					BppValue vdesc((BppRefObject*)pv, vtype);
					v = vdesc;
				}
				if (initialized && (!v.assign(BppValue(tmpEntry.getValue(), true))))
				{
					_error(BppString::format(CMP_TYPE_MISMATCH_FMT3,
						id.c_str(),
						getValueTypeName(v.getType()),
						getValueTypeName(tmpEntry.getValue().getType())).c_str(), &errCnt);
				}
				if (vtype == BVT_AUTO && (!initialized))
				{
					_error(CMP_MUST_INITIALIZE_AUTO, &errCnt);
				}
				v.setProtection(!isnamespace && protlevel ? protlevel : BVF_PUBLIC);

					if (isConst)
						v.setFlags(v.getFlags() | BVF_CONST);
				if (!pclass->addStaticDataMember(id.c_str(), v))
				{
					if (!pclass->defineStaticDataMember(id.c_str(), v))
					{
						_error(CMP_ILLEGAL_REDEF_TO_STATIC_DATA_MEMBER, &errCnt);
						errCnt++;
					}
					else _error(CMP_WARN_REDEF_STATIC_DATA_MEMBER, &errCnt);
				}
				if ((tkn == T_ANY) && (_scanner->getChar() == _T(',')))
				{
					if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER)) errCnt++;
					else
						id = _scanner->getTokenString();
					tkn = _scanner->getToken(errCnt);
				}
				else
				{
					_scanner->back();
					break;
				}
			} // end while
			isConst = false;
		}
		if (!funcdef)
		{
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		}
	} // end while member
	tkn = _scanner->getToken(errCnt);
	if ((tkn != T_ANY) || (_scanner->getChar() != _T(';')))
		_scanner->back();

	pclass->fixupDataDefs();
	return true;
}


BppScanner* BppCompiler::getActiveScanner() const
{
	BppScanner* pscan = _scanner;
	if (pscan) {
		while (pscan->_next) pscan = pscan->_next;
	}
	return pscan;
}




// globalvardef ::= <singlevardef> {"," <singlevardef> } ";". (vtype)
// singlevardef = <identifier> [ "=" <litexpr> ]
bool BppCompiler::_doGlobalVariable(const BppString& name, bool isConst, int& errCnt, BppValueType vtype, BppClass* pvClass)
{
	BppDictionaryEntry* entry = _symbols.find(name);
	if (entry)
		_error(BppString::format(CMP_ILLEGAL_GLOBAL_REDEF_FMT, name.c_str()).c_str(), &errCnt);
	else
	{
		entry = &_symbols[name];
		entry->setSymbolType(ST_SDATA);
	}
	if (vtype == BVT_STRONGOBJECT && pvClass)
	{
		BppVector* desc = new BppVector(2, true);
		(*desc)[IDX_CLASS] = pvClass;
		entry->setValue(BppValue((BppRefObject*)desc, vtype));
	}
	else entry->setValue(BppValue(vtype, BVF_NONE, true));
	if (vtype != BVT_NULL) ((BppValue&)entry->getValue()).setFlags(BVF_STRONGTYPED);
	while (_fetchRequire(T__DFUNCTION, errCnt, T_ANY))
	{
		BppChar c = _scanner->getChar();
		switch (c)
		{
		case _T(';'):
			return (errCnt == 0);
		case _T(','):
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
			{
				errCnt++;
				return false;
			}
			return _doGlobalVariable(_scanner->getTokenString(), isConst, errCnt, vtype, pvClass);
		case _T('='):
		case _T('('):
		{
			BppDictionaryEntry dummy(0, BppValue());
			if (_doLitExpr(_scanner->getToken(errCnt), dummy, errCnt)) {
				BppValue v(dummy.getValue(), dummy.getValue().isRefCounted());
				if (!entry->setValue(v, true))
				{
					_error(BppString::format(CMP_TYPE_MISMATCH_FMT3,
						name.c_str(),
						getValueTypeName(entry->getValue().getType()),
						getValueTypeName(dummy.getValue().getType())).c_str(), &errCnt);
				}
				if (isConst) {
					BppValue& vref = (BppValue&) entry->getValue();
					vref.setFlags(vref.getFlags() | BVF_CONST);
				}
			}
			if (c == _T('(') && (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))))
			{
				++errCnt;
				return false;
			}
		}
		break;
		default:
			_require(T__DFUNCTION, T_ANY, T_ANY, _T(';'));
			++errCnt;
			return false;
		}
	}
	errCnt++;
	return false;
}

// regularfunc ::= <identifier> <functionbody>
bool BppCompiler::_doRegularFunction(const BppString& name, int& errCnt, bool constResult,BppValueType rtype, BppClass* prClass)
{
	BppDictionaryEntry* sym = _symbols.find(name);
	if (sym)
	{
		if (sym->getSymbolType() != ST_SFUNCTION)
		{
			_error(CMP_ILLEGAL_REDEF_TO_FUNCTION, &errCnt);
			return false;
		}
	}
	else
	{
		sym = &_symbols[name];
		sym->setSymbolType(ST_SFUNCTION);
	}
	getActiveScanner()->_funcName = name;

	if (_showInfo)
		_print(_T("function %s\n"), name.c_str());

	/* compile the body of the function */
	BppVector* code = _codeFunctionBody(name, errCnt, constResult, 0, rtype, prClass);
	getActiveScanner()->_funcName.clear();
	if (code != 0)
		/* store the function definition */
		*sym = BppValue(code, true);
	return true;
}

// memberfunc ::= <classname> "::" <identifier> <functionbody>
/**
@param name Name of class
*/
bool BppCompiler::_doMemberFunction(const BppString& name, int& errCnt, bool constResult, BppValueType rtype, BppClass* prClass)
{
	BppString* selector = 0;
	BppClass* pclass = _findClass(name, errCnt/*,prClass*/);
	if (pclass == 0) return false;
	bool opdef = false;
	/* get the selector */
	BppToken tkn = _scanner->getToken(errCnt);
	if ((tkn == T_ANY) && (_scanner->getChar() == _T('~')))
	{
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
			errCnt++;
		else if (name != *(pclass->getClassName()))
		{
			_error(CMP_ILLEGAL_DTOR_DEF, &errCnt);
		}
		selector = new BppString(SYMN_dtor);
		selector->addRef();
	}
	else
	{
		if (tkn == T_OPERATOR)
			opdef = true;
		if (!(opdef || _require(T__DFUNCTION, T_IDENTIFIER, tkn)))
			errCnt++;
		else
		{
			selector = new BppString(opdef ? _getOperatorFunc(errCnt,pclass) : _scanner->getTokenString());
			selector->addRef();
			if (opdef && errCnt == 0)
			{
				BppClass* pvClass = 0;
				BppValueType vtype = findVTypeFromName(T__DFUNCTION ,*selector,errCnt, true, &pvClass);
				if (vtype != BVT_NULL && rtype != BVT_NULL)
				{
					_error(CMP_ILLEGAL_TYPE_BEFORE_CONVOP, &errCnt);
				}
				else if (vtype != BVT_NULL) { rtype = vtype; prClass = pvClass; }
			}
		}
	}
	if (selector == 0)
		return false;
	if (!(opdef || _fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('(')))) ++errCnt;
	if (_showInfo)
		_print(_T("member function %s::%s\n"), pclass->getClassName()->c_str(), selector->c_str());

	BppDictionaryEntry* entry = pclass->findMemberPL(BVF_PRIVATE, *selector, false);
	/* make sure the type matches the declaration */
	if (entry != NULL)
	{
		BppSymbolType st = entry->getSymbolType();
		if ((st != ST_FUNCTION) && (st != ST_SFUNCTION))
		{
			_error(CMP_ILLEGAL_REDEFINITION, &errCnt);
		}
	}
	else // new entry
	{
		entry = pclass->funcMembers()[selector];
		entry->setSymbolType(ST_FUNCTION);
	}
	getActiveScanner()->_funcName = BppString::format(_T("%s::%s"), pclass->getClassPath().c_str(), selector->c_str());
	/* compile the body of the function */
	BppVector* code = _codeFunctionBody(*selector, errCnt, constResult, pclass, rtype, prClass);
	selector->release();
	getActiveScanner()->_funcName.clear();
	//if (code == 0) return false;


	if (code)
	{
		/* store the new method */
		unsigned short protlevel = entry->getValue().getProtection();
		entry->setValue(BppValue(code, true));
		((BppValue&)entry->getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
		/* build OP_PINC / OP_PDEC methods if needed */
		const BppString& key = *(entry->getKey());
		selector = 0;
		if (key.compare(_T("OP_INC")) == 0)
			selector = new BppString(_T("OP_PINC"));
		else if (key.compare(_T("OP_DEC")) == 0)
			selector = new BppString(_T("OP_PDEC"));
		if (selector != 0)
		{
			protlevel = entry->getValue().getProtection();
			selector->addRef();
			BppDictionaryEntry* pentry = pclass->findMemberPL(BVF_PRIVATE, *selector, false);
			if (pentry == 0) // add post increment/decrement method if it not exists
			{
				pentry = pclass->funcMembers()[selector];
				pentry->setSymbolType(ST_FUNCTION);
				pentry->setValue(BppValue(code, true));
				((BppValue&)pentry->getValue()).setProtection(protlevel ? protlevel : BVF_PUBLIC);
			}
			selector->release();
		}
	}
	return true;
}



/**
<pre>
 functionbody ::= "(" [<argumentlist>][ ";" <localvarlist> ] ")" <blockstatement>
 argumentlist ::= <idlist>
 localvarlist ::= <idlist>
</pre>
*/
BppVector* BppCompiler::_codeFunctionBody(const BppString& name, int& errCnt, bool constResult, BppClass* ownerClass, BppValueType rtype, BppClass* prClass)
{
	BppFuncContext context;
	context.returnType = rtype;
	context.prClass = prClass;
	context.contextClass = ownerClass;
	context.returnConst = constResult;
	if (constResult && rtype == BVT_VOID)
		_error(CMP_VOID_CONST, &errCnt);
	BppVector* defargs = new BppVector();
	bool isdtor = false;
	/* add the implicit 'this' argument for member functions */
	if (ownerClass != NULL)
	{
		BppDictionaryEntry* pe = ownerClass->findMemberPL(BVF_PRIVATE, name, false);
		if (!(pe && (pe->getSymbolType() == ST_SFUNCTION)))
		{
			isdtor = (_tcscmp(name.c_str(), SYMN_dtor) == 0);
			context.arguments.push_back(new BppString(SYMN_this));
			defargs->push_back(BppValue());
		}
	}
	/* get the argument list */
	//BppValue dummy = defargs;
	_getIdList(context.arguments, ownerClass,_T(";)"), errCnt, defargs);
	if (isdtor && context.arguments.size() > 1)
	{
		_error(CMP_DTOR_CANT_HAVE_ARGS, &errCnt);
	}
	/* get temporary variables, at least 1 for exception value needed */
	int tcnt = 1;
	BppToken tkn = _scanner->getToken(errCnt);
	if ((tkn == T_ANY) && (_scanner->getChar() == _T(';')))
	{
		// add 1 to temporary count for exception value)
		tcnt = _getIdList(context.temporaries,ownerClass, _T(")"), errCnt) + 1;
		tkn = _scanner->getToken(errCnt);
	}
	if (!_require(T__DFUNCTION, T_ANY, tkn, _T(')')))
	{
		errCnt++;
		return 0;
	}
	tkn = _scanner->getToken(errCnt);
	if (tkn == T_CONST) {
		if (ownerClass) context.constThis = true;
		else
			_error(CMP_CONST_NONMEMBERFUNC, &errCnt);
		tkn = _scanner->getToken(errCnt);
	}
	if (tkn == T_ANY && _scanner->getChar() == _T(';')) // forward deklaration
	{
		return 0;
	}
	if (context.constThis) {
		context.arguments[0].setFlags(BVF_CONST);
		(*defargs)[0].setFlags(BVF_CONST);
	}

	_scanner->back();
	context.codeBuffer = new BppBuffer();
	context.codeVec = new BppVector(16);
	BppVector* result = context.codeVec;
	// store the bytecodes, class and function name as the first literals 
	result->push_back(BppValue(context.codeBuffer)); // buffer for bytecodes
	if (ownerClass != 0)
		result->push_back(BppValue(ownerClass));        // class (may be NULL!)
	else
		result->push_back(BppValue());
	result->push_back(BppValue()); // empty value for function state
//	context.contextClass = ownerClass;
	result->push_back(new BppString(name)); // function name
	result->push_back(defargs); // vector of default argument values
	result->push_back(BppValue(result, true, true)); // self reference
	if (tcnt > 0) // reserve space for temporaries
	{
		_codeIndexedOpcode(*result, OP_TSPACE, (BppNWord)tcnt, true);
	}


	context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('{'))) errCnt++;
	// compile statement block
	_codeBlock(context, errCnt);

	if (ownerClass != 0)
	{
		if (_tcscmp(ownerClass->getClassName()->c_str(), name.c_str()) == 0)
		{
			//       putcbyte(OP_PUSH);
			BppVarInfo varInfo = _findVariable(SYMN_this, context, false, errCnt);
			// return 'this' pointer on TOS in constructor
			_codeRValue(*result, varInfo);
		}
		else if ((ownerClass->getBaseClass() != 0) && (_tcscmp(SYMN_dtor, name.c_str()) == 0))
		{
			// push 'this' pointer
			context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
			BppVarInfo varInfo = _findVariable(SYMN_this, context, false, errCnt);
			_codeRValue(*result, varInfo);
			// push call to 'BC_dtor' (magic name of inherited dtor)
			context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
			result->push_back(new BppString(_T("BC_dtor")));
			_codeLiteral(*result, result->size() - 1);
			context.codeBuffer->push_back(BppOpCodeByte(OP_SEND, ODT_BYTE));
			context.codeBuffer->push_back(IDX_CLASS);
		}
	}
	// insert return statement if code doesn't already end with return
	size_t bsize = context.codeBuffer->size();
	if ((bsize < 2) ||
		((BppOpCodeByte((*context.codeBuffer)[bsize - 1]).opCode() != OP_RETURN) &&
			(BppOpCodeByte((*context.codeBuffer)[bsize - 2]).opCode() != OP_RETURN)))
	{
		if (context.returnType != BVT_NULL)
		{

			if (context.returnType == BVT_NIL)
			{
				context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
				context.codeBuffer->push_back(BppOpCodeByte(OP_NIL));
				if (context.returnConst) _codeOpcode(*context.codeVec, OP_SPCONST);
				context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN));
			}
			else {
				if (context.returnType == BVT_VOID)
				{
					context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH, ODT_BYTE));
					context.codeBuffer->push_back(BVT_VOID);
				}
				if (context.returnConst) _codeOpcode(*context.codeVec, OP_SPCONST);
				context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN, ODT_BYTE));
				context.codeBuffer->push_back(context.returnType);
			}
		}
		else {
			if (context.returnConst) _codeOpcode(*context.codeVec, OP_SPCONST);
			context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN));
		}

	}
	if (_genLocalVarInfo)
	{
		/*
			Erzeugt wird ein Vector mit zwei Elementen.
			Das erste Element ist ein Vektor mit den Namen aller
			formalen Parameter, das zweite ein Vektor mit den Namen
			der lokalen (temporären) Variablen. Die Reihenfolge der
			Vektorelemente entspricht jeweils der späteren Ablage
			der lokalen Variablen auf dem Stack.
			Existieren zusätzlich statische lokale Variablen, so wird an den
			zweiten Vektor ein zusätzliches Element vom Dictionary-Typ
			angehängt, das die Zuordnungen der Namen der statischen lokalen
			Variablen zu den Indizes im zurückgelieferten Bytecode-Vektor
			enthält.
		*/
		BppVector* info = new BppVector(2, true);
		(*info)[0] = new BppVector(context.arguments);
		BppVector* tmpVec = new BppVector(context.temporaries);

		(*info)[1] = tmpVec;
		if (context.statics.count() > 0)
		{
			BppDictionary* pstat = new BppDictionary(true);
			for (BppDictionaryEntry* p = context.statics.getFirst();
				p != 0; p = p->getNext())
			{
				BppDictionaryEntry* pnew = (*pstat)[p->getKey()];
				pnew->setValue(p->getValue());
			}
			tmpVec->push_back(pstat);
		}

		result->push_back(BppValue(info, true));
	}
	// free unused buffer space
	result->shrink();
	context.codeBuffer->shrink();
	if (context.statics.count() > 0)
		(*result)[IDX_CODE].setFlags(BVF_STATIC);
	/* show the generated code */
	if (_debugger != 0)
		_debugger->decodeProcedure(result);
	/* return the code object */
	return (result);
}

// statementblock ::= "{" { <statement> } "}"
void BppCompiler::_codeBlock(BppFuncContext& context, int& errCnt)
{
	BppToken tkn = _scanner->getToken(errCnt);
	if ((tkn != T_ANY) || (_scanner->getChar() != _T('}')))
	{
		do
		{
			_scanner->back();
			_codeStatement(context, errCnt);
			tkn = _scanner->getToken(errCnt);
			if (tkn == T_EOF)
			{
				_error(CMP_UNEXPECTED_EOF_IN_BLOCK, &errCnt);
				break;
			}
		} while ((tkn != T_ANY) || (_scanner->getChar() != _T('}')));
	}
	else
	{
		context.codeBuffer->push_back(BppOpCodeByte(OP_NIL));
	}
}

void BppCompiler::_codeSourceInfo(BppFuncContext& context, int&)
{
	bool newline = false;
	const BppChar* sname = _scanner->getSourceName();
	long lnum = _scanner->getLineNumber();
	if (_tcscmp(_currentSource.c_str(), sname) != 0)
	{
		newline = true;
		_currentSource = sname;
	}
	if (lnum != _currentLine)
	{
		newline = true;
		_currentLine = lnum;
	}
	if (!newline) return;
	size_t idx = context.codeVec->getStringIndex(sname);
	if (idx == BPP_NNPOS)
	{
		context.codeVec->push_back(new BppString(sname));
		idx = context.codeVec->size() - 1;
	}
	context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
	_codeLiteral(*(context.codeVec), (BppNWord)idx);
	_codeIndexedOpcode(*(context.codeVec), OP_NEWLINE, (BppNWord)lnum);
}



// statement ::= <blockstatement> | <if> | <while> | <dowhile> | <for> | 
// <break> | <continue> | <switch> | <return> | 
// ("TRON" ";") | ("TROFF" ";") | ("TRSTEP" ";") | 
// <try> | <throw> | (<staticvardefinition> | <varstatement> | <constvarstatement> | [<expr>]   ";" ) .
// <varstatement> ::= ("var" | "function" | <typename> | <classname> ) <vardefinition> .
// <constvarstatement> ::= "const" [("var" | "function" | <typename> | <classname> )] <vardefinition> .
// typename = "int" | "float" | "double" | "FILE" | "buffer" | "charbuffer" |
// "object" | "vector" | "string" | "dictionary" .
// classname::= (identifier of an already defined class)
void BppCompiler::_codeStatement(BppFuncContext& context, int& errCnt)
{
	if (_genLineInfo)
		_codeSourceInfo(context, errCnt);
	bool isConst = false;
	BppToken tkn = _scanner->getToken(errCnt);
doconst:
	switch (tkn)
	{
	case T_IF:           _codeIf(context, errCnt);        break;
	case T_WHILE:        _codeWhile(context, errCnt);     break;
	case T_DO:           _codeDoWhile(context, errCnt);   break;
	case T_FOR:          _codeFor(context, errCnt);       break;
	case T_BREAK:        _codeBreak(context, errCnt);     break;
	case T_CONTINUE:     _codeContinue(context, errCnt);  break;
	case T_SWITCH:       _codeSwitch(context, errCnt);    break;
	case T_YIELD:
	case T_RETURN:       _codeReturn(context, tkn == T_YIELD, errCnt);    break;
	case T_TRY:          _codeTry(context, errCnt);       break;
	case T_THROW:        _codeThrow(context, errCnt);     break;
	case T_CONST:
		if (isConst) {
			_require(T__DFUNCTION, T_IDENTIFIER, tkn);
			++errCnt;
		}
		isConst = true;
		tkn = _scanner->getToken(errCnt);
		goto doconst;
	case T_VAR:
	case T_AUTO:
	case T_FUNCTION:
		_codeVarDefinition(context, isConst, errCnt,findVTypeFromName(T__DFUNCTION,_scanner->getTokenString(),errCnt));
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		break;
	case T_STATIC:
	{
		BppValueType vtype = BVT_NULL;
		BppToken tnext = _scanner->getToken(errCnt);
		if (tnext == T_CONST) {
			isConst = true;
		    tnext = _scanner->getToken(errCnt);
		}
		BppString str = _scanner->getTokenString();
		if (tnext == T_IDENTIFIER || tnext == T_FUNCTION || tnext == T_AUTO)
		{
			BppClass* pclass = 0;
			vtype = findVTypeFromName(T__DFUNCTION, str, errCnt, true, &pclass,context.contextClass);
			if (vtype == BVT_NULL)
				_scanner->back();
		}
		else
		{
			_scanner->back();
		}
		_codeStaticVarDefinition(context, isConst, errCnt, vtype);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
	}
	break;
	case T_TRON:
		_codeOpcode(*context.codeVec, OP_TRON);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		break;
	case T_TROFF:
		_codeOpcode(*context.codeVec, OP_TROFF);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		break;
	case T_TRSTEP:
		_codeOpcode(*context.codeVec, OP_TRSTEP);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		break;
	case T_ANY:
	{
		BppChar c = _scanner->getChar();
		if (c == _T('{'))
		{
			_codeBlock(context, errCnt);
			return;
		}
		if (c == _T(';'))
			return;
	}
	case T_IDENTIFIER:
	{
		BppString str = _scanner->getTokenString();
		BppClass* pclass = 0;
		BppValueType vtype = findVTypeFromName(T__DFUNCTION, str, errCnt,  true, &pclass,context.contextClass);
		if (pclass) {
			BppScanner* pscn = getActiveScanner();
			str = parseNamePath(tkn, errCnt);
			pclass = _findClass(str, errCnt, context.contextClass, true);
			if (!pclass) {
				const BppChar* p = _tcsrchr(str.c_str(), _T(':'));
				if (p) {
					BppString lastPart = (const BppChar*)(p + 1);
					str.erase((--p) - str.c_str());
					pscn->_token = T_IDENTIFIER;
					pscn->_stringBuffer = lastPart.c_str();
					pscn->back();// letzter Teil - Variablen- oder Funktionsname
					pscn->_token = T_CC;
					pscn->back();
					pscn->_stringBuffer = str.c_str(); // Klassenname
					pscn->_token = T_IDENTIFIER;
				}
				vtype = BVT_NULL;
			}
			else {
				// Folgetoken (Variablen- oder Funktionsname) wieder vom Stack holen
				BppToken tnext = _scanner->getToken(errCnt);
				if (tnext == T_IDENTIFIER) // Variablendefinition
				{
					_scanner->back();
					getActiveScanner()->_stringBuffer = str.c_str();
					_codeVarDefinition(context, isConst, errCnt, vtype, pclass);
					if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
					break;
				}
				else { // Ausdruck
					_scanner->back(); // Klammer;					
					pscn->_token = T_IDENTIFIER;
					pscn->_stringBuffer = str.c_str();
					_scanner->back();
					_codeExpr(context, errCnt);
					if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
					break;
				}
			}
		}
		if (vtype != BVT_NULL || isConst)
		{

			_scanner->back();
			// Identifier-Token sichern
			BppScanner* scn = getActiveScanner();
			BppSavedToken* p = scn->_tokenStack;
			if (vtype != BVT_NULL) {
				scn->_tokenStack = scn->_tokenStack->next;
				p->next = 0;
			}
			BppToken tnext = _scanner->getToken(errCnt);
			if (tnext == T_IDENTIFIER)
			{
				BppScanner* scn = getActiveScanner();
				_scanner->back(); // save tnext
				if (p && vtype != BVT_NULL) {
					p->next = scn->_tokenStack;
					scn->_tokenStack = p;
				}
				if (vtype != BVT_NULL) _scanner->getToken(errCnt); // pop type name token
				_codeVarDefinition(context, isConst, errCnt, vtype, pclass);
				if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
				break;
			}
			else {
				_scanner->back();
				if (p) {
					p->next = scn->_tokenStack;
					scn->_tokenStack = p;
				}
				_codeExpr(context, errCnt);
				if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
				break;
			}
		}
	}
	default:
		_scanner->back();
		_codeExpr(context, errCnt);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		break;
	}
}

// <varstatement> := "var" <identifier> { "," <identifier> } [ "=" <expr2> ] .
// <vardef> ::= <identifier> [ "=" ]
void BppCompiler::_codeVarDefinition(BppFuncContext& context,  bool isConst,int& errCnt, BppValueType defType, BppClass* reqClass)
{
	BppScanner* pscan = getActiveScanner();
	BppString tname;
	if (defType != BVT_NULL)
		tname = _scanner->getTokenString();
	static const size_t sizeidx = 1;
	BppBuffer* buf = context.codeBuffer;
	BppOpCodeByte firstOpCode = (*buf)[0];
	if (firstOpCode.opCode() != OP_TSPACE) // sollte nie passieren (vgl. _codeFunctionBody)
	{
		_error(CMP_NO_LOCAL_VARSPACE, &errCnt);
	}
	BppToken t = _scanner->getToken(errCnt);
	// walk through definition list
	while (1)
	{
		if (!_require(T__DFUNCTION, T_IDENTIFIER, t))
		{
			++errCnt;
			break;
		}
		BppString name = _scanner->getTokenString();
		if (context.statics.find(name))
		{
			_error(BppString::format(CMP_STATIC_VAR_ALREADY_DEFINED_FMT, name.c_str()).c_str(), &errCnt);
		}
		if (context.arguments.getStringIndex(name.c_str()) != BPP_NNPOS)
		{
			_error(BppString::format(CMP_ARGUMENT_ALREADY_DEFINED_FMT, name.c_str()).c_str(), &errCnt);
		}
		if (context.temporaries.getStringIndex(name.c_str()) != BPP_NNPOS)
		{
			_error(BppString::format(CMP_LOCAL_VAR_ALREADY_DEFINED_FMT, name.c_str()).c_str(), &errCnt);
		}
		context.temporaries.push_back(new BppString(name));

		t = _scanner->getToken(errCnt);
		if ((t == T_ANY) && (_scanner->getChar() == _T('=')))
		{
			BppVarInfo varInfo;
			varInfo.index = context.temporaries.size() - 1;
			varInfo.type = VAR_LOCAL;
			varInfo.varvtype = defType;
			varInfo.reqClass = reqClass;
			if (defType == BVT_STRONGOBJECT && reqClass)
			{
				BppVector* v = new BppVector(2, true);
				(*v)[IDX_CODE] = new BppObject(reqClass, true);
				(*v)[IDX_CLASS] = reqClass;
				BppValue val((BppRefObject*)v, BVT_STRONGOBJECT);
				val.setFlags(val.getFlags() | BVF_STRONGTYPED);

				context.codeVec->push_back(val/*BppValue((BppRefObject*)new BppObject(reqClass,true), BVT_STRONGOBJECT)*/);
				BppVarInfo vinfo;
				vinfo.type = VAR_SLOCAL;
				vinfo.varvtype = BVT_OBJECT;
				vinfo.reqClass = reqClass;
				vinfo.index = context.codeVec->size() - 1;

				_codeVarInfo(*context.codeVec, VA_LOAD, vinfo); // load empty strong object
				vinfo = varInfo;

				_codeVarInfo(*context.codeVec, VA_STORE, vinfo); // store result
			}
			BppVarInfo rInfo;
			_codeExpr2_assign(context, rInfo, errCnt);
			_codeRValue(*context.codeVec, rInfo);

			_codeVarInfo(*context.codeVec, VA_STORE, varInfo); // store result
			if (isConst) {
				_codeIndexedOpcode(*context.codeVec, OP_MKCONST, varInfo.index);
				context.temporaries[varInfo.index].setFlags(BVF_CONST);
			}
		}
		else if ((t == T_ANY) && (_scanner->getChar() == _T('(')))
		{
			_scanner->back();
			BppVarInfo varInfo;
			varInfo.index = context.temporaries.size() - 1;
			varInfo.type = VAR_LOCAL;
			varInfo.varvtype = defType;
			BppVarInfo rInfo;
			pscan->_token = T_IDENTIFIER;
			switch (defType)
			{
			case BVT_OBJECT:
			case BVT_STRONGOBJECT:
				if (_tcscmp(tname.c_str(), TN_object) != 0)
				{
					pscan->_stringBuffer = tname.c_str();
					_scanner->back();
					_codeNew(context, rInfo, errCnt);
				}
				else
				{
					_scanner->getToken(errCnt); // eat '('
					_codeExpr1_comma(context, rInfo, errCnt);
					if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
				}
				break;
			case BVT_SBYTE:
			case BVT_BYTE:
			case BVT_SHORT:
			case BVT_USHORT:
			case BVT_INT:
			case BVT_UINT:
			case BVT_LONG:
			case BVT_ULONG:
			case BVT_INTPTR:
			case BVT_FLOAT:
			case BVT_DOUBLE:
			case BVT_STRING:
			case BVT_BUFFER:
			case BVT_CHARBUFFER:
			case BVT_VECTOR:
			case BVT_DICTIONARY:
				pscan->_stringBuffer = tname.c_str();
				_scanner->back();
				_codeExpr15_post(context, rInfo, errCnt);
				break;
			case BVT_FILE:
				pscan->_stringBuffer = _T("fopen");
				_scanner->back();
				_codeExpr15_post(context, rInfo, errCnt);
				break;
			default:
				_scanner->getToken(errCnt); // eat '('
				_codeExpr1_comma(context, rInfo, errCnt);
				if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
			}
			_codeRValue(*context.codeVec, rInfo);
			_codeVarInfo(*context.codeVec, VA_STORE, varInfo); // store result
			if (isConst)
				_codeIndexedOpcode(*context.codeVec, OP_MKCONST, varInfo.index);
		}
		else
		{
			_scanner->back();
			if (defType != BVT_NULL)
			{
				BppVarInfo varInfo;
				varInfo.index = context.temporaries.size() - 1;
				varInfo.type = VAR_LOCAL;
				varInfo.varvtype = defType;
				BppVarInfo rInfo;
				switch (defType)
				{
				case BVT_SBYTE:
				case BVT_BYTE:
				case BVT_SHORT:
				case BVT_USHORT:
				case BVT_INT:
				case BVT_UINT:
				case BVT_LONG:
				case BVT_ULONG:
				case BVT_INTPTR:
				case BVT_FLOAT:
				case BVT_DOUBLE:
					context.codeVec->push_back(BppValue(defType, BVF_NONE, false));
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				case BVT_BUFFER:
					context.codeVec->push_back(new BppBuffer());
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				case BVT_CHARBUFFER:
					context.codeVec->push_back(new BppCharBuffer());
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				case BVT_VECTOR:
					context.codeVec->push_back(new BppVector());
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				case BVT_DICTIONARY:
					context.codeVec->push_back(new BppDictionary());
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				case BVT_FILE:
					context.codeVec->push_back(BppValue((FILE*)0));
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				case BVT_OBJECT:
				case BVT_STRONGOBJECT:
					if (_tcscmp(tname.c_str(), TN_object) != 0)
					{
						pscan->_token = T_ANY;
						pscan->_charBuffer[0] = _T(')');
						_scanner->back();
						pscan->_charBuffer[0] = _T('(');
						_scanner->back();
						pscan->_stringBuffer = tname.c_str();
						pscan->_token = T_IDENTIFIER;
						_scanner->back();
						_codeNew(context, rInfo, errCnt);
					}
					else
					{
						context.codeVec->push_back(BppValue());
						_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
						rInfo.type = VAR_DONE;
					}
					break;
				case BVT_STRING:
					pscan->_stringBuffer = _T("");
					pscan->_token = T_STRING;
					_scanner->back();
					_codeExpr15_post(context, rInfo, errCnt);
					break;
				case BVT_AUTO:
					_error(CMP_MUST_INITIALIZE_AUTO, &errCnt);
				default:
					context.codeVec->push_back(BppValue());
					_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
					rInfo.type = VAR_DONE;
					break;
				}
				_codeRValue(*context.codeVec, rInfo);
				_codeVarInfo(*context.codeVec, VA_STORE, varInfo); // store result
				if (isConst)
					_error(CMP_MUST_INITIALIZE_CONST, &errCnt);
			}
		}
		t = _scanner->getToken(errCnt);
		if ((t == T_ANY) && (_scanner->getChar() == _T(',')))
		{
			t = _scanner->getToken(errCnt);
		}
		else
		{
			_scanner->back();
			break;
		}
	}
	size_t varcnt = context.temporaries.size();

	buf->setNW((size_t)sizeidx, (BppNWord)varcnt);
}

// <staticvarstatement> := "static" ["var" | <typename>] <identifier> [ "=" literal ] "," <identifier> [ "=" literal ]}  .
void BppCompiler::_codeStaticVarDefinition(BppFuncContext& context, bool isConst, int& errCnt, BppValueType defType, BppClass* defClass)
{
	BppToken t = _scanner->getToken(errCnt);
	if (defType == BVT_NULL && t == T_VAR)
		t = _scanner->getToken(errCnt);
	while (1)
	{
		if (!_require(T__DFUNCTION, T_IDENTIFIER, t))
		{
			++errCnt;
			break;
		}
		BppString name = _scanner->getTokenString();
		if (context.statics.find(name))
		{
			_error(BppString::format(CMP_STATIC_VAR_ALREADY_DEFINED_FMT, name.c_str()).c_str(), &errCnt);
		}
		if (context.arguments.getStringIndex(name.c_str()) != BPP_NNPOS)
		{
			_error(BppString::format(CMP_ARGUMENT_ALREADY_DEFINED_FMT, name.c_str()).c_str(), &errCnt);
		}
		if (context.temporaries.getStringIndex(name.c_str()) != BPP_NNPOS)
		{
			_error(BppString::format(CMP_LOCAL_VAR_ALREADY_DEFINED_FMT, name.c_str()).c_str(), &errCnt);
		}
		size_t varcnt = context.codeVec->size();
		if (defType == BVT_STRONGOBJECT && defClass)
		{
			BppVector* desc = new BppVector(2, true);
			(*desc)[IDX_CLASS] = defClass;
			context.codeVec->push_back(BppValue((BppRefObject*)desc, defType));
		}
		else
			context.codeVec->push_back(BppValue(defType, BVF_NONE, true));
		context.statics[name].setValue((BppInt)(varcnt));
		(*context.codeVec)[varcnt].setFlags(defType == BVT_NULL ? BVF_STATIC : BVF_STATIC | BVF_STRONGTYPED);
		t = _scanner->getToken(errCnt);
		BppChar chr = _scanner->getChar();
		if ((t == T_ANY) && ((chr == _T('=')) || (chr == _T('('))))
		{
			BppDictionaryEntry dummy(0, BppValue());
			_doLitExpr(_scanner->getToken(errCnt), dummy, errCnt);
			varcnt = context.codeVec->size();
			BppValue& dst = (*(context.codeVec))[varcnt - 1];
			if (!dst.assign(BppValue(dummy.getValue(), true)))
			{
				_error(BppString::format(CMP_TYPE_MISMATCH_FMT3,
					name.c_str(),
					getValueTypeName((*(context.codeVec))[varcnt - 1].getType()),
					getValueTypeName(dummy.getValue().getType())).c_str(), &errCnt);
			}
			if (isConst)
				dst.setFlags(dst.getFlags() | BVF_CONST);

			if (chr == _T('(') && (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')')))) ++errCnt;
		}
		else
		{
			if (defType == BVT_AUTO)
			{
				_error(CMP_MUST_INITIALIZE_AUTO, &errCnt);
			}
			_scanner->back();
		}
		t = _scanner->getToken(errCnt);
		if ((t == T_ANY) && (_scanner->getChar() == _T(',')))
		{
			t = _scanner->getToken(errCnt);
		}
		else
		{
			_scanner->back();
			break;
		}
	}
}
// throw ::= "throw" <expr> ";"
void BppCompiler::_codeThrow(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;
	if (context.catchLevel)
	{
		// inside of catch block throw may be called without expression
		BppToken t = _scanner->getToken(errCnt);
		if ((t == T_ANY) && (_scanner->getChar() == _T(';')))
		{
			// throw last exception again
			_codeTemporary(*context.codeVec, VA_LOAD, IDX_EXCEPTION);
			buf.push_back(BppOpCodeByte(OP_THROW));
			return;
		}
		_scanner->back();
	}
	_codeExpr(context, errCnt);
	buf.push_back(BppOpCodeByte(OP_THROW));
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
}

// try ::= "try" <blockstatement> "catch" "(" <identifier> ")" <blockstatement> 
void BppCompiler::_codeTry(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;
	buf.push_back(BppOpCodeByte(OP_TRY, ODT_NWORD));
	BppNWord nxt = (BppNWord)buf.push_backNW(0); // place holder for jump offset

	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('{'))) errCnt++;
	BppValue savedExeptionName = context.temporaries[IDX_EXCEPTION];

	// compile protected block   
	_codeBlock(context, errCnt);
	buf.push_back(BppOpCodeByte(OP_ENDTRY, ODT_NWORD));
	BppNWord end = (BppNWord)buf.push_backNW(0); // place holder for jump offset
	_fixupRefChain(context, nxt, buf.size());
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_CATCH)) errCnt++;
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('('))) errCnt++;
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER)) errCnt++;
	else
		context.temporaries[IDX_EXCEPTION] = new BppString(_scanner->getTokenString());
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('{'))) errCnt++;
	context.catchLevel++;
	// compile exception handler block
	_codeBlock(context, errCnt);
	buf.push_back(BppOpCodeByte(OP_ENDCATCH));
	context.catchLevel--;
	_fixupRefChain(context, end, buf.size());
	context.temporaries[IDX_EXCEPTION] = savedExeptionName;
}

// if ::= "if" <condition> <statement> [ "else" <statement> ]
void BppCompiler::_codeIf(BppFuncContext& context, int& errCnt)
{
	/* compile the test expression */
	_codeCondition(context, errCnt);

	BppBuffer& buf = *context.codeBuffer;
	/* skip around the 'then' clause if the expression is false */
	buf.push_back(BppOpCodeByte(OP_BRF, ODT_NWORD));
	BppNWord nxt = (BppNWord)buf.push_backNW(0); // place holder for jump offset

	/* compile the 'then' clause */
	_codeStatement(context, errCnt);

	/* compile the 'else' clause */
	BppToken tkn = _scanner->getToken(errCnt);
	if (tkn == T_ELSE)
	{
		buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
		BppNWord end = (BppNWord)buf.push_backNW(0); // place holder for jump offset
		_fixupRefChain(context, nxt, buf.size());
		_codeStatement(context, errCnt);
		nxt = end;
	}
	else
		_scanner->back();

	/* handle the end of the statement */
	_fixupRefChain(context, nxt, buf.size());
}

// while ::= "while" <condition> <statement>
void BppCompiler::_codeWhile(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;

	/* compile the test expression */
	BppNWord nxt = buf.size();
	_codeCondition(context, errCnt);

	/* skip around the loop body if the expression is false */
	buf.push_back(BppOpCodeByte(OP_BRF, ODT_NWORD));
	BppNWord end = (BppNWord)buf.push_backNW(0); // place holder for jump offset

	/* compile the loop body */
	BppNWord oldBreak = context.pushBreak(end);
	BppNWord oldCont = context.pushContinue(nxt);

	_codeStatement(context, errCnt);
	end = context.popBreak(oldBreak, end);
	context.popContinue(oldCont);

	/* branch back to the start of the loop */
	buf.push_back(BppOpCodeByte(OP_POPC));
	buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
	buf.push_backNW(nxt);

	/* handle the end of the statement */
	_fixupRefChain(context, end, buf.size());
}

// dowhile ::= "do" <statement> "while" <condition> ";"
void BppCompiler::_codeDoWhile(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;

	/* remember the start of the loop */
	BppNWord nxt = buf.size();

	/* compile the loop body */
	BppNWord oldBreak = context.pushBreak(0);
	BppNWord oldCont = context.pushContinue(nxt);

	_codeStatement(context, errCnt);
	BppNWord end = context.popBreak(oldBreak, 0);
	context.popContinue(oldCont);

	/* compile the test expression */
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_WHILE)) errCnt++;
	_codeCondition(context, errCnt);
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;

	/* branch to the top if the expression is true */
	buf.push_back(BppOpCodeByte(OP_POPC));
	buf.push_back(BppOpCodeByte(OP_BRT, ODT_NWORD));
	buf.push_backNW(nxt);

	/* handle the end of the statement */
	_fixupRefChain(context, end, buf.size());
}

// for ::= "for" "(" <varstatement> | <expr> ";" <expr> ";" <expr> ")" <statement> .
void BppCompiler::_codeFor(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;

	/* compile the initialization expression */
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('('))) errCnt++;
	BppToken tkn = _scanner->getToken(errCnt);
	switch (tkn)
	{
	case T_VAR:
		_codeVarDefinition(context,false, errCnt);
		break;
	case T_IDENTIFIER:
	case T_AUTO:
	{
		BppString tokstr = _scanner->getTokenString();
		BppString str = parseNamePath(tkn,errCnt); 
		BppClass* pclass = 0;
		BppValueType vtype = findVTypeFromName(T__DFUNCTION, str, errCnt, true, &pclass);
		if (vtype != BVT_NULL)
		{
			//_scanner->back(); // save type name token
			BppToken tnext = _scanner->getToken(errCnt);
			_scanner->back();
			BppScanner* pscn = getActiveScanner();
			pscn->_token = T_IDENTIFIER;
			pscn->_stringBuffer = str.c_str();
			pscn->_charBuffer[0] = _T('\0');
			if (tnext == T_IDENTIFIER)
			{
				//_scanner->back(); 
				//_scanner->getToken(errCnt); // pop type name token
				_codeVarDefinition(context, false, errCnt, vtype, pclass);
				//_scanner->back();
				break;
			}
		}
		else {
			BppScanner* pscn = getActiveScanner();
			pscn->_token = T_IDENTIFIER;
			pscn->_stringBuffer = tokstr.c_str();
			pscn->_charBuffer[0] = _T('\0');
		}
	}
	case T_ANY:
		if (_scanner->getChar() == _T(';')) {
			_scanner->back();
			break;
		}
	default:
		_scanner->back(/*true*/);
		_codeExpr(context, errCnt);
	}

	/* compile the test expression */
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
	BppNWord nxt = buf.size();
	tkn = _scanner->getToken(errCnt);
	if (tkn == T_ANY && _scanner->getChar() == _T(';'))
	{
		_scanner->back();
		buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
	}
	else
	{
		_scanner->back();
		_codeExpr(context, errCnt);
		/* branch to the loop body if the expression is true */
		buf.push_back(BppOpCodeByte(OP_BRT, ODT_NWORD));
	}
	BppNWord body = (BppNWord)buf.push_backNW(0);

	/* branch to the end if the expression is false */
	buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
	BppNWord end = (BppNWord)buf.push_backNW(0);

	/* compile the update expression */
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
	BppNWord update = buf.size();
	tkn = _scanner->getToken(errCnt);
	if (tkn != T_ANY || _scanner->getChar() != _T(')'))
	{
		_scanner->back();
		_codeExpr(context, errCnt);
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
	}

	/* branch back to the test code */
	buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
	buf.push_backNW(nxt);

	/* compile the loop body */
	_fixupRefChain(context, body, buf.size());
	BppNWord oldBreak = context.pushBreak(end);
	BppNWord oldCont = context.pushContinue(update);
	_codeStatement(context, errCnt);
	end = context.popBreak(oldBreak, end);
	context.popContinue(oldCont);

	/* branch back to the update code */
	buf.push_back(BppOpCodeByte(OP_POPC));
	buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
	buf.push_backNW(update);
	/* handle the end of the statement */
	_fixupRefChain(context, end, buf.size());
}

void BppCompiler::_codeSwitch(BppFuncContext& context, int& errCnt)
{

	BppBuffer& buf = *context.codeBuffer;

	/* compile the test expression */
	_codeCondition(context, errCnt);
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('{'))) errCnt++;

	buf.push_back(BppOpCodeByte(OP_SWITCH, ODT_NWORD));
	BppNWord end = (BppNWord)buf.push_backNW(0); // place holder for jump offset
	buf.push_back(BppOpCodeByte(OP_PUSH)); // reserve space for second value to compare
	BppNWord nxt = 0;
	BppNWord nxtBlock = 0;
	BppNWord cdefault = 0;
	BppNWord oldBreak = context.pushBreak(end);
	bool statementAllowed = false;
	BppToken tkn;
	while ((tkn = _scanner->getToken(errCnt)) != T_EOF)
	{
		if (tkn == T_ANY) {
			BppChar c = _scanner->getChar();
			if (c == _T('}')) break;
		}
		if (tkn == T_CASE)
		{
			statementAllowed = true;
			if (nxt != 0)
			{
				buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
				nxtBlock = (BppNWord)buf.push_backNW(0);
				_fixupRefChain(context, nxt, buf.size());
			}
			else
				nxtBlock = 0;
			buf.push_back(BppOpCodeByte(OP_POP)); // reserve space for second value to compare
			buf.push_back(BppOpCodeByte(OP_DUP)); // reserve space for second value to compare
			buf.push_back(BppOpCodeByte(OP_DUP)); // reserve space for second value to compare
			_codeExpr(context, errCnt);
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(':'))) errCnt++;
			buf.push_back(BppOpCodeByte(OP_EQ));
			buf.push_back(BppOpCodeByte(OP_BRF, ODT_NWORD));
			nxt = (BppNWord)buf.push_backNW(0);
			if (nxtBlock != 0)
			{
				_fixupRefChain(context, nxtBlock, buf.size());
				nxtBlock = 0;
			}
		}
		else if (tkn == T_DEFAULT)
		{
			if (cdefault != 0)
			{
				_error(CMP_MULTIPLE_DEFAULT_NOT_ALLOWED, &errCnt);
			}
			statementAllowed = true;
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(':'))) errCnt++;

			cdefault = (BppNWord)buf.size();
			if (nxtBlock != 0)
			{
				_fixupRefChain(context, nxtBlock, buf.size());
				nxtBlock = 0;
			}
		}
		else
		{
			if (!statementAllowed)
			{
				_error(CMP_CASE_OR_DEFAULT_EXPECTED, &errCnt);
			}
			_scanner->back();
			_codeStatement(context, errCnt);
		}
	}
	BppNWord cptr = (BppNWord)buf.size();
	end = context.popBreak(oldBreak, end);
	_fixupRefChain(context, end, cptr);
	if (nxt != 0)
	{
		if (cdefault != 0)
			_fixupRefChain(context, nxt, cdefault);
		else
			_fixupRefChain(context, nxt, cptr);
	}
	if (nxtBlock != 0)
		_fixupRefChain(context, nxtBlock, cptr);

	buf.push_back(BppOpCodeByte(OP_POPC));
}

// break ::= "break" ";"
void BppCompiler::_codeBreak(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;

	if (context.breakStack.size() > 0)
	{
		buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
		size_t idx = buf.push_backNW(context.popBreak(0, 0));
		context.pushBreak((BppNWord)idx);
	}
	else
	{
		_error(CMP_BREAK_OUTSIDE_LOOP, &errCnt);
	}
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
}

// continue ::= "continue" ";"
void BppCompiler::_codeContinue(BppFuncContext& context, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;

	if (context.breakStack.size() > 0)
	{
		buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
		buf.push_backNW(context.peekContinue());
	}
	else
	{
		_error(CMP_CONTINUE_OUTSIDE_LOOP, &errCnt);
	}
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
}

// return ::= "return"|"yield" [ <expr> ] ";"
void BppCompiler::_codeReturn(BppFuncContext& context, bool isYield, int& errCnt)
{
	BppToken tkn = _scanner->getToken(errCnt);
	BppOpCode oc = isYield ? OP_YIELD : OP_RETURN;
	if ((tkn != T_ANY) || (_scanner->getChar() != _T(';')))
	{
		if (context.returnType == BVT_VOID)
		{
			_error(CMP_RETURNVAL_NOT_ALLOWED, &errCnt);
		}
		if (context.returnType == BVT_NIL) {
			if (tkn != T_NIL) _error(CMP_INVALID_RETURNVAL_FOR_NULL, &errCnt);
		}
		else {
			_scanner->back();
			_codeExpr(context, errCnt);
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(';'))) errCnt++;
		}
	}
	else if (context.returnType > BVT_NIL)
	{
		_error(CMP_RETURNVAL_EXPECTED, &errCnt);
	}
	if (context.returnType == BVT_NIL) {
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
		context.codeBuffer->push_back(BppOpCodeByte(OP_NIL));
		if (context.returnConst) _codeOpcode(*context.codeVec, OP_SPCONST);
		_codeOpcode(*context.codeVec, oc);
		//context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN));
	}
	else if (context.returnType != BVT_NULL)
	{
		if (context.returnType == BVT_STRONGOBJECT && context.prClass) {
			context.codeVec->push_back(context.prClass);
			context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
			_codeLiteral(*context.codeVec, context.codeVec->size() - 1);
		}
		else if (context.returnType == BVT_VOID)
		{
			context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH, ODT_BYTE));
			context.codeBuffer->push_back(BVT_VOID);
		}
		//if (isYield)context.codeBuffer->push_back(BppOpCodeByte(OP_DUP, ODT_NONE));
		if (context.returnConst) _codeOpcode(*context.codeVec, OP_SPCONST);
		_codeIndexedOpcode(*context.codeVec, oc, context.returnType);
		//context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN, ODT_BYTE));
		//context.codeBuffer->push_back(context.returnType);
	}
	else {
		if (context.returnConst) _codeOpcode(*context.codeVec, OP_SPCONST);
		_codeOpcode(*context.codeVec, oc);
		//context.codeBuffer->push_back(BppOpCodeByte(OP_RETURN));
	}
	if (isYield)
		(*context.codeVec)[IDX_CODE].setFlags((*context.codeVec)[IDX_CODE].getFlags() | BVF_STATIC);
}

void BppCompiler::_fixupRefChain(BppFuncContext& context, BppNWord startIdx, BppNWord destIdx)
{
	BppBuffer& buf = *context.codeBuffer;
	BppNWord idx = startIdx;
	BppNWord bsize = (BppNWord)buf.size();
	while ((idx != 0) && (idx < bsize))
	{
#ifdef WINCE
		BppByte* pi = &(buf.data()[idx]);
		memcpy(&idx, pi, sizeof(BppNWord));
		memcpy(pi, &destIdx, sizeof(BppNWord));
#else
#ifdef GCC
		BppByte* pi = &(buf.data()[idx]);
		memcpy(&idx, pi, sizeof(BppNWord));
		memcpy(pi, &destIdx, sizeof(BppNWord));
#else
		BppNWord* pi = (BppNWord*)&(buf.data()[idx]);
		idx = *pi;
		*pi = destIdx;
#endif    
#endif
	}
}

// condition ::= "(" <expr> ")"
void BppCompiler::_codeCondition(BppFuncContext& context, int& errCnt)
{
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('('))) errCnt++;
	_codeExpr(context, errCnt);
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
}

void BppCompiler::_codeExpr(BppFuncContext& context, int& errCnt)
{
	BppVarInfo varInfo;
	_codeExpr1_comma(context, varInfo, errCnt);
	_codeRValue(*context.codeVec, varInfo);
}

// expr ::= <expr1>
// expr1 ::= <expr2> { "," <expr> }
/**
handles operator ','
*/
void BppCompiler::_codeExpr1_comma(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr2_assign(context, varInfo, errCnt);
	BppToken tkn = _scanner->getToken(errCnt);
	while ((tkn == T_ANY) && (_scanner->getChar() == _T(',')))
	{
		_codeRValue(*context.codeVec, varInfo);
		_codeExpr1_comma(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		tkn = _scanner->getToken(errCnt);
	}
	_scanner->back();
}


// expr2 ::= <expr3> <assignop> <expr>
// assignop ::= "=" | "+=" | "-=" | "*=" | "/=" | "%=" | "&=" | "|=" | "^=" | "~=" |  "<<=" | ">>=" | "=>"
/**
handles assignment operators
*/
void BppCompiler::_codeExpr2_assign(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{

	_codeExpr3_ternop(context, varInfo, errCnt); // handle the '?:' operator
	BppToken tkn;
	while ((((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('=')))
		|| tkn == T_ADDEQ
		|| tkn == T_SUBEQ
		|| tkn == T_MULEQ
		|| tkn == T_DIVEQ
		|| tkn == T_REMEQ
		|| tkn == T_BANDEQ
		|| tkn == T_BOREQ
		|| tkn == T_XOREQ
		|| tkn == T_BNOTEQ
		|| tkn == T_SHLEQ
		|| tkn == T_SHREQ
		//||     tkn == T_REFEQ
		//||     tkn == T_UNREFEQ
		)
	{
		if (varInfo.type == VAR_MEMBER)
		{
			if (!varInfo.isMutable && context.constThis &&
			//	varInfo.reqClass &&
				context.contextClass/* &&
				context.contextClass->isSubtypeOf(varInfo.reqClass)*/)
				varInfo.isConst = true;
		}
		if (!_checkLValue(varInfo)) errCnt++;
		switch (tkn)
		{
		case T_ANY: // simple assignment
		{
			if (varInfo.type == VAR_INDEX || varInfo.type == VAR_PROP)
				context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH)); // push right expression result
			// evaluate right expression
			BppVarInfo rInfo;
			_codeExpr1_comma(context, rInfo, errCnt);
			_codeRValue(*context.codeVec, rInfo);
			_codeVarInfo(*context.codeVec, VA_STORE, varInfo); // store result
		}
		break;
		case T_ADDEQ:
			_codeAssign(context, OP_ADD, varInfo, errCnt);
			break;
		case T_SUBEQ:
			_codeAssign(context, OP_SUB, varInfo, errCnt);
			break;
		case T_MULEQ:
			_codeAssign(context, OP_MUL, varInfo, errCnt);
			break;
		case T_DIVEQ:
			_codeAssign(context, OP_DIV, varInfo, errCnt);
			break;
		case T_REMEQ:
			_codeAssign(context, OP_REM, varInfo, errCnt);
			break;
		case T_BANDEQ:
			_codeAssign(context, OP_BAND, varInfo, errCnt);
			break;
		case T_BOREQ:
			_codeAssign(context, OP_BOR, varInfo, errCnt);
			break;
		case T_XOREQ:
			_codeAssign(context, OP_XOR, varInfo, errCnt);
			break;
		case T_BNOTEQ:
			_codeAssign(context, OP_BNOT, varInfo, errCnt);
			break;
		case T_SHLEQ:
			_codeAssign(context, OP_SHL, varInfo, errCnt);
			break;
		case T_SHREQ:
			_codeAssign(context, OP_SHR, varInfo, errCnt);
			break;
		} // end switch
		varInfo.type = VAR_DONE;
	} // end while
	_scanner->back();
}

void BppCompiler::_codeAssign(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt)
{
	_codeVarInfo(*context.codeVec, VA_DUP, varInfo); // duplicate index (??)
	_codeVarInfo(*context.codeVec, VA_LOAD, varInfo); // load variable

	context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH)); // push right expression result
	// evaluate right expression
	BppVarInfo rInfo;
	_codeExpr1_comma(context, rInfo, errCnt);
	_codeRValue(*context.codeVec, rInfo);

	// code operation
	context.codeBuffer->push_back(BppOpCodeByte(op));
	_codeVarInfo(*context.codeVec, VA_STORE, varInfo); // store result
}

// expr3 ::= <expr4> { "?" <expr> ":" <expr> }
/**
handle ?:-operator
*/
void BppCompiler::_codeExpr3_ternop(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr4_logor(context, varInfo, errCnt);
	BppToken tkn;
	BppBuffer& buf = *context.codeBuffer;
	while (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('?')))
	{
		_codeRValue(*context.codeVec, varInfo);
		buf.push_back(BppOpCodeByte(OP_BRF, ODT_NWORD));
		BppNWord nxt = buf.push_backNW(0);
		_codeExpr1_comma(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);

		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(':'))) errCnt++;

		buf.push_back(BppOpCodeByte(OP_BR, ODT_NWORD));
		BppNWord end = buf.push_backNW(0);
		_fixupRefChain(context, nxt, buf.size());
		_codeExpr1_comma(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		_fixupRefChain(context, end, buf.size());
	}
	_scanner->back();
}

// expr4 ::= <expr5> { "||" <expr5> }
/**
handle '||' operator
*/
void BppCompiler::_codeExpr4_logor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr5_logand(context, varInfo, errCnt);
	BppNWord end = 0;
	BppToken tkn;
	while ((tkn = _scanner->getToken(errCnt)) == T_OR)
	{
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_BRT, ODT_NWORD));
		end = context.codeBuffer->push_backNW(end);
		_codeExpr5_logand(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
	}
	_fixupRefChain(context, end, context.codeBuffer->size());
	_scanner->back();
}


// expr5 ::= <expr6> { "&&" <expr6> }
/**
handle '&&' operator
*/
void BppCompiler::_codeExpr5_logand(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr6_binor(context, varInfo, errCnt);
	BppNWord end = 0;
	BppToken tkn;
	while ((tkn = _scanner->getToken(errCnt)) == T_AND)
	{
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_BRF, ODT_NWORD));
		end = context.codeBuffer->push_backNW(end);
		_codeExpr6_binor(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
	}
	_fixupRefChain(context, end, context.codeBuffer->size());
	_scanner->back();
}

// expr6 ::= <expr7> { "|" <expr7> }
/**
handle '|' operator
*/
void BppCompiler::_codeExpr6_binor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppToken tkn;
	_codeExpr7_xor(context, varInfo, errCnt);
	while (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('|')))
	{
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr7_xor(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_BOR));
	}
	_scanner->back();
}


// expr7 ::= <expr8> [ "^" <expr8> }
/**
handle '^' operator
*/
void BppCompiler::_codeExpr7_xor(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppToken tkn;
	_codeExpr8_binand(context, varInfo, errCnt);
	while (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('^')))
	{
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr8_binand(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_XOR));
	}
	_scanner->back();
}

// expr8 ::= <expr9> { "&" <expr9> }
/**
handle '&' operator
*/
void BppCompiler::_codeExpr8_binand(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppToken tkn;
	_codeExpr9_eqops(context, varInfo, errCnt);
	while (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('&')))
	{
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr9_eqops(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_BAND));
	}
	_scanner->back();
}

// expr9 ::= <expr10> { ("==" | "!=") <expr10> }
/**
handle the '==' and '!=' operators
*/
void BppCompiler::_codeExpr9_eqops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr10_cmpops(context, varInfo, errCnt);
	BppToken tkn;
	BppOpCode op = OP_NIL;
	while ((tkn = _scanner->getToken(errCnt)) == T_EQ || tkn == T_NE)
	{
		switch (tkn)
		{
		case T_EQ: op = OP_EQ; break;
		case T_NE: op = OP_NE; break;
		}
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr10_cmpops(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(op));
	}
	_scanner->back();
}

// expr10 ::= <expr11> { ("<" | ">" | "<=" | ">="} <expr11>
/**
handle the '<', '<=', '>=' and '>' operators
*/
void BppCompiler::_codeExpr10_cmpops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr11_shift(context, varInfo, errCnt);
	BppToken tkn;
	BppOpCode op = OP_NIL;
	BppChar c = 0;
	while ((tkn = _scanner->getToken(errCnt)) == T_LE
		|| tkn == T_GE
		|| ((tkn == T_ANY) && (((c = _scanner->getChar()) == _T('>')) || (c == _T('<'))))
		)
	{
		switch (tkn)
		{
		case T_LE: op = OP_LE; break;
		case T_GE: op = OP_GE; break;
		case T_ANY: op = (c == _T('<')) ? OP_LT : OP_GT; break;
		}
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr11_shift(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(op));
	}
	_scanner->back();
}

// expr11 ::= <expr12> { ("<<" | ">>") <expr12> }
/**
handle the '<<' and '>>' operators
*/
void BppCompiler::_codeExpr11_shift(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr12_addops(context, varInfo, errCnt);
	BppToken tkn;
	BppOpCode op = OP_NIL;
	while ((tkn = _scanner->getToken(errCnt)) == T_SHL || tkn == T_SHR)
	{
		switch (tkn)
		{
		case T_SHL: op = OP_SHL; break;
		case T_SHR: op = OP_SHR; break;
		}
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr12_addops(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(op));
	}
	_scanner->back();

}

// expr12 ::= <expr13> { ("+" | "-") <expr13>}
/**
handle the '+' and '-' operators
*/
void BppCompiler::_codeExpr12_addops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr13_mulops(context, varInfo, errCnt);
	BppToken tkn;
	BppOpCode op = OP_NIL;
	BppChar c;
	while (((tkn = _scanner->getToken(errCnt)) == T_ANY)
		&& (((c = _scanner->getChar()) == _T('+')) || (c == _T('-'))))
	{
		switch (c)
		{
		case _T('+'): op = OP_ADD; break;
		case _T('-'): op = OP_SUB; break;
		}
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr13_mulops(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(op));
	}
	_scanner->back();
}

// expr13 ::= <expr14> { ("*" | "/" | "%") <expr14> }
/**
handle the '*','/' and '%' operators
*/
void BppCompiler::_codeExpr13_mulops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr14_preops(context, varInfo, errCnt);
	BppToken tkn;
	BppOpCode op = OP_NIL;
	BppChar c;
	while (((tkn = _scanner->getToken(errCnt)) == T_ANY)
		&& (((c = _scanner->getChar()) == _T('*')) || (c == _T('/')) || (c == _T('%'))))
	{
		switch (c)
		{
		case _T('*'): op = OP_MUL; break;
		case _T('/'): op = OP_DIV; break;
		case _T('%'): op = OP_REM; break;
		}
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));

		_codeExpr14_preops(context, varInfo, errCnt);
		_codeRValue(*context.codeVec, varInfo);
		context.codeBuffer->push_back(BppOpCodeByte(op));
	}
	_scanner->back();
}



// expr14 ::=  (("+" | "-" | "!" | "~") <expr14>) | (["++" | "--" | "new" | "delete" | "-" | "!" | "~"] <expr15>)
/**
do_expr14 - handle unary operators
*/
void BppCompiler::_codeExpr14_preops(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;
	BppVector& codeVec = *context.codeVec;
	BppToken tkn = _scanner->getToken(errCnt);
	switch (tkn)
	{
	case T_INC:
		_codePreIncDec(context, OP_INC, varInfo, errCnt);
		break;
	case T_DEC:
		_codePreIncDec(context, OP_DEC, varInfo, errCnt);
		break;
	case T_NEW:
		_codeNew(context, varInfo, errCnt);
		break;
	case T_DELETE:
		_codeDelete(context, varInfo, errCnt);
		break;
	case T_ANY:
	{
		BppChar c = _scanner->getChar();
		switch (c)
		{
		case _T('-'):
			_codeExpr14_preops(context, varInfo, errCnt);
			_codeRValue(codeVec, varInfo);
			buf.push_back(BppOpCodeByte(OP_NEG));
			break;
		case _T('!'):
			_codeExpr14_preops(context, varInfo, errCnt);
			_codeRValue(codeVec, varInfo);
			buf.push_back(BppOpCodeByte(OP_NOT));
			break;
		case _T('~'):
			_codeExpr14_preops(context, varInfo, errCnt);
			_codeRValue(codeVec, varInfo);
			buf.push_back(BppOpCodeByte(OP_BNOT));
			break;
		case _T('+'):
			_codeExpr14_preops(context, varInfo, errCnt);
			_codeRValue(codeVec, varInfo);
			buf.push_back(BppOpCodeByte(OP_PLUS));
			break;
		case _T('&'):
			_codeExpr14_preops(context, varInfo, errCnt);
			_codeRValue(codeVec, varInfo);
			buf.push_back(BppOpCodeByte(OP_MKREF)); // make reference
			break;
		case _T('*'):
			_codeExpr14_preops(context, varInfo, errCnt);
			_codeRValue(codeVec, varInfo);
			buf.push_back(BppOpCodeByte(OP_UNREF)); // make value from reference
			break;
		default:
			_scanner->back();
			_codeExpr15_post(context, varInfo, errCnt);
			return;
		}
	}
	break;
	default:
		_scanner->back();
		_codeExpr15_post(context, varInfo, errCnt);
		return;
	}
}

// expr15 ::= <primary>  <primarysuffix> { <primarysuffix> } .
// primarysuffix ::= <funcop> | <indexop> | <methodcall>  | "--" | "++" .
// methodcall ::= "->" <funcop> .
/**
handle function calls, index operations, method calls, postfix inc/dec
*/
void BppCompiler::_codeExpr15_post(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppToken tkn;
	BppChar c = 0;
	_codePrimary(context, varInfo, errCnt);
	while (
		(((tkn = _scanner->getToken(errCnt)) == T_ANY) &&
			(((c = _scanner->getChar()) == _T('(')) ||
				(c == _T('[')) ||
				(c == _T('.'))
				))
		|| tkn == T_MEMREF
		|| tkn == T_INC
		|| tkn == T_DEC
		)
	{
		switch (tkn)
		{
		case T_ANY:
			if (c == _T('('))
				_codeFunctionCall(context, varInfo, errCnt);
			else if (c == _T('['))
				_codeIndexedOp(context, varInfo, errCnt);
			else if (c == _T('.')) {
				_codePropertyOp(context, varInfo, errCnt);
			}
			break;
		case T_MEMREF:
			if (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T('(')))
				_codeMethodCall(0, varInfo, context, errCnt);
			else
			{
				if (!_require(T__DFUNCTION, tkn, T_IDENTIFIER)) errCnt++;
				_codeMethodCall(_scanner->getTokenString(), varInfo, context, errCnt);
			}
			break;
		case T_INC:
			_codePostIncDec(context, OP_PINC, varInfo, errCnt);
			break;
		case T_DEC:
			_codePostIncDec(context, OP_PDEC, varInfo, errCnt);
			break;
		}
	} // end while
	_scanner->back();
}


void BppCompiler::_codePreIncDec(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt)
{
	_codeExpr15_post(context, varInfo, errCnt);
	if (varInfo.type == VAR_MEMBER)
	{
		if (!varInfo.isMutable && context.constThis &&
			//	varInfo.reqClass &&
			context.contextClass/* &&
			context.contextClass->isSubtypeOf(varInfo.reqClass)*/)
			varInfo.isConst = true;
	}
	if (!_checkLValue(varInfo)) errCnt++;
	_codeVarInfo(*context.codeVec, VA_DUP, varInfo);
	_codeVarInfo(*context.codeVec, VA_LOAD, varInfo);
	context.codeBuffer->push_back(BppOpCodeByte(op));
	_codeVarInfo(*context.codeVec, VA_STORE, varInfo);
	varInfo.type = VAR_DONE;
}

void BppCompiler::_codePostIncDec(BppFuncContext& context, BppOpCode op, BppVarInfo& varInfo, int& errCnt)
{
	if (varInfo.type == VAR_MEMBER)
	{
		if (!varInfo.isMutable && context.constThis &&
			//	varInfo.reqClass &&
			context.contextClass/* &&
			context.contextClass->isSubtypeOf(varInfo.reqClass)*/)
			varInfo.isConst = true;
	}
	if (!_checkLValue(varInfo)) ++errCnt;
	_codeVarInfo(*context.codeVec, VA_DUP, varInfo);
	_codeVarInfo(*context.codeVec, VA_LOAD, varInfo);
	context.codeBuffer->push_back(BppOpCodeByte(OP_STOS));
	context.codeBuffer->push_back(BppOpCodeByte(op));
	_codeVarInfo(*context.codeVec, VA_STORE, varInfo);
	context.codeBuffer->push_back(BppOpCodeByte(OP_LDTOS));
	varInfo.type = VAR_DONE;
}


void BppCompiler::_codeNew(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{

	if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
	{
		errCnt++;
		return;
	}
	BppString classPath = parseNamePath(getActiveScanner()->_token, errCnt);
	BppClass* pclass = _findClass(classPath, errCnt, context.contextClass);
	BppVector& codeVec = *context.codeVec;
	codeVec.push_back(pclass);
	_codeLiteral(codeVec, codeVec.size() - 1);

	context.codeBuffer->push_back(BppOpCodeByte(OP_NEW));
	varInfo.type = VAR_DONE;

	_codeMethodCall((pclass ? pclass->getClassName()->c_str() : classPath.c_str()), varInfo, context, errCnt);
}

void BppCompiler::_codeDelete(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppBuffer& buf = *context.codeBuffer;
	BppVector& codeVec = *context.codeVec;
	_codePrimary(context, varInfo, errCnt); // parse expr behind delete
	_codeRValue(codeVec, varInfo);
	buf.push_back(BppOpCodeByte(OP_DELETE));
	varInfo.type = VAR_DONE;
}


static bool containsIdx(const BppDictionary& dict, int idx)
{
	for (BppDictionaryEntry* p = dict.getFirst(); p; p = p->getNext())
	{
		if (((BppInt)idx) == (BppInt)p->getValue()) return true;
	}
	return false;
}
// primary ::= <int> | <float> | <string> | "null" | "nil" | <variable> | <static_member> | <globalref> 
// variable ::= <identifier>
// static_member ::= <identifier> :: <identifier>
// globalref ::= "::" <identifier>
void BppCompiler::_codePrimary(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	BppVector& codeVec = *context.codeVec;
	BppToken tkn = _scanner->getToken(errCnt);
	switch (tkn)
	{
	case T_NUMBER:
	case T_FNUMBER:
	{
		BppValue v = _scanner->getNumberValue();
		varInfo.varvtype = v.getType();
		size_t idx = codeVec.rfind(v);

		if (idx == BPP_NNPOS || (containsIdx(context.statics, idx)))
		{
			idx = codeVec.size();
			codeVec.push_back(v);
		}
		_codeLiteral(codeVec, idx);
		varInfo.type = VAR_DONE;
	}
	break;
	case T_STRING:
	{
		// reuse string literal if it already exists
		const BppChar* str = _scanner->getTokenString();
		BppValue v(new BppString(str));
		varInfo.varvtype = v.getType();
		size_t idx = codeVec.rfind(v);
		if (idx == BPP_NNPOS || (containsIdx(context.statics, idx)))
		{
			idx = codeVec.size();
			codeVec.push_back(v);
		}
		_codeLiteral(codeVec, (BppNWord)idx);
	}
	varInfo.type = VAR_DONE;
	break;
	case T_NIL:
	{
		BppValue v;
		varInfo.varvtype = v.getType();
		size_t idx = codeVec.rfind(v);
		if (idx == BPP_NNPOS || (containsIdx(context.statics, idx)))
		{
			idx = codeVec.size();
			codeVec.push_back(v);
		}
		_codeLiteral(codeVec, idx);
		varInfo.type = VAR_DONE;
	}
	break;
	// putcbyte(OP_NIL);
//       break;
	case T_FUNCTION:
	case T_SHL:
	{
		BppDictionaryEntry dummy(0, BppValue());
		_doLiteralValue(tkn, dummy, errCnt);
		const BppValue& v = dummy.getValue();
		varInfo.varvtype = v.getType();
		size_t idx = codeVec.rfind(v);
		if (idx == BPP_NNPOS || (containsIdx(context.statics, idx)))
		{
			idx = codeVec.size();
			codeVec.push_back(v);
		}
		_codeLiteral(codeVec, idx);
		varInfo.type = VAR_DONE;
		break;
	}
	case T_IDENTIFIER:
	{
		BppString id = _scanner->getTokenString();
		BppDictionaryEntry* entry = _namedLiterals.find(id);
		if (entry != 0)
		{
			codeVec.push_back(entry->getValue());
			varInfo.varvtype = entry->getValue().getType();
			_codeLiteral(codeVec, codeVec.size() - 1);
			varInfo.type = VAR_DONE;
			break;
		}
		tkn = _scanner->getToken(errCnt);
		if (tkn == T_CC)
		{
			BppValue ownClass = codeVec[IDX_CLASS]; // save own class reference
			BppClass* pown = ownClass;
			BppClass* pother = _findClass(id, errCnt,context.contextClass); // temporary set other class
			if (pother == 0) return;
			while ((tkn = _scanner->getToken(errCnt)) == T_IDENTIFIER) 
			{
				BppString tmpid = id;
				tmpid += _T("::");
				tmpid += _scanner->getTokenString();
				BppClass* p = _findClass(tmpid, errCnt, context.contextClass,true);
				if (!p) {
					break;               
				}
				pother = p;
				id = tmpid;
				if ((tkn = _scanner->getToken(errCnt)) != T_CC) {
					break;
				}
			}
			_scanner->back();
			codeVec[IDX_CLASS] = pother;
			if (_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
			{
				BppDictionaryEntry* de = pother->findMember(pown, _scanner->getTokenString());
				if (de != 0)
				{
					BppSymbolType st = de->getSymbolType();
					if (st == ST_DATA)
					{
						if (!(pown && pown->isSubtypeOf(pother)))
						{
							_error(CMP_NOT_STATIC_MEMBER, &errCnt);
						}
					}
					if (st == ST_FUNCTION)
					{
						if (_scanner->getToken(errCnt) == T_ANY && _scanner->getChar() == _T('('))
						{
							_error(CMP_INVALID_NONVIRTUAL_MCALL, &errCnt);
						}
						_scanner->back();
					}
				}
				else
				{
					_error(CMP_NOT_CLASS_MEMBER, &errCnt);
				}
				varInfo = _findClassVariable(_scanner->getTokenString(), context, errCnt, pown);
			}
			else
			{
				errCnt++;
				varInfo = BppVarInfo();
			}
			codeVec[IDX_CLASS] = ownClass; // reset class reference
		}
		else
		{
			_scanner->back();
			varInfo = _findVariable(id.c_str(), context, false, errCnt);
		}
	}
	break;
	case T_CC:                   // erzwungener Globalverweis
		if (_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER))
			varInfo = _findVariable(_scanner->getTokenString(), context, true, errCnt);
		else
		{
			errCnt++;
			varInfo = BppVarInfo();
		}
		break;
	case T_LITEXPR:
	{
		BppDictionaryEntry dummy(0, BppValue());
		_doLiteralValue(tkn, dummy, errCnt);
		codeVec.push_back(dummy.getValue());
		varInfo.varvtype = dummy.getValue().getType();
		_codeLiteral(codeVec, codeVec.size() - 1);
		varInfo.type = VAR_DONE;
	}
	break;
	case T_ANY:
	{
		switch (_scanner->getChar())
		{
		case _T('('):
		{
			_codeExpr1_comma(context, varInfo, errCnt);
			if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
			return;
		}
		case _T('['):
		case _T('{'):
		{
			BppDictionaryEntry dummy(0, BppValue());
			_doLiteralValue(tkn, dummy, errCnt);
			codeVec.push_back(dummy.getValue());
			varInfo.varvtype = dummy.getValue().getType();
			_codeLiteral(codeVec, codeVec.size() - 1);
			varInfo.type = VAR_DONE;
			return;
		}
		}
	}
	case T_ERROR:
		_print(_scanner->getTokenString());
		errCnt++;
		break;
	default:
		_error(CMP_PRIMARY_EXPECTED, &errCnt);
		break;
	}
}

/**
@param name Path for inner class of <i>ownerclass</i> or full class path (if <i>ownerClass</i> is NULL
@param errCnt error counter
@param ownerClass Owner (outer) class of class specified by <i>name</i>.
@param allowNotFound Specifies if an error should signaled, if class not found (false) or not (true).
@return Pointer to class or NULL.
*/
BppClass* BppCompiler::_findClass(const BppString& name, int& errCnt, BppClass* ownerClass, bool allowNotFound)
{
	static const BppString CC = _T("::");
	size_t ccpos = name.find(CC);
	BppDictionaryEntry* classentry = 0;
	if (ownerClass && ccpos != 0) {
		classentry = _classes.find(BppString::format(_T("%s::%s"),
			ownerClass->getClassPath().c_str(),
			name.c_str()));
		if (!classentry) 
			return _findClass(name, errCnt, ownerClass->getOwnerClass(), allowNotFound);
		//if (!classentry && name.find(*ownerClass->getClassName()) == 0) {
		//	if (ccpos == BPP_NNPOS || ccpos == name.length()) {
		//		return _findClass(name, errCnt, ownerClass->getOwnerClass(), allowNotFound);
		//	}
		//}		    
	}
	if (!classentry) {
		if (ccpos == 0) classentry = _classes.find(name.c_str() + 2);
		else classentry = _classes.find(name);
	}
	if (classentry == 0)
	{
		if (!allowNotFound) _error(BppString::format(_T("%s %s"), CMP_UNDEFINED_CLASS, name.c_str()).c_str(), &errCnt);
		return 0;
	}
	return classentry->getValue();
}

BppVarInfo BppCompiler::_findVariable(const BppChar* name, BppFuncContext& context, bool global, int& errCnt)
{
	BppVarInfo result;

	size_t n = context.temporaries.getStringIndex(name);
	if ((!global) && (n != BPP_NNPOS))
	{
		result.type = VAR_LOCAL;
		result.index = n;
		result.isConst = (context.temporaries[n].getFlags() & BVF_CONST) != 0;
	}
	else if ((!global) && ((n = context.arguments.getStringIndex(name)) != BPP_NNPOS))
	{
		result.type = VAR_ARG;
		result.index = n;
		result.isConst = (context.arguments[n].getFlags() & BVF_CONST) != 0;
	}
	else
	{
		BppString nm = name;
		BppDictionaryEntry* pstatic = context.statics.find(nm);
		if ((!global) && (pstatic != 0))
		{
			result.type = VAR_SLOCAL;
			result.index = (size_t)(BppUInt)pstatic->getValue();
			result.isConst = ((*context.codeVec)[result.index].getFlags() & BVF_CONST) != 0;
		}
		else
		{
			if (!global)
			{
				result = _findClassVariable(name, context, errCnt, context.contextClass);
				if ((result.index != BPP_NNPOS) || (result.type == VAR_DONE))
					return result;
			}
			if (_tcscmp(name, SYMN_self) == 0)
			{
				_codeLiteral(*(context.codeVec), IDX_SELFREFERENCE);
				result.index = IDX_SELFREFERENCE;
				result.type = VAR_DONE;
				return result;
			}
			BppDictionaryEntry* sym = _symbols.find(name);
			if (sym)
			{
				result.type = VAR_VAR;

				// search for previous reference to sym first
				result.index = context.codeVec->getVarIndex(sym);
				if (result.index == BPP_NNPOS)
				{
					// add reference if not found
					context.codeVec->push_back(sym);
					result.index = context.codeVec->size() - 1;
				}
				result.isConst = (sym->getValue().getFlags() & BVF_CONST) != 0;
				return result;
			}
			sym = _classes.find(name);
			if (!sym)
			{
				_error(BppString::format(CMP_UNDEFINED_VAR_FMT, name).c_str(), &errCnt);
				return result;
			}
			const BppValue& classVal = sym->getValue();
			BppClass* pclass = classVal;
			result.index = BPP_NNPOS;
			n = context.codeVec->size();
			for (size_t i = 0; i < n; ++i)
			{
				if (((BppClass*)((*context.codeVec)[i])) == pclass)
				{
					result.index = i;
					break;
				}
			}
			if (result.index == BPP_NNPOS)
			{
				context.codeVec->push_back(sym->getValue());
				result.index = context.codeVec->size() - 1;
			}
			_codeLiteral(*context.codeVec, result.index);
			result.type = VAR_DONE;

		}
	}
	return result;
}


BppVarInfo BppCompiler::_findClassVariable(const BppChar* name, BppFuncContext& context, int& errCnt, BppClass* refClass)
{
	BppVarInfo result;
	BppClass* ownerClass = (*context.codeVec)[IDX_CLASS];
	if (ownerClass == 0) return result;

	bool isstaticself = false;
	if (_tcscmp(name, SYMN_self) == 0)
	{
		BppVarInfo selfresult;
		if (context.arguments.getStringIndex(SYMN_this) != BPP_NNPOS) {
			selfresult = _findVariable(SYMN_this, context, false, errCnt);

			if (selfresult.type == VAR_ARG)
			{
				_codeMethodCall(name, selfresult, context, errCnt);
				return selfresult;
			}
		}
		else isstaticself = true;

	}

	BppString nm = name;
	BppDictionaryEntry* entry = isstaticself ? (BppDictionaryEntry*)(*context.codeVec)[IDX_SELFREFERENCE] : ownerClass->findMember(refClass, nm);
	if (entry == 0) {

		return result;
	}
	switch (entry->getSymbolType())
	{
	case ST_DATA:
	{
		result.type = VAR_MEMBER;
		// index of member variable is stored in value of dictionary entry
		BppVector* v = entry->getValue();
		if (!v)
		{
			_error(CMP_INVALID_DATA_MEMBER, &errCnt);
			break;
		}
		result.index = (size_t)(BppUInt)(*v)[0];
		if (entry->getValue().getFlags() & BVF_MUTABLE)
			result.isMutable = true;
	}
	break;
	case ST_FUNCTION:
		if (_scanner->getToken(errCnt) == T_ANY && _scanner->getChar() == _T('('))
		{
			_scanner->back();
			result = _findVariable(SYMN_this, context, false, errCnt);
			_codeMethodCall(name, result, context, errCnt);
			break;
		}
		else
			_scanner->back();
	case ST_SDATA:
		result.type = VAR_VAR;
		// search for previous reference to entry first
		result.index = context.codeVec->getVarIndex(entry);
		if (result.index == BPP_NNPOS)
		{
			// add reference if not found
			context.codeVec->push_back(entry);
			result.index = context.codeVec->size() - 1;
		}
		if (entry->getValue().getFlags() & BVF_CONST)
			result.isConst = true;
		break;
	case ST_SFUNCTION:
		// search for previous reference to entry first
		result.index = context.codeVec->getVarIndex(entry);
		if (result.index == BPP_NNPOS)
		{
			// add reference if not found
			context.codeVec->push_back(entry);
			result.index = context.codeVec->size() - 1;
		}
		_codeVariable(*context.codeVec, VA_LOAD, result.index);
		result.type = VAR_DONE;
		break;
	}
	return result;
}

// code receiver value
void BppCompiler::_codeRValue(BppVector& codeVec, BppVarInfo& varInfo)
{
	switch (varInfo.type)
	{
	case VAR_ARG:
		_codeArgument(codeVec, VA_LOAD, varInfo.index);
		break;
	case VAR_LOCAL:
		_codeTemporary(codeVec, VA_LOAD, varInfo.index);
		break;
	case VAR_SLOCAL:
		_codeStaticLocal(codeVec, VA_LOAD, varInfo.index);
		break;
	case VAR_VAR:
		_codeVariable(codeVec, VA_LOAD, varInfo.index);
		break;
	case VAR_MEMBER:
		_codeMember(codeVec, VA_LOAD, varInfo.index);
		break;
	case VAR_INDEX:
		_codeIndexRef(codeVec, VA_LOAD);
		break;
	case VAR_PROP:
		_codePropertyRef(codeVec, VA_LOAD);
		break;
	default:
		break;
	}
	varInfo.type = VAR_DONE;
}

void BppCompiler::_codeVarInfo(BppVector& codeVec, BppVarAccessType act, BppVarInfo& varInfo)
{
	BppBuffer* cbuf = codeVec[IDX_CODE];
	///// nothing, variable doesn't exist
	//VAR_NONE = 0,
	//	/// argument
	//	VAR_ARG,
	//	/// temporary
	//	VAR_LOCAL,
	//	/// static local variable in function
	//	VAR_SLOCAL,
	//	/// global variable or static member variable
	//	VAR_VAR,
	//	/// member variable
	//	VAR_MEMBER,
	//	/// index variable
	//	VAR_INDEX,
	//	/// property member
	//	VAR_PROP,
	//	/// ignore - variable already processed
	//	VAR_DONE

	switch (varInfo.type)
	{
	case VAR_ARG:
		if (/*varInfo.type*/act == VA_STORE && varInfo.varvtype != BVT_NULL)
		{
			_codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
			_codeArgument(codeVec, act, varInfo.index);
			cbuf->push_back(BppOpCodeByte(OP_POP));
		}
		_codeArgument(codeVec, act, varInfo.index);
		break;
	case VAR_LOCAL:
		if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
		{
			_codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
			_codeTemporary(codeVec, act, varInfo.index);
			cbuf->push_back(BppOpCodeByte(OP_POP));
		}
		_codeTemporary(codeVec, act, varInfo.index);
		break;
	case VAR_SLOCAL:
		if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
		{
			_codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
			_codeStaticLocal(codeVec, act, varInfo.index);
			cbuf->push_back(BppOpCodeByte(OP_POP));
		}
		_codeStaticLocal(codeVec, act, varInfo.index);
		break;
	case VAR_VAR:
		if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
		{
			_codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
			_codeVariable(codeVec, act, varInfo.index);
			cbuf->push_back(BppOpCodeByte(OP_POP));
		}
		_codeVariable(codeVec, act, varInfo.index);
		break;
	case VAR_MEMBER:
		if (act == VA_STORE && varInfo.varvtype != BVT_NULL)
		{
			_codeIndexedOpcode(codeVec, OP_PUSH, varInfo.varvtype);
			_codeVariable(codeVec, act, varInfo.index);
			cbuf->push_back(BppOpCodeByte(OP_POP));
		}
		_codeMember(codeVec, act, varInfo.index);
		break;
	case VAR_INDEX:
		_codeIndexRef(codeVec, act);
		break;
	case VAR_PROP:
		_codePropertyRef(codeVec, act);
		break;
	default:
		break;
	}
}


/// compile an argument
void BppCompiler::_codeArgument(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
	switch (act)
	{
	case VA_LOAD:
		_codeIndexedOpcode(codeVec, OP_AREF, index);
		break;
	case VA_STORE:
		_codeIndexedOpcode(codeVec, OP_ASET, index);
		break;
	}
}
/// compile a temporary (local variable)
void BppCompiler::_codeTemporary(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
	switch (act)
	{
	case VA_LOAD:
		_codeIndexedOpcode(codeVec, OP_TREF, index);
		break;
	case VA_STORE:
		_codeIndexedOpcode(codeVec, OP_TSET, index);
		break;
	}
}

void BppCompiler::_codeStaticLocal(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
	switch (act)
	{
	case VA_LOAD:
		_codeIndexedOpcode(codeVec, OP_SREF, index);
		break;
	case VA_STORE:
		_codeIndexedOpcode(codeVec, OP_SSET, index);
		break;
	}
}

/// compile a variable reference
void BppCompiler::_codeVariable(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
	switch (act)
	{
	case VA_LOAD:
		_codeIndexedOpcode(codeVec, OP_REF, index);
		break;
	case VA_STORE:
		_codeIndexedOpcode(codeVec, OP_SET, index);
		break;
	}
}
void BppCompiler::_codeMember(BppVector& codeVec, BppVarAccessType act, BppNWord index)
{
	switch (act)
	{
	case VA_LOAD:
		_codeIndexedOpcode(codeVec, OP_MREF, index);
		break;
	case VA_STORE:
		_codeIndexedOpcode(codeVec, OP_MSET, index);
		break;
	}
}


/// compile a literal reference
void BppCompiler::_codeLiteral(BppVector& codeVec, BppNWord index)
{
	_codeIndexedOpcode(codeVec, OP_LIT, index);
}

/// compile an indexed reference
/**
Pushes opcode depending on <i>act</i> to end of code buffer (<i>codeVec</i>[IDX_CODE]).
@see _codeRValue, _codeVarInfo, BppVarAccessType
*/
void BppCompiler::_codeIndexRef(BppVector& codeVec, BppVarAccessType act)
{
	BppOpCode oc = OP_NOP;
	switch (act)
	{
	case VA_LOAD:
		oc = OP_VREF;
		break;
	case VA_STORE:
		oc = OP_VSET;
		break;
	case VA_PUSH:
		oc = OP_PUSH;
		break;
	case VA_DUP:
		oc = OP_DUP2;
		break;
	default:
		return;
	}
	BppBuffer* buf = codeVec[IDX_CODE];
	buf->push_back(BppOpCodeByte(oc));
}

/// compile an indexed reference
/**
Pushes opcode depending on <i>act</i> to end of code buffer (<i>codeVec</i>[IDX_CODE]).
@see _codeRValue, _codeVarInfo, BppVarAccessType
*/
void BppCompiler::_codePropertyRef(BppVector& codeVec, BppVarAccessType act)
{
	BppOpCode oc = OP_NOP;
	switch (act)
	{
	case VA_LOAD:
		oc = OP_PREF;
		break;
	case VA_STORE:
		oc = OP_PSET;
		break;
	case VA_PUSH:
		oc = OP_PUSH;
		break;
	case VA_DUP:
		oc = OP_DUP2;
		break;
	default:
		return;
	}
	BppBuffer* buf = codeVec[IDX_CODE];
	buf->push_back(BppOpCodeByte(oc));
}

void BppCompiler::_codeMethodCall(const BppChar* name, BppVarInfo& thisInfo, BppFuncContext& context, int& errCnt)
{
	// get the receiver value (this)
	_codeRValue(*context.codeVec, thisInfo);
	if (thisInfo.type != VAR_DONE)
	{
		_error(CMP_NO_THISREF_FOUND, &errCnt);
	}
	BppBuffer* buf = context.codeBuffer;
	// generate code to push the selector (member function name)
	size_t index = BPP_NNPOS;
	buf->push_back(BppOpCodeByte(OP_PUSH));
	if (name)
	{
		index = (_tcscmp(name, SYMN_self) == 0) ? IDX_SELFREFERENCE : context.codeVec->getStringIndex(name);
		if (index == BPP_NNPOS)
		{
			context.codeVec->push_back(new BppString(name));
			index = context.codeVec->size() - 1;
		}
		_codeLiteral(*context.codeVec, index);
	}
	else // evaluate expression to get method or method name 
	{
		_codeExpr(context, errCnt);
		index = context.codeVec->size() - 1;
		if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(')'))) errCnt++;
	}
	BppVector* defargs = 0;
	size_t ndefargs = 0;
	size_t nrequargs = 0;
	_getCalledFunctionInfo(context, index, defargs, ndefargs, nrequargs,true);
	int argc = 1; // first argument is 'this'
	// compile the argument list 
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T('('))) errCnt++;
	BppToken tkn = _scanner->getToken(errCnt);
	if (!((tkn == T_ANY) && (_scanner->getChar() == _T(')'))))
	{
		_scanner->back();
		do
		{
			buf->push_back(BppOpCodeByte(OP_PUSH));
			//size_t bsize = buf->size();
			_codeExpr2_assign(context, thisInfo, errCnt);
			_codeRValue(*context.codeVec, thisInfo);
			++argc;
			//_adjustAssignment(*buf,bsize);
		} while (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T(',')));
	}
	if (!_require(T__DFUNCTION, T_ANY, tkn, _T(')'))) errCnt++;
	if (argc < nrequargs)
		_error(BCI_TOO_FEW_ARGS, &errCnt);
	// do the call
	_codeIndexedOpcode(*context.codeVec, OP_SEND, (BppNWord)argc);
	thisInfo.type = VAR_DONE;
}

// funcop ::= "(" [ <funcargs> ] ")" .
// funcargs ::= <expr2> { "," <expr2> } .
void BppCompiler::_codeFunctionCall(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	/* get the value of the function */
	_codeRValue(*context.codeVec, varInfo);

	BppToken tkn = _scanner->getToken(errCnt);
	BppChar c = (tkn == T_ANY) ? _scanner->getChar() : 0;
	BppNWord n = 0;
	BppBuffer& buf = *context.codeBuffer;
	/* compile each argument expression */
	size_t ndefargs = 0;
	BppVector* defargs = 0;
	size_t nrequargs = 0;
	_getCalledFunctionInfo(context, varInfo.index, defargs, ndefargs, nrequargs);
	if (c != _T(')'))
	{
		_scanner->back();
		do
		{
			buf.push_back(BppOpCodeByte(OP_PUSH));
			//size_t bsize = buf.size();
			_codeExpr2_assign(context, varInfo, errCnt);
			_codeRValue(*context.codeVec, varInfo);
			++n;
			//_adjustAssignment(buf,bsize);
			tkn = _scanner->getToken(errCnt);
			c = (tkn == T_ANY) ? _scanner->getChar() : 0;
		} while (c == _T(','));
	}
	if (n < nrequargs)
		_error(BCI_TOO_FEW_ARGS, &errCnt);
	if (!_require(T__DFUNCTION, T_ANY, tkn, _T(')'))) errCnt++;
	_codeIndexedOpcode(*context.codeVec, OP_CALL, n);

	/* we've got an rvalue now */
	varInfo.type = VAR_DONE;
}

BppDictionaryEntry* BppCompiler::_getCalledFunctionInfo(const BppFuncContext& context, size_t varIndex, BppVector*& defargs, size_t& ndefargs, size_t& nrequargs, bool isMemberFunc) const
{
	defargs = 0;
	ndefargs = 0;
	nrequargs = 0;
	if (context.codeVec->size() > varIndex) {
		BppDictionaryEntry* pe = (*context.codeVec)[varIndex];
		if (pe) {
			const BppValue& v = pe->getValue();
			if (maskWeakRef(v.getType()) == BVT_BYTECODE) {
				BppVector* funcdef = v;
				if (funcdef && funcdef->size() > IDX_DEFAULTARGS) {
					defargs = (*funcdef)[IDX_DEFAULTARGS];
					if (defargs) {
						ndefargs = defargs->size();
						nrequargs = ndefargs;
						for (size_t i = 0; i < ndefargs; ++i)
						{
							if (((*defargs)[i].getFlags() & BVF_INVALID) != BVF_INVALID) {
								if (!isMemberFunc || i > 0) {
									nrequargs = i;
									break;
								}
							}
						}
					}
				}
			}
			return pe;
		}
	}
	return 0;
}

// indexop ::= "[" <expr> "]" .
/**
parses indexed operation
*/
void BppCompiler::_codeIndexedOp(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	//    int code_index();
	_codeRValue(*context.codeVec, varInfo);
	BppBuffer* buf = context.codeBuffer;
	size_t bsize = buf->size();
	if ((bsize > 1) && (BppOpCodeByte((*buf)[bsize - 2]).opCode() == OP_LIT))
	{
		_error(CMP_INDEXED_LITREF_NOT_ALLOWED, &errCnt);
	}
	context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
	_codeExpr(context, errCnt);
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_ANY, _T(']'))) errCnt++;
	varInfo.type = VAR_INDEX;
}

// indexop ::= "." <ident>  .
/**
parses indexed operation
*/
void BppCompiler::_codePropertyOp(BppFuncContext& context, BppVarInfo& varInfo, int& errCnt)
{
	//    int code_index();
	_codeRValue(*context.codeVec, varInfo);
	context.codeBuffer->push_back(BppOpCodeByte(OP_PUSH));
	if (!_fetchRequire(T__DFUNCTION, errCnt, T_IDENTIFIER)) errCnt++;

	// reuse string literal if it already exists
	const BppChar* str = _scanner->getTokenString();
	size_t idx = context.codeVec->getStringIndex(str);
	if (idx == BPP_NNPOS)
	{
		idx = context.codeVec->size();
		context.codeVec->push_back(new BppString(str));
	}
	_codeLiteral(*context.codeVec, (BppNWord)idx);
	varInfo.type = VAR_PROP;
}

// idlist ::= <identifier> { "," <identifier> } .
int BppCompiler::_getIdList(BppVector& dest, BppClass* contextClass, const BppChar* termStr, int& errCnt, BppVector* defaults, bool useTypeNames, const BppChar* ellipseSymbol)
{
	int cnt = 0;
	bool haveEllipse = false;
	BppToken tkn = _scanner->getToken(errCnt);
	if ((tkn != T_ANY) || (!_tcschr(termStr, _scanner->getChar())))
	{
		_scanner->back();
		do
		{
			if (haveEllipse) _error(CMP_ELLIPSE_MUST_BE_LAST, &errCnt);
			BppToken t = _scanner->getToken(errCnt);
			if (t == T_ELLIPSE) {
				haveEllipse = true;
				if (ellipseSymbol) {
					dest.push_back(new BppString(ellipseSymbol));
				}
				continue;
			}
			BppValueType vtype = BVT_NULL;
			BppClass* pvClass = 0;
			bool isConst = false;
			if (defaults != 0) {
				if (useTypeNames) {
					if (t == T_CONST) {
						isConst = true;
						t = _scanner->getToken(errCnt);
					}
					if (t == T_VAR)
						t = _scanner->getToken(errCnt);
					else if (t == T_IDENTIFIER || t == T_FUNCTION)
					{
						BppString name = _scanner->getTokenString();

						if (t != T_FUNCTION) {
							name = parseNamePath(t, errCnt);
						}
						vtype = findVTypeFromName(T__DFUNCTION, name, errCnt, true, &pvClass,contextClass);
						if (vtype != BVT_NULL)
							t = _scanner->getToken(errCnt);
					}
				}
			}
			if (!_require(T__DFUNCTION, T_IDENTIFIER, t))
			{
				errCnt++;
				continue;
			}
			const BppChar* p = _scanner->getTokenString();
			if (dest.getStringIndex(p) != BPP_NNPOS)
			{
				_error(BppString::format(CMP_DUPLICATED_IDENT_FMT, p).c_str(), &errCnt);
			}
			else
			{
				BppValue v = new BppString(p);
				if (isConst) v.setFlags(BVF_CONST);
				dest.push_back(BPPMOVE(v));
				unsigned char flags = isConst ? BVF_CONST : BVF_NONE;
				++cnt;
				if (defaults != 0)
				{
					if ((_scanner->getToken(errCnt) != T_ANY) || (_scanner->getChar() != _T('=')))
					{
						_scanner->back();
						defaults->push_back(BppValue(vtype, (BppValueFlags)(vtype == BVT_NULL ? BVF_NONE : BVF_STRONGTYPED), false));
						BppValue& vr = (*defaults)[defaults->size() - 1];

						vr.setFlags(vr.getFlags() | (BVF_INVALID | flags));
						continue;
					}
					BppDictionaryEntry tmp(0, BppValue());
					_doLitExpr(_scanner->getToken(errCnt), tmp, errCnt);
					if (vtype != BVT_NULL)
					{
						BppRefObject* pro = 0;
						if (vtype == BVT_STRONGOBJECT && pvClass)
						{
							BppVector* pdesc = new BppVector(2, true);
							(*pdesc)[IDX_CLASS] = pvClass;
							pro = pdesc;
						}
						BppValue tmpv(pro, vtype);
						tmpv.setFlags(BVF_STRONGTYPED);
						if (!tmpv.assign(tmp.getValue()))
						{
							BppString* pname = dest[dest.size() - 1];
							_error(BppString::format(CMP_TYPE_MISMATCH_FMT3,
								pname->c_str(),
								getValueTypeName(tmpv.getType()),
								getValueTypeName(tmp.getValue().getType())).c_str(), &errCnt);
						}
						tmpv.setFlags(tmpv.getFlags() | flags);
						defaults->push_back(BPPMOVE(tmpv));
					}
					else
					{
						BppValue tmpv = tmp.getValue();
						tmpv.setFlags(flags);
						defaults->push_back(BPPMOVE(tmp.getValue()));
					}
				}
			}
		} while (((tkn = _scanner->getToken(errCnt)) == T_ANY) && (_scanner->getChar() == _T(',')));
	}
	_scanner->back();
	return cnt;
}

/**
Pushes opcode with index argument, if index is greater than byte opcode (or forceNativeSize set) it is prefixed by
argument width code.
*/
void BppCompiler::_codeIndexedOpcode(BppVector& codeVec, BppOpCode opcode, BppNWord index, bool forceNativeSize)
{
	BppBuffer* buf = codeVec[IDX_CODE];
	BppOpCodeByte ocb = (BppByte)opcode;
	if (ocb.opDataType() != ODT_NONE)
	{
		buf->push_back(BppOpCodeByte(OP_NOP, ocb.opDataType()));
		ocb._v._op.opDataType = ODT_NONE;
	}
	if (forceNativeSize)
	{
		ocb._v._op.opDataType = ODT_NWORD;
		buf->push_back(ocb);
		buf->push_backNW(index);
	}
	else if ((index & 0xFF) == index)// Byte
	{
		ocb._v._op.opDataType = ODT_BYTE;
		buf->push_back(ocb);
		buf->push_back((BppByte)index);
	}
	else if ((index & 0xFFFF) == index)// Word
	{
		ocb._v._op.opDataType = ODT_WORD;
		buf->push_back(ocb);
		buf->push_backW(index);
	}
	else // DWord
	{
		ocb._v._op.opDataType = ODT_DWORD;
		buf->push_back(ocb);
		buf->push_backDW(index);
	}
}

size_t BppCompiler::_codeOpcode(BppVector& codeVec, BppOpCode opcode)
{
	BppBuffer* buf = codeVec[IDX_CODE];
	BppOpCodeByte ocb = (BppByte)opcode;
	if (ocb.opDataType() != ODT_NONE)
	{
		buf->push_back(BppOpCodeByte(OP_NOP, ocb.opDataType()));
		ocb._v._op.opDataType = ODT_NONE;
	}
	return buf->push_back(ocb);
}

bool BppCompiler::_fetchRequire(const BppChar* func, int& errCnt, BppToken rtoken, BppChar ch)
{
	return _require(func, rtoken, _scanner->getToken(errCnt), ch);
}

bool BppCompiler::_require(const BppChar* func, BppToken rtoken, BppToken token, BppChar ch)
{
	if ((token != rtoken) || ((token == T_ANY) && (ch != 0) && (ch != _scanner->getChar())))
	{
		if (rtoken == T_ANY)
		{
			BppChar chbuf[2];
			chbuf[0] = ch;
			chbuf[1] = 0;
			_error(BppString::format(CMP_EXPECTED_FOUND_FMT3, func, (const BppChar*)chbuf, _scanner->getTokenName(token)).c_str());
		}
		else
		{
			_error(BppString::format(CMP_EXPECTED_FOUND_FMT3, func, _scanner->getTokenName(rtoken), _scanner->getTokenName(token)).c_str());
		}
		return false;
	}
	return true;
}

bool BppCompiler::_checkLValue(const BppVarInfo& varInfo)
{
	if (varInfo.isConst || (varInfo.type == VAR_NONE) || (varInfo.type == VAR_DONE))
	{
		_error(varInfo.isConst ? BCI_ASSIGN_TO_CONST : CMP_LVALUE_EXPECTED);
		return false;
	}
	return true;
}

void BppCompiler::_error(const BppChar* msg, int* errCnt)
{
	_scanner->parseError(msg);
	_print(_T("%s"), _scanner->getTokenString());
	if (errCnt) ++(*errCnt);
}

BppValueType BppCompiler::findVTypeFromName(const BppChar* func, const BppString& name, int& errCnt, bool includeClasses, BppClass** classType, BppClass* ownerClass)
{
	const BppString nm = name.c_str();
	bool issigned = false;
	if (_tcscmp(nm.c_str(), _T("signed")) == 0)
		issigned = true;
	else if (_tcscmp(nm.c_str(), _T("unsigned")) != 0)
		return _vm.findVTypeFromName(nm, includeClasses, classType, ownerClass);
	if (_fetchRequire(func, errCnt, T_IDENTIFIER)) {
		const BppString tname = _scanner->getTokenString();
		BppValueType basetype = _vm.findVTypeFromName(tname);
		switch (basetype) {
		case BVT_BYTE:
			return issigned ? BVT_SBYTE : BVT_BYTE;
		case BVT_CHAR:
			return issigned ? BVT_CHAR : BVT_UCHAR;
		case BVT_SHORT:
			return issigned ? BVT_SHORT : BVT_USHORT;
		case BVT_INT:
			return issigned ? BVT_INT : BVT_UINT;
		case BVT_LONG:
			return issigned ? BVT_LONG : BVT_ULONG;
		}
	}
	_scanner->back();
	getActiveScanner()->_stringBuffer = issigned ? _T("int") : _T("uint");
	return issigned ? BVT_INT : BVT_UINT;
}

/**
* Funktion parses parts of qualified symbol name.
@param to Start token (current Token), should be of type T_IDENTIFIER.
@param errCnt error counter
@return qualified name parsed
*/
BppString BppCompiler::parseNamePath(BppToken tok, int& errCnt)
{
	BppString name = _scanner->getTokenString();
	if (tok == T_IDENTIFIER) {
		while (T_CC == (tok = _scanner->getToken(errCnt))) {
			if (_scanner->getToken(errCnt) != T_IDENTIFIER) {
				break;
			}
			name += _T("::");
			name += _scanner->getTokenString();
		}
		_scanner->back();
	}
	return name;
}

BppString BppCompiler::parseNamePathParts(BppToken tok, int& errCnt, BppVector& parts, bool storeLastPart, bool returnFullPath)
{
	BppString name = _scanner->getTokenString();
	if (tok == T_IDENTIFIER) {
	    parts.push_back(name);
		while (T_CC == (tok = _scanner->getToken(errCnt))) {
			if (_scanner->getToken(errCnt) != T_IDENTIFIER) {
				break;
			}
			name = _scanner->getTokenString();
			parts.push_back(name);
		}
		_scanner->back();
		size_t n = parts.size();
		if (returnFullPath && n>1) {
			name = ((BppString*)parts[0])->c_str();
			for (size_t i = 1; i < n; ++i) {
				name += _T("::");
				name += ((BppString*)parts[i])->c_str();
			}
		}
		if (!storeLastPart) parts.pop_back();
	}
	return name;
}




BppCompiler::TokenListVec::~TokenListVec()
{
	clear();
}

void BppCompiler::TokenListVec::push_back(BppSavedToken* lst)
{
	_listVec.push_back(BppValue(INTPTRCAST(lst), true));
}

BppSavedToken* BppCompiler::TokenListVec::at(size_t index) const
{
	return (BppSavedToken*)_listVec[index].ptr();
}

/**
Function appends all tokenlists behind item <i>index</i> to item <i/index</i>
end deletes items behind<i>index</i>.
@param index Start index of merging
*/
void BppCompiler::TokenListVec::mergeFrom(size_t index)
{
	size_t n = size();
	if (index+1 >= n) return;
	BppSavedToken* resultItem = 0;
	BppSavedToken* lastItem = 0;
	// laufe von index bis zum Vektorende
	for (size_t i = index; i < n; ++i) {
		// resultItem wird die Liste des ersten Eintrags ab index, der nicht Null ist.
		if (!resultItem) resultItem = at(i);
		if (!resultItem) continue;
		if (!lastItem) {
			// lastItem ist initial resultItem
			lastItem = resultItem;
			continue;
		}
		// laufe zum Listenende des ersten Eintrags ab index, der nicht Null ist
		while (lastItem->next) lastItem = lastItem->next;
		// hänge Liste des nächsten Eintrags an vorherige Liste an
		BppSavedToken* p = at(i);
		if (p) {
			lastItem->next = createCommaToken();
			lastItem = lastItem->next;
			lastItem->next = p;
			// setze Eintrag angehängten im Vektor auf Nill
			_listVec[i] = INTPTRCAST(0);
		}
	}
	// setze Eintrag von index auf Resultat ...
	_listVec[index] = INTPTRCAST(resultItem);
	// ... und lösche Rest hinter index 
	_listVec.erase(index + 1);
}

/**
* @return Comma token
* @note Caller is responsible to <tt>delete</tt> result, if no longer needed.
*/
BppSavedToken* BppCompiler::TokenListVec::createCommaToken() const
{
	BppSavedToken* result = new BppSavedToken();
	result->ttype = T_ANY;
	result->charBuffer = _T(',');
	return result;
}

/**
* @return Number of items in item vector.
*/
size_t BppCompiler::TokenListVec::size() const
{
	return _listVec.size();
}

/**
* Function deletes token list of each item and clears item vector.
*/
void BppCompiler::TokenListVec::clear()
{
	size_t n = size();
	for (size_t i = 0; i < n; ++i)
		delete at(i);
	_listVec.clear();
}
