#include "bppdbg.h"
#include <stdio.h>
#include "bppcomp.h"
#include "bppint.h"
#include "bppvm.h"
#include "bppprint.h"
#include "bppvec.h"
#include "bppdict.h"
#include "bppvalue.h"
#include "bppclass.h"
#include "bppobjct.h"
#include "bppbuf.h"
#include "bppconst.h"

struct OpCodePair
{
	BppOpCode opcode;
	const BppChar* name;
};

#define OCP(OPC) { OP_##OPC, _T(#OPC) }

static const OpCodePair _opCodeTab[] =
{
	/* opcodes */
	OCP(NOP), // 0x00, // no operation
	OCP(BRT), // 0x01, branch on true
	OCP(BRF), // 0x02, branch on false
	OCP(BR), // 0x03, branch unconditionally
	OCP(NIL), // 0x04, load top of stack with nil
	OCP(PUSH), // 0x05, push nil onto stack
	OCP(NOT), // 0x06, logical negate top of stack
	OCP(NEG), // 0x07, negate top of stack
	OCP(ADD), // 0x08, add top two stack entries
	OCP(SUB), // 0x09, subtract top two stack entries
	OCP(MUL), // 0x0A, multiply top two stack entries
	OCP(DIV), // 0x0B, divide top two stack entries
	OCP(REM), // 0x0C, remainder of top two stack entries
	OCP(BAND), // 0x0D, bitwise and of top two stack entries
	OCP(BOR), // 0x0E, bitwise or of top two stack entries
	OCP(XOR), // 0x0F, bitwise xor of top two stack entries
	OCP(BNOT), // 0x10, bitwise not of top two stack entries
	OCP(SHL), // 0x11, shift left top two stack entries
	OCP(SHR), // 0x12, shift right top two stack entries
	OCP(LT), // 0x13, less than
	OCP(LE), // 0x14, less than or equal to
	OCP(EQ), // 0x15, equal to
	OCP(NE), // 0x16, not equal to
	OCP(GE), // 0x17, greater than or equal to
	OCP(GT), // 0x18, greater than
	OCP(INC), // 0x19, increment
	OCP(DEC), // 0x1A, decrement
	OCP(LIT), // 0x1B, load literal
	OCP(RETURN), // 0x1C, return from interpreter
	OCP(CALL), // 0x1D, call a function
	OCP(REF), // 0x1E, load a variable value
	OCP(SET), // 0x1F, set the value of a variable
	OCP(VREF), // 0x20, load a vector element
	OCP(VSET), // 0x21, set a vector element
	OCP(MREF), // 0x22, load a member variable value
	OCP(MSET), // 0x23, set a member variable
	OCP(AREF), // 0x24, load an argument value
	OCP(ASET), // 0x25, set an argument value
	OCP(TREF), // 0x26, load a temporary variable value
	OCP(TSET), // 0x27, set a temporary variable
	OCP(TSPACE), // 0x28, allocate temporary variable space
	OCP(SEND), // 0x29, send a message to an object
	OCP(DUP2), // 0x2A, duplicate top two elements on the stack
	OCP(NEW), // 0x2B, create a new class object
	OCP(DELETE), // 0x2C,    delete object instance */
	OCP(DUP), // 0x2D duplicate TOS
	OCP(POP), // 0x2E, // remove topmost entry from stack
	OCP(MKREF), // 0x2F, // make reference from refcounted value
	OCP(TRY), // 0x30, enter try block
	OCP(ENDTRY), // 0x31,  leave try block
	OCP(THROW), // 0x32,  raise an exception
	OCP(ENDCATCH), // 0x33,  reset exception on end of catch block
	OCP(SWITCH), // 0x34,  enter switch block
	OCP(SREF), // 0x35, load static variable
	OCP(SSET), // < 0x36, set static variable
	OCP(PINC), // < 0x37, post increment
	OCP(PDEC), // < 0x38, post decrement
	OCP(PREF), // < 0x39, post increment
	OCP(PSET), // < 0x3A, post decrement
	OCP(POPC), // < 0x3B, POPC
	OCP(PLUS), // < 0x3C, unary plus
	OCP(UNREF), // < 0x3D, make value from reference
	OCP(STOS), // < 0x3E, store TOS
	OCP(LDTOS), // < 0x3E, load TOS
	OCP(MKCONST), // 0x40, make temporary const 
	OCP(SPCONST), // 0x40, make TOS const 
	OCP(YIELD), // yield
	// debug opcodes
	OCP(TRON), // 0xFC
	OCP(TROFF), // 0xFD
	OCP(TRSTEP), // 0xFE
	OCP(NEWLINE), // 0xFF
};

static const int _ocTabSize = sizeof(_opCodeTab) / sizeof(OpCodePair);


BppDebugger::BppDebugger(BppVirtualMachine& vm, IBppPrinter* printer)
:_vm(vm), _printer(printer), _interpreter(0), _currentProc(0),active(false)
{
	if (_printer == 0)
		_printer = &_vm.printer();
}

BppDebugger::~BppDebugger()
{
}

void BppDebugger::setInterpreter(BppInterpreter* interpreter)
{
	_interpreter = interpreter;
}
void BppDebugger::setCurrentProc(BppProcInfo* currentProc)
{
		IBppPrinter& pr = *_printer;

		_currentProc = currentProc;
	if (_currentProc != 0)
	{
		const BppChar* pname = _currentProc->getName();
		if (pname == 0)
			pname = _T("<noname>");
		const BppChar* classname = _T("");
		BppClass* pclass = (*_currentProc)[(unsigned)IDX_CLASS];
		if (pclass != 0)
			classname = pclass->getClassName()->c_str();
		pr(_T("-->current proc %s::%s, %d arguments\n"),classname,pname,_currentProc->getNargs());
	}
}

void BppDebugger::decodeProcedure(BppVector* codeVec)
{
	if ((codeVec == 0) || (codeVec->size() < 3)) return;
	BppVector* arguments = 0;
	BppVector* temporaries = 0;
	BppDictionary* statics = 0;
	ValPtr lastLit = codeVec->last();
	if (lastLit->getType() == BVT_LOCALVARINFO)
	{
		BppVector* p = *lastLit;
		arguments = (*p)[0];
		temporaries = (*p)[1];
		ValPtr lastTemp = temporaries->last();
		if (lastTemp->getType() == BVT_DICTIONARY)
			statics = *lastTemp;
	}
	const BppVector& cvec = *codeVec;
	BppBuffer* cbuf = cvec[IDX_CODE];
	if (cbuf == 0) return;
	BppByte* pcode = cbuf->data();
	if (pcode == 0) return;
	size_t codeSize = cbuf->size();
	BppString* name = cvec[IDX_NAME];
	BppClass* pclass = cvec[IDX_CLASS];
	IBppPrinter& pr = *_printer;
	if (pclass !=0) 
		pr(pclass->getClassName()->c_str());
	pr(_T("::"));
	if (name != 0)
		pr(name->c_str());
	else
		pr(_T("<unnamed>"));
	pr(_T(" (%u bytes)\n"),codeSize);
	size_t pc = 0;
	while (pc < codeSize)
	{
		BppOpCodeByte opcodebyte = pcode[pc];
	  unsigned ppc = pc;
	  BppOpCode opcode = (BppOpCode) opcodebyte._v._op.opCode;
	  if (opcode == OP_NOP && opcodebyte.opDataType() != ODT_NONE) // NOP + opDataType-Bits means extended code
	  {
		  BppByte ocb = opcodebyte;
		  opcodebyte = pcode[++pc];
		  opcode = (BppOpCode) (((BppByte) opcodebyte.opCode()) | ocb);
	  }
	  BppDWord opdata = 0;

		pr(_T("%04x: %02x %9s "),ppc,(unsigned)opcode,getOpCodeName(opcode));
		pc++;
	  switch (opcodebyte.opDataType())
	  {
	  case ODT_BYTE:
		  opdata = (BppDWord) (BppByte) pcode[pc++];
		  break;
	  case ODT_WORD:
		  opdata = BytePtr2W(pcode+pc);
		  pc += sizeof(BppWord);
		  break;
	  case ODT_DWORD:
		  opdata = BytePtr2DW(pcode+pc);
		  pc += sizeof(BppDWord);
		  break;
	  }
		switch(opcode)
		{
		case OP_BRT: // 0x01,  branch on true 
		case OP_BRF: // 0x02,  branch on false 
		case OP_BR: // 0x03,  branch unconditionally
		case OP_TRY:
		case OP_ENDTRY:
		case OP_SWITCH:
		  pr(_T("%04x "),opdata);
		break;
		case OP_LIT: // 0x1B,  load literal 
			{
				pr(_T("lit[%u] "),opdata);
				printValue(cvec[opdata]);
			}
			break;
		case OP_SREF: //  load static local variable */
		case OP_SSET: //  set static local variable */
			{
				pr(_T("svar[%u] "),opdata);
				if (statics)
				{
					for (BppDictionaryEntry* p = statics->getFirst();
						p != 0; p = p->getNext())
					{
						if (((BppUInt) p->getValue()) == opdata)
						{
							pr(_T("('%s') "),p->getKey()->c_str());
							break;
						}
					}
				}
				printValue(cvec[opdata]);
			}
			break;
		case OP_CALL: // 0x1D, call a function */
		case OP_SEND: // 0x29, send a message to an object */
			{
				BppNWord n = opdata;
				pr(_T("(%d arguments) "),n);
			}
			break;
		case OP_REF: // 0x1E,  load a variable value  
		case OP_SET: // 0x1F,  set the value of a variable 
			{
				BppNWord n = opdata;
					pr(_T("var[%u] "),n);
					printValue(cvec[n]);
			}
			break;
		case OP_MREF: // 0x22,  load a member variable value/
		case OP_MSET: // 0x23,  set a member variable
			{
				BppNWord n = opdata;
				pr(_T("member[%u] "),n);
				if (pclass)
				{
					const BppChar* p = pclass->findDataMemberFromIndex(n);
					if (p != 0) pr(_T("('%s') "),p);
				}
			}
			break;
		case OP_AREF: // 0x24,  load an argument value 
		case OP_ASET: // 0x25, set an argument value
			{
				BppNWord n = opdata;
				pr(_T("arg[%u] "),n);
				if (arguments && (n < arguments->size()))
				{
					BppString* s = (*arguments)[n];
					if (s != 0) pr(_T("('%s') "),s->c_str());
				}
			}
			break;
		case OP_TREF: // 0x26, load a temporary variable value
		case OP_TSET: // 0x27, set a temporary variable
			{
				BppNWord n = opdata;
				pr(_T("tmp[%d] "),n);
				if (temporaries && (n < temporaries->size()))
				{
					BppString* s = (*temporaries)[n];
					if (s != 0) pr(_T("('%s') "),s->c_str());
				}
			}
			break;
		case OP_TSPACE: // 0x28, allocate temporary variable space
			{
				BppNWord n = opdata;
				pr(_T("%u "),n);
			}
			break;
		case OP_NEWLINE: // 0x83, 
			{
				BppNWord n = opdata;
				pr(_T("line: %u "),n);
			}
			break;
	  case OP_PUSH:
			if (opcodebyte.opDataType() != ODT_NONE)
		  pr(_T("!%s "),getValueTypeName((BppValueType)(opdata & 0xFF)));
		 break;
	  case OP_RETURN:
	  case OP_YIELD:
			if (opcodebyte.opDataType() != ODT_NONE)
		  pr(_T("(%s)"),getValueTypeName((BppValueType)(opdata & 0xFF)));
		 break;
	  default:
			if (opcodebyte.opDataType() != ODT_NONE)
			 pr(_T("data: %lx "),opdata);
		} // end switch
		pr(_T("\n"));
	} // end while
}

void BppDebugger::traceInstruction(ValPtr sp, ValPtr bp, ValPtr fp, BppByte* pcode, unsigned pc)
{
		BppOpCodeByte opcodebyte = pcode[pc];
	  unsigned ppc = pc;
	  BppOpCode opcode = (BppOpCode) opcodebyte._v._op.opCode;
	  if (opcode == OP_NOP && opcodebyte.opDataType() != ODT_NONE) // NOP + opDataType-Bits means extended code
	  {
		  BppByte ocb = opcodebyte;
		  opcodebyte = pcode[++pc];
		  opcode = (BppOpCode) (((BppByte) opcodebyte.opCode()) | ocb);
	  }
	  BppDWord opdata = 0;
	  switch (opcodebyte._v._op.opDataType)
	  {
	  case ODT_BYTE:
		  opdata = (BppDWord) (BppByte) pcode[pc+1];
		  break;
	  case ODT_WORD:
		  opdata = BytePtr2W(pcode+pc+1);
		  break;
	  case ODT_DWORD:
		  opdata = BytePtr2DW(pcode+pc+1);
		  break;
	  }
	IBppPrinter& pr = *_printer;
	pr(_T("%04x: %02x %9s SP:%04x "),ppc,(unsigned)opcode,getOpCodeName(opcode),sp.current());
	pc++;
	switch(opcode)
	{
	case OP_NOP: // 0x00, no operation
	case OP_PUSH: // 0x05, push nil onto stack
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_BRT: // 0x01,  branch on true
		if (((BppInt)(*sp)) != 0)
			pr(_T("%04x "),BytePtr2NW(pcode+pc));
		else
			pr(_T("(skip) "));
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_BRF: // 0x02, branch on false
		if (((BppInt)(*sp)) == 0)
			(*_printer)(_T("%04x "),BytePtr2NW(pcode+pc));
		else
			(*_printer)(_T("(skip) "));
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_BR: // 0x03, branch unconditionally
	case OP_TRY:
	case OP_ENDTRY:
	case OP_SWITCH:
		(*_printer)(_T("%04x"),BytePtr2NW(pcode+pc));
		break;
	case OP_NIL: // 0x04, load top of stack with nil
	case OP_DUP:
	case OP_POPC:
	case OP_STOS:
	case OP_POP: // 0x2E, remove topmost entry from stack
		pr(_T("SP[0]:"));
		printValue(*sp,true);
		break;
	case OP_LDTOS:
		pr(_T("SP[0]:"));
		printValue(*sp, true);
		if (_currentProc) {
			pr(_T("VAL:"));
			printValue(_currentProc->tosTmp(), true);
		}
		break;
	case OP_NOT: // 0x06, logical negate top of stack
	case OP_NEG: // 0x07, negate top of stack
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_ADD: // 0x08, add top two stack entries
	case OP_SUB: // 0x09, subtract top two stack entries
	case OP_MUL: // 0x0A, multiply top two stack entries
	case OP_DIV: // 0x0B, divide top two stack entries
	case OP_REM: // 0x0C, remainder of top two stack entries
	case OP_BAND: // 0x0D, bitwise and of top two stack entries
	case OP_BOR: // 0x0E, bitwise or of top two stack entries
	case OP_XOR: // 0x0F, bitwise xor of top two stack entries
	case OP_SHL: // 0x11, shift left top two stack entries
	case OP_SHR: // 0x12, shift right top two stack entries
	case OP_LT: // 0x13, less than
	case OP_LE: // 0x14, less than or equal to
	case OP_EQ: // 0x15, equal to
	case OP_NE: // 0x16, not equal to
	case OP_GE: // 0x17, greater than or equal to
	case OP_GT: // 0x18, greater than
		pr(_T("SP[-1]:"));
		printValue(sp[-1]);
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_BNOT: // 0x10, bitwise not of top stack entries
	case OP_MKREF: // 0x46, make reference from TOS
	case OP_UNREF: //  make value from reference on TOS
	case OP_INC: // 0x19, increment
	case OP_DEC: // 0x1A, decrement
	case OP_PINC: // 0x37, post increment
	case OP_PDEC: // 0x38, post decrement
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_LIT: // 0x1B, load literal
		{
			BppNWord idx = opdata;
			if (_currentProc != 0)
			{
				pr(_T("LIT[%d]:"),idx);
				printValue((*_currentProc)[idx]);
			}
		}
		break;
	case OP_SREF: //  load static local
	case OP_SSET: //  set static local
		{
			BppNWord idx = opdata;
			if (_currentProc != 0)
			{
				pr(_T("svar[%d]:"),idx);
				const BppChar* pname = _currentProc->getLocalStaticVarName(idx);
				if (pname != 0) pr(_T("('%s') "),pname);
				printValue((*_currentProc)[idx]);
			}
		}
		break;
	case OP_RETURN: // 0x1C, return from interpreter
	case OP_YIELD:
		if (opdata == BVT_STRONGOBJECT && (BppClass*)(*sp) ) {
			pr(_T("SP[-1]:"));
			printValue(sp[-1]);
		}
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_CALL: // 0x1D, call a function
		{
			BppNWord n = opdata;
			ValPtr vp = sp;
			vp -= n;
			printValue(*vp);
			pr(_T("(%d arguments)"),n);
		}
		break;
	case OP_REF: // 0x1E, load a variable value
	case OP_SET: // 0x1F, set the value of a variable
		{
			BppNWord n = opdata;
			if (_currentProc != 0)
			{
				pr(_T("VAR[%d]:"),n);
				printValue((*_currentProc)[n],(opcode == OP_SET));
				if (opcode == OP_SET)
				{
					pr(_T("SP[0](val):"));
					printValue(*sp);
				}
			}
		}
		break;
	case OP_VREF: // 0x20, load a vector element
		pr(_T("SP[-1](src):"));
		printValue(sp[-1]);
		pr(_T("SP[0](index:)"));
		printValue(*sp);
		break;
	case OP_VSET: // 0x21, set a vector element
		pr(_T("SP[-2](dest):"));
		printValue(sp[-2]);
		pr(_T("SP[-1](index):"));
		printValue(sp[-1]);
		pr(_T("SP[0](val):"));
		printValue(*sp);
		break;
	case OP_PREF: // load property
		pr(_T("SP[-1](src):"));
		printValue(sp[-1]);
		pr(_T("SP[0](key:)"));
		printValue(*sp);
		break;
	case OP_PSET: // set property
		pr(_T("SP[-2](dest):"));
		printValue(sp[-2]);
		pr(_T("SP[-1](key):"));
		printValue(sp[-1]);
		pr(_T("SP[0](val):"));
		printValue(*sp);
		break;
	case OP_MREF: // 0x22, load a member variable value
	case OP_MSET: // 0x23, set a member variable
		{
			pr(_T("BP[0](obj):"));
			printValue(*bp);
			BppObject* obj = *bp;
			if (obj != 0)
			{
				BppNWord n = opdata;
				BppVector* mdata = (obj->getMemberData());
				if (mdata != 0)
				{
					BppValue& v = (*mdata)[n];
					pr(_T("member[%d]:"),n);
					BppClass* pcls = obj->getClass();
					if (pcls != 0)
					{
						const BppChar* p = pcls->findDataMemberFromIndex(n);
						if (p != 0)
							pr(_T("('%s') "),p);
					}
					printValue(v,(opcode == OP_MSET));
					if (opcode == OP_MSET)
					{
						pr(_T("SP[0](val):"));
						printValue(*sp);
					}
				} // mdata
			} // obj
		}
		break;
	case OP_AREF: // 0x24, load an argument value
	case OP_ASET: // 0x25, set an argument value
		{
			if (_currentProc == 0) break;
			int nargs = _currentProc->getNargs();
			BppNWord n = opdata;
			if (n >= (unsigned) nargs)
			{
				pr(_T("invalid index %d"),n);
				break;
			}
			const BppChar* pname = _currentProc->getArgumentName(n);
			if (pname != 0)
				 pr(_T("BP[%d](arg[%d] '%s'):"),n+1,n,pname);
			else
				 pr(_T("BP[%d](arg[%d]):"),n+1,n);
			printValue(bp[(int)n+1],(opcode == OP_ASET));
			if (opcode == OP_ASET)
			{
				pr(_T("SP[0](val):"));
				printValue(*sp);
			}
		}
		break;
	case OP_TREF: // 0x26, load a temporary variable value
	case OP_TSET: // 0x27, set a temporary variable
	case OP_MKCONST:
		{
			BppNWord n = opdata;
			const BppChar* pname = _currentProc->getLocalVarName(n);
			if (pname != 0)
				 pr(_T("FP[%d](tmp[%d] '%s'):"),n+1,n,pname);
			else
				 pr(_T("FP[%d](tmp[%d]):"),n+1,n);
			printValue(fp[(int)n+1],(opcode == OP_TSET));
			if (opcode == OP_TSET)
			{
				pr(_T("SP[0](val):"));
				printValue(*sp);
			}
		}
		break;
	case OP_TSPACE: // 0x28, allocate temporary variable space
		{
			BppNWord n = opdata;
			pr(_T("%d "),n);
		}
		break;
	case OP_SEND: // 0x29, send a message to an object
		{
			BppNWord n = opdata;
			ValPtr vp = sp;
			vp -= n;
			pr(_T("SP[%d](obj):"),-(int)n);
			printValue(*vp);
			pr(_T("SP[%d](name):"),-((int)n-1));
			printValue(vp[1]);
			pr(_T("(%d arguments)"),n);
		}
		break;
	case OP_DUP2: // 0x2A, duplicate top two elements on the stack
		pr(_T("SP[-1]:"));
		printValue(sp[-1]);
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_NEW: // 0x2B, create a new class object
		pr(_T("SP[0](class):"));
		printValue(*sp);
		break;
	case OP_DELETE: // 0x2C, delete object instance
		pr(_T("SP[0]:"));
		printValue(*sp);
		break;
	case OP_NEWLINE: // 0x83,
		{
			BppNWord idx = opdata;
			if (_currentProc != 0)
			{
				pr(_T("line: %d"),idx);
			}
		}
		break;

	}
	pr(_T("\n"));
}
void BppDebugger::printValue(const BppValue& val, bool safe)
{
	IBppPrinter& pr = *_printer;
	BppValueType vt = val.getType();
	if (val.getFlags() & BVF_CONST) pr(_T("const "));
	if (val.getFlags() & BVF_STRONGTYPED) pr(_T("!"));
	pr(_T("%s "), getValueTypeName(vt));
	switch (vt)
	{
	case BVT_BOOL:
		pr(val.asBool() ? _T("true") : _T("false"));
		break;
	case BVT_INT:
	case BVT_SHORT:
	case BVT_SBYTE:
#if (INT_MAX == SHRT_MAX)
		pr(_T("%ld"), (BppInt)val);
#else
		pr(_T("%d "), (BppInt)val);
#endif
		break;
	case BVT_UINT:
	case BVT_USHORT:
	case BVT_BYTE:
#if (INT_MAX == SHRT_MAX)
		pr(_T("%lu"), (BppUInt)val);
#else
		pr(_T("%u "), (BppUInt)val);
#endif
		break;
	case BVT_INTPTR:
		pr(_T("%p "),
#ifdef GCC
		(BppLong)
#else		 
			(BppIntPtr)
#endif
			val);
		break;
	case BVT_LONG:
#ifdef EMULATE_I64
	{
		BppChar buf[65];
		((BppLong)val).toa(buf);
		pr("%s ", buf);
	}
#else
		pr(_T("%lld "), (BppLong)val);
#endif
		break;
	case BVT_ULONG:
#ifdef EMULATE_I64
	{
		char buf[65];
		((BppULong)val).toa(buf);
		pr("%s ", buf);
}
#else
		pr(_T("%llu "), (BppULong)val);
#endif
		break;
	case BVT_DOUBLE:
	case BVT_FLOAT:
		pr(_T("%g "), (double)val);
		break;
	case BVT_FILE:
	{
		FILE* fp = val;
		if (fp == stdin) pr(_T("stdin "));
		else if (fp == stdout) pr(_T("stdout "));
		else if (fp == stderr) pr(_T("stderr "));
		else pr(_T("%p"), fp);
	}
	break;
	case BVT_CODE:
		//case BVT_CODEREF:
	{
		BPPUSERFUNC f = val;
		pr(_T("%p "), f);
	}
	break;
	case BVT_BYTECODE:
	case BVT_BUFFER:
	case BVT_CHARBUFFER:
	case BVT_VECTOR:
	case BVT_DICTIONARY:
	case BVT_BYTECODEREF:
	case BVT_BUFFERREF:
	case BVT_CHARBUFFERREF:
	case BVT_VECTORREF:
	case BVT_DICTIONARYREF:
	{
		BppRefObject* pref = val.getRef();
		if (!safe)
		{
			int cnt = 0;
			if (pref != 0)
				cnt = pref->getRefCnt();
			pr(_T("%p (#%d) "), pref, cnt);
		}
		else
			pr(_T("%p "), pref);
	}
	break;
	case BVT_CLASS:
	case BVT_CLASSREF:
	{
		BppClass* pclass = val;
		if (!safe)
		{
			int cnt = 0;
			if (pclass != 0)
				cnt = pclass->getRefCnt();
			pr(_T("%p (#%d) "), pclass, cnt);
			if (pclass != 0)
				pr(_T("(%s) "), pclass->getClassName()->c_str());
		}
		else
			pr(_T("%p "), pclass);
	}
	break;
	case BVT_OBJECT:
	case BVT_OBJECTREF:
	case BVT_STRONGOBJECT:
	{
		BppObject* obj = val;
		if (!safe)
		{
			int cnt = 0;
			if (obj != 0)
				cnt = obj->getRefCnt();
			pr(_T("%p (#%d) "), obj, cnt);
			if (obj != 0)
				pr(_T("(class %s) "), obj->getClass()->getClassName()->c_str());
		}
		else
			pr(_T("%p "), obj);
	}
	break;
	case BVT_STRING:
	case BVT_STRINGREF:
	{
		BppString* str = val;
		if (!safe)
		{
			int cnt = 0;
			if (str != 0)
				cnt = str->getRefCnt();
			pr(_T("%p (#%d) "), str, cnt);
			if (str != 0)
			{
				BppChar buf[21];
				_tcsncpy(buf, str->c_str(), 20);
				buf[20] = 0;
				if (str->length() > 20)
					pr(_T("(\"%s...\") "), buf);
				else
					pr(_T("\"%s\") "), buf);
			}
		}
		else
			pr(_T("%p "), str);
	}
	break;
	case BVT_VAR:
	{
		BppDictionaryEntry* de = val;
		pr(_T("%p "), de);
		if ((de != 0) && (!safe))
		{
			pr(_T("(key:%s value:"), de->getKey()->c_str());
			printValue(de->getValue());
			pr(_T(") "));
		}
	}
	break;
	default:
		if (val.isCharacterType()) {
			BppValue cval = BppValue::makeChar(val);
			BppChar b[2];
			b[0] = cval.asBppChar();
			b[1] = 0;
			pr(_T("'%s' "), b);
		}
	}
}

const BppChar* BppDebugger::getOpCodeName(BppOpCode opCode) const
{
	for (int i=0; i<_ocTabSize;i++)
		if (_opCodeTab[i].opcode == opCode) return _opCodeTab[i].name;
	return _T("<unknown>");
}

