#include "bppdict.h"


/**
@param key pointer to key string
@param val value of dictionary entry
@param stype symbol type (for use in internal symbol tables)
@note This constructor only increments reference counter of <i>key</i>, it doesn't make a copy! 
*/
BppDictionaryEntry::BppDictionaryEntry(BppString* key, const BppValue& val, BppSymbolType stype) BPPNOEXCEPT
:_key(key),_value(val),_dict(0),_stype(stype),_before(0),_next(0)
{
	if (_key) _key->addRef();
}

/**
@param key value of key
@param val value of dictionary entry
@param stype symbol type (for use in internal symbol tables)
*/
BppDictionaryEntry::BppDictionaryEntry(const BppString& key, const BppValue& val, BppSymbolType stype)
:_key(new BppString(key)),_value(val),_dict(0),_stype(stype), _before(0),_next(0)
{
	_key->addRef();
}

/**
@param cpy Dictionary entry to copy from.
@note Copy constructor only increments reference counter of key contained in <i>cpy</i>, it doesn't make a copy!
*/
BppDictionaryEntry::BppDictionaryEntry(const BppDictionaryEntry& cpy)
:_key(cpy._key),_value(cpy._value),_dict(0),_stype(cpy._stype), _before(0), _next(0)
{
	if (_key) _key->addRef();
}

/**
If coontained in a dictionary, destructor removes item from. Furthermore reference
counter of key (if any) is decremented.
*/
BppDictionaryEntry::~BppDictionaryEntry()
{
	if (_dict)
	{
		if (_before) _before->_next = _next;
		else _dict->_firstElement = _next;
		if (_next) _next->_before = _before;
		else _dict->_lastElement = _before;
#ifndef __NOSTL__
		if (_key)
		  _dict->_entryMap.erase(_key);
#endif

		_dict=0;
	}
	if (_key) _key->release();
}

/**
@param assignment dictionary entry to assign from
@note The assignment operator is mainly for use BppDictionaryEntry object
in the role of a variable. Normally it assigns both, the key (as variable name) and the value.
If object is part of a dictionary key of <i>assignment</i> is ignored but value assigned only.<br>
Please note that assignment of key means a reference to key is copied, not the character string of key.
*/
BppDictionaryEntry& BppDictionaryEntry::operator=(const BppDictionaryEntry& assignment)
{
	if (this == &assignment)
		return *this;
	if (!_dict)
	{
		BppString* p = assignment._key;
		if (p != 0) p->addRef();
		if (_key != 0 && _key != assignment._key) {
			_key->release();
		}
		_key = p;
	}
	_value = assignment._value;
	_stype = assignment._stype;
	return *this;
}

/**
@func Operator assigns value only, i.e. it leaves key unchanged. It is
for use to assign value to a dictionary entry. It does same as setValue member function.
@param assignment value to assign
@see setValue
*/
BppDictionaryEntry& BppDictionaryEntry::operator=(const BppValue& assignment)
{
	_value = assignment;
	return *this;
}
#ifdef CPP11
BppDictionaryEntry & BppDictionaryEntry::operator=(BppValue && assignment) BPPNOEXCEPT
{
	_value = BPPMOVE(assignment);
	return *this;
}
#endif
	
/**
@func Function assigns new value to dictionary entry.
*/
bool BppDictionaryEntry::setValue(const BppValue& val, bool typeSafe, bool copyFlags)
{
	if (typeSafe) {
		if (!_value.assign(val)) return false;
	}
	else _value = val;
	if (copyFlags) _value.setFlags(val.getFlags());
	return true;
}
#ifdef CPP11
bool BppDictionaryEntry::setValue(BppValue&& val) BPPNOEXCEPT
{
	_value = BPPMOVE(val);
	return true;
}
#endif
	//void setNext(BppDictionaryEntry* next);

bool BppDictionaryEntry::operator==(const BppDictionaryEntry& cmp) const
{
	if (&cmp == this) return true;
	if (cmp._stype != _stype) return false;
	if (_key != cmp._key)
	{
		if (_key == 0 || cmp._key == 0) return false;
		if ((*_key) != (*cmp._key)) return false;
	}
	return _value == cmp._value;
}

bool BppDictionaryEntry::operator!=(const BppDictionaryEntry& cmp) const
{
	return !(*this == cmp);
}


/**
@param sorted flag for sorted mode
*/
BppDictionary::BppDictionary(bool sorted)
:_firstElement(0),_lastElement(0),_sorted(sorted),_ownerClass(0)
#ifndef __NOSTL__
, _entryMap(*(new DictMapType))
#endif
{
}

/**
Destructor simply calls clear method.
@see clear
*/
BppDictionary::~BppDictionary()
{
	clear();
#ifndef __NOSTL__
	delete (&_entryMap);
#endif
}

/**
@return new copy of current dictionary
@note Caller is responsible for delete result if no longer needed.
*/
BppDictionary* BppDictionary::clone() const
{
	BppDictionary* result = new BppDictionary(_sorted);
	result->_ownerClass = _ownerClass;
	for (BppDictionaryEntry* p=_firstElement;p;p=p->_next)
	{
		BppValue& v = (BppValue&)((*result)[(const BppString&)(*(p->_key))]).getValue();
		v = BppValue(p->_value,true);
		v.setFlags(p->_value.getFlags());
	}
	return result;
}

/**
@return pointer to first entry (NULL on empty dictionary)
*/
BppDictionaryEntry* BppDictionary::getFirst() const BPPNOEXCEPT
{
	return _firstElement;
}

/**
@return pointer to last entry (NULL on empty dictionary)
*/
BppDictionaryEntry* BppDictionary::getLast() const BPPNOEXCEPT
{
	return _lastElement;
}

/**
@func Function searches for entry specified by <i>key</i>.
@param key key of entry to search for
@return pointer to entry if found, otherwise NULL
*/
BppDictionaryEntry* BppDictionary::find(const BppString& key) const BPPNOEXCEPT
{
#ifndef __NOSTL__
	DictMapType::iterator it = _entryMap.find(&key);
	if (it != _entryMap.end())
		return it->second;
#else
	for (BppDictionaryEntry* p = _firstElement; p != 0; p=p->_next)
	{
		if (p->_key != 0)
		{
			int c = p->_key->compare(key);
			if (c == 0)
				return p;
			if ((c > 0) && _sorted)
				break;
		}
	}
#endif
	return 0;
}

/**
@param key Key to get dictionary entry for.
@return Dictionary entry for <i>key</i>.
@note If <i>key</i> is a NULL pointer, operator returns NULL.<br>
If no entry for <i>key</i> is contained in dictionary, a new one is created,
otherwise pointer to existing entry is returned.
*/
BppDictionaryEntry* BppDictionary::operator[](BppString* key)
{
	if (key == 0)
		return 0;
	return &((*this)[*key]);
}

/**
@param key Key to get dictionary entry for.
@return Dictionary entry for <i>key</i>.
@note If no entry for <i>key</i> is contained in dictionary, a new one is created,
otherwise pointer to existing entry is returned.
*/
BppDictionaryEntry& BppDictionary::operator[](const BppString& key)
{
#ifndef __NOSTL__
	DictMapType::iterator it = _entryMap.lower_bound(&key);
	if (it != _entryMap.end())
		if (it->first->compare(key) == 0) return *(it->second);
	BppDictionaryEntry* pnew = new BppDictionaryEntry(key, BppValue());
	if ((!_sorted) || it == _entryMap.end()) {
		if (_lastElement) {
			pnew->_before = _lastElement;
			_lastElement->_next = pnew;
		}
		else _firstElement = pnew;
		_lastElement = pnew;
	}
	else {
		pnew->_before = it->second->_before;
		it->second->_before = pnew;
		pnew->_next = it->second;
		if (pnew->_before) pnew->_before->_next = pnew;
		else _firstElement = pnew;
	}
	pnew->_dict = this;
	_entryMap[pnew->getKey()] = pnew;
	return *pnew;
#else
	BppDictionaryEntry* pbefore = 0;
	for (BppDictionaryEntry* p = _firstElement; p != 0; p=p->_next)
	{

		if (p->_key != 0)
		{
			int c = p->_key->compare(key);
			if (c == 0)
				return *p;

			if ((c > 0) && _sorted)
				break;
		}

		if (_sorted && (p->_key != 0))
		{
			if (p->_key->compare(key) > 0) break;
		}

		pbefore = p;
	}
	BppDictionaryEntry* pnew = new BppDictionaryEntry(key,BppValue());
	pnew->_dict = this;
	if (pbefore != 0) // nicht erstes Element
	{
		pnew->_next = pbefore->_next;
	  if (!(pbefore->_next)) _lastElement = pnew;
	  pnew->_before = pbefore;
		pbefore->_next = pnew; // ans Ende hngen
	}
	else
	{
		pnew->_next = _firstElement;
	  if (_firstElement) {
		  _firstElement->_before = pnew;
		  pnew->_next = _firstElement;
	  }
	  if (_firstElement == _lastElement) _lastElement = pnew;
		_firstElement = pnew;
	}
	return *pnew;
#endif
}

/**
@param entry entry to erase
@return true if entry found and removed
*/
bool BppDictionary::erase(BppDictionaryEntry* entry)
{
	if (!entry) return false;
	if (entry->_dict != this) return false;
	delete entry;
	return true;
}

/**
@param key key of entry to erase
@return true if entry found and removed
*/
bool BppDictionary::erase(const BppString& key)
{
	return erase(find(key));
}

/**
Function removes (and deletes) all entries from dictionary.
*/
void BppDictionary::clear()
{
	while (_firstElement != 0)
	{
		BppDictionaryEntry* p = _firstElement;
		_firstElement = p->_next;
		p->_dict = 0;
		delete p;
	}
	_lastElement = 0;
#ifndef __NOSTL__
	_entryMap.clear();
#endif
}

/**
@return number of entries in dictionary
*/
unsigned BppDictionary::count() const BPPNOEXCEPT
{
#ifndef __NOSTL__
	return (unsigned) _entryMap.size();
#else
	unsigned result = 0;
	for (BppDictionaryEntry* p=_firstElement;p!=0;p=p->_next)
		result++;
	return result;
#endif
}

bool BppDictionary::operator==(const BppDictionary& cmp) const
{
	if (&cmp == this) return true;
	if (count() != cmp.count()) return false;
	BppDictionaryEntry* p1 = _firstElement;
	BppDictionaryEntry* p2 = cmp._firstElement;
	 while (p1 && p2)
	{
		if (*p1 != *p2) return false;
		p1 = p1->getNext();
		p2 = p2->getNext();
	}
	return true;
}
	
bool BppDictionary::operator!=(const BppDictionary& cmp) const
{
	return !(*this == cmp);
}
