
#include "bppobjct.h"
#include "bppclass.h"
#include "bppvec.h"
#include "bppdict.h"
#include "bppvm.h"
#include "bppconst.h"

/**
Constructor creates vector of member data and initializes all  data fields (except the 
class field) to null. 
@param cls Pointer to class defining this object.
@param isDummy For internal use only: It is used by Compiler to create empty strong typed object variable, see BppCompiler::_codeVarDefinition.
@note <p>If no class specified (i.e. <i>cls</i> == NULL) no data vector is
created. In this case a following call to getClass or getMemberData methods
will return NULL too.</p>
<p>The first element of data vector always contains pointer to defining class ob object.</p>
*/
BppObject::BppObject(BppClass* cls, bool isDummy)
	:_memberData(0), _nativeObject(0)
{
	if (cls != 0)
	{
		if (!isDummy || cls->getNotifyProc()) {
			_memberData = cls->getDefaultDataMembers().clone();
			_memberData->addRef();
			cls->notifyObjCreation(this, false);
		}
		else {
			_memberData = new BppVector();
			_memberData->addRef();
			_memberData->push_back(cls);
		}
	}
}


/**
Copy constructor simply copies member data vector of source object
specified by <i>cpy</i>.
@param cpy object to create copy from
*/
BppObject::BppObject(const BppObject& cpy)
:_memberData(0), _nativeObject(cpy._nativeObject)
{
		BppVector* p = cpy.getMemberData();
		if (p)
		{
			_memberData = p->clone();
			_memberData->addRef();
		}
		BppClass* pclass = cpy.getClass();
		if (pclass) pclass->notifyObjCreation(this, true);
}

/**
Destructor first calls notifyObjDelete method of defining class and than releases
member data vector.
@see BppClass:notifyObjDelete
*/
BppObject::~BppObject()
{
	if (_memberData)
	{
		BppClass* pclass = (*_memberData)[0];
		if (pclass != 0)
			pclass->notifyObjDelete(this);
		_memberData->release();
	}
}

BppObject::BppObject() BPPNOEXCEPT
:_nativeObject(0), _memberData(0)
{
}

/**
@return new copy of current vector
@note Caller is responsible for delete result if no longer needed.
*/
BppObject* BppObject::clone() const
{
	  BppObject* result = new BppObject();
		if (_memberData)
		{
		  result->_memberData = _memberData->clone();
			 result->_memberData->addRef();
		}
	  return result;
}
/**
@return pointer to defining class object or NULL (if none)
*/
BppClass* BppObject::getClass() const BPPNOEXCEPT
{
	return (_memberData) ? ((BppClass*)(*_memberData)[0]) : 0;
}

/**
@return pointer to member data vector or NULL (if none)
*/
BppVector* BppObject::getMemberData() const BPPNOEXCEPT
{
	return _memberData;
}

/**
Function tries to assign this object to value object <i>val</i> using static type compatibility rules as follows:
<ul>
<li>If <i>val</i> requires an strong typed object (is of type BVT_STRONGOBJECT), assignment is done if:
	<ul>
		<li>Type of current object is of same class type as required.</li>
		<li>Class of current object is derived from required class.</li>
		<li>Current object defines conversion operation into required class.</li>
	</ul>
</li>
<li>If <i>val</i> requires any object (type BVT_OBJECT) assignment is done.</li>
<li>If <i>val</i> requires any other type, current object must define conversion operation int that type.</li>
</ul>
@param val Value object used for destination. <i>val</i> must contain static type information!
@return true on success.
@note This method is internally used only by BppValue::assign method but may be useful
for implementations of extensions too.
*/
bool BppObject::safeAssignToValue(BppValue& val)
{ 
	BppClass* pclass = getClass();
	if (pclass)
	{
		BppVirtualMachine& vm = pclass->getVM();
		const BppChar* mname = 0;
		if (val.getType() == BVT_STRONGOBJECT)
		{
			BppVector* desc = val.getStrongObjDesc();
			if (!desc)
			{
				BppVector* v = new BppVector(2,true);
				(*v)[IDX_CLASS] = pclass;
				val = BppValue((BppRefObject*)v,BVT_STRONGOBJECT);
				val.setFlags(val.getFlags() | BVF_STRONGTYPED);
					 desc = val.getStrongObjDesc(); 
			}
			BppClass* reqClass = desc ? (BppClass*)(*desc)[IDX_CLASS] : (BppClass*)0;
			if (reqClass == 0)
			{
					 if (!desc) return false;
					// if (pclass->getRefCnt() < 2) pclass->addRef();
				(*desc)[IDX_CLASS] = pclass;
				if (getRefCnt() == 0) this->addRef();
				(*desc)[IDX_CODE] = this;
				return true;
			}
			if (pclass->isSubtypeOf(reqClass))
			{
				if (getRefCnt() == 0) this->addRef();
				(*desc)[IDX_CODE] = this;
				return true;
			}
			else mname = reqClass->getClassName()->c_str();
		}
		else if (val.getType() == BVT_OBJECT)
		{
			  val.assign(this);
			return true;
		}
		else
				mname = vm.getTypeNameFromVType(val.getType());
		if (mname)
		{
			BppValue res;
			if (vm.invokeMember(res,this,mname))
					return val.assign(res);
		}
	}
	return false;
}
