#include "bppregex.h"
#include "bppufunc.h"
#include "bppvec.h"
#include "bppfimpl.h"
#include "bppvalue.h"
#include "bppint.h"
#include "bppvm.h"
#include "bppprint.h"
#include "bppclass.h"
#include "bppmsg.h"
#include"bppobjct.h"
#include <stdlib.h>
#include <ctype.h>

BppRegEx::BppRegEx() :_trex(*new TRexpp()), _flags(0), _pos(0), _len(BPP_NNPOS)
{
}

/**
Copies pattern and flags from <i>cpy</i> and compiles expression if not empty.  All other members  set to default values.
@param cpy Source object
@see compile
*/
BppRegEx::BppRegEx(const BppRegEx & cpy) :_trex(*new TRexpp()), _pattern(cpy._pattern), _flags(cpy._flags), _pos(0), _len(BPP_NNPOS)
{
	if (_pattern.length()) compile(_pattern.c_str(),_flags);
}

/**
Copies pattern and flags from <i>cpy</i> and compiles expression if not empty. All other members  set to default values.
@param cpy Source object
@see compile
*/
BppRegEx & BppRegEx::operator=(const BppRegEx& cpy)
{
	if (&cpy != this) {
		_ptxt.clear();
		_pos = 0;
		_len = BPP_NNPOS;
	  _flags = cpy._flags;
		if (_pattern != cpy._pattern) {
			_pattern = cpy._pattern;
			_error.clear();
			if (_pattern.length()) compile(_pattern.c_str(),_flags);
		}
	}
	return *this;
}

/**
Deletes wrapped TRexpp instance.
*/
BppRegEx::~BppRegEx()
{
	delete &_trex;
}

/**
@param pat pattern to compile
@param flags Default flags for use with compiled pattern. Value should be ORed
combination of Flags enum members.
@return true on success. If result is false, getError may be called to get error description.
@note <i>flags</i> are stored only. They don't affect compilation of pattern and may be changed
later by setFlags method if needed.
@see setFlags
*/
bool BppRegEx::compile(const BppString & pat, unsigned flags)
{
	_flags = flags;
	_pos = 0;
	_len = BPP_NNPOS;
	_ptxt.clear();
	_error.clear();
	if (_pattern != pat) _pattern = pat;
	bool ok = _trex.Compile((const BppUChar*)_pattern.c_str(), flags);
	if (!ok) _error = (const BppChar*) _trex.GetError();
	return ok;
}

/**
Function sets flags for an already compiled expression.
@param flags New flags for use with compiled pattern. Value should be ORed
combination of Flags enum members.
@return true on success, i.e. if expression was compiled successful before.
@see compile
*/
bool BppRegEx::setFlags(unsigned flags)
{
	if (!_trex.setFlags(flags)) return false;
	_flags = flags;
	return true;
}

/**
@return Flags mask currently set.
*/
unsigned BppRegEx::getFlags() const
{
	return _flags;
}

/**
@return Error message of last failed call to compile.
*/
const BppString& BppRegEx::getError() const
{
	return _error;
}

/**
Function searches for first match in specified text.
@param txt Text to search in.
@param start Start position for search in <i>txt</i>.
@param len Length of substring in <i>text</i> used for search. If <i>len</i> is
negative or larger than length of <i>txt</i> searching is done until end of <i>txt</i>.
@return true if a match was found. If function returns <tt>true</tt>, #getSubExpressions(BppVector&)
can be called do get positions and values of match and its submatches (if any). Function #next()
may be called to get next match.
@note This function searches for first match only, regardless of RXF_FINDALL flag.
@par Example
@code
// Count number of digits in string
BppRegEx rex;
if (rex.compile(_T("\\d"))) {
	int cnt = 0;
	for(bool ok = rex.search(_T("abc123x4y5z"));ok;ok=rex.next()) ++cnt;
	printf(_T("Number of matches: %d\n"),cnt);
}
else printf(_T("%s\n"),getError().c_str();
@endcode
@see next, getSubExpressions
*/
bool BppRegEx::search(const BppString & txt, size_t start, size_t len)
{
	if (!_trex.Ok()) return false;
	_ptxt = txt;
	_pos = start;
	const BppUChar* ptxt = (const BppUChar*)_ptxt.c_str();
	size_t l = txt.length();
	if (start >= l) {
		_len = BPP_NNPOS;
		return false;
	}
	const BppUChar* p01 = ptxt + start;
	if (len > (l - start)) len = l - start;
	_len = len;
	const BppUChar* p02 = p01 + len;
	const BppUChar *p1, *p2;
	bool ok = _trex.SearchRange(p01, p02, &p1, &p2);
	if (ok) {
		_pos = p2 - ptxt;
	}
	return ok;
}

/**
Function is intended for use together with #search. It finds next match.
@return true if another match was found.
@see search
*/
bool BppRegEx::next()
{
	if (!_trex.Ok()) return false;
	size_t l = _ptxt.length();
	if (_pos >= l || _pos >= _len || l < _len) return false;
	const BppUChar* ptxt = (const BppUChar*)_ptxt.c_str();
	const BppUChar* p01 = ptxt+ _pos;
	const BppUChar* p02 = p01 + _len;
	const BppUChar *p1, *p2;
	bool ok = _trex.SearchRange(p01, p02, &p1, &p2);
	if (ok) {
		_pos = p2 - ptxt;
	}
	return ok;
}

/**
Function checks if expression matches entire text.
@return true on match. On match a call to getSubExpressions(BppVector&) can be done to get
any submatches.
@par Example
@code
// checks if adr is a valid mail address
bool checkMailAdress(const BppString& adr)
{
	BppRex rex;
	rex.compile("\w+[\\w\\.]*\w+@\\w+\\.\\a+");
	return rex.match(adr);
}
@endcode
*/
bool BppRegEx::match(const BppString & txt)
{
	if (!_trex.Ok()) return false;
	_ptxt = txt;
	_pos = 0;
	_len = _ptxt.length();
	const BppUChar* ptxt = (const BppUChar*) _ptxt.c_str();
		_pos = _ptxt.length();
		return _trex.Match(ptxt);
}

/**
Function gets matching substrings in <i>txt</i>.
@param txt Text to search for matches.
@param dst Vector to store results in. Funktion simply appends new string for each match
to <i>dest</i>.
@param start Start position of search in <i>txt</i>.
@param len Maximum length of search, if <i>len</i>+<i>start</i> exceeds lemgth of <i>txt</i>
or <i>len</i> is negative, search is done to end of <i>txt</i>.
@return Number of matching substrings appended to <i>dst</i>.
@note Behavior of method depends on RXF_FINDALL flag. If is's set, all matches are
appended to <i>dst</i>, otherwise search is aborted after first match.
@see getMatchPositions, search, next
*/
unsigned BppRegEx::getMatches(const BppString & txt, BppVector & dst, size_t start, size_t len)
{
	if (!_trex.Ok()) return 0;
	_ptxt = txt;
	size_t l = txt.length();
	if (start >= l) return 0;
	_pos = start;
	const BppUChar* p01 = (const BppUChar*) _ptxt.c_str() + start;
	if (len > l - start) len = l - start;
	_len = len;
	const BppUChar* p02 = p01 + len;
	const BppUChar *p1, *p2;
	unsigned result = 0;
	while (_trex.SearchRange(p01, p02, &p1, &p2)) {
		ptrdiff_t diff = p2 - p1;
		p01 = p2;
		BppChar* buf = (BppChar*)allocmem((diff + 1)*sizeof(BppChar));
		_tcsncpy(buf, (const BppChar*) p1, diff);
		buf[diff] = 0;
		dst.push_back(new BppString((const BppChar*) buf));
		freemem(buf);
	  ++result;
	 if (!(_flags & RXF_FINDALL)) break;
	}
	return result;
}

/**
Function gets positions and length of matching substrings in <i>txt</i>.
@param txt Text to search for matches.
@param dst Vector to store results in. For each match function appends a pair
(vector of two elements) to <i>dest</i>, where first element represents
position of match in <i>txt</i> and second its length.
@param start Start position of search in <i>txt</i>.
@param len Maximum length of search, if <i>len</i>+<i>start</i> exceeds lemgth of <i>txt</i>
or <i>len</i> is negative, search is done to end of <i>txt</i>.
@return Number of matching substrings appended to <i>dst</i>.
@note Behavior of method depends on RXF_FINDALL flag. If it's set, all matches are
appended to <i>dst</i>, otherwise search is aborted after first match.
@see getMatches, search, next
*/
unsigned BppRegEx::getMatchPositions(const BppString & txt, BppVector & dst, size_t start, size_t len)
{
	if (!_trex.Ok()) return 0;
	_ptxt = txt;
	size_t l = txt.length();
	if (start >= l) return 0;
	_pos = start;
	const BppUChar* ptxt = (const BppUChar*)_ptxt.c_str() + start;
	const BppUChar* p01 = ptxt;
	if (len > l - start) len = l - start;
	_len = len;
	const BppUChar* p02 = p01 + len;
	const BppUChar *p1, *p2;
	unsigned result = 0;
	while (_trex.SearchRange(p01, p02, &p1, &p2)) {
		ptrdiff_t diff = p2 - p1;
		BppVector* p = new BppVector(2, true);
		dst.push_back(p);
		(*p)[0] = (BppUInt)(p1 - ptxt);
		(*p)[1] = (BppUInt)(diff);
		++result;
		p01 = p2;
	  if (!(_flags & RXF_FINDALL)) break;
	}
	return result;
}

/**
Function gets submatch descriptors of last match.
@param dst Vector to store results in. For each submatch a dictionary of two elements is appended to <i>dst</i>, where
item 'pos' (BppInt) contains position and item 'text' (BppString) the matching text. First element appended is the main match itselfs.
@return Number of elements appended to <i>dst</i>.
@note Method is intended for use with #search, #next and #match functions..
@par Example
@code
// Find numbers behind a letter
BppRegEx rex;
if (rex.compile(_T("\\a(\\d+)"))) {
	for(bool ok = rex.search(_T("abc123x4y5z"));ok;ok=rex.next()) 
	{
		BppVector v;
		unsigned cnt = rex.getSubExpressions(v);
		for (unsigned i=0;i<cnt;++i) {
		 BppDictionary* pd = v[i];
		  printf(_T("%u pos: %d text: '%s'\n"),BppInt((*pd)[_T("pos")]),((BppString*)(*pd)[_T("text")])->c_str());
		}
	}
}
else printf(_T("%s\n"),getError().c_str();
@endcode
@see search, next, match
*/
unsigned BppRegEx::getSubExpressions(BppVector & dst)
{
	static const BppString text = _T("text");
	static const BppString pos = _T("pos");
	if (!_trex.Ok()) return 0;
	int n = _trex.GetSubExpCount();
	const BppUChar* p0 =(const BppUChar*) _ptxt.c_str();
	const BppUChar* p1;
	int len;
	unsigned result=0;
	ptrdiff_t pos0 = 0;
	for (int i = 0; i < n; ++i) {
		if (_trex.GetSubExp(i, &p1, &len)) {
		  ptrdiff_t posi = p1 - p0;
			++result;
			BppDictionary* pd = new BppDictionary();
			BppValue v = pd;
		 if (posi < pos0) {
			 posi = -1;
			 (*pd)[text] = new BppString();
			 (*pd)[pos] = BppInt(-1);
		 }
		 else {
			 BppChar* buf = (BppChar*)allocmem((len + 1) * sizeof(BppChar));
			 _tcsncpy(buf, (const BppChar*)p1, len);
			 buf[len] = 0;
			 (*pd)[text] = new BppString((const BppChar*)buf);
			 freemem(buf);
			 (*pd)[pos] = BppInt(posi);
			 pos0 = posi;
		 }
			dst.push_back(v);
		}
	}
	return result;
}


/**
Function combines #search, #next and #getSubExpressions functions. It iterates through specified text collecting all matches and their submatches.
@param txt Text to analyze.
@param dst Vector to store results into. For each match a new vector is appended to <i>dst</i>, each containing a vector of submatches.
@param start Start position of search in <i>txt</i>.
@param len Maximum length of search, if <i>len</i>+<i>start</i> exceeds lemgth of <i>txt</i>
or <i>len</i> is negative, search is done to end of <i>txt</i>.
@return Number of items (matches found) appended to <i>dst</i>.
@see getSubExpressions(BppVector&)
*/
unsigned BppRegEx::getSubExpressions(const BppString & txt, BppVector & dst, size_t start, size_t len)
{
	unsigned result = 0;
	for (bool ok = search(txt, start, len); ok; ok = next(), ++result) {
		BppVector* pv = new BppVector();
		dst.push_back(pv);
		getSubExpressions(*pv);
	}
	return result;
}

#ifndef DOXY
// Helper for replace implementation, builds format string for replacement text
// and fills argdesc with descriptors of submatch references.
static void buildFmtDef(BppString& fmt, BppVector& argdesc, const BppString& repl)
{
	static const BppString sfmt = _T("%s");
	size_t rl = repl.length();
	size_t pos0 = 0;
	for (long pos = repl.find(_T('$'), pos0); pos >= 0 && pos < (long)rl - 1; pos = repl.find(_T('$'), pos0))
	{
		if (pos > pos0) fmt += repl.substr(pos0, pos-pos0);
		pos0 = pos + 1;
		BppUChar c = repl[pos0];
		switch (c)
		{
		case '$':
			fmt += c;
			++pos0;
			break;
		case '&': // entire match
		case '0': // entire match
			fmt += sfmt;
			argdesc.push_back(BppInt(0));
			++pos0;
			break;
		case '`': // prefix
			fmt += sfmt;
			argdesc.push_back(BppInt(-1));
			++pos0;
			break;
		case '\'': // suffix
			fmt += sfmt;
			argdesc.push_back(BppInt(-2));
			++pos0;
			break;
		default:
			if (_istdigit(c)) {
				BppInt d = _ttol(repl.c_str() + pos0);
				if (d) {
					fmt += sfmt;
					argdesc.push_back(d);
					++pos0;
					while (d /= 10) {
						++pos0;
					}
				}
			}
			else fmt += _T('$');
		}
	}
	fmt += repl.substr(pos0);
}
#endif

/**
Function builds new string from <i>txt</i> replacing all matches in <i>txt</i> by <i>repl</i>.
@param txt Text to use for replace matches.
@param repl Replacement expression. The expression is a string, optionally containing following placeholders/variables:
<ul>
<li><b>$&</b> or <b>$0</b> - Inserts entire matching text.</li>
<li><b>$`</b> (back tick) - Inserts all text before match.</li>
<li><b>$'</b> (apostrophe) - Inserts all text behind match.</li>
<li><b>$<i>n</i></b> - Inserts text of <i>n</i>-th submatch, where <i>n</i> is a positive decimal number. If <i>n</i> exceeds number of submatches
available, string representation of <i>n</i> is inserted.</li>
<li><b>$$</b> - Inserts Dollar character ($).
</ul>
@return Result of replace operation
@see strReplace
*/
BppString BppRegEx::replace(const BppString & txt, const BppString& repl)
{
	BppString text = txt;
	strReplace(text, repl);
	return text;
}



/**
Function replaces all matches in <i>txt</i> by <i>repl</i>.
@param txt Text to replace matches in.
@param repl Replacement expression. The expression is a string, optionally containing following placeholders/variables:
<ul>
  <li><b>$&</b> or <b>$0</b> - Inserts entire matching text.</li>
  <li><b>$`</b> (back tick) - Inserts all text before match.</li>
  <li><b>$'</b> (apostrophe) - Inserts all text behind match.</li>
  <li><b>$<i>n</i></b> - Inserts text of <i>n</i>-th submatch, where <i>n</i> is a positive decimal number. If <i>n</i> exceeds number of submatches
  available, string representation of <i>n</i> is inserted.</li>
  <li><b>$$</b> - Inserts Dollar character ($).
</ul>
@return Number of matches replaced
@note Function modifies <i>txt</i> argument. Use #replace to create new string instead.
@see replace
*/
unsigned BppRegEx::strReplace(BppString & txt, const BppString& repl)
{
	_ptxt = txt;
	BppVector repparts;
	size_t l = txt.length();
	_pos = 0;
	_len = l;
	const BppUChar* ptxt = (const BppUChar*)_ptxt.c_str();
	const BppUChar* p01 = ptxt;
	const BppUChar* p02 = p01 + l;
	const BppUChar *p1, *p2;
	BppString fmt;
	BppVector argdesc;
	buildFmtDef(fmt, argdesc, repl);
	BppString result;
	unsigned cnt = 0;
	while (_trex.SearchRange(p01, p02, &p1, &p2))
	{
		++cnt;
		ptrdiff_t diff = p1 - p01;
		BppChar* buf = (BppChar*)allocmem((diff + 1) * sizeof(BppChar));
		_tcsncpy(buf, (const BppChar*)p01, diff);
		buf[diff] = 0;
		result += buf;
		freemem(buf);
		diff = p2 - p1;
		// Ersetzung einbauen
		unsigned dn = argdesc.size();
		if (dn == 0) result += fmt;
		else
		{
			BppVector subexp;
			unsigned nsubexp = getSubExpressions(subexp);
			BppVector args;
			for (unsigned i = 0; i < dn; ++i) {
				BppInt v = argdesc[i];
				switch (v) {
				case -2: // suffix
					args.push_back(new BppString((const BppChar*)p2));
					break;
				case -1: // prefix
					args.push_back(result);
					break;
				case 0: // full match
					buf = (BppChar*)allocmem((diff + 1) * sizeof(BppChar));
					_tcsncpy(buf, (const BppChar*)p1, diff);
					buf[diff] = 0;
					args.push_back(new BppString((const BppChar*)buf));
					freemem(buf);
					break;
				default: // subexp
					if (v <= nsubexp) {
						BppDictionary* pd = subexp[v - 1];
						args.push_back(pd->getFirst()->getValue());
					}
					else
					{
						BppString suffix;
						while (v >= 10) {
							suffix.insert(0, v % 10 + _T('0'));
							v /= 10;
							if (v <= nsubexp) {
								BppString* ps = subexp[v - 1];
								if (ps) suffix.insert(0, (const BppString&)*ps);
								break;
							}
						}
						args.push_back(suffix);
					}
				}
			}
			ValPtr argbegin = args.first();
			result += BppString(fmt.c_str(), args.size(), argbegin);
		}
		if (!(_flags & RXF_FINDALL)) break;
		p01 = p2;
		_pos = p2 - ptxt;
	}
	result += (const BppChar*)p2;
	if (cnt) txt = result;
	return cnt;
}


#ifndef DOXY

// Bpp_RegEx implementation

unsigned Bpp_RegEx::_idxPattern = 0;
unsigned Bpp_RegEx::_idxFlags = 0;

static BppUInt buildFlags(BppString* s) {
	BppUInt f = 0;
	if (s) {
		for (const BppChar* p = s->c_str(); *p; ++p) {
			switch (*p) {
			case _T('i'):
				f |= BppRegEx::RXF_IGNORECASE;
				break;
			case _T('m'):
				f |= BppRegEx::RXF_MULTILINE;
				break;
			case _T('g'):
				f |= BppRegEx::RXF_FINDALL;
				break;
			}
		}
	 }
	return f;
}
// RegEx(string pattern = null, string flags = null
bool Bpp_RegEx::ctor(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 3);
	BppValue& flagsVal = MEMBER(_idxFlags);
	BppValue& patternVal = MEMBER(_idxPattern);
	BppUInt f = 0; 
	if (argc == 3) {
		BppValue v(BVT_STRING, BVF_STRONGTYPED, false);
		if (!v.assign(ARG(2))) return interpreter->badType(ARG(2), v);
			f = buildFlags(v);
	}
	flagsVal = f;
	if (argc >= 2) {
		if (!patternVal.assign(ARG(1))) return interpreter->badType(ARG(1), patternVal);
	}
	BppString* pattern = patternVal;
	BppRegEx* p = new BppRegEx();
	if (pattern && pattern->length()) {
		if (!p->compile(*pattern, f)) {
			interpreter->returnError(p->getError().c_str(), pattern->c_str());
			delete p;
			return false;
		}
	}
	pThis->_nativeObject = p;
	*bp = pThis;
	sp = bp;
	return true;
}

bool Bpp_RegEx::dtor(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) delete p;
	pThis->_nativeObject = 0;
	sp = bp;
	return true;
}

bool Bpp_RegEx::setFlags(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	CHKVAR(bp[1]);
	BppValue v(BVT_STRING,BVF_STRONGTYPED,false);
	if (!v.assign(ARG(1))) return interpreter->badType(ARG(1), v);
	BppUInt f = buildFlags(v);
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) p->setFlags(f);
	MEMBER(_idxFlags) = f;
	bp->dispose();
	sp = bp;
	return true;
}

bool Bpp_RegEx::getFlags(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppUInt f = MEMBER(_idxFlags);
	BppString* s = new BppString();
	*bp = s;
	if (f & BppRegEx::RXF_IGNORECASE) (*s) += _T('i');
	if (f & BppRegEx::RXF_MULTILINE)  (*s) += _T('m');
	if (f & BppRegEx::RXF_FINDALL)    (*s) += _T("g");
	sp = bp;
	return true;
}

// bool compile(string pattern, string flags = null)
bool Bpp_RegEx::compile(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 3);
	CHKVAR(bp[1]);
	BppValue& flagsVal = MEMBER(_idxFlags);
	BppValue& patternVal = MEMBER(_idxPattern);
	BppUInt f = 0;
	if (argc == 3) {
		const BppValue& a = ARG(2);
		if (a.getType() != BVT_NULL) {
			BppValue v(BVT_STRING, BVF_STRONGTYPED, false);
			if (!v.assign(a)) return interpreter->badType(a, v);
			f = buildFlags(v);
			flagsVal = f;
		}
	}
	else f = flagsVal;

	const BppValue& ap = ARG(1);
	if (ap.getType() != BVT_NULL && (!patternVal.assign(ap))) return interpreter->badType(ap,patternVal);
	BppString* pattern = patternVal;
	if (!pattern) return false; // sollte nie passieren
	if (!pattern->length()) THROW(REX_PATTERN_EMPTY);
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (!p) {
		p = new BppRegEx();
		pThis->_nativeObject = p;
	}
	bool ok = p->compile(*pattern, f);
	*bp = BppValue::makeBool(ok);
	sp = bp;
	return true;
}

bool Bpp_RegEx::getError(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) *bp = new BppString(p->getError());
	else *bp = new BppString();
	sp = bp;
	return true;
}

// bool match(string txt)
bool Bpp_RegEx::match(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	CHKVAR(bp[1]);
	BppValue v(BVT_STRING,BVF_STRONGTYPED,false);
	BppString* txt = ARG(1);
	if (!txt) {
		if (!v.assign(ARG(1))) return interpreter->badType(ARG(1), v);
		txt = v;
	}
	if (!txt) return false;
	bool ok = false;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) ok = p->match(*txt);
	*bp = BppValue::makeBool(ok);
	sp = bp;
	return true;
}

// bool search(string txt, uint start = 0, uint len = NPOS)
bool Bpp_RegEx::search(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 4);
	CHKVAR(bp[1]);
	BppUInt start = 0;
	BppUInt len = BPP_NPOS;
	BppValue v(BVT_UINT,BVF_STRONGTYPED,false);
	if (argc == 4) {
		if (!v.assign(ARG(3))) return interpreter->badType(ARG(3),v);
		len = v;
	}
	if (argc > 2) {
		v = (BppUInt)0;
		if (!v.assign(ARG(2))) return interpreter->badType(ARG(2), v);
		start = v;
	}
	BppString* txt = ARG(1);
	if (!txt) {
		v = (BppString*)0;
		if (!v.assign(ARG(1))) return interpreter->badType(ARG(1), v);
		txt = v;
	}
	if (!txt) return false;
	bool ok = false;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) ok = p->search(*txt, start, len);
	*bp = BppValue::makeBool(ok);
	sp = bp;
	return true;
}

bool Bpp_RegEx::next(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(1, 1);
	CHKVAR(bp[1]);
	bool ok = false;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) ok = p->next();
	*bp = BppValue::makeBool(ok);
	sp = bp;
	return true;
}
// uint getMatches(string txt, vector dst, uint start = 0, uint len = -1)
bool Bpp_RegEx::getMatches(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 5);
	CHKVAR(bp[1]);
	BppUInt start = 0;
	BppUInt len = BPP_NPOS;
	BppValue v(BVT_INT,BVF_STRONGTYPED,false);
	if (argc == 5) {
		if (!v.assign(ARG(4))) return interpreter->badType(ARG(4),v);
		len = v;
	}
	if (argc > 3) {
		v = BppUInt(0);
		if (!v.assign(ARG(3))) return interpreter->badType(ARG(3), v);
		start = v;
	}
	BppValue vv(BVT_VECTOR,BVF_STRONGTYPED,false);
	BppVector* dst = ARG(2);
	if (!dst) {
		if (vv.assign(ARG(2))) dst = vv;
		if (!dst) return interpreter->badType(ARG(2), vv);
	}
	BppString* txt = ARG(1);
	if (!txt) {
		v = (BppString*)0;
		if (v.assign(ARG(1))) txt = vv;
		if (!txt) return interpreter->badType(ARG(1), v);
	}
	BppUInt cnt = 0;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) cnt = p->getMatches(*txt, *dst, start, len);
	*bp = cnt;
	sp = bp;
	return true;
}

// getMatchPositions(string txt, vector dst, uint start = 0, uint len = NPOS)
bool Bpp_RegEx::getMatchPositions(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 5);
	CHKVAR(bp[1]);
	BppUInt start = 0;
	BppUInt len = BPP_NPOS;
	BppValue v(BVT_INT, BVF_STRONGTYPED, false);
	if (argc == 5) {
		if (!v.assign(ARG(4))) return interpreter->badType(ARG(4), v);
		len = v;
	}
	if (argc > 3) {
		v = BppUInt(0);
		if (!v.assign(ARG(3))) return interpreter->badType(ARG(3), v);
		start = v;
	}
	BppValue vv(BVT_VECTOR, BVF_STRONGTYPED, false);
	BppVector* dst = ARG(2);
	if (!dst) {
		if (vv.assign(ARG(2))) dst = vv;
		if (!dst) return interpreter->badType(ARG(2), vv);
	}
	BppString* txt = ARG(1);
	if (!txt) {
		v = (BppString*)0;
		if (v.assign(ARG(1))) txt = vv;
		if (!txt) return interpreter->badType(ARG(1), v);
	}
	BppUInt cnt = 0;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) cnt = p->getMatchPositions(*txt, *dst, start, len);
	*bp = cnt;
	sp = bp;
	return true;
}

// uint getSubExpressions(vector dst)
// uint getSubExpressions(string txt, vector dst, uint start = 0, uint len = NPOS)
bool Bpp_RegEx::getSubExpressions(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	static const BppChar* fname = _T("getSubExpressions");
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 5);
	CHKVAR(bp[1]);

	BppValue vv(BVT_VECTOR, BVF_STRONGTYPED, false);
	BppVector* dst = 0;
	if (vv.assign(ARG((argc == 2) ? 1 : 2))) dst = vv;
	if (!dst) THROW(REX_INVALID_ARG_VECTOR_EXPECTED, fname);
	BppUInt cnt = 0;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (argc == 2) {
		if (p) cnt = p->getSubExpressions(*dst);
	}
	else {
		BppUInt start = 0;
		BppUInt len = BPP_NPOS;
		BppValue v(BVT_INT,BVF_STRONGTYPED,false);
		if (argc == 5) {
			if (!v.assign(ARG(4))) return interpreter->badType(ARG(4),v);
			len = v;
		}
		if (argc > 3) {
			v = BppUInt(0);
			if (!v.assign(ARG(3))) return interpreter->badType(ARG(3), v);
			start = v;
		}
		BppString* txt = ARG(1);
		if (!txt) {
			v = (BppString*)0;
			if (!v.assign(ARG(1))) return interpreter->badType(ARG(1), v);
			txt = v;
			if (!txt) return false;
		}
		if (p) cnt = p->getSubExpressions(*txt, *dst, start, len);
	}
	*bp = cnt;
	sp = bp;
	return true;
}


// string replace(string txt, string repl)
bool Bpp_RegEx::replace(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 3);
	CHKVAR(bp[1]);

	BppValue vt(BVT_STRING, BVF_STRONGTYPED,false);
	if (!vt.assign(ARG(1))) return interpreter->badType(ARG(1), vt);
	BppString* txt = vt;
	if (!txt) THROW(REX_TEXT_IS_NULL);

	BppValue vr(BVT_STRING, BVF_STRONGTYPED, false);
	if (!vr.assign(ARG(2))) return interpreter->badType(ARG(2), vr);
	BppString* repl = vr;
	if (!repl) THROW(REX_REPLACEMENT_IS_NULL);
	BppValue vresult = txt->clone();
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) p->strReplace(*((BppString*)vresult), *repl);
	*bp = vresult;
	sp = bp;
	return true;
}

//uint strReplace(string txt, string repl)
bool Bpp_RegEx::strReplace(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 3);
	CHKVAR(bp[1]);

		BppValue vt(BVT_STRING, BVF_STRONGTYPED, false);
		if (!vt.assign(ARG(1))) return interpreter->badType(ARG(1), vt);
		CHKVAR(ARG(1));
		BppString* txt = vt;
		if (!txt) THROW(REX_TEXT_IS_NULL);

		BppValue vr(BVT_STRING, BVF_STRONGTYPED, false);
		if (!vr.assign(ARG(2))) return interpreter->badType(ARG(2), vr);
		BppString* repl = vr;
		if (!repl) THROW(REX_REPLACEMENT_IS_NULL);
	BppUInt cnt = 0;
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (p) cnt = p->strReplace(*txt, *repl);
	*bp = cnt;
	sp = bp;
	return true;
}

bool Bpp_RegEx::op_call(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	return argc > 1 ? search(vm,argc,sp) : next(vm,argc,sp);
}


bool Bpp_RegEx::op_pref(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	static const BppChar* keys[] = {
		_T("flags"), _T("global"), _T("ignoreCase"), _T("multiline"),
		_T("source"), _T("text"), _T("pos"), _T("len")
	};
	static const size_t kcnt = sizeof(keys) / sizeof(BppChar*);
	CHKVM;
	CHKOBJECT;
	CHKARGC(2, 2);
	BppString* key = ARG(1);
	if (!key) THROW(REX_KEY_EXPECTED);
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (!p) THROW(REX_OBJ_NOT_INITIALIZED);
	const BppChar* pkey = key->c_str();
	unsigned flags = p->getFlags();
	size_t idx = 0;
	for (; idx < kcnt; ++idx) {
		if (_tcscmp(pkey, keys[idx]) == 0) break;
	}
	switch (idx) {
	case 0: // flags
	{
		BppString* s = new BppString();
		*bp = s;
		if (flags & BppRegEx::RXF_FINDALL) (*s) += _T('g');
		if (flags & BppRegEx::RXF_MULTILINE) (*s) += _T('m');
		if (flags & BppRegEx::RXF_IGNORECASE) (*s) += _T('i');
	}
	break;
	case 1: // global
		*bp = BppValue::makeBool(flags & BppRegEx::RXF_FINDALL);
		break;
	case 2: // ignoreCase
		*bp = BppValue::makeBool(flags & BppRegEx::RXF_IGNORECASE);
		break;
	case 3: // multiline
		*bp = BppValue::makeBool(flags & BppRegEx::RXF_MULTILINE);
		break;
	case 4: // source
		*bp = p->_pattern.clone();
		break;
	case 5: // text
		*bp = p->_ptxt.clone();
		break;
	case 6: // pos
		*bp = BppUInt(p->_pos);
		break;
	case 7: // len
		*bp = BppUInt(p->_len);
		break;
	default:
		THROW(REX_INVALID_KEY,pkey);
	}
	sp = bp;
	return true;
}

bool Bpp_RegEx::op_pset(BppVirtualMachine * vm, int argc, ValPtr & sp)
{
	static const BppChar* keys[] = {
		_T("flags"), _T("global"), _T("ignoreCase"), _T("multiline"),
		_T("text"), _T("pos"), _T("len")
	};
	static const size_t kcnt = sizeof(keys) / sizeof(BppChar*);
	CHKVM;
	CHKOBJECT;
	CHKARGC(3, 3);
	CHKVAR(bp[1]);
	BppString* key = ARG(1);
	if (!key) THROW(REX_KEY_EXPECTED);
	BppRegEx* p = (BppRegEx*)pThis->_nativeObject;
	if (!p) THROW(REX_OBJ_NOT_INITIALIZED);
	const BppChar* pkey = key->c_str();
	size_t idx = 0;
	for (; idx < kcnt; ++idx) {
		if (_tcscmp(pkey, keys[idx]) == 0) break;
	}
	BppValue vi(BVT_INT, BVF_STRONGTYPED, false);
	switch (idx) {
	case 0: // flags
	{
		BppString* s = ARG(2);
		if (!s) {
			BppValue vs(BVT_STRING, BVF_STRONGTYPED, false);
			if (!vs.assign(ARG(2))) return interpreter->badType(ARG(2), vs);
			s = vs;
		}
		if (!s) return false;
		p->setFlags(buildFlags(s));
	}
	break;
	case 1: // global
	{
		if (!vi.assign(ARG(2))) return interpreter->badType(ARG(2), vi);
		BppInt f = p->_flags;
		BppInt val = vi;
		if (val) p->setFlags(f | BppRegEx::RXF_FINDALL);
		else p->setFlags(f & ~BppRegEx::RXF_FINDALL );
	}
		break;
	case 2: // ignoreCase
	{
		if (!vi.assign(ARG(2))) return interpreter->badType(ARG(2), vi);
		BppInt f = p->_flags;
		BppInt val = vi;
		if (val) p->setFlags(f | BppRegEx::RXF_IGNORECASE);
		else p->setFlags(f & ~BppRegEx::RXF_IGNORECASE);
	}
	break;
	case 3: // multiline
	{
		if (!vi.assign(ARG(2))) return interpreter->badType(ARG(2), vi);
		BppInt f = p->_flags;
		BppInt val = vi;
		if (val) p->setFlags(f | BppRegEx::RXF_MULTILINE);
		else p->setFlags(f & ~BppRegEx::RXF_MULTILINE);
	}
	break;
	case 4: // text
	{
		BppString* s = ARG(2);
		if (!s) {
			BppValue vs(BVT_STRING, BVF_STRONGTYPED, false);
			if (!vs.assign(ARG(2))) return interpreter->badType(ARG(2), vs);
			s = vs;
		}
		if (!s) return false;
		p->_ptxt = *s;
		p->_len = -1;
		p->_pos = 0;
	}
	break;
	case 5: // pos
	{
		vi = BppUInt(0);
		if (!vi.assign(ARG(2))) return interpreter->badType(ARG(2), vi);
		BppUInt val = vi;
		p->_pos = val;
	}
	break;
	case 6: // len
	{
		if (!vi.assign(ARG(2))) return interpreter->badType(ARG(2), vi);
		BppUInt val = vi;
		p->_len = val;
	}
	break;
	default:
		THROW(REX_INVALID_KEY, pkey);
	}
	*bp = ARG(2); // pThis;
	sp = bp;
	return true;
}


void Bpp_RegEx::createClass(BppVirtualMachine & vm)
{
	BppClass* pClass = new BppClass(vm);
	if (!vm.registerNativeClass(_T("RegEx"), pClass)) return;
	pClass->addDataMember(_T("_pattern"),BppValue(BVT_STRING,BppValueFlags(BVF_PRIVATE|BVF_STRONGTYPED),true));
	pClass->addDataMember(_T("_flags"), BppValue(BVT_UINT, BppValueFlags(BVF_PRIVATE | BVF_STRONGTYPED), true));
	pClass->addMethod(_T("RegEx"), &ctor);
	pClass->addMethod(_T("dtor"), &dtor);
	pClass->addMethod(_T("setFlags"), &setFlags);
	pClass->addMethod(_T("getFlags"), &getFlags);
	pClass->addMethod(_T("compile"), &compile);
	pClass->addMethod(_T("getError"), &getError);
	pClass->addMethod(_T("match"), &match);
	pClass->addMethod(_T("search"), &search);
	pClass->addMethod(_T("next"), &next);
	pClass->addMethod(_T("getMatches"), &getMatches);
	pClass->addMethod(_T("getMatchPositions"), &getMatchPositions);
	pClass->addMethod(_T("getSubExpressions"), &getSubExpressions);
	pClass->addMethod(_T("replace"), &replace);
	pClass->addMethod(_T("strReplace"), &strReplace);
	pClass->addMethod(_T("OP_CALL"), &op_call);
	pClass->addMethod(_T("OP_PREF"), &op_pref);
	pClass->addMethod(_T("OP_PSET"), &op_pset);
	pClass->fixupDataDefs();
	_idxPattern = pClass->getMemberIndex(_T("_pattern"));
	_idxFlags = pClass->getMemberIndex(_T("_flags"));
	pClass->setNotifyProc(&notifyProc);
}



void BPPAPI Bpp_RegEx::notifyProc(BppVirtualMachine * vm, BppObject * obj, void ** nativeObject, BppObjectNotifyType reason)
{
	if ((!obj) || (!vm)) return;
	BppRegEx* p = (BppRegEx*)(obj->_nativeObject);
	switch (reason) {
	case BPP_OBJ_COPY:
	case BPP_OBJ_AFTERDESERIALIZE:
		 {
			BppString* pat = (*obj->getMemberData())[_idxPattern];
			BppUInt flags = (*obj->getMemberData())[_idxFlags];
			if (!p) obj->_nativeObject = p = new BppRegEx();
			if (pat) p->compile(*pat, flags);
			else p->setFlags(flags);
		}
		 break;
	case BPP_OBJ_DELETE:
		if (p) {
			delete p;
			obj->_nativeObject = 0;
		}
		break;
	}
}

#endif
