
#include <stdlib.h>
#include <ctype.h>
#include "bppstrng.h"
#include "bppbuf.h"
#include "bppufunc.h"
#include "bppvalue.h"
#include "bppvec.h"
#include "bppufunc.h"
#include "bppvalue.h"

BppString::BppString() BPPNOEXCEPT
:_data(0)
{
}

/**
Constructor copies character string of <i>cpy</i>.
@param cpy Other string to construct from.
*/
BppString::BppString(const BppString& cpy)
:BppRefObject(cpy),_data(0)
{
	size_t len = cpy.length();
	if (len > 0) 
	{
	  _data = dupstr(cpy._data);
	}
}

#ifdef CPP11
/**
Constructor moves character string of <i>cpy</i>.
@param cpy Other string to construct from.
*/
BppString::BppString(BppString&& cpy) BPPNOEXCEPT
	:_data(cpy._data)
{
	cpy._data = 0;
}
#endif
/**
@param src Ansi-C-string to copy characters from.
*/
BppString::BppString(const char* src)
:_data(0)
{
	if (src == 0)
		return;
	size_t len = strlen(src)+1;
	size_t bytes = len * sizeof(BppChar);
	_data = (BppChar*) allocmem(bytes);
#ifndef _UNICODE
		_tcscpy(_data,src);
#else
	mbstowcs(_data,src,len);
#endif
}

/**
@param src Buffer to copy bytes as characters from.
*/
BppString::BppString(const BppBuffer& src)
:_data(0)
{
	size_t len = src.size()+1;
	size_t bytes = (len+1) * sizeof(BppChar);
	_data = (BppChar*) allocmem(bytes);
#ifndef _UNICODE
		_tcsncpy(_data,(const char*)src.data(),len);
#else
	mbstowcs(_data,(const char*)src.data(),len);
#endif
	_data[len] = 0;
}

/**
@param src Character buffer to copy contents from.
*/
BppString::BppString(const BppCharBuffer& src)
:_data(0)
{
	size_t len = src.size();
	size_t bytes = (len+1) * sizeof(BppChar);
	_data = (BppChar*) allocmem(bytes);
	_tcsncpy(_data,src.data(),len+1);
	_data[len] = 0;
}

#ifdef WCHAR_MAX
/**
@param src Unicode-C-string to copy characters from.
@note This constructor doesn't exist in DOS builds.
*/
BppString::BppString(const wchar_t* src)
:_data(0)
{
	if (src == 0)
		return;
#ifdef _UNICODE
	size_t len = _tcslen(src)+1;
	size_t bytes = len * sizeof(BppChar);
	_data = (BppChar*) allocmem(bytes);
	_tcscpy(_data,src);
#else
	size_t bytes = wcstombs(0,src,0);
	_data = (BppChar*) allocmem(bytes+1);
	wcstombs((char*) _data,src,bytes+1);
#endif
}
#endif

/// format flags
static const BppChar* fmtFlags=_T("+-0 #");
/// digits for width specification
static const BppChar* fmtDigits=_T("1234567890");
/// supported type prefix characters
static const BppChar* fmtPrefixChars=_T("whlI2346FN");
/// supported type characters
static const BppChar* fmtTypeChars = _T("diouxXfegEGcspCSaA\1");

/**
This constructor is internally used for implementation of newstring B++ function.
It builds a formatted string in a printf-manner assuming variable values to be
contained in a vector where <i>args</i> points to first value.<br>
The <i>format</i> string is same as used for printf C-function, i.e.
it may vary in some details depending on compiler used to build B++ library.
@param format printf-like format string
@param nargs number of variable arguments in <i>args<(i>
@param args value pointer to start of variable value array
@note
<ul>
	<li>The <tt>%n</tt> variable specification in <i>format</i> is not supported
	in this implementation.</li>
	<li>The <i>format</i> string is splitted to parts each containing only a single
	variable. Each of that parts is formatted using BppString::format method.
	The result is a concatenation of formatted parts.</li>
	<li>The <i>nargs</i> argument ensures not to run out of <i>args</i> bounds
	in case of a mistake in <i>format</i> notation.</li>
</ul>
*/
BppString::BppString(const BppChar* format,unsigned nargs, ValPtr& args)
:_data(0)
{
	if ((format == 0) || (_tcslen(format) == 0))
		return;

	BppChar* ptokstr = dupstr(format);
	for (BppChar* p0 = _tcschr(ptokstr, _T('%')); p0; p0 = _tcschr(++p0, _T('%'))) {
		BppChar c = p0[1];
		if (!c) break;
		if (c == _T('%')) {
			p0[1] = _T('\1');
		}
	}
	// count tokens
	const BppChar* delim = _T("%");
	BppString fmtstr;
	unsigned iargs = 0;
	for(BppChar* p=_tcstok(ptokstr,delim);(p != 0) /*&& (iargs < nargs)*/;p=_tcstok(0,delim))
	{
		if ((_tcslen(p) == 0) && (p != ptokstr))
		{
			(*this)+=delim;
			continue;
		}
	  if (p[0] == _T('\1')) {
		  (*this) += delim;
		  (*this) += p+1;
		  continue;
	  }
	  if (iargs >= nargs) {
		  (*this) += p;
		  continue;
	  }
		const BppChar* pc=p;
#ifdef EMULATE_I64
	  bool left = false;
	  bool plus = false;
	  bool opt = false;
	  int width = 0;
	  int prec = 1;
	  const BppChar* ptmp = 0;
	  BppChar typeChr = 0;
#endif
	  while (_tcschr(fmtFlags,pc[0]) != 0) {
#ifdef EMULATE_I64
		  switch (pc[0])
		  {
		  case _T('-'):
			  left = true;
			  break;
		  case _T('+'):
			  plus = true;
			  break;
		  case _T('#'):
			  opt = true;
			  break;
		  }
#endif
		  ++pc;
	  }
#ifdef EMULATE_I64
	  ptmp = pc;
#endif
		while (_tcschr(fmtDigits,pc[0]) != 0) ++pc;
#ifdef EMULATE_I64
	  if (ptmp != pc) width = _ttoi(ptmp);
#endif
		if (*pc == _T('.'))
		{
			 ++pc;
#ifdef EMULATE_I64
		  ptmp = pc;
#endif
			 while (_tcschr(fmtDigits,pc[0]) != 0)  ++pc;
#ifdef EMULATE_I64
		  if (ptmp != pc) prec = _ttoi(ptmp);
#endif
		}
		while (_tcschr(fmtPrefixChars,pc[0]) != 0) ++pc;
		if ((_tcschr(fmtTypeChars,*pc) == 0) || (p == ptokstr))
		{
			(*this) += BppString((const BppChar*) p);
			continue;
		}
#ifdef EMULATE_I64
		  ptmp = pc+1;
		  typeChr = *pc;
#endif
		BppValue arg = args[(int)iargs];
		fmtstr = delim;
		fmtstr += BppString((const BppChar*) p);
		BppValueType vt = arg.getType();
		if (arg.isCharacterType() && vt != BVT_TCHAR) {
			arg = BppValue::makeChar(arg);
			vt = BVT_INT;
		}
		switch (vt)
		{
		case BVT_DOUBLE:
		case BVT_FLOAT:
			(*this) += BppString::format(fmtstr.c_str(),(double)arg);
			break;
		case BVT_INT:
		case BVT_UINT:
		case BVT_SHORT:
		case BVT_USHORT:
		case BVT_SBYTE:
		case BVT_BYTE:
			(*this) += BppString::format(fmtstr.c_str(),(BppInt)arg);
			break;
		case BVT_INTPTR:
		case BVT_UINTPTR:
			(*this) += BppString::format(fmtstr.c_str(),(BppIntPtr)arg);
			break;
		case BVT_LONG:
		case BVT_ULONG:
#ifdef EMULATE_I64
		  {
			  size_t l = fmtstr.length();
				  int base = 10;
			  bool upper = false;
			  bool positive = arg.getType() == BVT_ULONG;
				  bool flt = false;
			  bool chr = false;
			  switch (typeChr)
			  {
			  case 'X':
			  case 'P':
				  upper = true;
			  case 'x':
			  case 'p':
				  base = 16;
				  positive = true;
				  break;
			  case 'o':
				  base = 8;
			  case 'u':
				  positive = true;
				  break;
			  case 'd':
			  case 'i':
				  positive = false;
				  break;
			  case 'c':
			  case 'C':
				  chr = true;
				  break;
			  case 'f':
			  case 'e':
			  case 'E':
			  case 'g':
			  case 'G':
				  flt = true;
			  }
			  if (flt)
						(*this) += BppString::format(fmtstr.c_str(),(double)arg);
			  else if (chr)
						(*this) += BppString::format(fmtstr.c_str(),(int)(BppInt)arg);
			  else
			  {
				  size_t startpos = length();
				  if ((prec > 0) || arg.ptr())
				  {
					  BppChar buf[65];
					  if (positive)
						 (*this) += (const BppChar*) BppULong(arg).toa(buf,base,upper);
					  else
					  {
						 (*this) += (const BppChar*) BppLong(arg).toa(buf,base,upper);
						 if ((plus) && _istdigit(buf[0])) insert(startpos,_T('+'));
					  }
					 int flen = length() - startpos;
					 if (prec > 1)
					 {
								 int digits = flen;
						 size_t idx = startpos;
						 if ((!positive) && (!_istdigit((*this)[startpos])))
						 {
							 --digits;
							 ++idx;
						 }
						 int cnt = prec-digits;
						 if (cnt > 0)
						 {
							 insert(idx,_T('0'),cnt);
							 flen += cnt;
						 }
					 }
					 if (opt && base != 10)
					 { 
						 int cnt = 1;
						 insert(startpos,_T('0'));
						 if (base == 16)
						 {
							 ++cnt;
							 insert(startpos+1,upper ? _T('X') : _T('x'));
						 }
						 flen += cnt;
					 }
					 width -= flen;
					 if (width > 0)
					 {
						 if (left) startpos = length();
						 insert(startpos,_T(' '),width);
					 }
				  }
			  }
			  if (ptmp && (*ptmp))
				  (*this) += (const BppChar*) ptmp;
		  }
#else
			(*this) += BppString::format(fmtstr.c_str(),(BppLong)arg);
#endif
			break;
		default:
			(*this) += BppString::format(fmtstr.c_str(),arg.addr());
		}
		iargs++;
	}
	freemem(ptokstr);

}


/**
Destructor simply releases internal character array.
*/
BppString::~BppString()
{
	if (_data != 0)
		freemem(_data);
	_data = 0;
}

/**
@return new copy of current string
@note Caller is responsible for delete result if no longer needed.
*/
BppString* BppString::clone() const
{
	return new BppString(*this);
}

/**
@param assignment string to assign value from
@return reference to current string
*/
BppString& BppString::operator=(const BppString& assignment)
{
	if (&assignment != this)
	{
		BppChar* data = 0;
		if (assignment._data != 0)
		{
			data = dupstr(assignment._data);
		}
		if (_data != 0)
			freemem(_data);
		_data = data;
	}
	return *this;
}

#ifdef CPP11
/**
@param assignment string to move value from
@return reference to current string
*/
BppString& BppString::operator=(BppString&& assignment)
{
	if (&assignment != this && _data != assignment._data)
	{
		if (_data != 0)
			freemem(_data);
		_data = assignment._data;
		assignment._data = 0;
	}
	return *this;
}
#endif
/**
@param assignment C-string to assign value from
@return reference to current string
*/
BppString& BppString::operator=(const BppChar* assignment)
{
	BppChar* data = 0;
	if (assignment != 0)
	{
		data = dupstr(assignment);
	}
	if (_data != 0)
		freemem(_data);
	_data = data;
	return *this;
}

/**
@param appendix string to append
@return reference to current string
*/
BppString& BppString::operator+=(const BppString& appendix)
{
	return operator+=(appendix.c_str());
}

/**
@param appendix C-string to append
@return reference to current string
*/
BppString& BppString::operator+=(const BppChar* appendix)
{
	if (appendix == 0)
		return *this;
	if (_data == 0)
	  return operator=(appendix);
	size_t l = _tcslen(appendix);
	if (l > 0)
	{
		size_t newlen = _tcslen(_data)+l +1;
		size_t reqSize = newlen * sizeof(BppChar);
	  BppChar* p = (BppChar*) allocmem(reqSize);
	  if (_data) {
		  _tcscpy(p, _data);
		  freemem(_data);
	  }
	  _data = p;
	  _tcscat(_data, appendix);
	}
	return *this;
}

/**
@param appendix character to append
@return reference to current string
*/
BppString& BppString::operator+=(BppChar appendix)
{
	BppChar buf[2] = { 0,0 }; 
	buf[0] = appendix;
	return operator+=((const BppChar*)buf);
}

/**
@param appendix  string to append to current value
@return concatenated string
*/
BppString  BppString::operator+(const BppString& appendix) const
{
	BppString result = *this;
	result += appendix.c_str();
	return result;
}



/**
@param comparison string to compare with
@return <ul>
  <li>&lt;0 if current value less than <i>comparison</i></li>
  <li>0 if current value equals to <i>comparison</i></li>
  <li>&lt;0 if current value greater than <i>comparison</i></li>
</ul>
*/
int BppString::compare(const BppString& comparison) const BPPNOEXCEPT
{
	if (&comparison == this) return 0;
	return _tcscmp(c_str(),comparison.c_str());
}

/**
@param comparison string to compare with
@return <ul>
<li>&lt;0 if current value less than <i>comparison</i></li>
<li>0 if current value equals to <i>comparison</i></li>
<li>&lt;0 if current value greater than <i>comparison</i></li>
</ul>
*/
int BppString::icompare(const BppString& comparison) const BPPNOEXCEPT
{
	if (&comparison == this) return 0;
	return _tcsicmp(c_str(),comparison.c_str());
}

/**
@return number of characters contained in string (excluding terminal NUL character)
*/
size_t BppString::length() const BPPNOEXCEPT
{
	if (_data != 0)
		return  _tcslen(_data);
	return 0;
}


/**
@return single byte C-string representation of current value
*/
const char* BppString::a_str() const
{
	if (_data == 0)
		return "";
#ifdef _UNICODE
	static BppBuffer abuf;
	size_t bytes = (_tcslen(_data)+1) * sizeof(wchar_t);
	abuf.setCapacity(bytes);
	wcstombs((char*) abuf.data(),(const wchar_t*) _data,bytes);
	return (const char*) abuf.data();
#else
		return _data;
#endif
}

/**
@return wide character C-string representation of current value
*/
const wchar_t* BppString::w_str() const
{
#ifdef __OLDCPP__
	return (const wchar_t*) c_str();
#else
	if (_data == 0)
		return L"";
#ifdef _UNICODE
	return _data;
#else
	static BppBuffer wbuf;
	unsigned len = (unsigned) _tcslen(_data)+1;
	unsigned bytes = len * sizeof(wchar_t);
	wbuf.setCapacity(bytes);
	mbstowcs((wchar_t*) wbuf.data(),(const char*) _data,len);
	return (const wchar_t*) wbuf.data();
#endif // ! _UNICODE
#endif // ! __OLDCPP__
}

/**
@param cmp string to compare with
@return true if current value is less than value of <i>cmp</i>
@note Operator does a case sensitive alphanumeric compare. It is mainly
for internal use (needed to use BppString for key in std::map objects).
*/
bool BppString::operator<(const BppString& cmp) const BPPNOEXCEPT
{
  return compare(cmp) < 0;
}


bool BppString::operator== (const BppString& cmp) const BPPNOEXCEPT
{
	return compare(cmp) == 0;
}

bool BppString::operator!= (const BppString& cmp) const BPPNOEXCEPT
{
	return compare(cmp) != 0;
}

void BppString::clear()
{
	if (_data != 0)
		freemem(_data);
	_data = 0;
}

/**
Function searches for substring <i>s</i> in current string and returns position of its first
occurrence.
@param s string to search for
@param off index to start search from
@return position of <i>s</i> if found, otherwise BPP_NNPOS
*/
size_t BppString::find(const BppString& s, size_t off/* = 0*/) const BPPNOEXCEPT
{
	if (_data == 0) return BPP_NNPOS;
	size_t len = _tcslen(_data);
	if (off >= len) return BPP_NNPOS;
	BppChar* sp = &_data[off];
	const BppChar* p = _tcsstr(sp,s.c_str());
	if (p == 0) return BPP_NNPOS;
	return (p-_data);
}

/**
Function searches for character <i>c</i> in current string and returns position of its first
occurrence.
@param c character to search for
@param off index to start search from
@return position of <i>c</i> if found, otherwise BPP_NNPOS
*/
size_t BppString::find(BppChar c, size_t off/* = 0*/) const BPPNOEXCEPT
{
	if (_data == 0) return BPP_NNPOS;
	size_t len = _tcslen(_data);
	if (off >= len) return BPP_NNPOS;
	BppChar* sp = &_data[off];
	const BppChar* p = _tcschr(sp,c);
	if (p == 0) return BPP_NNPOS;
	return (p-_data);
}

/**
Function builds a new string containing substring of current one,
starting at position start and of <i>len</i> charecters length.
@param start start position of substring
@param len maximum langth of substring
@return substring
@note If <i>len</i>
greater than number of characters behind <i>start</i> position,
substring is built from <i>start</i> to end of current string.
*/
BppString BppString::substr(size_t start, size_t len) const
{
	BppString result;
	size_t mylen = length();
	if ((start < mylen) && (len != 0))
	{
		if (len > mylen || start+len > mylen)
			len = mylen - start;
		BppChar* buf = (BppChar*) allocmem((len+1) * sizeof(BppChar));
		_tcsncpy(buf,&_data[start],len+1);
		buf[len] = 0;
		result._data = buf;
	}
	return result;
}

/**
Function erases maximum <i>cnt</i> characters from current string, starting
at position <i>pos</i>.
@param pos start position for erase
@param cnt maximum number of characters to delete
@return reference to current string
@note If a negative value was specified for <i>cnt</i> or <i>cnt</i>
greater than number of characters behind <i>pos</i>,
all characters from <i>pos</i> to end of current string will be erased.
*/
BppString& BppString::erase(size_t pos, size_t cnt)
{
	size_t len = length();
	if ((len == 0) || (pos >= len)) return *this;
	if (cnt > (len - pos))
		cnt = len - pos;
	if (len - cnt == 0) {
		freemem(_data);
		_data = 0;
	}
	else if (cnt > 0)
	{
		size_t moveCnt = len - (pos + cnt);
		for (long i = 0; i < moveCnt; ++i)
		{
			_data[pos + i] = _data[pos + cnt + i];
		}
		len -= cnt;
		_data[len] = 0;
		_data = (BppChar*)reallocmem(_data, (len+1) * sizeof(BppChar)); 
	}
	return *this;

	//size_t len = length();
	//if (cnt > len) cnt = len;
	//if ((cnt > 0) && (pos < len))
	//{
	//	if (pos + cnt > len)
	//	{
	//		cnt = len-pos;
	//	}
	//	len -= cnt;
	//	if (len == 0)
	//	{
	//		freemem(_data);
	//		_data = 0;
	//	}
	//	else
	//	{
	//		BppChar* p1 = &_data[pos];
	//		BppChar* p2 = &_data[pos+cnt];
	//		size_t diff = len-pos;
	//		for (size_t i=0; i<=diff; ++i,++p1,++p2)
	//			*p1 = *p2;
	//		_data[len] = 0;
	//		_data = (BppChar*)reallocmem(_data,(len+1)*sizeof(BppChar));
	//	}
	//}
	//return *this;
}

/**
Function inserts <i>insertion</i> at position <i>pos</i> into current string.
@param insertion string to insert
@param pos insertion position
@return reference to current string
*/
BppString& BppString::insert(size_t pos, const BppString& insertion)
{
	size_t len = length();
	size_t ilen = insertion.length();
	if (ilen > 0)
	{
		if (pos == len) return ((*this)+=insertion);
		if (pos < len)
		{
			size_t newlen = len+ilen;
			_data = (BppChar*) reallocmem(_data,(newlen+1) * sizeof(BppChar));
			BppChar* p2 = &_data[newlen];
			BppChar* p1 = &_data[len];
			size_t diff = len-pos;
			for (size_t i=0; i<=diff;++i,--p1,--p2)
				*p2 = *p1;
			_tcsncpy(&_data[pos],insertion.c_str(),ilen);
		}
	}
	return *this;
}

/**
Function inserts <i>cnt</i> characters of value <i>insertion</i> into current string.
@param pos insertion position
@param insertion character value to insert
@param cnt number of characters to insert
@return reference to current string
*/
BppString& BppString::insert(size_t pos, BppChar insertion, size_t cnt /*= 1*/)
{
	size_t len = length();
	if (cnt > 0)
	{
		//if (pos == len) return ((*this)+=insertion);
		if (pos <= len)
		{
			size_t newlen = len+cnt;
			_data = (BppChar*) reallocmem(_data,(newlen+1) * sizeof(BppChar));
			if (_data == 0) errorExit(BP_ERR_NOMEM);
			BppChar* p2 = &_data[newlen]; 
			BppChar* p1 = &_data[len];
		 size_t ccnt = newlen-pos;
			for (size_t i=0; i<ccnt;++i,--p1,--p2)
			{
				*p2 = *p1;
			}
		 p1=_data+pos;
		 for (size_t j=0;j<cnt;++j,++p1) *p1 = insertion;
		 _data[newlen] = 0;
		}
	}
	return *this;
}

/**
Function splits current string into tokens using characters in <i>delim</i> for delimiters.
Tokens are appended to vector <i>dest</i>.
@param dest vector to store tokens into.
@param delim string containing delimiter characters.
@param storeEmptyTokens If <tt>true</tt>, empty tokens are stored in result too.
@param storeDelimiter If <tt>true</tt>, delimiter character between tokens are stored in result too.
@return Number of items appended to <i>dest</i>
@note <ul>
<li>If current string is empty, function does nothing.</li>
<li>Delimiters (if <i>storeDelimiter</i> set) are stored in result as character codes (i.e. of type <tt>BppInt</tt>).</li>
<li>If <i>storeEmptyTokens</i> is set and last character of string is a delimiter, empty
string is appended to result.</i>
</ul>
*/
size_t BppString::split(BppVector & dest, const BppString & delim, bool storeEmptyTokens, bool storeDelimiter) const
{
	size_t len = length();
	if (len == 0) return 0;
	size_t count = 0;
	size_t idxStart = 0;
	const BppChar* pdelimit = delim.c_str();
	size_t i = 0;
	while ((i = _tcscspn(_data + idxStart, pdelimit)) < len - idxStart)
	{
		if (i > 0)
		{
			dest.push_back(substr(idxStart, i));
			++count;
		}
		else if (storeEmptyTokens)
		{
			dest.push_back(new BppString());
			++count;
		}
		if (storeDelimiter)
		{
			BppChar c = _data[idxStart + i];
			dest.push_back((BppInt)c);
			++count;
		}
		//i++;
		idxStart += i + 1;
		if (idxStart >= len) break;
	}
	if (idxStart < len)
	{
		dest.push_back(new BppString(_data + idxStart));
		++count;
	}
	else if ((idxStart == len) && storeEmptyTokens)
	{
		dest.push_back(new BppString());
		++count;
	}
	return count;

}

/**
Function replaces all occurrences of <i>str</i> by <i>replacement</i>.
@param str substring to replace
@param replacement replacement string
@return reference to current string
*/
BppString& BppString::replace(const BppString& str, const BppString& replacement)
{
	size_t len = str.length();
	size_t rlen = replacement.length();
	if (len > 0)
	{
		int mode = rlen > len ? 2 : rlen < len ? 1 : 0;
		size_t pos = 0;
		while ((pos = find(str,pos)) != BPP_NNPOS)
	  {
		  switch (mode)
		  {
		  case 1: // rlen < len
			  erase((size_t)pos, len - rlen);
		  case 0: // rlen == len
			  if (rlen > 0) _tcsncpy(&_data[pos], replacement._data, rlen);
			  break;
		  default:
			  insert(pos,_T(' '),rlen - len);
			  _tcsncpy(&_data[pos], replacement._data, rlen);
		  }
		  pos += rlen;
	  }
	}
	return *this;
}

/**
Function replaces all occurrences of character <i>c</i> by <i>replacement</i>.
@param c character to replace
@param replacement replacement character
@return reference to current string
*/
BppString& BppString::replace(BppChar c, BppChar replacement) BPPNOEXCEPT
{
	for ( BppChar* p = _data; p && *p;++p)
	{
		if (*p == c ) *p = replacement;
	}
	return *this;
}

/**
Function creates new string from format string <i>fmt</i> and any
numper of variable values. It works similary as sprintf function of
C-library.
@param fmt format string
@param ... any number of values - must correspond to number of
variables in <i>fmt</i>.
@return formatted string
*/
BppString BppString::format(const BppChar* fmt,...)
{
	if (fmt==0) return BppString();
#if defined(_UNICODE) && defined(UNIX)
	static const BppString s = _T("%s");
	BppString fmtbuf = fmt;
	for (size_t p = fmtbuf.find(s); p != BPP_NNPOS; p = fmtbuf.find(s,++p)) {
		if (p == 0 || fmtbuf[p - 1] != _T('%'))
			fmtbuf.insert(p + 1, _T('l'));
	}
	fmt = fmtbuf.c_str();
#endif
	va_list args;
	va_start(args,fmt);
	int cnt = _vsctprintf(fmt,args);
	BppChar* buf = (BppChar*) allocmem((cnt+1) * sizeof(BppChar));
	va_end(args);
	va_start(args,fmt);
	_vstprintf_s(buf,cnt+1,fmt,args);
	va_end(args);
	BppString result;
	result._data = buf;
	return result;
}

/**
@return reference to current string
@note Behavior depends on current locale settings.
*/
BppString&  BppString::lower()
{
	if (_data) _tcslwr(_data);
	return *this;
}

/**
@return reference to current string
@note Behavior depends on current locale settings.
*/
BppString&  BppString::upper()
{
	if (_data) _tcsupr(_data);
	return *this;
}

/**
@param str character string to create string from
@return string containing lower case representation of <i>str</i>
@note Behavior depends on current locale settings.
*/
BppString  BppString::strlower(const BppChar * str)
{
	BppString s = str;
	s.lower();
	return s;
}

/**
@param str character string to create string from
@return string containing upper case representation of <i>str</i>
@note Behavior depends on current locale settings.
*/
BppString  BppString::strupper(const BppChar * str)
{
	BppString s = str;
	s.upper();
	return s;
}
