

#include "bppvec.h"
#include <stdlib.h>
#ifdef WINCE
#include <malloc.h>
#endif
#include "bppvalue.h"
#include "bppstrng.h"
#include "bppdict.h"
#include "bppufunc.h"

/// minimum possible vector capacity
static const unsigned MINCAPACITY = 2;

/**
Default constructor creates an empty vector if initial capacity #MINCAPACITY.
*/
BppVector::BppVector()
:_capacity(MINCAPACITY),_size(0)
{
	_data = (BppValue*) allocmem(_capacity*sizeof(BppValue));
	memset(_data,0,_capacity*sizeof(BppValue));
}

/**
Constructor creates a vector of initial capacity as specified by <i>capacity</i> argument.
@param capacity initial capacity of vector
@param init Initialization flag. If <tt>true</tt> vector of <i>capacity</i> elements is created
instead of empty vector.
*/
BppVector::BppVector(size_t capacity,bool init)
:_capacity(capacity>=MINCAPACITY ? capacity : MINCAPACITY),_size(0)
{
	_data = (BppValue*) allocmem(_capacity*sizeof(BppValue));
		memset(_data,0,_capacity*sizeof(BppValue));
		if (init) _size=capacity;
}

/**
Constructor makes a shallow copy of <i>cpy</i>, i.e. it creates a vector of same capacity
and size as <i>cpy</i> and assigns values from elements of <i>cpy</i> to elements of new vector.
@param cpy vector to copy from
*/
BppVector::BppVector(const BppVector& cpy)
:_capacity(cpy._capacity),_size(cpy._size)
{
	_data = (BppValue*)allocmem(_capacity * sizeof(BppValue));
	memset(_data, 0, _capacity * sizeof(BppValue));
	for (unsigned i = 0; i < _size; i++)
		_data[i] = cpy[i];
}
#ifdef CPP11
BppVector::BppVector(BppVector && cpy) BPPNOEXCEPT
:_capacity(cpy._capacity), _size(cpy._size),_data(cpy._data)
{
	cpy._data = 0;
	cpy._capacity = 0;
	cpy._size = 0;
}
#endif
/**
@return new copy of current vector
@note Caller is responsible for delete result if no longer needed.
*/
BppVector* BppVector::clone() const
{
	BppVector* result = new BppVector(_size,true);
	for (unsigned i=0;i<_size;++i)
	{
		BppValue& v = result->_data[i];
		const BppValue& vi = _data[i];
		BppVector* pv = vi;
		if (this == pv) {
			
			bool isweakref = !vi.isRefCounted();
			bool isbytecode = maskWeakRef(vi.getType()) == BVT_BYTECODE;
			v = BppValue(result,isbytecode,isweakref);
		}
		else v = BppValue(vi,true);
		v.setFlags(vi.getFlags());
	}
	return result;
}

/**
Destructor calls dispose method on all elements of vector and releases internal data field.
*/
BppVector::~BppVector()
{
	if (_data != 0)
	{
		for (unsigned i=0;i<_size;i++)
			_data[i].dispose();
		freemem(_data);
		_data = 0;
	}
}

/**
Assignment operator makes a shallow copy of <i>assignment</i>, i.e. it creates a vector of same capacity
and size as <i>assignment</i> and assigns values from elements of <i>assignment</i> to elements of new vector.
If vector is not empty before call, dispose is called on all old elements.
@param assignment vector to assign from. 
@see BppValue::dispose
*/
BppVector& BppVector::operator=(const BppVector& assignment)
{
	if (&assignment == this || assignment._data == _data)
		return *this;
	if (assignment._size < _size) erase(assignment._size);
	setCapacity(assignment._capacity);
	_size = assignment._size;
	for (unsigned i = 0; i < _size; ++i) _data[i] = assignment._data[i];
	return *this;
}
#ifdef CPP11
BppVector & BppVector::operator=(BppVector && assignment)
{
	if (&assignment == this || assignment._data == _data)
		return *this;
	for (unsigned j = 0; j<_size; j++)
		_data[j].dispose();
	freemem(_data);
	_data = assignment._data;
	_capacity = assignment._capacity;
	_size = assignment._size;
	assignment._data = 0;
	assignment._capacity = 0;
	assignment._size = 0;
	return *this;
}
#endif

/**
Function tries to set capacity of vector to specifies value.
@param capacity new capacity
@return capacity of vector after setting
@note Function never deletes any elements of vector, i.e. if vector has more elements
as value of <i>capacity</i> new capacity will be same as current size. 
*/
size_t BppVector::setCapacity(size_t capacity)
{
	if (capacity == 0) capacity = MINCAPACITY;
	if (capacity < _size)
	{
		capacity = _size;
	}
	void* p = reallocmem(_data, capacity * sizeof(BppValue));
	_data = (BppValue*)p;
	if (capacity > _capacity)
	{
		//size_t n = capacity - _capacity;
		 for (size_t i=_capacity;i<capacity;i++)
		    memset(&_data[i], 0, sizeof(BppValue));
	}
	_capacity = capacity;
	return _capacity;
}


/**
Function sets capacity to be same as current size.
@return new capacity
*/
size_t BppVector::shrink()
{
	return setCapacity(_size);
}

/**
Function  resizes vector to <i>size</i> elements.
If <i>size</i> is less than current size, elements are erased from end of vector.
If <i>size</i> is greater than current size, new elements of value <i>fill</i> are appended.
@param size New size (number of elements) of vector.
@param fill Value to use for new elements.
@return true on success, false in case of memory error. 
*/
bool BppVector::resize(size_t size, const BppValue & fill)
{
	if (_capacity < size) 
	   if (setCapacity(size) < size) return false;
	if (size > _size) insert(_size, fill, size - _size);
	else if (size < _size) erase(size);

	return true;
}


BppVector BppVector::operator+(const BppVector& app) const
{
	BppVector tmp(_size + app._size,true);
	unsigned i;
	for (i=0;i<_size;++i)
		tmp[i] = _data[i];
	for (i=0;i<app._size;++i)
		tmp[_size+i]= app._data[i];
	return tmp;
}

BppVector& BppVector::operator+=(const BppVector& app)
{
	if (&app == this || app._data == _data)
	{
		(*this) = (*this) + app;
	}
	else {
		size_t newsize = _size + app._size;
		if (_capacity < newsize)
			setCapacity(newsize);
		for (size_t i = 0; i < app._size; ++i)
			_data[_size + i] = app._data[i];
		_size += app._size;
	}
	return *this;
}

void BppVector::pop_back()
{
  if (_size == 0)
	  return;
  _data[--_size].dispose();
}

void BppVector::push_back(const BppValue& val)
{
	BppValue v = val;
	if (_size == _capacity)
		setCapacity(_size+1 + (_capacity >> 1));
	_data[_size++] = v;
}
#ifdef CPP11
void BppVector::push_back(BppValue&& val)
{
	if (_size == _capacity)
		setCapacity(_size + 1 + (_capacity >> 1));
	_data[_size++] = std::move(val);
}
#endif
void BppVector::push_front(const BppValue& val)
{
	insert(0,val);
}

void BppVector::pop_front()
{
	if (_size == 0) return;
	//BppValue tmp = _data[0];
	erase(0,1);
	// return tmp;
}

ValPtr BppVector::first()
{
	return ValPtr(this,(_size > 0)? 0 : -1);
}

ValPtr BppVector::last()
{
	return ValPtr(this,(int)_size -1);
}
  
void BppVector::clear()
{
	for (unsigned i=0; i<_size; i++)
		_data[i].dispose();
	_size = 0;
}

int BppVector::compare(const BppVector & cmp) const
{
	if (&cmp == this) return 0;
	//if (cmp._size != _size) return false;
	size_t s = cmp._size < _size ? cmp._size : _size;
	BppValue* p1 = _data;
	BppValue* p2 = cmp._data;
	for (size_t i = 0; i<s; ++i, ++p1, ++p2)
	{
		int r = p1->compare(*p2);
		if (r != 0) return r;
	}
	return cmp._size < _size ? 1 : cmp._size > _size ? -1 : 0;
}

BppVector& BppVector::erase(size_t pos, size_t cnt)
{
	if ((_size == 0) || (pos >= _size)) return *this;
	if (cnt > (_size-pos))
		cnt = _size-pos;
	if (cnt > 0)
	{
		size_t moveCnt = _size - (pos+cnt);
		for (size_t i = 0;i<moveCnt;++i)
		{
			if (i < cnt) 
				_data[pos+i].dispose();
			BppValue::rawcopy(_data[pos+i],_data[pos+cnt+i]);
		}
		BppValue v;
		while(cnt > 0)
		{
			BppValue::rawcopy(_data[--_size],v);
			--cnt;
		}
	}
	return *this;
}
// insert value into vector
BppVector& BppVector::insert(size_t pos, const BppValue& insertion, size_t cnt)
{
	if (pos > _size) pos = _size;
	if (cnt < 1) return *this;
	if (_size+cnt > _capacity)
	{
		setCapacity(_size+cnt + (_capacity >> 1));
	}

	BppValue insv(insertion,((BppVector*)insertion == this) ? true : false);
	
	if (pos < _size)
	{
		BppValue* psrc = _data+(_size-1);
		_size += cnt;
		BppValue* pdest = _data+(_size-1);
		size_t ccnt = _size-pos-1;
		size_t i;
		for (i = 0; i<ccnt; ++i,--pdest,--psrc)
		{
			BppValue::rawcopy(*pdest,*psrc);
		}
		pdest = _data+pos;
		memset(pdest,0,cnt*sizeof(BppValue));
		 for( i=0;i<cnt;++i,++pdest)
		 {
			*pdest = insv;
		 }
	}
	else
	{
		BppValue* pdest = _data+_size;
		_size += cnt;
		for(size_t i=0;i<cnt;++i,++pdest)
		{
			*pdest = insv;
		}
	}
	return *this;
}


/**
@param pos1 position of first element
@param pos2 position of second element
@param reference to vector
@note In case of invalid value of <i>pos1</i> or <i>pos2</i> method does nothing.
*/
BppVector& BppVector::swap(size_t pos1, size_t pos2) BPPNOEXCEPT
{
	if (pos1 != pos2 && pos1 < _size && pos2 < _size)
	{
		char p[sizeof(BppValue)];
		BppValue* pv = (BppValue*) p;
		 BppValue::rawcopy(*pv,_data[pos1]);
		BppValue::rawcopy(_data[pos1],_data[pos2]);
		BppValue::rawcopy(_data[pos2],*pv);
	}
	return *this;
}


/**
@param srcpos position of element to move
@param dstpos destination position
@param reference to vector
@note In case of invalid value of <i>srcpos</i> or <i>srcpos</i> method does nothing.
*/
BppVector& BppVector::move(size_t srcpos, size_t dstpos) BPPNOEXCEPT
{
	if (srcpos == dstpos || srcpos >= _size || dstpos >= _size) return *this;
	char p[sizeof(BppValue)];
	BppValue* pv = (BppValue*) p;
	BppValue::rawcopy(*pv,_data[srcpos]);
	if (srcpos < dstpos)  // shift left
	{
		for (size_t i = srcpos; i<dstpos;++i)
			BppValue::rawcopy(_data[i],_data[i+1]);
	}
	else // shift right
	{
		for (size_t i = srcpos; i>dstpos;--i)
			BppValue::rawcopy(_data[i],_data[i-1]);
	}
	BppValue::rawcopy(_data[dstpos],*pv);
	return *this;
}


/**
@param from Index position of start of subsequence.
@param cnt Maximum number of elements to copy into result.
@return Returns new <tt>vector</tt> containing shallow copy of specified part from current one.
*/
BppVector BppVector::slice(size_t from, size_t cnt) const
{
	size_t maxcnt = (from <= _size) ? _size - from : 0;
	if (cnt > maxcnt) cnt = maxcnt;
	if (cnt == 0) return BppVector();
	BppVector result(cnt,true);
	 for (size_t i=0;i<cnt;++i)
		result[i] = _data[from+i];
	return result;
}

bool BppVector::operator==(const BppVector& cmp) const
{
	if (&cmp == this) return true;
	if (cmp._size != _size) return false;
	BppValue* p1 = _data;
	BppValue* p2 = cmp._data;
	for (unsigned i=0;i<_size;++i,++p1,++p2)
	{
		if (*p1 != *p2) return false;
	}
	return true;
}

/// unequality operator
bool BppVector::operator!=(const BppVector& cmp) const
{
	return !(*this == cmp);
}
/// finds first index of <i>val</i> in vector beginning at <i>pos</i>;
size_t BppVector::find(const BppValue& val, size_t pos) const
{
	BppValue* p = _data + pos;
	for (size_t i=pos;i<_size;++i,++p)
	{
		if (*p == val) return i;
	}
	return BPP_NNPOS;
}

/// finds first index of <i>val</i> in vector beginning at <i>pos</i>;
size_t BppVector::rfind(const BppValue& val, size_t pos) const
{
	if (_size == 0) return BPP_NNPOS;
	if (pos >= _size)  pos = _size-1;
	BppValue* p = _data + pos;
	for (int i=pos;i>=0;--i,--p)
	{
		if (*p == val) return i;
		if (i == 0) break;
	}
	return BPP_NNPOS;
}

size_t BppVector::getStringIndex(const BppChar* str) const BPPNOEXCEPT
{
	if ((_data == 0) || (str == 0)) return BPP_NNPOS;
	for (size_t i=0; i<_size; i++)
	{
		BppString* s = _data[i];
		if (s != 0)
		{
			if (_tcscmp(s->c_str(),str) == 0)
				return i;
		}
	}
	return BPP_NNPOS;
}

size_t BppVector::getVarIndex(const BppDictionaryEntry* var) const
{
	if ((_data == 0) || (var == 0)) return BPP_NNPOS;
	for (size_t i=0; i<_size; i++)
	{
		BppDictionaryEntry* v = _data[i];
		if (v == var)
		{
				return i;
		}
	}
	return BPP_NNPOS;
}


ValPtr::~ValPtr() BPPNOEXCEPT
{
	_owner = 0;
	_current = -1;
}

ValPtr& ValPtr::operator=(const ValPtr& assignment) BPPNOEXCEPT
{
	if (&assignment != this)
	{
		_owner = assignment._owner;
		_current = assignment._current;
	}
	return *this;
}


ValPtr::operator BppValue*() BPPNOEXCEPT
{
	if ((_owner == 0) || (_current < 0)) return 0;
	if (_current >= (int) _owner->_size) return 0;
	return &_owner->_data[_current];
}

ValPtr& ValPtr::operator+=(int offs)
{
	if (_owner)
	{
		_current += offs;
		int diff = _current - ((int) _owner->_size) +1;
		while (diff > 0)
		{
			_owner->push_back(BppValue());
			diff--;
		}
	}
	return *this;
}

ValPtr& ValPtr::operator-=(int offs) BPPNOEXCEPT
{
	if (_owner)
	{
		_current -= offs;
	}
	return *this;
}

ValPtr& ValPtr::operator++()
{
	if (_owner != 0)
	{
		_current ++;
		int diff = _current - ((int) _owner->_size) +1;
		while (diff > 0)
		{
			_owner->push_back(BppValue());
			diff--;
		}
	}
	return *this;
}


ValPtr& ValPtr::operator--() BPPNOEXCEPT
{
	if (_owner)
		_current--;
	return *this;
}
#ifndef __OLDCPP__
const ValPtr ValPtr::operator++(int)
{
	ValPtr tmp = *this;
	++(*this);
	return tmp;
}

const ValPtr ValPtr::operator--(int) BPPNOEXCEPT
{
	ValPtr tmp = *this;
	if (_owner)
		_current--;
	return tmp;
}
#endif // ndef __OLDCPP__


bool ValPtr::operator==(const ValPtr& cmp) const BPPNOEXCEPT
{
	return ((_owner == cmp._owner) && (_current == cmp._current));
}

bool ValPtr::operator!=(const ValPtr& cmp) const BPPNOEXCEPT
{
	return ((_owner != cmp._owner) || (_current != cmp._current));
}

