#include <malloc.h>
#include <limits.h>
#include <stdlib.h>
#include "bppvm.h"
#include "bppprint.h"
#include "bppobjct.h"
#include "bppint.h"
#include "bppvec.h"
#include "bppclass.h"
#include "bppstrng.h"
#include "bppcomp.h"
#include "bppdbg.h"
#include "dictio.h"
#include "bppdfunc.h"
#include "bppufunc.h"
#ifndef NO_REGEX
#include "bppregex.h"
#endif
#include "bppstruc.h"
#include "bppmsg.h"

#ifdef WIN32
#include <Windows.h>
UINT BppVirtualMachine::_breakRequestMsg = RegisterWindowMessage(_T("BppBreakRequestMsg"));
UINT BppVirtualMachine::_dbgWaitMsg = RegisterWindowMessage(_T("BppDbgWaitMsg"));
UINT BppVirtualMachine::_lineChangedMsg = RegisterWindowMessage(_T("BppLineChangedMsg"));
#endif
#ifdef __MSDOS__
#ifndef NO_DYNLIB
	#include <usedll.h>

	typedef HDLL HMODULE;
#endif	
	typedef int BOOL;
#endif

#ifndef DOXY
struct TypeIdEntry
{
	const BppChar* key;
	BppValueType id;
};

// typeMapping is ordered by id values
static const TypeIdEntry _typeMapping[] = {
	{ _T("void"), BVT_VOID },
	{ _T("auto"), BVT_AUTO },
	{ _T("null"), BVT_NULL },
	{ _T("sbyte"), BVT_SBYTE },
	{ _T("bool"), BVT_BOOL },
	{ _T("char"), BVT_CHAR},
	{ _T("byte"), BVT_BYTE },
	{ _T("uchar"), BVT_UCHAR },
	{ _T("char8_t"), BVT_CHAR8 },
	{ _T("short"), BVT_SHORT },
	{ _T("ushort"), BVT_USHORT },
	{ _T("char16_t"), BVT_CHAR16 },
	{ _T("wchar_t"), BVT_WCHAR },
	{ _T("int"), BVT_INT },
	{ _T("intptr"), BVT_INTPTR },
	{ _T("uint"), BVT_UINT },
	{ _T("uintptr"), BVT_UINTPTR },
	{ _T("char32_t"), BVT_CHAR32 },
//	{ _T("tchar_t"), BVT_TCHAR },
	{ _T("long"), BVT_LONG },
	{ _T("ulong"), BVT_ULONG },
	{ _T("float"), BVT_FLOAT },
	{ _T("double"), BVT_DOUBLE },
	{ _T("FILE"), BVT_FILE },
	{ _T("function"), BVT_CODE },
	{ _T("object"), BVT_OBJECT },
	{ _T("Class"), BVT_CLASS },
	{ _T("string"), BVT_STRING },
	{ _T("buffer"), BVT_BUFFER },
	{ _T("charbuffer"), BVT_CHARBUFFER },
	{ _T("vector"), BVT_VECTOR },
	{ _T("function"), BVT_BYTECODE },
	{ _T("dictionary"), BVT_DICTIONARY }
};
static const size_t _typeMapSize = sizeof(_typeMapping) / sizeof(TypeIdEntry);

#endif

/**
Constructor initializes internal symbol tables for virtual machine, installs
default printer and registers predefined B++ functions for current instance.
@param printer Pointer to IBppPrinter derived object used for default printer.
If <i>printer</i> is <tt>NULL</tt> a dummy printer will be installed, i.e.
all output is discarded by default.
@note The printer installed here is used for debug and error output too if 
no special printers are specified therefore.
@see IBppPrinter, BppDummyPrinter, execute, compileFile, compileText
*/
BppVirtualMachine::BppVirtualMachine(IBppPrinter* printer)
:_genLineInfo(false),_genLocalVarInfo(false),_printer(printer),_interpreter(0),_ownsPrinter(printer == 0),_cmdargc(0),_cmdargv(0),
_classes(),_symbols(),_dbgWaitProc(0),_breakRequestProc(0),_lineChangedProc(0),_defaultInput(stdin),_currentLine(-1),_inDbg(false)
#ifdef WIN32
,_hwndDbg(0)
#endif
{
	_currentInput = &_defaultInput;
	if (!_printer)
		_printer = new BppDummyPrinter();
	_classes.addRef();
	_symbols.addRef();
	_modules.addRef();
	_currentSource.addRef();
	BppDefaultFunctions::registerFunctions(this);
#ifndef NO_REGEX
	Bpp_RegEx::createClass(*this);
#endif
	Bpp_CStruct::createClass(*this);
}

/**
Destructor releases all native modules loaded by virtual machine and clears
internal symbol tables.
*/
BppVirtualMachine::~BppVirtualMachine()
{
	_classes.clear();
	_symbols.clear();
#ifndef NO_DYNLIB
	for (BppDictionaryEntry* pentry= _modules.getFirst();pentry != 0;pentry = pentry->getNext())
		FreeLibrary((HMODULE)pentry->getValue().ptr());
#endif
	_modules.clear();
	if (_ownsPrinter)
		delete _printer;
	_printer = 0;
}

/**
@param name Typename, may be either a predefined type name from B++ or name of a class (if <i>includeClasses</i> is <tt>true</tt>).
@param includeClasses Flag specifies if <i>name</i> may be a class name too.
@param classType If set and <i>name</i> referres to a class, pointer to BppClass object is stored here.
@return Internal value type for type specified by <i>name</i>, if found, <tt>BVT_NULL</tt> otherwise.
@note If <i>includeClasses</i> is <tt>true</tt> and <i>name</i> is a class name, result is <tt>BVT_OBJECT</tt>, if
<i>classType</i> is NULL. If <i>classType</i> is set, result is <tt>BVT_STRONGOBJECT</tt> and pointer to BppClass object
is stored in <i>classType</i>.
*/
BppValueType BppVirtualMachine::findVTypeFromName(const BppString& name, bool includeClasses, BppClass** classType, BppClass* ownerClass)
{
	const BppString nm = name.c_str();
	if (includeClasses)
	{
		BppDictionaryEntry* pe = 0;
		if (ownerClass) {
			// Suche nach innerer Klasse von ownerClass
			const BppString ownerPath = ownerClass->getClassPath().c_str();
			const BppString key = BppString::format(_T("%s::%s"), ownerPath.c_str(), nm.c_str());
			pe = _classes.find(key);
			if (!pe && nm == *ownerClass->getClassName()) {
				// ownerClass selbst
				pe = _classes.find(ownerClass->getClassPath());
			}
			if (!pe) {
				// Suche nach einer anderen inneren Klasse in umgebenden Klassen von ownercClass
				for (BppClass* p = ownerClass->getOwnerClass(); p && !pe; p = p->getOwnerClass()) {
					const BppString clpath = p->getClassPath().c_str();
					const BppString k = BppString::format(_T("%s::%s"), clpath.c_str(), nm.c_str());
					pe = _classes.find(k);
				}
			}
		}
		if (!pe) pe = _classes.find(nm);
		if (pe)
		{
			if (classType)
			{
				*classType = (BppClass*)pe->getValue();
				return BVT_STRONGOBJECT;
			}
			return BVT_OBJECT;
		}
		else if (classType)
			classType = 0;
	}
	const BppChar* s = nm.c_str();
	for (size_t it=0;it<_typeMapSize;++it)
	{
		  const TypeIdEntry& entry = _typeMapping[it];
		if (_tcscmp(s,entry.key) == 0)
		{
			return _typeMapping[it].id;
		}
		if (entry.id >= _BVTREFMIN) break;
	}
	 return BVT_NULL;    
}

/**
@param vtype Type ID of any value type
@return Readable type name (type identifier useable in source code), if <i>vtype</i> specifies
public data type, otherwise <i>0</i>.
*/
const BppChar* BppVirtualMachine::getTypeNameFromVType(BppValueType vtype)
{
	vtype = maskWeakRef(vtype);
	for (size_t it=0;it<_typeMapSize;++it)
	{
		BppValueType id = _typeMapping[it].id;
		if (id == vtype)
		{
			return _typeMapping[it].key;
		}
		if (id >= vtype) return 0;
	}
	return 0;
}

bool BppVirtualMachine::getGenerateLineInfo() const
{
	return _genLineInfo;
}

void BppVirtualMachine::setGenerateLineInfo(bool genInfo)
{
	_genLineInfo = genInfo;
}

bool BppVirtualMachine::getGenerateLocalVarInfo() const
{
	return _genLocalVarInfo;
}

/**
If flag set, compiler generates additional information for names of local variables in functions for debugging purposes.
@param value of flag
*/
void BppVirtualMachine::setGenerateLocalVarInfo(bool genInfo)
{
	_genLocalVarInfo = genInfo;
}


/**
@return name of source for code instruction currently executed
@note Value is set by notifySourceLineChanged method if code contains line information.
@see getCurrentLine
*/
const BppString& BppVirtualMachine::getCurrentSource() const
{
	return _currentSource;
}

/**
@return line number of code instruction currently executed
@note Value is set by notifySourceLineChanged method if code contains line information.
*/
long BppVirtualMachine::getCurrentLine() const
{
	return _currentLine;
}

/**
Function releases all native modules loaded by virtual machine and clears
internal symbol tables. At last it re-registers predefined B++ functions for current instance.
@note This function should be used in order to clean up all precompiled stuff
in virtual machine without delete it, e.g. in order to compile and execute
a new program.
*/
void BppVirtualMachine::reinit()
{
	_classes.clear();
	_symbols.clear();
	_currentLine = -1;
	_currentSource.clear();
#ifndef NO_DYNLIB
	for (BppDictionaryEntry* pentry= _modules.getFirst();pentry != 0;pentry = pentry->getNext())
		FreeLibrary((HMODULE)pentry->getValue().ptr());
#endif
	_modules.clear();
	BppDefaultFunctions::registerFunctions(this);
#ifndef NO_REGEX
	Bpp_RegEx::createClass(*this);
#endif
	Bpp_CStruct::createClass(*this);
}

/**
Method tries to register a native function (written in C++) for use with B++ in current
virtual machine. Client code may use it to install user defined extension functions.
@param name Name of user defined function for use in B++. The name must be unique
in context of VM.
@param pfunc Function pointer. Please note that function must use STDCALL convention
on Win32 platforms.
@return <tt>true</tt> on success
@note Function fails if <i>name</i> is invalid, <i>pfunc</i> is NULL or
there is already a symbol defined for <i>name</i>.
*/
bool BppVirtualMachine::registerNativeFunction(const BppChar* name, BPPUSERFUNC pfunc)
{
	if ((!name) || (*name == _T('0')))
		return false;
	if (!pfunc)
		return false;
	BppDictionaryEntry& de = _symbols[name];
	if (de.getSymbolType() != ST_NONE) return false;
	de.setValue(pfunc);
	de.setSymbolType(ST_SFUNCTION);
	return true;
}

/**
Method tries to register a native class (written in C++) for use with B++ in current
virtual machine. Client code may use it to install user defined extension classes.
@param name Name of user defined function for use in B++. The name must be unique
in context of VM.
@param pClass Pointer to a class object. Class object must be created using
current virtual machine in its constructor.
@return <tt>true</tt> on success
@note Function fails if <i>name</i> is invalid, <i>pClass</i> is NULL or
there is already a symbol defined for <i>name</i>.
*/
bool BppVirtualMachine::registerNativeClass( const BppChar* name, BppClass* pClass )
{
	if ((!name) || (*name == _T('0')))
		return false;
	if (!pClass) return false;
	BppDictionaryEntry& de = _classes[name];
	if (de.getSymbolType() != ST_NONE) return false;
	pClass->setClassName(name);
	pClass->setIsNative(true);
	de.setValue(pClass);
	de.setSymbolType(ST_CLASS);
	return true;
}

/// name of destructor function
static const BppString dtorname = SYMN_dtor;
/**
Function is called from BppClass whenever an object instance is to be deleted.
It calls objects destructor (if any).
@param obj object to delete
*/
void BppVirtualMachine::notifyObjDelete(BppObject* obj)
{
	if (_interpreter)
	{
		BppClass* pclass = obj->getClass(); // should never return NULL
		if (pclass->findMemberPL(BVF_PRIVATE,dtorname) == 0) return; // return if no dtor defined
		ValPtr savedSP =_interpreter->_sp;
		_interpreter->_sp = _interpreter->_stack.last();
#ifndef NOCPPEXCEPTION  // catch all exceptions in destructor call
		try {
#endif
		++(_interpreter->_sp);  // push object to stack
		(*_interpreter->_sp) = BppValue(obj,true); // make reference to prevent from delete
		if (!_interpreter->send(0,dtorname.c_str()))
		{
			// TODO: send error message if dtor call fails
		}       

#ifndef NOCPPEXCEPTION
		} catch (...) {}
#endif
		// cleanup stack
		//while (_interpreter->_sp.current() > savedSP.current())
		//{
		//   _interpreter->_sp->dispose();
		//   --_interpreter->_sp;
		//}
		_interpreter->_sp = savedSP;
	}
}

/**
Function provides read/write access to variable for number of command line
arguments. It is typically used by an application during initializing
of VM.
@return writeable reference to variable for number of command line arguments
@see cmdargv
*/
int& BppVirtualMachine::cmdargc()
{
	return _cmdargc;
}

/**
Function provides read/write access to variable for array of command line
arguments. It is typically used by an application during initializing
of VM.
@return writeable reference to variable for array of command line arguments
@note <p>Client code should provide arguments array in order to make
argument related B++ functions useable.</p>
<p>Virtual machine does not copy nor delete arguments array, i.e. client
code is responsible to provide the array for entire life time of VM.</p>
@see cmdargc, Bpp_getargs, Bpp_getusrargs, Bpp_argcnt, Bpp_arg
*/
BppChar**& BppVirtualMachine::cmdargv()
{
	return _cmdargv;
}

/**
@return writeable reference to class table of VM
@note This function is intended for internal use only.
*/
BppDictionary& BppVirtualMachine::classes()
{
	return _classes;
}

/**
@return writeable reference to symbol table of VM
@note This function is intended for internal use only.
*/
BppDictionary& BppVirtualMachine::symbols()
{
	return _symbols;
}

/**
@return writeable reference to module table of VM
@note This function is intended for internal use only.
*/
BppDictionary& BppVirtualMachine::modules()
{
	return _modules;
}

/**
@return reference to default printer of VM
@note Using default printer allows formatted output operations
in implementation of native functions.
*/
IBppPrinter& BppVirtualMachine::printer()
{
	return *_printer;
}

/**
@return pointer to interpreter currently used by VM (or NULL if none)
@note This function is mainly for internal use. However, in some cases
access to interpreter may be useful in implementation of native
functions too, e.g. in order to get execution context or to emit
an error message.
@see BppInterpreter::getProcInfo, BppInterpreter::returnError
*/
BppInterpreter* BppVirtualMachine::interpreter() const
{
	return _interpreter;
}

/**

*/
void BppVirtualMachine::cloneSymbols(const BppVirtualMachine& other)
{
	_symbols.clear();
	_classes.clear();
	BppDictionaryEntry* pe;
#ifndef NO_DYNLIB	
	for (pe= _modules.getFirst();pe != 0;pe = pe->getNext())
		FreeLibrary((HMODULE)pe->getValue().ptr());
	_modules.clear();

	for (pe = other._modules.getFirst();pe != 0;pe = pe->getNext())
		loadNativeModule(pe->getKey()->c_str());
#endif
	for (pe = other._symbols.getFirst();pe != 0;pe = pe->getNext())
		(*_symbols[pe->getKey()]) = *pe;
	for (pe = other._classes.getFirst();pe != 0;pe = pe->getNext())
		(*_classes[pe->getKey()]) = *pe;
}

#ifndef DOXY
// Helper struct for compileXXX functions
// It ensures releasing of debugger and closing source file.
struct CompileHelper
{
	CompileHelper(FILE* fp, BppDebugger* dbg) : _file(fp),_dbg(dbg) {}
	~CompileHelper()
	{
		// Don't close _file because it is already done by BppFileStreamInput
		//if (_file) fclose(_file);
		if (_dbg) delete _dbg;
	}
	FILE* _file;
	BppDebugger* _dbg;
};
#endif

/**
@param filename Name of source file to compile (may contain directory path).
@param showinfo Flag for show names of classes and functions while compiling
@param decode Flag for debug output of generated bytecode
@param errPrinter Printer interface for output compiler (error) messages
@param dbgPrinter Printer interface for debug output
@return true on success
@note In order to create debug information in generated bytecode use #setGenerateLineInfo and #setGenerateLocalVarInfo.
*/
bool BppVirtualMachine::compileFile(const BppChar* filename, bool showinfo, bool decode, IBppPrinter* errPrinter, IBppPrinter* dbgPrinter)
{
	if (filename == 0) return false;
	FILE* fp = _tfopen(filename,_T("rt"));
	if (!fp) 
	{
		IBppPrinter* ppr = errPrinter ? errPrinter : _printer;
		(*ppr)(VM_ERR_OPEN_FILE_FMT,filename);
		return false;
	}
	BppCompiler comp(*this,errPrinter);
	comp.setShowInfo(showinfo);
	comp.setGenerateLineInfo(_genLineInfo);
	comp.setGenerateLocalVarInfo(_genLocalVarInfo);
	BppDebugger* dbg = 0;
	if (decode)
	{
		dbg = new BppDebugger(*this,dbgPrinter);
		comp.setDebugger(dbg);
	}
	CompileHelper hlp(fp,dbg);
	bool result = comp.compile(fp,filename);
	return result;
}

/**
@param src Source code to compile.
@param showinfo Flag for show names of classes and functions while compiling
@param decode Flag for debug output of generated bytecode
@param errPrinter Printer interface for output compiler (error) messages
@param dbgPrinter Printer interface for debug output
@return true on success
@note In order to create debug information in generated bytecode use #setGenerateLineInfo and #setGenerateLocalVarInfo.
*/
bool BppVirtualMachine::compileText(const BppChar* src, bool showinfo, bool decode, IBppPrinter* errPrinter, IBppPrinter* dbgPrinter)
{
	if (!src) return false;
	BppCompiler comp(*this,errPrinter);
	comp.setShowInfo(showinfo);
	comp.setGenerateLineInfo(_genLineInfo);
	comp.setGenerateLocalVarInfo(_genLocalVarInfo);
	BppDebugger* dbg = 0;
	if (decode)
	{
		dbg = new BppDebugger(*this,dbgPrinter);
		comp.setDebugger(dbg);
	}
	CompileHelper hlp(0,dbg);
	return comp.compile(src);
}

#ifndef DOXY
struct ExecHelper
{
	ExecHelper(BppInterpreter*& interpreter, BppDebugger* dbg) : _interpreter(interpreter),_dbg(dbg)
	{
		_interpreter->setDebugger(dbg);
	}
	~ExecHelper()
	{
		_interpreter->setDebugger(0);
		delete _interpreter;
		_interpreter = 0;
		if (_dbg) delete _dbg;
	}
	BppInterpreter*& _interpreter;
	BppDebugger* _dbg;
private:
	ExecHelper(const ExecHelper&);
	ExecHelper& operator=(const ExecHelper&);
};
#endif

/**
@param funcName Name of (entry point) function to execute
@param argc Number of arguments in <i>argv</i>
@param argv Array of argument strings
@param trace Flag for trace output while execution
@param errPrinter Printer for output of error messages. If <tt>NULL</tt>, default printer of VM is used instead.
@param dbgPrinter Printer used for trace output. If <tt>NULL</tt>, default printer of VM is used instead.
@return true on success.
*/
bool BppVirtualMachine::execute(const BppChar* funcName, int argc, const BppChar** argv, bool trace, IBppPrinter* errPrinter, IBppPrinter* dbgPrinter)
{
	if (!funcName) return false;
	_interpreter = new BppInterpreter(*this);
	if (_breakRequestProc != 0)
		_interpreter->setAllowBreakRequest(true);
	BppDebugger* dbg = 0;
	dbg = new BppDebugger(*this,dbgPrinter);
	dbg->active = trace;
	ExecHelper hlp(_interpreter,dbg);
	bool result = _interpreter->execute(funcName,argc,argv);
	_result = *(_interpreter->_sp);
	if (!result)
	{
		if (errPrinter != 0)
			(*errPrinter)(_T("%s"),_interpreter->errors().c_str());
		else
			(*_printer)(_T("%s"),_interpreter->errors().c_str());
	}
	return result;
}

/**
@return Return value of last call to #execute
*/
const BppValue& BppVirtualMachine::getResult() const
{
	return _result;
}

/**
@param filename Name of file (may contain directory path) to write compiled module into
@return true on success
*/
bool BppVirtualMachine::writeModule(const BppChar* filename)
{
	if (!filename) return false;
	BppDictionaryIO io(*this);
	return io.writeModule(filename);
}

/**
@param fp Handle of file to write compiled module into. File should be opened for binary write.
@return true on success
*/
bool BppVirtualMachine::writeModule(FILE* fp)
{
	if (!fp) return false;
	BppDictionaryIO io(*this);
	return io.writeModule(fp);
}

/**
@param filename Name of file (may contain directory path) to read module from
@return true on success
*/
bool BppVirtualMachine::readModule(const BppChar* filename)
{
	if (!filename) return false;
	BppDictionaryIO io(*this);
	return io.readModule(filename);
}

/**
@param fp Handle of file read module from. File should be opened for binary read.
@return true on success
*/
bool BppVirtualMachine::readModule(FILE* fp)
{
	if (!fp) return false;
	BppDictionaryIO io(*this);
	return io.readModule(fp);
}
#ifndef NO_DYNLIB
/**
@param dllname Name of native module (may contain path)
@return true on success
@note
<ul>
<li>If dllname contains an asterisk (*), it will be replaced by a build specific name suffix (see BPP_DLLWILD).</li>
<li>Function uses LoadLibrary for loading the module.</li>
<li>Native module must export a function <tt>RegisterNativeModule</tt> for initialization.
</ul>
*/
bool BppVirtualMachine::loadNativeModule(const BppChar* dllName)
{
	if (!dllName) return false;
	BppString _dllname = dllName;
	_dllname.replace(_T("*"),BPP_DLLWILD);
	if (_modules.find(dllName) != 0) return true;
#ifdef WINCE
	const BppChar* dllPath = searchPath(_dllname.c_str());
	if (dllPath == 0)
		dllPath = _dllname.c_str();
	HMODULE handle = LoadLibrary(dllPath);
#else
	HMODULE handle = LoadLibrary(_dllname.c_str());
#endif
	if (!handle) return false;
	BppRegisterNativeModuleFunc regFunc = (BppRegisterNativeModuleFunc) GetProcAddress(handle,_PT("RegisterNativeModule"));
	if (regFunc)
	{
	   _modules[dllName] = INTPTRCAST(handle);
		regFunc(this);
		return true;
	}
	FreeLibrary(handle);
	return false;
}
#endif // !NO_DYNLIB
#ifndef DOXY
struct DbgHelper
{
	DbgHelper(BppVirtualMachine* vm):_vm(vm) { _vm->setInDbg(true); }
	~DbgHelper() { _vm->setInDbg(false); }
	BppVirtualMachine* _vm;
};
#endif

/**
Method is called by interpreter before processing next bytecode instruction, if interpreter runs in step- or trace-mode.
It calls installed wait handler function, if any. Otherwise on Windows platforms it sends a debug wait message to debug window (if specified)
or simply returns <tt>false</tt>
@return Result of wait handler function or SendMessage call (in case of debug window) or <tt>false</tt> if none of both is installed.
@note Result of call sets (or resets) step mode for next instruction in interpreter.<br>
While method runs, _inDbg flag of vm is set.
@see setDbgWaitProc, setDbgWindow, setInDbg, isInDbg
*/
bool BppVirtualMachine::dbgWait() const
{
	DbgHelper dh((BppVirtualMachine*) this);
	if (_dbgWaitProc) return _dbgWaitProc((BppVirtualMachine*) this);
#ifdef WIN32
	if (_hwndDbg)
	{
		if (SendMessage(_hwndDbg,_dbgWaitMsg,(WPARAM)this,0) == _dbgWaitMsg)
			return true;
	}
#endif
	return false;
}

BppDbgWaitProc BppVirtualMachine::setDbgWaitProc(BppDbgWaitProc dbgWaitProc)
{
	BppDbgWaitProc oldProc = _dbgWaitProc;
	_dbgWaitProc = dbgWaitProc;
	return oldProc;
}



bool BppVirtualMachine::breakRequest() const
{
	DbgHelper dh((BppVirtualMachine*) this);
	if (_breakRequestProc) return _breakRequestProc((BppVirtualMachine*) this);
#ifdef WIN32
	if (_hwndDbg)
	{
		if (SendMessage(_hwndDbg,_breakRequestMsg,(WPARAM)this,0) == _breakRequestMsg)
			return true;
	}
#endif
	return false;
}

bool BppVirtualMachine::notifySourceLineChanged(const BppChar* sourceName, long line)
{
	DbgHelper dh((BppVirtualMachine*) this);
	_currentSource = sourceName;
	_currentLine = line;
	if (_lineChangedProc)
	{
		if (_lineChangedProc((BppVirtualMachine*)this,_currentSource.a_str(),line))
			return true;
	}
#ifdef WIN32
	else if (_hwndDbg)
	{
		if (SendMessage(_hwndDbg,_lineChangedMsg,(WPARAM)this,(LPARAM)line) == _lineChangedMsg)
			return true;
	}
#endif
	return _interpreter->_step;
}

BppBreakRequestProc BppVirtualMachine::setBreakRequestProc( BppBreakRequestProc breakRequestProc )
{
	BppBreakRequestProc oldProc = _dbgWaitProc;
	_breakRequestProc = breakRequestProc;
	if (_interpreter)
		_interpreter->setAllowBreakRequest(_breakRequestProc != 0);
	return oldProc;
}

BppLineChangedProc BppVirtualMachine::setLineChangedProc(BppLineChangedProc lineChangedProc)
{
	BppLineChangedProc oldProc = _lineChangedProc;
	_lineChangedProc = lineChangedProc;
	return oldProc;
}

/**
@return Input object currently set for VM.
@see setCurrentInput
*/
IBppInput* BppVirtualMachine::getCurrentInput() const
{
	return _currentInput;
}

/**
@param currentInput Pointer to Object of IBppInput type, used for character input.
If it is <tt>NULL</tt>, default input object (class BppFileStreamInput) for <tt>stdin</tt> will be used.
@return Input object in use before
@see getCurrentInput
*/
IBppInput* BppVirtualMachine::setCurrentInput( IBppInput* currentInput )
{
	 IBppInput* oldInput = _currentInput;
	 if (currentInput)
		 _currentInput = currentInput;
	 else
		 _currentInput = &_defaultInput;
	 return oldInput;
}
#ifdef WIN32
void BppVirtualMachine::setDbgWindow( HWND hwndDbg )
{
	_hwndDbg = hwndDbg;
	if (_interpreter)
		_interpreter->setAllowBreakRequest(_hwndDbg != 0);
}

UINT BppVirtualMachine::getBreakRequestMsg()
{
	return _breakRequestMsg;
}

UINT BppVirtualMachine::getDbgWaitMsg()
{
	return _dbgWaitMsg;
}

UINT BppVirtualMachine::getLineChangedMsg()
{
	return _lineChangedMsg;
}

#endif
void BppVirtualMachine::cancel()
{
	if (_interpreter)
		_interpreter->cancel();
}

bool BppVirtualMachine::invoke(BppValue& result, const BppChar* name, BppVector* args)
{
	if (!_interpreter) return false;
	return _interpreter->invoke(result,name,args);
}

bool BppVirtualMachine::invokeMember(BppValue& result, BppObject* pobj, const BppChar* name, BppVector* args)
{
  if (!_interpreter) return false;
  return _interpreter->invokeMember(result,pobj,name,args);
}

bool BppVirtualMachine::invokeStatic(BppValue& result, BppClass* pclass, const BppChar* name, BppVector* args)
{
  if (!_interpreter) return false;
  return _interpreter->invokeStatic(result,pclass,name,args);
}

bool BppVirtualMachine::isInDbg() const
{
	return _inDbg;
}

void BppVirtualMachine::setInDbg(bool inDbg)
{
	_inDbg = inDbg;
}

void BppVirtualMachine::addPredefinedSymbol(const BppChar * name, const BppChar * val)
{
	if (!name) return;
	if (val) _predefined[name] = BppString(val);
	else  _predefined[name] = BppValue();
}

bool BppVirtualMachine::removePredefinedSymbol(const BppChar* name)
{
	if (!name) return false;
	return _predefined.erase(name);
}

// ==============================================================================
//         C-API implementation
// ==============================================================================

#ifndef NO_CAPI
void* BPPAPI Bpp_CreateVM(void* pPrinter)
{
	return new BppVirtualMachine((IBppPrinter*) pPrinter);
}

int BPPAPI Bpp_ReleaseVM(void* pVM)
{
	BppVirtualMachine* p = (BppVirtualMachine*) pVM;
	if (p == 0)
		return 0;
	delete p;
	return 1;
}

int BPPAPI Bpp_GetGenerateLineInfo(void* pVM)
{
	BppVirtualMachine* p = (BppVirtualMachine*) pVM;
	if (!p)
		return 0;
	return (p->getGenerateLineInfo()) ? 1 : 0;
}

 /// Set value of line info generation flag.
 void BPPAPI Bpp_SetGenerateLineInfo(void* pVM, int bGenInfo)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*) pVM;
	if (!p)
		return;
	p->setGenerateLineInfo((0 != bGenInfo));
}

 int BPPAPI Bpp_GetGenerateLocalVarInfo(void* pVM)
{
	BppVirtualMachine* p = (BppVirtualMachine*) pVM;
	if (!p)
		return 0;
	return (p->getGenerateLocalVarInfo()) ? 1 : 0;
}

 void BPPAPI Bpp_SetGenerateLocalVarInfo(void* pVM, int bGenInfo)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*) pVM;
	if (!p)
		return;
	p->setGenerateLocalVarInfo((0 != bGenInfo));
}
 void BPPAPI Bpp_AddPredefinedSymbolA(void * pVM, char * pcszName, char * pcszVal)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*)pVM;
	 if (!p)
		 return;
	 BppString n = pcszName;
	 if (!pcszVal) {
		 p->addPredefinedSymbol(n.c_str());
		 return;
	 }
	 BppString v = pcszVal;
	 p->addPredefinedSymbol(n.c_str(),v.c_str());
 }

 int BPPAPI Bpp_RemovePredefinedSymbolA(void * pVM, char * pcszName)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*)pVM;
	 if (!p)
		 return 0;
	 BppString n = pcszName;
	 return p->removePredefinedSymbol(n.c_str());
 }

 void BPPAPI Bpp_ClearPredefinedSymbols(void* pVM)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*)pVM;
	 if (!p)
		 return;
	 p->predefines().clear();
 }

#ifndef __OLDCPP__
 void BPPAPI Bpp_AddPredefinedSymbolW(void * pVM, wchar_t * pcszName, wchar_t * pcszVal)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*)pVM;
	 if (!p)
		 return;
	 BppString n = pcszName;
	 if (!pcszVal) {
		 p->addPredefinedSymbol(n.c_str());
		 return;
	 }
	 BppString v = pcszVal;
	 p->addPredefinedSymbol(n.c_str(), v.c_str());
 }

 int BPPAPI Bpp_RemovePredefinedSymbolW(void * pVM, wchar_t * pcszName)
 {
	 BppVirtualMachine* p = (BppVirtualMachine*)pVM;
	 if (!p)
		 return 0;
	 BppString n = pcszName;
	 return p->removePredefinedSymbol(n.c_str());
 }
#endif

int BPPAPI Bpp_CompileFileA(void* pVM,
									 char* pcszFileName, 
									 int bShowInfo, 
									 int bDecode, 
									 void* pErrPrinter, 
									 void* pDbgPrinter)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString fName = (const char*) pcszFileName;
	return (pVirt->compileFile(fName.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}

#ifndef __OLDCPP__
int BPPAPI Bpp_CompileFileW(void* pVM,
									 wchar_t* pcszFileName, 
									 int bShowInfo, 
									 int bDecode, 
									 void* pErrPrinter, 
									 void* pDbgPrinter)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString fName = (const wchar_t*) pcszFileName;
	return (pVirt->compileFile(fName.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#endif

int BPPAPI Bpp_CompileTextA(void* pVM,
									 char* pcszSrc, 
									 int bShowInfo, 
									 int bDecode, 
									 void* pErrPrinter, 
									 void* pDbgPrinter)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString src = (const char*) pcszSrc;
	return (pVirt->compileText(src.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#ifndef __OLDCPP__
int BPPAPI Bpp_CompileTextW(void* pVM,
									 wchar_t* pcszSrc, 
									 int bShowInfo, 
									 int bDecode, 
									 void* pErrPrinter, 
									 void* pDbgPrinter)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString src = (const wchar_t*)pcszSrc;
	return (pVirt->compileText(src.c_str(),bShowInfo != 0,bDecode != 0,(IBppPrinter*) pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#endif
struct BppCharPtrArray
{
	BppChar** arrayPtr;
	int cnt;
	BppCharPtrArray(int argc, char** argv)
		:arrayPtr(0),cnt(argc)
	{
		if (argc > 0)
		{
			arrayPtr = (BppChar**) allocmem(argc * sizeof(BppChar*));
			for (int i=0;i<argc;i++)
			{
#ifndef _UNICODE
				arrayPtr[i] = strdup(argv[i]);
#else
				BppString s = (const char*)argv[i];
				arrayPtr[i] = wcsdup(s.w_str());
#endif
			}
		}
	}

#ifndef __OLDCPP__
	BppCharPtrArray(int argc, wchar_t** argv)
		:arrayPtr(0),cnt(argc)
	{
		if (argc > 0)
		{
			arrayPtr = (BppChar**) allocmem(argc * sizeof(BppChar*));
			for (int i=0;i<argc;i++)
			{
#ifdef _UNICODE
				arrayPtr[i] = wcsdup(argv[i]);
#else
				BppString s = (const wchar_t*)argv[i];
				arrayPtr[i] = strdup(s.a_str());
#endif
			}
		}
	}
#endif // NO_TCHAR_H

	~BppCharPtrArray()
	{
		if (arrayPtr)
		{
			for (int i=0;i<cnt;i++)
				freemem(arrayPtr[i]);
			freemem(arrayPtr);
		}
	}

	operator const BppChar**() const
	{
		return (const BppChar**) arrayPtr;
	}
};

int BPPAPI Bpp_ExecuteA(void* pVM,
								char* pcszFuncName, 
								int iArgc, 
								char** ppcszArgv,
								int bTrace,
								void* pErrPrinter, 
								void* pDbgPrinter)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString funcname = (const char*) pcszFuncName;
	BppCharPtrArray args(iArgc,ppcszArgv);
	return (pVirt->execute(funcname.c_str(),iArgc,args,bTrace != 0,(IBppPrinter*)pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}

#ifndef __OLDCPP__
int BPPAPI Bpp_ExecuteW(void* pVM,
								wchar_t* pcszFuncName, 
								int iArgc, 
								wchar_t** ppcszArgv,
								int bTrace,
								void* pErrPrinter, 
								void* pDbgPrinter)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString funcname = (const wchar_t*) pcszFuncName;
	BppCharPtrArray args(iArgc,ppcszArgv);
	return (pVirt->execute(funcname.c_str(),iArgc,args,bTrace != 0,(IBppPrinter*)pErrPrinter,(IBppPrinter*)pDbgPrinter)) ? 1 : 0;
}
#endif

int BPPAPI Bpp_WriteModuleA(void* pVM,
									 char* pcszFileName)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString fName = (const char*) pcszFileName;
	return (pVirt->writeModule(fName.c_str())) ? 1 : 0;
}

#ifndef __OLDCPP__
int BPPAPI Bpp_WriteModuleW(void* pVM,
									 wchar_t* pcszFileName)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString fName = (const wchar_t*) pcszFileName;
	return (pVirt->writeModule(fName.c_str())) ? 1 : 0;
}
#endif

int BPPAPI Bpp_ReadModuleA(void* pVM,
									char* pcszFileName)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString fName = (const char*) pcszFileName;
	return (pVirt->readModule(fName.c_str())) ? 1 : 0;
}

#ifndef __OLDCPP__
int BPPAPI Bpp_ReadModuleW(void* pVM,
									wchar_t* pcszFileName)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	BppString fName = (const wchar_t*) pcszFileName;
	return (pVirt->readModule(fName.c_str())) ? 1 : 0;
}
#endif

char* BPPAPI Bpp_GetCurrentSourceA(void* pVM)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return (char*) pVirt->getCurrentSource().a_str();
}

#ifndef __OLDCPP__
wchar_t* BPPAPI Bpp_GetCurrentSourceW(void* pVM)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return (wchar_t*) pVirt->getCurrentSource().w_str();
}
#endif

long BPPAPI Bpp_GetCurrentLine(void* pVM)
{
	if (!pVM) return -1;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return pVirt->getCurrentLine();
}

 int BPPAPI Bpp_GetIntResult(void* pVM)
 {
	 if (!pVM) return INT_MIN;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return (int) (BppInt) pVirt->getResult();
}

 void BPPAPI Bpp_Reinit(void* pVM)
 {
	  if (!pVM) return;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	pVirt->reinit();
}

 void* BPPAPI Bpp_CreateConsolePrinter()
 {
	 return new BppFileStreamPrinter(stdout);
 }

 void* BPPAPI Bpp_CreateCallbackPrinter(BppPrinterCallback pCallback)
 {
	 if (!pCallback) return 0;
	 return new BppCallbackPrinter(pCallback);
 }

int BPPAPI Bpp_ReleasePrinter(void* pPrinter)
{
	if (!pPrinter) return 0;
	delete (IBppPrinter*) pPrinter;
	return 1;
}

BppDbgWaitProc BPPAPI Bpp_SetDbgWaitProc(void* pVM, BppDbgWaitProc dbgWaitProc)
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return pVirt->setDbgWaitProc(dbgWaitProc);
}

BppBreakRequestProc BPPAPI Bpp_SetBreakRequestProc( void* pVM, BppBreakRequestProc breakRequestProc )
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return pVirt->setBreakRequestProc(breakRequestProc);
}

BppLineChangedProc BPPAPI Bpp_SetLineChangedProc( void* pVM, BppLineChangedProc lineChangedProc )
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return pVirt->setLineChangedProc(lineChangedProc);
}

void* BPPAPI Bpp_CreateCallbackInput( BppCharInputCallback pcCallback, BppStringInputCallback psCallback )
{
	 return new BppCallbackInput(pcCallback,psCallback);
}

int BPPAPI Bpp_ReleaseInput( void* pInput )
{
	 if (!pInput) return 0;
	 delete (IBppInput*) pInput;
	 return 1;
}

void* BPPAPI Bpp_SetCurrentInput( void* pVM, void* pInput )
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return pVirt->setCurrentInput((IBppInput*) pInput);
}

void* BPPAPI Bpp_GetCurrentInput( void* pVM )
{
	if (!pVM) return 0;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	return pVirt->getCurrentInput();
}

void BPPAPI Bpp_Cancel( void* pVM )
{
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	if (pVirt)
		pVirt->cancel();
}

#ifdef WIN32
void* BPPAPI Bpp_CreateWinMsgPrinter( HWND hwndDest )
{
	return new BppWinMsgPrinter(hwndDest);
}

UINT BPPAPI Bpp_GetPrintStringMsg()
{
	return BppWinMsgPrinter::getPrintStringMsg();
}

void* BPPAPI Bpp_CreateWinMsgInput( HWND hwndDest )
{
	return new BppWinMsgInput(hwndDest);
}

UINT BPPAPI Bpp_GetStringInputMsg()
{
	return BppWinMsgInput::getStringInputMsg();
}

UINT BPPAPI Bpp_GetCharInputMsg()
{
	return BppWinMsgInput::getCharInputMsg();
}

void BPPAPI Bpp_SetWinInputStringA( void* pMsgWinInput, char* pcszInputStr )
{
	if (!pMsgWinInput) return;
	BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
	p->setString(pcszInputStr);
}

void BPPAPI Bpp_SetWinInputStringW( void* pMsgWinInput, wchar_t* pcszInputStr )
{
	if (!pMsgWinInput) return;
	BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
	p->setString(pcszInputStr);
}

void BPPAPI Bpp_SetWinInputCharA( void* pMsgWinInput, char cInputChr )
{
	if (!pMsgWinInput) return;
	BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
	p->setChar(cInputChr);
}

void BPPAPI Bpp_SetWinInputCharW( void* pMsgWinInput, wchar_t cInputChr )
{
	if (!pMsgWinInput) return;
	BppWinMsgInput* p = (BppWinMsgInput*) pMsgWinInput;
	p->setChar(cInputChr);
}

void BPPAPI Bpp_SetDbgWindow( void* pVM, HWND hwndDbg )
{
	if (!pVM) return ;
	BppVirtualMachine* pVirt = (BppVirtualMachine*) pVM;
	pVirt->setDbgWindow(hwndDbg);
}

UINT BPPAPI Bpp_GetBreakRequestMsg()
{
	return BppVirtualMachine::getBreakRequestMsg();
}

UINT BPPAPI Bpp_GetDbgWaitMsg()
{
	return BppVirtualMachine::getDbgWaitMsg();
}

UINT BPPAPI Bpp_GetLineChangedMsg()
{
	return BppVirtualMachine::getLineChangedMsg();
}

char* BPPAPI Bpp_GetEnvA(char* pcszName)
{
	BppString nbuf = pcszName;
	static BppString buf = _tgetenv(nbuf.c_str());
	return (char*) buf.a_str();
}

void BPPAPI Bpp_SetEnvA(char* pcszName,char* pcszValue)
{
	BppString buf = pcszName;
	buf += _T('=');
	if (pcszValue) buf += pcszValue;
	_tputenv(buf.c_str());
}
#endif // WIN32

#ifndef __OLDCPP__
wchar_t* BPPAPI Bpp_GetEnvW(wchar_t* pcszName)
{
	BppString nbuf = pcszName;
	static BppString buf = _tgetenv(nbuf.c_str());
	return (wchar_t*) buf.w_str();
}

void BPPAPI Bpp_SetEnvW(wchar_t* pcszName,wchar_t* pcszValue)
{
	BppString buf = pcszName;
	buf += _T('=');
	if (pcszValue) buf += pcszValue;
	_tputenv((BppChar*)buf.c_str());
}
#endif

#endif // !NO_CAPI
