#include "dictio.h"
#include "bppprint.h"
#include "bppbuf.h"
#include "bppclass.h"
#include "bppvec.h"
#include "bppobjct.h"
#include "bppvalue.h"
#include "bppvm.h"
#include "bppconst.h"
#include "bppdfunc.h"
#include "bppmsg.h"
#include "bppufunc.h"

#define BPPETX _T("\x03")
#define BPPETXC '\x03'
static const BppChar* EmptyTag = BPPETX;
static const BppChar EmptyChar = BPPETXC;

struct ObjListEntry
{
	ObjListEntry() :ID(0), pval(0), next(0) {}
	~ObjListEntry()
	{
		if (next != 0)
		{
			delete next;
			next = 0;
		}
	}
	void* ID;
	BppValue val;
	BppValue* pval;
	ObjListEntry* next;
};

/**
Adds entry (ID and value) to _objlist, if not already contained.
@param ID Identificator of object (simply its address)
@param val Value containing the object
@return trus, if object was added to list.
@note If entry is added, val is copied, i.e. reference counter of object is incremented, if object is not e weak reference.
*/
bool BppDictionaryIO::addOLEntry(void* ID, const BppValue& val)
{
	ObjListEntry* ep = _objList;
	while (ep)
	{
		if (ep->ID == ID)
			return false;
		ep = ep->next;
	}
	ep = new ObjListEntry();
	ep->ID = ID;
	ep->val = val;
	ep->next = _objList;
	_objList = ep;
	return true;
}

void BppDictionaryIO::addORefEntry(void* ID, BppValue* pval)
{
	ObjListEntry* ep = new ObjListEntry();
	ep->ID = ID;
	ep->pval = pval;
	ep->next = _objRefList;
	_objRefList = ep;
}


void BppDictionaryIO::clearObjectLists()
{
	if (_objList != 0)
		delete _objList;
	if (_objRefList != 0)
		delete _objRefList;
	//if (_varList != 0)
	//	delete _varList;
	_objList = _objRefList = 0;
	//_varList = 0;
}

BppDictionaryIO::BppDictionaryIO(BppVirtualMachine& vm)
	:_classes(vm.classes()), _symbols(vm.symbols()), _print(vm.printer()), _vm(vm), _objList(0), _objRefList(0)/*, _varList(0)*/
{
	_classes.addRef();
	_symbols.addRef();
}

BppDictionaryIO::~BppDictionaryIO()
{
	clearObjectLists();
	_classes.release();
	_symbols.release();
}

bool BppDictionaryIO::writeModule(const BppString& filename)
{

	FILE* fp = _tfopen(filename.c_str(), _T("wb"));
	if (!fp)
	{
		_print(DIO_OPEN_WRITE_FAILED_FMT, filename.c_str());
		return false;
	}

	bool result = writeModule(fp);
	fclose(fp);
	return result;
}

bool BppDictionaryIO::writeModule(FILE* fp)
{
	if (!fp)
	{
		return false;
	}
	bool result = writeDictionary(fp, _vm.modules());
	if (result) {
		result = writeDictionary(fp, _classes, true);
	}
	if (result) {
		result = writeDictionary(fp, _symbols, true);
	}
	clearObjectLists();
	return result;
}

bool BppDictionaryIO::writeSingleValue(FILE* fp, BppValue& value)
{
	bool ok = (fp != 0);
	clearObjectLists();
	if (ok)
	{
		ok = writeValue(fp, value);
		clearObjectLists();
	}
	return ok;
}


// writes
// //- string: name of owner class (empty if none)
//   - byte: sorted flag (1 = sorted, 0= unsorted)
//   - all entries
//   - empty string as end marker
bool BppDictionaryIO::writeDictionary(FILE* fp, BppDictionary& dict, bool ignoreNativeFuncs)
{
	BppByte sorted = dict._sorted ? 1 : 0;
	bool ok = writeBytes(fp, &sorted, sizeof(BppByte));
	for (BppDictionaryEntry* pe = dict.getFirst(); ok && (pe != 0); pe = pe->getNext())
	{
		ok = writeDictEntry(fp, *pe, ignoreNativeFuncs);
	}
	if (ok) ok = writePtr(fp, 0); /* write null pointer as end sign */
	return ok;
}

// writes
// - string: key
// - BppSymbolType: symbol type
// - value
bool BppDictionaryIO::writeDictEntry(FILE* fp, BppDictionaryEntry& entry, bool ignoreNativeFuncs)
{
	static const BppChar* _internalGlobVars[] = { _T("stdin"), _T("stdout"),_T("stderr"),
		_T("__OSTYPE__"), _T("__UNICODE__"),_T("__PTRSIZE__") };
	static const size_t vcnt = sizeof(_internalGlobVars) / sizeof(BppChar*);
	BppSymbolType st = entry.getSymbolType();
	const BppValue& v = entry.getValue();
	const BppChar* key = entry.getKey()->c_str();
	BppValueType vtype = v.getType();

	bool isignore = false;
	if (ignoreNativeFuncs)
	{
		isignore = (vtype == BVT_CODE);
		if ((st == ST_SDATA) && (vtype == BVT_FILE))
		{
			for (size_t i = 0; !isignore && i < vcnt; ++i) {
				if (_tcscmp(_internalGlobVars[i], key) == 0)
					isignore = true;
			}
		}
	}
	// ignore native classes not refrenced in code (i.e. refcnt < 2)
	BppClass* pclass = v;
	if ((pclass != 0) && pclass->getIsNative()) {
		isignore = (pclass->getRefCnt() < 2);
	}
	// ignored classes are completely skipped
	if (isignore && pclass) return true;
	if (!writePtr(fp, INTPTRCAST(&entry)))
		return false; /* ID */
	// write key
	bool result = true;
	if (isignore) {
		BppString s(EmptyTag);
		s += key;
		result = writeString(fp, s.c_str());
	}
	else
		result = writeString(fp, key);
	// write symbol type
	if (result) result = writeBytes(fp, &st, 1);
	// write value
	if (result && !isignore) result = writeValue(fp, (BppValue&)v);
	if (!result)
		_print(DIO_WRITE_ENTRY_FAILED_FMT, key);
	return result;
}

// writes:
// - BppValueType: value type
// switch type
//   BVT_CLASS(REF):
//    - string: classname
//    if !writeref && not already written
//      - string: baseclass name (or "")
//      - writeDictionary(dataMembers)
//      - writeDictionary(funcMembers)
//      - BppInt: datasize
//    else
//      - string: "$" (end mark)
//   BVT_OBJECT(REF):
//    - pointer: to obj, used as ID
//    if not already written:
//      - string: classname
//      - write all member data entries (writeValue(val,writeClassRef=true))
//    else
//      - string: "$" (end mark)
//   BVT_BUFFER(REF):
//    - pointer: to buf, used as ID
//    if not already written:
//      - BppUInt: size (number of bytes)
//      - byte[size]: data
//    else
//      - BppInt: -1 (end mark)
//   BVT_CHARBUFFER(REF):
//    - pointer: to buf, used as ID
//    if not already written:
//      - BppUInt: size (number of characters)
//      - byte: chsize (size of character)
//      - byte[size * chsize]: data
//    else
//      - BppInt: -1 (end mark)
//   BVT_STRING(REF):
//    - pointer: to string, used as ID
//    if not already written:
//      - string: string content
//    else
//      - byte: 0xFF (end mark)
//   BVT_VECTOR(REF), BVT_BYTECODE(REF), BVT_LOCALVARINFO
//    - pointer: to vec, used as ID
//    if not already written:
//      - BppUInt: number of elements
//      - write all elements (writeValue(val,writeClassRef=true))
//    else
//      - BppInt: -1 (end mark)
//   BVT_CODE:
//      - string: name of native function
//   BVT_FILE:
//      - string: name of default file symbol
//   BVT_VAR:
//      - pointer: to dictionary, used as ID
//   BVT_DICTIONARY(REF):
//      - pointer: to dict, used as ID
//      if not already written:
//         - writeDictionary(...)
//      else:
//         - byte: 0xFF as end marker
//   BVT_(all integral types):
//      - BppInt value of type specific size (1 .. 8 bytes)
//   BVT_FLOAT:
//      - float: float value
//   BVT_DOUBLE:
//      - double: double value
bool BppDictionaryIO::writeValue(FILE* fp, BppValue& value, bool writeClassref)
{
	BppValueType vtype = value.getType();
	//BppValueType mvtype = maskWeakRef(vtype);
	unsigned char vflags = value.getFlags();
	writeBytes(fp, &vtype, 1);						// value type 1 Byte
	writeBytes(fp, &vflags, 1);						// value flags 1 Byte
	bool ok = true;
	switch (vtype)
	{
	case BVT_CLASS:
	case BVT_CLASSREF:
	{
		BppClass* pclass = value;
		if (pclass == 0) 
			return false;
		ok = writeString(fp, pclass->getClassPath().c_str());    // full class path (C-string)
		if ((!writeClassref) && ok && addOLEntry(pclass, value))
		{
			BppClass* pbase = pclass->getBaseClass();
			if (pbase == 0)
				ok = writeString(fp, _T(""));                   // empty string, if no baseclass exists
			else
				ok = writeString(fp, pbase->getClassPath().c_str()); // full class path of baseclass
			BppClass* powner = pclass->getOwnerClass(); 
			if (powner == 0)
				ok = writeString(fp, _T(""));                  // empty string if class is not owned by another class
			else
				ok = writeString(fp, powner->getClassPath().c_str()); // full class path of owner class

			if (ok) ok = writeDictionary(fp, pclass->dataMembers(), true); // dictionary of data members
			if (ok) ok = writeDictionary(fp, pclass->funcMembers(), true); //  dictionary of function members
			BppInt size = pclass->getDataSize();
			if (ok) ok = writeBytes(fp, &size, sizeof(BppInt)); // number of non-static data members
		}
		else
			ok = writeString(fp, EmptyTag); // Write empty tag behind classname, if class should is ignored or already written.
	}
	break;
	case BVT_OBJECT:
	case BVT_OBJECTREF:
	{

		BppObject* pobj = value;
		// use current address of object as ID
		ok = writePtr(fp, INTPTRCAST(pobj)); /* ID */
		if (pobj == 0) return true;
		if (ok)
		{
			if (addOLEntry(pobj, value)) // object not already written
			{
				pobj->getClass()->notifyObjSerialization(pobj, true);
				// write full class path of objects class
				ok = writeString(fp, pobj->getClass()->getClassPath().c_str());
				BppVector* mdata = pobj->getMemberData();
				// write non static member data
				BppUInt n = mdata->size() - 1;
				// write number of entries
				ok = ok && writeBytes(fp, &n, sizeof(BppUInt));
				// write member data entries, classes are written as references
				for (BppUInt i = 1; ok && (i <= n); i++)
					ok = writeValue(fp, (*mdata)[i], true);
			}
			else {
				// if already written, simply write EmptyTag instead of class path.
				ok = writeString(fp, EmptyTag);
            }
		}
	}
	break;
	case BVT_BUFFER:
	case BVT_BUFFERREF:
	{
		BppBuffer* pbuf = value;

		ok = writePtr(fp, INTPTRCAST(pbuf)); /* ID */
		if (pbuf == 0) return true;
		if (ok)
		{
			if (addOLEntry(pbuf, value))
			{
				BppUInt size = pbuf->size();
				ok = writeBytes(fp, &size, sizeof(BppUInt));
				if (ok) ok = writeBytes(fp, pbuf->data(), size);
			}
			else
			{
				BppInt endmark = -1;
				ok = writeBytes(fp, &endmark, sizeof(BppInt));
			}
		}
	}
	break;
	case BVT_CHARBUFFER:
	case BVT_CHARBUFFERREF:
	{
		BppCharBuffer* pbuf = value;

		ok = writePtr(fp, INTPTRCAST(pbuf)); /* ID */
		if (pbuf == 0) return true;
		if (ok)
		{
			if (addOLEntry(pbuf, value))
			{
				BppUInt size = pbuf->size();
				ok = writeBytes(fp, &size, sizeof(BppUInt));
				BppByte chsize = sizeof(BppChar);
				if (ok) ok = writeBytes(fp, &chsize, 1); // write character size
				if (ok) ok = writeBytes(fp, pbuf->data(), size * chsize);
			}
			else
			{
				BppInt endmark = -1;
				ok = writeBytes(fp, &endmark, sizeof(BppInt));
			}
		}
	}
	break;
	case BVT_STRING:
	case BVT_STRINGREF:
	{
		BppString* pbuf = value;

		ok = writePtr(fp, INTPTRCAST(pbuf)); /* ID */
		if (pbuf == 0) return true;
		if (ok)
		{
			if (addOLEntry(pbuf, value))
			{
				ok = writeString(fp, pbuf->c_str());
			}
			else
			{
				BppByte endmark = 0xFF;
				ok = writeBytes(fp, &endmark, sizeof(BppByte));
			}
		}
	}
	break;
	case BVT_BYTECODE:
	case BVT_BYTECODEREF:
	case BVT_LOCALVARINFO:
	case BVT_VECTOR:
	case BVT_VECTORREF:
	case BVT_STRONGOBJECT:
	{
		BppVector* pvec = vtype == BVT_STRONGOBJECT ? value.getStrongObjDesc() : (BppVector*)value;

		ok = writePtr(fp, INTPTRCAST(pvec)); /* ID */
		if (pvec == 0) return true;
		if (ok)
		{
			if (addOLEntry(pvec, value))
			{
				BppUInt size = pvec->size();
				ok = writeBytes(fp, &size, sizeof(BppUInt));
				for (BppUInt i = 0; ok && (i < size); i++)
				{
					if (i == IDX_FUNCSTATE && (vtype == BVT_BYTECODE || vtype == BVT_BYTECODEREF)) {
						BppValue dv;
						ok = writeValue(fp, dv, true);
					}
					else
					   ok = writeValue(fp, (*pvec)[i], true);
				}
			}
			else
			{
				BppInt endmark = -1;
				ok = writeBytes(fp, &endmark, sizeof(BppInt));
			}
		}
	}
	break;
	case BVT_DICTIONARY:
	case BVT_DICTIONARYREF:
	{
		BppDictionary* pdict = value;
		ok = writePtr(fp, INTPTRCAST(pdict)); /* ID */
		if (pdict == 0)
			return true;
		if (ok)
		{
			if (addOLEntry(pdict, value))
			{
				ok = writeDictionary(fp, *pdict);
			}
			else
			{
				BppByte endmark = 0xFF;
				ok = writeBytes(fp, &endmark, sizeof(BppByte));
			}
		}
	}
	break;
	case BVT_CODE:
	{
		BPPUSERFUNC p = value;
		//if (!p && !value.isNull())
		//	return false;
		ok = writePtr(fp, INTPTRCAST(p));
	}
	break;
	case BVT_FILE:
	{
		FILE* p = value;
		//if (!p) 
		//	return false;
		ok = writePtr(fp, INTPTRCAST(p));
	}
	break;
	case BVT_VAR: // reference to variable
	{
		BppDictionaryEntry* pvar = value;
		if (pvar == 0) 
			return false;
		ok = writePtr(fp, INTPTRCAST(pvar)); /* ID */
	}
	break;
	case BVT_INT:
	{
		BppInt v = value;
		ok = writeBytes(fp, &v, sizeof(BppInt));
	}
	break;
	case BVT_UINT:
	case BVT_CHAR8:
	case BVT_CHAR32:
	{
		BppUInt v = value;
		ok = writeBytes(fp, &v, sizeof(BppUInt));
	}
	break;
	case BVT_SHORT:
	{
		BppShort v = value;
		ok = writeBytes(fp, &v, sizeof(BppShort));
	}
	break;
	case BVT_USHORT:
	case BVT_CHAR16:
	{
		BppUShort v = value;
		ok = writeBytes(fp, &v, sizeof(BppUShort));
	}
	break;
	case BVT_WCHAR:
	{
		static const BppUInt s = sizeof(wchar_t);
		ok = writeBytes(fp, &s, 1);
		if (ok) {
			BppUInt v = value;
			ok = writeBytes(fp, &v, s);
		}
	}
	break;
	case BVT_SBYTE:
	case BVT_CHAR:
	{
		BppSByte v = value;
		ok = writeBytes(fp, &v, sizeof(BppSByte));
	}
	break;
	case BVT_BYTE:
	case BVT_BOOL:
	case BVT_UCHAR:
	{
		BppByte v = value;
		ok = writeBytes(fp, &v, sizeof(BppByte));
	}
	break;
	case BVT_LONG:
	{
		BppLong v = value;
		ok = writeBytes(fp, &v, sizeof(BppLong));
	}
	break;
	case BVT_ULONG:
	{
		BppULong v = value;
		ok = writeBytes(fp, &v, sizeof(BppULong));
	}
	break;
	case BVT_INTPTR:
	case BVT_UINTPTR:
	{
		BppIntPtr v = value;
		ok = writePtr(fp, v); /* ID */
	}
	break;
	case BVT_FLOAT:
	{
		float v = value;
		ok = writeBytes(fp, &v, sizeof(float));
	}
	break;
	case BVT_DOUBLE:
	{
		double v = value;
		ok = writeBytes(fp, &v, sizeof(double));
	}
	break;
	default: 
		break;
	}
	return ok;
}


// writes
// - byte: csize (size of character)
// - byte[(len+1) * csize]: string content
bool BppDictionaryIO::writeString(FILE* fp, const BppChar* str)
{
	// write one byte containing character size of string
	BppByte csize = sizeof(BppChar);
	fwrite(&csize, 1, 1, fp);
	if (str == 0)
	{
		// null makes an empty string
		BppChar c = 0;
		return (fwrite(&c, csize, 1, fp) == 1);
	}
	// write string including null character on end
	size_t len = _tcslen(str) + 1;
	return (fwrite(str, csize, len, fp) == len);
}

bool BppDictionaryIO::writePtr(FILE* fp, BppIntPtr ptr)
{
	BppByte psize = sizeof(BppIntPtr);
	fwrite(&psize, 1, 1, fp);
	return fwrite(&ptr, psize, 1, fp) == 1;
}

// writes
// - byte[cnt]: cnt bytes of data
bool BppDictionaryIO::writeBytes(FILE* fp, const void* src, BppUInt cnt)
{
	return (cnt == 0 || fwrite(src, cnt, 1, fp) == 1);
}

bool BppDictionaryIO::readModule(const BppString& filename)
{

	FILE* fp = _tfopen(filename.c_str(), _T("rb"));
	if (fp == 0)
	{
		_print(DIO_OPEN_READ_FAILED_FMT, filename.c_str());
		return false;
	}

	bool result = readModule(fp);
	fclose(fp);
	return result;
}

bool BppDictionaryIO::readModule(FILE* fp)
{

	if (!fp)
	{
		return false;
	}
	BppDictionary tmpModules;
	bool result = readDictionary(fp, tmpModules, true);
#ifndef NO_DYNLIB
	for (BppDictionaryEntry* pe = tmpModules.getFirst(); pe != 0; pe = pe->getNext())
	{
		if (!_vm.loadNativeModule(pe->getKey()->c_str()))
		{
			result = false;
			break;
		}
	}
#endif
	if (result)
	{
		result = readDictionary(fp, _classes, true);
		if (result) {
			for (BppDictionaryEntry* p = _classes.getFirst(); p; p = p->getNext())
			{
				BppClass* pc = p->getValue();
				if (pc) {
					pc->buildDefaultValues();
					pc->computeDataSize();
				}
			}
		}
	}
	if (result) {
		result = readDictionary(fp, _symbols, true);
	}
	if (result)
	{
		BppDefaultFunctions::registerSysVariables(&_vm);
		BppDefaultFunctions::registerFiles(&_vm);
	}
	dissolveObjRefs();
	clearObjectLists();
	return result;
}

bool BppDictionaryIO::readSingleValue(FILE* fp, BppValue& value)
{
	clearObjectLists();
	bool ok = (fp != 0);
	if (ok)
	{
		ok = readValue(fp, value);
		dissolveObjRefs();
		clearObjectLists();
	}
	return ok;
}

bool BppDictionaryIO::readDictionary(FILE* fp, BppDictionary& dict, bool readSortedFlag)
{
	if (readSortedFlag)
	{
		BppByte b;
		bool ok = readBytes(fp, &b, sizeof(BppByte));
		if (!ok || b > 1)
		{
#ifdef _DEBUG
			fpos_t pos;
			fgetpos(fp, &pos);
			_print(DIO_READ_DICT_SORTFLAG_FAILED_FMT, (int)b, (unsigned)pos);
#endif
			return false;
		}
		dict._sorted = (b == 1);
	}
	while (readDictEntry(fp, dict));
	return true;
}

bool BppDictionaryIO::readDictEntry(FILE* fp, BppDictionary& dict)
{
	bool ok = true;
	void* id = (void*)readPtr(fp, ok);/* ID */
	if (!ok || id == (void*)0) return false;
	BppString* key = readString(fp);
	if ((key == 0) /*|| (key->length() == 0)*/)
		return false;
	BppValue kv = key;
	bool ignored = (*key)[0] == EmptyChar;
	if (ignored) key->erase(0, 1);
	BppDictionaryEntry* entry = dict[key];
	BppSymbolType stype = ST_NONE;
	if (!readBytes(fp, &stype, 1))
		return false;
	entry->setSymbolType(stype);
	if (!ignored) {
		if (!readValue(fp, entry->_value))
			return false;
	}
	addOLEntry(id, entry);

	return true;
}

#define MKNUMORCHARVAL(NUMTYPE) if (vtype == NUMTYPE) value = BppValue(v); else value = BppValue::makeChar(v, vtype)

bool BppDictionaryIO::readValue(FILE* fp, BppValue& value, bool readClassref)
{
	unsigned char bytebuf;
	BppValueType vtype;
	if (!readBytes(fp, &bytebuf, 1))
		return false;
	vtype = (BppValueType)bytebuf;
	//BppValueType mvtype = maskWeakRef(vtype);
	if (!readBytes(fp, &bytebuf, 1))
		return false;
	bool ok = true;
	switch (vtype)
	{
	case BVT_CLASS:
	case BVT_CLASSREF:
	{
		BppString* className = readString(fp);
		if (className == 0)
			return false;
		BppValue vclass = className;
		//className->addRef();

		// baseName should be full path of base class, empty string or EmptyTag (if class is ignored)
		BppString* baseName = readString(fp);
		if (baseName == 0)
		{
			//className->release();
			return false;
		}
		BppValue vbase = baseName;
		//baseName->addRef();
		if ((!readClassref) && (_tcscmp(baseName->c_str(), EmptyTag) != 0))
		{
			BppClass* base = 0;
			BppClass* owner = 0;
			if (baseName->length() > 0)
			{
				// search for base class in class table
				BppDictionaryEntry* entry = _classes.find(*baseName);
				if (entry != 0)
					base = entry->getValue();
				else {
					// if base class not already exists, create new empty class for base and insert it into table
					const BppChar* pn = _tcsrchr(baseName->c_str(), _T(':'));
					base = new BppClass(_vm, pn ? new BppString(pn+1) : baseName);					
					_classes[*baseName] = base;
				}
			}
			//baseName->release();

			BppString* ownerName = readString(fp);
			if (ownerName == 0)
			{
				//className->release();
				return false;
			}
			BppValue vowner = ownerName;
			if (ownerName->length() > 0)
			{
				// search for owner class name in class table
				BppDictionaryEntry* entry = _classes.find(*ownerName);
				if (entry != 0)
					owner = entry->getValue();
				else {
					// if owner class not already exists, create new empty class for owner and insert it into table
					const BppChar* pn = _tcsrchr(ownerName->c_str(), _T(':'));
					owner = new BppClass(_vm, pn ? new BppString(pn+1) : ownerName);
					_classes[*ownerName] = owner;
				}
			}
			//ownerName->release();

			BppClass* newClass = 0;
			// search for class in class table
			BppDictionaryEntry* centry = _classes.find(*className);
			if (centry) {
				newClass = centry->getValue();
				if (newClass) {
					// class object already exists
					BppClass* curbase = newClass->getBaseClass();
					if (base && (!curbase)) newClass->setBaseClass(base);
					else if (base != curbase) {
#ifdef _DEBUG
						_print(DIO_ILLEGAL_BASE_REDEFINITION_FMT, className->c_str());
#endif
						//className->release();
						return false;
					}
					BppClass* curOwner = newClass->getOwnerClass();
					if (owner && (!curOwner)) newClass->setOwnerClass(base);
					else if (owner != curOwner) {
#ifdef _DEBUG
						_print(DIO_ILLEGAL_OWNER_REDEFINITION_FMT, className->c_str());
#endif
						//className->release();
						return false;
					}
				}
			}
			if (!newClass) {
				newClass = new BppClass(_vm, className, base, owner);
				centry = _classes[className];
				centry->setValue(newClass);
			}
			value = BppValue(newClass, (bool)(vtype == BVT_CLASSREF));
			if (centry) centry->setSymbolType(ST_CLASS);
			if (!(readDictionary(fp, newClass->dataMembers(), true) &&
				readDictionary(fp, newClass->funcMembers(), true))) {
				//className->release();
				return false;
			}
			BppInt sz = 0;
			if (!readBytes(fp, &sz, sizeof(BppInt))) {
				//className->release();
				return false;
			}
#ifdef _DEBUG
			if (sz > 0xff) {
				fpos_t pos;
				fgetpos(fp, &pos);

				_print(_T("Class too big:%s %d pos: %lld\n"), className->c_str(),sz,pos);
				//className->release();
				return false;
            }
#endif
			newClass->_size = sz;
			//className->release();

		}
		else
		{
			//baseName->release();
			BppDictionaryEntry* entry = _classes.find(*className);
			if ((!entry) || (!(BppClass*)entry->getValue())) {
				_classes[*className] = new BppClass(_vm, className);
				entry = _classes.find(*className);
				entry->setSymbolType(ST_CLASS);
			}
			ok = (entry != 0);
			if (ok)
				value = BppValue((BppClass*)entry->getValue(), (bool)(vtype == BVT_CLASSREF));
			//className->release();
		}
	}
	break;
	case BVT_OBJECT:
	case BVT_OBJECTREF:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok) break;
		if (id == 0)
		{
			value = BppValue((BppRefObject*)0, vtype);
			break;
		}
		BppString* className = readString(fp); /* classname */
		if (className == 0)
			return false;
		BppValue vclass = className;
		//className->addRef();
		if ((className->length() > 0) && ((*className)[0] != EmptyChar))
		{
			BppDictionaryEntry* entry = _classes.find(*className);
			BppClass* pclass = (entry != 0) ? (BppClass*)entry->getValue() : (BppClass*)0;
			if (!pclass) {
				_classes[*className] = new BppClass(_vm, className);
				entry = _classes.find(*className);
				if (entry) {
					entry->setSymbolType(ST_CLASS);
					pclass = (BppClass*)entry->getValue();
				}
			}
			if (pclass != 0)
			{
				BppObject* pobj = new BppObject(pclass);
				value = BppValue(pobj, (bool)(vtype == BVT_OBJECTREF));
				addOLEntry(id, value);
				BppVector* mdata = pobj->getMemberData();
				BppUInt n = mdata->size();
				BppUInt nr = 0;

				ok = ok && readBytes(fp, &nr, sizeof(BppUInt));
				if (ok && nr + 1 > n) {
					mdata->insert(BPP_NNPOS, BppValue(), (nr + 1) - n);
					n = nr + 1;
				}
				(*mdata)[0] = pclass;
				for (BppUInt i = 1; ok && (i < n) && (i<=nr); i++)
					ok = readValue(fp, (*mdata)[i], true);
				if (ok) pclass->notifyObjSerialization(pobj, false);
			}
			else {
				ok = false;
			}
		}
		else
		{
			value = BppValue((BppRefObject*)0, vtype);
			addORefEntry(id, &value);
		}
		//className->release();
	}
	break;
	case BVT_BUFFER:
	case BVT_BUFFERREF:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok) break;
		if (!id)
		{
			value = BppValue((BppRefObject*)0, vtype);
			break;
		}
		BppUInt size = 0;
		ok = readBytes(fp, &size, sizeof(BppUInt));
		if (!ok) break;
		if (((BppInt)size) != -1) // no end mark
		{
			BppBuffer* pbuf = new BppBuffer(size);
			value = BppValue(pbuf, (bool)(vtype == BVT_BUFFERREF));
			addOLEntry(id, value);
			pbuf->_size = size;
			ok = readBytes(fp, pbuf->_data, size);
		}
		else
		{
			value = BppValue((BppRefObject*)0, vtype);
			addORefEntry(id, &value);
		}
	}
	break;
	case BVT_CHARBUFFER:
	case BVT_CHARBUFFERREF:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok)
			break;
		if (!id)
		{
			value = BppValue((BppRefObject*)0, vtype);
			break;
		}
		BppUInt size = 0;
		ok = readBytes(fp, &size, sizeof(BppUInt));
		if (!ok)
			break;
		if (((BppInt)size) != -1) // no end mark
		{
			BppByte chsize = 0;
			ok = readBytes(fp, &chsize, sizeof(BppByte));
			if (!ok)
				break;
			BppCharBuffer* pbuf = new BppCharBuffer(size);
			value = BppValue(pbuf, (bool)(vtype == BVT_CHARBUFFERREF));
			addOLEntry(id, value);
			BppUInt cbuf = 0;
			for (BppUInt i = 0; ok && (i < size); i++)
			{
				ok = readBytes(fp, &cbuf, chsize);
				pbuf->push_back(cbufToBppChar(cbuf,chsize));
			}
		}
		else
		{
			value = BppValue((BppRefObject*)0, vtype);
			addORefEntry(id, &value);
		}
	}
	break;
	case BVT_STRING:
	case BVT_STRINGREF:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok)
			break;
		if (!id)
		{
			value = BppValue((BppRefObject*)0, vtype);
			break;
		}
		BppString* pstr = readString(fp);
		value = BppValue(pstr, (bool)(vtype == BVT_STRINGREF));
		if (pstr != 0)
			addOLEntry(id, value);
		else
			addORefEntry(id, &value);
	}
	break;
	case BVT_BYTECODE:
	case BVT_BYTECODEREF:
	case BVT_LOCALVARINFO:
	case BVT_VECTOR:
	case BVT_VECTORREF:
	case BVT_STRONGOBJECT:
	{
		bool isref = (vtype >= _BVTREFMIN);
		bool isstrongobj = (vtype == BVT_STRONGOBJECT);
		bool iscode = ((vtype == BVT_BYTECODE) || (vtype == BVT_BYTECODEREF) || (vtype == BVT_LOCALVARINFO));
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!id)
		{
			value = BppValue((BppRefObject*)0, vtype);
			break;
		}
		if (!ok)
			break;
		BppUInt size = 0;
		ok = readBytes(fp, &size, sizeof(BppUInt));
		if (!ok)
			break;
		if (((BppInt)size) != -1) // no end mark
		{
			BppVector* pvec = new BppVector(size, true);
			if (isstrongobj)
			{
				value = BppValue(vtype, BVF_STRONGTYPED, false);
				value.assignRef(pvec);
			}
			else
				value = BppValue(pvec, iscode, isref);
			addOLEntry(id, value);
			for (BppUInt i = 0; ok && (i < size); i++)
			{
				ok = readValue(fp, (*pvec)[i], true);
			}
		}
		else
		{
			value = BppValue((BppRefObject*)0, vtype);
			addORefEntry(id, &value);
		}
	}
	break;
	case BVT_DICTIONARY:
	case BVT_DICTIONARYREF:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok)
			break;
		if (id == 0)
		{
			value = BppValue((BppRefObject*)0, vtype);
			break;
		}
		BppByte sortflag = 0;
		ok = readBytes(fp, &sortflag, sizeof(BppByte));
		if (!ok)
			break;
		if (sortflag != 0xFF) // no end mark
		{
			BppDictionary* pdict = new BppDictionary((bool)(sortflag == 1));
			value = BppValue(pdict, (bool)(vtype == BVT_DICTIONARYREF));
			addOLEntry(id, value);
			ok = readDictionary(fp, *pdict, false);
		}
		else
		{
			value = BppValue((BppRefObject*)0, vtype);
			addORefEntry(id, &value);
		}
	}
	break;
	case BVT_CODE:
	case BVT_FILE:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok) 
			break;
		value = BppValue(vtype,BVF_NONE,false);
		if (id) addORefEntry(id, &value);
	}
	break;
	case BVT_VAR:
	{
		void* id = (void*)readPtr(fp, ok);/* ID */
		if (!ok) 
			break;
		value = BppValue((BppDictionaryEntry*)0);
		addORefEntry(id, &value);
	}
	break;
	case BVT_INT:
	{
		BppInt v = 0;
		ok = readBytes(fp, &v, sizeof(BppInt));
		if (ok) value = BppValue(v);
	}
	break;
	case BVT_UINT:
	case BVT_CHAR8:
	case BVT_CHAR32:
	{
		BppUInt v = 0;
		ok = readBytes(fp, &v, sizeof(BppUInt));
		if (ok) MKNUMORCHARVAL(BVT_UINT);
	}
	break;
	case BVT_SHORT:
	{
		BppShort v = 0;
		ok = readBytes(fp, &v, sizeof(BppShort));
		if (ok) value = BppValue(v);
	}
	break;
	case BVT_USHORT:
	case BVT_CHAR16:
	{
		BppUShort v = 0;
		ok = readBytes(fp, &v, sizeof(BppUShort));
		if (ok) MKNUMORCHARVAL(BVT_USHORT);
	}
	break;
	case BVT_SBYTE:
	case BVT_CHAR:
	{
		BppSByte v = 0;
		ok = readBytes(fp, &v, sizeof(BppSByte));
		if (ok) MKNUMORCHARVAL(BVT_SBYTE);
	}
	break;
	case BVT_BYTE:
	case BVT_UCHAR:
	{
		BppByte v = 0;
		ok = readBytes(fp, &v, sizeof(BppByte));
		if (ok) MKNUMORCHARVAL(BVT_BYTE);
	}
	break;
	case BVT_BOOL:
	{
		BppByte v = 0;
		ok = readBytes(fp, &v, sizeof(BppByte));
		if (ok) value = BppValue::makeBool(v != 0);
	}
	break;
	case BVT_WCHAR:
	{
		BppByte cs = 0;
		ok = readBytes(fp, &cs, sizeof(BppByte));
		if (ok) {
			BppUInt buf = 0;
			ok = readBytes(fp, &buf, cs);
			if (sizeof(wchar_t) != cs) {
				buf = cbufToWchar(buf, cs);
			}
			if (ok) value = BppValue::makeChar(buf, vtype);
		}
	}
	break;
	case BVT_LONG:
	{
		BppLong v = 0;
		ok = readBytes(fp, &v, sizeof(BppLong));
		if (ok) value = BppValue(v);
	}
	break;
	case BVT_ULONG:
	{
		BppULong v = 0L;
		ok = readBytes(fp, &v, sizeof(BppULong));
		if (ok) value = BppValue(v);
	}
	break;
	case BVT_INTPTR:
	{
		BppIntPtr v = readPtr(fp, ok);
		if (ok) value = BppValue(v, true);
	}
	break;
	case BVT_UINTPTR:
	{
		BppUIntPtr v = UINTPTRCAST(readPtr(fp, ok));
		if (ok) {
			value = BppValue(v, true);
		}
	}
	break;
	case BVT_FLOAT:
	{
		float v = 0;
		ok = readBytes(fp, &v, sizeof(float));
		if (ok)
			value = BppValue(v);
	}
	break;
	case BVT_DOUBLE:
	{
		double v = 0;
		ok = readBytes(fp, &v, sizeof(double));
		if (ok)
			value = BppValue(v);
	}
	break;
	default:
		value.dispose();
	}
	value.setFlags(bytebuf);
	return ok;
}


BppString* BppDictionaryIO::readString(FILE* fp)
{
	BppByte csize = 0;
	if (fread(&csize, 1, 1, fp) != 1)
		return 0;
	if (csize == 0xFF) return 0;  // flag for reference only
	if (csize == 0 || csize > 4)
	{
#ifdef _DEBUG
		fpos_t pos;
		fgetpos(fp, &pos);
		_print(DIO_ERROR_READ_STRING_FMT, (unsigned)pos);
#endif
		return 0;
	}
	BppCharBuffer buf(32); // default buffer size
	BppUInt cbuf = 0;
	while (!feof(fp))
	{
		fread(&cbuf, csize, 1, fp);
		buf.push_back(cbufToBppChar(cbuf,csize));
		if (cbuf == 0)
			break;
	}
	return new BppString((const BppChar*)buf.data());
}

BppIntPtr BppDictionaryIO::readPtr(FILE* fp, bool& ok)
{
	union {
		BppIntPtr p;
		char s[8];
	} buf;
	buf.p = 0;
	ok = fread(&buf, 1, 1, fp) == 1;
	if (ok) {
		size_t siz = (size_t)buf.s[0];
		ok = (siz == 4 || siz == 8);
		if (ok) {
			ok = fread(&buf, siz, 1, fp) == 1;
			if (ok) return buf.p;
		}
#ifdef _DEBUG
		fpos_t pos;
		fgetpos(fp, &pos);
		_print(DIO_READ_PTR_FAILED_FMT, (unsigned)pos);
#endif
	}
	return 0;
}


bool BppDictionaryIO::readBytes(FILE* fp, void* pbuf, BppUInt cnt)
{
	return (cnt == 0 || fread(pbuf, cnt, 1, fp) == 1);
}



void BppDictionaryIO::dissolveObjRefs()
{
	for (ObjListEntry* po = _objList; po != 0; po = po->next)
	{
		for (ObjListEntry* pr = _objRefList; pr; pr = pr->next)
		{
			if (pr->ID == po->ID)
			{
				if (pr->pval)
					pr->pval->assignRef(po->val.getRef());
			}
		}
	}
	clearObjectLists();
}


