/*
 * Created by SharpDevelop.
 * User: erik
 * Date: 01.10.2008
 * Time: 08:44
 * 
 */

using System;
using System.Windows.Forms;
using bp2builder.InterOp;

namespace bp2builder
{
	/// <summary>
	/// Base class of Bp2Builder documents.
	/// </summary>
	/// <remarks>
	/// This class provides an abstract document for use in Bp2Builder. 
	/// A Bp2Builder document usually referres to a disk file where
	/// document data are stored, e.g. a plain text file for source documents.
	/// Furthermore a document may be shown in a document list in GUI. For support that this class
	/// implements some helper functions and properties.
	/// </remarks>
	public abstract class Bp2Doc
	{

		/// <summary>
		/// Enumeration of possible document types
		/// </summary>
		public enum DocTypes
		{
			/// <summary>
			/// source text document
			/// </summary>
			Source,
			/// <summary>
			/// reference to precompiled module
			/// </summary>
			Module,
			/// <summary>
			/// B++ project file
			/// </summary>
			Project,
			/// <summary>
			/// any other document
			/// </summary>
			Any
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <remarks>
		/// Constructor initializes <c>_docType</c>c> member to <c>DocTypes.Any</c>
		/// </remarks>
		public Bp2Doc()
		{
			_docType = DocTypes.Any;
		}

		/// <summary>
		/// Get name of document file without directory path.
		/// </summary>
		public string FileName
		{
			get
			{
				if ((_location == null) || (_location.Length == 0))
				{
					return "";
				}
				int slashIdx = _location.LastIndexOf('/');
				if ((slashIdx >= 0) && (slashIdx < _location.Length-1))
					return _location.Substring(slashIdx+1);
				return _location;
			}
		}

		/// <summary>
		/// Get directory path of document file.
		/// </summary>
		public string FileDir
		{
			get
			{
				if ((_location == null) || (_location.Length == 0))
				{
					return "";
				}
				int slashIdx = _location.LastIndexOf('/');
				if ((slashIdx >= 0) && (slashIdx < _location.Length-1))
					return _location.Substring(0,slashIdx);
				return "";
			}
		}

		/// <summary>
		/// Get extension of document file name.
		/// </summary>
		public string Extension
		{
			get
			{
				string fname = FileName;
				int pointIdx = fname.LastIndexOf('.');
				if ((pointIdx < 0) || (pointIdx == fname.Length-1)) return "";
				return fname.Substring(pointIdx+1);
			}
		}

		/// <summary>
		/// Get plain name of document file (without extension).
		/// </summary>
		public string ShortName
		{
			get
			{
				string fname = FileName;
				int pointIdx = fname.LastIndexOf('.');
				if (pointIdx < 0) return fname;
				return fname.Substring(0,pointIdx);
			}
		}

		/// <summary>
		/// Complete location of document file (including directory path and extension).
		/// </summary>
		/// <value>complete file path</value>
		public string Location
		{
			get { return (null != _location) ? _location : ""; }
			set 
			{ 
				if (_location != value)
					Changed = true;
				_location = value; 
				if (null != _location)
					_location = _location.Replace('\\','/');
			}
		}

		/// <summary>
		/// Check changed state of document.
		/// </summary>
		public virtual bool Changed
		{
			get 
			{ 
				return _changed; 
			}

			set 
			{ 
				_changed = value; 
			}
		}

		/// <summary>
		/// Convert object to string.
		/// </summary>
		/// <returns>file name of document (without directory path)</returns>
		/// <remarks>
		/// This implementation allows direct use of document e.g.
		/// in a list box in GUI.
		/// </remarks>
		public override string ToString()
		{
			return FileName;
		}

		/// <summary>
		/// Equality check.
		/// </summary>
		/// <param name="obj">object (of Bp2Doc type) to compare with</param>
		/// <returns>true on equality</returns>
		/// <remarks>
		/// 
		/// </remarks>
		public override bool Equals(object obj)
		{
			Bp2Doc cmp = obj as Bp2Doc;
			if (null == cmp) return false;
			return (cmp.Location.ToLower() == Location.ToLower());
		}

		public override int GetHashCode()
		{
			return Location.ToLower().GetHashCode();
		}

		public DocTypes DocType
		{
			get { return _docType; }
			set { _docType = value; }
		}

		Bp2Printer ErrorPrinter
		{
			get { return _errorPrinter; }
			set { _errorPrinter = value; }
		}

		public virtual bool Rename(string newFileName)
		{
			if ((newFileName == null) || (newFileName == "")) return false;
			string newFullPath = FileDir;
			if (newFullPath != "")
				newFullPath += "/";
			newFileName = newFileName.Replace('\\','/');
			if (newFileName.StartsWith("/"))
				newFullPath += newFileName.Substring(1);
			else
				newFullPath += newFileName;
			if (System.IO.File.Exists(newFullPath)) return false;
			Location = newFullPath;
			return true;
		}

		public void ReportError(string msg)
		{
			if (null != _errorPrinter)
			{
				_errorPrinter.DoPrint(msg);
			}
		}

		public abstract bool Save();
		public abstract bool Load();

		string _location = "";
		DocTypes _docType;
		Bp2Printer _errorPrinter;
		bool _changed = false;

	}
}
